###########################################################################
#
# Copyright(c) 2011-2014 Qlogic Corporation, all rights reserved
# Copyright(c) 2014-2015 Qlogic Corporation, all rights reserved
# Proprietary and Confidential Information.
#
# Module Description:
# Display real-time Storm CPU usage information:
# Cpu usage counters, BRB drops, pause packets.
#                                    
###########################################################################

# profiler main function
proc profile {args} {
	global rdfFile
	global g_iterationTime
	global g_pciRegs
	global g_regsText
	
	#get RDF file name according to DID
	get_rdf_file

	# Parse the RDF file
	parse_rdf -clr -f $rdfFile
	
	initGlobals
	
	set parsedArgs [parseArgs $args];
	
	set frequency [lindex $parsedArgs 0]
	set iterations [lindex $parsedArgs 1]
	set isFile [lindex $parsedArgs 2]
	set logname [lindex $parsedArgs 3]
	set isHist [lindex $parsedArgs 4]
	set histname [lindex $parsedArgs 5]
	set samples [lindex $parsedArgs 6]
	
    if {$isFile} {
        set fileId [open $logname "w"]
    }
	
	if {$isHist == 0} {
		puts "\nDisplaying: Storm %busy (%store halt,%fin halted,%partial fin halted); BRB drops count; RX pause packets count; PCI header credits average; PCI data credits average"
	} else {
		initHists
	}

    for {set i 0} {$i < $iterations} {incr i} {	

		if {$isHist == 1} {
			hist $samples	
			puts -nonewline "\rSampling registers, iteration number is: [expr $i + 1]"
			flush stdout
			continue
		}
		
	    readCpuData; #to clear the registers
        set ticks1 [clock clicks -milliseconds]
		# measure cycles in $g_iterationTime milliseconds
		after $g_iterationTime
        set cpu [readCpuData]
        set ticks2 [clock clicks -milliseconds]
        set timeDelta [expr $ticks2 - $ticks1]
        
		# calculate busy data
        set line "\t    "
        foreach {storm data} $cpu {
            set busy [calcUtilization [lindex $data 0] $timeDelta $frequency]
            set store_halt [calcUtilization [lindex $data 1] $timeDelta $frequency]
            set fin_halt [calcUtilization [lindex $data 2] $timeDelta $frequency]
            set part_halt [calcUtilization [lindex $data 3] $timeDelta $frequency]
            set line [format "%s%s %2.0f(%2.0f,%2.0f,%2.0f)| " $line  $storm  $busy  $store_halt  $fin_halt  $part_halt ]
        }
		
		#Read registers' differences from initial value
		set countedRegs [countRegs]
		
		set line [format "%s %s" $line $countedRegs]
		
		foreach reg $g_pciRegs {
			set line  [format "%s %s: %s" $line $g_regsText($reg) [format 0x%x [getAveOfReg $reg]]]
		}
		
        puts -nonewline "\r$line, $i"
        
		if {$isFile} {
			puts $fileId "$line $i \n"		
        }
        
        flush stdout
    }
	
    puts "\r\n"
   
    if {$isFile} {
		close $fileId
    }
	
	if {$isHist == 1} {
		printHistsToFile $histname
	}
}

# Help info on "profile" command
proc profile_help {} {
    set help_info     "First mode - display real time Storm CPU utilization\n"
	append help_info  "Second mode - sample registers and save a histogram to file\n"
    append help_info  "First mode usage: profile \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\]\n"
	append help_info  "Second mode usage: profile -h <filename> \[-forever | -i <iterations>\] \[-s <samples>\]\n"
    append help_info  "Notes: \n"
    append help_info  "\t-forever runs in a loop without stopping (will never exit to user prompt)\n\n"
    append help_info  "\t-freq <frequency> update actual chips frequency for script calculations \n\n"
    append help_info  "\t-f <fileName>  writes results into file (first mode only)\n\n"
    append help_info  "\t-h <fileName>  creates histograms and writes into file\n\n"
	append help_info  "\t-s <samples> specifies number of samples per iteration (default is 1000 samples)\n\n"
    append help_info  "\t-i <iterations> specifies number of iteration to display profiling data (default is 60 iteration)\n\n"
    append help_info  "Example: profile -h histOut.txt -i 10"
    add_help profile $help_info
}

#error proc after bad argument while parsing
proc argError {arg} {
	error "Invalid argument: $arg \nFirst mode usage: profile \[-forever | -i <iterations>\] \[-f <fileName>\] \[-freq <frequency>\]\nSecond mode usage: profile -h <filename> \[-forever | -i <iterations>\] \[-sample <samples>\]\n"
}

#parse arguments for running the script, according to usage.
proc parseArgs {arguments} {
    
	global g_chipID
	global g_chipsFrequency
	
	set frequency -1
	set isFile 0;	
	set iterations 60
	set samples 1000
    set input_pending 0
    set args_counter 0
	set logname ""
	set isHist 0
	set histname ""
	set isIteration 0
	set isSample 0
	set isfrequency 0
	
    foreach arg $arguments {
        if {$input_pending == 0} {
            
            switch -- $arg {
                "-freq"     {
							  if {[scan [lindex $arguments [expr $args_counter + 1]] "%d" frequency] != 1 || $isfrequency == 1 } {argError $arg}; #checking input is integer                       
							  set isfrequency 1
							  set input_pending 1; #skip next argument
                            }
                "-f"     {
                              if {$isFile == 1} {argError $arg}
							  set isFile 1
							  set logname [lindex $arguments [expr $args_counter + 1]]
                              set input_pending 1
                            }
                "-h"     {
							  if {$isHist == 1} {argError $arg}
                              set isHist 1
							  set histname [lindex $arguments [expr $args_counter + 1]]
                              set input_pending 1
                            }
				"-s"	{
				                if {[scan [lindex $arguments [expr $args_counter + 1]] "%d" samples] != 1 || $isSample == 1} {argError $arg}
								set isSample 1
								set input_pending 1
							}
				"-i" {
				                if {[scan [lindex $arguments [expr $args_counter + 1]] "%d" iterations] != 1 || $isIteration == 1} {argError $arg}
								set isIteration 1
								set input_pending 1
							}
                "-forever"  { set iterations 10000000 }
                default {
                    argError $arg
                }
            }
            
        } else {
			set input_pending 0
        }
               
        incr args_counter         
    }
	
	if {$input_pending == 1} {		
		argError [lindex $arguments [expr $args_counter - 1]]
    }
	
	#only one mode of operation
	if {[expr $isHist == 1 && $isFile == 1]} {
		argError "-f and also -h"
	}
	
	if {$frequency == -1} {		
		set frequency $g_chipsFrequency($g_chipID)
    }
	
	return [list $frequency $iterations $isFile $logname $isHist $histname $samples]
}

#Return a boolean, if it is E3b0 or not.
proc isE3b0 {} {

	global g_chipID
	
	if {($g_chipID == 5773) || ($g_chipID == 5774) || ($g_chipID == 5793) || ($g_chipID == 5794)} {
		return 1
	}
	
	return 0
}

#Return a boolean, if it is E1/E1.5 or not.
proc isE1orE15 {} {

	global g_chipID
	
	if {($g_chipID == 5710) || ($g_chipID == 5711) || ($g_chipID == 5712)} {
		return 1
	}
	
	return 0
}

#Return a list of storm and his registers values 
proc readCpuData {} {
    global g_storms 

    set result {}

    foreach {storm sem} [array get g_storms] {
        lappend result $storm
		
		set reg_name _REG_SEM_FAST_IDLE_CYCLES 
        set data [getRegVal $sem$reg_name]
		
		set reg_name _REG_SEM_FAST_STORE_HALT_CYCLES 
        lappend data [getRegVal $sem$reg_name]
		
		set reg_name _REG_SEM_FAST_FIN_HALT_CYCLES
        lappend data [getRegVal $sem$reg_name]
		
		set reg_name _REG_SEM_FAST_PARTIAL_HALT_CYCLES
        lappend data [getRegVal $sem$reg_name]
        
		lappend result $data
    }
    return $result
}

#Return percentage of argument 'cycles' from all cycles in argument 'millisec' interval  
proc calcUtilization  {cycles millisec frequency} {
    
	set sec [expr ((double($millisec))/1000.)]
    set val [expr 100.0 * ((double($cycles))/($sec * (double($frequency))))]
    
    return $val
}

#Get a register's value, while considering its offset and bitmask as set in cfg file. Different flow for address regs and name regs.
proc getRegVal {reg} {
	global g_offsetRegsArgs
	
	if { ![info exists g_offsetRegsArgs($reg)]} {
		set bitMask 0xFFFFFFFF
		set offset 0
	} else {
		set bitMask [lindex $g_offsetRegsArgs($reg) 0]
		set offset [lindex $g_offsetRegsArgs($reg) 1]
	}

	if {[string first "REG_ADDR" $reg] != -1} {
		set addr [lindex [split $reg "_"] 2]
		set val [expr [expr ([reg read $addr] & $bitMask)] >> $offset]	
	} else {
		set val [expr [expr ([reg_read -n $reg] & $bitMask)] >> $offset]
	}
	
	return $val
}

#Return a string for printing the registers we want to count
proc countRegs {} {
	global g_regsText
	global g_regsToCount
		
	set toPrint ""
	foreach {reg initVal} [array get g_regsToCount] {
		set regVal [getRegVal $reg];
		 if { ![info exists g_regsText($reg)]} { 
			set g_regsText($reg) "Register $reg has incremented by"
		 }
		append toPrint "$g_regsText($reg): [format "%u" [expr $regVal - $initVal]]; "		
	}

	return $toPrint
}

#Calculate a register's average.
proc getAveOfReg {reg} {
	global g_pciSamples
	global g_pciSamplesInterval
	
	set sum 0
	for {set i 0} {$i < $g_pciSamples} {incr i} {
		incr sum [getRegVal $reg]
		after $g_pciSamplesInterval
	}
	
	return [expr $sum / $g_pciSamples]
}

#Initialization of histogram data structure
proc initHists {} {
	global g_rxHistRegs
	global g_txHistRegs
	global g_histsInMem
	array unset g_histsInMem *
	
	set regsToHist [concat $g_rxHistRegs $g_txHistRegs]
	foreach reg $regsToHist {
			set g_histsInMem($reg) {}
	}
}

#add a value to the register's histogram.
proc addValToHist {val regKey} {
	global g_histsInMem
	
	array unset regHist *
	array set regHist $g_histsInMem($regKey)	
	set keysList [array names regHist]; #list of the values that has been sampled
	set index [lsearch $keysList $val]
	if {$index == -1} {
		set g_histsInMem($regKey) [concat $g_histsInMem($regKey) [list $val 1]]	
	} else {
		set index [expr $index * 2]; #the index in the histogram list
		
		#increment g_histsInMem($regKey)[index+1] {lreplace list first last ?element element ...?}
		set g_histsInMem($regKey) [lreplace $g_histsInMem($regKey) [expr $index + 1] [expr $index + 1]  [expr [lindex $g_histsInMem($regKey) [expr $index + 1]] + 1]]
	}
}

#Root histogram function
proc hist {Samples} {
	global g_rxHistRegs
	global g_txHistRegs
	
	set regsToHist [concat $g_rxHistRegs $g_txHistRegs]
	for {set i 0} {$i < $Samples} {incr i} {
		foreach reg $regsToHist {
			set val [getRegVal $reg]
			addValToHist $val $reg
		}
	}
}

#Go over each of the registers and print the histogram from the run. Divided to two sections - Rx and Tx registers.
proc printHistsToFile {outFile} {
	global g_txHistRegs
	global g_rxHistRegs
		
	set fileId [open $outFile "w"]
	
	puts $fileId "\n\n----------------------RX data--------------------"
	foreach reg $g_rxHistRegs {
		printSingelHist $fileId $reg	
	}
	
	puts $fileId "\n\n----------------------TX data--------------------"
	foreach reg $g_txHistRegs {
		printSingelHist $fileId $reg	
	}

    close $fileId
}

#prints to file a single register's histogram.
proc printSingelHist {fileId reg} {
	global g_histsInMem
	global g_numberOfHistBuckets
	
	puts $fileId "Histogram on the values of register: $reg"
	
	array unset histArray *
	array set histArray $g_histsInMem($reg)	
	set orderedValues [lsort -integer [array names histArray]]
	if {[llength $orderedValues] > $g_numberOfHistBuckets} {
		getBuckets $orderedValues $fileId $g_histsInMem($reg)
	} else {
		foreach value $orderedValues {
			puts $fileId "Number of samples of value [format 0x%x $value]: $histArray($value)"		
		}
	}
	puts $fileId "==============================================================================="
	
}

#Divide the register's histogram into buckets (num of buckets can be set in cfg file) and print to file.
proc getBuckets {valuesList fileId regHist} {
	global g_numberOfHistBuckets
	
	array unset histArray *
	array set histArray $regHist
	
	set minVal [lindex $valuesList 0]
	set maxVal [lindex $valuesList end]
	
	set bucketSize [expr [expr $maxVal - $minVal] / $g_numberOfHistBuckets]
	
	array unset buckets *
	array set buckets {}
	foreach value $valuesList {
		set bucketMin [expr (($value - $minVal) / ($bucketSize))*$bucketSize + $minVal]; #Remember, the integer division do round down. So we get the bucket of the value
			if { ![info exists buckets($bucketMin)]} { 
				set buckets($bucketMin) $histArray($value)
			} else {
				set buckets($bucketMin) [expr $buckets($bucketMin) + $histArray($value)]			
			}
	}
	
	for {set i 0} {$i < $g_numberOfHistBuckets} {incr i} {
		set bucketMin [expr [expr $bucketSize * $i] + $minVal]
		set bucketMax [expr $bucketMin + $bucketSize]
		set numOfVals 0
		if {[info exists buckets($bucketMin)]} {
			set numOfVals $buckets($bucketMin)
		}
		puts $fileId "Number of values greater than [format 0x%x $bucketMin] and smaller than [format 0x%x $bucketMax]: $numOfVals"
	}
}

# add help
profile_help