/*********************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved.
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without
 * the prior written consent of QLogic Corporation.
 *
 *
 *  Name: tcl_vpd.c
 *
 *  Description: This file handles the VPD information text file:
 *               1) Parses the VPD text file, creates a VPD image to be stored in NVM
 *               2) Dumps an existing VPD image into a VPD text file.
 *               3) Added support for VPD interactive commands:
 *                 vpd read, vpd write, vpd show, vpd delete
 ***********************************************************************************/

#include <stdio.h>
#ifndef USER_LINUX
    #if !defined(UEFI) && !defined(__SunOS) && !defined(FREEBSD)
        #include <io.h>
    #endif
#else
    #include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "tcl_al.h"
#include "tcl_vpd.h"
#include "tcl_util.h"
#ifdef EDEBUG
    #include "dbg_os_utils.h"
#else
    #include "dg_os_util.h"
    #include "debug.h"
#endif
#include "common_nvm.h"
#include "utils.h"
#include "tcl_al_common.h"
#include "tcl_al_cmd_handlers.h"
#ifdef UEFI
    #include <sys/unistd.h>
    #include "debug.h"
#endif

typedef enum opers {
    VPD_READ = 0, 
    VPD_SHOW,
    VPD_WRITE,
    VPD_DELETE
} vpd_op_t;


static void usage_vpd(void)
{
    printf("Usage: vpd read <field name> \n");
    printf("# Shows the requested VPD field value\n");
    printf("# Inputs: <field name: (PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME)\n");

    printf("Usage: vpd write <field name> <value>\n");
    printf("# Change the requested VPD field value\n");
    printf("# Inputs: <field name: (PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME)\n");

    printf("Usage: vpd show\n");
    printf("# Show all VPD fields\n");

    printf("Usage: vpd delete <field name>\n");
    printf("# Delete the requested VPD field value\n");
    printf("# Inputs: <field name: (PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME)\n");

}

static void vpd_read_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "vpd read <field name>");
}


static void vpd_write_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "vpd write <field name> <value>");
}

static void vpd_show_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "vpd show");
}

static void vpd_delete_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "vpd delete <field name>");
}


static int tcl_al_parse_vpd_line(char *line, 
                                 char *vpd_field_name_ptr, 
                                 char *vpd_field_type_ptr, 
                                 char *vpd_field_data_ptr);


static int verify_vpd_field_name(char *field_name)
{
    /* Verify the field name is allowed */
    strtoupr(field_name);
    if ((strcmp( field_name, "PRODUCT NAME") != 0) && 
        (strcmp( field_name, "PRODUCT_NAME") != 0) &&
        (strcmp( field_name, "PN") != 0) && 
        (strcmp( field_name, "EC") != 0) && 
        (strcmp( field_name, "MN") != 0) &&
        (strcmp( field_name, "SN") != 0) &&
        (strcmp( field_name, "YA") != 0) &&
        (strcmp( field_name, "FN") != 0) &&
        (strcmp( field_name, "CC") != 0) &&
        (strcmp( field_name, "FC") != 0) &&
        (strcmp( field_name, "YB") != 0) &&
        (strncmp( field_name, "V", 1) != 0)) {
        return TCL_ERROR;
    } else {
        return TCL_OK;
    }
}

/*
 *------------------------------------------------------------------------
 * Description:
 *     Get VPD information from VPD image in extended dir into
 *     vpd_table_t type.
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */ 
static int get_vpd_info_from_image( Tcl_Interp *interp, vpd_table_t *vpd_table_p) 
{

    u32_t   length, byte_dumped, offset, image_idx;
    u8_t    *buf=NULL, big_endian=1, img_type = IMG_TYPE_ST;
    u32_t   i=0, tag_idx=0, j=0;
    char    tag_name[VPD_ID_TAG_MAX_LEN+1];
    u8_t    tag_len, *ptr;
    u16_t   len_remaining, product_name_len;
    u32_t   *ptr32, val32;


    /* Get VPD image offset in NVM and length */
    if (common_nvm_get_image_in_extended_dir_info(NVM_IMAGE_NAME_VPD, &offset, &length, &image_idx) != COMMON_NVM_STATUS_SUCCESS) {
        Tcl_AppendResult(interp, "Get VPD image offset and length failed!", NULL);
        return TCL_ERROR;
    }

    /* Allocate memory for dumping into buffer */
    buf = (u8_t *)Tcl_Alloc((img_type == IMG_TYPE_ST) ? length :
                            // for ATMEL - for every 256 bytes add another 8 bytes.
                            (length + (length / 256 * 8)));

    if (buf == NULL) {
        Tcl_AppendResult(interp, "No more host memory.", NULL);
        return TCL_ERROR;
    }

    /* Dump the VPD image into buf */
    if (common_nvm_dump_nvm_into_buf(buf, big_endian, length, offset, &byte_dumped) != COMMON_NVM_STATUS_SUCCESS) {
        Tcl_AppendResult(interp, "Dump VPD image from NVM into buffer failed!", NULL);
        return TCL_ERROR;;
    }

    /* Swap data bytes */
    for (i=0; i < length-4; i=i+4) {
        ptr32 = (u32_t *) (buf+i);
        val32 = htonl(*ptr32);
        *ptr32 = val32;
    }

    /* VPD info starts 4 bytes (after image version) from buffer start*/
    ptr = ((u8_t *) (buf+4));

    if (*ptr != (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_STRING)) {
        Tcl_AppendResult (interp, "Failed to get VPD info from image: Expected ID TAG 0x%x is missing! (read 0x%x)\n", 
                          (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_STRING), *ptr);
        return TCL_ERROR;
    }
    ptr++;

    /* Get the product name length */
    cmn_os_memcpy(&product_name_len, ptr, 2);
    vpd_table_p->product_name_len = product_name_len;
    ptr += 2;

    /* Get the product name */
    strncpy(vpd_table_p->product_name, (char *) ptr, product_name_len); 
    vpd_table_p->product_name[product_name_len]='\0';
    ptr += product_name_len;

    /* Verify the VPD-R TAG exist */    
    if (*ptr != (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_VPD_R)) {
        Tcl_AppendResult (interp, "Failed to get VPD info from image: Expected VPD-R TAG 0x%x is missing! (read 0x%x)\n", 
                          (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_VPD_R), *ptr);
        return TCL_ERROR;
    }
    ptr ++;

    /* Read the remaining length */
    cmn_os_memcpy(&len_remaining, ptr, 2);
    ptr += 2;

    /* Reset the no_tags counter */
    vpd_table_p->no_tags=0;

    /* Read the VPD fields */
    while (len_remaining > 0) {
        /* Copy the TAG name (e.g. PN, EC, MN ..) */
        strncpy(tag_name, (char *)ptr, 2);
        tag_name[2]='\0';
        strncpy(vpd_table_p->vpd_tags[tag_idx].tag_name, tag_name, 2);
        vpd_table_p->vpd_tags[tag_idx].tag_name[2]='\0';

        /* Increment the data ptr by 2 (since TAG name is 2 bytes) */
        ptr += 2;

        /* Copy the TAG Length */
        tag_len = *ptr;
        ptr++;            

        if (strcmp(tag_name, "RV") != 0) {

            /* Copy TAG data and length to vpd_info*/
            if (strcmp(vpd_table_p->vpd_tags[tag_idx].tag_name, "YB") != 0) {

                /* Get the TAG data */
                strncpy(vpd_table_p->vpd_tags[tag_idx].tag_data, (char *) ptr, tag_len);
                vpd_table_p->vpd_tags[tag_idx].tag_data[tag_len]='\0';
                ptr += tag_len;

            } else {
				/* YB tag is binary */
                for (j=0; j < tag_len; j++) {
                    vpd_table_p->vpd_tags[tag_idx].tag_data[j] = *ptr;
                    ptr++;
                }
            }

            len_remaining -= (tag_len + 3);
            vpd_table_p->vpd_tags[tag_idx].tag_len = tag_len;
            vpd_table_p->no_tags++;
            tag_idx++;

        } else {
            /* If got to RV tag, exit */
            break;
        }
        //DbgMessage(NULL, FATAL, "tag_idx=%d tag_name=%s tag_data=%s\n", tag_idx, tag_name, tag_data);
    }

    if (buf != NULL) Tcl_Free((char*)buf);
    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * Description:
 *     Implements "vpd read" command.
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */ 
static int tcl_al_vpd_read_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]  /* argument strings */
    )
{
    Tcl_Obj             *result_ptr = NULL;
    media_vpd_t         vpd_block;
    vpd_info_t          vpd_info_struct;
    u32_t               vpd_check=1;
    char                *tag_name;
    char                err_str[80];
    u32_t               nvm_offset, i, j;
    common_nvm_status_t status;
    u8_t                is_vpd_image;
    vpd_table_t         vpd_table;
    char                tmp_str[VPD_TAG_DATA_MAX_LEN];
    char                yb_data[4];


    // Prepare the result pointer
    result_ptr = Tcl_GetObjResult(interp);

    // Verify the number of arguments is 2
    if (objc != 2) {
        vpd_read_usage(interp);
        return TCL_ERROR;
    }

    // Save the first argument at tag_name
    tag_name = Tcl_GetString(objv[1]); 

    // Verify the VPD field name is recognized
    if (verify_vpd_field_name(tag_name) != TCL_OK) {
        sprintf (err_str, "vpd read failed! field name %s is incorrect\n", tag_name);
        print_msg (err_str, ESC_COLOR_RED);
        sprintf (err_str, "field name should be PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME\n");
        print_msg (err_str, ESC_COLOR_RED);
        return TCL_ERROR;
    }

    // Check if VPD image exists
    is_vpd_image = atoi(Tcl_GetVar2(interp, TCL_ENV, ENV_IS_VPD, TCL_GLOBAL_ONLY));
    DbgMessage(NULL, WARN, "tcl_al_vpd_read_cmd(): is_vpd_image=%d\n", is_vpd_image);

    // Get the request VPD field value */
    if (is_vpd_image) {
        // if VPD image exists in NVM, get this info from VPD image
        get_vpd_info_from_image(interp, &vpd_table);

        // If requested VPD info is PRODUCT_NAME, get it from vpd_table.product_name
        if (strcmp(tag_name, "PRODUCT_NAME") ==0) {

            Tcl_AppendStringsToObj(result_ptr, vpd_table.product_name, NULL);

        } else {

            // Else, find the requested VPD info which matches the tag_name
            for (i=0; i<vpd_table.no_tags; i++) {

                if (strcmp(tag_name, vpd_table.vpd_tags[i].tag_name) ==0) {

                    if (strcmp(vpd_table.vpd_tags[i].tag_name, "YB") != 0) {
                         Tcl_AppendStringsToObj(result_ptr, vpd_table.vpd_tags[i].tag_data, NULL);
                    } else {
                        // YB tag is binary
                        strcpy(tmp_str, "");
						yb_data[3] = '\0';
                        for (j=0; j<vpd_table.vpd_tags[i].tag_len; j++) {
                            sprintf(yb_data, "%02x ", vpd_table.vpd_tags[i].tag_data[j]);
                            strcat(tmp_str, yb_data);
                        }
                        Tcl_AppendStringsToObj(result_ptr, tmp_str, NULL);
                    }
                    break;
                }
            }
            // If requested field is not found, type a warning message
            if (i==vpd_table.no_tags) {
                Tcl_AppendStringsToObj(result_ptr, "Requested tag is not found in VPD info!", NULL);
            }
        }

    } else {
        // Set NVM offset of VPD
        nvm_offset = NVM_OFFSET(vpd);

        DbgMessage(NULL, WARN, "tcl_al_vpd_read_cmd(): nvm_offset=0x%x\n", nvm_offset);

        // Extract VPD data from offset in NVM
        status = common_nvm_read_nvram(nvm_offset, (u32_t*)(&vpd_block), sizeof(media_vpd_t));
        if ( status != COMMON_NVM_STATUS_SUCCESS ) {
            return TCL_ERROR;
        }

        // Save the VPD information within vpd_info_t struct
        common_nvm_get_vpd_info(&vpd_info_struct, &vpd_block, vpd_check);

        // Return the requested VPD field value (if exist)
        if (strcmp(tag_name, "PRODUCT_NAME") ==0) {
            if (strlen(vpd_info_struct.product_name)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.product_name, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field product_name does not exist", NULL);
            }
        } else if (strcmp(tag_name, "PN")==0) {
            if (strlen(vpd_info_struct.part_number)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.part_number, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field part_number does not exist", NULL);
            }
        } else if (strcmp(tag_name, "EC")==0) {
            if (strlen(vpd_info_struct.ec_number)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.ec_number, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field ec_number does not exist", NULL);
            }
        } else if (strcmp(tag_name, "SN")==0) {
            if (strlen(vpd_info_struct.serial_number)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.serial_number, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field serial_number does not exist", NULL);
            }
        } else if (strcmp(tag_name, "MN")==0) {
            if (strlen(vpd_info_struct.manufacturing_id)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.manufacturing_id, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field manufacturing_id does not exist", NULL);
            }
        } else if (strcmp(tag_name, "YA")==0) {
            if (strlen(vpd_info_struct.asset_tag)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.asset_tag, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field asset_tag does not exist", NULL);
            }
        } else if (strcmp(tag_name, "V0")==0) {
            if (strlen(vpd_info_struct.v0)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.v0, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field v0 does not exist", NULL);
            }
        } else if (strcmp(tag_name, "V1")==0) {
            if (strlen(vpd_info_struct.v1)) {
                Tcl_AppendStringsToObj(result_ptr, vpd_info_struct.v1, NULL);
            } else {
                Tcl_AppendStringsToObj(result_ptr, "VPD field v1 does not exist", NULL);
            }
        } else {
            Tcl_AppendStringsToObj(result_ptr, "Unknown VPD field\n", NULL);
        }
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * Description:
 *     Implements "vpd show" command.
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */ 
static int tcl_al_vpd_show_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]  /* argument strings */
    )
{
    Tcl_Obj             *result_ptr = NULL;
    media_vpd_t         vpd_block;
    vpd_info_t          vpd_info_struct;
    u32_t               vpd_check=1;
    u32_t               nvm_offset, i, j;
    common_nvm_status_t status;
    u8_t                is_vpd_image;
    vpd_table_t         vpd_table;
    char                tmp_str[VPD_TAG_DATA_MAX_LEN];
    char                yb_data[4];


    // Prepare the result pointer
    result_ptr = Tcl_GetObjResult(interp);

    // Verify the number of arguments is 2
    if (objc > 1) {
        vpd_show_usage(interp);
        return TCL_ERROR;
    }

    // Check if VPD image exists
    is_vpd_image = atoi(Tcl_GetVar2(interp, TCL_ENV, ENV_IS_VPD, TCL_GLOBAL_ONLY));
    DbgMessage(NULL, WARN, "tcl_al_vpd_show_cmd(): is_vpd_image=%d\n", is_vpd_image);

    Tcl_AppendStringsToObj(result_ptr, "VPD information: \n", NULL);

    // Get the request VPD field value */
    if (is_vpd_image) {
        // if VPD image exists in NVM, get this info from VPD image
        get_vpd_info_from_image(interp, &vpd_table);

        // printout the VPD fields
        Tcl_AppendStringsToObj(result_ptr, "PRODUCT_NAME = ", vpd_table.product_name, "\n",  NULL);
        for (i=0; i<vpd_table.no_tags; i++) {
            if (strcmp(vpd_table.vpd_tags[i].tag_name, "RV") == 0) {
                break;
            } else {
                if (strcmp(vpd_table.vpd_tags[i].tag_name, "YB") != 0) {
                    Tcl_AppendStringsToObj(result_ptr, vpd_table.vpd_tags[i].tag_name, " = ", vpd_table.vpd_tags[i].tag_data, "\n", NULL);
                } else {
                    // YB tag is binary
                    strcpy(tmp_str, "");
					yb_data[3] = '\0';
                    for (j=0; j<vpd_table.vpd_tags[i].tag_len; j++) {
                        sprintf(yb_data, "%02x ", vpd_table.vpd_tags[i].tag_data[j]);
                        strcat(tmp_str, yb_data);
                    }
                    Tcl_AppendStringsToObj(result_ptr, vpd_table.vpd_tags[i].tag_name, " = ", tmp_str, NULL);                    
                }
            }
        }

    } else {
        // Set NVM offset of VPD
        nvm_offset = NVM_OFFSET(vpd);

        DbgMessage(NULL, WARN, "tcl_al_vpd_show_cmd(): nvm_offset=0x%x\n", nvm_offset);

        // Extract VPD data from offset in NVM
        status = common_nvm_read_nvram(nvm_offset, (u32_t*)(&vpd_block), sizeof(media_vpd_t));
        if ( status != COMMON_NVM_STATUS_SUCCESS ) {
            return TCL_ERROR;
        }

        // Save the VPD information within vpd_info_t struct
        common_nvm_get_vpd_info(&vpd_info_struct, &vpd_block, vpd_check);

        // print the VPD fields (if exist)
        if (strlen(vpd_info_struct.product_name)) {
            Tcl_AppendStringsToObj(result_ptr, "PRODUCT_NAME = ", vpd_info_struct.product_name, "\n",  NULL);
        }
        if (strlen(vpd_info_struct.part_number)) {
            Tcl_AppendStringsToObj(result_ptr, "PN = ", vpd_info_struct.part_number, "\n", NULL);
        }
        if (strlen(vpd_info_struct.ec_number)) {
            Tcl_AppendStringsToObj(result_ptr, "EC = ", vpd_info_struct.ec_number, "\n", NULL);
        }
        if (strlen(vpd_info_struct.serial_number)) {
            Tcl_AppendStringsToObj(result_ptr, "SN = ", vpd_info_struct.serial_number, "\n", NULL);
        }
        if (strlen(vpd_info_struct.manufacturing_id)) {
            Tcl_AppendStringsToObj(result_ptr, "MN = ", vpd_info_struct.manufacturing_id, "\n", NULL);
        }
        if (strlen(vpd_info_struct.asset_tag)) {
            Tcl_AppendStringsToObj(result_ptr, "YA = ", vpd_info_struct.asset_tag, "\n", NULL);
        }
        if (strlen(vpd_info_struct.v0)) {
            Tcl_AppendStringsToObj(result_ptr, "V0 = ", vpd_info_struct.v0, "\n", NULL);
        }
        if (strlen(vpd_info_struct.v1)) {
            Tcl_AppendStringsToObj(result_ptr, "V1 = ", vpd_info_struct.v1, "\n", NULL);
        }
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * Description:
 *     Calculate total length required for VPD information
 *------------------------------------------------------------------------
 */ 
static u32_t calc_vpd_info_total_length(vpd_table_t *vpd_table_p) 
{
    u32_t total_len=0, i=0;

    total_len += vpd_table_p->product_name_len + 3; // product name + tag name + len
    total_len += 3; //for VPD_R tag
    DbgMessage(NULL, WARN, "calc_vpd_info_total_length(): total_len=%d\n", total_len);


    for (i=0; i<vpd_table_p->no_tags; i++) {
        // add to total length the tag len + 3 bytes (for tag name, tag len)
        total_len += vpd_table_p->vpd_tags[i].tag_len + 3;
    }
    // Note: 6 more bytes are required for RV tag, checksum, End tag
    total_len += 6;
    DbgMessage(NULL, WARN, "calc_vpd_info_total_length(): total_len=%d no_tags=%d\n", total_len, vpd_table_p->no_tags);

    return total_len;
}

/*
 *------------------------------------------------------------------------
 * Description:
 *     Create VPD image from vpd_table_t
 *------------------------------------------------------------------------
 */ 

static vpd_image_t* create_vpd_image ( vpd_table_t *vpd_table_p, u32_t vpd_table_size, u32_t vpd_image_rev)
{
    u8_t        *ptr=NULL, *checksum_start;
    vpd_image_t *vpd_image_ptr;
    u16_t       len_remaining=0;
    u32_t       tag_idx=0, i=0, j=0;
    u8_t        checksum;
    u32_t       required_vpd_size=0;


    /* Allocate memory for VPD image built from the parsed VPD text file */    
    vpd_image_ptr = (vpd_image_t *) cmn_os_alloc_virt(vpd_table_size+4);
    if ( vpd_image_ptr == NULL ) {
        DbgBreakMsg("ERROR: Failed to allocate memory for the VPD image!\n");
        goto handle_error;
    }

    DbgMessage(NULL, WARN, "create_vpd_image(): vpd_table_size=%d\n", vpd_table_size);

    /* Clean the VPD image */
    cmn_os_memset(vpd_image_ptr, 0, vpd_table_size);

    /* Set the image format revision */
    vpd_image_ptr->format_revision = vpd_image_rev;
    ptr = ((u8_t *) (vpd_image_ptr));
    ptr += 4;

    /* Save the ptr for checksum calculation */
    checksum_start = ptr;       

    /* Set the ID String TAG with the product name*/
    *ptr = (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_STRING);
    ptr++;

    cmn_os_memcpy(ptr, &vpd_table_p->product_name_len, 2);
    ptr += 2;
    strncpy((char *) ptr, vpd_table_p->product_name, vpd_table_p->product_name_len); 
    ptr += vpd_table_p->product_name_len;

    /****************  VPD-R *************/    
    *ptr = (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_VPD_R);
    ptr ++;

    /* Verify VPD table size is enough */
    required_vpd_size = calc_vpd_info_total_length(vpd_table_p);
    DbgMessage(NULL, WARN, "create_vpd_image(): vpd_table_size=%d required length=%d\n", vpd_table_size, required_vpd_size);
    if (vpd_table_size < required_vpd_size ) {
        DbgBreakMsg("ERROR: vpd_table_size is too small !\n");
        goto handle_error;
    }

    /* Length field should include the remaining no. bytes 
       VPD table size - product name length - 3 bytes (1 for ID tag, 2 for ID tag len*/
    len_remaining = vpd_table_size - vpd_table_p->product_name_len - 3;
    DbgMessage(NULL, WARN, "create_vpd_image(): len_remaining=%d\n", len_remaining);

    /* reduce 3 bytes of VPD_R tag*/
    len_remaining -= 3;
    /* Subtract one more for the END TAG */
    len_remaining -= 1;
    cmn_os_memcpy(ptr, &len_remaining, 2);
    ptr += 2;


    /* Copy the VPD fields information into a VPD image format */
    for (tag_idx=0; tag_idx < vpd_table_p->no_tags; tag_idx++) {

        /* Copy the TAG name (e.g. PN, EC, MN ..) */
        strncpy((char *)ptr, vpd_table_p->vpd_tags[tag_idx].tag_name, 2);
        /* Increment the data ptr by 2 (since TAG name is 2 bytes) */
        ptr += 2;

        /* Copy the TAG Length */
        *ptr = (u8_t) (vpd_table_p->vpd_tags[tag_idx].tag_len);
        ptr++;

        /* Copy the TAG data */
        if (strcmp(vpd_table_p->vpd_tags[tag_idx].tag_name, "YB") != 0) {
            strncpy((char *) ptr, vpd_table_p->vpd_tags[tag_idx].tag_data, vpd_table_p->vpd_tags[tag_idx].tag_len);  
            ptr += vpd_table_p->vpd_tags[tag_idx].tag_len;
        } else {
            
            for (j=0; j < vpd_table_p->vpd_tags[tag_idx].tag_len; j++) {
               *ptr = (u8_t) vpd_table_p->vpd_tags[tag_idx].tag_data[j];
               ptr++;
            }
        }
        len_remaining -= (vpd_table_p->vpd_tags[tag_idx].tag_len + 3);

        DbgMessage(NULL, WARN, "create_vpd_image(): tag_idx=%d tag_name=%s tag_len=%d tag_data=%s len_remaining=%d\n", 
                    tag_idx, vpd_table_p->vpd_tags[tag_idx].tag_name,
                    vpd_table_p->vpd_tags[tag_idx].tag_len,
                    vpd_table_p->vpd_tags[tag_idx].tag_data, len_remaining);

    }

    /* Add RV checksum field */
    strcpy((char *)ptr,"RV"); 
    ptr += 2;
    /* Remaining length */
    len_remaining -= 3;
    if (len_remaining > 255) {
        print_msg ("VPD file parsing error: remaining length at RV TAG exceeds 255\n", ESC_COLOR_RED);
        goto handle_error;
    }
    *ptr = (u8_t )(len_remaining);
    ptr++;

    /* Calculate the checksum */
    checksum = 0;
    while (checksum_start < ptr) {
        checksum += *checksum_start++;
    }
    *ptr = -checksum;
    ptr++;

    /* Fill the remaining bytes with reserved space 0x00*/
    if (len_remaining > 0) {
        for (i=0; i<len_remaining-1; i++) {
            *ptr=0;
            ptr++;
        }
    }

    /* End Tag */
    *ptr = (TAG_DATA_TYPE_SR_END << 3);

    return(void *)vpd_image_ptr;

handle_error:
    if (vpd_image_ptr) cmn_os_free_virt((int_ptr_t)vpd_image_ptr);
    return NULL;
}

/*
 * ------------------------------------------------------------------
 * Description:
 *     Implements "vpd write" command.
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */ 
static int tcl_al_vpd_write_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]  /* argument strings */
    )
{
    Tcl_Obj             *result_ptr = NULL;
    media_vpd_t         vpd_block;
    vpd_info_t          vpd_info_struct;
    u32_t               vpd_check=1;
    char                *tag_name;
    char                *tag_data;
    char                err_str[80];
    u32_t               nvm_offset, i, j, val32;
    common_nvm_status_t status;
    u8_t                is_vpd_image;
    vpd_table_t         vpd_table;
    vpd_image_t         *vpd_image_ptr;
    u32_t               vpd_image_offset, vpd_image_idx;
    u32_t               vpd_image_size, vpd_image_rev, vpd_table_size;
    u8_t                validate_versions, silent, val8;
    u32_t               vpd_info_total_len;
    char                *token_ptr;


    // Prepare the result pointer
    result_ptr = Tcl_GetObjResult(interp);

    // Verify the number of arguments is 3
    if (objc != 3) {
        vpd_write_usage(interp);
        return TCL_ERROR;
    }

    // Save the first argument at tag_name
    tag_name = Tcl_GetString(objv[1]); 

    // Verify the VPD field name is recognized
    if (verify_vpd_field_name(tag_name) != TCL_OK) {
        sprintf (err_str, "vpd write failed! field name %s is incorrect\n", tag_name);
        print_msg (err_str, ESC_COLOR_RED);
        sprintf (err_str, "field name should be PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME\n");
        print_msg (err_str, ESC_COLOR_RED);
        return TCL_ERROR;
    }

    // Save the second argument at tag_data
    tag_data = Tcl_GetString(objv[2]); 

    // Verify the field data length does not exceed the maximum allowed
    if (strlen(tag_data) > VPD_TAG_DATA_MAX_LEN) {
        sprintf (err_str, "field data exceeds the maximum length allowed %d \n", VPD_TAG_DATA_MAX_LEN );
        print_msg (err_str, ESC_COLOR_RED);
        return TCL_ERROR;
    }

    // Check if VPD image exists
    is_vpd_image = atoi(Tcl_GetVar2(interp, TCL_ENV, ENV_IS_VPD, TCL_GLOBAL_ONLY));
    DbgMessage(NULL, WARN, "tcl_al_vpd_write_cmd(): is_vpd_image=%d\n", is_vpd_image);

    // Get the request VPD field value */
    if (is_vpd_image) {
        // if VPD image exists in NVM, get this info from VPD image
        get_vpd_info_from_image(interp, &vpd_table);

        // If requested VPD field is PRODUCT_NAME, save it at vpd_table.product_name
        if (strcmp(tag_name, "PRODUCT_NAME") ==0) {
            vpd_table.product_name_len = strlen(tag_data);
            strncpy(vpd_table.product_name, tag_data, vpd_table.product_name_len);
            vpd_table.product_name[vpd_table.product_name_len] = '\0';
        } else {
            // Else, check if this field already exists
            for (i=0; i<vpd_table.no_tags; i++) {

                if (strcmp(tag_name, vpd_table.vpd_tags[i].tag_name) ==0) {
                    if (strcmp(tag_name, "YB") != 0) {
                        /* Tag is text */
                        strncpy(vpd_table.vpd_tags[i].tag_data, tag_data, strlen(tag_data));
                        vpd_table.vpd_tags[i].tag_data[strlen(tag_data)]='\0';
                        vpd_table.vpd_tags[i].tag_len = strlen(tag_data);
                    } else {
                        /* YB tag is binary */
                        token_ptr = strtok( tag_data, " " );
                        for (j=0; (token_ptr != NULL); j++) {
                            sscanf(token_ptr, "%2x", (int *) &val8);
                            vpd_table.vpd_tags[i].tag_data[j] = val8;
                            token_ptr = strtok( NULL, " ");
                        }
                        vpd_table.vpd_tags[i].tag_len = j;
                    }
                    break;
                }
            }
            // If requested field is not found, add a new field to vpd info
            if (i==vpd_table.no_tags) {
                Tcl_AppendStringsToObj(result_ptr, "VPD tag not found. Adding a new VDP tag ..\n", NULL);
                strncpy(vpd_table.vpd_tags[i].tag_name, tag_name, strlen(tag_name));
                vpd_table.vpd_tags[i].tag_name[strlen(tag_name)] = '\0';

                strncpy(vpd_table.vpd_tags[i].tag_data, tag_data, strlen(tag_data));
                vpd_table.vpd_tags[i].tag_data[strlen(tag_data)] = '\0';

                vpd_table.vpd_tags[i].tag_len = strlen(tag_data);
                vpd_table.no_tags++;
            }
        }

        /* Get VPD image information: offset, length, image_idx */
        if (common_nvm_get_image_in_extended_dir_info(NVM_IMAGE_NAME_VPD, &vpd_image_offset, &vpd_image_size, &vpd_image_idx) != COMMON_NVM_STATUS_SUCCESS) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to get VPD image information !!\n", NULL);
            return TCL_ERROR;
        }
        vpd_table_size = vpd_image_size - 8; //4 CRC bytes, 4 revision bytes

        /* Calclutae the required size for the updated VPD info */
        vpd_info_total_len = calc_vpd_info_total_length(&vpd_table);
        DbgMessage(NULL, WARN, "tcl_al_vpd_write_cmd(): vpd_info_total_len=%d vpd_table_size=%d\n",vpd_info_total_len, vpd_table_size);

        /* If the require size for the updated VPD info is larger the the deleted VPD image size,
           update the vpd_image_size to the nearst multiple of 4 */
        if (vpd_info_total_len > vpd_table_size ) {
            vpd_table_size = ((vpd_info_total_len/4) + 1) * 4;
            DbgMessage(NULL, WARN, "tcl_al_vpd_write_cmd(): Required vpd_image_size=%d\n", vpd_table_size);
        }

        /* Set vpd image_rev to 1*/
        vpd_image_rev = VPD_IMAGE_VERSION;

        /* Create VPD image buffer for programming */
        vpd_image_ptr = create_vpd_image (&vpd_table , vpd_table_size, vpd_image_rev);

        if ( vpd_image_ptr == NULL ) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to create VPD image!\n");
            return TCL_ERROR;
        }

        /* Delete VPD image from NVM */
        if (common_nvm_delete_vpd_image() != COMMON_NVM_STATUS_SUCCESS) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to delete VPD image from NVM !!\n", NULL);
            return TCL_ERROR;
        }

        /* Program the update VPD image in the extended dir */
        validate_versions = FALSE;
        silent = TRUE;
        if (common_nvm_prog_image_in_extended_dir((u8_t *) vpd_image_ptr, vpd_table_size+4, vpd_image_idx, 
                                                  validate_versions, silent, FALSE, FALSE) != COMMON_NVM_STATUS_SUCCESS) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to program updated VPD image !!\n", NULL);
            return TCL_ERROR;
        }


    } else {
        // Set NVM offset of VPD
        nvm_offset = NVM_OFFSET(vpd);

        DbgMessage(NULL, WARN, "tcl_al_vpd_write_cmd(): nvm_offset=0x%x\n", nvm_offset);

        // Extract VPD data from offset in NVM
        status = common_nvm_read_nvram(nvm_offset, (u32_t*)(&vpd_block), sizeof(media_vpd_t));
        if ( status != COMMON_NVM_STATUS_SUCCESS ) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to read from NVM offset for VPD !!\n", NULL);
            return TCL_ERROR;
        }

        // Get the VPD information into vpd_info_t struct
        common_nvm_get_vpd_info(&vpd_info_struct, &vpd_block, vpd_check);

        // Update the requested VPD field value in vpd_info_struct
        if (strcmp(tag_name, "PRODUCT_NAME") ==0) {
            if (strlen(tag_data) > MAX_PRODUCT_NAME_LENGTH) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD PRODUCT_NAME field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.product_name, tag_data, strlen(tag_data));
            vpd_info_struct.product_name[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "PN")==0) {
            if (strlen(tag_data) > MAX_PART_NUMBER_LENGTH) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD PN field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.part_number, tag_data, strlen(tag_data));
            vpd_info_struct.part_number[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "EC")==0) {
            if (strlen(tag_data) > MAX_EC_NUMBER_LENGTH) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD EC field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.ec_number, tag_data, strlen(tag_data));
            vpd_info_struct.ec_number[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "SN")==0) {
            if (strlen(tag_data) > MAX_SERIAL_NUMBER_LENGTH) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD SN field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.serial_number, tag_data, strlen(tag_data));
            vpd_info_struct.serial_number[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "MN")==0) {
            if (strlen(tag_data) > MAX_MANUFACT_ID_LENGTH) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD MN field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.manufacturing_id, tag_data, strlen(tag_data));
            vpd_info_struct.manufacturing_id[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "YA")==0) {
            if (strlen(tag_data) > MAX_ASSET_TAG_LENGTH) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD V0 field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.asset_tag, tag_data, strlen(tag_data));
            vpd_info_struct.asset_tag[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "V0")==0) {
            if (strlen(tag_data) > MAX_VENDDOR_SPECIFIC_V0) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD V0 field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.v0, tag_data, strlen(tag_data));
            vpd_info_struct.v0[strlen(tag_data)]='\0';

        } else if (strcmp(tag_name, "V1")==0) {
            if (strlen(tag_data) > MAX_VENDDOR_SPECIFIC_V1) {
                Tcl_AppendStringsToObj(result_ptr, "Input is too long for VPD V1 field !\n", NULL);
                return TCL_ERROR;
            }
            strncpy(vpd_info_struct.v1, tag_data, strlen(tag_data));
            vpd_info_struct.v1[strlen(tag_data)]='\0';

        } else {
            Tcl_AppendStringsToObj(result_ptr, "Unknown VPD field\n", NULL);
            return TCL_ERROR;
        }

        // Save the updated VPD information into vpd_block struct
        common_nvm_set_vpd_info(&vpd_info_struct, &vpd_block);

        // Program the updated VPD information into NVM
        status = common_nvm_program_nvm((u8_t *)&vpd_block, sizeof(media_vpd_t),
                                        nvm_offset, 0, &val32);
        if ( status != COMMON_NVM_STATUS_SUCCESS ) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to program updated VPD info to NVM !!\n", NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------
 * Description:
 *     Implements "vpd delete" command.
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */ 
static int tcl_al_vpd_delete_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]  /* argument strings */
    )
{
    Tcl_Obj             *result_ptr = NULL;
    char                *tag_name;
    char                err_str[80];
    u32_t               i, j, tag_len;
    u8_t                is_vpd_image;
    vpd_table_t         vpd_table;
    vpd_image_t         *vpd_image_ptr;
    u32_t               vpd_image_offset, vpd_image_idx;
    u32_t               vpd_image_size, vpd_image_rev, vpd_table_size;
    u8_t                validate_versions, silent;
    u32_t               vpd_info_total_len;
    u8_t                found_matching_tag=0;


    // Prepare the result pointer
    result_ptr = Tcl_GetObjResult(interp);

    // Verify the number of arguments is 2
    if (objc != 2) {
        vpd_delete_usage(interp);
        return TCL_ERROR;
    }

    // Save the first argument at tag_name
    tag_name = Tcl_GetString(objv[1]); 

    // Verify the VPD field name is recognized
    if (verify_vpd_field_name(tag_name) != TCL_OK) {
        sprintf (err_str, "vpd delete failed! field name %s is incorrect\n", tag_name);
        print_msg (err_str, ESC_COLOR_RED);
        sprintf (err_str, "field name should be PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx\n");
        print_msg (err_str, ESC_COLOR_RED);
        return TCL_ERROR;
    }


    // Check if VPD image exists
    is_vpd_image = atoi(Tcl_GetVar2(interp, TCL_ENV, ENV_IS_VPD, TCL_GLOBAL_ONLY));
    DbgMessage(NULL, WARN, "tcl_al_vpd_delete_cmd(): is_vpd_image=%d\n", is_vpd_image);

    // Get the request VPD field value */
    if (is_vpd_image) {
        // if VPD image exists in NVM, get this info from VPD image
        get_vpd_info_from_image(interp, &vpd_table);

        // If requested VPD field is PRODUCT_NAME, save it at vpd_table.product_name
        if (strcmp(tag_name, "PRODUCT_NAME") ==0) {
            Tcl_AppendStringsToObj(result_ptr, "You can not delete the Product Name tag from VPD !", NULL);
            return TCL_ERROR;
        } else {
            // Else, check if this field exists
            for (i=0; i<vpd_table.no_tags; i++) {

                DbgMessage(NULL, WARN, "tcl_al_vpd_delete_cmd(): i=%d tag_name=%s\n", i, vpd_table.vpd_tags[i].tag_name);

                if (strcmp(tag_name, vpd_table.vpd_tags[i].tag_name) ==0) {

                    DbgMessage(NULL, WARN, "tcl_al_vpd_delete_cmd(): Found a matching VPD tag to delete\n");
                    found_matching_tag=1;

                    /* Found a matching VPD tag*/
                    /* Delete this VPD tag info by moving all the following tags one by one*/
                    for (j=i; j<vpd_table.no_tags; j++) {
                        strncpy(vpd_table.vpd_tags[j].tag_name, vpd_table.vpd_tags[j+1].tag_name, 2);
                        vpd_table.vpd_tags[j].tag_name[2]='\0';
                        vpd_table.vpd_tags[j].tag_len = vpd_table.vpd_tags[j+1].tag_len;
                        tag_len = vpd_table.vpd_tags[j].tag_len;
                        strncpy(vpd_table.vpd_tags[j].tag_data, vpd_table.vpd_tags[j+1].tag_data, tag_len);
                        vpd_table.vpd_tags[j].tag_data[tag_len]='\0';
                    }
                    vpd_table.no_tags--;
                    DbgMessage(NULL, WARN, "tcl_al_vpd_delete_cmd(): vpd_table.no_tags=%d\n", vpd_table.no_tags);
                    break;
                }
            }
            // If requested field is not found, show a warning message
            if (!found_matching_tag) {
                Tcl_AppendStringsToObj(result_ptr, "VPD tag not found. Nothing to delete ..", NULL);
                return TCL_ERROR;
            }
        }

        /* Get VPD image information: offset, length, image_idx */
        if (common_nvm_get_image_in_extended_dir_info(NVM_IMAGE_NAME_VPD, &vpd_image_offset, &vpd_image_size, &vpd_image_idx) != COMMON_NVM_STATUS_SUCCESS) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to get VPD image information !!", NULL);
            return TCL_ERROR;
        }
        vpd_table_size = vpd_image_size - 8; //4 CRC bytes, 4 revision bytes

        /* Calclutae the required size for the updated VPD info */
        vpd_info_total_len = calc_vpd_info_total_length(&vpd_table);
        DbgMessage(NULL, WARN, "tcl_al_vpd_delete_cmd(): vpd_info_total_len=%d vpd_table_size=%d\n",vpd_info_total_len, vpd_table_size);

        /* If the require size for the updated VPD info is smaller then the deleted VPD image size,
           update the vpd_image_size to the nearst multiple of 4 */
        vpd_info_total_len = ((vpd_info_total_len/4) + 1) * 4;
        if (vpd_info_total_len < vpd_table_size) {
            vpd_table_size = vpd_info_total_len + 4; //4 bytes for the image revision
            DbgMessage(NULL, WARN, "tcl_al_vpd_delete_cmd(): Required vpd_image_size=%d\n", vpd_table_size);
        }

        /* Set vpd image_rev to 1*/
        vpd_image_rev = VPD_IMAGE_VERSION;

        /* Create VPD image buffer for programming */
        vpd_image_ptr = create_vpd_image (&vpd_table , vpd_table_size, vpd_image_rev);

        if ( vpd_image_ptr == NULL ) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to create VPD image!\n");
            return TCL_ERROR;
        }

        /* Delete VPD image from NVM */
        if (common_nvm_delete_vpd_image() != COMMON_NVM_STATUS_SUCCESS) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to delete VPD image from NVM !!", NULL);
            return TCL_ERROR;
        }

        /* Program the update VPD image in the extended dir */
        validate_versions = FALSE;
        silent = TRUE;
        if (common_nvm_prog_image_in_extended_dir((u8_t *) vpd_image_ptr, vpd_table_size+4, vpd_image_idx, 
                                                  validate_versions, silent, FALSE, FALSE) != COMMON_NVM_STATUS_SUCCESS) {
            Tcl_AppendStringsToObj(result_ptr, "Failed to program updated VPD image !!", NULL);
            return TCL_ERROR;
        }


    } else {
        Tcl_AppendStringsToObj(result_ptr, "VPD image does not exist. You can not delete VPD tag !!", NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}
/*
 *------------------------------------------------------------------------
 * Description:
 *     Implements the "vpd" command, which can read/write/show/delete
 *     VPD field in NVM.
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */
int tcl_al_vpd_cmd (
                   IN    ClientData client_data, /* Not used. */
                   IN    Tcl_Interp *interp,     /* current interpreter */
                   IN    int objc,               /* number of arguments */
                   IN    Tcl_Obj *CONST objv[]   /* argument strings */
                   )
{  
    static char *option_names[] = { "read", "show", "write", "delete", "help", (char *)NULL };

    enum options {
        VPD_READ, VPD_SHOW, VPD_WRITE, VPD_DELETE, VPD_HELP
    };
    int index;

    tcl_al_log_command(VPD_CMD, objc, objv);

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option=(read| show| write | delete | help) arg...");
        return TCL_ERROR;
    }
#ifdef EDEBUG
    if (Tcl_GetIndexFromObj(interp, objv[1], (const char **)option_names, "option", TCL_EXACT, &index) != TCL_OK) {
#else
    if (Tcl_GetIndexFromObj(interp, objv[1], option_names, "option", TCL_EXACT, &index) != TCL_OK) {
#endif
        return TCL_ERROR;
    }

    switch ((enum options)index) {
    case VPD_READ:
        return tcl_al_vpd_read_cmd(interp, objc-1, objv+1); 
    case VPD_SHOW:
        return tcl_al_vpd_show_cmd(interp, objc-1, objv+1); 
    case VPD_WRITE:
        return tcl_al_vpd_write_cmd(interp, objc-1, objv+1);         
    case VPD_DELETE:
        return tcl_al_vpd_delete_cmd(interp, objc-1, objv+1);
    case VPD_HELP:
        usage_vpd();
        return TCL_OK;
    default:
        return TCL_ERROR;
    }

    return TCL_OK; 
}

/*
*------------------------------------------------------------------------
* Description:
*     Clean up the "vpd" command.
* 
* Returns:
*     A standard Tcl result.
*------------------------------------------------------------------------
*/
int tcl_al_vpd_cleanup(
    IN   Tcl_Interp *interp
    )
{
    return Tcl_DeleteCommand(interp, VPD_CMD);
}


/*
*------------------------------------------------------------------------
* tcl_al_vpd_usage --
*
*     Add usage information on "vpd" command 
*------------------------------------------------------------------------
*/
int tcl_al_vpd_usage(
    IN  char *buf,              //Pointer to the buffer of the command usage
    IN  u32_t *length           //Pointer to the length result of the command usage
    )
{   
    u32_t offset = 0;
    *length = 0;

    *length += sprintf( &buf[offset+*length], "Read, write or delete a vpd field, show all vpd fields\n\n");

    *length += sprintf( &buf[offset+*length], "Usage: vpd read <field name> \n");
    *length += sprintf( &buf[offset+*length], "# Shows the requested VPD field value\n");
    *length += sprintf( &buf[offset+*length], "# Inputs: <field name: (PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME)\n");

    *length += sprintf( &buf[offset+*length], "Usage: vpd write <field name> <value>\n");
    *length += sprintf( &buf[offset+*length], "# Change the requested VPD field value\n");
    *length += sprintf( &buf[offset+*length], "# Inputs: <field name: (PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx or PRODUCT_NAME)\n");

    *length += sprintf( &buf[offset+*length], "Usage: vpd show\n");
    *length += sprintf( &buf[offset+*length], "# Show all VPD fields\n");

    *length += sprintf( &buf[offset+*length], "Usage: vpd delete <field name>\n");
    *length += sprintf( &buf[offset+*length], "# Delete the requested VPD field value\n");
    *length += sprintf( &buf[offset+*length], "# Inputs: <field name: (PN/EC/MN/SN/YA/FN/CC/FC/YB/Vx)\n");

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_read_vpd_text_file: 
 * Parse the VPD information text file
 * 
 * Input: 
 * Specify the name of the file that contains the VPD information
 *
 * return: 
 * NULL on error; otherwise, a handle used for other operations.
 *------------------------------------------------------------------------
 */
void * tcl_al_read_vpd_text_file(char *filename, u32_t *byte_read)
{
    vpd_image_t *vpd_image_ptr = NULL;
    FILE *fd;
    char *cptr = NULL;
    char line[1024];
    char err_str[80];
    u32_t val32, tag_idx=0, i=0;
    char tmp_str[1024];
    char field_name[VPD_ID_TAG_MAX_LEN+1];
    char field_type[VPD_TYPE_MAX_LEN+1];
    char field_data[VPD_TAG_DATA_MAX_LEN+1];
    vpd_table_t vpd_table;
    u32_t is_vpd_r=0, is_vpd_w=0, vpd_rev=0;
    u32_t required_vpd_size=0, vpd_table_size;
    u8_t is_product_name_field=0, product_name_exist=0;
	int val8;
    char *token_ptr;


#if !defined (UEFI) && !defined (WIN_DIAG)
    val32 = access(filename, R_OK | W_OK | F_OK);
    if (val32 == -1) {
        return NULL;
    }
#endif

    fd = fopen(filename, "r");
    if (fd == NULL) {
        return NULL;
    }

    /* Initialize local variables */
    field_name[0]='\0';
    field_type[0]='\0';
    field_data[0]='\0';
    *byte_read = 0;


    while (!feof(fd)) {
        if (fgets(line, 1024, fd)) {
            /* Skip comment lines */
            if (line[0] == '#') continue;
            if (line[0] == ';') continue;

            /* Skip lines which do not include '=' or ',' */
            if ((strchr(line, '=') == NULL) && (strchr(line, ',') == NULL)) continue;

            /* Get the VPD text file revision number */
            strcpy(tmp_str, line);
            strtoupr(tmp_str);
            if (strstr(tmp_str, "REVISION")) {
                /* Get the value after the '=' char */
                cptr = strchr(tmp_str, '=');
                for (cptr++; *cptr != ' '; cptr++);
                val32 = strtoul(&cptr[1], &cptr, 16);
                vpd_rev = val32;
                //printf("VPD version = %d \n", vpd_rev);
                if (vpd_rev != VPD_IMAGE_VERSION ) {
                    sprintf(err_str, "VPD file parsing error: VPD-R version is not as expected version %d\n", VPD_IMAGE_VERSION);
                    print_msg (err_str, ESC_COLOR_RED);
                    return NULL;
                }
                required_vpd_size += sizeof(u32_t);
            }

            /* Look for the [REGION=VPD_R/VPD_W] title */
            strcpy(tmp_str, line);
            strtoupr(tmp_str);

            if (strstr(tmp_str, "REGION")) {
                if (strstr( tmp_str, "VPD-R")!=NULL) {
                    //printf("Found VPD-R region title\n");
                    is_vpd_r=1;
                    is_vpd_w=0;

                    if (tcl_al_get_vpd_r_size(line, &vpd_table_size) == TCL_OK) {

                        if ((vpd_table_size % 4) != 0) {
                            print_msg ("VPD file parsing error: VPD-R size should be a multiple of 4\n", ESC_COLOR_RED);
                            return NULL;
                        }
                        if (vpd_table_size > MAX_VPD_R_SIZE) {
                            sprintf(err_str, "VPD file parsing error: VPD-R size exceed maximum allowed size %d\n", MAX_VPD_R_SIZE);
                            print_msg (err_str, ESC_COLOR_RED);
                            return NULL;
                        }
                        //printf("VPD_R size = %d\n", vpd_table_size);

                    } else {
                        print_msg ("VPD file parsing error: VPD-R size invalid\n", ESC_COLOR_RED);
                        return NULL;
                    }
                } else if (strstr( tmp_str, "VPD-W")!=NULL) {
                    //printf("Found VPD-W region title\n");
                    is_vpd_w=1;
                    is_vpd_r=0;
                } else {
                    print_msg ("VPD file parsing error: Unknown region title, should be either VPD-R or VPD-W\n", ESC_COLOR_RED);
                    return NULL;
                }
            } else {
                if (is_vpd_r) {
                    /* Parse VPD_R fields */
                    if (tcl_al_parse_vpd_line(line, field_name, field_type, field_data) == TCL_OK) {
                        /*
                        printf("INFO: parsing the line %s\n", line);
                        printf("INFO: into field_name %s field_type %s field_data %s\n", field_name, field_type, field_data);
                        */

                        /* Verify the field name does not exceed the allowed length */
                        if (strlen(field_name) > VPD_ID_TAG_MAX_LEN) {
                            sprintf (err_str, "VPD file parsing error: field name (%s) string length is longer than %d\n", field_name, VPD_ID_TAG_MAX_LEN);
                            print_msg (err_str, ESC_COLOR_RED);
                            return NULL;
                        }

                        /* Verify the field name is allowed */
                        if (verify_vpd_field_name(field_name) != TCL_OK) {
                            sprintf (err_str, "VPD file parsing error: field name %s is incorrect\n", field_name);
                            print_msg (err_str, ESC_COLOR_RED);
                            sprintf (err_str, "VPD file parsing error: field name should be PRODUCT NAME/PN/EC/MN/SN/FN/CC/FC/YB/Vx\n");
                            print_msg (err_str, ESC_COLOR_RED);
                            return NULL;
                        }

                        /* Save the Produce Name for VPD image ID string TAG*/
                        if (strcmp( field_name, "PRODUCT NAME") == 0) {
                            is_product_name_field=1;
                        } else {
                            is_product_name_field=0;
                            /* Verify the field name does not exceed the allowed length */
                            if (strlen(field_name) > VPD_TAG_NAME_MAX_LEN) {
                                sprintf (err_str, "VPD file parsing error: field name (%s) string length is longer than %d\n", field_name, VPD_TAG_NAME_MAX_LEN);
                                print_msg (err_str, ESC_COLOR_RED);
                                return NULL;
                            }
                        }

                        /* Verify the field type is either Text/Binary */
                        strcpy(tmp_str, field_type);
                        strtolwr(tmp_str);
                        if ((strcmp( tmp_str, "text") != 0) && (strcmp( tmp_str, "binary") != 0)) {
                            sprintf (err_str, "VPD file parsing error: field type %s is incorrect\n", field_type);
                            print_msg (err_str, ESC_COLOR_RED);
                            print_msg ("VPD file parsing error: field type should be Text or Binary\n", ESC_COLOR_RED);
                            return NULL;
                        }

                        /* Verify the field data does not exceed the allowed length */
                        if (strlen(field_data) > VPD_TAG_DATA_MAX_LEN) {
                            sprintf (err_str, "VPD file parsing error: field data string length is longer than %d\n", VPD_TAG_DATA_MAX_LEN);
                            print_msg (err_str, ESC_COLOR_RED);
                            return NULL;
                        }

                        /* Verify the field data is not empty */
                        if (strlen(field_data) == 0) {
                            print_msg ("VPD file parsing error: field data is empty\n", ESC_COLOR_RED);
                            return NULL;
                        }

                        /* For debug */
                        /*
                        printf("VPD_R info: idx=%d field_name=[%s] field_type=[%s] field_data=[%s]\n", 
                                            tag_idx, field_name, field_type, field_data);
                        */

                        /* Get the product name */
                        if (is_product_name_field) {
                            strcpy(vpd_table.product_name, field_data);
                            vpd_table.product_name_len = strlen(vpd_table.product_name);
                            vpd_table.product_name[vpd_table.product_name_len]='\0';
                            product_name_exist=1;
                            required_vpd_size += strlen(field_data);
                        } else {
                            /* Copy the VPD field information into the local vpd_table struct */
                            strncpy(vpd_table.vpd_tags[tag_idx].tag_name, field_name,2);
                            vpd_table.vpd_tags[tag_idx].tag_name[2]='\0';
                            
                            if ((strcmp( tmp_str, "text") == 0)) {
                                /* Copy Text data into tag_data */
                                vpd_table.vpd_tags[tag_idx].tag_len = strlen(field_data);
                                strncpy(vpd_table.vpd_tags[tag_idx].tag_data, field_data, strlen(field_data));
                                vpd_table.vpd_tags[tag_idx].tag_data[strlen(field_data)]='\0';
                            } else {
                                /* Copy Binary data into tag_data */
                                token_ptr = strtok( field_data, " " );
                                for (i=0; (token_ptr != NULL); i++) {
                                    sscanf(token_ptr, "%2x", &val8);
                                    vpd_table.vpd_tags[tag_idx].tag_data[i] = (u8_t)val8;
                                    token_ptr = strtok( NULL, " ");
                                }
                                vpd_table.vpd_tags[tag_idx].tag_len = i;
                            }

                            /* Add to required size for VPD image this field length + 3 bytes (for TAG name, tag len) */
                            required_vpd_size += vpd_table.vpd_tags[tag_idx].tag_len + 3;
                            tag_idx ++;
                        }

                    } else {
                        sprintf (tmp_str, "VPD file incorrect format: in line [%s]\n", strtok( line, " \r\t\n"));
                        print_msg (tmp_str, ESC_COLOR_RED);
                        print_msg ("Should include field_name=field_type(Text/Binary), \"field_data\" - as well as newline at end of file\n", ESC_COLOR_RED);
                        return NULL;
                    }
                } else {
                    if (is_vpd_w) {
                        print_msg ("VPD-W fields are not supported\n", ESC_COLOR_RED);
                        return NULL;
                    }
                }
            }
        }
    }


    /* Verify there are number of fields > 0 in VPD text file*/
    if (tag_idx == 0) {
        print_msg ("VPD file parsing error: either empty or doesn't include [REGION=VPD-R] title\n", ESC_COLOR_RED);
        return NULL;
    }

    /* Verify that VPD text file includes a Product Name */
    if (!product_name_exist) {
        print_msg ("VPD file parsing error: Missing Product Name\n", ESC_COLOR_RED);
        return NULL;
    }

    /* Save the number of VPD tags */
    vpd_table.no_tags = tag_idx;
    //printf("Number of VPD TAGs is %d\n", vpd_table.no_tags);

    /* Verify the required size for all fields is less then "size" declared in the VPD text file 
       Note: 6 more bytes are required for RV tag */ 
    required_vpd_size += 6;
    DbgMessage(NULL, WARN, "vpd_table_size=%d required_vpd_size=%d\n", vpd_table_size, required_vpd_size);
    if (required_vpd_size > vpd_table_size) {
        sprintf (tmp_str, "VPD image size required (%d) is less than declared size in file (%d)\n", 
                 required_vpd_size, vpd_table_size);
        print_msg (tmp_str, ESC_COLOR_RED);
        return NULL;
    }

    vpd_image_ptr = create_vpd_image (&vpd_table , vpd_table_size, vpd_rev);
    *byte_read = vpd_table_size+4;

    fclose(fd);

    if ( vpd_image_ptr == NULL ) {
        sprintf (tmp_str,"Failed to create VPD image!\n");
        print_msg (tmp_str, ESC_COLOR_RED);
        return NULL;
    }

    return vpd_image_ptr;

}

/* tcl_al_parse_vpd_line 
 * 
 * Parse a line from VPD text file into three fields: 
 *                                      field_name, field_type, field_data
 * 
 * Ignore white spaces at the begining and ending of field if exit
 *
*/
int tcl_al_parse_vpd_line(char *line, char *vpd_field_name_ptr, char *vpd_field_type_ptr, char *vpd_field_data_ptr)
{
    char *token_ptr;
    char vpd_line[1024];

    // Find first occurance of the token '='
    strcpy(vpd_line, line);
    token_ptr = strtok( vpd_line, "=");
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }

    remove_white_spaces(token_ptr, strlen(token_ptr), vpd_field_name_ptr);

    // Find first occurance of the token ','
    token_ptr = strtok( NULL, "," );
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }

    remove_white_spaces(token_ptr, strlen(token_ptr), vpd_field_type_ptr);

    // Find first occurance of the token "
    token_ptr = strtok( NULL, "\"");
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }
    remove_white_spaces(token_ptr, strlen(token_ptr), vpd_field_data_ptr);

    // Find second occurance of the token ";
    token_ptr = strtok( NULL, "\"");
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }
    // Incase there are spcaes between , and " 
    if (strlen(vpd_field_data_ptr) == 0) {
        remove_white_spaces(token_ptr, strlen(token_ptr), vpd_field_data_ptr);
    }

    return TCL_OK;
}


/* tcl_al_get_vpd_r_size 
 * 
 * Get the VPD-R size from the REGION line 
 * 
 *
*/
int tcl_al_get_vpd_r_size(char *line, u32_t *size)
{
    char *token_ptr=NULL;
    char size_str[80];
    int tmp;

    // Find first occurance of the token '='
    token_ptr = strtok( line, "=");
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }

    // Find first occurance of the token ','
    token_ptr = strtok( NULL, "," );
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }

    // Find first occurance of the token "
    token_ptr = strtok( NULL, "]");
    if (token_ptr == NULL) {
        return TCL_ERROR;
    }

    remove_white_spaces(token_ptr, strlen(token_ptr), size_str);

    tmp = atoi(size_str);
    if (tmp < 0) {
        return TCL_ERROR;
    } else {
        *size = (u32_t) tmp;
        return TCL_OK;
    }
}


/*
 *------------------------------------------------------------------------
 * tcl_al_write_vpd_text_file: 
 * read the nvm image VPD, parse it and output to file
 * 
 * input: specify the name of the output file 
 * return: 
 *------------------------------------------------------------------------
 */

int tcl_al_write_vpd_text_file(char *filename, u8_t *buf, u32_t size)
{
    FILE        *fd=NULL;
    u32_t       file_size=0;
    u8_t        *ptr=NULL;
    char        tag_name[VPD_ID_TAG_MAX_LEN+1];
    char        tag_data[VPD_TAG_DATA_MAX_LEN+1];
    u8_t        tag_len;
    u16_t       len_remaining, product_name_len;
    char        err_str[80];
    vpd_image_t *vpd_image_ptr=NULL;
    u32_t        i, j;
    u32_t       *ptr32=NULL, val32;
    char        yb_data[3];

    if ((size % 4) != 0) {
        print_msg ("VPD file dumping error: VPD-R size should be a multiple of 4\n", ESC_COLOR_RED);
        return TCL_ERROR;
    }
    if (size > MAX_VPD_R_SIZE+8) {
        sprintf(err_str, "VPD file dumping error: VPD-R size exceed maximum allowed size %d\n", MAX_VPD_R_SIZE);
        print_msg (err_str, ESC_COLOR_RED);
        return TCL_ERROR;
    }

    vpd_image_ptr = (vpd_image_t *) buf;

    /* Swap revision bytes */
    vpd_image_ptr->format_revision = htonl(vpd_image_ptr->format_revision);

    /* Swap data bytes */
    for (i=0; i < size-4; i=i+4) {
        ptr32 = (u32_t *) &(vpd_image_ptr->vpd_data[i]);
        val32 = htonl(*ptr32);
        *ptr32 = val32;
    }

    /* Point to the begining of VPD data */
    ptr = (u8_t *) &(vpd_image_ptr->vpd_data[0]);

    /* For debugging */
    /*
    printf("VPD data:\n");
    for (i=0; i < 50; i++) {
        if (i%4==0) {
            printf("\t");
        }
        printf("%x", *(ptr+i));
    }
    printf("\n");
    */


    fd = fopen(filename, "a");
    if (fd) {
        /* If file already exists, exit do not allow to overwrite the file */
        fseek (fd, 0L, SEEK_END);
        file_size = ftell (fd);
        fseek (fd, 0L, SEEK_SET);      
        if (file_size > 0) {
            print_msg ("Dumping VPD image failed ! output file allready exists\n", ESC_COLOR_RED);
            return TCL_ERROR;
        }

        /* Verify the ID String TAG exist*/

        if (*ptr != (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_STRING)) {
            sprintf (err_str, "VPD file dumping error: Expected ID TAG 0x%x is missing (read 0x%x)\n", 
                     (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_STRING), *ptr);
            print_msg (err_str, ESC_COLOR_RED);
            fclose (fd);
            return TCL_ERROR;
        }
        ptr++;

        fprintf(fd, "#############################################################################\n");
        fprintf(fd, ";                                                                            \n");
        fprintf(fd, ";  Example VPD text file                                                     \n");
        fprintf(fd, ";                                                                            \n");
        fprintf(fd, ";                                                                            \n"); 
        fprintf(fd, ";  Notes:                                                                    \n");
        fprintf(fd, ";  1. Each Comment line starts with either \"#\" or ;                        \n");
        fprintf(fd, ";  2. Text file include VPD-R, VPD-W regions                                 \n");
        fprintf(fd, ";  3. Each region starts with the title [REGION=VPD-R] or [REGION=VPD-W]     \n");
        fprintf(fd, ";  4. Each line includes: Field name=Field_type (Text/Binary), \"Field data\"\n");
        fprintf(fd, ";  5. Supports both upper and lower case.                                    \n");
        fprintf(fd, "#############################################################################\n\n");

        fprintf(fd, "# Revision for future enhancements. Currently, only revision 1 is supported  \n");
        fprintf(fd, "Revision = %d\n\n", vpd_image_ptr->format_revision);

        /* Size is according to image size - 8 
          (4 bytes for image revision, 4 bytes for CRC*/
        fprintf(fd, "[REGION=VPD-R, %d]\n", size-8);

        /* Get the product name length */
        cmn_os_memcpy(&product_name_len, ptr, 2);
        ptr += 2;

        /* Get the product name */
        tag_data[0]='\0';
        strncpy(tag_data, (char *) ptr, product_name_len); 
        tag_data[product_name_len]='\0';
        fprintf(fd, "Product Name=Text,\"%s\"\n", tag_data);
        ptr += product_name_len;

        /* Verify the VPD-R TAG exist */    
        if (*ptr != (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_VPD_R)) {
            sprintf (err_str, "VPD file dumping error: Expected VPD-R TAG 0x%x is missing (read 0x%x)\n", 
                     (TAG_LARGE_RESOURCE | TAG_DATA_TYPE_LR_VPD_R), *ptr);
            print_msg (err_str, ESC_COLOR_RED);
            fclose (fd);
            return TCL_ERROR;
        }
        ptr ++;

        /* Read the remaining length */
        cmn_os_memcpy(&len_remaining, ptr, 2);
        ptr += 2;

        /* Verify remaining length does not exceed image size */
        if (len_remaining > size-8) {
            sprintf (err_str, "VPD file dumping error: VPD-R length (0x%x) > image size (0x%x)\n", len_remaining, size-8);
            print_msg (err_str, ESC_COLOR_RED);
            fclose (fd);
            return TCL_ERROR;
        }

        /* Read the VPD fields */
        while (len_remaining > 0) {
            /* Copy the TAG name (e.g. PN, EC, MN ..) */
            strncpy(tag_name, (char *)ptr, 2);
            tag_name[2]='\0';

            /* Increment the data ptr by 2 (since TAG name is 2 bytes) */
            ptr += 2;

            /* Copy the TAG Length */
            tag_len = *ptr;
            ptr++;            

            if (strcmp(tag_name, "RV") != 0) {
                /* Get the TAG data */    
                if (strcmp(tag_name, "YB") != 0) {                    
                    strncpy(tag_data, (char *) ptr, tag_len);
                    tag_data[tag_len]='\0';
                    ptr += tag_len;
                    fprintf(fd, "%s=Text,\"%s\"\n", tag_name, tag_data);

                } else {
    				/* YB tag is binary */
                    strcpy(tag_data, "");
                    for (j=0; j < tag_len; j++) {                        
                        sprintf(yb_data, "%02x", *ptr);
                        strcat(tag_data, yb_data);
                        ptr++;
                        if (j != (tag_len-1)) {
                            strcat(tag_data, " ");
                        }
                    }
                    fprintf(fd, "%s=Binary,\"%s\"\n", tag_name, tag_data);                                  
                }
                len_remaining -= (tag_len + 3);

            } else {
                /* If got to RV tag, exit */
                break;
            }

        }

        printf("VPD image was parsed succesfully into %s\n", filename);

        fclose(fd);
        return TCL_OK;
    } else {
        return TCL_ERROR;
    }

}



