/****************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved
 * Proprietary and Confidential Information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of QLogic Corporation. 
 *
 * Name:        hmacsha1.c
 *
 * Description: Compute a secure hash using HMAC-SHA1.
 *  
 * 
 ****************************************************************************/

#ifdef BOOT_CODE_ONLY
#include "bc_common.h"
#endif
#ifdef USER_LINUX
#include <sys/types.h>
#endif
#include "bcmtype.h"
#include "dg_os_util.h"



#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

/* local function defines */
#define ROTL(n,X)               ( ( ( X ) << (n) ) | ( ( X ) >> ( 32 - (n) ) ) ) /* Rotate variable 'X' to the left by 'n' bits */


/******************************************************************
    SHA1 constent/structure/function defines
******************************************************************/
#define SHA1_DIGESTSIZE  20  /* 160-bits */
#define SHA1_BLOCKSIZE   64  /* process 64 bytes at a time */

typedef struct {
    u32_t digest[5];        /* Message digest */
    u32_t countLo, countHi; /* 64-bit bit count */
    u32_t data[16];         /* SHA1 data buffer */
    u32_t Endianness;
} SHA1_CTX;

/* The SHA1 f()-functions. */
#define f1(x,y,z)   ( z ^ ( x & ( y ^ z ) ) )           /* Rounds  0-19 */
#define f2(x,y,z)   ( x ^ y ^ z )                       /* Rounds 20-39 */
#define f3(x,y,z)   ( ( x & y ) | ( z & ( x | y ) ) )   /* Rounds 40-59 */
#define f4(x,y,z)   ( x ^ y ^ z )                       /* Rounds 60-79 */
/* The SHA1 Mysterious Constants */
#define K1          0x5A827999L                         /* Rounds  0-19 */
#define K2          0x6ED9EBA1L                         /* Rounds 20-39 */
#define K3          0x8F1BBCDCL                         /* Rounds 40-59 */
#define K4          0xCA62C1D6L                         /* Rounds 60-79 */
/* SHA1 initial values */
#define h0init      0x67452301L
#define h1init      0xEFCDAB89L
#define h2init      0x98BADCFEL
#define h3init      0x10325476L
#define h4init      0xC3D2E1F0L

/******************************************************************
    routine prototype defines 
******************************************************************/
void        hmac_sha1 ( u8_t* k, u32_t lk, u8_t* d, u32_t ld, u8_t* out, u32_t t);


static void  SHA1Init(SHA1_CTX *);
static u32_t fn_k(u32_t t);
static u32_t fn_f(u32_t t, u32_t x, u32_t y, u32_t z);
static void  SHA1Transform( u32_t *digest, u32_t *data );
static void  SHA1Update(SHA1_CTX *, u8_t *buffer, u32_t count);
static void  SHA1Final(u8_t *output, SHA1_CTX *);
static void  SHA1toByte(u8_t *output, u32_t *input, u32_t len);


#ifdef BIG_ENDIAN
#define longReverse(buffer, byteCount, Endianness)
#define endianTest(endianness)
#else
/******************************************************************
    When run on a little-endian CPU we need to perform byte
    reversal on an array of long words.
******************************************************************/
static void longReverse(u32_t *buffer, u32_t byteCount, u32_t Endianness )
{
    u32_t value;

    if (Endianness==TRUE) return;
    byteCount /= sizeof( u32_t );
    while( byteCount-- )
        {
        value = *buffer;
        value = ( ( value & 0xFF00FF00L ) >> 8  ) | ( ( value & 0x00FF00FFL ) << 8 );
        *buffer++ = ( value << 16 ) | ( value >> 16 );
        }
}

/******************************************************************
    Find out if the code is running on little or big endian
******************************************************************/
static void endianTest(u32_t *endian_ness)
{
    u32_t n=0x01000000;
    *endian_ness = *(u8_t *)&n;
}

#endif



/******************************************************************
    Initialize the SHA1 values *
******************************************************************/
static void SHA1Init(SHA1_CTX *shactx)
{
    endianTest(&shactx->Endianness);
    /* Set the h-vars to their initial values */
    shactx->digest[ 0 ] = h0init;
    shactx->digest[ 1 ] = h1init;
    shactx->digest[ 2 ] = h2init;
    shactx->digest[ 3 ] = h3init;
    shactx->digest[ 4 ] = h4init;

    /* Initialise bit count */
    shactx->countLo = shactx->countHi = 0;
}

/******************************************************************
    SHA-1 K value selector
******************************************************************/
static u32_t fn_k(u32_t t)
{
    if (t < 20)
        return (K1);

    if (t < 40)
        return (K2);

    if (t < 60)
        return (K3);
    else
        return (K4);
}

/******************************************************************
    SHA-1 Function selector
******************************************************************/
static u32_t fn_f(u32_t t, u32_t x, u32_t y, u32_t z)
{
    if (t < 20)
        return f1(x,y,z);

    if (t < 40)
        return f2(x,y,z);

    if (t < 60)
        return f3(x,y,z);
    else
        return f4(x,y,z);
}

/******************************************************************
    This is where number crunching is happening.  See FIP180-1 for
    more details.
******************************************************************/
static void SHA1Transform( u32_t *digest, u32_t *data )
{
    u32_t t,s,temp, A, B, C, D, E;
    u32_t w[ 16 ];                  /* Expanded data */

    /* Set up first buffer and local data buffer */
    A = digest[ 0 ];
    B = digest[ 1 ];
    C = digest[ 2 ];
    D = digest[ 3 ];
    E = digest[ 4 ];

    dg_os_memcpy( (u8_t *)w, (u8_t *)data, SHA1_BLOCKSIZE );

    for (t=0; t<80; t++)
    {
        s = t & 0xf;

        if (t >= 16)
        {
            w[s] = ROTL(1, (w[(s+13)&0xf]) ^ (w[(s+8)&0xf]) ^ (w[(s+2)&0xf]) ^ (w[s]) );
        }

        temp = ROTL(5, A) + fn_f(t, B, C, D) + E + w[s] + fn_k(t);

        E = D;
        D = C;
        C = ROTL(30, B);
        B = A;
        A = temp;
    }

    /* Build message digest */
    digest[ 0 ] += A;
    digest[ 1 ] += B;
    digest[ 2 ] += C;
    digest[ 3 ] += D;
    digest[ 4 ] += E;
}

/******************************************************************
    Update SHA1 for a block of data
******************************************************************/
static void SHA1Update(SHA1_CTX *shactx, u8_t *buffer, u32_t count)
{
    u32_t tmp;
    u32_t dataCount;

    /* Update bitcount */
    tmp = shactx->countLo;
    if ( ( shactx->countLo = tmp + ( ( u32_t ) count << 3 ) ) < tmp )
        shactx->countHi++;             /* Carry from low to high */
    shactx->countHi += count >> 29;

    /* Get count of bytes already in data */
    dataCount = (u32_t)(tmp >> 3) & 0x3F;

    /* Handle any leading odd-sized chunks */
    if( dataCount )
        {
        u8_t *p = ( u8_t * ) shactx->data + dataCount;

        dataCount = SHA1_BLOCKSIZE - dataCount;
        if( count < dataCount )
            {
            dg_os_memcpy( p, buffer, count );
            return;
            }
        dg_os_memcpy( p, buffer, dataCount );
        longReverse( shactx->data, SHA1_BLOCKSIZE, shactx->Endianness);
        SHA1Transform( shactx->digest, shactx->data );
        buffer += dataCount;
        count -= dataCount;
        }

    /* Process data in SHA1_BLOCKSIZE chunks */
    while( count >= SHA1_BLOCKSIZE )
        {
        dg_os_memcpy( (u8_t *)shactx->data, (u8_t *)buffer, SHA1_BLOCKSIZE );
        longReverse( shactx->data, SHA1_BLOCKSIZE, shactx->Endianness );
        SHA1Transform( shactx->digest, shactx->data );
        buffer += SHA1_BLOCKSIZE;
        count -= SHA1_BLOCKSIZE;
        }

    /* Handle any remaining bytes of data. */
    dg_os_memcpy( (u8_t *)shactx->data, (u8_t *)buffer, count );
    }

/*  */

/******************************************************************
    Final wrapup - pad to SHA1_BLOCKSIZE-byte boundary with the bit
    pattern 1 0* (64-bit count of bits processed, MSB-first)
******************************************************************/
static void SHA1Final(u8_t *output, SHA1_CTX *shactx)
{
    u32_t count;
    u8_t *dataPtr;

    /* Compute number of bytes mod 64 */
    count = ( u32_t ) shactx->countLo;
    count = ( count >> 3 ) & 0x3F;

    /* Set the first char of padding to 0x80.  This is safe since there is
       always at least one byte free */
    dataPtr = ( u8_t * ) shactx->data + count;
    *dataPtr++ = 0x80;

    /* Bytes of padding needed to make 64 bytes */
    count = SHA1_BLOCKSIZE - 1 - count;

    /* Pad out to 56 mod 64 */
    if( count < 8 )
        {
        /* Two lots of padding:  Pad the first block to 64 bytes */
        dg_os_memset( dataPtr, 0, count );
        longReverse( shactx->data, SHA1_BLOCKSIZE, shactx->Endianness );
        SHA1Transform( shactx->digest, shactx->data );

        /* Now fill the next block with 56 bytes */
        dg_os_memset( (u8_t *)shactx->data, 0, SHA1_BLOCKSIZE - 8 );
        }
    else
        /* Pad block to 56 bytes */
        dg_os_memset( dataPtr, 0, count - 8 );

    /* Append length in bits and transform */
    shactx->data[ 14 ] = shactx->countHi;
    shactx->data[ 15 ] = shactx->countLo;

    longReverse( shactx->data, SHA1_BLOCKSIZE - 8, shactx->Endianness );
    SHA1Transform( shactx->digest, shactx->data );

    /* Output to an array of bytes */
    SHA1toByte(output, shactx->digest, SHA1_DIGESTSIZE);

    /* Zeroise sensitive stuff */
    dg_os_memset((u8_t *)shactx, 0, sizeof(shactx));
}

/******************************************************************
    Output SHA1 digest in byte array
******************************************************************/
static void SHA1toByte(u8_t *output, u32_t *input, u32_t len)
{    /*  */
    u32_t i, j;

    for(i = 0, j = 0; j < len; i++, j += 4) 
    {
        output[j+3] = (u8_t)( input[i]        & 0xff);
        output[j+2] = (u8_t)((input[i] >> 8 ) & 0xff);
        output[j+1] = (u8_t)((input[i] >> 16) & 0xff);
        output[j  ] = (u8_t)((input[i] >> 24) & 0xff);
    }
}

/******************************************************************
    Function to compute the HMAC-SHA1 digest
******************************************************************/
/*  */
void
hmac_sha1 ( 
u8_t*    k,     /* secret key */
u32_t      lk,    /* length of the key in bytes */
u8_t*    d,     /* data */
u32_t      ld,    /* length of data in bytes */
u8_t*    out,   /* output buffer, at least "t" bytes */
u32_t      t      /* output buffer size */
)
{
    SHA1_CTX ctx;
    u8_t    isha[SHA1_DIGESTSIZE], osha[SHA1_DIGESTSIZE];
    u8_t    key[SHA1_DIGESTSIZE];
    u8_t    buf[SHA1_BLOCKSIZE];
    u32_t     i;

    if (lk > SHA1_BLOCKSIZE) {
        SHA1Init(&ctx);
        SHA1Update(&ctx, k, lk);
        SHA1Final(key, &ctx);
        k = key;
        lk = SHA1_DIGESTSIZE;
    }

    /* Pad the key for inner digest */
    for (i = 0 ; i < lk ; ++i) buf[i] = k[i] ^ 0x36;
    for (i = lk ; i < SHA1_BLOCKSIZE ; ++i) buf[i] = 0x36;

    /**** Inner Digest ****/
    SHA1Init(&ctx);
    SHA1Update(&ctx, buf, SHA1_BLOCKSIZE);
    SHA1Update(&ctx, d, ld);
    SHA1Final(isha, &ctx);

    /* Pad the key for outter digest */
    for (i = 0 ; i < lk ; ++i) buf[i] = k[i] ^ 0x5C;
    for (i = lk ; i < SHA1_BLOCKSIZE ; ++i) buf[i] = 0x5C;

    /**** Outter Digest ****/
    SHA1Init(&ctx);
    SHA1Update(&ctx, buf, SHA1_BLOCKSIZE);
    SHA1Update(&ctx, isha, SHA1_DIGESTSIZE);
    SHA1Final(osha, &ctx);

    dg_os_memcpy (out, osha, t);
}


