
 /******************************************************************************
 * The information contained in this file is confidential and proprietary to
 * QLogic Corporation.  No part of this file may be reproduced or
 * distributed, in any form or by any means for any purpose, without the
 * express written permission of QLogic Corporation.
 *
 * (c) COPYRIGHT 2011-2014 Broadcom Corporation, ALL RIGHTS RESERVED.
 * (c) COPYRIGHT 2014-2015 QLogic Corporation, ALL RIGHTS RESERVED.
 *
 *
 * Module Description:
 *
 *
 ******************************************************************************/
#include "man_types.h"
#include "dbg_os_uart.h"
#include "dbg_os_utils.h"
#include "hw_access.h"

#define NVRAM_TIMEOUT_COUNT                     30000
#define DUART_FUNCTION_NUMBER                   0  /* When working over DUART we'll use function 0 */

#define DUART_CHIP_REV      nvm_dev_info.chip_id
#define DUART_NVM_PAGE_SZ   nvm_dev_info.nvm_page_size
#define DUART_NVM_SIZE      nvm_dev_info.nvm_size


#define MAX_RETRIES 5
#define WAIT_MS     500


/*********************** Globals *********************/
static struct {
    u32_t chip_id;
    u32_t nvm_size;
    u32_t nvm_page_size;
} nvm_dev_info = {0};


/********************* Access functions *************/

__inline static int duart_reg_rd(
    IN   ua_handle_t h,
    IN   u32_t       reg_offset,
    OUT  u32_t     * val_p
  )
{
   ua_status_t stat;
   u8_t i;

   for (i=0; i< MAX_RETRIES; i++) { 
       stat = dbg_os_duart_read(h, reg_offset, val_p);
       if (stat==UA_STATUS_OK) break;
       dbg_msleep(WAIT_MS);
   }

   if ( stat != UA_STATUS_OK ) {
       printf("ERROR - duart_reg_rd\n");
   }

   return stat;
}

__inline static int duart_reg_wr(
    IN   ua_handle_t h,
    IN   u32_t       reg_offset,
    IN   u32_t       val
  )
{
   ua_status_t stat;
   u8_t i;

   for (i=0; i< MAX_RETRIES; i++) { 
       stat = dbg_os_duart_write(h, reg_offset, val, 1);    //wait_for_echo is set to 1
       if (stat==UA_STATUS_OK) break;      
       dbg_msleep(WAIT_MS);
    }

    if ( stat != UA_STATUS_OK ) {
       printf("ERROR - duart_reg_wr\n");
    }

    return stat;
}

/*******************************************************************************
 * Description:
 *
 * Return:
 ******************************************************************************/
static dbg_status_t
acquire_nvram_lock(
    ua_handle_t h
    )
{
    dbg_status_t status;
    u32_t j, cnt;
    u32_t val;

    /* Adjust timeout for emulation/FPGA */
    cnt = NVRAM_TIMEOUT_COUNT;
    if (DUART_CHIP_REV == CHIP_REV_EMUL) cnt *= 100;

    val = 0;

    /* Request access to the flash interface. */
    if (duart_reg_wr(h, MCP_REG_MCPR_NVM_SW_ARB, MCPR_NVM_SW_ARB_ARB_REQ_SET1 << DUART_FUNCTION_NUMBER)) {
        return DBG_STATUS_FAILURE;
    }
    
    for(j = 0; j < cnt*10; j++)
    {
        if (duart_reg_rd(h, MCP_REG_MCPR_NVM_SW_ARB,&val)) {
            return DBG_STATUS_FAILURE;
        }
        if(val & (MCPR_NVM_SW_ARB_ARB_ARB1 << DUART_FUNCTION_NUMBER))
        {
            break;
        }

        os_if_usleep(5);
    }

    if(val & (MCPR_NVM_SW_ARB_ARB_ARB1 << DUART_FUNCTION_NUMBER))
    {
        status = DBG_STATUS_SUCCESS;
    }
    else
    {
        printf("Cannot get access to nvram interface\n");
        status = DBG_STATUS_FAILURE;
    }

    return status;
} /* acquire_nvram_lock */



/*******************************************************************************
 * Description:
 *
 * Return:
 ******************************************************************************/
static void
release_nvram_lock(
    ua_handle_t h)
{
    u32_t j, cnt;
    u32_t val;

	/* Adjust timeout for emulation/FPGA */
	cnt = NVRAM_TIMEOUT_COUNT;
	if (DUART_CHIP_REV == CHIP_REV_EMUL) cnt *= 100;

    /* Relinquish nvram interface. */
    duart_reg_wr(h, MCP_REG_MCPR_NVM_SW_ARB, (MCPR_NVM_SW_ARB_ARB_REQ_CLR1 << DUART_FUNCTION_NUMBER));

    val = 0;

    for(j = 0; j < cnt; j++)
    {
		duart_reg_rd(h, MCP_REG_MCPR_NVM_SW_ARB,&val);
		if(!(val & (MCPR_NVM_SW_ARB_ARB_ARB1 << DUART_FUNCTION_NUMBER)))
        {
            break;
        }

        os_if_usleep( 5);
    }

    if (val & (MCPR_NVM_SW_ARB_ARB_ARB1 << DUART_FUNCTION_NUMBER)) {
        printf("ERROR - release_nvram_lock\n");
    }
} /* release_nvram_lock */


/*******************************************************************************
* Description:
*
* Return:
******************************************************************************/
static dbg_status_t
enable_nvram_access(
	ua_handle_t h)
{
	u32_t val;

	duart_reg_rd(h, MCP_REG_MCPR_NVM_ACCESS_ENABLE, &val);

	/* Enable both bits, even on read. */
	duart_reg_wr(h, MCP_REG_MCPR_NVM_ACCESS_ENABLE, val | MCPR_NVM_ACCESS_ENABLE_EN | MCPR_NVM_ACCESS_ENABLE_WR_EN);

	return DBG_STATUS_SUCCESS;
} /* enable_nvram_access */



/*******************************************************************************
* Description:
*
* Return:
******************************************************************************/
static dbg_status_t
disable_nvram_access(
	ua_handle_t h)
{
	u32_t val;

	duart_reg_rd(h, MCP_REG_MCPR_NVM_ACCESS_ENABLE, &val);

	/* Disable both bits, even after read. */
	duart_reg_wr(h, MCP_REG_MCPR_NVM_ACCESS_ENABLE, val & ~(MCPR_NVM_ACCESS_ENABLE_EN | MCPR_NVM_ACCESS_ENABLE_WR_EN));

	return DBG_STATUS_SUCCESS;
} /* disable_nvram_access */


/*******************************************************************************
 * Description:
 *
 * Return:
 ******************************************************************************/
static dbg_status_t
nvram_read_dword(
    ua_handle_t h,
    u32_t offset,
    u32_t *ret_val,
    u32_t nvram_flags)
{
    dbg_status_t status;
    u32_t cmd_flags;
    u32_t val;
    u32_t j, cnt;

    /* Build the command word. */
    cmd_flags = nvram_flags | MCPR_NVM_COMMAND_DOIT ;

    /* Need to clear DONE bit separately. */
    duart_reg_wr(h, MCP_REG_MCPR_NVM_COMMAND, MCPR_NVM_COMMAND_DONE);

    /* Address of the NVRAM to read from. */
	duart_reg_wr(h, MCP_REG_MCPR_NVM_ADDR, offset & MCPR_NVM_ADDR_NVM_ADDR_VALUE);

    /* Issue a read command. */
	duart_reg_wr(h, MCP_REG_MCPR_NVM_COMMAND, cmd_flags);

    /* Adjust timeout for emulation/FPGA */
	cnt = NVRAM_TIMEOUT_COUNT;
	if (DUART_CHIP_REV == CHIP_REV_EMUL) cnt *= 100;

    /* Wait for completion. */
    status = DBG_STATUS_FAILURE;
    for(j = 0; j < cnt; j++)
    {
        os_if_usleep( 5);
		duart_reg_rd(h, MCP_REG_MCPR_NVM_COMMAND, &val);
        if(val & MCPR_NVM_COMMAND_DONE)
        {
			duart_reg_rd(h, MCP_REG_MCPR_NVM_READ, &val);

            /* Change to little endian. */
            #if defined(LITTLE_ENDIAN)
            val = ((val & 0xff) << 24) | ((val & 0xff00) << 8) |
                ((val & 0xff0000) >> 8) | ((val >> 24) & 0xff);
            #endif

            *ret_val = val;

            status = DBG_STATUS_SUCCESS;

            break;
        }
    }

    return status;
} /* nvram_read_dword */



/*******************************************************************************
 * Description:
 *
 * Return:
 ******************************************************************************/
static dbg_status_t
nvram_write_dword(
    ua_handle_t h,
    u32_t offset,
    u32_t val,
    u32_t nvram_flags)
{
    dbg_status_t status;
    u32_t cmd_flags;
    u32_t j, cnt;

    /* Build the command word. */
    cmd_flags = nvram_flags | MCPR_NVM_COMMAND_DOIT | MCPR_NVM_COMMAND_WR;

    /* Change to little endian. */
    #if defined(LITTLE_ENDIAN)
    val = ((val & 0xff) << 24) | ((val & 0xff00) << 8) |
        ((val & 0xff0000) >> 8) | ((val >> 24) & 0xff);
    #endif

    /* Need to clear DONE bit separately. */
    duart_reg_wr(h, MCP_REG_MCPR_NVM_COMMAND, MCPR_NVM_COMMAND_DONE);

    /* Write the data. */
	duart_reg_wr(h, MCP_REG_MCPR_NVM_WRITE, val);

    /* Address of the NVRAM to write to. */
	duart_reg_wr(h, MCP_REG_MCPR_NVM_ADDR, offset & MCPR_NVM_ADDR_NVM_ADDR_VALUE);

    /* Issue the write command. */
    duart_reg_wr(h, MCP_REG_MCPR_NVM_COMMAND, cmd_flags);

    /* Adjust timeout for emulation/FPGA */

	cnt = NVRAM_TIMEOUT_COUNT;
	if (DUART_CHIP_REV == CHIP_REV_EMUL) cnt *= 100;

    /* Wait for completion. */
    status = DBG_STATUS_FAILURE;
    for(j = 0; j < cnt; j++)
    {
        os_if_usleep( 5);

        duart_reg_rd(h, MCP_REG_MCPR_NVM_COMMAND, &val);
        if(val & MCPR_NVM_COMMAND_DONE)
        {
            status = DBG_STATUS_SUCCESS;
            break;
        }
    }

    return status;
} /* nvram_write_dword */



/************** External interfaces ********************************************/


/*******************************************************************************
 * Description:
 *
 * Return:
 ******************************************************************************/
dbg_status_t
dbg_dev_duart_nvm_read(
    ua_handle_t h,
    u32_t offset,
    u32_t *ret_buf,
    u32_t buf_size
    )
{
    dbg_status_t status;
    u32_t cmd_flags;

   
    if((buf_size & 0x03) || (offset & 0x03))
    {
        printf("Invalid paramter1.\n");

        return DBG_STATUS_FAILURE;
    }

	// TODO what is the nvram total size
    if( offset + buf_size > DUART_NVM_SIZE && 0)
    {
        printf("Invalid paramter2.\n");

        return DBG_STATUS_FAILURE;
    }

    /* Request access to the flash interface. */
    status = acquire_nvram_lock(h);
    if(status != DBG_STATUS_SUCCESS)
    {
        return status;
    }


	/* Enable access to flash interface */
	status = enable_nvram_access(h);
	if(status != DBG_STATUS_SUCCESS)
	{
        release_nvram_lock(h);
		return status;
	}

	/* Read the first word. */
	cmd_flags = MCPR_NVM_COMMAND_FIRST;
    while(buf_size > sizeof(u32_t) && status == DBG_STATUS_SUCCESS)
    {
        status = nvram_read_dword(h, offset, ret_buf, cmd_flags);

        /* Advance to the next dword. */
        offset += sizeof(u32_t);
        ret_buf++;
        buf_size -= sizeof(u32_t);
		cmd_flags = 0;
    }

    if(status == DBG_STATUS_SUCCESS)
    {
        cmd_flags |= MCPR_NVM_COMMAND_LAST;
        status = nvram_read_dword(h, offset, ret_buf, cmd_flags);
    }

	/* Disable access to flash interface */
	disable_nvram_access(h);

    release_nvram_lock(h);   

    return status;
} /* lm_nvram_read */


/*******************************************************************************
 * Description:
 *
 * Return:
 ******************************************************************************/
dbg_status_t dbg_dev_duart_nvm_write(
    ua_handle_t h,
    u32_t offset,
    u32_t *data_buf,
    u32_t buf_size
    )
{
    dbg_status_t status;
    u32_t cmd_flags;
    u32_t written_so_far;
    u32_t *ptr32, addr;


	if(offset & 0x03)
    {
        printf("Invalid paramter.\n");

        return DBG_STATUS_FAILURE;
    }
	// TODO what is the nvram total size
    if( offset + buf_size > DUART_NVM_SIZE )
    {
        printf("Invalid paramter.\n");

        return DBG_STATUS_FAILURE;
    }

    status = DBG_STATUS_SUCCESS;

    /* Request access to the flash interface. */
    status = acquire_nvram_lock(h);
    if(status != DBG_STATUS_SUCCESS)
        return status;

    /* Enable access to flash interface */
    status = enable_nvram_access(h);
    if(status != DBG_STATUS_SUCCESS)
    {
        release_nvram_lock(h);
        return status;
    }

    written_so_far = 0;
    cmd_flags = MCPR_NVM_COMMAND_FIRST;
    addr = offset;
    ptr32 = data_buf;
    while (written_so_far < buf_size)
    {
        if (written_so_far == (buf_size - 4))
            cmd_flags |= MCPR_NVM_COMMAND_LAST;
        else if (((addr & 0xff) + 4) == NVRAM_PAGE_SIZE)        // else if (((addr + 4) % NVRAM_PAGE_SIZE) == 0)
            cmd_flags |= MCPR_NVM_COMMAND_LAST;
        else if ((addr & 0xff) == 0)                            // else if ((addr % NVRAM_PAGE_SIZE) == 0)
            cmd_flags |= MCPR_NVM_COMMAND_FIRST;
        nvram_write_dword(h, addr, *ptr32, cmd_flags);
        ptr32++;
        addr += 4;
        written_so_far += 4;
        cmd_flags = 0;
    }
    /* Disable access to flash interface */
    disable_nvram_access(h);
    release_nvram_lock(h);


    return status;

} /* lm_nvram_write */



void dbg_dev_duart_nvm_set_dev_info(
    IN    u32_t chip_id,
    IN    u32_t nvm_size,
    IN    u32_t nvm_page_size
    )
{
    nvm_dev_info.chip_id       = chip_id;
    nvm_dev_info.nvm_page_size = nvm_page_size;
    nvm_dev_info.nvm_size      = nvm_size;

    printf("chip_id=0x%x, nvm_page_size=%d, nvm_size=%d\n", chip_id, nvm_page_size, nvm_size);
}
