#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "tcl.h"
#include "bcmtype.h"
#include "edebug_types.h"
#include "dbg_os_utils.h"
#include "dbg_hmem_def.h"
#include "util.h"

#define ADDR_TABLE_SIZE_DEFAULT 3

/* global vars */
static u32_t addr_table_size;
static alloc_entry_t *addr_table;

static void clean_entry ( alloc_entry_t *entry );
static u32_t find_empty_entry ( alloc_entry_t **table_ptr );


dbg_status_t dbg_hmem_read(
    IN   u32_t low_addr_val, 
    IN   u32_t high_addr_val,
    IN   access_method_t method,
    OUT  u32_t * val_p
    )
{
    u32_t mask = DWORD_ADDR_MASK32;
    volatile u16_t   *ptr16 = NULL; 
    volatile u8_t    *ptr8  = NULL;
    volatile u32_t   *ptr32 = NULL;
    volatile void  *void_ptr= NULL;


    switch (method) {
    case ACCESS_8BIT:
        mask = 0xffffffff;
        break;
    case ACCESS_16BIT:
        mask = WORD_ADDR_MASK32;
        break;
    default: ;
    }

    void_ptr = (volatile void *)((int_ptr_t)LONG_ADDR(low_addr_val & mask, high_addr_val));

    /* Get the memory content */
    switch (method) 
    {
        case ACCESS_8BIT:
            ptr8 = (volatile u8_t *)void_ptr;
            *val_p = (u32_t)(*ptr8);
            break;
        case ACCESS_16BIT:
            ptr16 = (volatile u16_t *)void_ptr;
            *val_p = (u32_t)(*ptr16);
            break;
        case ACCESS_32BIT:
            ptr32 = (volatile u32_t *)void_ptr;
            *val_p = *ptr32;
            break;
        default:
            return DBG_STATUS_INVALID_PARAMETER;
    }

    return DBG_STATUS_SUCCESS;
}

dbg_status_t dbg_hmem_alloc_mem(
    IN    u32_t nbytes, 
    IN    char* descr_str, 
    IN    u32_t memtype, 
    OUT   u32_t * low_addr_p,
    OUT   u32_t * high_addr_p
    )
{
    dbg_virt_addr_t actual_ptr;
    dbg_phys_addr_t phys_ptr;
    u32_t i, entry, *u32ptr, actual_nbytes;
    u32_t align_diff;
    char local_buf[MAX_HMEM_DESC_STR_SIZE+2];
    dbg_status_t status;

    *high_addr_p = 0;

    if ((memtype != MEM_VIRT) && (memtype != MEM_PHYS))
    {
        return DBG_STATUS_INVALID_PARAMETER;
    }

    
    /* Find an empty entry from table */
    entry = find_empty_entry(&addr_table);
    if (entry == -1)
    {
        return DBG_STATUS_RESOURCE;
    }

    /* Allocate the memory */
    /* Ensure enough memory address to fulfill two requirements: 
     * 64-bit aligned address and multiple of four bytes, that is 8 bytes aligned. */
    actual_nbytes = ((nbytes+7)&0xfffffff8)+8;
    if (memtype == MEM_VIRT)
    {
        actual_ptr = (dbg_virt_addr_t)(uintptr_t)malloc(actual_nbytes);
    }
    else
    {
        /* Physical memory wanted */
        status = dbg_os_alloc_pcontig(actual_nbytes, &actual_ptr, &phys_ptr);
        if (status != DBG_STATUS_SUCCESS)
        {
            actual_ptr = 0;
        }
    }

    if (actual_ptr == 0)
    {
        return DBG_STATUS_RESOURCE;
    }

    /* Check if the address is 64-bit aligned, if so, move the pointer. */
    align_diff = (0x8 - ((u32_t)(actual_ptr) & 0x7)) & 0x7;

    /* Append it */
    addr_table[entry].actual_vaddr = actual_ptr;
    addr_table[entry].aligned_vaddr = actual_ptr + align_diff;
    /* nbytes still presents what the user thinks it should be, 
     * not 8 bytes more. */
    addr_table[entry].nbytes = nbytes;
    addr_table[entry].actual_nbytes = actual_nbytes;
    addr_table[entry].mem_type = memtype;


    if (descr_str == NULL)
    {
        /* Make up a name for it */
        sprintf(local_buf, "tmp%04x", entry);
        descr_str = local_buf;
    }
    strncpy(addr_table[entry].desc_str, descr_str, MAX_HMEM_DESC_STR_SIZE);
    if (memtype == MEM_PHYS)
    {
        addr_table[entry].actual_paddr = (dbg_phys_addr_t)phys_ptr;
        addr_table[entry].aligned_paddr = (dbg_phys_addr_t)(phys_ptr + align_diff);
        //addr_table[entry].handle = handle;
    }

    /* Initialize memory to some funny pattern, used for debug only */
    for (i=0, u32ptr=(u32_t *)(uintptr_t)addr_table[entry].aligned_vaddr; i < nbytes; i+=4, u32ptr++)
    {
        *u32ptr = 0xdeadbeef;
    }

    /* Set the result */
    *low_addr_p  = U64_LO(actual_ptr + align_diff);
    *high_addr_p = U64_HI(actual_ptr + align_diff);

    return DBG_STATUS_SUCCESS;
}



dbg_status_t dbg_hmem_free(
    IN   u32_t low_addr_val, 
    IN   u32_t high_addr_val
    )
{
    u32_t i;
    dbg_virt_addr_t virt_addr;
    dbg_status_t dbg_result;


    if ( ( low_addr_val == 0 ) && ( high_addr_val == 0 ) ) {
        return DBG_STATUS_INVALID_PARAMETER;
    }

    virt_addr = (dbg_virt_addr_t)LONG_ADDR(low_addr_val, high_addr_val);

    /* Find an matching entry from table */
    for ( i = 0; i < addr_table_size; i++ )
    {
        if ( addr_table[i].aligned_vaddr == virt_addr ) break;
    }

    if ( i == addr_table_size )
    {
        /* None is found, return error */
        return DBG_STATUS_INVALID_PARAMETER;
    }

    /* Free the memory according to the obtained address */
    if (addr_table[i].mem_type == MEM_PHYS)
    {
        dbg_result = dbg_os_free_pcontig(addr_table[i].actual_vaddr);
        if ( dbg_result != DBG_STATUS_SUCCESS ) {
            return dbg_result;
        }
    }
    else if (addr_table[i].mem_type == MEM_VIRT)
    {
        free((char *)(uintptr_t)addr_table[i].actual_vaddr);
    }
    else if (addr_table[i].mem_type == MEM_MAP)
    {
        /* Call the dg_os_unmap_iomem() API to unmap */
        dbg_result = dbg_os_unmap_iomem(addr_table[i].actual_vaddr,
					addr_table[i].actual_nbytes);
        if ( dbg_result != DBG_STATUS_SUCCESS ) {
            return dbg_result;
        }
    }
    /* Remove that entry */
    clean_entry(&addr_table[i]);

    return DBG_STATUS_SUCCESS;
}



dbg_status_t dbg_hmem_find_entry(
    IN   u32_t    low_addr,
    IN   u32_t    high_addr,
    OUT  alloc_entry_t *entry_p
    )
{
    u32_t i;
    dbg_virt_addr_t virt_addr = (dbg_virt_addr_t)LONG_ADDR(low_addr, high_addr);

    for ( i = 0; i < addr_table_size; i++ ) {
        if ( addr_table[i].aligned_vaddr == virt_addr ) {
            memcpy(entry_p, addr_table + i, sizeof(alloc_entry_t));
            return DBG_STATUS_SUCCESS;
        }
    }

    return DBG_STATUS_FAILURE;
}

dbg_status_t dbg_hmem_map(
    IN   u32_t        low_addr, 
    IN   u32_t        high_addr, 
    IN   u32_t        nbytes,
    OUT  dbg_virt_addr_t *virt_addr_p
    )
{

    dbg_phys_addr_t ptr = (dbg_phys_addr_t) LONG_ADDR(low_addr, high_addr);
    char local_buf[MAX_HMEM_DESC_STR_SIZE+2];
    u32_t entry;
    dbg_status_t rc;

    printf("dbg_hmem_map\n");

    /* Find an empty entry from table */
    entry = find_empty_entry(&addr_table);
    if (entry == -1)
    {
        return DBG_STATUS_RESOURCE;
    }

    /* Call the dg_os_map_iomem API to map: there may be a handle returned 
     * as well */
    rc = dbg_os_map_iomem(nbytes, ptr, virt_addr_p);

    if ((rc != DBG_STATUS_SUCCESS) || (*virt_addr_p == 0))
    {
        //Tcl_SetObjResult(interp, Tcl_NewStringObj("Can't map the physical memory to virtual space", -1));
        return DBG_STATUS_FAILURE;
    }

    /* Append it */
    addr_table[entry].actual_vaddr = *virt_addr_p;
    addr_table[entry].aligned_vaddr = *virt_addr_p;
    addr_table[entry].nbytes = nbytes;
    addr_table[entry].actual_nbytes = nbytes;
    addr_table[entry].mem_type = MEM_MAP;
    addr_table[entry].actual_paddr = ptr;
    addr_table[entry].aligned_paddr = ptr;
    sprintf(local_buf, "tmp%04x", entry);
    strncpy(addr_table[entry].desc_str, local_buf, MAX_HMEM_DESC_STR_SIZE);

    return DBG_STATUS_SUCCESS;
}

/*************************************************************************************************/

static void
clean_entry (
    alloc_entry_t *entry
    )
{

    entry->dev_idx = 0;
    entry->mem_type = MEM_NONE;
    entry->actual_vaddr = 0;
    entry->actual_paddr = 0;
    entry->aligned_vaddr = 0;
    entry->aligned_paddr = 0;
    entry->desc_str[0] = '\0';
    entry->nbytes = 0;
    entry->actual_nbytes = 0;
}

static u32_t
find_empty_entry (
    alloc_entry_t **table_ptr
    )
{
    u32_t i, j;

    /* Find an empty entry from table */
    for (i=0; i<addr_table_size; i++)
    {
        if ((*table_ptr)[i].aligned_vaddr == 0) break;
    }

    if (i == addr_table_size)       /* Still need i here */
    {
        /* Table out of space, expand it. */
        (*table_ptr) = (alloc_entry_t *)realloc((char *)(uintptr_t)(*table_ptr), 
                                          sizeof(alloc_entry_t)*(addr_table_size+ADDR_TABLE_SIZE_DEFAULT));
        if ((*table_ptr) == NULL)
        {
            return -1;
        }
        for (j=0; j<ADDR_TABLE_SIZE_DEFAULT; j++)
        {
            /* Initialize the newly allocated entries. */
            clean_entry(&(*table_ptr)[addr_table_size+j]);
        }
        addr_table_size += ADDR_TABLE_SIZE_DEFAULT;
    }
    return i;
}


dbg_status_t dbg_hmem_init_env(void)
{
    u32_t i;

    printf("dbg_hmem_init_env\n");

    /* Build a table to keep track of what's allocated */
    addr_table = (alloc_entry_t *)malloc(ADDR_TABLE_SIZE_DEFAULT * sizeof(alloc_entry_t));
    if ( addr_table == NULL ) {
        return DBG_STATUS_RESOURCE;
    }
    addr_table_size = ADDR_TABLE_SIZE_DEFAULT;
    for (i=0; i<addr_table_size; i++) 
    {
        clean_entry(&addr_table[i]);
    }

    return DBG_STATUS_SUCCESS;
}

/*
 *------------------------------------------------------------------------
 * toe_fini_hmem_cmd --
 *     Finalize Tcl "hmem" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
dbg_status_t dbg_hmem_cleanup_env(void)
{
    u32_t i;
    dbg_virt_addr_t ptr;
    dbg_status_t res = DBG_STATUS_SUCCESS;

    printf("dbg_hmem_cleanup_env\n");

    for (i=0; i<addr_table_size; i++)
    {
        if (addr_table[i].mem_type == MEM_PHYS)
        {
            if ( dbg_os_free_pcontig(addr_table[i].actual_vaddr) != DBG_STATUS_SUCCESS ) {
                res = DBG_STATUS_FAILURE;
            }
            clean_entry(&addr_table[i]);
        }
        else if (addr_table[i].mem_type == MEM_VIRT)
        {
            ptr = addr_table[i].actual_vaddr;
            free((char *)(uintptr_t)ptr);
            clean_entry(&addr_table[i]);
        }
        else if (addr_table[i].mem_type == MEM_MAP)
        {
            /* Call the dg_os_unmap_iomem() API to unmap */
            if ( dbg_os_unmap_iomem(addr_table[i].actual_vaddr,
				   addr_table[i].actual_nbytes) != DBG_STATUS_SUCCESS ) {
                res = DBG_STATUS_FAILURE;
            }
        }
    }
    free((char *)(uintptr_t)addr_table);
    addr_table = NULL;
    addr_table_size = 0;

    return res;
}

