#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "edebug_types.h"
#include "bcmtype.h"

/*
 * instruction encoding
 */
typedef struct normal_bits
{
#if defined(BIG_ENDIAN)
    u32_t   opcode:6;
    u32_t   rs:5;
    u32_t   rt:5;
    u32_t   imm:16;
#elif defined(LITTLE_ENDIAN)
    u32_t   imm:16;
    u32_t   rt:5;
    u32_t   rs:5;
    u32_t   opcode:6;
#endif /* _ENDIAN */
} normal_bits_t;


typedef struct special_bits
{
#if defined(BIG_ENDIAN)
    u32_t   zero:6;
    u32_t   rs:5;
    u32_t   rt:5;
    u32_t   rd:5;
    u32_t   imm:5;
    u32_t   opcode:6;
#elif defined(LITTLE_ENDIAN)
    u32_t   opcode:6;
    u32_t   imm:5;
    u32_t   rd:5;
    u32_t   rt:5;
    u32_t   rs:5;
    u32_t   zero:6;
#endif /* _ENDIAN */
} special_bits_t;


typedef struct regimm_bits
{
#if defined(BIG_ENDIAN)
    u32_t   zero:6;
    u32_t   rs:5;
    u32_t   opcde:5;
    u32_t   offset:16;
#elif defined(LITTLE_ENDIAN)
    u32_t   offset:16;
    u32_t   opcde:5;
    u32_t   rs:5;
    u32_t   zero:6;
#endif /* _ENDIAN */
} regimm_bits_t;


/*
 * MIPS32 instruction class
 */
typedef enum inst_class
{
    ISC_NORMAL = 64,
    ISC_SPECIAL = 0,
    ISC_REGIMM = 1,
    ISC_COP0 = 16,
    ISC_COP1 = 17,
    ISC_COP2 = 18,
    ISC_COP3 = 19,
    ISC_SPECIAL2 = 28,
    ISC_OBSOLETE = 128,
    ISC_FUTURE = 129,
    ISC_ASE = 130,
    ISC_RESERVED = 131,
    ISC_VENDOR = 132,
    ISC_MOVCI = 133,
} inst_class_t;


typedef enum inst_field
{
    ISF_RESERVED = 0,
    ISF_JUMP     = 1,       /* J addr */
    ISF_DIRECT   = 2,       /* BEQ rs, rt, offset */
    ISF_INDIRECT = 3,       /* LB rt, offset(rs) */
    ISF_SPECIAL  = 4,       /* SLT rd, rs, rt */
    ISF_REGIMM   = 5,       /* TLTI rs, immediate */
    ISF_JR       = 6,       /* JR rs */
} inst_field_t;


/*
 * MIPS32 opcode descriptor
 */
typedef struct inst_opcode
{
    inst_class_t isc;
    char*        mnemonic;
    u8_t         rs_valid;
    u8_t         rt_valid;
    u8_t         rd_valid;
    u8_t         imm_valid;
    inst_field_t ift;
    u8_t         imm_unshift;
    u8_t         is_branch;
} inst_opcode_t;


/*
 * MIPS32 opcode table
 */
inst_opcode_t _nop = { ISC_NORMAL, "nop", 0, 0, 0, 0, ISF_DIRECT, 0 };

inst_opcode_t opcode_table[] =
{
    /* 000 000 */ { ISC_SPECIAL,  "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 000 001 */ { ISC_REGIMM,   "", 0, 0, 1, 1, ISF_RESERVED, 0, 0 }, 
    /* 000 010 */ { ISC_NORMAL,   "j", 0, 0, 0, 0, ISF_JUMP, 0, 0 }, 
    /* 000 011 */ { ISC_NORMAL,   "jal", 0, 0, 0, 0, ISF_JUMP, 0, 0 },
    /* 000 100 */ { ISC_NORMAL,   "beq", 1, 1, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 000 101 */ { ISC_NORMAL,   "bne", 1, 1, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 000 110 */ { ISC_NORMAL,   "blez", 1, 0, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 000 111 */ { ISC_NORMAL,   "bgtz", 1, 0, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 001 000 */ { ISC_NORMAL,   "addi", 1, 1, 1, 1, ISF_DIRECT, 0, 0 }, 
    /* 001 001 */ { ISC_NORMAL,   "addiu", 1, 1, 1, 1, ISF_DIRECT, 1, 0 }, 
    /* 001 010 */ { ISC_NORMAL,   "slti", 1, 1, 1, 1, ISF_DIRECT, 0, 0 }, 
    /* 001 011 */ { ISC_NORMAL,   "sltiu", 1, 1, 1, 1, ISF_DIRECT, 0, 0 }, 
    /* 001 100 */ { ISC_NORMAL,   "andi", 1, 1, 1, 1, ISF_DIRECT, 1, 0 }, 
    /* 001 101 */ { ISC_NORMAL,   "ori", 1, 1, 1, 1, ISF_DIRECT, 1, 0 }, 
    /* 001 110 */ { ISC_NORMAL,   "xori", 1, 1, 1, 1, ISF_DIRECT, 1, 0 }, 
    /* 001 111 */ { ISC_NORMAL,   "lui", 0, 1, 1, 1, ISF_DIRECT, 1, 0 }, 
    /* 010 000 */ { ISC_COP0,     "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 001 */ { ISC_COP1,     "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 010 */ { ISC_COP2,     "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 011 */ { ISC_COP3,     "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 100 */ { ISC_OBSOLETE, "beql", 1, 1, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 010 101 */ { ISC_OBSOLETE, "bnel", 1, 1, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 010 110 */ { ISC_OBSOLETE, "blezl", 1, 0, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 010 111 */ { ISC_OBSOLETE, "bgtzl", 1, 0, 1, 1, ISF_DIRECT, 0, 1 }, 
    /* 011 000 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 001 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 010 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 011 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 100 */ { ISC_SPECIAL2, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 101 */ { ISC_ASE,      "jalx", 0, 0, 0, 0, ISF_JUMP, 0, 0 }, 
    /* 011 110 */ { ISC_ASE,      "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 111 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 100 000 */ { ISC_NORMAL,   "lb", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 },
    /* 100 001 */ { ISC_NORMAL,   "lh", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 100 010 */ { ISC_NORMAL,   "lwl", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 100 011 */ { ISC_NORMAL,   "lw", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 100 100 */ { ISC_NORMAL,   "lbu", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 100 101 */ { ISC_NORMAL,   "lhu", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 100 110 */ { ISC_NORMAL,   "lwr", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 100 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 000 */ { ISC_NORMAL,   "sb", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 101 001 */ { ISC_NORMAL,   "sh", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 101 010 */ { ISC_NORMAL,   "swl", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 101 011 */ { ISC_NORMAL,   "sw", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 101 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 101 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 110 */ { ISC_NORMAL,   "swr", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 101 111 */ { ISC_NORMAL,   "cache", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 },
    /* 110 000 */ { ISC_NORMAL,   "ll", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 110 001 */ { ISC_NORMAL,   "lwc1", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 110 010 */ { ISC_VENDOR,   "lwc2", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 110 011 */ { ISC_NORMAL,   "pref", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 110 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 110 101 */ { ISC_NORMAL,   "ldc1", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 110 110 */ { ISC_VENDOR,   "ldc2", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 110 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 000 */ { ISC_NORMAL,   "sc", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 111 001 */ { ISC_NORMAL,   "swc1", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 111 010 */ { ISC_VENDOR,   "swc2", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 111 011 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 101 */ { ISC_NORMAL,   "sdc1", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 111 110 */ { ISC_VENDOR,   "sdc2", 1, 1, 1, 1, ISF_INDIRECT, 0, 0 }, 
    /* 111 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
};


/*
 * MIPS32 special class opcode table
 */
inst_opcode_t special_table[] =
{
    /* 000 000 */ { ISC_NORMAL,   "sll", 0, 1, 1, 1, ISF_SPECIAL, 0, 0 }, 
    /* 000 001 */ { ISC_MOVCI,    "movci", 0, 0, 1, 0, ISF_RESERVED, 0, 0 }, 
    /* 000 010 */ { ISC_NORMAL,   "srl", 0, 1, 1, 1, ISF_SPECIAL, 0, 0 }, 
    /* 000 011 */ { ISC_NORMAL,   "sra", 0, 1, 1, 1, ISF_SPECIAL, 0, 0 }, 
    /* 000 100 */ { ISC_NORMAL,   "sllv", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 000 101 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 000 110 */ { ISC_NORMAL,   "srlv", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 000 111 */ { ISC_NORMAL,   "srav", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 001 000 */ { ISC_NORMAL,   "jr", 1, 0, 0, 0, ISF_JR, 0, 0 }, 
    /* 001 001 */ { ISC_NORMAL,   "jalr", 1, 0, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 001 010 */ { ISC_NORMAL,   "movz", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 001 011 */ { ISC_NORMAL,   "movn", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 001 100 */ { ISC_NORMAL,   "syscall", 0, 0, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 001 101 */ { ISC_NORMAL,   "break", 0, 0, 0, 0, ISF_SPECIAL, 0, 0 },     
    /* 001 110 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 001 111 */ { ISC_NORMAL,   "sync", 0, 0, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 010 000 */ { ISC_NORMAL,   "mfhi", 0, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 010 001 */ { ISC_NORMAL,   "mthi", 1, 0, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 010 010 */ { ISC_NORMAL,   "mflo", 0, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 010 011 */ { ISC_NORMAL,   "mtlo", 1, 0, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 010 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 101 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 110 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 010 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 000 */ { ISC_NORMAL,   "mult", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 011 001 */ { ISC_NORMAL,   "multu", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 011 010 */ { ISC_NORMAL,   "div", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 011 011 */ { ISC_NORMAL,   "divu", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 011 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 101 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 110 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 011 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 100 000 */ { ISC_NORMAL,   "add", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 001 */ { ISC_NORMAL,   "addu", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 010 */ { ISC_NORMAL,   "sub", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 011 */ { ISC_NORMAL,   "subu", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 100 */ { ISC_NORMAL,   "and", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 101 */ { ISC_NORMAL,   "or", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 110 */ { ISC_NORMAL,   "xor", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 100 111 */ { ISC_NORMAL,   "nor", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 101 000 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 001 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 010 */ { ISC_NORMAL,   "slt", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 101 011 */ { ISC_NORMAL,   "sltu", 1, 1, 1, 0, ISF_SPECIAL, 0, 0 }, 
    /* 101 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 101 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 110 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 101 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 110 000 */ { ISC_NORMAL,   "tge", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 110 001 */ { ISC_NORMAL,   "tgeu", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 110 010 */ { ISC_NORMAL,   "tlt", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 110 011 */ { ISC_NORMAL,   "tltu", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 110 100 */ { ISC_NORMAL,   "teq", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 110 101 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 110 110 */ { ISC_NORMAL,   "tne", 1, 1, 0, 0, ISF_SPECIAL, 0, 0 }, 
    /* 110 111 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 000 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 001 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 010 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 011 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 100 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 101 */ { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 110 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
    /* 111 111 */ { ISC_FUTURE,   "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 }, 
};


/*
 * MIPS32 regimm class opcode table 
 */
inst_opcode_t regimm_table[] =
{
    /* 00 000 */  { ISC_NORMAL,   "bltz", 1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 00 001 */  { ISC_NORMAL,   "bgez", 1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 00 010 */  { ISC_OBSOLETE, "bltzl", 1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 00 011 */  { ISC_OBSOLETE, "bgezl", 1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 00 100 */  { ISC_RESERVED,  "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 00 101 */  { ISC_RESERVED,  "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 00 110 */  { ISC_RESERVED,  "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 00 111 */  { ISC_RESERVED,  "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 01 000 */  { ISC_NORMAL,   "tgei", 1, 0, 1, 1, ISF_REGIMM, 0, 0 },
    /* 01 001 */  { ISC_NORMAL,   "tgeiu", 1, 0, 1, 1, ISF_REGIMM, 0, 0 },
    /* 01 010 */  { ISC_NORMAL,   "tlti", 1, 0, 1, 1, ISF_REGIMM, 0, 0 },
    /* 01 011 */  { ISC_NORMAL,   "tltiu", 1, 0, 1, 1, ISF_REGIMM, 0, 0 },
    /* 01 100 */  { ISC_NORMAL,   "teqi", 1, 0, 1, 1, ISF_REGIMM, 0, 0 },
    /* 01 101 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 01 110 */  { ISC_NORMAL,   "tnei", 1, 0, 1, 1, ISF_REGIMM, 0, 0 },
    /* 01 111 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 10 000 */  { ISC_NORMAL,   "bltzal",  1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 10 001 */  { ISC_NORMAL,   "bgezal",  1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 10 010 */  { ISC_OBSOLETE, "bltzall", 1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 10 011 */  { ISC_OBSOLETE, "bgezall",  1, 0, 1, 1, ISF_REGIMM, 0, 1 },
    /* 10 100 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 10 101 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 10 110 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 10 111 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 000 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 001 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 010 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 011 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 100 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 101 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 110 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
    /* 11 111 */  { ISC_RESERVED, "", 0, 0, 0, 0, ISF_RESERVED, 0, 0 },
};

/*
 *
 */
u32_t
calc_imm(
    u8_t unshift,
    u8_t is_branch,
    u32_t imm,
    u32_t pc
    )
{
    if( unshift )
    {
        if( is_branch )
        {
            return pc + imm + 4;
        }
        else
        {
            return imm;
        }
    }
    else
    {
        if( is_branch )
        {
            return pc + (imm << 2) + 4;
        }
        else
        {
            return imm << 2;
        }
    }
}

/*
 * disassemble 32-bit word
 */
char*
dbg_fdisasm(
    u32_t inst,
    u32_t pc
    )
{
    normal_bits_t* nbits;
    inst_opcode_t* iop;
    special_bits_t* sbits = NULL;
    regimm_bits_t* rbits = NULL;
    static char str[80];
    void * tmp_ptr = &inst;

    nbits = (normal_bits_t*)tmp_ptr;
    iop = (inst == 0) ? &_nop : &opcode_table[nbits->opcode];

    /* 1st level lookup */
    switch( iop->isc )
    {
        case ISC_SPECIAL:
            sbits = (special_bits_t*)nbits;
            iop = &special_table[sbits->opcode];
            break;
        case ISC_REGIMM:
            rbits = (regimm_bits_t*)nbits;
            iop = &regimm_table[nbits->opcode];
            break;
        case ISC_NORMAL:
        case ISC_OBSOLETE:
        case ISC_COP0:
        case ISC_COP1:
        case ISC_COP2:
        case ISC_COP3:
        case ISC_SPECIAL2:
        case ISC_FUTURE:
        case ISC_ASE:
        case ISC_RESERVED:
        case ISC_VENDOR:
        default:
            break;
    }

    /* address and raw opcode */
    sprintf( str, "%08x:      %08x        ", pc, inst );

    /* actual decode */
    switch( iop->ift )
    {
        case ISF_JUMP:      /* J addr */
            sprintf( strchr(str, 0), "%-8.8s", iop->mnemonic );
            sprintf( strchr(str, 0), "0x%x", 
                     ((inst & 0x03ffffff)<<2) | (pc & 0xf0000000) );
            break;
        case ISF_DIRECT:    /* BEQ rs, rt, offset */
            sprintf( strchr(str, 0), "%-8.8s", iop->mnemonic );
            if( iop->rt_valid )
                sprintf( strchr(str, 0), "r%d,", nbits->rt );
            if( iop->rs_valid )
                sprintf( strchr(str, 0), "r%d,", nbits->rs );
            if( iop->imm_valid )
                sprintf( strchr(str, 0), "0x%x", 
                         calc_imm( iop->imm_unshift, iop->is_branch,
                                   nbits->imm, pc ) );
            break;
        case ISF_INDIRECT:  /* LB rt, offset(rs) */
            sprintf( strchr(str, 0), "%-8.8sr%d,0x%x(r%d)", 
                     iop->mnemonic, nbits->rt, nbits->imm, nbits->rs );
            break;
        case ISF_SPECIAL:   /* SLT rd, rs, rt */
            sprintf( strchr(str, 0), "%-8.8s", iop->mnemonic );
            if( iop->rd_valid )
                sprintf( strchr(str, 0), "r%d,", sbits->rd );
            if( iop->rs_valid )
                sprintf( strchr(str, 0), "r%d,", sbits->rs );
            if( iop->rt_valid )
                sprintf( strchr(str, 0), "r%d", sbits->rt );
            if( iop->imm_valid )
                sprintf( strchr(str, 0), ",%d", sbits->imm );
            break;
        case ISF_REGIMM:    /* TLTI rs, immediate */
            sprintf( strchr(str, 0), "%-8.8s", iop->mnemonic );
            if( iop->rs_valid )
                sprintf( strchr(str, 0), "r%d,", rbits->rs );
            if( iop->imm_valid )
                sprintf( strchr(str, 0), "0x%x",
                         calc_imm( iop->imm_unshift, iop->is_branch,
                                   rbits->offset , pc ) );
            break;
        case ISF_JR:        /* JR rs */
            sprintf( strchr(str, 0), "%-8.8sr%d", iop->mnemonic,
                     sbits->rs );
            break;
        case ISF_RESERVED:
        default:
            sprintf( strchr(str, 0), "0x%x", inst );
            break;
    }

    return str;
}

