#include <stdio.h>
#include <string.h>
#include "man_types.h"
#include "dev_if.h"
#include "dbg_man.h"
#include "hw_access.h"
#include "os_types.h"

#define MAX_RETRIES 5
#define WAIT_MS     500

static u32_t no_pci_last_path_used = 0;


static dbg_status_t duart_reg_read(
    u32_t   addr,
    u32_t * val32_p
    )
{
    dbg_status_t status;
    u8_t i, duart_port = dbg_get_duart_port();

    for (i=0; i< MAX_RETRIES; i++) {    
        status=dbg_duart_read(duart_port, addr, val32_p);
        if (status==DBG_STATUS_SUCCESS) break;        
        dbg_msleep(WAIT_MS);
    }
	
	if (status != DBG_STATUS_SUCCESS) {
	    printf ("Error - duart_reg_read failed\n");
	}
    
    return status;
}

static dbg_status_t duart_reg_write(
    u32_t   addr,
    u32_t   val32,
    u8_t    wait_for_echo
    )
{
    dbg_status_t status;
    u8_t i, duart_port = dbg_get_duart_port();

    for (i=0; i< MAX_RETRIES; i++) {  
        status=dbg_duart_write(duart_port, addr, val32, wait_for_echo);
        if (status==DBG_STATUS_SUCCESS) break;        
        dbg_msleep(WAIT_MS);
    }
	
	if (status != DBG_STATUS_SUCCESS) {
	    printf ("Error - duart_reg_write failed\n");
	}

    return status;
}

dbg_status_t
dbg_reg_read(
    u32_t addr,
    u32_t * val32_p
    )
{
    dbg_status_t status;
    u32_t val, result, i;

    if ( tcl_al_get_no_pci_mode() == 0) {
        if ( ! XDBG_CUR_DEV_IS_VALID ) {
            return DBG_STATUS_FAILURE;
        } else {
            dev_if_t * cur_dev = dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].drv_dev_if_p;


	    /* if device is hidden, perform indirect reading instead */
            if ( XDBG_CUR_DEV_IS_HIDDEN ) {
                return DBG_STATUS_FAILURE;
            }
            else {
                return dbg_dev_reg_read(cur_dev, addr, val32_p);
                }
        }
    } else {
        if (xdbg_cur_dev !=   BAD_DEVICE_HNDL) {
            /* if we want to read register from another path, first write
               to the appropriate register */
            if (no_pci_last_path_used !=  dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.path_num) {                
                val = 1 << 31;
                val |= dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.path_num << 20;
                
                // Verify that DBU register has changed before continueing
                for (i=0; i<MAX_RETRIES; i++) {
                    duart_reg_write(DBU_REG_DBU_VFID_CFG_OFFSET, val, 1);
                    duart_reg_read(DBU_REG_DBU_VFID_CFG_OFFSET, &result); 
                    if (result == val) break;                                    
                    dbg_msleep(WAIT_MS);
                }
                if (result != val) {
				    printf("Error - Failed to set read DBU_REG_DBU_VFID_CFG_OFFSET register !\n");
                    return DBG_STATUS_FAILURE;
                }

                no_pci_last_path_used = dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.path_num;
            }

            // bit22-24 of the address holds the device number 
            addr |= 0x400000*(dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.device_num);
        }

        return duart_reg_read(addr, val32_p);
    }

    return status;
}

dbg_status_t
dbg_reg_write(
    u32_t addr,
    u32_t val,
    u8_t  wait_for_echo
    )
{
    dbg_status_t status;
    u32_t path_reg_val, i, result;

    if ( tcl_al_get_no_pci_mode() == 0 ) {
        if ( ! XDBG_CUR_DEV_IS_VALID ) {
            return DBG_STATUS_FAILURE;
        } else {
            dev_if_t * cur_dev = XDBG_CUR_DEV_IF;

	    /* if device is hidden, perform indirect writing instead */
            if ( XDBG_CUR_DEV_IS_HIDDEN ) {
                return DBG_STATUS_FAILURE;
            }
            else {
                return dbg_dev_reg_write(cur_dev,
                                         addr,
                                         val);
            }
        }
    } else {
        if (xdbg_cur_dev !=   BAD_DEVICE_HNDL) {
            /* if we want to read/write register from another path, first write
               to the appropriate register */
            if (no_pci_last_path_used !=  dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.path_num) {
                path_reg_val = 1 << 31;
                path_reg_val |= dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.path_num << 20;               

                // Verify that DBU register has changed before continueing
                for (i=0; i<MAX_RETRIES; i++) {
                    duart_reg_write(DBU_REG_DBU_VFID_CFG_OFFSET, path_reg_val, 1);
                    duart_reg_read(DBU_REG_DBU_VFID_CFG_OFFSET, &result); 
                    if (result == path_reg_val) break;
					dbg_msleep(WAIT_MS);
                }
                if (result != path_reg_val) {
				    printf("Error - Failed to set read DBU_REG_DBU_VFID_CFG_OFFSET register !\n");					
                    return DBG_STATUS_FAILURE;
                }

                no_pci_last_path_used = dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.path_num;
            }

            /* bit22-24 of the address holds the device number */            
            addr |= 0x400000*(dbg_pci_devices[PF_IDX_BY_HDL(xdbg_cur_dev)].pci_info.device_num);
        }

        return duart_reg_write(addr, val, wait_for_echo);
    }

    return status;
}

dbg_status_t
dbg_reg_dread(
    u32_t addr,
    u32_t * val32_p
    )
{
    if ( tcl_al_get_no_pci_mode() == 0 ) {
        if ( ! XDBG_CUR_DEV_IS_VALID ) {
            return DBG_STATUS_FAILURE;
        } else {
            dev_if_t * cur_dev = XDBG_CUR_DEV_IF;
            return dbg_dev_reg_dread(cur_dev,
                                     addr,
                                     val32_p);
        }
    } else {
        return duart_reg_read(addr, val32_p);
    }
}

dbg_status_t
dbg_reg_dwrite(
    u32_t addr,
    u32_t val
    )
{
    if ( tcl_al_get_no_pci_mode() == 0 ) {
        if ( ! XDBG_CUR_DEV_IS_VALID ) {
            return DBG_STATUS_FAILURE;
        } else {
            dev_if_t * cur_dev = XDBG_CUR_DEV_IF;
            return dbg_dev_reg_dwrite(cur_dev,
                                     addr,
                                     val);
        }
    } else {
        return duart_reg_write(addr, val, 1); //wait for echo is set to 1
    }
}



