#include "man_types.h"
#include "dev_if.h"
#include "hw_access.h"
#include "dbg_serport.h"
#include "esc_seq.h"
#include "dbg_os_utils.h"
#include "./../../../../common/include/prints.h"

#define MAX_READ_RETRIES 5
#define WAIT_MS          500

uart_info_t uart_info = {0};

void dbg_init_uart_info_array(void)
{
        int i;

        for (i = 0; i < MAX_UART_HANDLES; i++)
        {
                uart_info.uart_handles[i] = -1;
        }
}

dbg_status_t dbg_set_duart_port(
    u8_t port
    )
{
    /* Port should be opened before the redirection */
	if ( uart_info.uart_handles[port] == -1 ) {
        return DBG_STATUS_CON_CLOSED;
    }
    uart_info.duart_port = port;
    return DBG_STATUS_SUCCESS;
}

u8_t dbg_get_duart_port(void)
{
	return uart_info.duart_port;
}

u8_t dbg_uart_is_opened(void)
{
        return uart_info.uart_io_handle != 0;
}

ua_handle_t dbg_uart_get_io_handle(void)
{
        return uart_info.uart_io_handle;
}

void * dbg_uart_get_io_handle_ref(void)
{
        return &(uart_info.uart_io_handle);
}

void dbg_serport_cleanup(void)
{
        u32_t index;

        /* Close all COM port(s) */
        for (index = 0; index < MAX_UART_HANDLES; index++)
        {
                if (uart_info.uart_handles[index] != -1)
                {
                    dbg_os_uart_close(uart_info.uart_handles[index]);
                    uart_info.uart_handles[index] = -1;
                }
        }
        /* Make sure the global variable points at nothing also. */
        uart_info.uart_io_handle = 0;  /* indicate no COM port used */
        //uart_info.uart_stat_handle = 0;  /* indicate no COM port used */
}

dbg_status_t dbg_serport_open (
    u8_t * setting,
    u32_t port,
    u32_t baud
    )
{
        ua_status_t rc;
        ua_handle_t  uart_handle;

        rc = dbg_os_uart_open(&uart_handle, port, baud, (char *)setting, UAM_TEXT);

        if (rc == UA_STATUS_OK)
        {
                uart_info.uart_handles[port] = uart_handle;
                printf("Opened port %d settings: %d,%s\n", port, baud, setting);
                return DBG_STATUS_SUCCESS;
        }
        else {
                switch (rc)
                {
                case UA_STATUS_PORT_INUSE:
                        printf("Error: port %d is in use\n", port);
                        break;
                case UA_STATUS_NODEV:
                        printf("Error: port %d is not available\n", port);
                        break;
                case UA_STATUS_BAD_BAUD:
                        printf("Error: port %d wrong baud rate setting\n", port);
                        break;
                case UA_STATUS_BAD_DATABIT:
                        printf("Error: port %d wrong data bit setting\n", port);
                        break;
                case UA_STATUS_BAD_PARITY:
                        printf("Error: port %d wrong parity setting\n", port);
                        break;
                case UA_STATUS_BAD_STOPBIT:
                        printf("Error: port %d wrong stop bit setting\n", port);
                        break;
                default:
                        printf("Error: failed to open port %d\n", port);
                        break;
                }
        }
        return DBG_STATUS_FAILURE;
}


dbg_status_t dbg_serport_close (void)
{
        u32_t rc, index;
	dbg_status_t return_value = DBG_STATUS_SUCCESS;
	u8_t closed_flag=0;

	uart_info.print_buf[0] = 0;
	uart_info.auxprint_buf[0] = 0;

	/* Even if any close error happened, no redirection is desired to the end
	   user. */
	uart_info.verbose_mode &= ~V_SER_PORT;

	for (index = 0; index < MAX_UART_HANDLES; index++)
	{
		/* Skip the non-opened port */
		if (uart_info.uart_handles[index] == -1)
			continue;

		if ((rc = dbg_os_uart_close(uart_info.uart_handles[index])) == UA_STATUS_OK)
		{
                        uart_info.uart_handles[index] = -1;  /* indicate no COM port used */
			closed_flag =1;
		}
		else
		{
			/* Even in the case of closing error, still reset the handle to
			 * avoid using the ill-opened port for redirection. */
			uart_info.uart_handles[index] = -1;

			/* Set the error return value. */
			return_value = DBG_STATUS_FAILURE;

			/* Report the error */
			//printf("Error: port %d can not be closed \n", index);
		}

	}

	if (closed_flag==0) {
		printf("serial port was not closed:  there were no serial ports open\n");
	}

        /* Make sure the global variable points at nothing also. */
        uart_info.uart_io_handle = 0;  /* indicate no COM port used */
        //uart_info.uart_stat_handle = 0;  /* indicate no COM port used */

        return return_value;
}


dbg_status_t dbg_serport_redirect(
    u32_t io_port
    )
{
	uart_info.uart_io_handle = uart_info.uart_handles[io_port];

	/* If the I/O handle is 0 which means the I/O port has not been opened yet,
	 * no redirection can be enabled because nothing can work without I/O.
	 */
	if (uart_info.uart_io_handle == 0)
	{
		return DBG_STATUS_CON_CLOSED;
	} else {
		uart_info.verbose_mode |= V_SER_PORT;
	}

    return DBG_STATUS_SUCCESS;
}


dbg_status_t dbg_duart_read(
    u8_t      io_port,
    u32_t     addr,
    u32_t   * val32_p
    )
{
        return dbg_os_duart_read(uart_info.uart_handles[io_port], addr, val32_p) == UA_STATUS_OK ? DBG_STATUS_SUCCESS : DBG_STATUS_FAILURE;
}

dbg_status_t dbg_duart_write(
    u8_t      io_port,
    u32_t     addr,
    u32_t     val32,
    u8_t      wait_for_echo
    )
{
        return dbg_os_duart_write(uart_info.uart_handles[io_port], addr, val32, wait_for_echo) == UA_STATUS_OK ? DBG_STATUS_SUCCESS : DBG_STATUS_FAILURE;
}

dbg_status_t dbg_duart_nvm_read(
    u8_t      io_port,
    u32_t     addr,
    u32_t   * buf,
    u32_t     buf_sz
    )
{
        return dbg_dev_duart_nvm_read(uart_info.uart_handles[io_port], addr, buf, buf_sz);
}

dbg_status_t dbg_duart_nvm_write(
    u8_t      io_port,
    u32_t     addr,
    u32_t   * buf,
    u32_t     buf_sz
    )
{
        return dbg_dev_duart_nvm_write(uart_info.uart_handles[io_port], addr, buf, buf_sz);
}

dbg_status_t dbg_setup_duart_nvm(void)
{
    u8_t duart_port = dbg_get_duart_port();
    dbg_status_t stat;
    u32_t total_nvm_size = 0, nvm_psize = NVRAM_PAGE_SIZE, chip_id = 0, val = 0, i;

    /* Initialization of DUART_NVRAM internals */
    for (i=0; i<MAX_READ_RETRIES; i++) {
        stat = dbg_duart_read(duart_port, MCP_REG_MCPR_NVM_CFG4, &total_nvm_size);
        if (stat == DBG_STATUS_SUCCESS) break;
        dbg_msleep(WAIT_MS);
    }
    if ( stat != DBG_STATUS_SUCCESS ) {
        printf("Failed to read total NVRAM size\n");
        return stat;
    }

    total_nvm_size = NVRAM_1MB_SIZE << (total_nvm_size & MCPR_NVM_CFG4_FLASH_SIZE);

    /* Get the chip revision id and number. */
	/* chip num:16-31, rev:12-15, metal:4-11, bond_id:0-3 */
    for (i=0; i<MAX_READ_RETRIES; i++) {
        stat = dbg_duart_read(duart_port, MISC_REG_CHIP_NUM, &val);
        if (stat == DBG_STATUS_SUCCESS) break;
        dbg_msleep(WAIT_MS);
    }
    if ( stat != DBG_STATUS_SUCCESS ) {
        printf("Failed to read chip number");
        return stat;
    }
    chip_id = (val & 0xffff) << 16;

    
    for (i=0; i<MAX_READ_RETRIES; i++) {
        stat = dbg_duart_read(duart_port, MISC_REG_CHIP_REV, &val);
        if (stat == DBG_STATUS_SUCCESS) break;
        dbg_msleep(WAIT_MS);
    }
    if ( stat != DBG_STATUS_SUCCESS ) {
        printf("Failed to read chip revision number");
        return stat;
    }
    chip_id |= (val & 0xf) << 12;

    for (i=0; i<MAX_READ_RETRIES; i++) {
        stat = dbg_duart_read(duart_port, MISC_REG_CHIP_METAL, &val);
        if (stat == DBG_STATUS_SUCCESS) break;
        dbg_msleep(WAIT_MS);
    }
    if ( stat != DBG_STATUS_SUCCESS ) {
        printf("Failed to read chip metal version");
        return stat;
    }
    chip_id |= (val & 0xff) << 4;

    /* TBD: check chip_id */
    chip_id |= (val & 0xf);

    dbg_dev_duart_nvm_set_dev_info(chip_id, total_nvm_size, nvm_psize);

    return DBG_STATUS_SUCCESS;
}

dbg_status_t dbg_duart_get_chip_id(u32_t *chip_id_p)
{
    u8_t duart_port = dbg_get_duart_port();
    dbg_status_t stat;
    u32_t val=0;

    stat = dbg_duart_read(duart_port, MISC_REG_CHIP_NUM, &val);
    if ( stat != DBG_STATUS_SUCCESS ) {
        return stat;
    }
    *chip_id_p = (val & 0xffff);

    return DBG_STATUS_SUCCESS;

}

u32_t dbg_duart_get_nvm_size(void)
{
    return 0x100000;
}

