#ifndef __BROADCOM_DBG_HMEM_DEF_H__
#define __BROADCOM_DBG_HMEM_DEF_H__

#include "bcmtype.h"
#include "os_types.h"

#ifndef CACHE_LINE_SIZE_MASK
#define CACHE_LINE_SIZE_MASK        0x3f
#endif

/* Maximal length of hmem memory description */
#define MAX_HMEM_DESC_STR_SIZE     30    

typedef enum _access_method_t {
    ACCESS_8BIT, ACCESS_16BIT, ACCESS_32BIT
} access_method_t;

typedef enum _pattern_t {
    PTTN_FIXED, PTTN_INC1
} pattern_t;

typedef struct _alloc_entry_t {
    u32_t dev_idx;       /* Device that allocates the memory, 0 if user mode memory */
    u32_t mem_type;      /* Can be MEM_VIRT, MEM_PHYS, MEM_MAP */
    dbg_virt_addr_t actual_vaddr;  /* Virtual address */
    dbg_phys_addr_t actual_paddr;  /* Physical address, NULL if user mode memory */
    dbg_virt_addr_t aligned_vaddr; /* Aligned virtual address */
    dbg_phys_addr_t aligned_paddr; /* Aligned physical address */
    u32_t nbytes;        /* byte count user wants, not necessarily the actual size */
    u32_t actual_nbytes; /* byte count of memory allocated */
    //u32_t handle;        /* Handle to the allocated memory, -1 if user mode memory */
    char desc_str[MAX_HMEM_DESC_STR_SIZE+2];  /* Description string. +2 is for CR and NULL chars */
} alloc_entry_t;



/**
 * Description:
 *      Reads the defined number of bytes from the given address.
 * 
 * Returns:
 *      DIAG_STATUS_INVALID_PARAMETER if "method" has an unknown value
 *      DIAG_STATUS_SUCCESS 
 */ 
dbg_status_t dbg_hmem_read(
    IN    u32_t       low_addr_val,      /* low dword of the address to read from */ 
    IN    u32_t       high_addr_val,     /* high dword of the address to read from */ 
    IN    access_method_t method,        /* access method, defining the number of bytes to read */
    OUT   u32_t     * val_p              /* output buffer */
    );

/**
 * Description:
 *      Writes the defined number of bytes to the given address.
 * 
 * Returns:
 *      DIAG_STATUS_INVALID_PARAMETER if "method" has an unknown value
 *      DIAG_STATUS_SUCCESS 
 */ 
dbg_status_t dbg_hmem_write(
    IN    u32_t       low_addr_val,      /* low dword of the address */ 
    IN    u32_t       high_addr_val,     /* high dword of the address */ 
    IN    access_method_t method,        /* access method, defining the number of bytes to write */
    IN    void      * val_p              /* Input buffer */
    );

/**
 * Description:
 *      Fills the memory range with the "incremental" pattern. 
 * Memory range will be filled by incrementing 
 * each byte's value by 1 starting from 0.
 * 
 * Returns:
 *      void
 */ 
void dbg_hmem_fill_inc(
    IN   u32_t low_addr_val, 
    IN   u32_t high_addr_val,
    IN   u32_t nbytes
    );

/**
 * Description:
 *      Fills the memory range with the given "pattern_value". 
 * "pattern" may have two values: PTTN_FIXED and PTTN_INC1.
 * If its value is PTTN_INC1, memory range will be filled by incrementing 
 * each byte's value by 1 starting from 0.
 * 
 * Returns:
 *      void
 */ 
void dbg_hmem_fill(
    IN   u32_t        low_addr,        /* low dword of the address */ 
    IN   u32_t        high_addr,       /* high dword of the address */ 
    IN   u32_t        nbytes,          /* size of the memory range */
    IN   u32_t        pattern_value,   /* pattern */
    IN   pattern_t    pattern_type     /* pattern type */
    );

/**
 * Description:
 *      Allocates a new memory buffer of the given size,
 * of the given type and having the given description string.
 * 
 * -  Memory type may be MEM_VIRT or MEM_PHYS.
 * -  If description string is NULL, a default
 * description string will be assigned to the newly allocated buffer.
 * 
 * Returns:
 *      DIAG_STATUS_INVALID_PARAMETER if memory type is wrong.
 *      DIAG_STATUS_RESOURCE if there is no free space in a registry table
 * or there is no enough memory to allocate a new buffer.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_hmem_alloc_mem(
    IN    u32_t      nbytes,          /* size of the requested buffer */
    IN    char     * descr_str,       /* description string */
    IN    u32_t      memtype,         /* memory type */
    OUT   u32_t    * low_addr_p,      /* virtual address of the new buffer (low dword) */
    OUT   u32_t    * high_addr_p      /* virtual address of the new buffer (high dword) */
    );

/**
 * Description:
 *      Releases a buffer with the given virtual address.
 * 
 * Returns:
 *      DIAG_STATUS_INVALID_PARAMETER if invalid virtual address is given.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_hmem_free(
    IN    u32_t      low_addr,     /* low dword of the address */
    IN    u32_t      high_addr     /* high dword of the address */
    );

/**
 * Description:
 *      This function is needed to provide the upper level 
 * information of how big array it has to allocate in 
 * order to pass it then to dg_hmem_get_used function.
 * 
 * Returns:
 *      Number of none-empty entries in the registry table.
 */ 
u32_t dbg_hmem_get_entries_count(
    void
    );

/**
 * Description:
 *      Finds an entry in the registry table by its virtual address.
 * 
 * Returns:
 *      DIAG_STATUS_SUCCESS if the entry was found.
 *      DIAG_STATUS_FAILURE otherwise.
 */ 
dbg_status_t dbg_hmem_find_entry(
    IN    u32_t           low_addr,   /* low dword of the address */
    IN    u32_t           high_addr,  /* high dword of the address */
    OUT   alloc_entry_t * entry_p     /* output buffer for the found entry */
    );

/**
 * Description:
 *      Maps the physical memory region into application address space.
 * 
 * Returns:
 *      DIAG_STATUS_RESOURCE if there is no enough resources to complete the operation.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_hmem_map(
    IN    u32_t         low_addr,     /* low dword of the address */
    IN    u32_t         high_addr,    /* high dword of the address */
    IN    u32_t         nbytes,       /* size of the region to map */
    OUT   dbg_virt_addr_t * virt_addr_p   /* output buffer (virtual address of the mapped buffer) */
    );

/**
 * Description:
 *      Initializes the hmem infrastructure: allocates registry table, initializes 
 * global variables.
 * 
 * Returns:
 *      DIAG_STATUS_RESOURCE if initialisation failed.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_hmem_init_env(
    void
    );

/**
 * Description:
 *      Cleans up the hmem infrastructure: frees all used bufferes,
 * releases the registry table.
 * 
 * Returns:
 *      DIAG_STATUS_SUCCESS if all was okay.
 *      DIAG_STATUS_FAILURE otherwise.
 */     
dbg_status_t dbg_hmem_cleanup_env(
    void
    );

#endif /* __BROADCOM_DBG_HMEM_DEF_H__ */
