#ifndef __DBG_MANAGER_H__
#define __DBG_MANAGER_H__

#include "edebug_types.h"
#include "man_types.h"

dbg_status_t dbg_pci_scan(void);
dbg_status_t dbg_non_pci_scan(void);
u32_t dbg_get_total_dev(void);
void dg_set_total_br_dev(u32_t no_br_dev);
u32_t dbg_get_total_br_dev(void);

/**
 * Description:
 *      disable statistics gathering by operational driver
 * 
 * Returns:
 *      DIAG_STATUS_FAILURE if operation fails.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_stats_disable();

/**
 * Description:
 *      disable statistics gathering by operational driver
 * 
 * Returns:
 *      DIAG_STATUS_FAILURE if operation fails.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_stats_disable();

/**
 * Description:
 *      enable statistics gathering by operational driver
 * 
 * Returns:
 *      DIAG_STATUS_FAILURE if operation fails.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_stats_enable();

/**
 * Description:
 *      shows NIC statistics 
 * 
 * Returns:
 *      DBG_STATUS_FAILURE if operation fails.
 *      DBG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_stats_show(); 

/**
 * Description:
 *      shows NIC statistics 
 * 
 * Returns:
 *      DBG_STATUS_FAILURE if operation fails.
 *      DBG_STATUS_SUCCESS otherwise.
 */ 

dbg_status_t dbg_dev_stats_show(dev_if_t * cur_dev);

/**
 * Description:
 *      Selects the device with the given index (starting from 0) 
 * in the "device" command list.
 * 
 * Returns:
 *      DIAG_STATUS_FAILURE if operation fails.
 *      DIAG_STATUS_SUCCESS otherwise.
 */ 
dbg_status_t dbg_select_dev_num(
        IN    u32_t        dev_num  /* device index */
        );

dbg_status_t dbg_pcicfg_read(
        IN    u32_t        addr,      /* register offset */  
        OUT   u32_t      * ret_val_p  /* output buffer */
        );

void dbg_pcicfg_setdut(
    u16_t domain,
    u8_t bus,
    u8_t  device,           
    u8_t  function   
    );

void dbg_pcicfg_getdut(
    u16_t *domain,
    u8_t *bus,
    u8_t *device,
    u8_t *function
    );

dbg_status_t dbg_pcicfg_process_link_control_reg(
        pci_info_t  * pci_info_ar, 
        u32_t i, 
        u32_t ret_val
        );

pci_info_t * dbg_get_br_pci_info(
    int dev_idx
    );

/**
 * Description:
 *      Suspends the execution (busy wait) for "cnt" miliseconds.
 * 
 * Returns:
 *      void
 */ 
void dbg_msleep(
    IN    u32_t        cnt        /* number of msec to sleep */
    );

/**
 * Description:
 *      Suspends the execution (busy wait) for "cnt" microseconds.
 * 
 * Returns:
 *      void
 */ 
void dbg_usleep(
    IN    u32_t        cnt        /* number of usec to sleep */
    );

/**
 * Description:
 *      Reads the dword from the register at given offset on the GRC of the "current" device.
 * Uses the default read mechanism. 
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS if the operation succeeds.
 */ 
dbg_status_t dbg_reg_read(
    IN    u32_t        addr,     /* register offset */
    OUT   u32_t      * val32_p   /* output buffer */
    );

/**
 * Description:
 *      Reads the dword from the register at given offset on the GRC of the "current" device.
 * Uses the gateway registers (0x78, 0x80) to access the GRC.
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS if the operation succeeds.
 */ 
dbg_status_t dbg_reg_iread(
    IN    u32_t        addr,     /* register offset */
    OUT   u32_t      * val32_p   /* output buffer */
    );

/**
 * Description:
 *      Reads the dword from the register at given offset on the GRC of the "current" device.
 * Uses PCI memory cycles (direct access the GRC mapping in the application address space).
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS if the operation succeeds.
 */ 
dbg_status_t dbg_reg_dread(
    IN    u32_t        addr,     /* register offset */
    OUT   u32_t      * val32_p   /* output buffer */
    );



/**
 * Description:
 *      Writes the dword to the register at given offset on the GRC of the "current" device.
 * Uses the default write mechanism. 
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS if the operation succeeds.
 */ 
dbg_status_t dbg_reg_write(
    IN    u32_t        addr,     /* register offset */
    IN    u32_t        val,      /* value to write */
    IN    u8_t         wait_for_echo /* 1 - wait for echo , 0 - do not wait for echo */
    );

/**
 * Description:
 *      Writes the dword to the register at given offset on the GRC of the "current" device.
 * Uses the gateway registers (0x78, 0x80) to access the GRC.
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS if the operation succeeds.
 */ 
dbg_status_t dbg_reg_dwrite(
    IN    u32_t        addr,     /* register offset */
    IN    u32_t        val       /* value to write */
    );

/**
 * Description:
 *      Cleans up the diagnostics internals.
 * 
 * Returns:
 *      void
 */ 
dbg_status_t dbg_cleanup(
    void
    );

/**
 * Description:
 *      Cleans up the data structures for the device with the given index.
 * 
 * Returns:
 *      void
 */
dbg_status_t dbg_cleanup_one_br_device (
    IN   u32_t   dev_idx    /* device index */
    );

dbg_status_t dbg_cleanup_one_device(
	dbg_dev_info_t * dev_info
	);

dbg_status_t dbg_dev_get_board_type(u32_t  index);

dbg_status_t dbg_dev_get_bc_ver(u32_t  index);

dbg_status_t dbg_dev_get_shmem_base(u32_t index);

u8 dbg_dev_get_abs_func_num(u32_t num_paths, u32_t index);

/*
*------------------------------------------------------------------------
* dg_number_of_everest_functions --
*
* Result:
*     number of functions (1 or 2)
*
* please note:
* ***********
*  this function assumes that all everest functions in this machine are following device ids. 
*  therefore,  the second everest function should be 1 following or 1 previous to the current device.
*  if this assumption is not current,  than also function dg_select_other_func should be corrected.
*------------------------------------------------------------------------
*/
u32_t dbg_number_of_everest_functions (void);

/**
* Description:
*      Returns the port number.
* 
* Returns:
*      DIAG_STATUS_SUCCESS on success.
*	   result contains the check result.
*/ 
dbg_status_t dbg_get_port_num(
							  u32_t				*result
							  );

/**
* Description:
*      Returns the function number.
* 
* Returns:
*      DIAG_STATUS_SUCCESS on success.
*	   result contains the result.
*/ 
dbg_status_t dbg_get_func_num(u32_t *result);

#endif
