
#ifndef __DBG_SERPORT_H__
#define __DBG_SERPORT_H__

#include "bcmtype.h"
#include "dbg_os_uart.h"
#include "edebug_types.h"
#include "disputil.h"

/**
 * Opens the given serial port
 * 
 * Returns: 
 *      DBG_STATUS_SUCCESS in case of success. Appropriate error
 *      code in case of serial port opening failure.
 */ 
dbg_status_t dbg_serport_open(
    IN    u8_t   * setting,   /* connection settings: "8N1", "7E1" or "8O1" */ 
    IN    u32_t    port,      /* port number: 1 - COM1, 2 - COM2, ... */
    IN    u32_t    baud       /* baud rate: 1200, 9600, 19200, 38400, 57600 or 115200 */
    );

/**
 * Closes all opened ports. Invalidates all handles.
 * Should be used before xDiag exits.
 */ 
void dbg_serport_cleanup(void);

/**
 * Closes all opened ports. Invalidates all handles.
 * Sets up I/O functions to work with console instead of serial. 
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS if all ports were successfully closed,
 *      Appropriate error otherwise
 */ 
dbg_status_t dbg_serport_close(void);


/**
 * Redirects I/O to the given serial port.
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS in case of success,
 *      Appropriate error otherwise
 */ 
dbg_status_t dbg_serport_redirect(
    IN    u32_t    io_port       /* serial port number */
    );

/**
 * Reads DWORD from the given address in GRC using debug UART
 * interface connected to the given serial port.
 * 
 * Returns:
 *      DIAG_STATUS_SUCCESS in case of success,
 *      Appropriate error otherwise
 */ 
dbg_status_t dbg_duart_read(
    IN    u8_t      io_port,    /* number of the serial port connected to the debug UART */
    IN    u32_t     addr,       /* offset in GRC */
    OUT   u32_t   * val32_p     /* output buffer */
    );

/**
 * Writes DWORD to the given address in GRC using debug UART
 * interface connected to the given serial port.
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS in case of success,
 *      Appropriate error otherwise
 */ 
dbg_status_t dbg_duart_write(
    IN    u8_t      io_port,    /* number of the serial port connected to the debug UART */
    IN    u32_t     addr,       /* offset in GRC */
    IN    u32_t     val32,      /* value to write */
    IN    u8_t      wait_for_echo
    );

/**
 * Returns:
 *      TRUE if dbg_serport_derirect has been successfully
 *      called.
 */
u8_t dbg_uart_is_opened(void);

/**
 * Returns:
 *      Current I/O handle (set by dbg_serport_redirect).
 */ 
ua_handle_t dbg_uart_get_io_handle(void);

/**
 * Returns:
 *      Pointer to the current I/O handle.
 */ 
void * dbg_uart_get_io_handle_ref(void);

/**
 * Sets serial port number, which should be used
 * to work with Debug UART port.
 * 
 * Returns: 
 *      DBG_STATUS_CON_CLOSED if the given port hasn't been
 *      opened yet. DBG_STATUS_SUCCESS in case of success.
 */ 
dbg_status_t dbg_set_duart_port(
    IN    u8_t port   /* serial port number (1 for COM1, 2 - COM2, ...) */
    );

/**
 * Inits uart_info global array to empty values 
 */ 
void dbg_init_uart_info_array(void);

/**
 * Gets serial port number, which should be used
 * to work with Debug UART port.
 * 
 * Returns: 
 *      Port numeber
 */ 
u8_t dbg_get_duart_port(void);

/**
 * Reads a buffer from the given address in NVRAM using debug UART
 * interface connected to the given serial port.
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS in case of success,
 *      Appropriate error otherwise
 */ 
dbg_status_t dbg_duart_nvm_read(
    IN    u8_t      io_port,     /* number of the serial port connected to the debug UART */ 
    IN    u32_t     addr,        /* offset in NVRAM */                                         
    OUT   u32_t   * buf,         /* output buffer */                                         
    IN    u32_t     buf_sz       /* buffer size */
    );

/**
 * Writes a buffer to the given address in NVRAM using debug UART
 * interface connected to the given serial port.
 * 
 * Returns:
 *      DBG_STATUS_SUCCESS in case of success,
 *      Appropriate error otherwise
 */ 
dbg_status_t dbg_duart_nvm_write(
    IN    u8_t      io_port,     /* number of the serial port connected to the debug UART */ 
    IN    u32_t     addr,        /* offset in NVRAM */                                       
    IN    u32_t   * buf,         /* output buffer */                                         
    IN    u32_t     buf_sz       /* buffer size */                                           
    );

#endif /* __DBG_SERPORT_H__ */
