#include "man_types.h"
#include "dbg_os_utils.h"
#include "uart.h"
#include "os_utils.h"
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#define DUART_MAX_READ_BUF_LEN    30  /* Maximal size of transaction
                                         Currently it's 20 but we dont' want to fail if hw suddenly behaves
                                         a bit not as written in the spec. */
#define DUART_PROMPT_CHAR         '>' /* Prompt character */
#define DUART_ERROR_CHAR          '?' /* Prompt character */
#define DUART_READ_MAX_RETRY_NUM  5  /* Maximum number of retrys on read */


/**
 * Reads 32 bit from GRC using the debug UART port:
 *
 *  - Writes "r<addr>\r\n", where "addr" is an offset in GRC in hex.
 *  - Reads the echoed symbols
 *  - Reads the charachters until ">" character is read.
 * If ">" was finally read - interpret the previously read value as 32-bit hex value,
 * otherwise ("?" was read) means there was an illegal input - fatal error.
 */
ua_status_t dbg_os_duart_read(
    IN    ua_handle_t   h,
    IN    u32_t         addr,
    OUT   u32_t       * val_p
    )
{
    u8_t buf[DUART_MAX_READ_BUF_LEN + 1];
    s32_t i, j;
    ua_status_t uart_stat;
    u16_t outlen = 0, cur_out_len = 0;
    u32_t str_len;
#define NUM_BITS_TO_WAIT  5*8  /* If we didn't get any charachter yet - let's sleep for the time it get 5 chars to arrive */
    u32_t wait_time_us;

    wait_time_us = (u32_t)((1000000.0* NUM_BITS_TO_WAIT)/((long double)os_if_uart_get_baud(h)));

    /* Empty the UART buffer if there is a junk from the previous operations */
    os_if_uart_flush_rb(h, DUART_MAX_READ_BUF_LEN);

    /* Create a command */
    sprintf((char *)buf, "r%x\r\n", addr);

    str_len = strlen((char *)buf);

    /* Write the command */
    uart_stat = os_if_uart_write(h, buf, str_len);
    if ( uart_stat != UA_STATUS_OK) {
        return uart_stat;
    }

    /* Wait for output to become available */
    os_if_usleep(wait_time_us);

    /* Read the echo */
    for ( i = 0, outlen = 0; i < DUART_READ_MAX_RETRY_NUM ; i++) {
        uart_stat = os_if_uart_read(h, buf + outlen, str_len - outlen, &cur_out_len);
        outlen += cur_out_len;
        if ( uart_stat != UA_STATUS_OK ) {
            return uart_stat;
        }

        if ( outlen == str_len ) {
            break;
        } else {
            os_if_usleep(wait_time_us);
        }
    }

    buf[outlen] = '\0';
    /* Check that we read the needed number of characters */
    if ( outlen < str_len ) {
        return UA_STATUS_FAILURE;
    }

    /* Read the result untill the prompt char */
    for ( i=0; i < DUART_MAX_READ_BUF_LEN + 1; i++ ) {
        /* Read the output characters: one by one */
        for ( j = 0, outlen = 0; (j < DUART_READ_MAX_RETRY_NUM) && (outlen < 1); j++) {
            uart_stat = os_if_uart_read(h, buf + i, 1, &outlen);
            if ( uart_stat != UA_STATUS_OK ) {
                break;
            } else if ( outlen == 0 ) {
                os_if_usleep(wait_time_us);
            }
        }

        if ( ( uart_stat != UA_STATUS_OK ) || ( outlen != 1 ) ) {
            return UA_STATUS_FAILURE;
        }

        /* Parse the character */
        if ( isxdigit(buf[i]) ) {
            continue;
        } else if ( buf[i] == DUART_PROMPT_CHAR ) {
            /* Skip the leading prompt character */
            if ( i == 0 ) {
                i--;
                continue;
            }
            buf[i] = '\0';
            if (sscanf((char *)buf, "%x", val_p) != 1) {                
                return UA_STATUS_FAILURE;
            }
            return UA_STATUS_OK;
        } else if ( ( buf[i] == '\n' ) || ( buf[i] == '\r' ) ) {
            i--;
            continue;
        } else if( buf[i] == DUART_ERROR_CHAR ) {
            return UA_STATUS_FAILURE;
        }
          else {
            return UA_STATUS_FAILURE;
        }
    }

    /* We shouldn't get here: i == DUART_MAX_READ_BUF_LEN
       and end of input hasn't been reached */
    return UA_STATUS_FAILURE;
}

/**
 * Writes 32 bit to GRC using the debug UART port:
 *
 *  - Writes "w<addr> <val>\r\n", where "addr" and "val" is an offset in GRC in hex.
 *  - Reads the echoed symbols
 *  - Reads the charachters until ">" character is read.
 *  - If "?" was read - report an error otherwise report a success.
 */
ua_status_t dbg_os_duart_write(
    IN    ua_handle_t   h,
    IN    u32_t         addr,
    IN    u32_t         val,
    IN    u8_t          wait_for_echo
    )
{
    u8_t buf[DUART_MAX_READ_BUF_LEN + 1];
    ua_status_t uart_stat;
    u32_t str_len, echo_len = 0;
    u16_t outlen = 0;
    u32_t wait_time_us;    

    wait_time_us = (u32_t)((1000000.0* NUM_BITS_TO_WAIT)/((long double)os_if_uart_get_baud(h)));

    /* Create a command */
    sprintf((char *)buf, "w%x %x\r\n", addr, val);

    str_len = strlen((char *)buf);

    /* Write the command */
    uart_stat = os_if_uart_write(h, buf, str_len);
    if ( uart_stat != UA_STATUS_OK) {
        return uart_stat;
    }

    /* Incase writing to DBU special registers (config, timing) to change duart rate, do not wait for echo */
    if (!wait_for_echo) {
        return uart_stat;
    }

    /* Read the echo */
    while ( 1 ) {        
        outlen = 0;
        uart_stat = os_if_uart_read(h, buf, 1, &outlen);
        if ( uart_stat != UA_STATUS_OK ) {
            return uart_stat;
        }
        if ( outlen == 0 ) {            
            os_if_usleep(wait_time_us);
            continue;
        }
        if ( buf[0] == DUART_PROMPT_CHAR ) {
            if ( echo_len < str_len ) {
                echo_len = 0;
                continue;
            }
            break;
        } else if ( buf[0] == DUART_ERROR_CHAR ) {
            return UA_STATUS_FAILURE;
        }

        echo_len++;
    }


    return UA_STATUS_OK;
}
