#ifndef __DBG_OS__UTILS_H__
#define __DBG_OS__UTILS_H__

#include "edebug_types.h"
#include "dbg_os_uart.h"
#include "os_utils.h"

/* Memory allocation types */
//typedef u16_t  mem_type_t;
#define MEM_NONE                       0   /* None */
#define MEM_VIRT                       1   /* Virtual */
#define MEM_PHYS                       2   /* Physical */
#define MEM_MAP						   3   /* Mapping physical into user (virtual) address space */

#ifndef F_OK
#define F_OK 00
#endif
#ifndef X_OK
#define X_OK 01
#endif
#ifndef W_OK
#define W_OK 02
#endif
#ifndef R_OK
#define R_OK 04
#endif

#define cmn_os_alloc_virt	dbg_os_alloc_virt
#define cmn_os_free_virt	dbg_os_free_virt
#define cmn_os_memset    	dbg_os_memset
#define cmn_os_memcpy   	dbg_os_memcpy

/**
 * Attach the user defined handler to SIGINT signal.
 *
 * Returns:
 * 	DIAG_STATUS_SUCCESS if succeeded.
 */
static __inline dbg_status_t dbg_os_ctl_c_attach(
	IN   cb_t ctl_c_handler  /* Ctl-C handler */
	)
{
	return os_if_ctl_c_attach(ctl_c_handler);
}

/**
 * Description:
 *      Reads the register (DWORD) from PCI configuration space
 * using PCI configuration cycle from device with the given PCI address.
 *
 * Returns:
 *      DIAG_STATUS_SUCCESS if register was read successfully.
 */
__inline static dbg_status_t dbg_os_pcicfg_read(
    IN    u16_t   domain,
    IN    u8_t    bus,
    IN    u8_t    device,
    IN    u8_t    function,
    IN    u32_t   register_number,        /* register offset */
    OUT   u32_t * dword_read              /* output buffer */
    )
{

    return os_if_read_configuration_dword(domain, bus, device,function,register_number, dword_read);
}

static __inline dbg_status_t dbg_os_alloc_pcontig(
     IN    u32_t           nbytes,         /* size of the requested chunk*/
     OUT   dbg_virt_addr_t   * virt_addr_p,    /* virtual address of the allocated chunk */
     OUT   dbg_phys_addr_t   * phy_addr_p      /* physical chunk of the allocated chunk */
     )
{
    return os_if_alloc_pcontig(nbytes, virt_addr_p, phy_addr_p);
}

dbg_status_t dbg_os_free_pcontig(
     IN    dbg_virt_addr_t     virt_addr       /* virtual address of the chunk to release */
     );

dbg_status_t dbg_os_unmap_iomem (
     IN    dbg_virt_addr_t    virt_addr,        /* virtual adress of the memory chunk to release */
     IN    u32_t               nbytes            /* should be the same as in map_iomem */
     );

dbg_status_t dbg_os_map_iomem (
     IN    u32_t           nbytes,           /* the size of memory in bytes */
     IN    dbg_phys_addr_t     phy_addr,         /* physical address */
     OUT   dbg_virt_addr_t   * virt_addr_p       /* pointer to a virtual address pointer */
     );

static __inline void dbg_os_change_color(
    IN    u8_t color,        /* Color to set */
    IN    u8_t fore_or_back  /* Forground or background */
    )
{
	os_if_change_color(color,fore_or_back);
}

static __inline void dbg_os_restore_color(void)
{
        os_if_restore_color();
}

static __inline u32_t dbg_os_le32_to_cpu(
	u32_t val
	)
{
	return os_if_le32_to_cpu(val);
}

static __inline u32_t dbg_os_cpu_to_le32(
	u32_t val
	)
{
	return os_if_cpu_to_le32(val);
}

static __inline u32_t dbg_os_le16_to_cpu(
	u32_t val
	)
{
	return os_if_le16_to_cpu(val);
}

static __inline u32_t dbg_os_cpu_to_le16(
	u32_t val
	)
{
	return os_if_cpu_to_le16(val);
}

static __inline u32_t dbg_os_be32_to_cpu(
	u32_t val
	)
{
	return os_if_be32_to_cpu(val);
}

static __inline u32_t dbg_os_cpu_to_be32(
	u32_t val
	)
{
	return os_if_cpu_to_be32(val);
}

static __inline u32_t dbg_os_be16_to_cpu(
	u32_t val
	)
{
	return os_if_be16_to_cpu(val);
}

static __inline u32_t dbg_os_cpu_to_be16(
	u32_t val
	)
{
	return os_if_cpu_to_be16(val);
}

dbg_status_t dbg_os_get_total_dev(
    OUT   u32_t          * dev_found_p       /* number of found devices */
    );

dbg_status_t dbg_os_pci_scan(
    OUT   pci_info_t     * pci_info_ar      /* pci info array pointer to be filled */
    );

/**
 * Description:
 *      Suspends the execution (busy wait) for "cnt" miliseconds.
 *
 * Returns:
 *      void
 */
void dbg_os_msleep(
     IN    u32_t           cnt        /* number of msec to sleep */
     );

/**
 * Description:
 *      Suspends the execution (busy wait) for "cnt" mcroseconds.
 *
 * Returns:
 *      void
 */
void dbg_os_usleep(
     IN    u32_t           cnt        /* number of usec to sleep */
     );

/********* UART related functions */
/*
 * functions prototype
 */
ua_status_t
dbg_os_uart_init(
    void
    );

ua_status_t
dbg_os_uart_open(
    ua_handle_t* h,
    u8_t port,          /* 1 (COM1), 2 (COM2), 3, 4 */
    u32_t baud,         /* 9600, 19200, 38400, 57600, 115200 */
    char* setting,      /* "8N1", "7E1", "8O1" */
    u8_t mode           /* UAM_TEXT, UAM_BINARY */
    );

ua_status_t
dbg_os_uart_close(
    ua_handle_t h
    );

ua_status_t
dbg_os_uart_write(
    ua_handle_t h,
    u8_t* buf,
    u16_t len
    );

ua_status_t
dbg_os_uart_read(
    ua_handle_t h,
    u8_t* buf,
    u16_t len,
    u16_t* outlen
    );

void dbg_os_swap_bytes(
					u32_t  *buf,
					u32_t  buf_size
					);

#endif

void *dbg_os_alloc_virt(
	u32_t size);

void dbg_os_free_virt(
	int_ptr_t virt_addr);

void *dbg_os_memset(
	void  * buf,
	u8_t	ch,
	u32_t   buf_size
	);

void *dbg_os_memcpy(
	IN   void   	* dest,
	IN   const void * src,
	IN   u32_t  	  bnum
	);

								   

