#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "man_types.h"
#include "tcl_al_common.h"
#include "dbg_man.h"

/*
 *------------------------------------------------------------------------
 * Description:
 *     Prints the usage of the "assert" command, which reads from lm_device:
 *     last_index - latest assert index offset.
 *	   get_offset - the assert information offset (according to it's index)
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */

static void usage_assert()
{
	printf("Usage: assert last_index_offset <instance> \n");
	printf("# Get the latest assert index offset address.\n");
	printf("# Inputs: <instance> t (for tsem), x (for xsem), u (for usem) or c (for csem)\n\n");			
}
/*
 *------------------------------------------------------------------------
 * tcl_al_assert_usage --
 *     Add usage and help information on "assert" command 
 *------------------------------------------------------------------------
 */
int tcl_al_assert_usage(
	IN	char *buf,				//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{	
	u32_t offset = 0;
	*length = 0;

	*length += sprintf( &buf[offset+*length], "Read the assert last index offset and assert information offset\n\n");

	*length += sprintf( &buf[offset+*length], "Usage: assert last_index_offset <instance> \n");
	*length += sprintf( &buf[offset+*length], "Get the latest assert index offset address.\n");
	*length += sprintf( &buf[offset+*length], "Inputs: <instance> t (for tsem), x (for xsem), u (for usem) or c (for csem)\n\n");	
											  
	*length += sprintf( &buf[offset+*length], "Usage: assert get_offset <instance> <index> \n");
	*length += sprintf( &buf[offset+*length], "Get the assert information offset according to it's index.\n");	
	*length += sprintf( &buf[offset+*length], "Inputs: <instance> t (for tsem), x (for xsem), u (for usem) or c (for csem)\n");
	*length += sprintf( &buf[offset+*length], "         <index> \n\n");				  	

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_assert_last_index --
 *
 *     Implements the new Tcl "assert last_index" command.
 *     Get the latest assert index offset. 
 *
 * Inputs: 
 * objv[1] - instance x(for xsem, t (for tsem), u (for usem), c (for csem)
 *
 * Results:
 *     A standard Tcl result 
 *------------------------------------------------------------------------
 */
static int tcl_al_assert_last_index_offset (ClientData client_data, Tcl_Interp *interp, 
                  int objc, Tcl_Obj *CONST objv[])
{
	Tcl_Obj *result_ptr = NULL;	
	char*   instance;
	char result_str[16];   
	u32_t	last_index_offset;		
	dbg_status_t status;    

    
	/* Verify the number of arguments is 2 */
	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "<instance>");
		return TCL_ERROR;
	}
	
	result_ptr = Tcl_GetObjResult(interp);
	
	/* Get the instance from the first argument*/
	instance = Tcl_GetString(objv[1]);   	
	
	/* Get the last assert index offset */    		
	status = dbg_assert_get_last_assert_index_offset(instance, &last_index_offset);
	if (status != DBG_STATUS_SUCCESS) {
		Tcl_AppendStringsToObj(result_ptr, "Failed to use dg_lmdev_get_last_assert_index_offset", (char *)NULL);
		return TCL_ERROR;
	}
	else {
		 sprintf(result_str, "0x%x", last_index_offset);
		 Tcl_SetStringObj(result_ptr, result_str, strlen(result_str));
		 return TCL_OK;				
	}           
}

/*
 *------------------------------------------------------------------------
 * Description:
 *     Implements the "assert" command, which reads from the lm_device: 
 *     last_index - latest assert index offset.
 *	   get_offset - the assert information offset (according to it's index)
 * 
 * Returns:
 *     A standard Tcl result.
 *------------------------------------------------------------------------
 */
int tcl_al_assert_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{ 
	static char *option_names[] = { "last_index_offset", "help", (char *)NULL };
	int i;

	enum options {LAST_INDEX, GET_OFFSET, HELP};
	int index;	
	
	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "option=(last_index_offset| get_offset|help) arg...");
		return TCL_ERROR;
	}
	
	if (Tcl_GetIndexFromObj(interp, objv[1], (const char**)option_names, "option", TCL_EXACT, &index) != TCL_OK) {
		return TCL_ERROR;
	}
	
	switch ((enum options)index) {
	case LAST_INDEX:
	    return tcl_al_assert_last_index_offset(client_data, interp, objc-1, objv+1); 	
	case HELP:
	    usage_assert();
	    return TCL_OK;
	default:
	    return TCL_ERROR;
	}
	
	return TCL_OK; 
}


