#include <string.h>
#include "man_types.h"
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "man_types.h"
#include "tcl_al_common.h"
#include "esc_seq.h"
#include "hw_access.h"

extern edebug_eng_mode_t eng_mode;

typedef enum {
    DEFAULT_CMD = 0,
    SELECT_CMD,
    LIST_CMD,
    LIST_INCLUDING_HIDDEN_CMD,
    VF_SHOW_CMD,
    REFRESH_CMD,
    SELECT_BDF_CMD,
    INVALID_PARAMS
} device_cmd_type_t;

void tcl_al_update_env(Tcl_Interp *interp, u32_t dev_num);
void disp_dev_title(void);
void disp_dev_info(Tcl_Interp *interp, int dev_num );

/* holds num of QLogic unhidden devices */
static int num_not_hidden_devices=-1;

static void usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "device <devNum> or device bdf <domain> <bus> <device> <function>");
}

/*
 *------------------------------------------------------------------------
 * tcl_al_device_usage --
 *     Add usage information on "device" command
 *------------------------------------------------------------------------
 */
int tcl_al_device_usage(
                       IN  char *buf,              //Pointer to the buffer of the command usage
                       IN  u32_t *length           //Pointer to the length result of the command usage
                       )
{
    u32_t offset = 0;
    *length = 0;

    *length += sprintf( &buf[offset+*length], "Display brief information on all devices or on a selected device\n\n");
    *length += sprintf( &buf[offset+*length], "Usage: device [<dev_num>|-a|bdf <bus> <device> <function>]\n" );
    *length += sprintf( &buf[offset+*length], "If -a option is given, the command will display all devices\n");
    *length += sprintf( &buf[offset+*length], "     including hidden devices\n");
    *length += sprintf( &buf[offset+*length], "If bdf option is given, the command will go to the device\n");
    *length += sprintf( &buf[offset+*length], "     with supplied BDF values, if can find it\n");
    *length += sprintf( &buf[offset+*length], "\n");
    *length += sprintf( &buf[offset+*length], "Notes: \n");
    *length += sprintf( &buf[offset+*length], "Without any arguments, the command displays information on all devices.\n" );
    *length += sprintf( &buf[offset+*length], "When denNum is provided, the command displays information regarding the selected device.\n" );
    return TCL_OK;
}

/**
 *
 * @param interp
 * @param objc Number of parameters minus 1 (real parameters).
 * @param objv Actual parameters of the command (without command
 *             name itself.
 *
 * @return Command type (device_cmd_type_t).
 */
static device_cmd_type_t device_parse_cmd(
                                         IN    Tcl_Interp *interp,     /* current interpreter */
                                         IN    int objc,               /* number of arguments */
                                         IN    Tcl_Obj *CONST objv[]   /* argument strings */
                                         )
{
    char *option;

    if (objc == 0) {
        /* device */
        return DEFAULT_CMD;
    } else {
        /* device -l */
        option = Tcl_GetString(objv[0]);
        if (strcmp(option, "-l") == 0)
            return LIST_CMD;

        if (strcmp(option, "-a") == 0)
            return LIST_INCLUDING_HIDDEN_CMD;

        /* device refresh */
        if (strcmp(option, "refresh") == 0)
            return REFRESH_CMD;

        /* device vf_show */
        if (strcmp(option, "vf_show") == 0)
            return VF_SHOW_CMD;

        /* device vf_show */
        if (strcmp(option, "bdf") == 0)
            return SELECT_BDF_CMD;

        /* device N */
        if (objc == 1) {
            return SELECT_CMD;
        }

        if (objc == 2) {
            option = Tcl_GetString(objv[1]);
            if (strcmp(option, "-no_display") == 0) {
                return SELECT_CMD;
            }
        }
    }

    return INVALID_PARAMS;
}

static int device_default_cmd(
                             IN    Tcl_Interp *interp     /* current interpreter */
                             )
{
    int dev_num = -1;
    u32_t no_total_br_dev=0;

    no_total_br_dev = dbg_get_total_br_dev();

    disp_dev_title();

    num_not_hidden_devices = 0;

    for (dev_num = 0; (u32_t)dev_num < no_total_br_dev; dev_num++) {
        /* PCI mode */
        if ((tcl_al_get_no_pci_mode() == 0) && (eng_mode != EDEBUG_VIRTUAL_NVM_MODE)) {
            if (BR_PCI_DEV_PRESENT(dev_num) && !BR_PCI_DEV_HIDDEN(dev_num)) {
                num_not_hidden_devices++;
                disp_dev_info(interp, dev_num);
            }
        }
        /* non PCI mode - valid devices are consecutive in the array */
        else {
            if (!BR_PCI_DEV_HIDDEN(dev_num)) {
                num_not_hidden_devices++;
                disp_dev_info(interp, dev_num);
            }
        }
    }

    return TCL_OK;
}

static int device_list_including_hidden_cmd(
                                           IN    Tcl_Interp *interp     /* current interpreter */
                                           )
{
    int dev_num = -1;
    u32_t no_total_br_dev=0;

    no_total_br_dev = dbg_get_total_br_dev();

    disp_dev_title();

    for (dev_num = 0; (u32_t)dev_num < no_total_br_dev; dev_num++) {
        if ( BR_PCI_DEV_PRESENT(dev_num) || tcl_al_get_no_pci_mode()) {
            disp_dev_info(interp, dev_num);
        }
        set_tcl_prompt(interp, EDEBUG_ENG_MODE);
    }
    return TCL_OK;
}



static int device_set_according_to_bdf(
                                      IN    Tcl_Interp *interp,     /* current interpreter */
                                      IN    int objc,               /* number of arguments */
                                      IN    Tcl_Obj *CONST objv[]   /* argument strings */
                                      )
{
    int res, i;
    int domain=-1, bus=-1, device=-1, function=-1;
    u16_t cur_domain, orig_domain;
    u8_t cur_bus, cur_device, cur_function;
    u8_t orig_bus, orig_device, orig_function;
    char cmd_str[64];

    if (objc < 5 ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("Must receive 4 numbers representing domain/bus/device/function.", -1));                               
        return TCL_ERROR;
    }

    res = Tcl_GetIntFromObj(interp, objv[1], &domain);
    if ( res != TCL_OK ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("Must receive 4 numbers representing domain/bus/device/function.", -1));                               
        return TCL_ERROR;
    }

    res = Tcl_GetIntFromObj(interp, objv[2], &bus);
    if ( res != TCL_OK ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("Must receive 4 numbers representing domain/bus/device/function.", -1));                               
        return TCL_ERROR;
    }
    res = Tcl_GetIntFromObj(interp, objv[3], &device);
    if ( res != TCL_OK ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("Must receive 4 numbers representing domain/bus/device/function.", -1));                               
        return TCL_ERROR;
    }
    res = Tcl_GetIntFromObj(interp, objv[4], &function);
    if ( res != TCL_OK ) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("Must receive 4 numbers representing domain/bus/device/function.", -1));                               
        return TCL_ERROR;
    }

    /* get current device values */
    dbg_pcicfg_getdut(&orig_domain, &orig_bus, &orig_device, &orig_function);

    /* search for the relevant device */
    for (i = 1; i <= dbg_get_total_br_dev(); i++) {
        sprintf(cmd_str, "device %d", i);
        res = Tcl_Eval(interp, cmd_str);
        if (res == TCL_OK) {
            dbg_pcicfg_getdut(&cur_domain, &cur_bus, &cur_device, &cur_function);
            if ((cur_domain == domain) && (cur_bus == bus) && (cur_device == device) && (cur_function == function)) {
                break;
            }
        }
    }

    if (i > dbg_get_total_br_dev()) {
        /* go back to original device */
        sprintf(cmd_str, "device bdf %d %d %d %d", orig_domain, orig_bus, orig_device, orig_function);
        res = Tcl_Eval(interp, cmd_str);

        /* return error */
        Tcl_SetObjResult(interp, Tcl_NewStringObj("Could not find device with that BDF", -1));                               
        return TCL_ERROR;
    }

    return TCL_OK;        
}


static int device_select_cmd(
                            IN    Tcl_Interp *interp,     /* current interpreter */
                            IN    int objc,               /* number of arguments */
                            IN    Tcl_Obj *CONST objv[]   /* argument strings */
                            )
{
    char *str;
    int pf_dev = -1, vf_dev = -1, res;
    u32_t dev_num;

    str = Tcl_GetString(objv[0]);
    if (!str)
        return TCL_ERROR;

    if (strchr(str, '.')) {
        if (sscanf(str, "%d.%d", &pf_dev, &vf_dev) != 2) {
            Tcl_SetObjResult(interp, Tcl_NewStringObj("Bad format in argument", -1));
            return TCL_ERROR;
        }

        //BW
#if 0

        if ((pf_dev < 1) || ((u32_t)pf_dev > dg_get_total_br_dev()) ||
            (vf_dev < 1) || (vf_dev > dg_get_vfs_num(pf_dev - 1))) {
            Tcl_SetObjResult(interp, Tcl_NewStringObj("Invalid device number", -1));
            return TCL_ERROR;
        }

        if (dg_select_vf_dev_num(pf_dev - 1, vf_dev - 1) != DIAG_STATUS_SUCCESS)
            goto error_exit;

        dev_num = VF_NUM_LONG(pf_dev, vf_dev);
#endif
    } else {
        if (!strchr(str, 'H')) {
            res = Tcl_GetIntFromObj(interp, objv[0], &pf_dev);
            if (( res != TCL_OK ) || ( pf_dev < 1 ) || ((u32_t)pf_dev > dbg_get_total_br_dev())) {
                Tcl_SetObjResult(interp, Tcl_NewStringObj("Invalid device number.", -1));

                /* Refresh the prompt */
                set_tcl_prompt(interp, EDEBUG_ENG_MODE);

                return TCL_ERROR;
            }
        } else {
            /* if device is hidden, you can select it with H postfix, for exmaple: "device 3H" */
            if ((strlen(str) != 2) ||
                (sscanf(str, "%dH", &pf_dev) != 1) ||
                ( pf_dev < 1 ) ||
                (((u32_t)pf_dev > dbg_get_total_br_dev())) ||
                (!BR_PCI_DEV_HIDDEN(pf_dev-1))) {
                Tcl_SetObjResult(interp, Tcl_NewStringObj("Invalid device number.", -1));
                return TCL_ERROR;
            }
        }


        if (dbg_select_dev_num(pf_dev - 1) != DBG_STATUS_SUCCESS)
            goto error_exit;
    }

    dev_num = pf_dev;
	
    /* Update environment settings according to the new selected device */
    tcl_al_update_env(interp, pf_dev - 1);

    /* Refresh the prompt */
    set_tcl_prompt(interp, EDEBUG_ENG_MODE);

    if (BR_PCI_DEV_HIDDEN(pf_dev-1) && !tcl_al_get_no_pci_mode()) {
        return TCL_ERROR;
    } else {
        return TCL_OK;
    }

    error_exit:
#ifndef WIN_DIAG
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Failed to select the device: check device number", -1));
#else
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Failed to select the device: check device is installed and enabled", -1));
#endif
    return TCL_ERROR;
}

/**
 * Description:
 *      "device" command cleanup handler.
 * Returns:
 *      A standard TCL result.
 */
int tcl_al_device_cleanup(
                         IN   Tcl_Interp *interp
                         )
{
    return Tcl_DeleteCommand(interp, DEVICE_CMD);;
}

int tcl_al_device_cmd (
                      IN    ClientData client_data, /* Not used. */
                      IN    Tcl_Interp *interp,     /* current interpreter */
                      IN    int objc,               /* number of arguments */
                      IN    Tcl_Obj *CONST objv[]   /* argument strings */
                      )
{
    static u8_t first_call = 1;
    device_cmd_type_t cmd_type;

    /* Setup all device, this will also collect device info */
    if ( first_call ) {
        //BW dbg_setup_all();
        first_call = 0;
    }

    /* Parse command type */
    cmd_type = device_parse_cmd(interp,objc - 1,objv + 1);

    switch (cmd_type) {
    case DEFAULT_CMD:
        return device_default_cmd(interp);
    case LIST_CMD:
        // return device_list_cmd(interp);
    case LIST_INCLUDING_HIDDEN_CMD:
        return device_list_including_hidden_cmd(interp);
    case SELECT_CMD:
        return device_select_cmd(interp, objc - 1, objv + 1);
    case SELECT_BDF_CMD:
        return device_set_according_to_bdf(interp, objc - 1, objv + 1);
    case VF_SHOW_CMD:
        // return device_vf_show(interp);
    case REFRESH_CMD:
        // return device_refresh(interp);
    default:
        return TCL_ERROR;
    }

    return TCL_OK;
}

void tcl_al_update_env(Tcl_Interp *interp, u32_t dev_num)
{
    Tcl_Obj *obj_ptr;
    char tmp_str[16];

    /* DEV holds dev_num starting from 1 and not from 0 */
    sprintf(tmp_str, "%d", dev_num+1);
    Tcl_SetVar2(interp, TCL_ENV, ENV_DEV, tmp_str, TCL_GLOBAL_ONLY);

    /* Set the $::current(BUS) */
    sprintf(tmp_str, "0x%x", BR_DEV_INFO(dev_num)->pci_info.pci_bus);
    obj_ptr = Tcl_NewStringObj(tmp_str, strlen(tmp_str));
    Tcl_SetVar2Ex(interp, TCL_ENV, ENV_BUS, obj_ptr, 0);

    /* Set the $::current(DID) */
    sprintf(tmp_str, "0x%x", BR_DEV_INFO(dev_num)->pci_info.device_id);
    obj_ptr = Tcl_NewStringObj(tmp_str, strlen(tmp_str));
    Tcl_SetVar2Ex(interp, TCL_ENV, ENV_DID, obj_ptr, 0);

    /* Set the $::current(DEFAULT_DID) only if it's value is valid */
    sprintf(tmp_str, "%d", BR_DEV_INFO(dev_num)->pci_info.device_id);
    Tcl_SetVar2(interp, TCL_ENV, "DEFAULT_DID", tmp_str, TCL_GLOBAL_ONLY);

    /* Set the $::current(DEFAULT_DID) only if it's value is valid */
    sprintf(tmp_str, "%s", BR_DEV_INFO(dev_num)->pci_info.revision);
    Tcl_SetVar2(interp, TCL_ENV, ENV_CHIP_REV, tmp_str, TCL_GLOBAL_ONLY);

    /* Set the $::current(FUNC) */
    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sprintf(tmp_str, "%d", 0);
    } else {
        sprintf(tmp_str, "%d", BR_DEV_INFO(dev_num)->pci_info.pci_func / dbg_get_num_paths());
    }
    obj_ptr = Tcl_NewStringObj(tmp_str, strlen(tmp_str));
    Tcl_SetVar2Ex(interp, TCL_ENV, ENV_FUNC, obj_ptr, 0);

    /* Set the $::current(PCI_FUNC) */
    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sprintf(tmp_str, "%d", 0);
    } else {
        sprintf(tmp_str, "%d", BR_DEV_INFO(dev_num)->pci_info.pci_func);
    }
    obj_ptr = Tcl_NewStringObj(tmp_str, strlen(tmp_str));
    Tcl_SetVar2Ex(interp, TCL_ENV, ENV_PCI_FUNC, obj_ptr, 0);

    /* Set the $::current(PATH) */
    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sprintf(tmp_str, "%d", 0);
    } else {
        sprintf(tmp_str, "%d", BR_DEV_INFO(dev_num)->pci_info.pci_func & 0x1);
    }
    obj_ptr = Tcl_NewStringObj(tmp_str, strlen(tmp_str));
    Tcl_SetVar2Ex(interp, TCL_ENV, ENV_PATH, obj_ptr, 0);

    /* Set the $::current(DRV_STATE) */
    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sprintf(tmp_str, "%s", "NONE");
    } else {
        sprintf(tmp_str, "%s", "RUN");
    }
    obj_ptr = Tcl_NewStringObj(tmp_str, strlen(tmp_str));
    Tcl_SetVar2Ex(interp, TCL_ENV, ENV_DRV_STATE, obj_ptr, 0);

    /* Set the NVM_SIZE */
    if (tcl_al_get_no_pci_mode() == 0)
        sprintf(tmp_str, "%d", common_nvm_nvm_size() & 0xfffff000);
    else
        sprintf(tmp_str, "%d", 0);


    Tcl_SetVar2(interp, TCL_ENV, ENV_NVM_SIZE, tmp_str, TCL_GLOBAL_ONLY);

    /* Chip Number */
    Tcl_SetVar2(interp, TCL_ENV, ENV_CHIP_NUM, BR_DEV_INFO(dev_num)->pci_info.asic_rev, 0);

    /* Set the $::current(TOTAL_DEV) */
    sprintf(tmp_str, "%d", dbg_get_total_br_dev());

    Tcl_SetVar2(interp, TCL_ENV, ENV_TOTAL_DEV, tmp_str, TCL_GLOBAL_ONLY);


    if (common_nvm_tcl_al_get_no_pci_mode() == 0 && !BR_PCI_DEV_HIDDEN(dev_num)) {
        /* Set the $::current(IS_VPD_IMAGE) */
        sprintf(tmp_str, "%d", is_vpd_image_exist());
        Tcl_SetVar2(interp, TCL_ENV, ENV_IS_VPD, tmp_str, TCL_GLOBAL_ONLY);
    }
}

void disp_dev_title(void)
{
    int i;

    if (eng_mode == EDEBUG_VIRTUAL_NVM_MODE) {
        printf("%-2s: %6s %s %s %s %s\n",
               "C ","Brd", " path", "port", "dev", "        MAC       ");
    } else if (tcl_al_get_no_pci_mode() == 0) {
        printf("%-2s: %6s:%-5s %-11s %-4s %5s %s %s %s %s %s %s %s %s\n",
               "C ","Brd", "Rv", "    Bus","PCI", "  Spd    ","       MAC       ", " bc ver    ", "Config.   ", "Serial Number  ",
               "i/f    ", "st.", "drv ver   ", "driver name");
    } else {
        printf("%-2s: %6s:%s %s %s %s %s %s %s\n",
               "C ","Brd", "Rv", "path", "port", "dev", "        MAC       ", "Config. ", "Serial Number  "
              );
    }
}

/*-----------------------------------------------------------------------------
 * disp_dev_info --
 *
 *     Give the detailed information about a specific device.
 *
 * ----------------------------------------------------------------------------
 */
void disp_dev_info(Tcl_Interp *interp, int dev_num )
{
    pci_info_t *pci_info_p = NULL;
    char buf[10], bdf[16], dev_num_str[64], configuration[64], eth_name[64];
    char full_config[256];
    u16_t vfs_num = 0;
    int br_dev_num = dev_num;
    FILE *fd;       
	u32_t silent_rev, val;
	char *rev_id_ctrl; 

    /* in PCI mode, get index of QLogic device within global array of all devices */
    if (tcl_al_get_no_pci_mode() == 0) {
        br_dev_num = DBG_DEV_INDEX(dev_num);
    }

    /* get pci_info from the real index */
    pci_info_p = &(dbg_pci_devices[br_dev_num].pci_info);

    if (pci_info_p == NULL) {
        printf ("Invalid device number %d. Check configuration.\n", dev_num);
        return;
    }

    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sc_change_color(ESC_COLOR_BROWN, ESC_FOREGROUND);
    } else {
        sc_restore_color();
    }

    sprintf(bdf, "%02x:%02x:%02x:%02x", pci_info_p->pci_domain, pci_info_p->pci_bus, pci_info_p->pci_device, pci_info_p->pci_func);

    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sprintf(dev_num_str, "%dH", dev_num+1);
    } else {
        sprintf(dev_num_str, "%d", dev_num+1);
    }

    if (pci_info_p->pxe_enable==1) {
        sprintf(buf, "M%c,", pci_info_p->pxe_agent);
    } else {
        buf[0]='\0';
    }

    sprintf(configuration, "%s%s%s%s",
            (pci_info_p->wol_enable==1) ? "W,":"",
            buf,
            (pci_info_p->mfw_enable==1) ? "MF,":"",
            pci_info_p->pxe_speed);

    if (BR_PCI_DEV_HIDDEN(dev_num)) {
        sprintf(configuration, "XXXXXXX");
        sprintf(pci_info_p->mac_addr, "XX:XX:XX:XX:XX:XX");
        sprintf(pci_info_p->serial_number, "XXXXXXX");
    }


    /* display eth_name only if it's not too long */
    if (strlen(pci_info_p->eth_name) > 7) {
        sprintf(eth_name, "long...");
    } else {
        strcpy(eth_name, pci_info_p->eth_name);
    }

    /* allow maximum 10 chars of driver_ver to be shown */
    pci_info_p->driver_ver[10]='\0';

	/* CQ73635 nvm cfg 233 - Revision ID control */
	dbg_reg_read(MISC_REG_CHIP_TEST_REG, &val);
	silent_rev = val & 0xff;

	if ((silent_rev != 0) && 
		((pci_info_p->device_id == CHIP_NUM_57810) ||
		 (pci_info_p->device_id == CHIP_NUM_57810_MF) ||
		 (pci_info_p->device_id == CHIP_NUM_57811) ||
		 (pci_info_p->device_id == CHIP_NUM_57811_MF) ||
		 (pci_info_p->device_id == CHIP_NUM_57811_VF) ||
		 (pci_info_p->device_id == CHIP_NUM_57800) ||
		 (pci_info_p->device_id == CHIP_NUM_57800_MF))) {

		/* Update environment settings according to the new selected device */
		tcl_al_update_env(interp, dev_num);
		nvm_cfg_get(interp, IDX_REV_ID_CTRL);
		rev_id_ctrl = Tcl_GetString(Tcl_GetObjResult(interp));
		
		if (strcmp(rev_id_ctrl, "0") == 0) {
			/* Preserve - do nothing */
		} else if (strcmp(rev_id_ctrl, "1") == 0) {
			/* Display actual revision ID (silent rev) */
			strcpy(pci_info_p->revision, "B1");
		} else if (strcmp(rev_id_ctrl, "2") == 0) {
			/* Force B0 */
			strcpy(pci_info_p->revision, "B0");
		} else if (strcmp(rev_id_ctrl, "3") == 0) {
			/* Force B1 */
			strcpy(pci_info_p->revision, "B1");
		}
	}

    if (eng_mode == EDEBUG_VIRTUAL_NVM_MODE) {
        sprintf(full_config,
                "%-2s: %5s:   %-4d %-4d %-3d  %s\n",
                dev_num_str,
                pci_info_p->asic_rev,
                pci_info_p->path_num,
                pci_info_p->port_id,
                pci_info_p->device_num_in_port,
                pci_info_p->mac_addr
               );          
    } else if (tcl_al_get_no_pci_mode() == 0) {
        sprintf(full_config, 
                "%-2s: %6s:%-5s %s %s %-7s %-18s %-10s %-10s %-15s %-7s D%d  %-10s %s\n",
                dev_num_str,
                pci_info_p->asic_rev,
                pci_info_p->revision,
                bdf,
                pci_info_p->pcie_lane_width,
                pci_info_p->pcie_lane_speed,
                pci_info_p->mac_addr,
                pci_info_p->bc_ver,
                configuration,
                pci_info_p->serial_number,
                eth_name,
                pci_info_p->eth_state,
                pci_info_p->driver_ver,
                pci_info_p->driver_name
               );
    } else {
        sprintf(full_config,
                "%-2s: %5s:%s   %-4d %-4d %-3d  %s %-8s %-15s\n",
                dev_num_str,
                pci_info_p->asic_rev,
                pci_info_p->revision,
                pci_info_p->path_num,
                pci_info_p->port_id,
                pci_info_p->device_num_in_port,
                pci_info_p->mac_addr,
                configuration,
                pci_info_p->serial_number
               );                
    }
    printf(full_config);

    if (tcl_al_get_special_log_mode()) {
        fd = fopen(SPECIAL_LOG_FILE, "a");
        if (fd) {
            fseek (fd, 0L, SEEK_END);
            if (dev_num == 0) {
                fprintf(fd, "device list output\n");
                fprintf(fd, "==================\n\n");
            }
            fprintf(fd, "%s", full_config);
        }
        fclose(fd);
    }
}












