#include <string.h>
#include <stdlib.h>
#include "bcmtype.h"
#include "tcl.h"
#include "tcl_al_cmd_handlers.h"
#include "tcl_al.h"
#include "tcl_help.h"
#include "dbg_os_utils.h"

#define MAX_HELP_LEN	        8192		        // maximum length of the help on all commands
#define MAX_CMDS		100			// maximum number of eDebug commands

/* Global arrays of cmd_names: command names and cmd_help_ptrs: pointers to each command help string */
static char *cmd_names[MAX_CMDS];
static char *cmd_help_ptrs[MAX_CMDS];	

/* The current index of the command to be inserted to the cmd_names[] and cmd_help_ptrs[] arrays */
static u32_t	current_idx=0;				

static void usage(Tcl_Interp *interp);
/**
 * Description:
 *     Initialize the cmd_names[], cmd_help_ptrs[] arrays 
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
int tcl_al_init_help_array (void) 
{
	memset(cmd_names, 0, sizeof(char*)*MAX_CMDS);
	memset(cmd_help_ptrs, 0, sizeof(char*)*MAX_CMDS);

	return TCL_OK;
}

/**
 * Description:
 *     Iterates over cmd_names and searches for cmdName
 * 
 * Returns:
 *     1 - if cmdName exists in cmd_names
 *     0 - otherwise
 */ 
static __inline int tcl_al_cmd_exists(	
	IN    char *cmdName)
{
	u32_t i;
	for (i=0; i< current_idx; i++) {
		if (strcmp(cmd_names[i], cmdName) == 0) {
			return 1;
		}
	}
	return 0;
}

/**
 * Description:
 *     Adds the usage information of the command to cmd_names[], cmd_help_ptrs[] arrays
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
int tcl_al_add_usage_info (
    IN char *cmdName,				// Name of the command, to add help about
    IN help_callback_t cb_proc,			// Callback procedure to be used to add help and usage on this command
    IN u32_t help_len				// Length of the help and usgae info string
    )
{   
	u32_t len;	

	if(tcl_al_cmd_exists(cmdName))
	   return TCL_OK;

	/* Allocate memory for the command name */
	cmd_names[current_idx] = (char *) malloc(sizeof(u8_t)*(strlen(cmdName) + 1));	

	/* copy the command name into the cmd_names[] array */
	strcpy( cmd_names[current_idx], cmdName);	

	/* Allocate memory for the help and usage string */
	cmd_help_ptrs[current_idx] = (char *) malloc(help_len + 1);	

	/* Use the callback function to update the usage buffer of this command in the cmd_help_ptrs[] array */
	cb_proc( cmd_help_ptrs[current_idx], &len);	

	current_idx++;

	return TCL_OK;
}


/**
 * Description:
 *     Implements "help" command.
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
int tcl_al_help_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{   		
	char	*name_ptr;				//local pointer used to copy the name string to local help buffer
	char	*usage_ptr;				//local pointer used to copy the usage string to local buffers
	char	usage_buf[MAX_USAGE_LEN];		//local buffer for the request command usage
	char	help_buf[MAX_HELP_LEN];			//local buffer for help on whole commands
		
	int	cmd_idx=-1;				//The command index of the command we are looking for help
							//in the cmds_help[] array
	u32_t	index=0, i=0;				
	Tcl_Obj	*ret_obj;


	/**
	* Check the number of arguments - if no arguments, print help on all commands
	*/ 
	if (objc < 2) {    
		for (cmd_idx = 0; (cmd_idx < MAX_CMDS) && (cmd_names[cmd_idx] != NULL) ; cmd_idx++)
		{	
			/* Copy the command name to the help buffer */
			i=0;			
			name_ptr = cmd_names[cmd_idx];			
			while (*name_ptr != '\0')
			{				            
				help_buf[index++] = *name_ptr++;                								
			}			
			help_buf[index++] = '\t';				
			
			/* Copy the first line of the usage of the command to the help buffer */
			i=0;            
			usage_ptr = cmd_help_ptrs[cmd_idx];
			while (*usage_ptr != '\n')
			{					                
				help_buf[index++] = *usage_ptr;                												
				usage_ptr++;
			}
			help_buf[index++] = '\n';
		}
		help_buf[index]='\0';
		
		/* Now return the help string back to the console */
		ret_obj = Tcl_GetObjResult( interp );
		Tcl_AppendStringsToObj( ret_obj, help_buf, NULL );        
		return TCL_OK;
	}

	/**
	* Find the command index, of the command we are looking for, in the cmds_help[] array
	*/ 	
	if (Tcl_GetIndexFromObj(interp, objv[1], (const char **)cmd_names, "command", TCL_EXACT, &cmd_idx) != TCL_OK) {
		Tcl_AppendResult(interp, "Unknown command", NULL);
		return TCL_ERROR;
		}    
	else {
		/* If found the command in cmd_names[] array, print help and usage of the specific command */
		sprintf( &usage_buf[0], "%s\n", cmd_help_ptrs[cmd_idx]);		
        		
		/* Now return the help string back to the console */
		ret_obj = Tcl_GetObjResult( interp );
		Tcl_AppendStringsToObj( ret_obj, usage_buf, NULL );
		return TCL_OK;		
	}
}

int tcl_al_help_cleanup(
    IN   Tcl_Interp *interp
    )
{
	u32_t cmd_idx;

	Tcl_DeleteCommand(interp, HELP_CMD);

	/* Free the allocated memory for cmd_names[] array */
	for (cmd_idx = 0; (cmd_names[cmd_idx] != NULL) && (cmd_idx < MAX_CMDS); cmd_idx++) {
		free ((char *) cmd_names[cmd_idx]);
	}

	/* Free the allocated memory for cmd_help_ptrs[] array */
	for (cmd_idx = 0; (cmd_help_ptrs[cmd_idx] != NULL) && (cmd_idx < MAX_CMDS); cmd_idx++) {
		free ((char *) cmd_help_ptrs[cmd_idx]);
	}

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_help_usage --
 *     Add usage information on "add_help" command 
 *------------------------------------------------------------------------
 */
int tcl_al_help_usage(
	IN	char *buf,				//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{	
	u32_t offset = 0;
	*length = 0;

	*length += sprintf( &buf[offset+*length], "Get help information on a command (also for TCL commands)\n\n");
	*length += sprintf( &buf[offset+*length], "Usage: help <comand name> \n" );
        
	*length += sprintf( &buf[offset+*length], "Note: When no arguments, help will print the list of commands and their description.\n" );    
	*length += sprintf( &buf[offset+*length], "Help will include also TCL commands of scripts which have been sourced.\n" );    
	  
	return TCL_OK;
}

/**
 * Description:
 *     Implements "add_help" command, which adds help information on a new command.
 *	   Used for adding a new command from a TCL script and help information about it.
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
int tcl_al_add_help_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{   		
	char *cmdName;
	char *cmdHelp;
	int help_len;

	/**
	* Verify the number of arguments include: add_help <command name> <help information>
	*/ 
	if (objc < 3) {
		usage(interp);
		return TCL_ERROR;        
	}

	/* Get the command name and help information from the input arguments */
	cmdName = Tcl_GetString(objv[1]);	
	cmdHelp = Tcl_GetString(objv[2]);
	help_len = sizeof(u8_t)*(strlen(cmdHelp) + 1);


	if(tcl_al_cmd_exists(cmdName))
		return TCL_OK;

	/* Allocate memory for the command name */
	cmd_names[current_idx] = (char *) malloc(sizeof(u8_t)*(strlen(cmdName) + 1));

	/* copy the command name into the cmd_names[] array */
	strcpy( cmd_names[current_idx], cmdName);	

	/* Allocate memory for the command usage string */
	cmd_help_ptrs[current_idx] = (char *) malloc(help_len + 1);
	
	/* copy the help information pointer into the cmd_help_ptrs[] array */
	strcpy( cmd_help_ptrs[current_idx], cmdHelp);	

	current_idx++;

	return TCL_OK;
}

static void usage(Tcl_Interp *interp)
{
	Tcl_WrongNumArgs(interp, 0, NULL, "add_help <command_name> <help string>");
}

int tcl_al_add_help_cleanup(
    IN   Tcl_Interp *interp
    )
{	

	Tcl_DeleteCommand(interp, ADD_HELP_CMD);
	
	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_add_help_usage --
 *     Add usage information on "add_help" command 
 *------------------------------------------------------------------------
 */
int tcl_al_add_help_usage(
	IN	char *buf,			//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{	
	u32_t offset = 0;
	*length = 0;

	*length += sprintf( &buf[offset+*length], "Add help information on a new command (usally for TCL commands)\n\n");
	*length += sprintf( &buf[offset+*length], "Usage: add_help <comand name> <help string>>\n" );
        
	*length += sprintf( &buf[offset+*length], "Note: The first line should include a one-line help information.\n" );    
	*length += sprintf( &buf[offset+*length], "Next, will include the usage information.\n" );    
	  
    return TCL_OK;
}


