/*************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved.
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without
 * the prior written consent of QLogic Corporation.
 *
 *
 *  Name: tcl_hex.c
 *
 *  Description: Implementation of the "hex16" and "hex32" commands,
 *				 Which are used to display value in hexdecimal format.
 *
 *
 ****************************************************************************/

#include <string.h>
#include "man_types.h"
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "tcl_al_common.h"
#include "util.h"
#include "disputil.h"
#include "dbg_man.h"


/* global vars */

/* hex command */
/*
 *------------------------------------------------------------------------
 * tcl_al_hex32_cmd --
 *
 *     Implements the new Tcl "hex32" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
int
tcl_al_hex32_cmd(client_data, interp, objc, objv)
    ClientData client_data; /* Not used. */
    Tcl_Interp *interp;     /* current interpreter */
    int objc;               /* number of arguments */
    Tcl_Obj *CONST objv[];  /* argument strings */
{
    char buf[20];
    u32_t value;

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "value");
        return TCL_ERROR;
    }

    /* Get the value */
    if (Tcl_GetIntFromObj(interp, objv[1], (int *)&value) == TCL_ERROR)
    {
        return TCL_ERROR;
    }

    sprintf(buf, "0x%08x", value);

    Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, -1));
    return TCL_OK;

}

/*
 *------------------------------------------------------------------------
 * tcl_al_hex16_cmd --
 *
 *     Implements the new Tcl "hex16" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     The returning value will have the upper 16 bits masked off.
 *------------------------------------------------------------------------
 */
int
tcl_al_hex16_cmd(client_data, interp, objc, objv)
    ClientData client_data; /* Not used. */
    Tcl_Interp *interp;     /* current interpreter */
    int objc;               /* number of arguments */
    Tcl_Obj *CONST objv[];  /* argument strings */
{
    char buf[20];
    u32_t value;

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "value");
        return TCL_ERROR;
    }

    /* Get the value */
    if (Tcl_GetIntFromObj(interp, objv[1], (int *)&value) == TCL_ERROR)
    {
        return TCL_ERROR;
    }

    sprintf(buf, "0x%04x", value & 0xffff);

    Tcl_SetObjResult(interp, Tcl_NewStringObj(buf, -1));
    return TCL_OK;

}

/*
 *------------------------------------------------------------------------
 * tcl_al_hex_usage --
 *     Add usage information on "hex" command
 *------------------------------------------------------------------------
 */
int tcl_al_hex_usage(
	IN	char *buf,				//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{
    u32_t offset = 0;
    *length = 0;

    *length += sprintf( &buf[offset+*length], "Returns value in hexadecimal form.\n\n");

    *length += sprintf( &buf[offset+*length], "Usage: hex32 <value>\n" );
    *length += sprintf( &buf[offset+*length], "       hex16 <value>\n" );

    return TCL_OK;
}



/*
 *------------------------------------------------------------------------
 * tcl_al_hex_cleanup
 *
 * Description:
 *      "HEX" command cleanup handler.
 *
 * Returns:
 *      A standard TCL result.
 *------------------------------------------------------------------------
 */
int tcl_al_hex_cleanup(
    IN   Tcl_Interp *interp
    )
{
    Tcl_DeleteCommand(interp, HEX_CMD"16");
	Tcl_DeleteCommand(interp, HEX_CMD"32");
	return TCL_OK;
}



