#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "man_types.h"
#include "tcl_al_common.h"
#include "util.h"
#include "disputil.h"
#include "dbg_man.h"
#include "dbg_hmem_def.h"
#include "mem_track.h"

/* Used for fill command */
#define PTTN_KEYWORD_INC "increment"


/* global vars */
static char *mem_str[] = {"None", "Virt", "Phys", "Mapped"};

/* Needed for "show" use case */
static u32_t old_addr_low = 0;
static u32_t old_addr_high = 0;
static u32_t old_count = 4;


 /*------------------------------------------------------------------------
 * Description:
 *     "hmem" command usage info.
 *
 * Returns:
 *     void
 *------------------------------------------------------------------------
 */
static void usage_hmem()
{
    printf("Usage: hmem read [-b|-s] <vaddr>\n" );
    printf("       # Read byte (-b), word (-s) or dword (default) by dereferencing host virtual memory\n" );
    printf("       hmem write [-b|-s] <vaddr> <32bit_value>\n" );
    printf("       # Write byte [-b], word [-s] or dword by dereferencing host virtual memory\n" );
    printf("       hmem show <vaddr> <byte_count>\n" );
    printf("       hmem fill <vaddr> <byte_count> <pattern>\n" );
    printf("       hmem alloc <bytecount> [description_string]\n" );
    printf("       # Allocate host virtual (user space) memory\n" );
    printf("       hmem palloc <bytecount> [description_string]\n" );
    printf("       # Allocate host physical (kernel space) memory, returns virtual address\n" );
    printf("       hmem paddr <vaddr>\n" );
    printf("       # Obtain physical mem addr from kernel mem, given the virt addr\n" );
    printf("       hmem free <vaddr>\n" );
    printf("       # Free allocated (or unmap mapped) memory\n" );
    printf("       hmem map <low32phyaddr> [high32phyaddr] <size>\n" );
    printf("       # Map kernel memory to user space (virtual)\n" );
    printf("       # Do NOT map memory allocated by \"hmem palloc\"\n" );
    printf("       hmem inuse\n" );
    printf("       # Display currently allocated/mapped memory blocks\n" );
    printf("            <pattern>: increment, or any 32-bit integer\n" );
    printf("Note: Display according to $%s(%s) setting\n", TCL_SYS_ENV, ENV_DISP_BYTE );
    printf("      description_string is limited to %d characters, no space\n", MAX_HMEM_DESC_STR_SIZE);
    printf("Example: hmem read 0\n" );
    printf("         hmem write 0 0x100\n" );
    printf("         hmem show 0 0x60\n" );
    printf("         hmem fill 0x356000 0x60 increment\n" );

}

int get_vaddr(
    IN  Tcl_Interp *interp,     /* current interpreter */
    IN  Tcl_Obj *CONST objv[],   /* argument strings */
	IN	u32_t arg_idx,
	OUT	u32_t *low_addr,
	OUT	u32_t *high_addr
    )
{
	Tcl_Obj *obj;	
        int list_len;

        /*get_vaddr may be called from 'hmem read', which allows sending only the lower 32-bits of the address 
          (the upper 32 bits are assumed to be 0). We therefor accept either a list of length 2 or a scalar.*/
        Tcl_ListObjLength(interp, objv[arg_idx], &list_len);
        if ((list_len == 0) || (list_len > 2)) {
                Tcl_WrongNumArgs(interp, 0, NULL, "vaddr: scalar or list of length 2 is expected");
                return TCL_ERROR;
        }

        Tcl_ListObjIndex(interp, objv[arg_idx], 0, &obj);
        if (Tcl_GetIntFromObj(interp, obj, (int *)low_addr) == TCL_ERROR) {
                Tcl_WrongNumArgs(interp, arg_idx, objv, "vaddr: low 32-bit int value is expected");
                return TCL_ERROR;
        }

        if (list_len > 1)
        {
            Tcl_ListObjIndex(interp, objv[arg_idx], 1, &obj);
            if (Tcl_GetIntFromObj(interp, obj, (int *)high_addr) == TCL_ERROR) {
                    Tcl_WrongNumArgs(interp, arg_idx, objv, "vaddr: high 32-bit int value is expected");
                    return TCL_ERROR;
            }
        }
        else
        {
            *high_addr = 0;
        }


	return TCL_OK;
}

/**
 * Description:
 *     Implements the new Tcl "hmem read" command.
 *
 * Returns:
 *     A standard Tcl result
 */
static int hmem_read_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    u32_t  arg_idx, value;
    u32_t low_addr_val, high_addr_val;
    char  *option=NULL;
    access_method_t method=ACCESS_32BIT;
    Tcl_Obj *obj;
    dbg_status_t dbg_status;
    int tcl_status;

    arg_idx = 1;

    /* Parse -b/-s options */
    option = Tcl_GetString(objv[arg_idx]);
    if (strcmp(option, "-b")==0)
    {
        method = ACCESS_8BIT;
        arg_idx++;
    }
    else if (strcmp(option, "-s")==0)
    {
        method = ACCESS_16BIT;
        arg_idx++;
    }

    if ((tcl_status = get_vaddr(interp, objv, arg_idx, &low_addr_val, &high_addr_val)) != TCL_OK)
	return tcl_status;

    dbg_status = dbg_hmem_read(low_addr_val, high_addr_val,method, &value);
    if ( dbg_status != DBG_STATUS_SUCCESS ) {
        printf("Read operation failed \n");
        return TCL_ERROR;
    }
    
    /* Retrieve the 32-bit value from memory */
    obj = Tcl_GetObjResult(interp);
    common_nvm_set_hex_int_object(&obj, value);    

    return TCL_OK;
}

/**
 * Description:
 *     Implements the new Tcl "hmem palloc" command.
 *
 * Returns:
 *     A standard Tcl result
 */
static int hmem_alloc_phys_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    u32_t bytecount;
    char *str;
    dbg_status_t dbg_result;
    u32_t low_addr, high_addr;
    Tcl_Obj * objPtr1, *objPtr2, *listObj;

    /* Argument check */
    if (objc < 2) 
    {
        Tcl_WrongNumArgs(interp, 1, objv, "bytecount [description_name]");
        return TCL_ERROR;
    }

    /* Check for the name string */
    str = (objc > 2) ? Tcl_GetString(objv[2]) : NULL;

    /* Get the bytes count */
    if (Tcl_GetIntFromObj(interp, objv[1], (int *)&bytecount) == TCL_ERROR) 
    {
        Tcl_WrongNumArgs(interp, 1, objv, "bytecount [description_name]");
        return TCL_ERROR;
    }

    /* Allocate the physically contigious memory buffer */
    dbg_result = dbg_hmem_alloc_mem(bytecount, str, MEM_PHYS, &low_addr, &high_addr);
    if ( dbg_result != DBG_STATUS_SUCCESS ) {
        printf("Memory allocation failed \n");
        return TCL_ERROR;
    }


    listObj = Tcl_NewListObj(0, NULL);
    objPtr1 = Tcl_NewIntObj(high_addr);
    objPtr2 = Tcl_NewIntObj(low_addr);
    common_nvm_set_hex_int_object(&objPtr1, high_addr);    
    common_nvm_set_hex_int_object(&objPtr2, low_addr);        
    Tcl_ListObjAppendElement(interp, listObj, objPtr2);
    Tcl_ListObjAppendElement(interp, listObj, objPtr1);
    Tcl_SetObjResult(interp, listObj);
    return TCL_OK;
}

/**
 * Description:
 *     Implements the new Tcl "hmem free" command.
 *
 * Returns:
 *     A standard Tcl result
 */
static int hmem_free_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    u32_t low_addr_val, high_addr_val;
    dbg_status_t dbg_status;
    int tcl_status;
    int arg_idx = 1;

    /* Argument check */
    if ( objc != 2 ) 
    {
        Tcl_WrongNumArgs(interp, 1, objv, "vaddr");
        return TCL_ERROR;
    }

    if ((tcl_status = get_vaddr(interp, objv, arg_idx, &low_addr_val, &high_addr_val)) != TCL_OK)
	return tcl_status;

    dbg_status = dbg_hmem_free(low_addr_val, high_addr_val);
    if ( dbg_status != DBG_STATUS_SUCCESS ) {
        printf("Failed to release the memory \n");
        return TCL_ERROR;
    }

    
    return TCL_OK;

}

/**
 * Description:
 *     Implements the new Tcl "hmem paddr" command.
 *
 * Returns:
 *     A standard Tcl result
 */
static int hmem_get_phys_addr_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    u32_t low_addr_val, high_addr_val=0;
    dbg_status_t dbg_stat;
    Tcl_Obj *objPtr, *objPairPtr[2];
    alloc_entry_t entry;
    int arg_idx = 1;
	int tcl_status;

    /* Argument check */
    if ( objc != 2 ) 
    {
        Tcl_WrongNumArgs(interp, 1, objv, "vaddr");
        return TCL_ERROR;
    }

    if ((tcl_status = get_vaddr(interp, objv, arg_idx, &low_addr_val, &high_addr_val)) != TCL_OK)
	return tcl_status;

    dbg_stat = dbg_hmem_find_entry(low_addr_val, high_addr_val, &entry);
    if ( ( dbg_stat != DBG_STATUS_SUCCESS ) || ( entry.mem_type != MEM_PHYS ) )
    {
        /* None found or invalid entry type, return error */
        printf("The requested entry wasn't found \n");
        return TCL_ERROR;
    }

    /* Set the result */
    objPtr = Tcl_NewListObj(0, NULL);
    objPairPtr[0] = Tcl_NewIntObj(U64_LO(entry.aligned_paddr));  // lo 32-bits
    objPairPtr[1] = Tcl_NewIntObj(U64_HI(entry.aligned_paddr));  // hi 32-bits
    common_nvm_set_hex_int_object(&objPairPtr[0], U64_LO(entry.aligned_paddr));    
    common_nvm_set_hex_int_object(&objPairPtr[1], U64_HI(entry.aligned_paddr));       
    objPtr = Tcl_NewListObj(2, objPairPtr);
    Tcl_SetObjResult(interp, objPtr);

    return TCL_OK;
}

/**
 * Description:
 *     "hmem" cleanup handler.
 *
 * Returns:
 *     A standard Tcl result
 */
int tcl_al_hmem_cleanup(
    IN   Tcl_Interp *interp
    )
{
    dbg_status_t res;
    int stat = TCL_OK;

    res = dbg_hmem_cleanup_env();
    if ( res != DBG_STATUS_SUCCESS ) {
        stat = TCL_ERROR;
    }

    stat = Tcl_DeleteCommand(interp, HMEM_CMD);
    return stat;
}



/**
 * Description:
 *     A main "hmem" function. 
 *
 * Returns:
 *     A standard Tcl result
 */
int tcl_al_hmem_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    static char *option_names[] = { "read", "free", "palloc", "paddr", "help",
                                    (char *)NULL };
    enum options {READ, FREE, PALLOC, PADDR, HELP};
    int index;

    tcl_al_log_command(HMEM_CMD, objc, objv);

    #ifndef _WINDOWS
        printf("host mem opertaions are supported only in Windows\n");
        printf("for more details contact ediag/edebug team\n");
        return TCL_ERROR;
    #endif        

    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option=(read|free|palloc|paddr arg...");
        return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[1], (const char **)option_names, "option",TCL_EXACT, 
                            &index) != TCL_OK) {
        return TCL_ERROR;
    }
    switch ((enum options)index) {
        case READ:
            return hmem_read_cmd(client_data, interp, objc-1, objv+1); //DONE       
        case FREE:
            return hmem_free_cmd(client_data, interp, objc-1, objv+1);  //DONE        
        case PALLOC:
            return hmem_alloc_phys_cmd(client_data, interp, objc-1, objv+1);  //DONE
        case PADDR:
            return hmem_get_phys_addr_cmd(client_data, interp, objc-1, objv+1); //DONE        
        case HELP:
            usage_hmem();
            return TCL_OK;
        default:
            return TCL_ERROR;
    }
    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * tcl_al_hmem_usage --
 *     Add usage information on "hmem" command 
 *------------------------------------------------------------------------
 */
int tcl_al_hmem_usage(
	IN	char *buf,				//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{	
	u32_t offset = 0;
	*length = 0;

	*length += sprintf( &buf[offset+*length], "Operations such as allocate, fill, free of host virtual memory\n\n");
    *length += sprintf( &buf[offset+*length], "Usage: hmem alloc <byte_count> [<description>]\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem fill <vaddr> <byte_count> <pattern>\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem free <vaddr>\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem inuse\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem paddr <vaddr>\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem palloc <byte_count> [<description>]\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem read <vaddr>\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem show <vaddr> <byte_count>\n" );
	*length += sprintf( &buf[offset+*length], "		  hmem write <vaddr> <value>\n" );
    *length += sprintf( &buf[offset+*length], "		  hmem memcpy <dest_low32addr> [<dest_high32addr>] <src_low32addr> [<src_high32addr>] <size>\n" );
	*length += sprintf( &buf[offset+*length], "Note: palloc - allocates a continuous memory region\n" );  
	*length += sprintf( &buf[offset+*length], "Note: paddr - returns the physical address of a buffer with virtual address\n" );  

	  
    return TCL_OK;
}
