/*********************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved.
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved.
 * Contains proprietary and confidential information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without
 * the prior written consent of QLogic Corporation.
 *
 *
 *  Name: tcl_hw_set.c
 *
 *  Description: 
 *               1) Parses the HW_SET text file
 *               2) Creates a HW_SET image to be stored in NVM
 *               3) Dumps an existing HW_SET image into a HW_SET text file.
 *               
 ***********************************************************************************/

#include <stdio.h>
#ifndef USER_LINUX
#if !defined(UEFI) && !defined(__SunOS) && !defined(FREEBSD)
#include <io.h>
#endif
#else
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "tcl_hw_set.h"
#include "man_types.h"
#include "tcl_util.h"
#include "dbg_os_utils.h"
#include "common_nvm.h"
#include "utils.h"

#ifdef UEFI
#include <sys/unistd.h>
#endif

#ifndef F_OK
#define F_OK 00
#endif
#ifndef X_OK
#define X_OK 01
#endif
#ifndef W_OK
#define W_OK 02
#endif
#ifndef R_OK
#define R_OK 04
#endif


/*
 *------------------------------------------------------------------------
 * tcl_al_parse_hw_set_line :
 * 
 * Parse a line from HW_SET text file into three fields: 
 * field_name, field_type, field_data
 * 
 * Ignore white spaces at the begining and ending of field if exit
 * 
 * ------------------------------------------------------------------------
 */ 
static int tcl_al_parse_hw_set_line(char *line, char *reg_type_p, u32_t *bank_no_p, u32_t *pf_no_p, 
                                    u32_t *addr_p, u32_t *data_p, u32_t *reset_type_p, char *op_p)
{
	   char *token_ptr;
       char  tmp_str[80];
       char  err_str[80];
             
	   // Find first occurance of comma - reg_type field
       token_ptr = strtok( line, ",");
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }
       
	   remove_white_spaces(token_ptr, strlen(token_ptr), reg_type_p);

       // Verify reg_type is GRC/PHY/PCI
       if ((strcmp(reg_type_p, "GRC") != 0) && (strcmp(reg_type_p, "PHY") != 0) && (strcmp(reg_type_p, "PCI") != 0)) {
           sprintf (err_str, "Parsing HW_SET.TXT file failed! reg_type %s is not allowed\n", reg_type_p);
           print_msg (err_str, ESC_COLOR_RED);
           return TCL_ERROR;
       }
       
	   // Find second occurance of comma - bank_no field
       token_ptr = strtok( NULL, "," );
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }
       
       // Verify bank_no is a number
	   remove_white_spaces(token_ptr, strlen(token_ptr), tmp_str);
       if (!isdigit(tmp_str[0])) {
           sprintf (err_str, "Parsing HW_SET.TXT file failed! bank_no is not a number\n");
           print_msg (err_str, ESC_COLOR_RED);
           return TCL_ERROR;
       }
       *bank_no_p = atol(tmp_str);
       

       // Find third occurance of comma - pf_no field
       token_ptr = strtok( NULL, "," );
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }
       
       // Verify PF is 0-8
       remove_white_spaces(token_ptr, strlen(token_ptr), tmp_str);
       if (!isdigit(tmp_str[0])) {
           sprintf (err_str, "Parsing HW_SET.TXT file failed! PF is not a number\n");
           print_msg (err_str, ESC_COLOR_RED);
           return TCL_ERROR;
       }
       *pf_no_p = atol(tmp_str);
       if (*pf_no_p > 8) {
           sprintf (err_str, "Parsing HW_SET.TXT file failed! PF number %d is not allowed\n", *pf_no_p);
           print_msg (err_str, ESC_COLOR_RED);
           return TCL_ERROR;
       }

       // Find fourth occurance of comma - addr field
       token_ptr = strtok( NULL, "," );
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }      

       remove_white_spaces(token_ptr, strlen(token_ptr), tmp_str);       
       sscanf (tmp_str, "0X%X", addr_p);

       // Find fifth occurance of comma - data field
       token_ptr = strtok( NULL, "," );
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }    
       
       remove_white_spaces(token_ptr, strlen(token_ptr), tmp_str);
       sscanf (tmp_str, "0X%X", data_p);

       // Find sixth occurance of comma - reset type field
       token_ptr = strtok( NULL, "," );
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }      
              
       remove_white_spaces(token_ptr, strlen(token_ptr), tmp_str);
       sscanf (tmp_str, "0X%X", reset_type_p);
       if (*reset_type_p > 0xF)  {
           sprintf (err_str, "Parsing HW_SET.TXT file failed! Reset type 0x%x is not allowed\n", *reset_type_p);
           print_msg (err_str, ESC_COLOR_RED);
           return TCL_ERROR;
       }

       // Find last field - operation field
       token_ptr = strtok( NULL, " \r\t\n" );
	   if (token_ptr == NULL) {
		   return TCL_ERROR;
	   }        

       // Verify operation is R/W/RMW
       remove_white_spaces(token_ptr, strlen(token_ptr), op_p);
        if ((strcmp(op_p, "W") != 0) && (strcmp(op_p, "R") != 0) && 
           (strcmp(op_p, "RMW_SET") != 0)  && (strcmp(op_p, "RMW_CLR") != 0)) {
           sprintf (err_str, "Parsing HW_SET.TXT file failed! operation %s is not allowed\n", op_p);
           print_msg (err_str, ESC_COLOR_RED);
           return TCL_ERROR;
       }       
            
	   return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_read_hw_set_text_file: 
 * Parse the HW_SET information text file
 * 
 * Input: 
 * Specify the name of the file that contains the HW_SET information
 *
 * return: 
 * NULL on error; otherwise, a handle used for other operations.
 *------------------------------------------------------------------------
 */
void * tcl_al_read_hw_set_text_file(char *filename, u32_t *byte_read)
{
    
    FILE *fd;
    char line[512], err_str[80], tmp_str[512];
    char reg_type[80], op[80];
    u32_t val32, idx=0, hw_set_image_size=0;
    u32_t bank_no=0, pf_no=0;
    u32_t addr, data, reset_type;
    hw_set_info_t hw_sets[MAX_HW_SETS];    
    hw_set_image_t *hw_set_image_ptr=NULL;
  
    
#if !defined (UEFI) && !defined (WIN_DIAG)
    val32 = access(filename, R_OK | W_OK | F_OK);
    if (val32 == -1)
    {
        return NULL;
    }
#endif
	
    fd = fopen(filename, "r");
    if (fd == NULL)
    {
        return NULL;
    }

	/* Initialize local variables */	
    *byte_read = 0;
    
    while (!feof(fd))
    {
        if (fgets(line, 1024, fd))
        {
			/* Skip comment lines */
			if (line[0] == '#') continue;
            if (line[0] == ';') continue;
            
			/* Skip lines which do not include '=' or ',' */
            if ((strchr(line, '=') == NULL) && (strchr(line, ',') == NULL)) continue;
           	
            /* Change to upper case */                     
            strcpy(tmp_str, line);
            strtoupr(tmp_str);
                    
            /* Parse one line  at a time */
			if (tcl_al_parse_hw_set_line(tmp_str, &reg_type[0], &bank_no, &pf_no, &addr, &data, &reset_type, &op[0]) == TCL_OK)
            {   
                /*           
                DbgMessage7(NULL, WARN, "reg_type=%s bank_no=%d pf_no=%d addr=0x%x data=0x%x reset_type=0x%x op=%s\n", 
                              reg_type, bank_no, pf_no, addr, data, reset_type, op); 
                */ 
                if (strcmp(reg_type,"GRC")==0) {
                    hw_sets[idx].reg_type = GRC_REG_TYPE;
                } else if (strcmp(reg_type,"PHY")==0) {
                    hw_sets[idx].reg_type = PHY_REG_TYPE;
                } else if (strcmp(reg_type,"PCI")==0){
                    hw_sets[idx].reg_type = PCI_REG_TYPE;
                } else {
                    hw_sets[idx].reg_type = 0;
                    sprintf (err_str, "HW_SET file incorrect reg type: in line [%s]\n", line);
                    print_msg (err_str, ESC_COLOR_RED);				
                    return NULL;
                }

                hw_sets[idx].bank_num = bank_no;
                hw_sets[idx].pf_num = pf_no;
                hw_sets[idx].reg_addr = addr;
                hw_sets[idx].reg_data = data;
                hw_sets[idx].reset_type = reset_type;

                if (strcmp(op,"W")==0) {
                    hw_sets[idx].operation = WRITE_OP;
                } else if (strcmp(op, "R")==0) {
                    hw_sets[idx].operation = READ_OP;
                } else if (strcmp(op, "RMW_SET")==0) {
                    hw_sets[idx].operation = RMW_SET_OP;
                } else if (strcmp(op, "RMW_CLR")==0) {
                    hw_sets[idx].operation = RMW_CLR_OP;
                } else {
                    hw_sets[idx].operation = 0;
                    sprintf (err_str, "HW_SET file incorrect operation: in line [%s]\n", line);
                    print_msg (err_str, ESC_COLOR_RED);				
                    return NULL;
                }
                idx++;                              
                                
            } else {
                sprintf (err_str, "HW_SET file incorrect format: in line [%s]\n", line);
				print_msg (err_str, ESC_COLOR_RED);				
				return NULL;
			}                            
        }
    } 
    fclose(fd);
      
    // Allocate memory for the parsed file structure hw_set_image_t  
    hw_set_image_size =  sizeof(hw_set_image_t)+ (sizeof(hw_set_info_t) * (idx-1));
	hw_set_image_ptr = (hw_set_image_t *) malloc(hw_set_image_size);

    // Clean the HW_SET image */
    memset(hw_set_image_ptr, 0, hw_set_image_size);
	
    if ( hw_set_image_ptr == NULL ) {
        DbgBreakMsg("HW_SET: Memory allocation failed!\n");
        goto handle_error;
    }    
    
    // Set the image format revision */
    hw_set_image_ptr->format_version = HW_SET_IMAGE_VERSION;	

    // Save the number of hw_sets parsed from HW_SET file 
	hw_set_image_ptr->no_hw_sets = idx++;

    // Copy all the hw_sets[] info into the hw_set_image
	for (idx=0; idx < hw_set_image_ptr->no_hw_sets; idx++ ) {       
		memcpy((u8_t *) &hw_set_image_ptr->hw_sets[idx], (u8_t *) &hw_sets[idx], sizeof (hw_set_info_t));           
	}

	// Calculate the number of bytes in image
	*byte_read = sizeof(hw_set_image_t)+ (sizeof(hw_set_info_t) * (idx-1));
	
	
    return (void *)hw_set_image_ptr;

handle_error:
    if (hw_set_image_ptr) free((char *)hw_set_image_ptr);
    return NULL;

    
}



/*
 *------------------------------------------------------------------------
 * tcl_al_write_hw_set_test_file: 
 * Read the nvm image HW_SET, parse it and output to file
 * 
 * Input: specify the name of the output file 
 *------------------------------------------------------------------------
 */

int tcl_al_write_hw_set_text_file(char *filename, u8_t *buf)
{
	FILE *fd=NULL;
	hw_set_image_t *hw_set_image_ptr = NULL;
	u32_t idx=0, file_size=0;
    char err_str[80], reg_type[80], op[80];    
    hw_set_info_t hw_set_cmd;
        
    fd = fopen(filename, "a");
    if (fd)
	{
		/* If file already exists, exit do not allow to overwrite the file */
		fseek (fd, 0L, SEEK_END);
		file_size = ftell (fd);
		fseek (fd, 0L, SEEK_SET);      
		if (file_size > 0) {
			print_msg ("Parsing HW_SET image failed ! output file allready exists\n", ESC_COLOR_RED);
			return TCL_ERROR;
		}

		hw_set_image_ptr = (hw_set_image_t *) buf;             
      
		fprintf(fd, "###########################################################################\n");
        fprintf(fd, "#                                                                          \n");
        fprintf(fd, "#  Example HW_SET text file                                                \n");
        fprintf(fd, "#                                                                          \n");
        fprintf(fd, "#                                                                          \n");
        fprintf(fd, "#  Notes:                                                                  \n");
        fprintf(fd, "#                                                                          \n");
        fprintf(fd, "#  1. Each Comment line starts with either # or ;                          \n");  
        fprintf(fd, "#                                                                          \n");  
        fprintf(fd, "#  2. Each line includes:                                                  \n");
        fprintf(fd, "#     Register type, Bank no., PF no., Address, Data, Reset type, Operation\n"); 
        fprintf(fd, "#                                                                          \n");
        fprintf(fd, "#  3. Register types allowed: GRC, PHY, PCI                                \n");
        fprintf(fd, "#                                                                          \n"); 
        fprintf(fd, "#  4. Reset types: 0x1-POR, 0x2-HARD, 0x4-CORE, 0x8-MCP RESET              \n");
        fprintf(fd, "#     Image supports combined types, e.g. 0xC for CORE and MCP RESET       \n");
        fprintf(fd, "#                                                                          \n"); 
        fprintf(fd, "#  5. Operations allowed: R-read, W-write, RMW_SET-read, modify write set, \n");
        fprintf(fd, "#                         RMW_CLR-read, modify write clear                 \n");  
        fprintf(fd, "###########################################################################\n\n");
        
                                                
        for (idx=0; idx < htonl(hw_set_image_ptr->no_hw_sets); idx++) {
           memcpy((u8_t *) &hw_set_cmd, (u8_t *) &hw_set_image_ptr->hw_sets[idx], sizeof (hw_set_info_t));    
           
           switch (htonl(hw_set_cmd.reg_type)) {
           case GRC_REG_TYPE:
               strcpy(reg_type, "GRC");
               break;
           case PHY_REG_TYPE:
               strcpy(reg_type, "PHY");
               break;
           case PCI_REG_TYPE:
               strcpy(reg_type, "PCI");
               break;
           default:
               sprintf (err_str, "Reading HW_SET image failed! reg_type 0x%x is not allowed\n", hw_set_cmd.reg_type);
               print_msg (err_str, ESC_COLOR_RED);
               return TCL_ERROR;
           }
           
                      
           switch (htonl(hw_set_cmd.operation)) {
           case READ_OP:
               strcpy(op, "R");
               break;
           case WRITE_OP:
               strcpy(op, "W");
               break;
           case RMW_SET_OP:
               strcpy(op, "RMW_SET");
               break;
           case RMW_CLR_OP:
               strcpy(op, "RMW_CLR");
               break;           
           default:
               sprintf (err_str, "Reading HW_SET image failed! operation 0x%x is not allowed\n", hw_set_cmd.operation);
               print_msg (err_str, ESC_COLOR_RED);
               return TCL_ERROR;
           }            

           fprintf(fd, "%s, %d, %d, 0x%x, 0x%x, 0x%x, %s\n", reg_type, htonl(hw_set_cmd.bank_num), htonl(hw_set_cmd.pf_num), 
                                                             htonl(hw_set_cmd.reg_addr), htonl(hw_set_cmd.reg_data),
                                                             htonl(hw_set_cmd.reset_type), op);          													
                                   		
		}
		
		printf("HW_SET image was parsed succesfully into %s\n", filename);         
        fclose(fd);
		return TCL_OK;
	} else {
		return TCL_ERROR;
	}

}

