 /****************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved
 * Proprietary and Confidential Information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without
 * the prior written consent of QLogic Corporation.
 *
 * Name:        tcl_mcp.c
 *
 * Description: Runs the "mcp" command for the MCP in the everest chip:
 *
 *              The led command is used to control/monitor mcp block.
 *
 ****************************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "man_types.h"
#include "./../../../common/include/parser.h"
#include "tcl_firmware.h"
#include "bcmtype.h"
#include "edebug_types.h"
#include "tcl.h"
#include "tcl_al.h"
#include "tcl_al_common.h"
#include "57712_reg.h"
#include "license.h"
#include "hw_access.h"
#include "dbg_mcp.h"
#include "mcp_shmem.h"
#include "dbg_mcp_disasm.h"
#include "ncsi_basic_types.h"
#include "util.h"
#include "shmem.h"
#include "dbg_shmem.h"
#include "mcp_shmem.h"
#include "ncsi_cmds.h"
#include "nvm_map.h"
#include "tcl_mcp.h"
#include "code_swap_def.h"
#include "os_utils.h"

#define CPU_VIEW_SCRATCHPAD_BASE        0x08000000
#define CPU_STATE_VALUE_CLEAR           0xffffffff
#define CPU_LOAD_DISP_SAMPLE_RATE       10

/* Translate MCP SPAD offset from MCP view to driver view (and vice versa)*/
#define MCP_HOST_ADDR(spad_addr)       ((u32_t)(spad_addr) - CPU_VIEW_SCRATCHPAD_BASE + dreg->spad_base)
#define IS_MCP_SPAD_ADDR(addr)       (((addr) & CPU_VIEW_SCRATCHPAD_BASE) == CPU_VIEW_SCRATCHPAD_BASE)
#define MCP_SPAD_ADDR(host_addr)       ((u32_t)(host_addr) + CPU_VIEW_SCRATCHPAD_BASE - dreg->spad_base)

/* debugger registers typedef */
typedef struct debug_regs
{
    u32_t mode;
    u32_t mode_value_halt;
    u32_t mode_value_sstep;
    u32_t state;
    u32_t state_value_clear;
    u32_t gpr0;
    u32_t evmask;
    u32_t pc;
    u32_t inst;
    u32_t bp;
    u32_t spad_base;
    u32_t mips_view_base;
    u32_t host_view_base;
    u32_t ct_mbox_base;
} debug_regs_t;

//
// mailboxes for communication between diag and firmware.
//
typedef struct _debug_mb_t
{
    u32_t   signature;
        #define MCPDBG_MB_SIGNATURE         0x4646414e

    u32_t   command;
        #define DBGMCP_MB_SEQ_MASK          0x0000ffff

        #define DBGMCP_MB_CMD_MASK          0xffff0000
        #define DBGMCP_MB_CMD_WRITE         0x00010000
        #define DBGMCP_MB_CMD_READ          0x00020000

    u32_t   status;
        // lower 16 bits are sequnce number copied from command
        #define DBGMCP_MB_STATUS_MASK       0xffff0000
        #define DBGMCP_MB_STATUS_OK         0x00000000
        #define DBGMCP_MB_STATUS_ALIGNMENT  0x00010000
        #define DBGMCP_MB_STATUS_FAIL       0x00020000

    u32_t   offset;

    u32_t   value;

} debug_mb_t;

const nig_dest_mac_addr_t NigDestMacRegAddrsPort0[NCSI_MAC_ADDRESS_MAX] = 
{
    {NIG_REG_LLH0_DEST_MAC_0_0, NIG_REG_LLH0_DEST_MAC_0_1},
    {NIG_REG_LLH0_DEST_MAC_1_0, NIG_REG_LLH0_DEST_MAC_1_1},
    {NIG_REG_LLH0_DEST_MAC_2_0, NIG_REG_LLH0_DEST_MAC_2_1}, 
    #if (NCSI_MAC_ADDRESS_MAX > 3) 
    {NIG_REG_LLH0_DEST_MAC_3_0, NIG_REG_LLH0_DEST_MAC_3_1},
    #endif
    #if (NCSI_MAC_ADDRESS_MAX > 4) 
    {NIG_REG_P0_LLH_DEST_MAC_4_0, NIG_REG_P0_LLH_DEST_MAC_4_1},
    {NIG_REG_P0_LLH_DEST_MAC_5_0, NIG_REG_P0_LLH_DEST_MAC_5_1},
    #endif
};
const nig_dest_mac_addr_t NigDestMacRegAddrsPort1[NCSI_MAC_ADDRESS_MAX] = 
{
    {NIG_REG_LLH1_DEST_MAC_0_0, NIG_REG_LLH1_DEST_MAC_0_1},  
    {NIG_REG_LLH1_DEST_MAC_1_0, NIG_REG_LLH1_DEST_MAC_1_1},  
    {NIG_REG_LLH1_DEST_MAC_2_0, NIG_REG_LLH1_DEST_MAC_2_1},  
    #if (NCSI_MAC_ADDRESS_MAX > 3) 
    {NIG_REG_LLH1_DEST_MAC_3_0, NIG_REG_LLH1_DEST_MAC_3_1},       
    #endif
    #if (NCSI_MAC_ADDRESS_MAX > 4) 
    {NIG_REG_P1_LLH_DEST_MAC_4_0, NIG_REG_P1_LLH_DEST_MAC_4_1}, 
    {NIG_REG_P1_LLH_DEST_MAC_5_0, NIG_REG_P1_LLH_DEST_MAC_5_1}, 
    #endif
};

const nig_reg_t nig_reg[PORT_MAX] =
{
    // Port 0
    {
        /*.DEST_MAC        =*/ NigDestMacRegAddrsPort0,
        /*.MCP_MASK        =*/ NIG_REG_LLH0_MCP_MASK,        
        /*.VLAN_ID         =*/ {NIG_REG_LLH0_VLAN_ID_0, NIG_REG_LLH0_VLAN_ID_1},
        /*.EGRESS_FIFO_EMPTY =*/ NIG_REG_EGRESS_MNG0_FIFO_EMPTY,
        /*.EGRESS_IN_EN      =*/ NIG_REG_EGRESS_UMP0_IN_EN,  
        /*.INGRESS_IN_EN     =*/ NIG_REG_INGRESS_UMP0_IN_EN, 
        /*.INGRESS_OUT_EN    =*/ NIG_REG_INGRESS_UMP0_OUT_EN,
        /*.BRB_MCP_MASK      =*/ NIG_REG_LLH0_BRB1_MCP_MASK,
#if (EVEREST_CHIP_NUM != CHIP_NUM_57710)
        /*.MCP_MASK_MF       =*/ NIG_REG_LLH0_MCP_MASK_MF,
        /*.BRB_MCP_MASK_MF   =*/ NIG_REG_LLH0_BRB1_MCP_MASK_MF,
#endif
#if (EVEREST_CHIP_NUM == CHIP_NUM_57712)
        // Following is only available for new features in E2 (e.g. Host-BMC interface)
        /*.HOST_FWD_MCP_MASK       =*/ NIG_REG_P0_TLLH_MCP_MASK,
        /*.HOST_FWD_MCP_MASK_EXT   =*/ NIG_REG_P0_TLLH_MCP_MASK_EXT,
        /*.HOST_EGRESS_FIFO_EMPTY =*/ NIG_REG_P0_TX_MNG_HOST_FIFO_EMPTY,
        /*.BRB_MCP_MASK_A         =*/ NIG_REG_P0_LLH_BRB1_MCP_MASK_A,
        /*.MCP_MASK_A             =*/ NIG_REG_P0_LLH_MCP_MASK_A,
        /*.HOST_FWD_NTWK_DRV_MASK      =*/ NIG_REG_P0_TLLH_NTWK_DRV_MASK,
        /*.HOST_FWD_NTWK_MCP_MASK      =*/ NIG_REG_P0_TLLH_NTWK_MCP_MASK, 
        /*.HOST_DONT_FWD_NTWK_DRV_MASK =*/ NIG_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK,
        /*.HOST_DONT_FWD_NTWK_MCP_MASK =*/ NIG_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK,
        /*.TX_MNG_DESTINATION         =*/ NIG_REG_P0_TX_MNG_DESTINATION,         
        /*.TX_MNG_HOST_CTRL           =*/ NIG_REG_P0_TX_MNG_HOST_CTRL,           
        /*.TX_MNG_HOST_OUTER_VLAN_MSB =*/ NIG_REG_P0_TX_MNG_HOST_OUTER_VLAN_MSB, 
        /*.TX_MNG_HOST_FIFO_TM        =*/ NIG_REG_P0_TX_MNG_HOST_FIFO_TM,        
        /*.TX_MNG_HOST_ENABLE         =*/ NIG_REG_P0_TX_MNG_HOST_ENABLE,         
        /*.TX_HOST_MNG_ENABLE         =*/ NIG_REG_P0_TX_HOST_MNG_ENABLE,    
        /*.GLOBAL_ERROR_MASK          =*/ NIG_REG_P0_TLLH_GLOBAL_ERROR_MASK,   
        /*.FIFO_EMPTY                 =*/ NIG_REG_P0_TLLH_FIFO_EMPTY,          
        /*.FIFO_FULL                  =*/ NIG_REG_P0_TLLH_FIFO_FULL,           
        /*.RSLT_BUF_FULL              =*/ NIG_REG_P0_TLLH_RSLT_BUF_FULL,       
        /*.TLLH_MNG_PKT_RCVD          =*/ NIG_REG_P0_STAT_TLLH_MNG_PKT_RCVD,   
        /*.TLLH_MNG_OCTETS_RCVD       =*/ NIG_REG_P0_STAT_TLLH_MNG_OCTETS_RCVD,
        /*.HBUF_PACKET_DISCARD        =*/ NIG_REG_P0_STAT_HBUF_PACKET_DISCARD, 
        /*.MNG_HOST_PACKET_OUT        =*/ NIG_REG_P0_STAT_MNG_HOST_PACKET_OUT, 
        /*.MNG_HOST_OCTET_OUT         =*/ NIG_REG_P0_STAT_MNG_HOST_OCTET_OUT,  
#endif
        /*.MNG_PACKET_INP             =*/ NIG_REG_STAT0_MNG_PACKET_INP,
        /*.MNG_OCTET_INP              =*/ NIG_REG_STAT0_MNG_OCTET_INP,
        /*.MNG_DISCARD                =*/ NIG_REG_STAT0_MNG_DISCARD, 
        /*.MNG_PACKET_OUT             =*/ NIG_REG_STAT0_MNG_PACKET_OUT,
        /*.MNG_OCTET_OUT              =*/ NIG_REG_STAT0_MNG_OCTET_OUT,
    },
    // Port 1
    {
        /*.DEST_MAC        =*/ NigDestMacRegAddrsPort1,
        /*.MCP_MASK        =*/ NIG_REG_LLH1_MCP_MASK ,        
        /*.VLAN_ID         =*/ {NIG_REG_LLH1_VLAN_ID_0, NIG_REG_LLH1_VLAN_ID_1},
        /*.EGRESS_FIFO_EMPTY =*/ NIG_REG_EGRESS_MNG1_FIFO_EMPTY,
        /*.EGRESS_IN_EN      =*/ NIG_REG_EGRESS_UMP1_IN_EN,  
        /*.INGRESS_IN_EN     =*/ NIG_REG_INGRESS_UMP1_IN_EN, 
        /*.INGRESS_OUT_EN    =*/ NIG_REG_INGRESS_UMP1_OUT_EN,
        /*.BRB_MCP_MASK      =*/ NIG_REG_LLH1_BRB1_MCP_MASK,
#if (EVEREST_CHIP_NUM != CHIP_NUM_57710)
        /*.MCP_MASK_MF       =*/ NIG_REG_LLH1_MCP_MASK_MF,
        /*.BRB_MCP_MASK_MF   =*/ NIG_REG_LLH1_BRB1_MCP_MASK_MF,
#endif
#if (EVEREST_CHIP_NUM == CHIP_NUM_57712)
        // Following is only available for new features in E2 (e.g. Host-BMC interface)
        /*.HOST_FWD_MCP_MASK       =*/ NIG_REG_P1_TLLH_MCP_MASK,
        /*.HOST_FWD_MCP_MASK_EXT   =*/ NIG_REG_P1_TLLH_MCP_MASK_EXT,
        /*.HOST_EGRESS_FIFO_EMPTY =*/ NIG_REG_P1_TX_MNG_HOST_FIFO_EMPTY,
        /*.BRB_MCP_MASK_A         =*/ NIG_REG_P1_LLH_BRB1_MCP_MASK_A,
        /*.MCP_MASK_A             =*/ NIG_REG_P1_LLH_MCP_MASK_A,
        /*.HOST_FWD_NTWK_DRV_MASK      =*/ NIG_REG_P1_TLLH_NTWK_DRV_MASK,
        /*.HOST_FWD_NTWK_MCP_MASK      =*/ NIG_REG_P1_TLLH_NTWK_MCP_MASK, 
        /*.HOST_DONT_FWD_NTWK_DRV_MASK =*/ NIG_REG_P1_TLLH_NTWK_DONT_FWD_DRV_MASK,
        /*.HOST_DONT_FWD_NTWK_MCP_MASK =*/ NIG_REG_P1_TLLH_NTWK_DONT_FWD_MCP_MASK,
        /*.TX_MNG_DESTINATION         =*/ NIG_REG_P1_TX_MNG_DESTINATION,         
        /*.TX_MNG_HOST_CTRL           =*/ NIG_REG_P1_TX_MNG_HOST_CTRL,           
        /*.TX_MNG_HOST_OUTER_VLAN_MSB =*/ NIG_REG_P1_TX_MNG_HOST_OUTER_VLAN_MSB, 
        /*.TX_MNG_HOST_FIFO_TM        =*/ NIG_REG_P1_TX_MNG_HOST_FIFO_TM,        
        /*.TX_MNG_HOST_ENABLE         =*/ NIG_REG_P1_TX_MNG_HOST_ENABLE,         
        /*.TX_HOST_MNG_ENABLE         =*/ NIG_REG_P1_TX_HOST_MNG_ENABLE,         
        /*.GLOBAL_ERROR_MASK          =*/ NIG_REG_P1_TLLH_GLOBAL_ERROR_MASK,   
        /*.FIFO_EMPTY                 =*/ NIG_REG_P1_TLLH_FIFO_EMPTY,          
        /*.FIFO_FULL                  =*/ NIG_REG_P1_TLLH_FIFO_FULL,           
        /*.RSLT_BUF_FULL              =*/ NIG_REG_P1_TLLH_RSLT_BUF_FULL,       
        /*.TLLH_MNG_PKT_RCVD          =*/ NIG_REG_P1_STAT_TLLH_MNG_PKT_RCVD,   
        /*.TLLH_MNG_OCTETS_RCVD       =*/ NIG_REG_P1_STAT_TLLH_MNG_OCTETS_RCVD,
        /*.HBUF_PACKET_DISCARD        =*/ NIG_REG_P1_STAT_HBUF_PACKET_DISCARD, 
        /*.MNG_HOST_PACKET_OUT        =*/ NIG_REG_P1_STAT_MNG_HOST_PACKET_OUT, 
        /*.MNG_HOST_OCTET_OUT         =*/ NIG_REG_P1_STAT_MNG_HOST_OCTET_OUT,  
#endif
        /*.MNG_PACKET_INP             =*/ NIG_REG_STAT1_MNG_PACKET_INP,
        /*.MNG_OCTET_INP              =*/ NIG_REG_STAT1_MNG_OCTET_INP,
        /*.MNG_DISCARD                =*/ NIG_REG_STAT1_MNG_DISCARD, 
        /*.MNG_PACKET_OUT             =*/ NIG_REG_STAT1_MNG_PACKET_OUT,
        /*.MNG_OCTET_OUT              =*/ NIG_REG_STAT1_MNG_OCTET_OUT,
    }
};


/* sub-command function typedef */
typedef dbg_status_t (*mcp_sub_cmd_t)( ClientData client_data,
                          Tcl_Interp* interp,
                          int objc,
                          Tcl_Obj* CONST objv[],
                          debug_regs_t* dreg );

/*
 * local function prototypes
 */
static dbg_status_t
mcp_run(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_halt(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_gpr(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_ctrl(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_load(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_mem(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_step(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_disasm(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_trace(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_access(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );


static dbg_status_t
mcp_ff(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_fill(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_free_disasm(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_reset(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_unbreak(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

static dbg_status_t
mcp_decode(
    ClientData client_data,
    Tcl_Interp* interp,
    int objc,
    Tcl_Obj* CONST objv[],
    debug_regs_t* dreg
    );

/* end of sub-commands */

/*
 * supporting functions
 */
dbg_status_t
run_halt(
    debug_regs_t* dreg,
    u8_t run
    );

static dbg_status_t
display_gpr(
    debug_regs_t* dreg
    );

static dbg_status_t
modify_gpr(
    debug_regs_t* dreg,
    u32_t reg_index,
    u32_t value
    );

static dbg_status_t
display_ctrl(
    debug_regs_t* dreg,
    char* reg_name
    );

static dbg_status_t
modify_ctrl(
    debug_regs_t* dreg,
    char* reg_name,
    u32_t value
    );

static dbg_status_t
parse_firmware_header(
    char* filename,
    firmware_t* fw
    );

static dbg_status_t
free_firmware_header(
    firmware_t* fw
    );

dbg_status_t
write_spad(
    debug_regs_t* dreg,
    u32_t mips_addr,
    u32_t value
    );

dbg_status_t
read_spad(
    debug_regs_t* dreg,
    u32_t mips_addr,
    u32_t* value
    );

static char*
read_and_disasm(
    debug_regs_t* dreg,
    u32_t addr
    );

/*
 * Local variables
 */
static char* sub_cmd_names[] =
{
    "run",
    "halt",
    "gpr",
    "ctrl",
    "step",
    "load",
    "mem",
    "disasm",
    "trace",
    "access",
    "fill",
    "ff",
    "free_disasm",
    "reset",
    "decode",
    "unbreak",
    (char *)NULL
};

static mcp_sub_cmd_t sub_cmd[] =
{
    mcp_run,
    mcp_halt,
    mcp_gpr,
    mcp_ctrl,
    mcp_step,
    mcp_load,
    mcp_mem,
    mcp_disasm,
    mcp_trace,
    mcp_access,
    mcp_fill,
    mcp_ff,
    mcp_free_disasm,
    mcp_reset,
    mcp_decode,
    mcp_unbreak
};

#define SIZEOF_SUB_CMD  (sizeof(sub_cmd)/sizeof(sub_cmd[0]))


/*
 *------------------------------------------------------------------------
 * tcl_al_mcp_usage
 *------------------------------------------------------------------------
 */
int tcl_al_mcp_usage (
    IN char *buf,       /* Usage char buffer */
    IN u32_t *length    /* Usage char buffer length */
    )
{
    u32_t offset = 0;
    *length = 0;

    *length += sprintf( &buf[offset+*length], "Debugger for on-chip MIPS processor\n\n");
    *length += sprintf( &buf[offset+*length], "Usage: mcp run\n" );
    *length += sprintf( &buf[offset+*length], "       mcp halt\n" );
    *length += sprintf( &buf[offset+*length], "       mcp gpr  [reg value]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp ctrl [reg value]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp step [number of steps] [none|both|ctrl|gpr]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp load\n" );
    *length += sprintf( &buf[offset+*length], "       mcp mem [addr [value1 [value2 ... ]]]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp disasm [addr [count]]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp free_disasm [addr [count]]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp trace\n" );
    *length += sprintf( &buf[offset+*length], "       mcp access addr [value]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp fill [len [value]]\n" );
    *length += sprintf( &buf[offset+*length], "       mcp reset\n" );
    *length += sprintf( &buf[offset+*length], "       mcp unbreak\n" );
    *length += sprintf( &buf[offset+*length], "       mcp decode [mcp|nig|addr|scheduler|swims <cnt>]\n" );

    return TCL_OK;
}

/* mcp command */
/*
 *------------------------------------------------------------------------
 * mcp_cmd --
 *
 *     Implements the new Tcl "mcp" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
int tcl_al_mcp_cmd(
    IN  ClientData client_data, /* Not used. */
    IN  Tcl_Interp* interp,     /* current interpreter */
    IN  int objc,               /* number of arguments */
    IN  Tcl_Obj* CONST objv[]   /* argument strings */
    )
{
    int index;
    int rc;
    static debug_regs_t dreg;
    char *resp_TclGetVar2;
    u32_t chip_num;

    /* "mcp" command is not supported for VFs - break */    
#if 0
    if (dbg_cur_dev_is_vf()) {
        Tcl_SetObjResult(interp, Tcl_NewStringObj("\"mcp\" command is not supported for VFs", -1));
        return TCL_ERROR;
    }
#endif

    if( objc < 2 )
    {
        /* XXX: need to use the what for displaying usage */
        Tcl_WrongNumArgs( interp, 1, objv, "option=(select) arg..." );
        return TCL_ERROR;
    }

    rc = Tcl_GetIndexFromObj( interp, objv[1], (const char**)sub_cmd_names, "option",
                              0, &index );

    if( (rc != TCL_OK) || (index >= SIZEOF_SUB_CMD) ||
        (sub_cmd[index] == NULL) )
    {
        return TCL_ERROR;
    }

    dreg.mode               = MCP_REG_MCPR_CPU_MODE;
    dreg.mode_value_halt    = MCPR_CPU_MODE_SOFT_HALT;
    dreg.mode_value_sstep   = MCPR_CPU_MODE_STEP_ENA;
    dreg.state              = MCP_REG_MCPR_CPU_STATE;
    dreg.state_value_clear  = CPU_STATE_VALUE_CLEAR;
    dreg.gpr0               = MCP_REG_MCPR_CPU_REG_FILE;
    dreg.evmask             = MCP_REG_MCPR_CPU_EVENT_MASK;
    dreg.pc                 = MCP_REG_MCPR_CPU_PROGRAM_COUNTER;
    dreg.inst               = MCP_REG_MCPR_CPU_INSTRUCTION;
    dreg.bp                 = MCP_REG_MCPR_CPU_HW_BREAKPOINT;

    // Incase of NO_PCI mode, read the environment variable ENV_CHIP_NUM
    // and set dreg.spad_base according to the chip number
    if (tcl_al_get_no_pci_mode() == 0) {
        if (!is_e1x(interp)) {
               dreg.spad_base          = MCP_A_REG_MCPR_SCRATCH;
        } else {
                dreg.spad_base          = MCP_REG_MCPR_SCRATCH;
        }
    } else {
        // incase of NO_PCI mode, read the chip_num from MISC register
        // and set dreg.spad_base according to the chip number        
        if ( dbg_reg_read(MISC_REG_CHIP_NUM, &chip_num) != DBG_STATUS_SUCCESS ) {
            printf("Failed to read chip number from UART\n");
            return TCL_ERROR;
        }
        
        if ((chip_num == CHIP_NUM_5712) || (chip_num == CHIP_NUM_5712E) || 
            (chip_num == CHIP_NUM_5713) || (chip_num == CHIP_NUM_5713E) ||
            (chip_num == CHIP_NUM_57800) || (chip_num == CHIP_NUM_57800_MF) || (chip_num == CHIP_NUM_57810_MF) || 
            (chip_num == CHIP_NUM_57840_4_10) || (chip_num == CHIP_NUM_57840_2_20) ||
            (chip_num == CHIP_NUM_57810) || (chip_num == CHIP_NUM_57810_MF) || 
            (chip_num == CHIP_NUM_57840_MF) || (chip_num == CHIP_NUM_57811) ||
            (chip_num == CHIP_NUM_57811_MF) || (chip_num == CHIP_NUM_57811_VF) ||
            (chip_num == CHIP_NUM_57840_OBSOLETE) || (chip_num == CHIP_NUM_57840_MF_OBSOLETE)) {        
            dreg.spad_base          = MCP_A_REG_MCPR_SCRATCH;
        } else {
            dreg.spad_base          = MCP_REG_MCPR_SCRATCH;
        }
    }
 
    dreg.mips_view_base     = CPU_VIEW_SCRATCHPAD_BASE;
    dreg.host_view_base     = dreg.spad_base - dreg.mips_view_base;

    rc = sub_cmd[index]( client_data, interp, objc-1, objv+1, &dreg );
    return rc;
}

int tcl_al_mcp_cleanup(
    IN   Tcl_Interp *interp
    )
{
    return Tcl_DeleteCommand(interp, MCP_CMD);
}

/*
 *------------------------------------------------------------------------
 * mcp_run --
 *
 *     Implements the new Tcl "mcp run" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_run(
    IN   ClientData client_data,
    IN   Tcl_Interp* interp,
    IN   int objc,
    IN   Tcl_Obj* CONST objv[],
    IN   debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    int rc;

    rc_ptr = Tcl_GetObjResult(interp);

    rc = run_halt( dreg, 1 );
    if( rc != TCL_OK )
    {
        Tcl_AppendStringsToObj( rc_ptr, "Failed to run the mcp", NULL);
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * mcp_halt --
 *
 *     Implements the new Tcl "mcp halt" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_halt(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    int rc;

    rc_ptr = Tcl_GetObjResult(interp);

    rc = run_halt( dreg, 0 );
    if( rc != TCL_OK )
    {
        Tcl_AppendStringsToObj( rc_ptr, "Failed to halt the mcp", NULL);
        return TCL_ERROR;
    }

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * mcp_gpr --
 *
 *     Implements the new Tcl "mcp gpr" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_gpr(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    int rc;
    u32_t reg_index;
    u32_t value;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv,
                         "[(1...31) value]");
        return TCL_ERROR;
    }
    else if( objc == 3 )
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&reg_index );
        Tcl_GetIntFromObj( interp, objv[2], (int*)&value );
        rc = modify_gpr( dreg, reg_index, value );
        if( rc != TCL_OK )
            return TCL_ERROR;
    }

    rc = display_gpr( dreg );

    return (rc == TCL_OK) ? TCL_OK : TCL_ERROR;
}

/*
 *------------------------------------------------------------------------
 * mcp_ctrl --
 *
 *     Implements the new Tcl "mcp ctrl" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_ctrl(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    char* reg_name;
    int rc;
    u32_t value;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv,
                         "[(mode|state|evmask|pc|inst|bp) [value]]");
        return TCL_ERROR;
    }
    else if( objc >= 2 )
    {
        reg_name = Tcl_GetString(objv[1]);
        if( objc > 2 )
            Tcl_GetIntFromObj( interp, objv[2], (int*)&value );
    }
    else
    {
        reg_name = NULL;
    }

    if( objc > 2 )
    {
        rc = modify_ctrl( dreg, reg_name, value );
        if( rc != TCL_OK )
            return TCL_ERROR;
    }

    rc = display_ctrl( dreg, reg_name );

    return (rc == TCL_OK) ? TCL_OK : TCL_ERROR;
}

/*
 *------------------------------------------------------------------------
 * mcp_load --
 *
 *     Implements the new Tcl "mcp load" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_load(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    int rc;
    char* filename;
    firmware_t fw;
    u32_t i;
    u32_t j;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc != 2 )
    {
        Tcl_WrongNumArgs(interp, 1, objv,
                         "firmware filename must be specified");
        return TCL_ERROR;
    }

    filename = Tcl_GetString(objv[1]);
    rc = parse_firmware_header( filename, &fw );
    if( rc != TCL_OK )
    {
        /* Need to free the allocated parts of the header (if any) */
        free_firmware_header(&fw);
        Tcl_AppendStringsToObj( rc_ptr, "Firmware file ", filename, " not found.", NULL);
        return TCL_ERROR;
    }

    run_halt( dreg, 0 );

    /* Text first */
    printf( "   Writing Text section to address 0x%x offset ",
            fw.text_addr );
    for( i=j=0; i < fw.text_len; i+=4, ++j )
    {
        if (i % CPU_LOAD_DISP_SAMPLE_RATE == 0)
        {
            printf("0x%04x\b\b\b\b\b\b",i);
            fflush(stdout);
        }
        write_spad( dreg, fw.text_addr+i, fw.text[j] );
    }
    printf("0x%04x\n",i);

    /* Rodata */
    printf( "   Writing Rodata section to address 0x%x offset ",
            fw.rodata_addr );
    for( i=j=0; i < fw.rodata_len; i+=4, ++j )
    {
        if (i % CPU_LOAD_DISP_SAMPLE_RATE == 0)
        {
            printf("0x%04x\b\b\b\b\b\b",i);
            fflush(stdout);
        }
        write_spad( dreg, fw.rodata_addr+i, fw.rodata[j] );
    }
    printf("0x%04x\n",i);

    /* Data */
    printf( "   Writing Data section to address 0x%x offset ",
            fw.data_addr );
    for( i=j=0; i < fw.data_len; i+=4, ++j )
    {
        if (i % CPU_LOAD_DISP_SAMPLE_RATE == 0)
        {
            printf("0x%04x\b\b\b\b\b\b",i);
            fflush(stdout);
        }
        write_spad( dreg, fw.data_addr+i, fw.data[j] );
    }
    printf("0x%04x\n",i);

    /* Sbss */
    if( fw.sbss != NULL )
    {
        printf( "   Writing SBSS section to address 0x%x offset ",
                fw.sbss_addr );
        for( i=j=0; i < fw.sbss_len; i+=4, ++j )
        {
            if (i % CPU_LOAD_DISP_SAMPLE_RATE == 0)
            {
                printf("0x%04x\b\b\b\b\b\b",i);
                fflush(stdout);
            }
            write_spad( dreg, fw.sbss_addr+i, fw.sbss[j] );
        }
        printf("0x%04x\n",i);
    }

    /* Bss */
    if( fw.bss != NULL )
    {
        printf( "   Writing BSS section to address 0x%x offset ",
                fw.bss_addr );
        for( i=j=0; i < fw.bss_len; i+=4, ++j )
        {
            if (i % CPU_LOAD_DISP_SAMPLE_RATE == 0)
            {
                printf("0x%04x\b\b\b\b\b\b",i);
                fflush(stdout);
            }
            write_spad( dreg, fw.bss_addr+i, fw.bss[j] );
        }
        printf("0x%04x\n",i);
    }

    /* Sdata */
    if( fw.sdata != NULL )
    {
        printf( "   Writing Sdata section to address 0x%x offset ",
                fw.sdata_addr );
        for( i=j=0; i < fw.sdata_len; i+=4, ++j )
        {
            if (i % CPU_LOAD_DISP_SAMPLE_RATE == 0)
            {
                printf("0x%04x\b\b\b\b\b\b",i);
                fflush(stdout);
            }
            write_spad( dreg, fw.sdata_addr+i, fw.sdata[j] );
        }
        printf("0x%04x\n",i);
    }

    free_firmware_header(&fw);

    /* clear pre-fetch instruction */
    dbg_reg_dwrite( dreg->inst, 0 );
    dbg_reg_dwrite( dreg->pc, fw.start_addr );

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * mcp_step --
 *
 *     Implements the new Tcl "mcp step" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_step(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    u32_t value;
    u32_t pc;
    u32_t steps;
    u32_t step_idx;
    static enum
    {
        NONE,
        CTRL,
        GPR,
        BOTH
    } rdisp_mode = BOTH;
    char* p;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv, "[number of steps] [(none|ctrl|gpr|both)]");
        return TCL_ERROR;
    }
    if ( objc >= 2)
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&steps );
    }
    else
    {
        steps = 1;
    }
    if( objc == 3 )
    {
        p = Tcl_GetString(objv[2]);
        if( strcmp(p, "none") == 0 )
            rdisp_mode = NONE;
        else if( strcmp(p, "ctrl") == 0 )
            rdisp_mode = CTRL;
        else if( strcmp(p, "gpr") == 0 )
            rdisp_mode = GPR;
        else if( strcmp(p, "both") == 0 )
            rdisp_mode = BOTH;
    }

    /* Doing x steps (1 by default) */
    printf("Running %d step%s\n", steps, (steps > 1)? "s":"");
    for (step_idx = 0; step_idx < steps; step_idx++)
    {
        dbg_reg_dread( dreg->mode, &value );
        value |= dreg->mode_value_halt | dreg->mode_value_sstep;
        dbg_reg_dwrite( dreg->mode, value );
    }

    switch( rdisp_mode )
    {
        case NONE:
            break;
        case CTRL:
            display_ctrl( dreg, NULL );
            break;
        case GPR:
            display_gpr( dreg );
            break;
        case BOTH:
            display_gpr( dreg );
            display_ctrl( dreg, NULL );
            break;
    }

    dbg_reg_dread( dreg->pc, &pc );
    printf( "%s\n", read_and_disasm( dreg, pc ) );

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * ascii --
 *
 *     Return ASCII character if the input character is printable.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
u8_t
ascii(
    IN  u8_t c
    )
{
    return isgraph(c) ? c : '.';
}


/*
 *------------------------------------------------------------------------
 * mcp_mem --
 *
 *     Implements the new Tcl "mcp mem" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_mem(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    static u32_t addr = 0;
    u32_t value;
    static u32_t n32w = 16;
    int rc;
    u32_t i;
    char abuf[24];

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc > 2 )
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&addr );
        // make it more forgiving when entering address
        addr = (addr & 0xFFFFF) | CPU_VIEW_SCRATCHPAD_BASE;
        for( --objc, ++objv; objc > 1; --objc, ++objv, addr += 4 )
        {
            // XXX allow register name as the address
            Tcl_GetIntFromObj( interp, objv[1], (int*)&value );
            rc = write_spad( dreg, addr, value );
            if( rc != TCL_OK )
                return TCL_ERROR;
        }
    }
    else
    {
        if( objc == 2 )
        {
            // XXX allow register name as the address
            Tcl_GetIntFromObj( interp, objv[1], (int*)&addr );
            // make it more forgiving when entering address
            addr = (addr & 0xFFFFF) | CPU_VIEW_SCRATCHPAD_BASE;
        }

        for( i = 0, abuf[0]=0; i < n32w; ++i, addr += 4 )
        {
            if( i % 4 == 0 )
                printf( "%08x:  ", addr );

            rc = read_spad( dreg, addr, &value );
            if( rc != TCL_OK )
                return TCL_ERROR;

            printf( "%08x ", value );
            sprintf( strchr(abuf, 0), "%c%c%c%c", ascii(value>>24),
                     ascii((value>>16)&0x0ff), ascii((value>>8)&0x0ff),
                     ascii(value&0x0ff) );

            if( i % 4 == 3 )
            {
                printf( "  %s\n", abuf );
                abuf[0] = 0;
            }
        }
        if( i % 4 != 0 )
        {
            while( i % 4 != 0 )
                printf( "         " );
            printf( "  %s\n", abuf );
        }
    }

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * run_halt --
 *
 *     Put the MCP into RUN or HALT state.
 *
 * Output:
 *     TCL_OK -- if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
dbg_status_t
run_halt(
    IN   debug_regs_t* dreg,
    IN   u8_t run
    )
{
    int rc;
    u32_t value;

    rc = dbg_reg_dread( dreg->mode, &value );
    if( rc != TCL_OK )
        return rc;

    if( run )
    {
        value &= ~dreg->mode_value_halt;
        rc = dbg_reg_dwrite( dreg->state,
                            dreg->state_value_clear );
        if( rc != TCL_OK )
            return rc;
    }
    else
    {
        value |= dreg->mode_value_halt;
    }

    rc = dbg_reg_dwrite( dreg->mode, value );
    if( rc != TCL_OK )
        return rc;

    if( !run )
    {
        rc = dbg_reg_dwrite( dreg->state,
                            dreg->state_value_clear );
        if( rc != TCL_OK )
            return rc;
    }

    return rc;
}

/*
 *------------------------------------------------------------------------
 * free_firmware_header --
 *
 *     Free any structures that referenced by any field within firmare_t.
 *
 * Output:
 *     TCL_OK
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
free_firmware_header(
    INOUT  firmware_t* fw
    )
{
    free((char *)fw->text);
    if (fw->rodata) free((char *)fw->rodata);
    if (fw->data) free((char *)fw->data);
    if (fw->sbss) free((char *)fw->sbss);
    if (fw->bss) free((char *)fw->bss);
    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * parse_firmware_header --
 *
 *     Parse the firmware .h file.
 *
 * Output:
 *     TCL_OK if success
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
parse_firmware_header(
    IN   char* filename,
    OUT  firmware_t* fw
    )
{
    FILE* fp;
    int rc;
    char* p;
    u32_t linenum;
    char* var;
    u32_t  value;
    u32_t* parray;
    u32_t* pindex;
    u32_t count;
    u32_t** pp_buf;
    u32_t* p_len = NULL;
    static char line[4096];

    memset( fw, 0, sizeof(*fw) );

    fp = fopen( filename, "rt" );
    if( fp == NULL )
    {
        printf("   Cannot open firmware image \"%s\"\n", filename);
        return TCL_ERROR;
    }

    rc = TCL_OK;
    linenum = 1;
    parray = NULL;
    pindex = NULL;

    while( ( fgets( line, sizeof(line)-1, fp ) != NULL ) &&
           ( rc == TCL_OK ) )
    {
        p = find_word( line, NULL );
        switch( *p )
        {
            case 'i':
            case 'U':
            case 'u':
                if( ( strncmp( p, "int", 3 ) != 0 ) &&
                    ( strncmp( p, "U32", 3 ) != 0 ) &&
                    ( strncmp( p, "u32_t", 5 ) != 0 ) )
                {
                    printf("   Unknown keyword at line %d:\n", linenum);
                    printf("     --->\"%s\"<---\n", line);
                    rc = TCL_ERROR;
                    break;
                }

                pp_buf = NULL;

                p = next_word( p, NULL );
                parse_var_value( p, &var, &value );

                if( find_right_substring( var, "ReleaseMajor" ) )
                {
                    fw->ver_major = value;
                }
                else if( find_right_substring( var, "ReleaseMinor" ) )
                {
                    fw->ver_minor = value;
                }
                else if( find_right_substring( var, "ReleaseFix" ) )
                {
                    fw->ver_fix = value;
                }
                else if( find_right_substring( var, "StartAddr" ) )
                {
                    fw->start_addr = value;
                }
                else if( find_right_substring( var, "TextAddr" ) )
                {
                    fw->text_addr = value;
                }
                else if( find_right_substring( var, "TextLen" ) )
                {
                    fw->text_len = value;
                }
                else if( find_right_substring( var, "RodataAddr" ) )
                {
                    fw->rodata_addr = value;
                }
                else if( find_right_substring( var, "RodataLen" ) )
                {
                    fw->rodata_len = value;
                }
                else if( find_right_substring( var, "SDataAddr" ) )
                {
                    fw->sdata_addr = value;
                }
                else if( find_right_substring( var, "SDataLen" ) )
                {
                    fw->sdata_len = value;
                }
                else if( find_right_substring( var, "DataAddr" ) )
                {
                    fw->data_addr = value;
                }
                else if( find_right_substring( var, "DataLen" ) )
                {
                    fw->data_len = value;
                }
                else if( find_right_substring( var, "Text" ) )
                {
                    pp_buf = &fw->text;
                    p_len = &fw->text_len;
                    pindex = &fw->text_index;
                }
                else if( find_right_substring( var, "Rodata" ) )
                {
                    pp_buf = &fw->rodata;
                    p_len = &fw->rodata_len;
                    pindex = &fw->rodata_index;
                }
                else if( find_right_substring( var, "Data" ) )
                {
                    pp_buf = &fw->data;
                    p_len = &fw->data_len;
                    pindex = &fw->data_index;
                }
                else if( find_right_substring( var, "Sbss" ) )
                {
                    pp_buf = &fw->sbss;
                    p_len = &fw->sbss_len;
                    pindex = &fw->sbss_index;
                }
                else if( find_right_substring( var, "Bss" ) )
                {
                    pp_buf = &fw->bss;
                    p_len = &fw->bss_len;
                    pindex = &fw->bss_index;
                }
                else if( find_right_substring( var, "Sdata" ) )
                {
                    pp_buf = &fw->sdata;
                    p_len = &fw->sdata_len;
                    pindex = &fw->sdata_index;
                }

                if( pp_buf )
                {
                    if( *p_len == 0 )
                    {
//                        printf("   size is still unknown at line %d:\n", linenum);
//                        printf("     --->\"%s\"<---\n", line);
//                        rc = TCL_ERROR;
//                        break;
                    }
                    *pp_buf = (U32*)malloc( *p_len + 10 );
                    if( *pp_buf == NULL )
                    {
                        printf("   Unable to allocate memory for Text at line %d:\n", linenum);
                        printf("     --->\"%s\"<---\n", line);
                        rc = TCL_ERROR;
                        break;
                    }
                    parray = *pp_buf;
                    *pindex = 0;
                }
                break;

            case '0':
                if( parray != NULL && pindex != NULL )
                {
                    for( ; ; )
                    {
                        integer_from_string( p, parray+*pindex, &count );
                        if( count > 0 )
                        {
                            *pindex += 1;
                        }
                        else
                        {
                            break;
                        }
                        p = next_word( p, NULL );
                    }
                }
                break;


            case '#':
            case '\0':
            case '/':
            case ' ':
            case '}':
            case '\r':
            case '\n':
            case 't': /* typedef */
                /* ignore the line */
                break;

            default:
                printf("   Cannot parse line %d:\n", linenum);
                printf("     --->\"%s\"<---\n", line);
                rc = TCL_ERROR;
                break;
        }

        ++linenum;
    }

    fclose( fp );

    return rc;
}


/*
 *------------------------------------------------------------------------
 * write_spad --
 *
 *     Write a 32-bit word into scratch pad
 *
 * Output:
 *     TCL_OK if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
dbg_status_t
write_spad(
    IN  debug_regs_t* dreg,
    IN  u32_t mips_addr,
    IN  u32_t value
    )
{
    return dbg_reg_dwrite( mips_addr + dreg->host_view_base, value );
}

/*
 *------------------------------------------------------------------------
 * display_gpr --
 *
 *     Display the General Purpose Registers.
 *
 * Output:
 *     TCL_OK if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
display_gpr(
    IN  debug_regs_t* dreg
    )
{
    u32_t i;
    u32_t base;
    u32_t value;
    static u32_t prev_val[32];
    int rc;

    for( i = 0, base = dreg->gpr0; i < 32; base += 4 )
    {
        rc = dbg_reg_dread(base, &value);
        if( rc != TCL_OK )
            return rc;
        printf( "  %sR%-2d: %08x  ", ((prev_val[i] == value)? " ":"*"),i, value );
        prev_val[i] = value;
        if( ++i % 4 == 0 )
            printf( "\n" );
    }

    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * modify_gpr --
 *
 *     Change the General Purpose Register.
 *
 * Output:
 *     TCL_OK if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
modify_gpr(
    IN  debug_regs_t* dreg,
    IN  u32_t reg_index,
    IN  u32_t value
    )
{
    if( reg_index >= 32 )
        return TCL_ERROR;

    return dbg_reg_dwrite( dreg->gpr0 + reg_index * 4, value );
}

/*
 *------------------------------------------------------------------------
 * display_ctrl --
 *
 *     Display the MCP Control Registers.
 *
 * Output:
 *     TCL_OK if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
display_ctrl(
    IN  debug_regs_t* dreg,
    IN  char* reg_name
    )
{
    u32_t value;
    static u32_t prev_val[6];

    if( !reg_name || (strcmp(reg_name, "mode") == 0) )
    {
        dbg_reg_dread( dreg->mode, &value );
        printf( "%s%6s: %08x ", ((prev_val[0] == value)? " ":"*"), "mode", value );
        prev_val[0] = value;
    }

    if( !reg_name || (strcmp(reg_name, "state") == 0) )
    {
        dbg_reg_dread( dreg->state, &value );
        printf( "%s%6s: %08x ", ((prev_val[1] == value)? " ":"*"), "state", value );
        prev_val[1] = value;
    }

    if( !reg_name || (strcmp(reg_name, "evmask") == 0) )
    {
        dbg_reg_dread( dreg->evmask, &value );
        printf( "%s%6s: %08x ", ((prev_val[2] == value)? " ":"*"), "evmask", value );
        prev_val[2] = value;
    }

    if( !reg_name || (strcmp(reg_name, "pc") == 0) )
    {
        dbg_reg_dread( dreg->pc, &value );
        printf( "%s%6s: %08x ", ((prev_val[3] == value)? " ":"*"), "pc", value );
        prev_val[3] = value;
    }

    if( !reg_name )
        printf( "\n" );

    if( !reg_name || (strcmp(reg_name, "inst") == 0) )
    {
        dbg_reg_dread( dreg->inst, &value );
        printf( "%s%6s: %08x ", ((prev_val[4] == value)? " ":"*"), "inst", value );
        prev_val[4] = value;
    }

    if( !reg_name || (strcmp(reg_name, "bp") == 0) )
    {
        dbg_reg_dread( dreg->bp, &value );
        printf( "%s%6s: %08x ", ((prev_val[5] == value)? " ":"*"), "bp", value );
        prev_val[5] = value;
    }

    printf( "\n" );

    return TCL_OK;
}



/*
 *------------------------------------------------------------------------
 * modify_ctrl --
 *
 *     Change the MCP Control Register.
 *
 * Output:
 *     TCL_OK if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
modify_ctrl(
    IN  debug_regs_t* dreg,
    IN  char* reg_name,
    IN  u32_t value
    )
{
    if( (strcmp(reg_name, "mode") == 0) )
        dbg_reg_dwrite( dreg->mode, value );

    if( (strcmp(reg_name, "state") == 0) )
        dbg_reg_dwrite( dreg->state, value );

    if( (strcmp(reg_name, "evmask") == 0) )
        dbg_reg_dwrite( dreg->evmask, value );

    if( (strcmp(reg_name, "pc") == 0) )
        dbg_reg_dwrite( dreg->pc, value );

    if( (strcmp(reg_name, "inst") == 0) )
        dbg_reg_dwrite( dreg->inst, value );

    if( (strcmp(reg_name, "bp") == 0) )
        dbg_reg_dwrite( dreg->bp, value );

    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * read_spad --
 *
 *     Read a 32-bit word from scratch pad
 *
 * Output:
 *     TCL_OK if successful
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
dbg_status_t
read_spad(
    IN  debug_regs_t* dreg,
    IN  u32_t mips_addr,
    OUT u32_t* value
    )
{
    return dbg_reg_dread( mips_addr + dreg->host_view_base, value );
}

/*
 *------------------------------------------------------------------------
 * read_and_disasm --
 *
 *     Disassemble one instruction at specified location.
 *
 * Results:
 *     Pointer to character string.
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static char*
read_and_disasm(
    IN  debug_regs_t* dreg,
    IN  u32_t addr
    )
{
    u32_t inst;

    read_spad( dreg, addr, &inst );

    return dbg_fdisasm( inst, addr );
}

/*
 *------------------------------------------------------------------------
 * mcp_disasm --
 *
 *     Implements the internal "mcp disasm" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_disasm(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    static u32_t addr;
    u32_t count;

    rc_ptr = Tcl_GetObjResult(interp);

    count = 10;

    if( objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv, "[addr]");
        return TCL_ERROR;
    }
    else if( objc >= 2 )
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&addr );
        if( objc == 3 )
        {
            Tcl_GetIntFromObj( interp, objv[2], (int*)&count );
        }
    }

    for( ; count != 0; --count, addr += sizeof(u32_t) )
    {
        printf( "%s\n", read_and_disasm( dreg, addr ) );
    }

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * mcp_trace --
 *
 *     Prints the debug trace buffer
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_trace(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
#define LINE_PER_SCREEN                 24      //The defintion of LINE_CNT (18) is too small.
#define TRACE_SIZE                      0x0800
#define hton32(a)   ( (((a)>>24)&0x0ff) | (((a)>>8)&0x0ff00) | \
                      (((a)<<8)&0x0ff0000) | ((a)<<24) )

    u32_t shmem;
    u32_t trace_start;
    u32_t val;
    u32_t idx;
    u32_t byte_idx;
    u32_t trace_idx;
    u8_t  path_id, remainder = 0;
    u32_t cur_dword;
    u32_t fst_dword;
    static u32_t validate_sig = 0;
    Tcl_Obj* rc_ptr;
    rc_ptr = Tcl_GetObjResult(interp);

    // Reading the FW private data offset

    // The trace debug reside behind shmem_base of path0. In case of path1, we
    // need to get the shmem_base of other other path which reside in shmem2

    if (tcl_al_get_no_pci_mode()) {
        path_id = XDBG_CUR_DEV->pci_info.path_num;
    }
    else {
        dbg_reg_read(BAR_ME_REGISTER, &val);
        path_id = (u8_t)((val & ME_REG_ABS_PF_NUM) >> ME_REG_ABS_PF_NUM_SHIFT) & 1;
    }	
	
    if (path_id == 0) {
        dbg_reg_read(MISC_REG_SHARED_MEM_ADDR, &shmem);
    } else {
        u32_t shmem2_base;
        dbg_reg_read(MISC_REG_GENERIC_CR_1, &shmem2_base);
        dbg_reg_read(shmem2_base + OFFSETOF_SHMEM2_OTHER_SHMEM_BASE_ADDR, &shmem);
    }

    if ( shmem == 0 )
    {
        Tcl_AppendStringsToObj( rc_ptr,
                                "Illegal shared memory address",
                                NULL);
        return TCL_ERROR;
    }
    printf("shmem=0x%x\n", shmem);
    // convert shmem base address from host-based to MCP spad view base
    trace_start = MCP_SPAD_ADDR(shmem - TRACE_SIZE);
    printf("trace_start=0x%x\n", trace_start);

    if (validate_sig == 0)
    {
        read_spad( dreg, trace_start, &val );
        if( val != MFW_TRACE_SIGNATURE )
        {
            //printf("the signature value is 0x%x (should be 0x%x)\n", val, TRACE_SIGNATURE);
            Tcl_AppendStringsToObj( rc_ptr,
                                    "No trace buffer found in scratch-pad.",
                                    NULL);
            return TCL_ERROR;
        }
        validate_sig = 1; /* The signature is valid */
    }

    /* Retrive the current offset in the buffer */
    read_spad( dreg, trace_start + 4 /*4: signature*/, &trace_idx);
    printf("current trace index 0x%x (offset %d=0x%x)\n", trace_idx,
           trace_idx - trace_start, trace_idx - trace_start);
    if (trace_idx != (trace_idx & ~0x3))
    {
        remainder = trace_idx & 0x3;
        trace_idx = trace_idx & ~0x3;
    }

    if (trace_idx < trace_start || trace_idx > trace_start + TRACE_SIZE)
    {
        Tcl_AppendStringsToObj( rc_ptr,
                                "Trace index is out of range.",
                                NULL);
        return TCL_ERROR;
    }

    read_spad(dreg, trace_idx, &cur_dword);
    fst_dword = cur_dword;

    /* Print the first dword relevant bytes */
    for (byte_idx = remainder; byte_idx < sizeof(u32_t); byte_idx++)
    {
        u8_t byte = (fst_dword >> ((3 - byte_idx) * 8)) & 0xff;
        printf("%c", byte);
    }
    fflush(stdout);

    /* Read from the index until the end of the buffer (older messages) */
    for (idx = trace_idx + 4; idx < TRACE_SIZE + trace_start; idx+=4)
    {
        read_spad(dreg, idx, &cur_dword);
        for (byte_idx = 0; byte_idx < 4; byte_idx++)
        {
            u8_t byte = (cur_dword >> ((3 - byte_idx) * 8)) & 0xff;
            printf("%c", byte);
        }
        fflush(stdout);
    }

    /* Read from the beginning of the buffer until the current index (newer messages) */
    for (idx = trace_start + 8; idx < trace_idx; idx+=4)
    {
        read_spad(dreg, idx, &cur_dword);
        for (byte_idx = 0; byte_idx < 4; byte_idx++)
        {
            u8_t byte = (cur_dword >> ((3 - byte_idx) * 8)) & 0xff;
            printf("%c", byte);
        }
    }
    /* Print the remainder */
    for (idx = 0; idx <= remainder; idx++)
    {
        printf("%c", (fst_dword >> ((3 - idx) * 8)) & 0xff);
    }

    /* Extra line feed at the end */
    printf("\n");
    fflush(stdout);

    return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * mcp_access --
 *
 *     Implements the internal "mcp access" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
#define CPU_UCFIO_BASE                      0x80000000
static dbg_status_t
mcp_access(
    IN     ClientData client_data,
    IN     Tcl_Interp* interp,
    IN     int objc,
    IN     Tcl_Obj* CONST objv[],
    INOUT  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    u32_t addr;
    u32_t value=0;
    u32_t sig;
    u32_t seq;
    u32_t cmd;
    u32_t timeout_ms;
    u32_t shmem;
    u32_t need_help = FALSE;
    char tmp_str[16];
    // this pointer is used just to get the right offset in the scratch pad
    // it should not be access directly
    debug_mb_t * debug_mb_offset;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc == 1 || objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv, "[addr [value | help]]");
        return TCL_ERROR;
    }
    else
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&addr );

        if( objc == 3 && (Tcl_GetIntFromObj(interp, objv[2], (int *)&value) == TCL_OK))
        {
            cmd = DBGMCP_MB_CMD_WRITE;
            Tcl_GetIntFromObj( interp, objv[2], (int*)&value );
        }
        else
        {
            char * help_str;
            cmd = DBGMCP_MB_CMD_READ;

            if (objc == 3)
            {
                help_str =  Tcl_GetStringFromObj(objv[2], NULL);

                if (strcmp("help", help_str) == 0)
                {
                    need_help = TRUE;
                }
            }
        }
    }

    // Check if the FW supports the direct access (validate the signature)
    // ///////////////////////////////////////////////////////////////////

    // Reading the FW private data offset
    dbg_reg_read(MISC_REG_SHARED_MEM_ADDR, &shmem);
    if( shmem == 0 )
    {
        Tcl_AppendStringsToObj( rc_ptr,
                                "No shared memory address",
                                NULL);
        return TCL_ERROR;
    }

    dbg_reg_read((u32_t)((int_ptr_t)&(((shmem_region_t*)((int_ptr_t)shmem))->fw_info_fio_offset)), &dreg->ct_mbox_base );

    // Reading the FW private data first field (the debug mailbox offset)
    read_spad( dreg, dreg->ct_mbox_base, &dreg->ct_mbox_base );

    // Reading the debug mailbox offset
    read_spad( dreg, dreg->ct_mbox_base, &sig );

    if( sig != MCPDBG_MB_SIGNATURE )
    {
        //printf("the signature value (addr 0x%x) is 0x%x (should be 0x%x)\n",
        //       dreg->ct_mbox_base, sig, MCPDBG_MB_SIGNATURE);
        Tcl_AppendStringsToObj( rc_ptr,
                                "No debug mail box found in scratch-pad.",
                                NULL);
        return TCL_ERROR;
    }

    debug_mb_offset = (debug_mb_t*)((int_ptr_t)((debug_mb_t *)((int_ptr_t)(dreg->ct_mbox_base))));

    // Update the command sequence number
    read_spad( dreg, (u32_t)((int_ptr_t)(&debug_mb_offset->command)), &seq );
    // the new command sequnce number is one more than it used to be
    cmd |= ++seq & DBGMCP_MB_SEQ_MASK;

    // Issuing the command
    write_spad( dreg, (u32_t)((int_ptr_t)(&debug_mb_offset->value)), value);
    write_spad( dreg, (u32_t)((int_ptr_t)(&debug_mb_offset->offset)), addr);
    write_spad( dreg, (u32_t)((int_ptr_t)(&debug_mb_offset->command)), cmd);

    // Waiting for the response
    // @@@TBD we should use yield() and get_ts/rdtsc/delta_ms or some other
    //          time measuring function in order to wait in a more polite
    //          way.
    // @@@TBD we should add a counter for timeout errors (in case the FW is stack)
    timeout_ms = 1000;
    do
    {
        read_spad(dreg, (u32_t)((int_ptr_t)(&debug_mb_offset->status)), &seq);
        dbg_os_msleep(1);
    } while (((seq & DBGMCP_MB_SEQ_MASK) != (cmd & DBGMCP_MB_SEQ_MASK))
              && (timeout_ms--));
    /*printf("cmd = 0x%08x, current seq(0x%08x) = 0x%08x\n",
           cmd, (u32_t)&debug_mb_offset->status, seq);*/

    if (((seq & DBGMCP_MB_STATUS_MASK) == DBGMCP_MB_STATUS_OK)      &&
        ((seq & DBGMCP_MB_SEQ_MASK) == (cmd & DBGMCP_MB_SEQ_MASK))   )
    {
        // Read the result
        if ((cmd & DBGMCP_MB_CMD_MASK) == DBGMCP_MB_CMD_READ)
        {
            read_spad( dreg, (u32_t)((int_ptr_t)(&debug_mb_offset->value)), &value);
            sprintf(tmp_str, "0x%x", value);
            Tcl_SetStringObj(rc_ptr, tmp_str, strlen(tmp_str));
        }
        return TCL_OK;
    }
    else
    {
        Tcl_AppendStringsToObj( rc_ptr, "Failed to access FastIO space.",
                                NULL);
        return TCL_ERROR;
    }
}

/*
 *------------------------------------------------------------------------
 * mcp_ff --
 *
 *     Implements the internal "mcp ff" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_ff(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    u32_t value;
    /* Both macros uses "u32_t value" without defining it */
    #define MCP_REG_VALUE_PRINT(reg)                                         \
        dbg_reg_dread( reg, &value );                     \
        printf( #reg "(%08lx)              = %08lx\n",                         \
                (unsigned long)reg, (unsigned long)value );

    /* The macro assumes that "u32_t value" contains the reg value */
    #define MCP_BIT_VALUE_PRINT(bit)                                         \
        if ((bit) & (value))                                                 \
            printf( "\t" #bit "(%08lx) is set\n", (unsigned long)bit);


    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_MODE);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_LOCAL_RST);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_STEP_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_PAGE_0_DATA_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_PAGE_0_INST_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_MSG_BIT1);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_INTERRUPT_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_SOFT_HALT);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_BAD_DATA_HALT_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_BAD_INST_HALT_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_FIO_ABORT_HALT_ENA);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_MODE_SPAD_UNDERFLOW_HALT_ENA);

    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_STATE);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_BREAKPOINT);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_BAD_INST_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_PAGE_0_DATA_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_PAGE_0_INST_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_BAD_DATA_ADDR_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_BAD_PC_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_ALIGN_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_FIO_ABORT_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_SOFT_HALTED);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_SPAD_UNDERFLOW);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_INTERRRUPT);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_DATA_ACCESS_STALL);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_INST_FETCH_STALL);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_STATE_BLOCKED_READ);

    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_EVENT_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_BREAKPOINT_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_BAD_INST_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_PAGE_0_DATA_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_PAGE_0_INST_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_BAD_DATA_ADDR_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_BAD_PC_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_ALIGN_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_FIO_ABORT_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_SOFT_HALTED_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_SPAD_UNDERFLOW_MASK);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_EVENT_MASK_INTERRUPT_MASK);

    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_PROGRAM_COUNTER);
    printf( "\t%s\n", read_and_disasm( dreg, value ) );

    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_INSTRUCTION);
    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_DATA_ACCESS);
    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_INTERRUPT_ENABLE);
    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_INTERRUPT_VECTOR);
    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_INTERRUPT_SAVED_PC);

    MCP_REG_VALUE_PRINT(MCP_REG_MCPR_CPU_HW_BREAKPOINT);
    MCP_BIT_VALUE_PRINT(MCPR_CPU_HW_BREAKPOINT_DISABLE);


    return TCL_OK;
#undef MCP_REG_VALUE_PRINT
#undef MCP_BIT_VALUE_PRINT
}

/*
 *------------------------------------------------------------------------
 * mcp_fill --
 *
 *     Implement the Tcl command that fills the scratch pad.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_fill(
    IN   ClientData client_data,
    IN   Tcl_Interp* interp,
    IN   int objc,
    IN   Tcl_Obj* CONST objv[],
    IN   debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    u32_t len;
    u32_t value;
    u32_t addr;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv, "[len [value]]");
        return TCL_ERROR;
    }

    len = 16*1024/sizeof(u32_t);
    value = 0;
    if( objc > 1 )
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&len );
        if( objc == 3 )
        {
            Tcl_GetIntFromObj( interp, objv[2], (int*)&value );
        }
    }

    for( addr = dreg->mips_view_base; len > 0; --len, ++addr )
    {
        write_spad( dreg, addr, value );
    }

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * mcp_free_disasm --
 *
 *     Implements the internal "mcp free disasm" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_free_disasm(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    static u32_t addr;
    u32_t count;

    rc_ptr = Tcl_GetObjResult(interp);

    count = 10;

    if( objc > 3 )
    {
        Tcl_WrongNumArgs(interp, 1, objv, "[addr]");
        return TCL_ERROR;
    }
    else if( objc >= 2 )
    {
        Tcl_GetIntFromObj( interp, objv[1], (int*)&addr );
        if( objc == 3 )
        {
            Tcl_GetIntFromObj( interp, objv[2], (int*)&count );
        }
    }

    for( ; count != 0; --count, addr += sizeof(u32_t) )
    {
        u32_t inst;
        dbg_reg_dread(addr, &inst);
        printf( "%s\n", dbg_fdisasm( inst, addr ) );
    }

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * mcp_reset --
 *
 *     Implements the internal "mcp reset" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_reset(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
#if 0
    if (dbg_reset_mcp() != DBG_STATUS_SUCCESS) {
        printf("Failed to reset MCP \n");
    }
    return TCL_OK;
#endif
    printf("mcp reset is not supported in edebug as it will cause MCP to clear mailbox seq num\n");
    printf("while driver will not do that. Therefore, it will cause them to be out of sync\n");

    return TCL_ERROR;    
}

/*
 *------------------------------------------------------------------------
 * mcp_unbreak --
 *
 *     Allow MCP to "step" out of an halt instruction.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static dbg_status_t
mcp_unbreak(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    int rc;
    u32_t value;

    rc_ptr = Tcl_GetObjResult(interp);

    // First read the state to make sure we are in the halt state
    dbg_reg_dread( dreg->state, &value );
    if ((value & 0x4) == 0)
    {
        Tcl_AppendStringsToObj( rc_ptr, "Failed to unbreak the mcp: Not in halt state!!", NULL);
        return TCL_ERROR;
    }

    // read PC counter
    dbg_reg_dread( dreg->pc, &value );
    // clear the break by writing NOP to the PC in scratchpad
    rc = write_spad( dreg, value, 0);
    if( rc != TCL_OK )
        return TCL_ERROR;

    // do a single step
    dbg_reg_dread( dreg->mode, &value );
    dbg_reg_dwrite( dreg->mode, value |(MCPR_CPU_MODE_STEP_ENA| MCPR_CPU_MODE_SOFT_HALT) );
    dbg_reg_dwrite( dreg->state, 0xFFFFFFFF );
    dbg_reg_dwrite( dreg->mode, value &~(MCPR_CPU_MODE_STEP_ENA| MCPR_CPU_MODE_SOFT_HALT) );
    Tcl_AppendStringsToObj( rc_ptr, "Stepping out of MCP breakpoint", NULL);

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 *
 * DecodeMcpRegs --
 *
 *     Decode a list of MCP GRCs that is interested by MCP FW
 *
 * Results:
 *     nothing
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
typedef enum
{
    VIEW_GRC_CPU_CTRL       = 1<<0,
    VIEW_GRC_EGRESS         = 1<<1,
    VIEW_GRC_INGRESS        = 1<<2,
    VIEW_GRC_MISC           = 1<<3,
}view_mask_t;

typedef struct
{
    u32_t       addr;
    view_mask_t view_mask;
}grc_reg_info_t;
void DecodeMcpRegs(Tcl_Obj* rc_ptr, u32_t mask)
{
    dbg_status_t  status;
    u32_t index, addr, data;
    static const grc_reg_info_t grc_reg_list[] =
    {
        {MCP_REG_MCPR_PORT4MODE_EN,                 VIEW_GRC_CPU_CTRL},
        {MCP_REG_MCPR_CPU_MODE,                     VIEW_GRC_CPU_CTRL},
        {MCP_REG_MCPR_CPU_STATE,                    VIEW_GRC_CPU_CTRL},
        {MCP_REG_MCPR_CPU_EVENT_MASK,               VIEW_GRC_CPU_CTRL},
        {MCP_REG_MCPR_CPU_LAST_BRANCH_ADDR,         VIEW_GRC_CPU_CTRL},
        {MCP_REG_MCPR_HMCPQ_FTQ_CTL,                VIEW_GRC_MISC},
        {MCP_REG_MCPR_MCPQ_FTQ_CTL,                 VIEW_GRC_MISC},
        {MCP_REG_MCPR_UMP_UMP_CMD,                  VIEW_GRC_EGRESS|VIEW_GRC_INGRESS},
        {MCP_REG_MCPR_UMP_UMP_CONFIG,               VIEW_GRC_EGRESS|VIEW_GRC_INGRESS},
        {MCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC,     VIEW_GRC_EGRESS},
        {MCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC_AC,  VIEW_GRC_EGRESS},
        {MCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS, VIEW_GRC_EGRESS},
        {MCP_REG_MCPR_UMP_UMP_EGRESS_BURST_STATUS,  VIEW_GRC_EGRESS},
        {MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_PTRS,     VIEW_GRC_EGRESS},
        {MCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC,    VIEW_GRC_INGRESS},
        {MCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC_AC, VIEW_GRC_INGRESS},
        {MCP_REG_MCPR_UMP_UMP_INGRESS_BURST_STATUS, VIEW_GRC_INGRESS},
        {MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_PTRS,    VIEW_GRC_INGRESS},
        {MCP_REG_MCPR_MCP_ATTENTION_STATUS,         VIEW_GRC_MISC},
        {MCP_REG_MCPR_MCP_HEARTBEAT_STATUS,         VIEW_GRC_MISC},
        {MCP_REG_MCPR_NVM_STATUS,                   VIEW_GRC_MISC},
        {MCP_REG_MCPR_SMBUS_EVENT_STATUS,           VIEW_GRC_MISC},
    };
    for (index = 0; index < sizeof(grc_reg_list)/sizeof(grc_reg_info_t); index++)
    {
        // do view filtering if needed
        if (mask & grc_reg_list[index].view_mask)
        {
            addr = grc_reg_list[index].addr;
            status =dbg_reg_read(addr, &data);
            if (status == DBG_STATUS_SUCCESS)   
                // display help only if we look at single reg
                DecodeMcpGrcRegDetail(rc_ptr, addr, data, FALSE);
        }
    }
}// DecodeMcpRegs

/*
 *------------------------------------------------------------------------
 * ump_fifo_flat_space_hack --
 *
 *     When decoding the ingress/egress fifo flat space, the 2-byte padding
 *     doesn't exist.  This cause message decoding not to work.  This routine 
 *     will insert 2-byte padding and shift everything by accordingly.
 *
 * Results:
 *     None.                       
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static void
ump_fifo_flat_space_hack(
    u32_t *buffer,              
    u32_t size)
{
    u32_t index;
    for (index = size; index !=0 ; index--)
    {
        buffer[index] = (buffer[index] >> 16) + (buffer[index-1] << 16);
    }
    buffer[0] >>= 16;
}// ump_fifo_flat_space_hack

/*
 *------------------------------------------------------------------------
 * DecodeNigGrcRegDetail --
 *
 *     Decode NIG GRC register with details
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static void DecodeNigGrcRegDetail(Tcl_Obj* rc_ptr, u32_t addr, u32_t data)
{             
    char disp_str[1000];
    switch (addr)
    {
        case NIG_REG_LLH0_MCP_MASK: case NIG_REG_LLH1_MCP_MASK:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
                addr, data, "MCP_MASK Netwk->BMC filtering", 
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_BRCST               ? "BRCST On|":"",       
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_MLCST               ? "MLCST On|":"",         
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_UNCST               ? "UNCST On|":"",         
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_MAC0                ? "MAC0 On|":"",        
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_MAC1                ? "MAC1 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_MAC2                ? "MAC2 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_ARP                 ? "ARP On|":"",           
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_IP0                 ? "IP0 On|":"",           
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_IP1                 ? "IP1 On|":"",           
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_IP2                 ? "IP2 On|":"",           
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_NTBS_U_SRC          ? "NTBS_U_SRC On|":"",  
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_NTBS_T_SRC          ? "NTBS_T_SRC On|":"",    
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_RMCP                ? "RMCP On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_DHCP                ? "DHCP On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_NTBS_U_DST          ? "NTBS_U_DST On|":"",    
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_UDP0                ? "UDP0 On|":"",        
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_UDP1                ? "UDP1 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_UDP2                ? "UDP2 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_NTBS_T_DST          ? "NTBS_T_DST On|":"",    
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_TCP0                ? "TCP0 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_TCP1                ? "TCP1 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_TCP2                ? "TCP2 On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_VLAN_ID0            ? "VLAN_ID0 On|":"",    
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_VLAN_ID1            ? "VLAN_ID1 On|":"",      
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_VLAN_ID2            ? "VLAN_ID2 On|":"",      
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_VLAN                ? "VLAN On|":"",          
                      data &    NIG_LLH0_MCP_MASK_REG_LLH0_MCP_MASK_NO_VLAN             ? "NO_VLAN On|":"");
            break;    
        case NIG_REG_EGRESS_MNG0_FIFO_EMPTY: case NIG_REG_EGRESS_MNG1_FIFO_EMPTY:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s %s",
                addr, data, "EGRESS_MNG_FIFO_EMPTY\n\tBMC->Netwk Rx Egress Fifo: ", 
                  data                                ? "Empty":"Not Empty");
            break;
        case NIG_REG_EGRESS_UMP0_IN_EN:           case NIG_REG_EGRESS_UMP1_IN_EN:  
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s %s",
                addr, data,  "NIG_REG_EGRESS_UMP_IN_EN\n\tInput enable for TX UMP mgnt packet port", 
                data ? "ENABLE": "DISABLE");
            break;
        case NIG_REG_INGRESS_UMP0_IN_EN:         case NIG_REG_INGRESS_UMP1_IN_EN: 
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s %s",
                addr, data, "INGRESS_UMP_IN_EN\n\tInput enable for RX UMP port request", 
                data ? "ENABLE": "DISABLE");
            break;
        case NIG_REG_INGRESS_UMP0_OUT_EN:        case NIG_REG_INGRESS_UMP1_OUT_EN:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s %s",
                addr, data, "INGRESS_UMP_OUT_EN\n\t Output enable for RX UMP port response: ", 
                data ? "ENABLE": "DISABLE");
            break;
        case NIG_REG_LLH0_BRB1_MCP_MASK:         case NIG_REG_LLH1_BRB1_MCP_MASK:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
                addr, data,"BRB1_MCP_MASK \n\tNetwk->Host filtering", 
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_MAC0                  ? "MAC0 On|":"",                 
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_MAC1                  ? "MAC1 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_MAC2                  ? "MAC2 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_ARP                   ? "ARP On|":"",                    
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_IP0                   ? "IP0 On|":"",                    
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_IP1                   ? "IP1 On|":"",                    
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_IP2                   ? "IP2 On|":"",                    
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_NTBIOS_UDP_SRC        ? "NTBIOS_UDP_SRC On|":"",       
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_NTBIOS_TCP_SRC        ? "NTBIOS_TCP_SRC On|":"",         
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_RMCP                  ? "RMCP On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_DHCP                  ? "DHCP On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_NTBIOS_UDP_DST        ? "NTBIOS_UDP_DST On|":"",         
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_UDP0                  ? "UDP0 On|":"",                 
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_UDP1                  ? "UDP1 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_UDP2                  ? "UDP2 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_NTBIOS_TCP_DST        ? "NTBIOS_TCP_DST On|":"",         
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_TCP0                  ? "TCP0 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_TCP1                  ? "TCP1 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_TCP2                  ? "TCP2 On|":"",                   
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_VLAN_ID0              ? "VLAN_ID0 On|":"",             
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_VLAN_ID1              ? "VLAN_ID1 On|":"",               
                      data & NIG_LLH0_BRB1_MCP_MASK_REG_LLH0_BRB1_MCP_MASK_VLAN_ID2              ? "VLAN_ID2 On|":"");    
            break;
        case NIG_REG_LLH0_MCP_MASK_MF:           case NIG_REG_LLH1_MCP_MASK_MF:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s",
                addr, data, "MCP_MASK_MF \n\tNetwk->BMC filtering", 
                     data & NIG_LLH0_MCP_MASK_MF_REG_LLH0_MCP_MASK_OUTER_VLAN               ? "OUTER_VLAN On|":""   ,         
                     data & NIG_LLH0_MCP_MASK_MF_REG_LLH0_MCP_MASK_NO_OUTER_VLAN            ? "NO_OUTER_VLAN On|":"",         
                     data & NIG_LLH0_MCP_MASK_MF_REG_LLH0_MCP_MASK_OUTER_VLAN_ID            ? "OUTER_VLAN_ID On|":"",         
                     data & NIG_LLH0_MCP_MASK_MF_REG_LLH0_MCP_MASK_MAC3                     ? "MAC3 On|":""         ,         
                     data & NIG_LLH0_MCP_MASK_MF_REG_LLH0_MCP_MASK_IPV6_MLCST               ? "IPV6_MLCST On|":""   ,         
                     data & NIG_LLH0_MCP_MASK_MF_REG_P0_LLH_MCP_MASK_PF_OUTER_VLAN          ? "PF_OUTER_VLAN On|":"");  
            break;
        case NIG_REG_LLH0_BRB1_MCP_MASK_MF:      case NIG_REG_LLH1_BRB1_MCP_MASK_MF:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s", 
                addr, data, " \n\tNetwk->Host filtering",  
                    data & NIG_LLH0_BRB1_MCP_MASK_MF_REG_LLH0_BRB1_MCP_MASK_MAC3                 ? "MAC3 On|":"",          
                    data & NIG_LLH0_BRB1_MCP_MASK_MF_REG_LLH0_BRB1_MCP_MASK_OUTER_VLAN_ID        ? "OUTER_VLAN_ID On|":"", 
                    data & NIG_LLH0_BRB1_MCP_MASK_MF_REG_LLH0_BRB1_MCP_MASK_IPV6_MLCST           ? "IPV6_MLCST On|":"");
            break;
        case NIG_REG_P0_TLLH_MCP_MASK:           case NIG_REG_P1_TLLH_MCP_MASK:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
                addr, data, "TLLH_MCP_MASK\n\tHost->BMC filtering", 
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_BRCST                    ? "BRCST On|":"",        
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_MLCST                    ? "MLCST On|":"",        
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_UNCST                    ? "UNCST On|":"",        
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_MAC0                     ? "MAC0 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_MAC1                     ? "MAC1 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_MAC2                     ? "MAC2 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_ARP                      ? "ARP On|":"",          
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_IP0                      ? "IP0 On|":"",          
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_IP1                      ? "IP1 On|":"",          
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_IP2                      ? "IP2 On|":"",          
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_NTBS_U_SRC               ? "NTBS_U_SRC On|":"",   
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_NTBS_T_SRC               ? "NTBS_T_SRC On|":"",   
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_RMCP                     ? "RMCP On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_DHCP                     ? "DHCP On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_NTBS_U_DST               ? "NTBS_U_DST On|":"",   
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_UDP0                     ? "UDP0 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_UDP1                     ? "UDP1 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_UDP2                     ? "UDP2 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_NTBS_T_DST               ? "NTBS_T_DST On|":"",   
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_TCP0                     ? "TCP0 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_TCP1                     ? "TCP1 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_TCP2                     ? "TCP2 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_VLAN_ID0                 ? "VLAN_ID0 On|":"",     
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_VLAN_ID1                 ? "VLAN_ID1 On|":"",     
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_VLAN_ID2                 ? "VLAN_ID2 On|":"",     
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_VLAN                     ? "VLAN On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_NO_VLAN                  ? "NO_VLAN On|":"",      
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_MAC3                     ? "MAC3 On|":"",         
                      data & NIG_P0_TLLH_MCP_MASK_REG_P0_TLLH_MCP_MASK_IPV6_MLCST               ? "IPV6_MLCST On|":"");
            break;
        case NIG_REG_P0_TLLH_MCP_MASK_EXT:       case NIG_REG_P1_TLLH_MCP_MASK_EXT:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s", 
                addr, data, "MCP_MASK_EXT\n\tHost->BMC filtering",
                data & NIG_P0_TLLH_MCP_MASK_EXT_REG_P0_TLLH_MCP_MASK_EXT_MAC4                   ? "MAC4 On|":"",                    
                data & NIG_P0_TLLH_MCP_MASK_EXT_REG_P0_TLLH_MCP_MASK_EXT_MAC5                   ? "MAC5 On|":"",                    
                data & NIG_P0_TLLH_MCP_MASK_EXT_REG_P0_TLLH_MCP_MASK_EXT_ETHERTYPE0             ? "ETHERTYPE0 On|":"",              
                data & NIG_P0_TLLH_MCP_MASK_EXT_REG_P0_TLLH_MCP_MASK_EXT_ETHERTYPE1             ? "ETHERTYPE1 On|":"",              
                data & NIG_P0_TLLH_MCP_MASK_EXT_REG_P0_TLLH_MCP_MASK_EXT_ALLMLCST               ? "ALLMLCST On|":"");       
            break;
        case NIG_REG_P0_TX_MNG_HOST_FIFO_EMPTY:  case NIG_REG_P1_TX_MNG_HOST_FIFO_EMPTY:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s %s",
                addr, data, "TX_MNG_HOST_FIFO_EMPTY\n\tBMC->Host Rx Egress Fifo: ", 
                  data                                ? "Empty":"Not Empty");
            break;
        case NIG_REG_P0_LLH_BRB1_MCP_MASK_A:     case NIG_REG_P1_LLH_BRB1_MCP_MASK_A:
            sprintf(disp_str,  
                 "\n(NIG%05x=%08x) %s\n%s%s%s%s", 
                addr, data, "LLH_BRB1_MCP_MASK_A\n\tMulti Field Register",
                data & NIG_P0_LLH_BRB1_MCP_MASK_A_REG_P0_LLH_BRB1_MCP_MASK_A_MAC4               ? "MAC4 On|":"",               
                data & NIG_P0_LLH_BRB1_MCP_MASK_A_REG_P0_LLH_BRB1_MCP_MASK_A_MAC5               ? "MAC5 On|":"",               
                data & NIG_P0_LLH_BRB1_MCP_MASK_A_REG_P0_LLH_BRB1_MCP_MASK_A_ETHERTYPE0         ? "ETHERTYPE0 On|":"",         
                data & NIG_P0_LLH_BRB1_MCP_MASK_A_REG_P0_LLH_BRB1_MCP_MASK_A_ETHERTYPE1         ? "ETHERTYPE1 On|":"");  
            break;
        case NIG_REG_P0_LLH_MCP_MASK_A:          case NIG_REG_P1_LLH_MCP_MASK_A:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s", 
                addr, data, "LLH_MCP_MASK_A\n\tMulti Field Register",
                data & NIG_P0_LLH_MCP_MASK_A_REG_P0_LLH_MCP_MASK_A_MAC4                         ? "MAC4 On|":"",             
                data & NIG_P0_LLH_MCP_MASK_A_REG_P0_LLH_MCP_MASK_A_MAC5                         ? "MAC5 On|":"",             
                data & NIG_P0_LLH_MCP_MASK_A_REG_P0_LLH_MCP_MASK_A_ETHERTYPE0                   ? "ETHERTYPE0 On|":"",       
                data & NIG_P0_LLH_MCP_MASK_A_REG_P0_LLH_MCP_MASK_A_ETHERTYPE1                   ? "ETHERTYPE1 On|":"",       
                data & NIG_P0_LLH_MCP_MASK_A_REG_P0_LLH_MCP_MASK_A_ALLMLCST                     ? "ALLMLCST On|":"");
            break;
        case NIG_REG_P0_TLLH_NTWK_DRV_MASK:      case NIG_REG_P1_TLLH_NTWK_DRV_MASK:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s", 
                addr, data, "TLLH_NTWK_DRV_MASK\n\tHost Forward Netwk filtering",
                 data & NIG_P0_TLLH_NTWK_DRV_MASK_REG_P0_TLLH_NTWK_DRV_MASK_BRCST               ? "BRCST On|":"",         
                 data & NIG_P0_TLLH_NTWK_DRV_MASK_REG_P0_TLLH_NTWK_DRV_MASK_MLCST               ? "MLCST On|":"",         
                 data & NIG_P0_TLLH_NTWK_DRV_MASK_REG_P0_TLLH_NTWK_DRV_MASK_UNCST               ? "UNCST On|":"",         
                 data & NIG_P0_TLLH_NTWK_DRV_MASK_REG_P0_TLLH_NTWK_DRV_MASK_VLAN                ? "VLAN On|":"",          
                 data & NIG_P0_TLLH_NTWK_DRV_MASK_REG_P0_TLLH_NTWK_DRV_MASK_NO_VLAN             ? "NO_VLAN On|":"",       
                 data & NIG_P0_TLLH_NTWK_DRV_MASK_REG_P0_TLLH_NTWK_DRV_MASK_ALLMLCST            ? "ALLMLCST On|":"");
            break;
        case NIG_REG_P0_TLLH_NTWK_MCP_MASK:      case NIG_REG_P1_TLLH_NTWK_MCP_MASK: 
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
                addr, data, "TLLH_NTWK_MCP_MASK\n\tHost Forward Netwk filtering", 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_MAC0             ? "MAC0 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_MAC1             ? "MAC1 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_MAC2             ? "MAC2 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_ARP              ? "ARP On|":"",                  
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_IP0              ? "IP0 On|":"",                  
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_IP1              ? "IP1 On|":"",                  
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_IP2              ? "IP2 On|":"",                  
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_NTBIOS_UDP_SRC   ? "NTBIOS_UDP_SRC On|":"",       
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_NTBIOS_TCP_SRC   ? "NTBIOS_TCP_SRC On|":"",       
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_RMCP             ? "RMCP On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_DHCP             ? "DHCP On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_NTBIOS_UDP_DST   ? "NTBIOS_UDP_DST On|":"",       
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_UDP0             ? "UDP0 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_UDP1             ? "UDP1 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_UDP2             ? "UDP2 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_NTBIOS_TCP_DST   ? "NTBIOS_TCP_DST On|":"",       
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_TCP0             ? "TCP0 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_TCP1             ? "TCP1 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_TCP2             ? "TCP2 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_VLAN_ID0         ? "VLAN_ID0 On|":"",             
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_VLAN_ID1         ? "VLAN_ID1 On|":"",             
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_VLAN_ID2         ? "VLAN_ID2 On|":"",             
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_MAC3             ? "MAC3 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_IPV6_MLCST       ? "IPV6_MLCST On|":"",           
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_MAC4             ? "MAC4 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_MAC5             ? "MAC5 On|":"",                 
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_ETHERTYPE0       ? "ETHERTYPE0 On|":"",           
                data & NIG_P0_TLLH_NTWK_MCP_MASK_REG_P0_TLLH_NTWK_MCP_MASK_ETHERTYPE1       ? "ETHERTYPE1 On|":"");
            break;
        case NIG_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK: case NIG_REG_P1_TLLH_NTWK_DONT_FWD_DRV_MASK:
            sprintf(disp_str,  

                 "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s", 
                addr, data, "TLLH_NTWK_DONT_FWD_DRV_MASK\n\tHost Don't Forward Netwk filtering",
                data & NIG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_BRCST      ? "BRCST On|":"",            
                data & NIG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_MLCST      ? "MLCST On|":"",            
                data & NIG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_UNCST      ? "UNCST On|":"",            
                data & NIG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_VLAN       ? "VLAN On|":"",             
                data & NIG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_NO_VLAN    ? "NO_VLAN On|":"",          
                data & NIG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_REG_P0_TLLH_NTWK_DONT_FWD_DRV_MASK_ALLMLCST   ? "ALLMLCST On|":"");  
            break;
        case NIG_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK: case NIG_REG_P1_TLLH_NTWK_DONT_FWD_MCP_MASK:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s",
                addr, data, "TLLH_NTWK_DONT_FWD_MCP_MASK\n\tHost Don't Forward Netwk filtering", 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_MAC0            ? "MAC0 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_MAC1            ? "MAC1 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_MAC2            ? "MAC2 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_ARP             ? "ARP On|":"",                  
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_IP0             ? "IP0 On|":"",                  
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_IP1             ? "IP1 On|":"",                  
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_IP2             ? "IP2 On|":"",                  
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_NTBIOS_UDP_SRC  ? "NTBIOS_UDP_SRC On|":"",       
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_NTBIOS_TCP_SRC  ? "NTBIOS_TCP_SRC On|":"",       
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_RMCP            ? "RMCP On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_DHCP            ? "DHCP On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_NTBIOS_UDP_DST  ? "NTBIOS_UDP_DST On|":"",       
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_UDP0            ? "UDP0 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_UDP1            ? "UDP1 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_UDP2            ? "UDP2 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_NTBIOS_TCP_DST  ? "NTBIOS_TCP_DST On|":"",       
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_TCP0            ? "TCP0 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_TCP1            ? "TCP1 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_TCP2            ? "TCP2 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_VLAN_ID0        ? "VLAN_ID0 On|":"",             
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_VLAN_ID1        ? "VLAN_ID1 On|":"",             
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_VLAN_ID2        ? "VLAN_ID2 On|":"",             
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_MAC3            ? "MAC3 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_IPV6_MLCST      ? "IPV6_MLCST On|":"",           
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_MAC4            ? "MAC4 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_MAC5            ? "MAC5 On|":"",                 
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_ETHERTYPE0      ? "ETHERTYPE0 On|":"",           
                data & NIG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_REG_P0_TLLH_NTWK_DONT_FWD_MCP_MASK_ETHERTYPE1      ? "ETHERTYPE1 On|":"");
            break;
        case NIG_REG_P0_TX_MNG_DESTINATION:      case NIG_REG_P1_TX_MNG_DESTINATION:         
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s",
                addr, data, "TX_MNG_DESTINATION\n\tDestination of the packet that will be loaded through the UMP-to-NIG interface", 
                       data ? "HOST":"BMC");
            break;
        case NIG_REG_P0_TX_MNG_HOST_CTRL:        case NIG_REG_P1_TX_MNG_HOST_CTRL:           
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s %d %s %d %s %d %s %d",
                addr, data, "TX_MNG_HOST_CTRL\n\tMulti Field Register", 
                "\n\tSIDEBAND_INFO:",           (      data & NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_SIDEBAND_INFO)>>NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_SIDEBAND_INFO_SIZE,
                "\n\tPORT_ID_TAG:",             (      data & NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_PORT_ID_TAG) >> NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_PORT_ID_TAG_SIZE  ,
                "\n\tVNIC_NUM:",                (      data & NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_VNIC_NUM) >> NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_VNIC_NUM_SIZE        ,
                "\n\tOUTER_VLAN_HEADER_INSERT:",(      data & NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_OUTER_VLAN_HEADER_INSERT) >> NIG_P0_TX_MNG_HOST_CTRL_REG_P0_TX_MNG_HOST_OUTER_VLAN_HEADER_INSERT_SIZE);
            break;
        case NIG_REG_P0_TX_MNG_HOST_OUTER_VLAN_MSB:  case NIG_REG_P1_TX_MNG_HOST_OUTER_VLAN_MSB: 
            //  These bits compose of 16-bit Ethertype; 3-bit priority; and 1-bit canonical format indicator.  
            //  The 12 LSB bits of the inserted VLAN header is selected from the per-PF VLAN IDs based on the configured VNIC number",
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s%s %04x %s %x %s %x %s %03x",
                addr, data, "TX_MNG_HOST_OUTER_VLAN_MSB\n\tThe outer VLAN hdr to be inserted into MCP pkts that are fwd to host",
                "\n\tTag Protocol ID (TPID):",    data>>16,
                "\n\tPriority Code Point (PCP):",   (data & 0x7)>>13,
                "\n\tCanonical Format Indicator (CFI)::",   (data & 0x1)>>12 ,
                "\n\tVLAN ID::", (data & 0xFFF));
            break;
        case NIG_REG_P0_TX_MNG_HOST_FIFO_TM:     case NIG_REG_P1_TX_MNG_HOST_FIFO_TM:        
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s %d",
                addr, data, "TX_MNG_HOST_FIFO_TM\n\tTM for MCP TX FIFO used for fowarding pkts to the host",
                       data);
            break;
        case NIG_REG_P0_TX_MNG_HOST_ENABLE:      case NIG_REG_P1_TX_MNG_HOST_ENABLE:         
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %s",
                addr, data, "TX_MNG_HOST_FIFO_TM\n\tTM for MCP TX FIFO used for fowarding pkts to the host",
                      data ? "ENABLE": "DISABLE");
            break;
        case NIG_REG_P0_TX_HOST_MNG_ENABLE:      case NIG_REG_P1_TX_HOST_MNG_ENABLE:         
            //  Set this bit to enable the routing of mgnt pkts from PBF interface toward MCP 
            //  when the criteria for the MCP filters are met.  All pkts from PBF are forwarded 
            //  to the network when this bit is cleared."            
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %s",
                addr, data, "\nTX_HOST_MNG_ENABLE\n\tHost-to-MCP path enable: ",
                      data ? "ENABLE": "DISABLE");
            break;
        case NIG_REG_P0_TLLH_GLOBAL_ERROR_MASK:     case NIG_REG_P1_TLLH_GLOBAL_ERROR_MASK: 
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s\n%s%s",
                addr, data, "GLOBAL_ERROR_MASK\n\tMulti Field Register", 
                data & NIG_P0_TLLH_GLOBAL_ERROR_MASK_REG_P0_TLLH_MCP_ERROR_MASK     ?"MCP_ERROR_MASK On|":"",            
                data & NIG_P0_TLLH_GLOBAL_ERROR_MASK_REG_P0_TLLH_NTWK_ERROR_MASK    ?"NTWK_ERROR_MASK On|":"");  
            break;
        case NIG_REG_P0_STAT_TLLH_MNG_PKT_RCVD:  case NIG_REG_P1_STAT_TLLH_MNG_PKT_RCVD:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s:%d",
                addr, data, "STAT_TLLH_MNG_PKT_RCVD\n\tTx stats : mgnt pkts received  in TLLH",
                       data);
            break;
        case NIG_REG_P0_STAT_TLLH_MNG_OCTETS_RCVD:    case NIG_REG_P1_STAT_TLLH_MNG_OCTETS_RCVD:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s:%d",
                addr, data, "STAT_TLLH_MNG_OCTETS_RCVD\n\tTx stats : mgnt octets received  in TLLH",
                       data);
            break;
        case NIG_REG_P0_STAT_HBUF_PACKET_DISCARD:     case NIG_REG_P1_STAT_HBUF_PACKET_DISCARD: 
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s:%d",
                addr, data, "STAT_HBUF_PACKET_DISCARD\n\tTx stats : mgnt pkts discarded due to lack of buffer space in HBUF",
                       data);
            break;
        case NIG_REG_P0_TLLH_FIFO_EMPTY:    case NIG_REG_P1_TLLH_FIFO_EMPTY:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %s",
                addr, data, "TLLH_FIFO\n\t Host->BMC fifo",
                       data    ? "Empty": "Not Empty");
            break;
        case NIG_REG_P0_TLLH_FIFO_FULL:    case NIG_REG_P1_TLLH_FIFO_FULL:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %s",
                addr, data, "TLLH_FIFO\n\tHost->BMC fifo",
                       data     ? "Full": "Not Full");
            break;
        case NIG_REG_P0_TLLH_RSLT_BUF_FULL:    case NIG_REG_P1_TLLH_RSLT_BUF_FULL:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %s",
                addr, data, "TLLH_RSLT_BUF_FULL TLLH\n\tlocal parsing result buffer Host->BMC ",
                       data   ? "Full": "Not Full");
            break;
        case NIG_REG_P0_STAT_MNG_HOST_PACKET_OUT:    case NIG_REG_P1_STAT_MNG_HOST_PACKET_OUT:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, "STAT_MNG_HOST_PACKET_OUT\n\tnumber of pkts loaded into MCP TX FIFO for forwarding to the host",
                       data);
            break;
        case NIG_REG_P0_STAT_MNG_HOST_OCTET_OUT:    case NIG_REG_P1_STAT_MNG_HOST_OCTET_OUT:
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, "STAT_MNG_HOST_OCTET_OUT\n\tnumber of bytes loaded into MCP TX FIFO for forwarding to the host",
                       data);
            break;
        case NIG_REG_STAT0_MNG_PACKET_INP:   case NIG_REG_STAT1_MNG_PACKET_INP:       
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, "MNG_PACKET_INP \n\tRx stats :   NETW->MCP pkts",
                       data);
            break;
        case NIG_REG_STAT0_MNG_OCTET_INP :   case NIG_REG_STAT1_MNG_OCTET_INP :       
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, "MNG_OCTET_INP\n\tRx stats :   NETW->MCP bytes",
                       data);
            break;
        case NIG_REG_STAT0_MNG_DISCARD   : case NIG_REG_STAT1_MNG_DISCARD   :       
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, "MNG_DISCARD \n\tRx stats :   NETW->MCP discarded due to lack of buffer space",
                       data);
            break;
        case NIG_REG_STAT0_MNG_PACKET_OUT: case NIG_REG_STAT1_MNG_PACKET_OUT:       
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, "MNG_PACKET_OUT\n\tTx stats :   MCP->NETW pkts",
                       data);
            break;
        case NIG_REG_STAT0_MNG_OCTET_OUT : case NIG_REG_STAT1_MNG_OCTET_OUT :       
            sprintf(disp_str,  
                "\n(NIG%05x=%08x) %s: %d",
                addr, data, " MNG_OCTET_OUT\n\tTx stats :   MCP->NETW bytes",
                       data);
            break;
        default:
            sprintf(disp_str, "\n(NIG%05x=%08x)Unsupported NIG reg  !!!!", addr, data);
            break;
    }
    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
}// DecodeNigGrcRegDetail        
       

/*
 *------------------------------------------------------------------------
 * DecodeNigRegs --
 *
 *     Decode a list of NIG GRCs that is interested by MCP FW
 *
 * Results:
 *     nothing
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
void DecodeNigRegs(Tcl_Obj* rc_ptr)
{
    dbg_status_t  status;
    char disp_str[100];
    dbg_dev_info_t *cur_dev_p = XDBG_CUR_DEV;
    u32_t index, addr, data, port;    
    u32_t mac_address_count = is_e1x(globalInterp) ? 4:6;
    status = dbg_get_port_num(&port);
    printf("current port num = %x\n", port);
    if (port > 2) 
    {
        Tcl_AppendStringsToObj( rc_ptr, "Something is very wrong!!!", NULL);
        return;
    }
    addr = nig_reg[port].MCP_MASK;                 status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].MCP_MASK_MF;              status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].EGRESS_FIFO_EMPTY;        status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].EGRESS_IN_EN;             status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].INGRESS_IN_EN;            status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].INGRESS_OUT_EN;           status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].BRB_MCP_MASK;             status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].BRB_MCP_MASK_MF;          status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    if (!is_e1x(globalInterp))  // E2 or E3only
    {
        // Following is only available for new features in E2 (e.g. Host-BMC interface)
        addr = nig_reg[port].HOST_FWD_MCP_MASK;       status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HOST_FWD_MCP_MASK_EXT;   status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HOST_EGRESS_FIFO_EMPTY;  status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].BRB_MCP_MASK_A;          status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].MCP_MASK_A;              status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HOST_FWD_NTWK_DRV_MASK;  status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HOST_FWD_NTWK_MCP_MASK;  status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HOST_DONT_FWD_NTWK_DRV_MASK; status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HOST_DONT_FWD_NTWK_MCP_MASK; status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TX_MNG_DESTINATION;          status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TX_MNG_HOST_CTRL;           status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TX_MNG_HOST_OUTER_VLAN_MSB; status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TX_MNG_HOST_FIFO_TM;        status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TX_MNG_HOST_ENABLE;         status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TX_HOST_MNG_ENABLE;        status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].GLOBAL_ERROR_MASK;         status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].FIFO_EMPTY;                status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].FIFO_FULL;                 status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].RSLT_BUF_FULL;             status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TLLH_MNG_PKT_RCVD;         status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].TLLH_MNG_OCTETS_RCVD;      status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].HBUF_PACKET_DISCARD;       status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].MNG_HOST_PACKET_OUT;       status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
        addr = nig_reg[port].MNG_HOST_OCTET_OUT;        status =dbg_reg_read(addr, &data);
        if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    } // E2 only
    addr = nig_reg[port].MNG_PACKET_INP;       status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].MNG_OCTET_INP ;       status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].MNG_DISCARD   ;       status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].MNG_PACKET_OUT;       status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);
    addr = nig_reg[port].MNG_OCTET_OUT ;       status =dbg_reg_read(addr, &data);
    if (status == DBG_STATUS_SUCCESS)   DecodeNigGrcRegDetail(rc_ptr, addr, data);

    // MAC and VLAN ID info
    for (index = 0; index < mac_address_count; index ++)
    {
        u32_t hi_word, lo_word;
        status = dbg_reg_read(nig_reg[port].DEST_MAC[index].HiAddr, &hi_word);
        status &= dbg_reg_read(nig_reg[port].DEST_MAC[index].LoAddr, &lo_word);
        if (status == DBG_STATUS_SUCCESS) 
        {
            sprintf(disp_str, "\nMAC address [%d] = %02x:%02x:%02x:%02x:%02x:%02x",index, (hi_word>>8)&0xFF, hi_word & 0xFF, 
                (lo_word>>24)&0xFF,  (lo_word>>16)&0xFF,  (lo_word>>8)&0xFF, lo_word & 0xFF);
            Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
        }
    }
    for (index = 0; index < NCSI_VLAN_TAG_COUNT; index ++)
    {
        addr = nig_reg[port].VLAN_ID[index];                status =dbg_reg_read(addr, &data);
        sprintf(disp_str, "\nVLAN ID [%d] = %08x", index, data);
        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
    }
} // DecodeNigRegs

/*
 *------------------------------------------------------------------------
 * mcp_decode --
 *
 *     Implements the "mcp decode" command, which checks if MCP exist.
 *     Decode NIG or MCP GRCs, shmem, ncsi
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
// round byte count to the word count
#define RND(byte_cnt)  (((byte_cnt)+sizeof(u32_t)-1)/sizeof(u32_t))
static dbg_status_t
mcp_decode(
    IN  ClientData client_data,
    IN  Tcl_Interp* interp,
    IN  int objc,
    IN  Tcl_Obj* CONST objv[],
    IN  debug_regs_t* dreg
    )
{
    Tcl_Obj* rc_ptr;
    char * block_name;
    dbg_status_t  status;
    char disp_str[100];
    int rc;

    rc_ptr = Tcl_GetObjResult(interp);

    if( objc < 2 )
    {
        Tcl_WrongNumArgs(interp, 1, objv, "[shmem|nig|mcp|ncsi|scheduler|swims|<addr>]");
        return TCL_ERROR;
    }
    else
    {
        u32_t shmem_base, shmem2_base;
        u32_t addr, data, value, path_id, index, port_id, chnl_id;
        block_name =  Tcl_GetStringFromObj(objv[1], NULL);
        if  (strcmp("ncsi", block_name) == 0) 
        {   
            union {
                NcsiPersistentData_t    PersistentData;                   
                NcsiChannelData_t       ChannelData;    
            } buf;

            dbg_get_path_num(&path_id);
            dbg_reg_read(MISC_REG_SHARED_MEM_ADDR, &shmem_base);

            if (! shmem_base ) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem base address !!", NULL);
                return TCL_ERROR;
            }
            // convert shmem base address from host-based to MCP spad view base
            shmem_base = MCP_SPAD_ADDR(shmem_base);

            status = dbg_get_port_num(&port_id);
            chnl_id = COMPUTE_CHNL(path_id, port_id);

            // read NcsiPersistentData    (only exist for PATH_0 mgmtfw_state)
            if (path_id == 0)
            {
                Tcl_AppendStringsToObj( rc_ptr, "NCSI persistent data: \n", NULL);
                addr =  shmem_base + OFFSETOF(shmem_region_t, mgmtfw_state);
                for( index = 0 ; index < sizeof(NcsiPersistentData_t)/sizeof(u32_t); ++index, addr+=4 )
                {
                    rc = read_spad( dreg, addr, &value );
                    if( rc != TCL_OK )
                        return TCL_ERROR;
                    ((u32*)&buf)[index] = value;
                }
                DecodeNcsiPersistentData(rc_ptr, &buf.PersistentData);
                // Then find the NcsiChannelData in PersistentData structure
                addr =  ((u32_t *)buf.PersistentData.pChannelSpecific)[chnl_id];
                if (! IS_MCP_SPAD_ADDR(addr)) 
                {
                    sprintf(disp_str,  "Invalid MCP spad address 0x%x for MCP spad\n", addr);
                    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                    return TCL_ERROR;
                }
                sprintf(disp_str,  "Channal data Mem for chnl %d (path = %d, port = %d)\n", chnl_id, path_id, port_id);
                Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                for( index = 0 ; index < sizeof(NcsiChannelData_t)/sizeof(u32_t); ++index, addr+=4 )
                {
                    rc = read_spad( dreg, addr, &value );
                    if( rc != TCL_OK )
                        return TCL_ERROR;
                    ((u32*)&buf)[index] = value;
                }
                DecodeNcsiChannelData(rc_ptr, &buf.ChannelData);
            }
            else  //  PATH_1 mgmtfw_state only contains channel specific data for PATH_1
            {
                addr =  shmem_base + OFFSETOF(shmem_region_t, mgmtfw_state) + sizeof(NcsiChannelData_t) * port_id;
                sprintf(disp_str,  "Channal data Mem for chnl %d (path = %d, port = %d)\n", chnl_id, path_id, port_id);
                Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                for( index = 0 ; index < sizeof(NcsiChannelData_t)/sizeof(u32_t); ++index, addr+=4 )
                {
                    rc = read_spad( dreg, addr, &value );
                    if( rc != TCL_OK )
                        return TCL_ERROR;
                    ((u32*)&buf)[index] = value;
                }
                DecodeNcsiChannelData(rc_ptr, &buf.ChannelData);
            }

        }
        else if  (strcmp("shmem", block_name) == 0) 
        {   
            union {
                struct shmem_region    shmem;                   
                struct shmem2_region   shmem2;    
                u32_t                  data[1];  
            } buf;
            u32_t retain_bitmap_addr;
            
            dbg_get_path_num(&path_id);
            dbg_reg_read(MISC_REG_SHARED_MEM_ADDR, &shmem_base);
            if (path_id == 0) {
                dbg_reg_read(MISC_REG_GENERIC_CR_0, &shmem2_base);
            } else {
                dbg_reg_read(MISC_REG_GENERIC_CR_1, &shmem2_base);
            }

            sprintf(disp_str,  "Shared Mem for path %d:\n", path_id);
            Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
            if (!shmem_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem base address !!", NULL);
                return TCL_ERROR;
            }
            // convert shmem base address from host-based to MCP spad view base
            shmem_base = MCP_SPAD_ADDR(shmem_base);
            for (index = 0; index < sizeof(struct shmem_region)/sizeof(u32_t); index++)
            {
                read_spad( dreg, shmem_base+index*sizeof(u32_t), (u32_t *)(&buf)+ index );
            } 
            DecodeShmemRegion(rc_ptr, &buf.shmem);

            sprintf(disp_str,  "Shared Mem 2 for path %d:\n", path_id);
            Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);

            if (!shmem2_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem2 base address !!", NULL);
                return TCL_ERROR;
            }
            // convert shmem2 base address from host-based to MCP spad view base
            shmem2_base = MCP_SPAD_ADDR(shmem2_base);
            
            for (index = 0; index < sizeof(struct shmem2_region)/sizeof(u32_t); index++)
            {
                read_spad( dreg, shmem2_base+index*sizeof(u32_t), (u32_t *)(&buf)+ index );
            } 
            DecodeShmem2Region(rc_ptr, &buf.shmem2);
            retain_bitmap_addr = buf.shmem2.nvm_retain_bitmap_addr;

            for (index = 0; index < NVM_RETAIN_BITMAP_SIZE; index++)
            {
                dbg_reg_read(retain_bitmap_addr+index*sizeof(u32_t), buf.data + index );
            }
            DecodeRetainAddr(rc_ptr, path_id, buf.data);

        }
        else if (strcmp("temperature", block_name) == 0)
        {
            union {
                struct shmem2_region   shmem2;  
                u32_t                  data[1];  
            } buf;

            common_nvm_get_path_num(&path_id);
            dbg_reg_read(MISC_REG_SHARED_MEM_ADDR, &shmem_base);
            if (path_id == 0) {
                dbg_reg_read(MISC_REG_GENERIC_CR_0, &shmem2_base);
            } else {
                dbg_reg_read(MISC_REG_GENERIC_CR_1, &shmem2_base);
            }
             if (!shmem2_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem2 base address !!", NULL);
                return TCL_ERROR;
            }
            // convert shmem2 base address from host-based to MCP spad view base
            shmem2_base = MCP_SPAD_ADDR(shmem2_base);
            
            for (index = 0; index < sizeof(struct shmem2_region)/sizeof(u32_t); index++)
            {
                read_spad( dreg, shmem2_base+index*sizeof(u32_t), (u32_t *)(&buf)+ index );
            } 
            DecodeTemperature(rc_ptr, &buf.shmem2);
        }            
        else if  (strcmp("npar", block_name) == 0) 
        {   
            u32_t mf_cfg_base;
            struct mf_cfg buf;
            
            dbg_reg_read(MISC_REG_SHARED_MEM_ADDR, &shmem_base);

            mf_cfg_base = shmem_base +  OFFSETOF(struct shmem_region, func_mb) + E2_FUNC_MAX*sizeof(struct drv_func_mb);

            sprintf(disp_str,  "NPAR :\n");
            Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
            if (!mf_cfg_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid npar cfg base address !!", NULL);
                return TCL_ERROR;
            }
            // convert shmem base address from host-based to MCP spad view base
            mf_cfg_base = MCP_SPAD_ADDR(mf_cfg_base);
            for (index = 0; index < sizeof(struct mf_cfg)/sizeof(u32_t); index++)
            {
                read_spad( dreg, mf_cfg_base+index*sizeof(u32_t), (u32_t *)(&buf)+ index );
            } 
            DecodeMfCfg(rc_ptr, &buf);
        }
        else if (strcmp("nig", block_name) == 0) 
        {   
            DecodeNigRegs(rc_ptr);         
        }
        else if (strcmp("mcp", block_name) == 0)
        {
            u32_t mask;
            if (objc < 3 || (Tcl_GetIntFromObj(interp, objv[2], (int *)&mask) != TCL_OK))
                mask = (u32_t)(-1);  // display everything, no view filtering
            DecodeMcpRegs(rc_ptr, mask);         
        }
        else if (strcmp("egress", block_name) == 0)
        {
            #define EGRESS_MIN_PKT_LENGTH    40
            #define EGRESS_MAX_PKT_LENGTH    256
            #define EGRESS_FLAT_SPACE_SIZE  (MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_FLAT_SPACE_COUNT*sizeof(u32_t))
            u32_t pkt_buffer[EGRESS_MAX_PKT_LENGTH/sizeof(u32_t)];
            u32_t * last_pkt;
            u32_t addr, pkt_size, temp, rd_ptr,wr_ptr, temp_rd_ptr;

            if( objc >= 3 && Tcl_GetIntFromObj( interp, objv[2], (int*)&value )!= TCL_ERROR)
            {
                u32_t pkt_index, max_pkt = 1;
                // translate offset to rd_ptr
                rd_ptr = RND((u32_t)value) & 0x7ff;

                if( objc >= 4 && Tcl_GetIntFromObj( interp, objv[3], (int*)&value )!= TCL_ERROR)
                {
                    max_pkt = (u32_t)value;
                    if (max_pkt > EGRESS_FLAT_SPACE_SIZE/EGRESS_MIN_PKT_LENGTH)
                    {
                        Tcl_AppendStringsToObj( rc_ptr, "\nRequesting too many packet\n", NULL);
                        max_pkt = 1;  // requesting too many packet
                    } 
                }

                addr = MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_FLAT_SPACE;
                for (pkt_index = 0; pkt_index < max_pkt; pkt_index ++)
                {
                    sprintf(disp_str,  "\nEgress Pkt at loc 0x%x (%x):\n", (u32_t)(addr + rd_ptr * sizeof(u32_t)), rd_ptr);
                    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                    dbg_reg_read( addr + rd_ptr * sizeof(u32_t), &pkt_size);
                    if (pkt_size == 0 || pkt_size > EGRESS_MAX_PKT_LENGTH)
                    {
                        sprintf(disp_str,  "\nIncorrect Egress Pkt length = %u (rd_ptr = %08x)!!\n", pkt_size, rd_ptr);
                        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        break;
                    }
                    else
                    {
                        if (++rd_ptr * sizeof(u32_t) >= EGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                        for (index = 0;   index < RND(pkt_size) ; index++)
                        {
                            dbg_reg_read( addr + rd_ptr * sizeof(u32_t), pkt_buffer+index);
                            if (++rd_ptr * sizeof(u32_t) >= EGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                        }
                        if (index < RND(pkt_size))
                        {
                            sprintf(disp_str, "\nIncomplete Egress Pkt length: %d out of %d bytes !!\n", (u32_t)(index * sizeof(u32_t)), pkt_size); 
                            Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        }
                        else
                        {
                            ump_fifo_flat_space_hack(pkt_buffer, index);              
                            DecodeNcsiCmdPkt(rc_ptr, (NcsiRmiiCmdPkt_t *)pkt_buffer);
                        }
                    }
                }
            
            }
            else  // default setting (no offset input)
            {
                // read egress fifo ptrs
                addr = MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_PTRS;
                status = dbg_reg_read(addr, &data);
                if (status == DBG_STATUS_SUCCESS)   
                    DecodeMcpGrcRegDetail(rc_ptr, addr, data, FALSE);
                rd_ptr = ((data & MCPR_UMP_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_RD_PTR) >> 0 ) ;
                wr_ptr = ((data & MCPR_UMP_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_WR_PTR) >> 16) ;
                if ((rd_ptr & 0x800) != (wr_ptr & 0x800))
                {
                    Tcl_AppendStringsToObj( rc_ptr, "\nEgress Read/Write Pointer wrap around detected. \n", NULL);
                }

                addr = MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_FLAT_SPACE;
                // We need to filter out the wrap around bit in RD/WR pointer
                rd_ptr =  rd_ptr & 0x7FF;
                wr_ptr =  wr_ptr & 0x7FF;

                if (rd_ptr == wr_ptr)    // no unprocessed pkt
                {
                    temp_rd_ptr = rd_ptr;
                    // Show last process Egress Pkt
                    Tcl_AppendStringsToObj( rc_ptr, "\nLast Processed Egress Pkt :\n", NULL);
                    last_pkt = &pkt_buffer[EGRESS_MAX_PKT_LENGTH/sizeof(u32_t)];
                    for (index = 0; index < EGRESS_MAX_PKT_LENGTH/sizeof(u32_t); index++)
                    {
                        last_pkt --;
                        temp_rd_ptr = (temp_rd_ptr == 0) ? EGRESS_FLAT_SPACE_SIZE / sizeof(u32_t) - 1 : temp_rd_ptr - 1;
                        dbg_reg_read(addr + temp_rd_ptr * sizeof(u32_t), last_pkt);
                        if (index >= EGRESS_MIN_PKT_LENGTH/sizeof(u32_t) && RND(*last_pkt) == index)
                        {
                            last_pkt++;
                            break;  // found a full packet in flat space
                        }
                    }
                    if (index >= EGRESS_MAX_PKT_LENGTH/sizeof(u32_t))
                    {
                        Tcl_AppendStringsToObj( rc_ptr, "\nEgress Pkt length too long !!\n", NULL);
                    }
                    else
                    {
                        ump_fifo_flat_space_hack(last_pkt, index);              
                        DecodeNcsiCmdPkt(rc_ptr, (NcsiRmiiCmdPkt_t *)last_pkt);
                    }
                }
                temp = 1;  // use as counter
                while (rd_ptr != wr_ptr)
                {
                    sprintf(disp_str, "\nUnprocessed Egress Pkt #%d:\n", temp++); 
                    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                    dbg_reg_read(addr + rd_ptr * sizeof(u32_t), &pkt_size);
                    if (pkt_size > sizeof(pkt_buffer) )
                    {
                        sprintf(disp_str,  "\nIncorrect Egress Pkt length = %d!!\n", pkt_size);
                        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        break;
                    }
                    if (++rd_ptr * sizeof(u32_t) >= EGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                    for (index = 0;   index < RND(pkt_size) && rd_ptr != wr_ptr ; index++)
                    {
                        dbg_reg_read(addr + rd_ptr * sizeof(u32_t), pkt_buffer + index);
                        if (++rd_ptr * sizeof(u32_t) >= EGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                    }
                    if (index < RND(pkt_size))
                    {
                        sprintf(disp_str,  "\nIncomplete Egress Pkt length: %d out of %d bytes !!\n", (u32_t)(index * sizeof(u32_t)), pkt_size);
                        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        break;
                    }
                    else
                    {
                        ump_fifo_flat_space_hack(pkt_buffer, index);              
                        DecodeNcsiCmdPkt(rc_ptr, (NcsiRmiiCmdPkt_t *)pkt_buffer);
                    }
                }
            }
        }
        else if(strcmp("ingress", block_name) == 0)
        {
            #define INGRESS_MIN_PKT_LENGTH    40
            #define INGRESS_MAX_PKT_LENGTH    256
            #define INGRESS_FLAT_SPACE_SIZE  (MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_FLAT_SPACE_COUNT*sizeof(u32_t))
            u32_t pkt_buffer[INGRESS_MAX_PKT_LENGTH/sizeof(u32_t)];
            u32_t * last_pkt;
            u32_t addr, pkt_size, temp, rd_ptr,wr_ptr, temp_rd_ptr;
            if( objc >= 3 && Tcl_GetIntFromObj( interp, objv[2], (int*)&value )!= TCL_ERROR)
            {
                u32_t pkt_index, max_pkt = 1;
                // translate offset to rd_ptr
                rd_ptr = RND((u32_t)value) & 0x3ff;

                if( objc >= 4 && Tcl_GetIntFromObj( interp, objv[3], (int*)&value )!= TCL_ERROR)
                {
                    max_pkt = (u32_t)value;
                    if (max_pkt > INGRESS_FLAT_SPACE_SIZE/INGRESS_MIN_PKT_LENGTH)
                    {
                        Tcl_AppendStringsToObj( rc_ptr, "\nRequesting too many packet\n", NULL);
                        max_pkt = 1;  // requesting too many packet
                    } 
                }
                addr = MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_FLAT_SPACE;
                for (pkt_index = 0; pkt_index < max_pkt; pkt_index ++)
                {
                    sprintf(disp_str, "\nIngress Pkt at loc 0x%x (%x):\n", (u32_t)(addr + rd_ptr * sizeof(u32_t)), rd_ptr); 
                    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                    dbg_reg_read( addr + rd_ptr * sizeof(u32_t), &pkt_size);
                    if (pkt_size == 0 || pkt_size > INGRESS_MAX_PKT_LENGTH)
                    {
                        sprintf(disp_str,  "\nIncorrect Ingress Pkt length = %u (rd_ptr = %08x)!!\n", pkt_size, rd_ptr);
                        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        break;
                    }
                    else
                    {
                        if (++rd_ptr * sizeof(u32_t) >= INGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                        for (index = 0;   index < RND(pkt_size) ; index++)
                        {
                            dbg_reg_read( addr + rd_ptr * sizeof(u32_t), pkt_buffer+index);
                            if (++rd_ptr * sizeof(u32_t) >= INGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                        }
                        if (index < RND(pkt_size))
                        {
                            sprintf(disp_str,  "\nIncomplete Ingress Pkt length: %d out of %d bytes !!\n", (u32_t)(index * sizeof(u32_t)), pkt_size);
                            Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        }
                        else
                        {
                            ump_fifo_flat_space_hack(pkt_buffer, index);              
                            DecodeNcsiRmiiCmdRspPkt(rc_ptr, (NcsiRmiiCmdRspPkt_t *)pkt_buffer);
                        }
                    }
                }
            }
            else  // default setting (no offset input)
            {
                // read ingress fifo ptrs
                addr = MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_PTRS;
                status = dbg_reg_read(addr, &data);
                if (status == DBG_STATUS_SUCCESS)   
                    DecodeMcpGrcRegDetail(rc_ptr, addr, data, FALSE);
                rd_ptr = ((data & MCPR_UMP_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_RD_PTR) >> 0 ) ;
                wr_ptr = ((data & MCPR_UMP_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_WR_PTR) >> 16) ;

                addr = MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_FLAT_SPACE;
                // We need to filter out the wrap around bit in RD/WR pointer
                rd_ptr =  rd_ptr & 0x3FF;
                wr_ptr =  wr_ptr & 0x3FF;


                if (rd_ptr == wr_ptr)  // no unprocessed pkt
                {
                    temp_rd_ptr = rd_ptr;
                    // Show last process Ingress Pkt
                    Tcl_AppendStringsToObj( rc_ptr, "\nLast Processed Ingress Pkt :\n", NULL);
                    last_pkt = &pkt_buffer[INGRESS_MAX_PKT_LENGTH/sizeof(u32_t)];
                    for (index = 0; index < INGRESS_MAX_PKT_LENGTH/sizeof(u32_t); index++)
                    {
                        last_pkt --;
                        temp_rd_ptr = (temp_rd_ptr == 0) ? INGRESS_FLAT_SPACE_SIZE / sizeof(u32_t) - 1 : temp_rd_ptr - 1;
                        dbg_reg_read(addr + temp_rd_ptr * sizeof(u32_t), last_pkt);
                        if (index >= INGRESS_MIN_PKT_LENGTH/sizeof(u32_t) && RND(*last_pkt) == index)
                        {
                            last_pkt++;                                            
                            break;  // found a full packet in flat space
                        }
                    }
                    if (index >= INGRESS_MAX_PKT_LENGTH/sizeof(u32_t))
                    {
                        Tcl_AppendStringsToObj( rc_ptr, "\nIngress Pkt length too long !!\n", NULL);
                    }
                    else
                    {
                        ump_fifo_flat_space_hack(last_pkt, index);              
                        DecodeNcsiRmiiCmdRspPkt(rc_ptr, (NcsiRmiiCmdRspPkt_t *)last_pkt);
                    }
                }

                temp = 1;  // use as counter
                while (rd_ptr != wr_ptr)
                {
                    sprintf(disp_str,  "\nUnprocessed Ingress Pkt #%d:\n", temp++);
                    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                    dbg_reg_read(addr + rd_ptr * sizeof(u32_t), &pkt_size);
                    if (pkt_size > sizeof(pkt_buffer))
                    {
                        sprintf(disp_str, "\nIncorrect Ingress Pkt length = %d!!\n", pkt_size); 
                        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        break;
                    }
                    if (++rd_ptr * sizeof(u32_t) >= INGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                    for (index = 0;   index < RND(pkt_size) && rd_ptr != wr_ptr ; index++)
                    {
                        dbg_reg_read(addr + rd_ptr * sizeof(u32_t), pkt_buffer + index);
                        if (++rd_ptr * sizeof(u32_t) >= INGRESS_FLAT_SPACE_SIZE)   rd_ptr = 0;
                    }
                    if (index < RND(pkt_size))
                    {
                        sprintf(disp_str,  "\nIncomplete Ingress Pkt length: %d out of %d bytes !!\n", (u32_t)(index * sizeof(u32_t)), pkt_size);
                        Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
                        break;
                    }
                    else
                    {
                        ump_fifo_flat_space_hack(pkt_buffer, index);              
                        DecodeNcsiRmiiCmdRspPkt(rc_ptr, (NcsiRmiiCmdRspPkt_t *)pkt_buffer);
                    }
                }
            }
        }
        else if(strcmp("scheduler", block_name) == 0)
        {   
            struct multi_thread_t buf;
            u32 stack_pointer, index, stack_depth[NUM_THREADS];
            // Both paths point to the same multi-thread structure.
            dbg_reg_read(MISC_REG_GENERIC_CR_0, &shmem2_base);
           
            if (!shmem2_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem baes address !!", NULL);
                return TCL_ERROR;
            }
            dbg_reg_read(shmem2_base + OFFSETOF(shmem2_region_t, multi_thread_data_offset), &addr);
            
            addr =  MCP_SPAD_ADDR(addr); 
            for( index = 0 ; index < sizeof(struct multi_thread_t)/sizeof(u32_t); ++index, addr+=4 )
            {
                rc = read_spad( dreg, addr, &value );
                if( rc != TCL_OK )
                    return TCL_ERROR;
                ((u32*)&buf)[index] = value;
            }
            for (index = 0; index < NUM_THREADS; index++)  { // u32 array 
                for (stack_pointer = (((struct multi_thread_t *)(&buf))->thread[index].stack_guard_addr + 4);
                     stack_pointer < (((struct multi_thread_t *)(&buf))->thread[index].stack_guard_addr + THREAD_STACK_SIZE - 4);
                     stack_pointer += 4) {
                //  printf("stack_pointer = 0x%x -> 0x%x\n", stack_pointer, MCP_HOST_ADDR(stack_pointer));
                     rc = dbg_reg_read( MCP_HOST_ADDR(stack_pointer), &value );
                     if( rc != DBG_STATUS_SUCCESS )
                         return TCL_ERROR;
                     if( value != STACK_FILL )
                         break;
                }
                stack_depth[index] = (THREAD_STACK_SIZE - (stack_pointer - (((struct multi_thread_t *)(&buf))->thread[index].stack_guard_addr + 4)));
            }
            DecodeScheduler(rc_ptr, &buf, stack_depth);
        }
        else if  (strcmp("lfa", block_name) == 0)
        {   
            u32 lfa_addr;
            struct shmem_lfa buf;
            u32 *buf_ptr = (u32 *)&buf;

            // Both paths point to the same multi-thread structure.

            common_nvm_get_path_num(&path_id);
            
            if (path_id == 0) {
                dbg_reg_read(MISC_REG_GENERIC_CR_0, &shmem2_base);
            } else {
                dbg_reg_read(MISC_REG_GENERIC_CR_1, &shmem2_base);
            }
            
            if (!shmem2_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem baes address !!", NULL);
                return TCL_ERROR;
            }
            dbg_get_port_num(&port_id);
            dbg_reg_read(shmem2_base + OFFSETOF(shmem2_region_t, lfa_host_addr[port_id]), &lfa_addr);

            for (index = 0; index < sizeof(struct shmem_lfa)/sizeof(u32_t); index++)
            {
                dbg_reg_read(lfa_addr+index*sizeof(u32_t), buf_ptr + index);
            }
            DecodeLfa(rc_ptr, path_id, port_id, &buf);
        }
        else if  (strcmp("swims", block_name) == 0) 
        {   
            swim_meta_t swim_meta;
            swim_stack_t swim_stack[NUM_THREADS];
            u32 index;
            u32 addr;
            u32 image_size;
            u32 crc_computed, crc_in_scratch;
            u8 buff[0x10000];

            // read swim static init address from shmem2
            dbg_reg_read(MISC_REG_GENERIC_CR_0, &shmem2_base);
            if (!shmem2_base) 
            {
                Tcl_AppendStringsToObj( rc_ptr, "Invalid shmem baes address !!", NULL);
                return TCL_ERROR;
            }
            dbg_reg_read(shmem2_base + OFFSETOF(shmem2_region_t, swim_base_addr), &addr);

            // fill structures by reading them from scratchpad
            for (index = 0; index < sizeof(swim_meta_t)/sizeof(u32_t); ++index, addr+=4)
            {
                rc = read_spad(dreg, addr, &value);
                if (rc != TCL_OK)
                    return TCL_ERROR;
                ((u32*)&swim_meta)[index] = value;
            }
            for (index = 0; index < (sizeof(swim_stack_t)*NUM_THREADS)/sizeof(u32_t); ++index, addr+=4)
            {
                rc = read_spad(dreg, addr, &value);
                if (rc != TCL_OK)
                    return TCL_ERROR;
                ((u32*)&swim_stack)[index] = value;
            }

            // calculate crc of image in srcatchpad, and read the crc expected from end of scratchpad
            image_size = swim_meta.info[swim_meta.cur_loaded_swim].image_size;
            for (index = 0, addr = 0x8000040; addr < (0x8000040 + image_size); ++index, addr+=4)
            {
                rc = read_spad(dreg, addr, &value);
                if (rc != TCL_OK)
                    return TCL_ERROR;
                ((u32*)&buff)[index] = value;
            }
            crc_computed = 0xffffffff;
            common_nvm_compute_crc(buff, image_size-4, 0, &crc_computed);
            crc_computed = os_if_cpu_to_le32(~htonl(crc_computed));
            read_spad(dreg, 0x8000040 + image_size - 4, &crc_in_scratch);

            // decode structures and print to user
            DecodeSwims(rc_ptr, &swim_meta, swim_stack, crc_computed, crc_in_scratch);            
        }
        else if (Tcl_GetIntFromObj(interp, objv[1], (int *)&addr) != TCL_ERROR)
        {
            u32_t index, count;
            if (objc < 3 || (Tcl_GetIntFromObj(interp, objv[2], (int *)&count) != TCL_OK))
                count = 1;

            addr &= ~0x3;  // make sure addr is in word-boundary
            for (index = 0; index < count; index++, addr+= 4)
            {
                // check to see if NIG or MCP GRC addr
                if (addr >= 0x10000 && addr < 0x20000)   // NIG GRC
                {
                    status = dbg_reg_read(addr, &data);
                    if (status == DBG_STATUS_SUCCESS)   
                        DecodeNigGrcRegDetail(rc_ptr, addr, data);
                }
                else if (addr >= 0x80000 && addr < 0xC0000)   // MCP GRC
                {
                    status = dbg_reg_read(addr, &data);
                    if (status == DBG_STATUS_SUCCESS)   
                        // display help only if we look at single reg
                        DecodeMcpGrcRegDetail(rc_ptr, addr, data, count==1);
                }
                else
                {
                    Tcl_AppendStringsToObj( rc_ptr, "Not supported yet!!!", NULL);
                    return TCL_ERROR;
                }
            }
        }
        else
        {
            Tcl_AppendStringsToObj( rc_ptr, "Unrecognized argument (use mcp or nig or <addr>)!!!", NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;

}





