#include <string.h>
#include "man_types.h"
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "man_types.h"
#include "tcl_al_common.h"
#include "util.h"
#include "disputil.h"
#include "dbg_man.h"

/* global vars */
static u32_t old_offset = 0;
static u32_t old_length = 4;


typedef enum opers {
    READ = 0, 
    SHOW
} reg_op_t;

static void usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "pcicfg read|write|show <args>");
}

static void read_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "pcicfg read <addr>");
}

static void show_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "pcicfg show <begin addr> <nbytes>");
}


/*********************************************************************/

/**
 * Description:
 *     Implements "pcicfg show" command.
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
static int pcicfg_show_cmd (
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    u32_t *ptr, bytecount, maxcount, addr, i, tmp_addr;
    dbg_status_t status;

    if ( objc > 2 ) {
        show_usage(interp);
        return TCL_ERROR;
    }

    switch (objc)
    {
        case 0:  /* pcicfg show */
                addr = (old_offset + old_length) & DWORD_ADDR_MASK32;
                bytecount = old_length;
                break;
        case 1: /* pcicfg show <addr> */
                if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR)
                    return TCL_ERROR;
                bytecount = 4;
                break;
        case 2: /* pcicfg show <addr> <nbytes> */
                if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR)
                    return TCL_ERROR;
                if (Tcl_GetIntFromObj(interp, objv[1], (int *)&bytecount) == TCL_ERROR)
                    return TCL_ERROR;
                addr &= DWORD_ADDR_MASK32;

                /* Set the cap on how much to dump */
                maxcount = MAX_PRINTF_BUFFER / CHARS_PER_LINE * 32;
                if (DISPLAY_IN_BYTES(interp)) maxcount /= 2;

                if (bytecount > maxcount)
                {
                    printf("Bytecount exceeded, only 0x%x bytes will be displayed\n", maxcount);
                    bytecount = maxcount;
                }
                bytecount = ALIGN_UP4_32(bytecount);
                break;
        default:
                Tcl_SetObjResult(interp, Tcl_NewStringObj("Uncaught error", 0));
                return TCL_ERROR;
    }

    ptr = (u32_t *) Tcl_Alloc (bytecount);
    for (i=0, tmp_addr=addr; i<bytecount/4; i++, tmp_addr+=4)
    {
        status = dbg_pcicfg_read(tmp_addr, &(ptr[i]));
        if (status != DBG_STATUS_SUCCESS) {
            Tcl_SetObjResult(interp, Tcl_NewStringObj("Bad device or address", 0));
            Tcl_Free( (void*)ptr );
            return TCL_ERROR;
        }
    }

    if (DISPLAY_IN_BYTES(interp))
    {
        disp_8bit_array( (u8_t *)ptr, addr, bytecount );
    }
    else
    {
        disp_32bit_array( ptr, addr, bytecount/4, 0 );
    }
    Tcl_Free( (void*)ptr );

    /* Store the address and length for the next command */
    old_offset = addr;
    old_length = bytecount;

    /* Clear the result object */
    Tcl_SetObjResult(interp, Tcl_NewStringObj("", 0));

    return TCL_OK;
}


/*********************************************************************/

/**
 * Description:
 *     Implements "pcicfg read" command.
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
static int pcicfg_read_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    Tcl_Obj *result_ptr = NULL;
    u32_t addr;
    /*u8_t val8;
    u16_t val16;*/
    u32_t val32;
    dbg_status_t status;
    char result_str[16];

    /**
     * Check arguments list. We'll support only the simpliest memory read operation:
     * a read from one address.
     */ 
    if (objc < 1 ) {
        read_usage(interp);
        return TCL_ERROR;
    }

    /**
     * Object which will be used to create a return value
     */ 
    result_ptr = Tcl_GetObjResult(interp);

    /**
     * Parse the address
     */ 
    if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR) {
        return TCL_ERROR;
    }

    /**
     * Parse the number of bytes to read
     */ 
    /*if (Tcl_GetIntFromObj(interp, objv[1], (int *)&nbytes) == TCL_ERROR) {
        return TCL_ERROR;
    }*/

/*    printf("BUS: %d, Device: %d, Function: %d\n", dev_data.pci_info.bus_num, dev_data.pci_info.device,
           dev_data.pci_info.func);
*/
    /*switch(nbytes) {
    case 1:
        if ( my_read_configuration_byte(dev_data.pci_info.bus_num, 
                                        ((dev_data.pci_info.device << 3) | dev_data.pci_info.func), 
                                        addr, &val8) != 0 ){
            printf("PCI cycle failed\n");
            return TCL_ERROR;
        }
        Tcl_SetIntObj(result_ptr, val8);
        break;
    case 2:
        if ( my_read_configuration_word(dev_data.pci_info.bus_num, 
                                        ((dev_data.pci_info.device << 3) | dev_data.pci_info.func), 
                                        addr, &val16) != 0 ){
            printf("PCI cycle failed\n");
            return TCL_ERROR;
        }
        Tcl_SetIntObj(result_ptr, val16);
        break;
    case 4: */
        status = dbg_pcicfg_read(addr, &val32);

        if ( status == DBG_STATUS_SUCCESS )
            sprintf(result_str, "0x%x", val32);
        else         
            /* if operation failed return a phony value of 0xFFFFFFFF
               in order to be consistent with xdiag and common scripts */            
            sprintf(result_str, "0x%x", 0xffffffff);
        
        
    /*    break;
    default:
        Tcl_WrongNumArgs(interp, 1, objv, "1, 2 or 4");
        return TCL_ERROR;
    }
*/
    //Tcl_SetIntObjDispType(result_ptr, DISP_32_HEX);
    Tcl_SetStringObj(result_ptr, result_str, strlen(result_str));
    return TCL_OK;
}

/**
 * Description:
 *     Implements "pcicfg" command.
 * 
 * Returns:
 *     A standard Tcl result.
 */ 
int tcl_al_pcicfg_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
    static char *option_names[] = { "read", "show", (char *)NULL };
    int cmd_idx;

    /**
     * Check the number of arguments - must be at least 2 (pcicfg <op. type>  [<address>] [<value (for write)>])
     */ 
    if (objc < 2) {
        usage(interp);
        return TCL_ERROR;
    }

    /**
     * Get op. type - find the op. index in option_names array
     */ 
    if (Tcl_GetIndexFromObj(interp, objv[1], (const char**)option_names, "option", TCL_EXACT, 
                            &cmd_idx) != TCL_OK) {
        return TCL_ERROR;
    }

    switch (cmd_idx) {
    case READ:
        return pcicfg_read_cmd(interp, objc - 2, &(objv[2])); /* Cut the "read/write" prefix */    
    case SHOW:
        return pcicfg_show_cmd(interp, objc - 2, &(objv[2])); /* Cut the "read/write" prefix */
    default:
        Tcl_AppendResult(interp, "Unknown operation type",
                         NULL);
        return TCL_ERROR;
    }
}

/**
 * Description:
 *      "pcicfg" cleanup handler.
 * Returns:
 *      Standard TCL result.
 */ 
int tcl_al_pcicfg_cleanup(
    IN   Tcl_Interp *interp
    )
{
    
    return Tcl_DeleteCommand(interp, PCICFG_CMD);
}

/*
 *------------------------------------------------------------------------
 * tcl_al_pcicfg_usage --
 *     Add usage information on "pcicfg" command 
 *------------------------------------------------------------------------
 */
int tcl_al_pcicfg_usage(
	IN	char *buf,				//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{	
	u32_t offset = 0;
	*length = 0;

	*length += sprintf( &buf[offset+*length], "Read data from PCI configuration space using PCI config cycle.\n\n");
    *length += sprintf( &buf[offset+*length], "Usage: pcicfg read <addr>\n" );
    *length += sprintf( &buf[offset+*length], "      pcicfg show <begin_addr> <nbytes>\n" );
    	  
    return TCL_OK;
}

