#include <string.h>
#include "man_types.h"
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "man_types.h"
#include "tcl_al_common.h"
#include "util.h"
#include "disputil.h"
#include "dbg_man.h"
#include "os_utils.h"

/* externals */
extern u32_t uart_com;
extern u32_t uart_baud;
extern u8_t is_logfile;
extern char *logfile_name;

/* global vars */
static u32_t old_offset = 0;
static u32_t old_length = 4;

typedef enum opers {
    READ = 0,
    DREAD,
    SHOW,
    WRITE,
    DWRITE
} reg_op_t;

static void usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "reg read|dread|write|dwrite|show <args>");
}

static void read_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "reg read|dread <addr>");
}

static void write_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "reg write|dwrite <addr> <value>");
}

static void show_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "reg show[<begin addr> [<nbytes>] ]");
}

/**
 * Description:
 *     Implements "reg read" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
static int reg_read_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[],  /* argument strings */
    IN    int cmd                 /* Command type */
    )
{
        Tcl_Obj *result_ptr = NULL;
        u32_t addr;
        /*u8_t val8;
        u16_t val16;*/
        u32_t val32;
        char result_str[16];
        FILE *fd;

        /**
        * Check arguments list. We'll support only the simpliest memory read operation:
        * a read from one address.
        */
        if (objc < 1 ) {
                read_usage(interp);
                return TCL_ERROR;
        }

        /**
        * Object which will be used to create a return value
        */
        result_ptr = Tcl_GetObjResult(interp);

        /**
        * Parse the address
        */
        if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR) {
                return TCL_ERROR;
        }

        switch (cmd) {
        case READ:
                if ( dbg_reg_read(addr, &val32) != DBG_STATUS_SUCCESS ){
                        return TCL_ERROR;
                }
                break;
        case DREAD:
                if ( dbg_reg_dread(addr, &val32) != DBG_STATUS_SUCCESS ){
                        return TCL_ERROR;
                }
                break;
        }

        sprintf(result_str, "0x%x", val32);

        Tcl_SetStringObj(result_ptr, result_str, strlen(result_str));

        /* Record the result to the log file */
        if (is_logfile) {
            fd = os_if_log_fopen(logfile_name);
            fprintf(fd, "%s", result_str);
            fclose(fd);
        }   

    return TCL_OK;
}

/**
 * Description:
 *     Implements "reg write" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
static int reg_write_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[],  /* argument strings */
    IN    int cmd                 /* Command type */
    )
{
        Tcl_Obj *result_ptr = NULL;
        u32_t addr, val;
        u8_t  wait_for_echo=1;      //By default wait for echo on write request

        /**
        * Check arguments list. We'll support only the simpliest memory read operation:
        * a read from one address.
        */
        if (objc < 2) {
                write_usage(interp);
                return TCL_ERROR;
        }

        /**
        * Object which will be used to create a return value
        */
        result_ptr = Tcl_GetObjResult(interp);

        /**
        * Parse the address
        */
        if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR) {
                return TCL_ERROR;
        }

        /**
        * Parse the value
        */
        if (Tcl_GetIntFromObj(interp, objv[1], (int *)&val) == TCL_ERROR) {
                return TCL_ERROR;
        }

        if (objc == 3) {
                /* If arg 2 given "no_echo" set the wait_for_echo flag to zero */
                if (strcmp("-no_echo", Tcl_GetStringFromObj(objv[2], NULL)) == 0) {
                    wait_for_echo = 0;
                }
        }


        switch (cmd) {
        case WRITE:
                if ( dbg_reg_write(addr, val, wait_for_echo) != DBG_STATUS_SUCCESS ){
                    return TCL_ERROR;
                }
                break;
        case DWRITE:
                if ( dbg_reg_dwrite(addr, val) != DBG_STATUS_SUCCESS ){
                    return TCL_ERROR;
                }
                break;
        }

        return TCL_OK;
}

/**
 * Description:
 *     Implements "reg show" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
static int reg_show_cmd (
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[],  /* argument strings */
    IN    int cmd                 /* Command type */
    )
{
        u32_t *ptr, bytecount, maxcount, addr, i, tmp_addr;
        dbg_status_t status;

        if ( objc > 2 ) {
                show_usage(interp);
                return TCL_ERROR;
        }

        switch (objc)
        {
        case 0:
                addr = (old_offset + old_length) & DWORD_ADDR_MASK32;
                bytecount = old_length;
                break;
        case 1:
                if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR)
                        return TCL_ERROR;
                bytecount = 4;
                break;
        case 2:
                if (Tcl_GetIntFromObj(interp, objv[0], (int *)&addr) == TCL_ERROR)
                        return TCL_ERROR;
                if (Tcl_GetIntFromObj(interp, objv[1], (int *)&bytecount) == TCL_ERROR)
                        return TCL_ERROR;
                addr &= DWORD_ADDR_MASK32;

                /* Set the cap on how much to dump */
                maxcount = MAX_PRINTF_BUFFER / CHARS_PER_LINE * 32;

                if (DISPLAY_IN_BYTES(interp)) maxcount /= 2;

                if (bytecount > maxcount)
                {
                    printf("Bytecount exceeded, only 0x%x bytes will be displayed\n", maxcount);
                    bytecount = maxcount;
                }
                bytecount = ALIGN_UP4_32(bytecount);
                break;
        default:
                Tcl_SetObjResult(interp, Tcl_NewStringObj("Uncaught error", 0));
                return TCL_ERROR;
        }

        ptr = (u32_t *) Tcl_Alloc (bytecount);
        switch (cmd) {
        case SHOW:
                for ( i = 0, tmp_addr = addr; i < bytecount/4; i++, tmp_addr += 4)
                {
                    status = dbg_reg_read(tmp_addr, &(ptr[i]));
                    if (status != DBG_STATUS_SUCCESS) {
                        Tcl_SetObjResult(interp, Tcl_NewStringObj("Bad device or address", 0));
                        Tcl_Free( (void*)ptr );
                        return TCL_ERROR;
                    }
                }
                break;
        }


        if (DISPLAY_IN_BYTES(interp))
        {
                disp_8bit_array( (u8_t *)ptr, addr, bytecount );
        }
        else
        {
                disp_32bit_array( ptr, addr, bytecount/4, 0 );
        }
        Tcl_Free( (void*)ptr );

        /* Store the address and length for the next command */
        old_offset = addr;
        old_length = bytecount;

        /* Remove the list from result_ptr */
        Tcl_SetObjResult(interp, Tcl_NewStringObj("", 0));

        return TCL_OK;
}




/**
 * Description:
 *     Implements "reg" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
int tcl_al_reg_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
        static char *option_names[] = { "read", "dread", "show", "write", "dwrite", (char *)NULL };
        int cmd_idx;
        int rc;

        /**
        * Check the number of arguments - must be at least 2 (reg <op. type>  [<address>] [<value (for write)>])
        */
        if (objc < 2) {
                usage(interp);
                return TCL_ERROR;
        }

        /**
        * Get op. type - find the op. index in option_names array
        */
        if (Tcl_GetIndexFromObj(interp, objv[1], (const char**)option_names, "option", TCL_EXACT,
                            &cmd_idx) != TCL_OK) {
                return TCL_ERROR;
        }

        switch (cmd_idx) {
        case READ:
        case DREAD:
                rc = reg_read_cmd(interp, objc - 2, &(objv[2]), cmd_idx); /* Cut the "read/write" prefix */
                if ((tcl_al_get_no_pci_mode() != 0) && rc) {
                    tcl_al_reinit_non_pci_mode(interp, uart_com, uart_baud);
                    printf("\n\n serial connection was broken. Recovered... \n\n");
                    return reg_read_cmd(interp, objc - 2, &(objv[2]), cmd_idx);
                } else {
                    return TCL_OK;
                }
        case WRITE:
        case DWRITE:
                rc = reg_write_cmd(interp, objc - 2, &(objv[2]), cmd_idx); /* Cut the "read/write" prefix */
                if ((tcl_al_get_no_pci_mode() != 0) && rc) {
                    tcl_al_reinit_non_pci_mode(interp, uart_com, uart_baud);
                    printf("\n\n serial connection was broken. Recovered... \n\n");
                    return reg_write_cmd(interp, objc - 2, &(objv[2]), cmd_idx);
                } else {
                    return TCL_OK;
                }
        case SHOW:
                rc = reg_show_cmd(interp, objc - 2, &(objv[2]), cmd_idx); /* Cut the "read/write" prefix */
                if ((tcl_al_get_no_pci_mode() != 0) && rc) {
                    tcl_al_reinit_non_pci_mode(interp, uart_com, uart_baud);
                    printf("\n\n serial connection was broken. Recovered... \n\n");
                    return reg_show_cmd(interp, objc - 2, &(objv[2]), cmd_idx);
                } else {
                    return TCL_OK;
                }
        default:
                Tcl_AppendResult(interp, "Unknown operation type",
                                 NULL);
                return TCL_ERROR;
        }
}

int tcl_al_reg_cleanup(
    IN   Tcl_Interp *interp
    )
{
    Tcl_DeleteCommand(interp, REG_CMD);
    return TCL_OK;
}
/*
 *------------------------------------------------------------------------
 * tcl_al_reg_usage --
 *     Add usage information on "reg" command
 *------------------------------------------------------------------------
 */
int tcl_al_reg_usage(
	IN	char *buf,				//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{
        u32_t offset = 0;
        *length = 0;

        *length += sprintf( &buf[offset+*length], "Read, show or write a 32-bit value to a register\n\n");
        *length += sprintf( &buf[offset+*length], "Usage: reg read <addr>\n" );
        *length += sprintf( &buf[offset+*length], "       reg write <addr> <value>\n" );
        *length += sprintf( &buf[offset+*length], "       reg show [<addr> [bytecount]]\n" );
        *length += sprintf( &buf[offset+*length], "       reg dread <addr>\n" );
        *length += sprintf( &buf[offset+*length], "       reg dwrite <addr> <value>\n" );
        *length += sprintf( &buf[offset+*length], "       reg iread <addr>\n" );
        *length += sprintf( &buf[offset+*length], "       reg iwrite <addr> <value>\n" );
        *length += sprintf( &buf[offset+*length], "       reg ishow [<addr> [bytecount]]\n" );
        *length += sprintf( &buf[offset+*length], "Note: All subcommands with 'i' refer to indirect access.\n" );
        *length += sprintf( &buf[offset+*length], "      All subcommands with 'd' refer to direct access.\n" );

        return TCL_OK;
}
