
/******************************************************************************/
/* The information contained in this file is confidential and proprietary to  */
/* QLogic Corporation.  No part of this file may be reproduced or             */
/* distributed, in any form or by any means for any purpose, without the      */
/* express written permission of QLogic Corporation.                          */
/*                                                                            */
/* (c) COPYRIGHT 2011-2014 Broadcom Corporation, ALL RIGHTS RESERVED.         */
/* (c) COPYRIGHT 2014-2015 QLogic Corporation, ALL RIGHTS RESERVED.           */
/*                                                                            */
/* Module Description:  processes the serial port command to do console       */
/*                      redirection.                                          */
/*                                                                            */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "os_types.h"
#include "tcl.h"
#include "bcmtype.h"
#include "esc_seq.h"
#include "tcl_al_cmd_handlers.h"
#include "dbg_serport.h"
#include "tcl_al_common.h"
#include "tcl_al.h"
#include "hw_access.h"

#define MAX_RETRIES 3
#define WAIT_MS     50

/* forward declarations */
static int validate_port_is_open(void);

/* holds if we are still before first serial port command */
u32_t firstSerialPortCommand = 1;

/*
 *------------------------------------------------------------------------
 * tcl_al_serport_usage --
 *     CB for "help" command to display help
 *
 * Results:
 *     None.
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
int tcl_al_serport_usage(
	char *buf,
	u32_t *len
	) {
	u32_t offset = 0;

	*len = 0;
	*len += sprintf(&buf[offset + *len], "Serial port I/O.\n\n");
	*len += sprintf(&buf[offset + *len], "Usage: serialport open <port#> [baudrate [settings]] \n");
	*len += sprintf(&buf[offset + *len], "                            # Open COM port \n");
	*len += sprintf(&buf[offset + *len], "                            # Default values for baudrate and settings are:\n");
	*len += sprintf(&buf[offset + *len], "                            # 115200 and 8N1 respectively\n");
	*len += sprintf(&buf[offset + *len], "                            # Supported baud rates are 600, 1200, 4800, 9600, 19200,\n");
	*len += sprintf(&buf[offset + *len], "                            #  38400, 57600, 115200\n");
	*len += sprintf(&buf[offset + *len], "       serialport redirect duart\n");
	*len += sprintf(&buf[offset + *len], "                            # Redirect I/O or GRC transations to COM port\n");
	*len += sprintf(&buf[offset + *len], "                            # as defined by $sys(IO_PORT) and $sys(DUAT_PORT)\n");
	*len += sprintf(&buf[offset + *len], "       serialport close     # Close all COM ports and redirect back to console\n");
	*len += sprintf(&buf[offset + *len], "Example:\n");
	*len += sprintf(&buf[offset + *len], "         :> serialport open 1 115200 8N1\n");
	*len += sprintf(&buf[offset + *len], "         :> serialport open 2 1200 8N1\n");
	*len += sprintf(&buf[offset + *len], "         :> set sys(IO_PORT)    1\n");
	*len += sprintf(&buf[offset + *len], "         :> set sys(DUART_PORT) 2\n");
	*len += sprintf(&buf[offset + *len], "         :> serialport redirect duart\n");

	return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * tcl_al_serport_cleanup --
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
int
tcl_al_serport_cleanup(
	Tcl_Interp *interp     /* current interpreter */
	) {
	dbg_serport_cleanup();

	/* De-register the command from Tcl */
	Tcl_DeleteCommand(interp, SERPORT_CMD);
	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_serport_open_cmd
 *
 *     Implements the new Tcl "serialport open" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static int tcl_al_serport_open_cmd(
	ClientData client_data,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
	) {
	u32_t baud, port;
	unsigned char setting[5] = "8N1";
	u32_t	settings_str__len = 0;
	u8_t *usr_settings_str;
	dbg_status_t stat;

	baud = UA_DEFAULT_BAUD;
	setting[3] = '\0';

	if (objc < 2) {

		Tcl_WrongNumArgs(interp, 1, objv, "?serial_port_number?");
		return TCL_ERROR;
	}

	/* Get serial port number */
	if (Tcl_GetIntFromObj(interp, objv[1], (int *)&port) == TCL_ERROR) return TCL_ERROR;

	if (objc >= 3) {
		/* parse for baudrate */
		if (Tcl_GetIntFromObj(interp, objv[2], (int *)&baud) == TCL_ERROR) return TCL_ERROR;
	}
	if (objc >= 4) {
		/* parse for settings */
		usr_settings_str = (u8_t *)Tcl_GetStringFromObj(objv[3], (int *)&settings_str__len);
		if (settings_str__len < 3) {
			printf("Error: value for setting is incorrect. Example values are 8N1, 7E1, 8O1 \n");
			return TCL_ERROR;
		}

		setting[0] = *((char *)usr_settings_str);
		setting[1] = *((char *)(usr_settings_str + 1));
		setting[2] = *((char *)(usr_settings_str + 2));
	}

	printf("opening com port %d baud %d\n", port, baud);
	stat = dbg_serport_open(setting, port, baud);
	if (stat != DBG_STATUS_SUCCESS) {
		return TCL_ERROR;
	}

	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_serialport_close_cmd
 *
 *     Implements the new Tcl "serialport close" command.  This command will
 *     attempt to close the openned serial port and also redirects command
 *     prompt back to stdio.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static int tcl_al_serport_close_cmd(
	ClientData client_data,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
	) {
	if (objc != 1) {
		Tcl_WrongNumArgs(interp, 1, objv, ":too many args");
		return TCL_ERROR;
	}

	if (dbg_serport_close() != DBG_STATUS_SUCCESS) {
		return TCL_ERROR;
	}


	return TCL_OK;
}

static int redirect_duart(
	Tcl_Interp *interp
	) {
	Tcl_Obj *port_obj;
	dbg_status_t stat;
	int io_port;

	/* Set up the uart_io_handle. */
	port_obj = Tcl_GetVar2Ex(interp, TCL_SYS_ENV, ENV_DUART_PORT, 0);

	if (Tcl_GetIntFromObj(interp, port_obj, &io_port) == TCL_OK) {
		if ((io_port > MAX_UART_HANDLES) || (io_port < 0)) {
			printf("$::sys(DUART_PORT) value is illegal: must be between 0 and %d\n", MAX_UART_HANDLES);
			return TCL_ERROR;
		}
		printf("Serial port %d is used for DUART redirection\n", io_port);
	} else {
		printf("$::sys(DUART_PORT) must be set!\n");
		return TCL_ERROR;
	}

	/* If the DUART handle is 0 which means the I/O port has not been opened yet,   
	 * no redirection can be enabled because nothing can work without I/O.
	 */
	stat = dbg_set_duart_port(io_port);
	switch (stat) {
	case DBG_STATUS_CON_CLOSED:
		printf("Error: redirection failed because the serial port (%d) has not been opened.\n", io_port);
		return TCL_ERROR;
	case DBG_STATUS_SUCCESS:
		if (validate_port_is_open() != TCL_OK) {
			printf("ERROR - validate_port_is_open failed !\n");
			return TCL_ERROR;
		}

		if (dbg_setup_duart_nvm() != DBG_STATUS_SUCCESS) {
			printf("redirect_duart: dg_setup_duart_nvm failed !\n");
			// Don't fail incase failed to set DUART NVM
			//return TCL_ERROR;
		}

		break;

	default:
		printf("redirect_duart: serial port open failure\n");
		return TCL_ERROR;
	}

	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * validate_port_is_open
 *
 *     Validates com port is opened correctly by reading one register
 *     and validating its value
 *
 * Results:
 *     TCL_OK if register value is fine, otherwise TCL_ERROR
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static int validate_port_is_open(void) {
	u32_t tmpVal = (u32_t)-1;
	dbg_status_t status = (dbg_status_t)DBG_STATUS_FAILURE;
	u32_t num_retries = MAX_RETRIES;

	while ((num_retries > 0) && (status != DBG_STATUS_SUCCESS)) {
		status = dbg_reg_read(BAR_ME_REGISTER, &tmpVal);
		if ((status == DBG_STATUS_SUCCESS) && (tmpVal != -1)) {
			break;
		}
		num_retries--;
		dbg_msleep(WAIT_MS);
	}

	if ((status != DBG_STATUS_SUCCESS) || ((tmpVal != 0) && (tmpVal != 0x1))) {
		return TCL_ERROR;
	} else {
		return TCL_OK;
	}
}

/*
 *------------------------------------------------------------------------
 * get_serport_chip_id
 *
 *     Get chip id via serial port
 *
 *------------------------------------------------------------------------
 */
int get_serport_chip_id(u32_t *chip_id_p) {

	if (dbg_duart_get_chip_id(chip_id_p) != DBG_STATUS_SUCCESS) {
		printf("get_serport_chip_id: dbg_duart_get_chip_id failed !\n");
		return TCL_ERROR;
	}
	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * tcl_al_serport_redirect_cmd
 *
 *     Implements the new Tcl "serialport redirect" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static int tcl_al_serport_redirect_cmd(
	ClientData client_data,
	Tcl_Interp *interp,
	int objc,
	Tcl_Obj *CONST objv[]
	) {

	if (objc != 2) {
		Tcl_WrongNumArgs(interp, 1, objv, ":too many args");
		return TCL_ERROR;
	} else {
		/* Check that the option key is as expected */
		char *str_val = Tcl_GetStringFromObj(objv[1], NULL);
		if (strcmp(str_val, "duart") != 0) {
			printf("Invalid option: %s\n", str_val);
			return TCL_ERROR;
		}

		return redirect_duart(interp);
	}
}

/*------------------------------------------------------------------------
 * tcl_al_serport_cmd
 *
 *     Implements the new Tcl "serialport" command.
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
int
tcl_al_serport_cmd(client_data, interp, objc, objv)
ClientData client_data; /* Not used. */
Tcl_Interp *interp;     /* current interpreter */
int objc;               /* number of arguments */
Tcl_Obj *CONST objv[];  /* argument strings */
{
	static char *option_names[] = { "open", "redirect", "close",
		(char *)NULL };
	enum options {OPEN, REDIRECT, CLOSE};
	int index;

	/* if it is first serialport command, init appropraite array */
	if (firstSerialPortCommand) {
		dbg_init_uart_info_array();
		firstSerialPortCommand = 0;
	}

	if (objc < 2) {
		Tcl_WrongNumArgs(interp, 1, objv, "option=(open|redirect|close)");
		return TCL_ERROR;
	}

	if (Tcl_GetIndexFromObj(interp, objv[1], (const char **)option_names, "option", 0,
							&index) != TCL_OK) {
		return TCL_ERROR;
	}
	switch ((enum options)index) {
	case OPEN:
		return tcl_al_serport_open_cmd(client_data, interp, objc - 1, objv + 1);
		break;
	case REDIRECT:
		return tcl_al_serport_redirect_cmd(client_data, interp, objc - 1, objv + 1);
		break;
	case CLOSE:
		return tcl_al_serport_close_cmd(client_data, interp, objc - 1, objv + 1);
		break;
	default:
		return TCL_ERROR;
	}
}

