#include <string.h>
#include "man_types.h"
#include "tcl.h"
#include "tclDecls.h"
#include "tcl_al.h"
#include "tcl_al_common.h"
#include "util.h"
#include "disputil.h"
#include "dbg_man.h"

typedef enum opers {
    ENABLE = 0,
    DISABLE,
    SHOW
} stats_op_t;

static void usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "stats enable|disable|show");
}

static void enable_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "stats enable");
}

static void disable_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "stats disable");
}

static void show_usage(Tcl_Interp *interp)
{
    Tcl_WrongNumArgs(interp, 0, NULL, "stats show");
}

/**
 * Description:
 *     Implements "stats enable" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
static int stats_enable_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[],  /* argument strings */
    IN    int cmd                 /* Command type */
    )
{
        /**
        * Check arguments list. We'll support only the simpliest memory read operation:
        * a read from one address.
        */
        if (objc != 0) {
                enable_usage(interp);
                return TCL_ERROR;
        }

        dbg_stats_enable();

        return TCL_OK;
}

/**
 * Description:
 *     Implements "stats disable" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
static int stats_disable_cmd(
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[],  /* argument strings */
    IN    int cmd                 /* Command type */
    )
{
        /**
        * Check arguments list. We'll support only the simpliest memory read operation:
        * a read from one address.
        */
        if (objc != 0) {
                disable_usage(interp);
                return TCL_ERROR;
        }

        dbg_stats_disable();

        return TCL_OK;
}

/**
 * Description:
 *     Implements "stats show" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
static int stats_show_cmd(
	IN    Tcl_Interp *interp,     /* current interpreter */
	IN    int objc, 			  /* number of arguments */
	IN    Tcl_Obj *CONST objv[],  /* argument strings */
	IN    int cmd   			  /* Command type */
	)
{
		/**
		* Check arguments list. We'll support only one statistcis format
		* (per-OS). 
		*/
		if (objc != 0) {
				show_usage(interp);
				return TCL_ERROR;
		}

		if (tcl_al_get_no_pci_mode() != 0) {
			printf("stats show command is not supported in serial mode\n");
			return TCL_ERROR;
		}
		dbg_stats_show();

		return TCL_OK;
}

/**
 * Description:
 *     Implements "stats" command.
 *
 * Returns:
 *     A standard Tcl result.
 */
int tcl_al_stats_cmd (
    IN    ClientData client_data, /* Not used. */
    IN    Tcl_Interp *interp,     /* current interpreter */
    IN    int objc,               /* number of arguments */
    IN    Tcl_Obj *CONST objv[]   /* argument strings */
    )
{
        static char *option_names[] = { "enable", "disable", "show", (char *)NULL };
        int cmd_idx;

        /**
        * Check the number of arguments - must be at least two
        * (stats <op.type>)
        */
        if (objc < 2) {
                usage(interp);
                return TCL_ERROR;
        }

        /**
        * Get op. type - find the op. index in option_names array
        */
        if (Tcl_GetIndexFromObj(interp, objv[1], (const char**)option_names, "option", TCL_EXACT,
                            &cmd_idx) != TCL_OK) {
                return TCL_ERROR;
        }

        switch (cmd_idx) {
        case ENABLE:
                return stats_enable_cmd(interp, objc - 2, &(objv[2]), cmd_idx);
        case DISABLE:
                return stats_disable_cmd(interp, objc - 2, &(objv[2]), cmd_idx);
		case SHOW:
                return stats_show_cmd(interp, objc - 2, &(objv[0]), cmd_idx);
        default:
                Tcl_AppendResult(interp, "Unknown operation type", NULL);
                return TCL_ERROR;
        }
}

int tcl_al_stats_cleanup(
    IN   Tcl_Interp *interp
    )
{
        Tcl_DeleteCommand(interp, STATS_CMD);
        return TCL_OK;
}
/*
 *------------------------------------------------------------------------
 * tcl_al_stats_usage --
 *     Add usage information on "stats" command
 *------------------------------------------------------------------------
 */
int tcl_al_stats_usage(
	IN	char *buf,			//Pointer to the buffer of the command usage
	IN	u32_t *length			//Pointer to the length result of the command usage
	)
{
        u32_t offset = 0;
        *length = 0;

        *length += sprintf( &buf[offset+*length], "Enable or disable stats gathering, show NIC stats\n\n");
        *length += sprintf( &buf[offset+*length], "Usage: stats enable\n" );
        *length += sprintf( &buf[offset+*length], "       stats disable\n" );
        *length += sprintf( &buf[offset+*length], "       stats show\n" );

        return TCL_OK;
}
