#include <stdlib.h>
#include <string.h>
#include "man_types.h"
#include "bcmtype.h"
#include "tcl_al.h"
#include "tcl_al_common.h"
#include "tcl_util.h"
#include "tcl.h"
#include "tclDecls.h"
#include "esc_seq.h"
#include "errcode.h"
#include "os_utils.h"

extern u8_t is_logfile;
extern char *logfile_name;

int print_dbg(char * fmt, ...)
{
   va_list arglist;
   int rc = 0;
   Tcl_Channel outChannel;
   char  print_buf[MAX_PRINTF_BUFFER]; 
   FILE* fd; 
   
   va_start(arglist, fmt);
   vsprintf(print_buf, fmt, arglist);
   va_end(arglist);
   
   fprintf(stdout, "%s", print_buf);
   fflush(stdout);
   
   if (is_logfile) {
       fd = os_if_log_fopen(logfile_name);
       fprintf(fd, "%s", print_buf);         
       fclose(fd);
   }   
   
   return(rc);
}

void set_tcl_prompt(
    IN   Tcl_Interp* interp,
    IN   edebug_eng_mode_t eng_mode
    )
{
        const char *dev_num;
        char prompt[80];

        if ((eng_mode && (eng_mode == EDEBUG_NO_INIT_MODE)) || (dbg_get_total_br_dev() == 0)) {
            sprintf(prompt, "puts -nonewline \"no_devices:> \"");
            Tcl_SetVar2(interp, "tcl_prompt1", NULL, prompt, TCL_GLOBAL_ONLY);
            return;
        }

        dev_num = Tcl_GetVar2(interp, TCL_ENV, ENV_DEV, TCL_GLOBAL_ONLY);

        if (tcl_al_get_no_pci_mode() == 0) {
            /* Set the prompt */
            if ((*dev_num - '1' >= 0 ) && (BR_PCI_DEV_HIDDEN(*dev_num - '1'))) {
                sc_change_color(ESC_COLOR_BROWN, ESC_FOREGROUND);
                sprintf(prompt, "puts -nonewline \"%s.HIDDEN:> \"", dev_num);
            }
            else {
                sc_restore_color();
                sprintf(prompt, "puts -nonewline \"%s:> \"", dev_num);
            }
        }
        else {
             /* Set the prompt */
            sprintf(prompt, "puts -nonewline \"%s-SERIAL:> \"", dev_num);
        }

        Tcl_SetVar2(interp, "tcl_prompt1", NULL, prompt, TCL_GLOBAL_ONLY);
}

/*
 *------------------------------------------------------------------------
 * mac_to_hex:
 *
 * Convert the lower three bytes of the MAC address to hex number
 *------------------------------------------------------------------------
 */
u32_t mac_to_hex (char *mac_str)
{
	int t, i, shift=2;
	u32_t hex=0;

	if (mac_str == NULL)
		return 0;


	for(i=0; i< (int)strlen(mac_str); i++)
	{
		if(mac_str[i] == ':')
		{
			shift = shift-1;
			continue;
		}
		t = strtol(&mac_str[i], NULL, 16);
		hex |= (t << shift*8);
		i++;
	}

	return hex;
}



/*
 *------------------------------------------------------------------------
 * check_mac_addr:
 *
 * Check that mac & iscsi mac addresses are not idential.
 * Check also that the mac addresses are sequential.
 *
 * Inputs:
 * interp				- Tcl interperter
 * mac_addr_p			- pointer to the primary MAC address string
 * iscsi_mac_addr_p		- pointer to the iscsi MAC address string
 * error_code			- the returned error code
 *------------------------------------------------------------------------
 */
u32_t check_mac_addr(Tcl_Interp *interp, char *mac_addr_p, char *iscsi_mac_addr_p, errcode_t* err_code)
{
	if ((*mac_addr_p != '\0') || (*iscsi_mac_addr_p != '\0'))
    {

	   if ((strncmp(iscsi_mac_addr_p, mac_addr_p, 9) != 0) ||
          ((mac_to_hex(iscsi_mac_addr_p+9)) - (mac_to_hex(mac_addr_p+9)) != 1))
		{
			printf( "\nError: Both Primary and iSCSI MAC addresses need to be consecutive.\n");
            *err_code = ERR_MAC_TOO_DIFF;
			return TCL_ERROR;
		}
    }
	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * generate_new_mac:
 *
 * Generate a new MAC addresses from source MAC address,
 * which is +incr from source MAC address.
 *------------------------------------------------------------------------
 */
void generate_new_mac(char *src_mac_str, char *new_mac_str, u8_t incr)
{
    int i;
    u32_t hex=0;
    unsigned char temp;
    char tmpchar[2];
	char tmpStr[30];

    strcpy(new_mac_str, src_mac_str);

    hex = mac_to_hex(src_mac_str+9);
    hex += incr;

    for (i=2; i>=0; i--)
    {
        temp = (unsigned char)(hex >> (i*8)) & 0xff;
        sprintf(tmpchar, "%02x", temp);
        strncpy(new_mac_str+9+(2-i)*3, tmpchar, 2);
    }

    sprintf(tmpStr, "Automatic generated MAC address  %s\n", new_mac_str);
    printf(tmpStr);

    return;
}


 /*
 *------------------------------------------------------------------------
 * program_mac_address:
 *
 * Program the requested MAC addresses (primary,iScsi,rdma) on the requested
 * device using the prg_mac_addr utility
 *
 * Inputs:
 * interp - Tcl interperter
 * mac_addr_str		- pointer to the input MAC address string
 * mac_addr_type	- ADDR_TYPE_PRIMARY or ADDR_TYPE_ISCSI or ADDR_TYPE_RDMA
 * dev_idx			- device number
 * error_code		- the returned error code
 *------------------------------------------------------------------------
 */
 u32_t program_mac_addresses(Tcl_Interp *interp, char *mac_addr_str, char *iscsi_mac_addr_str, char *rdma_mac_addr_str, u8_t dev_idx, errcode_t* err_code)
 {
	/* program the primary mac address. */
	if (*mac_addr_str != '\0')
	{
       if (prg_mac_addr(interp, mac_addr_str, ADDR_TYPE_PRIMARY, dev_idx) != TCL_OK) {
          *err_code = ERR_PRG_PRIM_MAC_ADDR;
          return TCL_ERROR;
       }
	}

	/* program the iSCSI mac address . */
	if (*iscsi_mac_addr_str != '\0')
	{
       if (prg_mac_addr(interp, iscsi_mac_addr_str, ADDR_TYPE_ISCSI, dev_idx) != TCL_OK) {
           *err_code = ERR_PRG_ISCSI_MAC_ADDR;
		   return TCL_ERROR;
       }
	}

	/* program the RDMA mac address . */
	if (*rdma_mac_addr_str != '\0')
	{
       if (prg_mac_addr(interp, rdma_mac_addr_str, ADDR_TYPE_RDMA, dev_idx) != TCL_OK) {
           *err_code = ERR_PRG_RDMA_MAC_ADDR;
		   return TCL_ERROR;
       }
	}
	return TCL_OK;

 }

/*
 *------------------------------------------------------------------------
 * prg_mac_addr:
 *
 * Program the requested MAC address (primary/iScsi/rdma) on the requested
 * device using "nvm cfg" command options.
 *
 * Inputs:
 * interp - Tcl interperter
 * mac_addr_str		- pointer to the input MAC address string
 * mac_addr_type	- ADDR_TYPE_PRIMARY or ADDR_TYPE_ISCSI or ADDR_TYPE_RDMA
 * dev_idx			- device number
 *------------------------------------------------------------------------
 */
u32_t prg_mac_addr(Tcl_Interp *interp, char *mac_addr_str, u8_t mac_addr_type, u8_t dev_idx)
{
	char buffer[80];
	int tcl_eval_return;


	buffer[0]='\0';

    /* Execute the command "device $id". */
	execute_device(interp, dev_idx);

	/* Use the relevant "nvm cfg" command, according to the MAC address type */
	if (mac_addr_type == ADDR_TYPE_PRIMARY) {
		  sprintf(&buffer[0], "nvm cfg %d=%s", MAC_OPTION, mac_addr_str);
	} else if (mac_addr_type == ADDR_TYPE_ISCSI) {
		  sprintf(&buffer[0], "nvm cfg %d=%s", ISCSI_MAC_OPTION, mac_addr_str);
	} else {
		  sprintf(&buffer[0], "nvm cfg %d=%s", RDMA_MAC_OPTION, mac_addr_str);
	}

	Tcl_ResetResult( interp );
	tcl_eval_return = Tcl_Eval(interp, &buffer[0]);

	if (tcl_eval_return != TCL_OK)
	{
	  if (mac_addr_type == ADDR_TYPE_PRIMARY)
	  {
		 sprintf( &buffer[0],
			"\nWarning: experience the following error when programming Mac address:%s \n", mac_addr_str);
	  }
	  else if (mac_addr_type == ADDR_TYPE_ISCSI)
	  {
		 sprintf( &buffer[0],
			"\nWarning: experience the following error when programming ISCSI Mac address:%s \n", mac_addr_str);
	  }
	  else
	  {
		 sprintf( &buffer[0],
			"\nWarning: experience the following error when programming RDMA Mac address:%s \n", mac_addr_str);
	  }
	  printf( &buffer[0]);
	  return TCL_ERROR;
	}

	return TCL_OK;
}


/*
 *------------------------------------------------------------------------
 * nvm_upgrade:
 * upgrade nvm bootcode file
 *
 * Inputs:
 * interp	- Tcl interperter
 * bc_type  - BC or BC1 or BC2
 * bc_name	- boot code file name
 * force_upgrade - Should we force upgrade 0 or 1 (even if version is older than exists)
 * ibc		- For isci code, should we burn configuration block 0
 * ibc2		- For isci code, should we burn configuration block 1
 * ibp		- For isci code, should we burn configuration utility software
 *------------------------------------------------------------------------
 */
u32_t nvm_upgrade(Tcl_Interp *interp, u8_t image_type, char *image_name, u8_t force_upgrade, u8_t ibc, u8_t ibc2, u8_t ibp)
{
	char   buffer[80];
	u32_t  s_len=0;


	if (force_upgrade) {
		s_len += sprintf(&buffer[s_len], "nvm upgrade -F");
	} else {
		s_len += sprintf(&buffer[s_len], "nvm upgrade");
	}

	switch (image_type) {
		case BC_IMAGE:
			s_len += sprintf(&buffer[s_len], " -bc %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! Boot code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case BC1_IMAGE:
			s_len += sprintf(&buffer[s_len], " -bc1 %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! Boot code 1 %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case BC2_IMAGE:
			s_len += sprintf(&buffer[s_len], " -bc2 %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! Boot code 2 %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case IPMI_IMAGE:
			s_len += sprintf(&buffer[s_len], " -ipmi %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! IPMI code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case UMP_IMAGE:
			s_len += sprintf(&buffer[s_len], " -ump %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! UMP code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case MBA_IMAGE:
			s_len += sprintf(&buffer[s_len], " -mba %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! MBA code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case MBI_IMAGE:
			s_len += sprintf(&buffer[s_len], " -mbi %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! MBI code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
        case IB_IPV6_IMAGE:
        case IB_IPV4N6_IMAGE:
		case IB_IMAGE:
			if (ibc) {
				s_len += sprintf(&buffer[s_len], " -ibc %s", image_name);
			} else if (ibc2) {
                s_len += sprintf(&buffer[s_len], " -ibc2 %s", image_name);
			} else if (ibp) {
                s_len += sprintf(&buffer[s_len], " -ibpe %s", image_name);
			} else {
                switch (image_type) {
                    case IB_IPV6_IMAGE:
                        s_len += sprintf(&buffer[s_len], " -ib_ipv6 %s", image_name);
                        break;
                    case IB_IPV4N6_IMAGE:
                        s_len += sprintf(&buffer[s_len], " -ib_ipv4n6 %s", image_name);
                        break;
                    case IB_IMAGE:
                        s_len += sprintf(&buffer[s_len], " -ib %s", image_name);
                }
            }
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! IB code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case L2B_IMAGE:
			s_len += sprintf(&buffer[s_len], " -l2 %s", image_name);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! L2 code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
		case NCSI_IMAGE:
			s_len += sprintf(&buffer[s_len], " -ncsi %s", image_name);
			printf("command: %s\n", buffer);
            if (Tcl_Eval(interp,  buffer) != TCL_OK) {
                sprintf(&buffer[0], "Warning! NCSI code %s not programmed!\n", image_name);
			    printf(&buffer[0]);
			    return TCL_ERROR;
	        }
			break;
	}

	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * nvm_cfg:
 * Use nvm cfg to change desired option (with option_num) to desired option_value
 *
 * Inputs:
 * interp	- Tcl interperter
 * option_num  - option number of the option to configure
 * option_value	- option value to configure
 *------------------------------------------------------------------------
 */
u32_t nvm_cfg(Tcl_Interp *interp, u8_t option_num, u8_t option_value)
{
	char   buffer[80];

	switch (option_num) {
		case WOL_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to configure WOL!\n");
				return TCL_ERROR;
			}
			break;
		case MBA_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to configure MBA(PXE)!\n");
				return TCL_ERROR;
			}
			break;
		case MFW_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to configure management firmware!\n");
				return TCL_ERROR;
			}
			break;
		case MBA_PROTOCOL_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to configure MBA protocol!\n");
				return TCL_ERROR;
			}
			break;
		case MBA_VLAN_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning!  Failed to configure MBA VLAN!\n");
				return TCL_ERROR;
			}
			break;
		case MBA_VLAN_VAL_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to configure MBA VLAN value!\n");
				return TCL_ERROR;
			}
			break;
		case PN_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=", option_num);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to get PN value!\n");
				return TCL_ERROR;
			}
			break;
		case MBA_BOOT_RETRY_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {
				printf("Warning! Failed to configure MBA BOOT RETRY value!\n");
				return TCL_ERROR;
			}
			break;

		case IFFE_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);  
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {		    
				print_msg("Warning! Failed to configure IFFE!\n", ESC_COLOR_RED);
				return TCL_ERROR;
			}
			break;

		case REV_ID_CTRL_OPTION:
			sprintf(&buffer[0], "nvm cfg %d=%1d", option_num, option_value);  
			if (Tcl_Eval(interp,  buffer) != TCL_OK) {		    
				print_msg("Warning! Failed to configure rev ID ctrl!\n", ESC_COLOR_RED);
				return TCL_ERROR;
			}
			break;
	}

    return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * nvm_cfg:
 * Use nvm cfg = to change desired option current value
 *
 * Inputs:
 * interp	- Tcl interperter
 * option_num  - option number of the option to configure
 *------------------------------------------------------------------------
 */
int nvm_cfg_get(Tcl_Interp *interp, u8_t option_num)
{
    char   buffer[80] = "";

    sprintf(&buffer[0], "nvmcfgT::nvm_cfg_no_print %d=", option_num);

    return Tcl_Eval(interp, buffer);
}

/*
 *------------------------------------------------------------------------
 * nvm_cfg_serial_no:
 * Configure serial number
 *
 * Inputs:
 * interp	- Tcl interperter
 * serial_no_p  - pointer to serial number string
 *------------------------------------------------------------------------
 */
u32_t nvm_cfg_serial_no(Tcl_Interp *interp, char *serial_no_p)
{
	char   buffer[80];

	sprintf(&buffer[0], "nvm cfg %d=%s", SERIAL_NUMBER_OPTION, serial_no_p);
	if (Tcl_Eval(interp,  buffer) != TCL_OK) {
			printf("Warning! Failed to configure serial number value!\n");
			return TCL_ERROR;
	}

	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * nvm_prg:
 * Program the NVM
 *
 * Inputs:
 * interp	- Tcl interperter
 * file_name_p  - pointer to file name
 *------------------------------------------------------------------------
 */
u32_t nvm_prg(Tcl_Interp *interp, char *file_name_p, u8_t match_dev_id, u8_t use_raw_img, u8_t preserve_ec, u8_t preserve_sn)
{
	char   buffer[80];
	char   tmp_str[80];
    FILE *fd;

    fd = fopen(file_name_p, "r");
    if (fd == NULL)
    {
        printf("Warning! File does not exist!\n");
    }
    fclose(fd);

	strcpy(&tmp_str[0], "nvm prg ");
	if ( match_dev_id ) {
		// match the id first before program the raw image to NVRAM
        strcat(&tmp_str[0], "-idmatch ");
	}
	if ( preserve_ec ) {
		// match the id first before program the raw image to NVRAM
        strcat(&tmp_str[0], "-skip_ec ");
	}
	if ( preserve_sn ) {
		// match the id first before program the raw image to NVRAM
        strcat(&tmp_str[0], "-skip_sn ");
	}

	if ( use_raw_img ) {
		// match the id first before program the raw image to NVRAM
        strcat(&tmp_str[0], "-raw ");
	}

    sprintf(&buffer[0], "%s 0 %s", &tmp_str[0], file_name_p);

	if (Tcl_Eval(interp,  buffer) != TCL_OK) {
			printf("Warning! Failed to program the NVM!\n");
			return TCL_ERROR;
	}

	return TCL_OK;
}

/*
 *------------------------------------------------------------------------
 * nvm_crc:
 * Chcek the NVM
 *
 * Inputs:
 * interp	- Tcl interperter
 * block_name_p  - pointer to block name to check
 *------------------------------------------------------------------------
 */
u32_t nvm_crc(Tcl_Interp *interp, char *block_name_p)
{
	char   buffer[80];

	// Run "nvm crc"
	sprintf(&buffer[0], "nvm crc %s", block_name_p);
	// Verify command has been performed successfully
	if (Tcl_Eval(interp,  buffer) != TCL_OK) {
			/* Normal: that block may not exist in this image. */
	}
	else if (strstr(Tcl_GetStringResult(interp), "OK"))
    {
        /* Got OK, do nothing */
    }
    else
    {
        sprintf(&buffer[0], "Warning! NVRAM CRC check fails on block %s!\n", block_name_p);
        printf(&buffer[0]);
        return TCL_ERROR;
	}
	return TCL_OK;
}
/*
 *------------------------------------------------------------------------
 * execute_device:
 *
 * Execute the "device" command
 *
 * Inputs:
 * interp  - Tcl interperter
 * dev_idx - device number
 *------------------------------------------------------------------------
 */
u32_t execute_device(Tcl_Interp *interp, u8_t dev_idx)
{
	char buffer[256];
	int tcl_eval_return;
	Tcl_Obj *result_ptr = NULL;
	char *tmp_p;


	/* Execute the command "device $id". */
	sprintf(&buffer[0], "device %d", dev_idx);
	Tcl_ResetResult( interp );
	tcl_eval_return = Tcl_Eval(interp, &buffer[0]);
	result_ptr = Tcl_GetObjResult(interp);

	/* Verify that "device" command has completed successfully, otherwise print an error message */
	if (tcl_eval_return != TCL_OK)
	{
	  tmp_p = Tcl_GetString( result_ptr );
	  sprintf( &buffer[0],
			"\nWarning: experience the following error when executing device command:%s\n", tmp_p);

	  printf( &buffer[0]);
	  return TCL_ERROR;
	} else {
		return TCL_OK;
	}
}

/*
 *------------------------------------------------------------------------
 * show_devices:
 *
 * Show the list of devices
 *
 * Inputs:
 * interp  - Tcl interperter
 * dev_idx - device number
 *------------------------------------------------------------------------
 */
int tcl_al_show_devices(Tcl_Interp *interp)
{
	char buffer[256];
	int tcl_eval_return;

	/* If in NO_PCI mode, return without showing devices */
    if (strcmp(Tcl_GetVar2(interp, TCL_SYS_ENV, ENV_NO_PCI, TCL_GLOBAL_ONLY), "1") == 0) {
        return TCL_OK;
    }

	/* Show the devices. */
	sprintf(&buffer[0], "device");
	Tcl_ResetResult( interp );
	tcl_eval_return = Tcl_Eval(interp, &buffer[0]);

	if (tcl_eval_return != TCL_OK)
	{
		return TCL_ERROR;
	} else {
		return TCL_OK;
	}
}


/*
*------------------------------------------------------------------------
* is_zlr_machine:
*
* Check if machine is ZLR machine, by using "is_zlr" TCL utility
*
* Inputs:
* interp  - Tcl interperter
*
* Output: 1 - if ZLR machine
*------------------------------------------------------------------------
*/
u8_t is_zlr_machine (Tcl_Interp *interp)
{
    Tcl_Eval(interp, "is_zlr");
    return atoi(Tcl_GetString(Tcl_GetObjResult(interp)));
}

/*
*------------------------------------------------------------------------
* is_fpga:
*
* Check if running on FPGA
*
* Inputs:
* interp  - Tcl interperter
*
* Output: 1 - if running on FPGA
*------------------------------------------------------------------------
*/
u8_t is_fpga (Tcl_Interp *interp)
{
    return (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_REV, 0), "FPGA", 4) == 0) ? 1:0;

}

/*
*------------------------------------------------------------------------
* is_emul:
*
* Check if running on emulation
*
* Inputs:
* interp  - Tcl interperter
*
* Output: 1 - if running on emulation
*------------------------------------------------------------------------
*/
u8_t is_emul (Tcl_Interp *interp)
{
    return (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_REV, 0), "EMUL", 4) == 0) ? 1:0;
}

/*
*------------------------------------------------------------------------
* split_dev:
*
* Splits device string of format <pf dev>.<vf_dev> into
* pf_dev_num and vf_dev_num
*
* Inputs:
* dev_obj - device string
*
* Output: pf_dev_num and vf_dev_num
*------------------------------------------------------------------------
*/
int split_dev(const char *dev_str, u32_t* pf_dev_num, u32_t* vf_dev_num)
{
    *pf_dev_num = 0;
    *vf_dev_num = 0;
    if (strstr(dev_str, ".")) {
        if (sscanf(dev_str, "%d.%d", pf_dev_num, vf_dev_num) != 2 )
            goto error_exit;

        /* Sanity check */
        if ((! *vf_dev_num) && (! *pf_dev_num))
            goto error_exit;
    } else {
        if (sscanf(dev_str, "%d", pf_dev_num) != 1)
            goto error_exit;

        /* Sanity check */
        if (! *pf_dev_num)
            goto error_exit;
    }

    return TCL_OK;

error_exit:
    printf("Bad format in device argument");
    return TCL_ERROR;
}

/* Covert string to upper case */
int strtoupr(char *str)
{
    while(*str)
    {
        *str = toupper(*str);
        str++;
    }
    return 0;
}


/* Covert string to lower case */
int strtolwr(char *str)
{
    while(*str)
    {
        *str = tolower(*str);
        str++;
    }
    return 0;
}

/*
*------------------------------------------------------------------------
* is_e2: 
* 
* Check if running on everest 2 (CHIP_NUM 57712/E or 57713/E)
* 
* Inputs: 
* interp  - Tcl interperter
*
* Output: 1 - if running on everest 2
*------------------------------------------------------------------------
*/
u8_t is_e2 (Tcl_Interp *interp)
{    
    if ((strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57712", 5) == 0) ||
	    (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57713", 5) == 0)) {
            return 1;
        } else {
            return 0;
        }
}

/*
*------------------------------------------------------------------------
* is_e1x: 
* 
* Check if running on E1 / E1H (CHIP_NUM 57710 or 57711 or 57711E)
* 
* Inputs: 
* interp  - Tcl interperter
*
* Output: 1 - if running on everest 2
*------------------------------------------------------------------------
*/
u8_t is_e1x (Tcl_Interp *interp)
{    
    if ((strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57710", 5) == 0) ||
	    (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57711", 5) == 0)) {
            return 1;
        } else {
            return 0;
        }
}

/*
*------------------------------------------------------------------------
* is_e1: 
* 
* Check if running on E1 (CHIP_NUM 57710)
*------------------------------------------------------------------------
*/
u8_t is_e1 (Tcl_Interp *interp)
{    
    return (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57710", 5) == 0) ? 1:0;       
}

/*
*------------------------------------------------------------------------
* is_e1h: 
* 
* Check if running on E1H (57711 or 57711E)
*------------------------------------------------------------------------
*/
u8_t is_e1h (Tcl_Interp *interp)
{    
    return (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57711", 5) == 0) ? 1:0;
}

/*
*------------------------------------------------------------------------
* is_e3: 
* 
* Check if running on everest 3
* 
* Inputs: 
* interp  - Tcl interperter
*
* Output: 1 - if running on everest 3
*------------------------------------------------------------------------
*/
u8_t is_e3 (Tcl_Interp *interp)
{    
    if ((strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57810", 5) == 0) ||
	    (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57840", 5) == 0) ||
        (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57800", 5) == 0) ||
        (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57811", 5) == 0) ||
	(strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "578XX", 5) == 0)) {
            return 1;
        } else {
            return 0;
        }
}

/*
*------------------------------------------------------------------------
* is_rev_b0:
*
* Check if running on revision B
*
* Inputs:
* interp  - Tcl interperter
*
* Output: 1 - if running on FPGA
*------------------------------------------------------------------------
*/
u8_t is_rev_b (Tcl_Interp *interp)
{
    return (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_REV, 0), "B", 1) == 0) ? 1:0;

}

/*
*------------------------------------------------------------------------
* is_e3_57840: 
* 
* Check if running on everest 3 (CHIP_NUM 57840)
*------------------------------------------------------------------------
*/
u8_t is_e3_57840 (Tcl_Interp *interp)
{    
    return (strncmp (Tcl_GetVar2(interp, TCL_ENV, ENV_CHIP_NUM, 0), "57840", 5) == 0) ? 1:0;        
}


/* remove_white_spaces
 * inputs: 
 * token_ptr - points to the begining of the input field terminated by the token (',')
 *             may include white spaces           
 * field_ptr - pointer to the output field, result string w/o spaces
 *           
*/

void remove_white_spaces (char *token_ptr, u32_t field_len, char *field_ptr) 
{
	char *cptr;
	u32_t i=0;

	cptr = field_ptr;

	// Ignore white spaces at the beginning
	if (*token_ptr == ' ') {
		for ( ; (i<field_len) && (*token_ptr == ' '); token_ptr++) {
			i++;
		}
	}

	// remove count of white spaces at the begining from field_len
	field_len -= i;

	// Ignore white spaces at the end
	if (field_len > 0) {
		i = field_len - 1;
	}
	while ( (i > 0) && (*(token_ptr+i) == ' ')) {
		i--;
	}
	*(cptr+i+1) = '\0';

	// Copy the string from end to begining
	while (i > 0) {
		*(cptr+i) = *(token_ptr+i);
		i--;
	}

	// first char (i=0)
	*cptr = *token_ptr;
}

