
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/endian.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <fcntl.h>
#include "os_types.h"
#include "edebug_types.h"
#include "dev_if.h"
#include "edebug_if.h"
#include "57712_reg.h"

#define DRVSIM_NVRAM_1MB_SIZE   0x20000 /* 1M bit in bytes */

int dev_mem_fd = -1;
static void **mapped_area_arr;
static int perform_cleanup = 0;

/* read 32 bits value from given grc addr */
dbg_status_t freebsd_reg_read(struct os_params os_params,
                              u32_t            addr,
                              u32_t *          value)
{
	*value = htole32(*((volatile u32_t *)((mapped_area_arr[os_params.br_idx] + addr))));
    return DBG_STATUS_SUCCESS;
}

/* write 32 bits value to given grc addr */
dbg_status_t freebsd_reg_write(struct os_params os_params,
                               u32_t            addr,
                               u32_t            value)
{
	*((volatile u32_t *)(mapped_area_arr[os_params.br_idx] + addr)) = htole32(value);
    return DBG_STATUS_SUCCESS;
}

/* read from nvram */
dbg_status_t freebsd_nv_read(struct os_params os_params,
                             u32_t            addr,
                             u32_t *          value,
                             u32_t            size)
{
    struct ifreq ifr;
    struct bxe_nvram_data *data;
    int alloc_len;

    alloc_len = (sizeof(struct bxe_nvram_data) + size - sizeof(u32_t));

    if ((data = (struct bxe_nvram_data *)calloc(1, alloc_len)) == NULL)
    {
        return DBG_STATUS_FAILURE;
    }

    data->op     = BXE_IOC_RD_NVRAM;
    data->offset = addr;
    data->len    = size;

    memset(&ifr, 0, sizeof(ifr));
    strcpy(ifr.ifr_name, os_params.eth_name);
    ifr.ifr_data = (caddr_t)data;

    if (ioctl(os_params.ioctl_fd, SIOCGPRIVATE_0, &ifr) < 0) {
        printf("ERROR: BXE_IOC_RD_NVRAM ioctl failed\n");
        free(data);
        return DBG_STATUS_FAILURE;
    }

    memcpy(value, data->value, size);

    free(data);

    return DBG_STATUS_SUCCESS;
}

/* write to nvram */
dbg_status_t freebsd_nv_write(struct os_params os_params,
                              u32_t            addr,
                              u32_t *          value,
                              u32_t            size)
{
    struct ifreq ifr;
    struct bxe_nvram_data *data;
    int alloc_len;
    u32_t *tmp;
    int i;

    alloc_len = (sizeof(struct bxe_nvram_data) + size - sizeof(u32_t));

    if ((data = (struct bxe_nvram_data *)calloc(1, alloc_len)) == NULL)
    {
        return DBG_STATUS_FAILURE;
    }

    data->op     = BXE_IOC_WR_NVRAM;
    data->offset = addr;
    data->len    = size;
    memcpy(data->value, value, size);

    /*
     * byte swapping should be performed because bxe expects to receive the
     * bytes in byte swapped order
     */
    tmp = (u32_t *)data->value;
    for (i = 0; i < (size / 4); i++) {
        tmp[i] = ntohl(tmp[i]);
    }

    memset(&ifr, 0, sizeof(ifr));
    strcpy(ifr.ifr_name, os_params.eth_name);
    ifr.ifr_data = (caddr_t)data;

    if (ioctl(os_params.ioctl_fd, SIOCGPRIVATE_0, &ifr) < 0) {
        printf("ERROR: BXE_IOC_WR_NVRAM ioctl failed\n");
        free(data);
        return DBG_STATUS_FAILURE;
    }

    free(data);

    return DBG_STATUS_SUCCESS;
}

/* get NIC statistics */
dbg_status_t freebsd_stats_show(struct os_params os_params)
{
    char *strings;
    u64_t *counters;
    u32_t size_str, size_cnt;
    union bxe_stats_show_data descriptor;
    struct ifreq ifr;
    int i;

    memset(&descriptor, 0, sizeof(descriptor));
    descriptor.op = BXE_IOC_STATS_SHOW_NUM;

    memset(&ifr, 0, sizeof(ifr));
    strcpy(ifr.ifr_name, os_params.eth_name);
    ifr.ifr_data = (caddr_t)&descriptor;

    if (ioctl(os_params.ioctl_fd, SIOCGPRIVATE_0, &ifr) < 0) {
        printf("ERROR: BXE_IOC_STATS_SHOW_NUM ioctl failed\n");
        return DBG_STATUS_FAILURE;
    }

    if (descriptor.desc.num < 1) {
        printf("ERROR: Cannot get NIC statistics (num < 1)\n");
        return DBG_STATUS_FAILURE;
    }

    /* allocate memory for the statistics strings and counters */

    size_str = (descriptor.desc.num * descriptor.desc.len);
    size_cnt = (descriptor.desc.num * sizeof(u64_t));

    strings = (char *)calloc(1, size_str);
    if (strings == NULL) {
        printf("ERROR: Cannot get NIC statistics (malloc failed)\n");
        return DBG_STATUS_FAILURE;
    }

    counters = (u64_t *)calloc(1, size_cnt);
    if (counters == NULL) {
        printf("ERROR: Cannot get NIC statistics (malloc failed)\n");
        free(strings);
        return DBG_STATUS_FAILURE;
    }

    /* prepare and send second ioctl to get the statistics strings */
    *((u32_t *)strings) = BXE_IOC_STATS_SHOW_STR;

    memset(&ifr, 0, sizeof(ifr));
    strcpy(ifr.ifr_name, os_params.eth_name);
    ifr.ifr_data = (caddr_t)strings;

    if (ioctl(os_params.ioctl_fd, SIOCGPRIVATE_0, &ifr) < 0) {
        printf("ERROR: BXE_IOC_STATS_SHOW_STR ioctl failed\n");
        free(strings);
        free(counters);
        return DBG_STATUS_FAILURE;
    }

    /* prepare and send third ioctl to get the statistics counters */
    *((u32_t *)counters) = BXE_IOC_STATS_SHOW_CNT;

    memset(&ifr, 0, sizeof(ifr));
    strcpy(ifr.ifr_name, os_params.eth_name);
    ifr.ifr_data = (caddr_t)counters;

    if (ioctl(os_params.ioctl_fd, SIOCGPRIVATE_0, &ifr) < 0) {
        printf("ERROR: BXE_IOC_STATS_SHOW_CNT ioctl failed\n");
        free(strings);
        free(counters);
        return DBG_STATUS_FAILURE;
    }

    /* pretty-print the statistics per-driver */
    printf("NIC statistics:\n");
    for (i = 0; i < descriptor.desc.num; i++) {
        printf("     %.*s: %llu\n",
               descriptor.desc.len,
               &strings[i*descriptor.desc.len],
               counters[i]);
    }

    free(strings);
    free(counters);

    return DBG_STATUS_SUCCESS;
}

dbg_status_t freebsd_cleanup(struct _dev_if_t * dev_if)
{
    if (perform_cleanup) {
        free(mapped_area_arr);
        close(dev_mem_fd);
        perform_cleanup = 0;
    }

    close(dev_if->os_params.ioctl_fd);
    return DBG_STATUS_SUCCESS;
}

/* initialize Everest's dispatch table  */
dbg_status_t everest_dev_if_init_dispatch(int        br_idx,
                                          pci_info_t pci_info,
                                          dev_if_t * dev_if)
{
    u32_t val;
    int fd;

    dev_if->os_params.reg_bar_address = pci_info.region0_base_addr;
    dev_if->os_params.reg_bar_size    = pci_info.region0_size;
    dev_if->os_params.br_idx          = br_idx;
    strcpy(dev_if->os_params.eth_name, pci_info.eth_name);

    if ((dev_if->os_params.ioctl_fd = socket(PF_INET, SOCK_DGRAM, 0)) <= 0)
    {
        printf("ERROR - failed to open socket\n");
        perror("socket");
        return DBG_STATUS_FAILURE;
    }

    if (dev_mem_fd == -1) {
        mapped_area_arr = (void *)malloc(sizeof(void *) *
                                         dbg_get_total_br_dev());

        dev_mem_fd = open("/dev/mem", O_RDWR);
        if (dev_mem_fd == -1) {
            printf("ERROR: Failed to open /dev/mem\n");
            close(dev_if->os_params.ioctl_fd);
            return DBG_STATUS_FAILURE;
        }

        perform_cleanup = 1;
    }

    mapped_area_arr[dev_if->os_params.br_idx] =
        mmap(NULL,
             dev_if->os_params.reg_bar_size,
             (PROT_READ | PROT_WRITE),
             MAP_SHARED,
             dev_mem_fd,
             (off_t)dev_if->os_params.reg_bar_address);
    if (mapped_area_arr[dev_if->os_params.br_idx] == MAP_FAILED) {
        printf("ERROR: Failed mapping device /dev/mem\n");
        free(mapped_area_arr);
        close(dev_mem_fd);
        close(dev_if->os_params.ioctl_fd);
        return DBG_STATUS_FAILURE;
    }

    dev_if->dev_if_reg_read     = freebsd_reg_read;
    dev_if->dev_if_reg_write    = freebsd_reg_write;
    dev_if->dev_if_get_dev_info = NULL;
    dev_if->dev_if_cleanup      = freebsd_cleanup;
    dev_if->dev_if_nv_read      = freebsd_nv_read;
    dev_if->dev_if_nv_write     = freebsd_nv_write;
    dev_if->dev_if_stats_show   = freebsd_stats_show;

    /* Read NVM SIZE */
    freebsd_reg_read(dev_if->os_params, MCP_REG_MCPR_NVM_CFG4, &val);
    dev_if->os_params.geeprom_length =
        (DRVSIM_NVRAM_1MB_SIZE << (val & MCPR_NVM_CFG4_FLASH_SIZE));

    return DBG_STATUS_SUCCESS;
}

