#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <net/if.h>
#include <inttypes.h>
#include "edebug_types.h"
#include "dbg_os_utils.h"
#include "ethtool-copy.h"
#include "/usr/include/linux/sockios.h"
#include "util.h"
#include "os_types.h"
#include "dev_if.h"
#include "edebug_if.h"
#include "57712_reg.h"

#define DRVSIM_NVRAM_1MB_SIZE	0x20000	/* 1M bit in bytes */
#define NVRAM_PAGE_SIZE     	256

int dev_mem_fd = -1;
static void **mapped_area_arr;
static int perform_cleanup = 0;

static dbg_status_t linux_map_bar(struct os_params os_params)
{
	if (-1 == dev_mem_fd) {
		mapped_area_arr = (void*)malloc(sizeof(void*) * dbg_get_total_br_dev());
		dev_mem_fd = open("/dev/mem", O_RDWR);
		if (dev_mem_fd == -1) {
			printf("failed to open /dev/mem\n");
			return DBG_STATUS_FAILURE;
		}
		perform_cleanup = 1;
	}

	mapped_area_arr[os_params.br_idx] = mmap(NULL, 
				       os_params.reg_bar_size,
				       PROT_READ | PROT_WRITE,
				       MAP_SHARED,
				       dev_mem_fd,
				       (off_t) os_params.reg_bar_address);
	if (mapped_area_arr[os_params.br_idx] == MAP_FAILED) {
		printf("failed mapping device /dev/mem\n");
		return DBG_STATUS_FAILURE;
	}
}

static dbg_status_t linux_open_ioctl_socket(struct os_params *p_os_params)
{
        int fd;
        struct ifreq ifr;
        int err;
	struct ethtool_drvinfo drvinfo;

        /* Setup our control structures. */
	memset(&ifr, 0, sizeof(ifr));
	strcpy(ifr.ifr_name, p_os_params->eth_name);

        /* Open control socket. */
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (fd < 0) {
		printf("Cannot get control socket");
		return DBG_STATUS_FAILURE;
	}      

	drvinfo.cmd = ETHTOOL_GDRVINFO;
	ifr.ifr_data = (caddr_t)&drvinfo;
	err = ioctl(fd, SIOCETHTOOL, &ifr);
	if (err < 0) {
		printf("Cannot get driver information");
		return DBG_STATUS_FAILURE;
	}

	if (p_os_params->geeprom_length <= 0)
		p_os_params->geeprom_length = drvinfo.eedump_len;

	if (drvinfo.eedump_len < p_os_params->geeprom_length)
		p_os_params->geeprom_length = drvinfo.eedump_len;       

	p_os_params->fd_ethtool = fd;

	return DBG_STATUS_SUCCESS;
}

/* read 32 bits value from given grc addr */
dbg_status_t linux_reg_read(
    IN  struct os_params   	os_params,
    IN  u32_t       	  	addr,
    OUT u32_t*      		value
    )
{
	*value = os_if_cpu_to_le32(*((volatile u32_t *)((mapped_area_arr[os_params.br_idx] + addr)))); 
	return DBG_STATUS_SUCCESS;
}

dbg_status_t linux_reg_write(
    IN  struct os_params   	os_params,
    IN  u32_t       		addr,
    IN  u32_t       		value
    )
{
	*((volatile u32_t *)(mapped_area_arr[os_params.br_idx] + addr)) = os_if_cpu_to_le32(value);
	return DBG_STATUS_SUCCESS;
}

dbg_status_t linux_nv_read(
    IN struct os_params 	os_params,
    IN  u32_t			addr,
    OUT u32_t*			value,
    IN  u32_t       		size
    )
{
	struct ifreq ifr;
	struct ethtool_eeprom *eeprom;
        int err;

	 /* Setup our control structures. */
	memset(&ifr, 0, sizeof(ifr));
	strcpy(ifr.ifr_name, os_params.eth_name);

	eeprom = calloc(1, sizeof(*eeprom) + size);
	if (!eeprom) {
		printf("Cannot allocate memory for EEPROM data");
		return DBG_STATUS_FAILURE;
	}

	eeprom->cmd = ETHTOOL_GEEPROM;
	eeprom->len = size;
	eeprom->offset = addr;
	ifr.ifr_data = (caddr_t)(eeprom);

	err = ioctl(os_params.fd_ethtool, SIOCETHTOOL, &ifr);
        if (err < 0) {
                printf("Cannot get EEPROM data\n");
                return DBG_STATUS_FAILURE;
        }

	memcpy(value, eeprom->data, size);

	free(eeprom);

	return DBG_STATUS_SUCCESS;
}

/* write 32 bits value to given grc addr */
dbg_status_t linux_nv_write(
    IN struct os_params     	os_params,
    IN  u32_t			addr,
    IN  u32_t*			value,
    IN  u32_t       	size
    )
{
	struct ifreq ifr;
	struct ethtool_eeprom *eeprom;
        int err, i;
	u32_t *tmpArray;

	if (!size) {
		return DBG_STATUS_SUCCESS;
	}

	 /* Setup our control structures. */
	memset(&ifr, 0, sizeof(ifr));
	strcpy(ifr.ifr_name, os_params.eth_name);

	eeprom = calloc(1, sizeof(*eeprom)+size);
	if (!eeprom) {
		printf("Cannot allocate memory for EEPROM data");
		return DBG_STATUS_FAILURE;
	}

	eeprom->cmd = ETHTOOL_SEEPROM;
	eeprom->len = size;
	eeprom->offset = addr;
	eeprom->magic = 0;
	memcpy(eeprom->data, value, size);

	/* byte swapping should be performed because bnx2x expects to receive the bytes in byte swapped order */
	tmpArray = (u32_t *) (eeprom->data);	
	for (i = 0; i < (size/4); i++) {		
		tmpArray[i]=ntohl(tmpArray[i]);
		
	}	

	ifr.ifr_data = (caddr_t)eeprom;
	
	err = ioctl(os_params.fd_ethtool, SIOCETHTOOL, &ifr);
	if (err < 0) {
		printf("Cannot set EEPROM data");
		return DBG_STATUS_FAILURE;
	}

	free(eeprom);	

	return DBG_STATUS_SUCCESS;
}

/* write 32 bits value to given grc addr */
dbg_status_t linux_cleanup(
    IN  struct _dev_if_t   	*dev_if
    ) 
{
	if (perform_cleanup) {
			close(dev_mem_fd);		
			free(mapped_area_arr);	
			perform_cleanup = 0;	
	}
	
	return DBG_STATUS_SUCCESS;
}

static dbg_status_t drvsim_nvram_read_dword(struct os_params os_params, u32 offset, u32 *ret_val, u32 is_first, u32 is_last)
{
	int i;
	u32 cmd_flags,val;

	/* build the command word */
	cmd_flags = MCPR_NVM_COMMAND_DOIT;
    if (is_first) cmd_flags |= MCPR_NVM_COMMAND_FIRST;
    if (is_last) cmd_flags |= MCPR_NVM_COMMAND_LAST;

	/* need to clear DONE bit separately */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_COMMAND, MCPR_NVM_COMMAND_DONE);

	/* address of the NVRAM to read from */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_ADDR, (offset & MCPR_NVM_ADDR_NVM_ADDR_VALUE));

	/* issue a read command */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_COMMAND, cmd_flags);

	/* wait for completion */
	*ret_val = 0;
	for (i = 0; i < 1000; i++) {
		dbg_os_usleep(5);
        linux_reg_read(os_params, MCP_REG_MCPR_NVM_COMMAND, &val);

		if (val & MCPR_NVM_COMMAND_DONE) {
            linux_reg_read(os_params, MCP_REG_MCPR_NVM_READ, &val);
            /* Big to little endian */
			*ret_val = dbg_os_be32_to_cpu(val);
			return 0;
		}
	}
	return DBG_STATUS_TIMEOUT;
}

static dbg_status_t drvsim_nvram_write_dword(struct os_params os_params, u32 offset, u32 val,
				   u32 is_first, u32 is_last)
{
	int i, rc, cmd_flags;

	/* build the command word */
	cmd_flags = MCPR_NVM_COMMAND_DOIT | MCPR_NVM_COMMAND_WR;
    if (is_first) cmd_flags |= MCPR_NVM_COMMAND_FIRST;
    if (is_last) cmd_flags |= MCPR_NVM_COMMAND_LAST;

	/* need to clear DONE bit separately */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_COMMAND, MCPR_NVM_COMMAND_DONE);

	/* write the data */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_WRITE, dbg_os_cpu_to_be32(val));

	/* address of the NVRAM to write to */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_ADDR,
	       (offset & MCPR_NVM_ADDR_NVM_ADDR_VALUE));

	/* issue the write command */
	linux_reg_write(os_params, MCP_REG_MCPR_NVM_COMMAND, cmd_flags);

	/* wait for completion */
	for (i = 0; i < 1000; i++) {
		dbg_os_usleep(5);
		linux_reg_read(os_params, MCP_REG_MCPR_NVM_COMMAND, &val);
		if (val & MCPR_NVM_COMMAND_DONE) return DBG_STATUS_SUCCESS;
	}

	return DBG_STATUS_TIMEOUT;
}

dbg_status_t drvsim_nv_read(
    IN struct os_params 	os_params,
    IN  u32_t			addr,
    OUT u32_t*			value,
    IN  u32_t       	size
    )
{
    u32 offset;
    int rc;

    if ((addr & 0x3) || (size & 3) || 
        (addr + size) > os_params.geeprom_length)
    {
           return DBG_STATUS_INVALID_PARAMETER;
    }

    // Read integral number of dwords
    for (offset = addr; offset < (addr + size); offset += 4) {
        rc = drvsim_nvram_read_dword(os_params, offset, value++, (offset == addr), (offset == (addr + size - 4)));
        if (rc != DBG_STATUS_SUCCESS) return rc;
    }
   
	return DBG_STATUS_SUCCESS;
}

/* write 32 bits value to given grc addr */
dbg_status_t drvsim_nv_write(
    IN struct os_params     	os_params,
    IN  u32_t			addr,
    IN  u32_t*			value,
    IN  u32_t       	size
    )
{
    dbg_status_t rc;
    int offset;

    if ((addr & 0x3) || (size & 3) || 
        (addr + size) > os_params.geeprom_length)
    {
           return DBG_STATUS_INVALID_PARAMETER;
    }

    // Write integral number of dwords
    for (offset = addr; offset < (addr + size); offset += 4) {
        u32 is_first = ((offset == addr) || (offset % NVRAM_PAGE_SIZE == 0));
        u32 is_last = (offset == (addr + size - 4) || offset % NVRAM_PAGE_SIZE == (NVRAM_PAGE_SIZE-4));
        rc = drvsim_nvram_write_dword(os_params, offset, *value++, is_first, is_last);
        if (rc != DBG_STATUS_SUCCESS) return rc;
    }

    return DBG_STATUS_SUCCESS;
}

/* get NIC statistics via ethtool */
dbg_status_t linux_stats_show(
	IN  struct os_params	os_params)
{
#if 1
	
	struct ifreq ifr;
	struct ethtool_drvinfo drvinfo;
	struct ethtool_gstrings *strings;
	struct ethtool_stats *stats;
	unsigned int n_stats, sz_str, sz_stats, i;
	int fd, err;



	/* setup ifr control structure */
	memset(&ifr, 0, sizeof(ifr));
	strcpy(ifr.ifr_name, os_params.eth_name);
	
	/* prepare and send first ioctl to get the number of statistics */
	drvinfo.cmd = ETHTOOL_GDRVINFO;
	ifr.ifr_data = (caddr_t)&drvinfo;

	err = ioctl(os_params.fd_ethtool, SIOCETHTOOL, &ifr);
	if (err < 0) {
		printf("Cannot get NIC statistics (first ioctl failed)\n");
		return DBG_STATUS_FAILURE;
	}
	n_stats = drvinfo.n_stats;
	if (n_stats < 1) {
		printf("Cannot get NIC statistics (n_stats < 1)\n");
		return DBG_STATUS_FAILURE;
	}

	/* allocate memory for the statistics */
	sz_str = n_stats * ETH_GSTRING_LEN;
	sz_stats = n_stats * sizeof(u64);
	strings = calloc(1, sz_str + sizeof(struct ethtool_gstrings));
	stats = calloc(1, sz_stats + sizeof(struct ethtool_stats));
	if (!strings || !stats) {
		printf("Cannot get NIC statistics (memory allocation failed)\n");
		return DBG_STATUS_FAILURE;
	}

	/* prepare and send second ioctl to get the statistics strings */
	strings->cmd = ETHTOOL_GSTRINGS;
	strings->string_set = ETH_SS_STATS;
	strings->len = n_stats;
	ifr.ifr_data = (caddr_t) strings;
	err = ioctl(os_params.fd_ethtool, SIOCETHTOOL, &ifr);
	if (err < 0) {
		printf("Cannot get NIC statistics (strings ioctl failed)\n");
		free(strings);
		free(stats);
		return DBG_STATUS_FAILURE;
	}

	/* prepare and send third ioctl to get the statistics counters */
	stats->cmd = ETHTOOL_GSTATS;
	stats->n_stats = n_stats;
	ifr.ifr_data = (caddr_t) stats;
	err = ioctl(os_params.fd_ethtool, SIOCETHTOOL, &ifr);
	if (err < 0) {
		printf("Cannot get NIC statistics (stats ioctl failed)\n");
		free(strings);
		free(stats);
		return DBG_STATUS_FAILURE;
	}

	/* pretty-print the statistics per-driver */
	printf("NIC statistics:\n");
	for (i = 0; i < n_stats; i++) {
		printf("	 %.*s: %" PRIu64 "\n",
			ETH_GSTRING_LEN,
			&strings->data[i * ETH_GSTRING_LEN],
			stats->data[i]);
	}

	free(strings);
	free(stats);
#endif

	return DBG_STATUS_SUCCESS;
}

/* initialize Everest's dispatch table  */
dbg_status_t everest_dev_if_init_dispatch(
    IN  int		br_idx,
    IN  pci_info_t      pci_info,
    OUT dev_if_t*       dif         /* dispatch table to be initialized */
    )
{
	/* init OS dependent parameters from pci_info */
	dif->os_params.reg_bar_address = pci_info.region0_base_addr;
	dif->os_params.reg_bar_size = pci_info.region0_size;
	dif->os_params.br_idx = br_idx;
	dif->os_params.geeprom_length = -1;
	sprintf(dif->os_params.eth_name, "%s", pci_info.eth_name);

	/* perform one time initializations */
	linux_map_bar(dif->os_params);

	/* Dispatch table */ 
	dif->dev_if_reg_read = linux_reg_read;
	dif->dev_if_reg_write = linux_reg_write;
	dif->dev_if_cleanup = linux_cleanup;
	dif->dev_if_get_dev_info = NULL;
	dif->dev_if_stats_show = linux_stats_show;  

    if (!strcmp(pci_info.eth_name, "drvsim"))
    {
        u32_t val;

        dif->dev_if_nv_read = drvsim_nv_read;
        dif->dev_if_nv_write = drvsim_nv_write;

        /* read nmram size */
        linux_reg_read(dif->os_params, MCP_REG_MCPR_NVM_CFG4, &val);
        dif->os_params.geeprom_length = DRVSIM_NVRAM_1MB_SIZE << (val & MCPR_NVM_CFG4_FLASH_SIZE);
    } else {
        linux_open_ioctl_socket(&(dif->os_params));
        dif->dev_if_nv_read = linux_nv_read;
        dif->dev_if_nv_write = linux_nv_write;
    }
	
	return DBG_STATUS_SUCCESS;
}

