
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stropts.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "os_types.h"
#include "edebug_types.h"
#include "dev_if.h"
#include "edebug_if.h"
#include "57712_reg.h"

#define DRVSIM_NVRAM_1MB_SIZE   0x20000 /* 1M bit in bytes */

/* read 32 bits value from given grc addr */
dbg_status_t solaris_reg_read(struct os_params os_params,
                              u32_t            addr,
                              u32_t *          value)
{
    struct strioctl strioc;
    struct bnxe_reg_data data;

    memset(&data, 0, sizeof(data));
    data.offset = addr;
    data.value  = 0;

    strioc.ic_cmd    = GIOCBNXEREG;
    strioc.ic_timout = 10;
    strioc.ic_len    = sizeof(data);
    strioc.ic_dp     = (char *)&data;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        return DBG_STATUS_FAILURE;
    }

    *value = data.value;
    return DBG_STATUS_SUCCESS;
}

/* write 32 bits value to given grc addr */
dbg_status_t solaris_reg_write(struct os_params os_params,
                               u32_t            addr,
                               u32_t            value)
{
    struct strioctl strioc;
    struct bnxe_reg_data data;

    memset(&data, 0, sizeof(data));
    data.offset = addr;
    data.value  = value;

    strioc.ic_cmd    = SIOCBNXEREG;
    strioc.ic_timout = 10;
    strioc.ic_len    = sizeof(data);
    strioc.ic_dp     = (char *)&data;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        return DBG_STATUS_FAILURE;
    }

    return DBG_STATUS_SUCCESS;
}

/* read 32 bits value from given pci config space addr */
dbg_status_t solaris_pcicfg_read_common(int     fd,
                                        u32_t   addr,
                                        u32_t * value)
{
    struct strioctl strioc;
    struct bnxe_reg_data data;

    memset(&data, 0, sizeof(data));
    data.offset = addr;
    data.value  = 0;

    strioc.ic_cmd    = GIOCBNXEPCI;
    strioc.ic_timout = 10;
    strioc.ic_len    = sizeof(data);
    strioc.ic_dp     = (char *)&data;

    if (ioctl(fd, I_STR, &strioc) < 0)
    {
        return DBG_STATUS_FAILURE;
    }

    *value = data.value;
    return DBG_STATUS_SUCCESS;
}

dbg_status_t solaris_pcicfg_read(struct os_params os_params,
                                 u32_t            addr,
                                 u32_t *          value)
{
    return solaris_pcicfg_read_common(os_params.ioctl_fd, addr, value);
}

dbg_status_t solaris_pcicfg_read_ifname(char *  ifname,
                                        u32_t   addr,
                                        u32_t * value)
{
    char dev_path[128];
    dbg_status_t rc;
    int fd;

    strcpy(dev_path, "/dev/");
    strcat(dev_path, ifname);

    if ((fd = open(dev_path, O_RDONLY)) <= 0)
    {
        printf("ERROR - failed to open device %s (%d)\n", dev_path, errno);
        return DBG_STATUS_FAILURE;
    }

    rc = solaris_pcicfg_read_common(fd, addr, value);

    close(fd);

    return rc;
}

/* read from nvram */
dbg_status_t solaris_nv_read(struct os_params os_params,
                             u32_t            addr,
                             u32_t *          value,
                             u32_t            size)
{
    struct strioctl strioc;
    struct bnxe_nvram_data * data;

    int alloc_len = (sizeof(struct bnxe_nvram_data) + size - sizeof(u32_t));

    if ((data = (struct bnxe_nvram_data *)malloc(alloc_len)) == NULL)
    {
        return DBG_STATUS_FAILURE;
    }

    memset(data, 0, alloc_len);
    data->offset     = addr;
    data->num_of_u32 = (size / sizeof(u32_t));

    strioc.ic_cmd    = GIOCBNXENVRM;
    strioc.ic_timout = 10;
    strioc.ic_len    = alloc_len;
    strioc.ic_dp     = (char *)data;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        free(data);
        return DBG_STATUS_FAILURE;
    }

    memcpy(value, data->value, size);

    free(data);
    return DBG_STATUS_SUCCESS;
}

/* write to nvram */
dbg_status_t solaris_nv_write(struct os_params os_params,
                              u32_t            addr,
                              u32_t *          value,
                              u32_t            size)
{
    struct strioctl strioc;
    struct bnxe_nvram_data * data;

    int alloc_len = (sizeof(struct bnxe_nvram_data) + size - sizeof(u32_t));

    if ((data = (struct bnxe_nvram_data *)malloc(alloc_len)) == NULL)
    {
        return DBG_STATUS_FAILURE;
    }

    memset(data, 0, alloc_len);
    data->offset     = addr;
    data->num_of_u32 = (size / sizeof(u32_t));

    memcpy(data->value, value, size);

    strioc.ic_cmd    = SIOCBNXENVRM;
    strioc.ic_timout = 10;
    strioc.ic_len    = alloc_len;
    strioc.ic_dp     = (char *)data;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        free(data);
        return DBG_STATUS_FAILURE;
    }

    free(data);
    return DBG_STATUS_SUCCESS;
}

/* get NIC statistics */
dbg_status_t solaris_stats_show(struct os_params os_params)
{
    char *strings;
    u64_t *counters;
    u32_t size_str, size_cnt;
    b10_stats_show_data_t descriptor;
    struct strioctl strioc;
    int i;

    memset(&descriptor, 0, sizeof(descriptor));
    descriptor.op = STATS_SHOW_TYPE_NUM;

    memset(&strioc, 0, sizeof(strioc));
    strioc.ic_cmd    = GIOCBNXESTATS;
    strioc.ic_timout = 10;
    strioc.ic_len    = sizeof(descriptor.desc);
    strioc.ic_dp     = (char *)&descriptor.desc;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        printf("ERROR: Failed to get NIC statistics (num+len ioctl failed)\n");
        return DBG_STATUS_FAILURE;
    }

    if (descriptor.desc.num < 1) {
        printf("ERROR: Cannot get NIC statistics (num < 1)\n");
        return DBG_STATUS_FAILURE;
    }

    /* allocate memory for the statistics strings and counters */

    size_str = (descriptor.desc.num * descriptor.desc.len);
    size_cnt = (descriptor.desc.num * sizeof(u64_t));

    strings = (char *)calloc(1, size_str);
    if (strings == NULL) {
        printf("ERROR: Cannot get NIC statistics (malloc failed)\n");
        return DBG_STATUS_FAILURE;
    }

    counters = (u64_t *)calloc(1, size_cnt);
    if (counters == NULL) {
        printf("ERROR: Cannot get NIC statistics (malloc failed)\n");
        free(strings);
        return DBG_STATUS_FAILURE;
    }

    /* prepare and send second ioctl to get the statistics strings */
    *((u32_t *)strings) = STATS_SHOW_TYPE_STR;

    memset(&strioc, 0, sizeof(strioc));
    strioc.ic_cmd    = GIOCBNXESTATS;
    strioc.ic_timout = 10;
    strioc.ic_len    = size_str;
    strioc.ic_dp     = (char *)strings;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        printf("ERROR: Failed to get NIC statistics (strings ioctl failed)\n");
        free(strings);
        free(counters);
        return DBG_STATUS_FAILURE;
    }

    /* prepare and send third ioctl to get the statistics counters */
    *((u32_t *)counters) = STATS_SHOW_TYPE_CNT;

    memset(&strioc, 0, sizeof(strioc));
    strioc.ic_cmd    = GIOCBNXESTATS;
    strioc.ic_timout = 10;
    strioc.ic_len    = size_cnt;
    strioc.ic_dp     = (char *)counters;

    if (ioctl(os_params.ioctl_fd, I_STR, &strioc) < 0)
    {
        printf("ERROR: Failed to get NIC statistics (counters ioctl failed)\n");
        free(strings);
        free(counters);
        return DBG_STATUS_FAILURE;
    }

    /* pretty-print the statistics per-driver */
    printf("NIC statistics:\n");
    for (i = 0; i < descriptor.desc.num; i++) {
        printf("     %.*s: %llu\n",
               descriptor.desc.len,
               &strings[i*descriptor.desc.len],
               counters[i]);
    }

    free(strings);
    free(counters);

    return DBG_STATUS_SUCCESS;
}

dbg_status_t solaris_cleanup(struct _dev_if_t * dev_if)
{
    close(dev_if->os_params.ioctl_fd);
    return DBG_STATUS_SUCCESS;
}

/* initialize Everest's dispatch table  */
dbg_status_t everest_dev_if_init_dispatch(int        br_idx,
                                          pci_info_t pci_info,
                                          dev_if_t * dev_if)
{
    char dev_path[128];
    u32_t val;
    int fd;

    strcpy(dev_path, "/dev/");
    strcat(dev_path, pci_info.eth_name);

    if ((fd = open(dev_path, O_RDONLY)) <= 0)
    {
        printf("ERROR - failed to open device %s (%d)\n", dev_path, errno);
        return DBG_STATUS_FAILURE;
    }

    dev_if->os_params.ioctl_fd = fd;

    dev_if->dev_if_reg_read     = solaris_reg_read;
    dev_if->dev_if_reg_write    = solaris_reg_write;
    dev_if->dev_if_get_dev_info = NULL;
    dev_if->dev_if_cleanup      = solaris_cleanup;
    dev_if->dev_if_nv_read      = solaris_nv_read;
    dev_if->dev_if_nv_write     = solaris_nv_write;
    dev_if->dev_if_stats_show   = solaris_stats_show;

    /* Read NVM SIZE */
    solaris_reg_read(dev_if->os_params, MCP_REG_MCPR_NVM_CFG4, &val);
    dev_if->os_params.geeprom_length =
        (DRVSIM_NVRAM_1MB_SIZE << (val & MCPR_NVM_CFG4_FLASH_SIZE));

    return DBG_STATUS_SUCCESS;
}

