#include <stdio.h>
#include <windows.h>
#include <ntstatus.h>
#include "os_types.h"
#include "edebug_types.h"
#include "dev_if.h"
#include "edebug_if.h"
#include "57712_reg.h"

#define DRVSIM_NVRAM_1MB_SIZE	0x20000	/* 1M bit in bytes */
#define NVRAM_PAGE_SIZE     	256

static struct reg_access {
	void *mapped_area;
};

static struct reg_access *reg_access;

static int perform_cleanup = 0;

/* read 32 bits value from given grc addr */
dbg_status_t windows_reg_read(
    IN  struct os_params   	os_params,
    IN  u32_t       	  	addr,
    OUT u32_t*      		value
    )
{
    BOOL success;
    u32_t data;
    DWORD bytes_returned;
    chip_read_offset_t read_offset;

    read_offset.offset = (u64_t)addr;
    read_offset.num_of_u32 = 1;

    success = DeviceIoControl(
        (HANDLE)(os_params.dev),
        CHIP_IOC_READ_REG,
        &read_offset,
        sizeof(read_offset),
        &data,                          // v   OUT
        sizeof(data),                   // v
        &bytes_returned,                // v   OUT
        NULL);                          // v

    if(!success)
    {
        return DBG_STATUS_FAILURE;
    }

    *value = data;
    return DBG_STATUS_SUCCESS;
}

/* write 32 bits value to given grc addr */
dbg_status_t windows_reg_write(
    IN  struct os_params   	os_params,
    IN  u32_t       		addr,
    IN  u32_t       		value
    )
{
	BOOL success;
    DWORD bytes_returned;
    chip_write_data_t write_data;

    write_data.offset = (u64_t)addr;
    write_data.num_of_u32 = 1;
    write_data.data[0] = value;

    success = DeviceIoControl(
        (HANDLE)(os_params.dev),
        CHIP_IOC_WRITE_REG,
        &write_data,
        sizeof(write_data),
        NULL,
        0,
        &bytes_returned,
        NULL);

    return success ? DBG_STATUS_SUCCESS : DBG_STATUS_FAILURE;
}

/* read 32 bits value from given pci config space addr */
dbg_status_t windows_pcicfg_read(
    IN  struct os_params   	os_params,
    IN  u32_t       	  	addr,
    OUT u32_t*      		value
    )
{
    BOOL success;
    u32_t data;
    DWORD bytes_returned;
    chip_read_offset_t read_offset;

    read_offset.offset = (u64_t)addr;
    read_offset.num_of_u32 = 1;

    success = DeviceIoControl(
        (HANDLE)(os_params.dev),
        CHIP_IOC_READ_PCICFG,
        &read_offset,
        sizeof(read_offset),
        &data,                          // v   OUT
        sizeof(data),                   // v
        &bytes_returned,                // v   OUT
        NULL);                          // v

    if(!success)
    {
        return DBG_STATUS_FAILURE;
    }

    *value = data;
    return DBG_STATUS_SUCCESS;
}

/* read from nvram */
dbg_status_t windows_nv_read(
    IN   struct os_params   	os_params,
    IN   u32_t       		addr,
    OUT  u32_t*       		value,
	IN   u32_t				size
    )
{
	BOOL success;
    chip_read_offset_t read_offset;
	DWORD bytes_returned;

	read_offset.offset = (u64_t)addr;
	read_offset.num_of_u32 = size/4;

    success = DeviceIoControl(
        (HANDLE)(os_params.dev),
        CHIP_IOC_READ_NVM,
        &read_offset,
        sizeof(read_offset),
        value,
        size,
        &bytes_returned,
        NULL);

	if (!success)
	{
		printf("windows nv_read opertaion failed\n");
		return DBG_STATUS_FAILURE;

	}

    return DBG_STATUS_SUCCESS;
}

/* write to nvram */
dbg_status_t windows_nv_write(
    IN   struct os_params   	os_params,
    IN   u32_t       		addr,
    IN   u32_t*      		value,
	IN   u32_t				size
    )
{
	BOOL success;
    chip_write_data_t *p_write_data;
	DWORD bytes_returned;
	u32_t bytes_left = size;

	p_write_data=malloc(sizeof(chip_write_data_t)+size);

    /* must return success if size==0 as otherwise if last chunk will be zero
     * sized the whole upgrade will fail
     */
    if (!size)
    {
        return DBG_STATUS_SUCCESS;
    }

	if (!p_write_data)
	{
		printf("memory allocation failed\n");
		return DBG_STATUS_FAILURE;
	}

	p_write_data->offset = (u64_t)addr;
	p_write_data->num_of_u32 = size/4;

	memcpy(p_write_data->data, value, size);

	success = DeviceIoControl(
		(HANDLE)(os_params.dev),
		CHIP_IOC_WRITE_NVM,
		p_write_data,
		sizeof(chip_write_data_t)+size,
		NULL,
		0,
		&bytes_returned,
		NULL);

	if (!success)
	{
		printf("windows nv_write opertaion failed\n");
		return DBG_STATUS_FAILURE;

	}

    return DBG_STATUS_SUCCESS;
}

/* get NIC statistics via evbd */
dbg_status_t windows_stats_show(
	IN  struct os_params	os_params)
{
	char *strings;
	u64_t *counters;
	unsigned int size_str, size_cnt;
	int i, err, bytes_returned, opcode;
	b10_stats_show_data_t descriptor;
	DWORD err_code;

	opcode = STATS_SHOW_TYPE_NUM;
	err = DeviceIoControl(
		(HANDLE)(os_params.dev),
		CHIP_IOC_STATS_SHOW,
		&opcode,
		sizeof(opcode),
		&descriptor.desc, 					  // v   OUT
		sizeof(descriptor.desc),  			  // v   OUT
		&bytes_returned,
		NULL);

	if (!err)  {
		err_code = GetLastError();
		printf("Cannot get NIC statistics (num+len ioctl failed), error code = %x\n", err_code);
		return DBG_STATUS_FAILURE;
	}

	if (descriptor.desc.num < 1) {
		printf("Cannot get NIC statistics (num < 1)\n");
		return DBG_STATUS_FAILURE;
	}

	/* allocate memory for the statistics strings and counters */
	size_str = descriptor.desc.num * descriptor.desc.len;
	size_cnt = descriptor.desc.num * sizeof(u64_t);
	strings = calloc(1, size_str);
	if (strings == NULL) {
		printf("Cannot get NIC statistics (memory allocation failed)\n");
		return DBG_STATUS_FAILURE;
	}
	counters = calloc(1, size_cnt);
	if (counters == NULL) {
		printf("Cannot get NIC statistics (memory allocation failed)\n");
		free(strings);
		return DBG_STATUS_FAILURE;
	}

	/* prepare and send second ioctl to get the statistics strings */
	opcode = STATS_SHOW_TYPE_STR;
	err = DeviceIoControl(
		(HANDLE)(os_params.dev),
		CHIP_IOC_STATS_SHOW,
		&opcode,
		sizeof(opcode),
		strings,					   // v   OUT
		size_str,   			 // v   OUT
		&bytes_returned,
		NULL);

	if (!err)  {
		err_code = GetLastError();
		printf("Cannot get NIC statistics (strings ioctl failed), error code = %x\n", err_code);
		free(strings);
		free(counters);
		return DBG_STATUS_FAILURE;
	}

	if (bytes_returned != size_str)  {
		printf("Cannot get NIC statistics (in strings ioctl - bytes returned = %d, expected = %d)\n", bytes_returned, size_str);
		free(strings);
		free(counters);
		return DBG_STATUS_FAILURE;
	}

	/* prepare and send third ioctl to get the statistics counters */
	opcode = STATS_SHOW_TYPE_CNT;
	err = DeviceIoControl(
		(HANDLE)(os_params.dev),
		CHIP_IOC_STATS_SHOW,
		&opcode,
		sizeof(opcode),
		counters,   					// v   OUT
		size_cnt,   			 // v   OUT
		&bytes_returned,
		NULL);

	if (!err)  {
		err_code = GetLastError();
		printf("Cannot get NIC statistics (counters ioctl failed), error code = %x\n", err_code);
		free(strings);
		free(counters);
		return DBG_STATUS_FAILURE;
	}

	if (bytes_returned != size_cnt)  {
		printf("Cannot get NIC statistics (in counters ioctl - bytes returned = %d, expected = %d)\n", bytes_returned, size_cnt);
		free(strings);
		free(counters);
		return DBG_STATUS_FAILURE;
	}
	/* pretty-print the statistics per-driver */
	printf("NIC statistics:\n");
	for (i = 0; i < descriptor.desc.num; i++) {
		printf("	 %.*s: %llu\n",
			descriptor.desc.len,
			&strings[i*descriptor.desc.len],
			counters[i]);
	}

	free(strings);
	free(counters);

	return DBG_STATUS_SUCCESS;
}

dbg_status_t windows_cleanup(
    IN  struct _dev_if_t   	*dev_if
    )
{
	return DBG_STATUS_SUCCESS;
}

static dbg_status_t windows_get_dev_info(
    IN   struct os_params   os_params,
    OUT  pci_info_t			*p_pci_info
    )
{
    BOOL success;
    DWORD bytes_returned;
    b10_device_info_t b10_dev_info;
    u32_t i;

    success = DeviceIoControl(
        (HANDLE)(os_params.dev),
        B10_IOC_GET_DEVICE_INFO,
        NULL,
        0,
        &b10_dev_info,
        sizeof(b10_device_info_t),
        &bytes_returned,
        NULL);

    p_pci_info->device_id =  b10_dev_info.did;

    p_pci_info->svid =  b10_dev_info.svid;

    sprintf(p_pci_info->driver_ver, "%d.%d.%d.%d",
            (b10_dev_info.drv_ver & 0xFF000000) >> 24,
            (b10_dev_info.drv_ver & 0x00FF0000) >> 16,
            (b10_dev_info.drv_ver & 0x0000FF00) >> 8,
            (b10_dev_info.drv_ver & 0x000000FF));

    sprintf(p_pci_info->fw_ver, "%d.%d.%d.%d",
            (b10_dev_info.fw_ver & 0xFF000000) >> 24,
            (b10_dev_info.fw_ver & 0x00FF0000) >> 16,
            (b10_dev_info.fw_ver & 0x0000FF00) >> 8,
            (b10_dev_info.fw_ver & 0x000000FF));

    sprintf(p_pci_info->pcie_lane_width, "PCIE-%d", b10_dev_info.pcie_width);

    /* PCIE speed */
    switch (b10_dev_info.pcie_speed) {
        case PCIE_SPEED_2_5_G:
            sprintf(p_pci_info->pcie_lane_speed, "%s", "2.5");
            break;
        case PCIE_SPEED_5_G:
            sprintf(p_pci_info->pcie_lane_speed, "%s", "5.0");
            break;
	case PCIE_SPEED_8_G:
            sprintf(p_pci_info->pcie_lane_speed, "%s", "8.0");
            break;
        default:			
            sprintf(p_pci_info->pcie_lane_speed, "unknown");
    }

   // dev_info_p->pxe_agent = '\0';         // not available TBD
   sprintf(p_pci_info->mac_addr, "%02X:%02X:%02X:%02X:%02X:%02X",
            b10_dev_info.hw_mac_addr[0],
            b10_dev_info.hw_mac_addr[1],
            b10_dev_info.hw_mac_addr[2],
            b10_dev_info.hw_mac_addr[3],
            b10_dev_info.hw_mac_addr[4],
            b10_dev_info.hw_mac_addr[5]);

    // b10_dev_info.nic_cfg not available TBD
    p_pci_info->pxe_enable  = ((b10_dev_info.nic_cfg & NIC_CFG_PXE_ENABLE) == NIC_CFG_PXE_ENABLE) ? 1 : 0;
    p_pci_info->wol_enable  = ((b10_dev_info.nic_cfg & NIC_CFG_WOL_ENABLE) == NIC_CFG_WOL_ENABLE) ? 1 : 0;
    p_pci_info->mfw_enable  = ((b10_dev_info.nic_cfg & NIC_CFG_MFW_ENABLE) == NIC_CFG_MFW_ENABLE) ? 1 : 0;

    switch ( b10_dev_info.nic_cfg & NIC_CFG_PXE_SPEED_MASK ) {
    case PXE_SPEED_10MB_FD:
        sprintf(p_pci_info->pxe_speed, "10FD");
        break;
    case PXE_SPEED_10MB_HD:
        sprintf(p_pci_info->pxe_speed, "10HD");
        break;
    case PXE_SPEED_100MB_HD:
        sprintf(p_pci_info->pxe_speed, "100HD");
        break;
    case PXE_SPEED_100MB_FD:
        sprintf(p_pci_info->pxe_speed, "100FD");
        break;
    case PXE_SPEED_1GB:
        sprintf(p_pci_info->pxe_speed, "1GFD");
        break;
    case PXE_SPEED_AUTO:
        sprintf(p_pci_info->pxe_speed, "auto");
        break;
    default:
        sprintf(p_pci_info->pxe_speed, "unknown");
    }

    return DBG_STATUS_SUCCESS;
}

/* initialize Everest's dispatch table  */
dbg_status_t everest_dev_if_init_dispatch(
	IN  int 	br_idx,
	IN  pci_info_t  	pci_info,
	OUT dev_if_t*   	dif 		/* dispatch table to be initialized */
	)
{
	HANDLE dev;
	u32_t val;

	/* open handle towards the device */
	dev = CreateFile(pci_info.eth_name, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
	dif->os_params.dev = dev;

	/* TMP TMP */
	dif->os_params.geeprom_length = 0x1000000;

	dif->dev_if_reg_read = windows_reg_read;
	dif->dev_if_reg_write = windows_reg_write;
	dif->dev_if_cleanup = windows_cleanup;
	dif->dev_if_get_dev_info = windows_get_dev_info;
	dif->dev_if_nv_read = windows_nv_read;
	dif->dev_if_nv_write = windows_nv_write;
	dif->dev_if_stats_show = windows_stats_show;

	/* Read NVM SIZE */
	windows_reg_read(dif->os_params, MCP_REG_MCPR_NVM_CFG4, &val);
	dif->os_params.geeprom_length = DRVSIM_NVRAM_1MB_SIZE << (val & MCPR_NVM_CFG4_FLASH_SIZE);

	return DBG_STATUS_SUCCESS;
}

