#include <stdio.h>
#include <string.h>
#include "man_types.h"
#include "os_utils.h"
#include "edebug_types.h"
#include "pcics_reg_driver.h"
#include "hw_access.h"

static s32_t total_pci_dev = -1;
static s32_t total_interfaces = -1;

static dbg_status_t parse_lspci_info(
    OUT   pci_info_t  * pci_info_ar
    );

static dbg_status_t get_resources_info(
    OUT   pci_info_t  * pci_info_ar
    );

static dbg_status_t get_drvsim_info(
    OUT   pci_info_t  * pci_info_ar
    );

static dbg_status_t parse_ifconfig_ethtool_info(
    OUT   pci_info_t  * pci_info_ar
    );

static dbg_status_t parse_ethtool_info(
    OUT   pci_info_t  * pci_info_ar,
    IN    char    * if_name,
    IN    char    * mac_addr
    );


#define DEV_DRVSIM_FILE                 "/dev/drvsim_577xx"
#define PROC_PCI_DEVICES_FILE           "/proc/bus/pci/devices"
#define PROC_PCI_DEVICES_MAX_LINE_LEN   400
#define DEVICE_FOLDER_MAX_LEN           200
#define FILE_NAME_MAX_LEN               220
#define PCI_BDF_BUS_MASK                0xFF00
#define PCI_BDF_BUS_SHIFT               8
#define PCI_BDF_DEVICE_MASK             0xF8
#define PCI_BDF_DEVICE_SHIFT            5
#define PCI_BDF_FUNC_MASK               0x7
#define PCI_BDF_FUNC_SHIFT              0


dbg_status_t os_if_get_total_dev(
    OUT   u32_t       * dev_found_p          /* number of found devices */
    )
{
        char line[200];
        FILE* pipe;
        u32_t result = 0;

        if (total_pci_dev >= 0) {
		*dev_found_p = total_pci_dev;
		return DBG_STATUS_SUCCESS;
	}

        /* get number of devices from lspci output */
        pipe = popen("lspci -n", "r");
        if (!pipe) {
                printf("BW error!\n");
                return DBG_STATUS_FAILURE;
        }

        /* count number of lines in lspci to get num of devices */
        while (fgets(line, sizeof(line), pipe)) {
                result++;
        }

        *dev_found_p = result;
        total_pci_dev = *dev_found_p;

        pclose(pipe);

	return DBG_STATUS_SUCCESS;
}

dbg_status_t os_if_pci_scan(
    OUT   pci_info_t  * pci_info_ar         /* array to return the PCI devices info */
    )
{
        dbg_status_t status = DBG_STATUS_SUCCESS;
        u32_t i, ret_val;
        char tmp[40]="xxxxxxxxxxxxxxxxxxxx";

        status = parse_lspci_info(pci_info_ar);
        if (DBG_STATUS_SUCCESS != status) {
                return status;
        }

        status = get_resources_info(pci_info_ar);
        if (DBG_STATUS_SUCCESS != status) {
                return status;
        }

        status = parse_ifconfig_ethtool_info(pci_info_ar);
        if (DBG_STATUS_SUCCESS != status) {
                return status;
        }
        
        status = get_drvsim_info(pci_info_ar);
        if (DBG_STATUS_SUCCESS != status) {
                return status;
        }

        for (i = 0; i < dbg_get_total_dev(); i++) {
                if (IS_EVEREST(pci_info_ar[i])) {
                        /* set current DUT that pci config space approach will be right */
                        dbg_pcicfg_setdut(pci_info_ar[i].pci_domain, pci_info_ar[i].pci_bus, pci_info_ar[i].pci_device, pci_info_ar[i].pci_func);

                        /* read and parse link_control register */
                        dbg_pcicfg_read(PCICFG_LINK_CONTROL, &ret_val);
                        dbg_pcicfg_process_link_control_reg(pci_info_ar, i, ret_val);

                        /* read and parse SRIOV info in config space */
                        dbg_pcicfg_read(PCI_CFG_SRIOV_BASE + SRIOV_NUM_VFS_OFFSET, &ret_val);
                        if (ret_val) {
                                pci_info_ar[i].total_vfs = ret_val & 0xFF;
                                dbg_pcicfg_read (PCI_CFG_SRIOV_BASE + SRIOV_FIRST_VF_STRIDE_OFFSET, &ret_val);
                                pci_info_ar[i].first_vf_offset = ret_val & 0xFF;
                                pci_info_ar[i].vf_stride = (ret_val & 0xFF00) >> 16;
                        }

                        /* read and parse SVID info from config space */
                        dbg_pcicfg_read(PCICFG_SUBSYSTEM_VENDOR_ID_OFFSET, &ret_val);
                        pci_info_ar[i].svid = (u32)((u16) ret_val);
                }
        }

        return DBG_STATUS_SUCCESS;
}

static dbg_status_t get_resources_info(
    OUT   pci_info_t  * pci_info_ar
    )
{
        FILE* resources_fd;
	char resources_file_name[100];
	char line[100];
	u64_t bar_start, bar_end;
        int i;

        for (i = 0; i < total_pci_dev; i++) {
                /* for each EVEREST device - parse resource file for bars data */
                if (IS_EVEREST(pci_info_ar[i])) {
                        /* open resource file of the device in order to read bar address and length */
                	sprintf(resources_file_name, "/sys/bus/pci/devices/%04x:%02x:%02x.%01x/resource",
							pci_info_ar[i].pci_domain, pci_info_ar[i].pci_bus, pci_info_ar[i].pci_device, pci_info_ar[i].pci_func);
                	resources_fd = fopen(resources_file_name, "rt");
					
                	/* read bar0 adress from first line of resource file */
                	fgets(line, sizeof(line)-1, resources_fd);
                	sscanf(line, "0x%lx 0x%lx", &bar_start, &bar_end);
                        pci_info_ar[i].region0_base_addr = bar_start;
                        pci_info_ar[i].region0_size = bar_end - bar_start + 1;

                        /* read bar2 adress from third line of resource file */
                	fgets(line, sizeof(line)-1, resources_fd);
                        fgets(line, sizeof(line)-1, resources_fd);
                	sscanf(line, "0x%lx 0x%lx", &bar_start, &bar_end);
                        pci_info_ar[i].region2_base_addr = bar_start;
                        pci_info_ar[i].region2_size = bar_end - bar_start + 1;

                        /* close file */
                        fclose(resources_fd);
                }
        }

        return DBG_STATUS_SUCCESS;
}

static dbg_status_t parse_lspci_info(
    OUT   pci_info_t  * pci_info_ar
    )
{
        char line[200];
        char cmd[60];
        char *token;
        FILE *pipe;
        u32_t i, tmp, line_num, region0, region2, tmp2, br_devices = 0;

        /* BDF and vendor/device details are retrieved through lspci */
        pipe = popen("lspci -n", "r");
        if (!pipe) {
                printf("BW error!\n");
                return DBG_STATUS_FAILURE;
        }

        i = 0;

        /* parse each line in the output according to lspci format */
        while (fgets(line, sizeof(line), pipe)) {				

                /* lspci -n can have to formats, depends whether domain is 0 */
                /* first try to read in the format where domain num is not 0 */
                if (7 != (sscanf(line, "%4x:%2x:%2x.%1x %4x: %4x:%4x",
                                 &(pci_info_ar[i].pci_domain),
                                 &(pci_info_ar[i].pci_bus),
                                 &(pci_info_ar[i].pci_device),
                                 &(pci_info_ar[i].pci_func),
                                 &tmp,
                                 &(pci_info_ar[i].vendor_id),
                                 &(pci_info_ar[i].device_id)))) {

                                        /* if reached here coulnd't read in first format */
                                        /* try second format */
                                        if (6 != (sscanf(line, "%x:%x.%x %x: %x:%x",
                                                 &(pci_info_ar[i].pci_bus),
                                                 &(pci_info_ar[i].pci_device),
                                                 &(pci_info_ar[i].pci_func),
                                                 &tmp,
                                                 &(pci_info_ar[i].vendor_id),
                                                 &(pci_info_ar[i].device_id)))) {


                                                        if (6 != (sscanf(line, "%x:%x.%x Class %x: %x:%x",
                                                         &(pci_info_ar[i].pci_bus),
                                                         &(pci_info_ar[i].pci_device),
                                                         &(pci_info_ar[i].pci_func),
                                                         &tmp,
                                                         &(pci_info_ar[i].vendor_id),
                                                         &(pci_info_ar[i].device_id)))) {


                                                                /* if reached here couldn't read in both format */
                                                                /* ERROR */
                                                                printf("error parsing lspci -n!\n");
                                                                return DBG_STATUS_FAILURE;
                                                        }
                                                }

                                                /* in this format domain number is 0*/
                                                pci_info_ar[i].pci_domain = 0;

                                }
                i++;				
        }

        pclose(pipe);

        for (i = 0; i < total_pci_dev; i++) {
                /* hold D0 state by default */
                pci_info_ar[i].eth_state = ETH_D0;				

                /* for each EVEREST device - parse verbose lspci for more data */
                if (IS_EVEREST(pci_info_ar[i])) {
                        br_devices++;					

                        /* build lspci cmd for specific BDF */
                        sprintf(cmd, "lspci -s %04x:%02x:%02x.%0x -vv",
								pci_info_ar[i].pci_domain,
                                pci_info_ar[i].pci_bus,
                                pci_info_ar[i].pci_device,
                                pci_info_ar[i].pci_func);

                        pipe= popen(cmd, "r");
                        if (!pipe) {
                                printf("BW error!\n");
                                return DBG_STATUS_FAILURE;
                        }

					

                        while (fgets(line, sizeof(line), pipe)) {
                                /* split line to words */
                                token=strtok(line, " ");

                                if (token) {
                                        /* in Flags line, parse IRQ info */
                                        if (0 != strstr(token, "Interrupt:")) {
                                                while (token) {
                                                        token=strtok(NULL, " ");
                                                        if (token) {
                                                                if (0 != strstr(token, "IRQ")) {
                                                                        token=strtok(NULL, " ");
                                                                        if (token) {
                                                                                sscanf(token, "%d", &(pci_info_ar[i].irq));
                                                                        }
                                                                        break;
                                                                }
                                                        }
                                                }
                                        }

                                        /* in Status line, parse interface power state */
                                        if (0 != strstr(token, "Status:")) {
                                               token = strtok(NULL, " ");
                                               if (token) {
                                               
                                                       if (0 == strcmp(token, "D0")) {
                                                               pci_info_ar[i].eth_state = ETH_D0;
                                                       }
                                                       if (0 == strcmp(token, "D1")) {
                                                               pci_info_ar[i].eth_state = ETH_D1;
                                                       }
                                                       if (0 == strcmp(token, "D2")) {
                                                               pci_info_ar[i].eth_state = ETH_D2;
                                                       }
                                                       if (0 == strcmp(token, "D3")) {
                                                               pci_info_ar[i].eth_state = ETH_D3;
                                                       }
                                               }
                                        }
                                }
                        }

                        pclose(pipe);
                }
        }

        dbg_set_total_br_dev(br_devices);

        return DBG_STATUS_SUCCESS;
}

static dbg_status_t get_drvsim_info(
    OUT   pci_info_t  * pci_info_ar
    )
{
    /* Read DriverSim supported devices from module */
    dbg_status_t res = DBG_STATUS_SUCCESS;
    FILE *drvsim_fd;
    char buf[40];
    int bus,dev,func;
    u32 i;

    if (!(drvsim_fd = fopen(DEV_DRVSIM_FILE, "r"))) return DBG_STATUS_SUCCESS;       /* No DriverSim installed on SUT */
    while (fgets(buf, sizeof(buf), drvsim_fd)) {
        if (sscanf(buf, "%d:%d:%d", &bus, &dev, &func) != 3) {
            res = DBG_STATUS_FAILURE;      /* Unknown format */
            break;
        }
        /* lookup (b,d,f) in pci array */
        for (i = 0; i < total_pci_dev; i++) {
            if (pci_info_ar[i].pci_bus == bus &&
                pci_info_ar[i].pci_device == dev &&
                pci_info_ar[i].pci_func == func)
            {
                /* register pci device as driversim */
                sprintf(pci_info_ar[i].eth_name,"drvsim");
                break;
            }
        }
        if (i == total_pci_dev) {
            res = DBG_STATUS_FAILURE;      /* (b,d,f) not found in list of devices */
            break;
        }
    }
    fclose(drvsim_fd);
    return res;
}

static dbg_status_t parse_ifconfig_ethtool_info(
    OUT   pci_info_t  * pci_info_ar
    )
{
    FILE *pipe;
	char line[200], if_name[10];
	char *mac_addr=NULL;
	char *token;
	char *is_next_line_if_name="first line", *is_mac_addr_line=NULL;

        /* get all valid interfaces through ifconfig cmd */
        pipe= popen("ifconfig -a", "r");
        if (!pipe) {
                printf("BW error!\n");
                return DBG_STATUS_FAILURE;
        }

	/* parse each line in the output to find interface names */
	while (fgets(line, sizeof(line), pipe)) {
		
		/* mac addr is detailed in ifconfig after the "HWaddr" or "ether" prefix */
		is_mac_addr_line = strstr(line, "HWaddr");
		if (is_mac_addr_line) {
			mac_addr = strtok(is_mac_addr_line, "HWaddr ");
		} else {
			is_mac_addr_line = strstr(line, "ether");
			if (is_mac_addr_line) {
				mac_addr = strtok(is_mac_addr_line, "ether ");
			}
		}

		/* Get if_name */
		token = strtok(line, " ");
		if (is_next_line_if_name) {
			strcpy(if_name, token);
			/* In Red-Hat 7.0 if_name includes :, remove it */
			if (strstr(if_name, ":")) {
				if_name[strlen(if_name)-1]='\0';
			}
		}

		/* If found mac_addr and if_name != "lo"/"sit0" parse_ethtool_info */
		if ((is_mac_addr_line) && strcmp(if_name, "lo")&& strcmp(if_name, "sit0")) {
			
			if (parse_ethtool_info(pci_info_ar, if_name, mac_addr) != DBG_STATUS_SUCCESS) {
				pclose(pipe);
				return DBG_STATUS_FAILURE;
			}
		}

		/* interafce names appears in the first line and after blank lines */
		is_next_line_if_name = strstr(line, "\n");
	}

	pclose(pipe);

	return DBG_STATUS_SUCCESS;
}

static dbg_status_t parse_ethtool_info(
    OUT   pci_info_t  * pci_info_ar,
    IN    char    * if_name,
    IN    char    * mac_addr
    )
{
        char line[200];
        char cmd[60];
        char *token;
        FILE *pipe;
        u32_t i, domain=-1, bus=-1, device=-1, func=-1, index=-1;


        /* if second line is "No data available" then we won't be able to perform ethtool -i to the device */
        sprintf(cmd, "ethtool %s", if_name);
        pipe= popen(cmd, "r");
        if (!pipe) {
                printf("BW error!\n");
                return DBG_STATUS_FAILURE;
        }
        fgets(line, sizeof(line), pipe);
        fgets(line, sizeof(line), pipe);
        if (strcmp("No data available\n", line) != 0) {
        
                /* build ethtool get info command for the input interface */
                sprintf(cmd, "ethtool -i %s", if_name);
                pipe= popen(cmd, "r");
                if (!pipe) {
                        printf("BW error!\n");
                        return DBG_STATUS_FAILURE;
                }


                /* parse each line in the output to find index in pci_info_ar */
                while (fgets(line, sizeof(line), pipe)) {
                                token=strtok(line, " ");

                                if (0 == strcmp("bus-info:", token)) {
                                        token=strtok(NULL, " ");
                                        sscanf(token, "%4x:%2x:%2x.%1x",
                                               &domain, &bus, &device, &func);
                                }
                }

                index = -1;

                if (domain != -1 && bus != -1 && device != -1 && func != -1) {
                        /* find BDF in pci_info_arr */
                        for (i = 0; i < total_pci_dev; i++) {
                                if (pci_info_ar[i].pci_domain == domain &&
                                    pci_info_ar[i].pci_bus == bus &&
                                    pci_info_ar[i].pci_device == device &&
                                    pci_info_ar[i].pci_func == func) {
                                        index = i;
                                        break;
                                }
                        }
                }

                pclose(pipe);

                /* if found device in array, insert additional data to it from ethtool and mac_addr */
                if (index != -1 ) {
                        /* insert mac addr to right index */
                        memcpy(pci_info_ar[i].mac_addr, mac_addr, MAC_ADDR_LEN);

                        /* resend ethtool command */
                        pipe= popen(cmd, "r");
                        if (!pipe) {
                                printf("BW error!\n");
                                return DBG_STATUS_FAILURE;
                        }

                        sscanf(if_name, "%s", &(pci_info_ar[i].eth_name[0]));

                        /* parse each line in the output */
                        while (fgets(line, sizeof(line), pipe)) {
                                        token=strtok(line, " ");

                                        if (0 == strcmp("driver:", token)) {
                                                token=strtok(NULL, " ");
                                                sscanf(token, "%s", &(pci_info_ar[i].driver_name[0]));
                                        }
                                        if (0 == strcmp("version:", token)) {
                                                token=strtok(NULL, " ");
                                                sscanf(token, "%s", &(pci_info_ar[i].driver_ver[0]));
                                        }
                                        if (0 == strcmp("firmware-version:", token)) {
                                                char fw_ver[512];
                                                int fw_ver_offset = 0;

                                                token=strtok(NULL, " ");
                                                while (token) {
                                                        strcpy(fw_ver + fw_ver_offset, token);
                                                        fw_ver_offset += strlen(token);
                                                        fw_ver[fw_ver_offset]='_';
                                                        fw_ver_offset++;
                                                        token=strtok(NULL, " ");
                                                }

                                                if (fw_ver_offset) {
                                                        sscanf(fw_ver, "%s", &(pci_info_ar[i].fw_ver[0]));
                                                } else {
                                                        sscanf("invalid", "%s", &(pci_info_ar[i].fw_ver[0]));
                                                }
                                        }


                        }
                        pclose(pipe);
                }
        }

        return DBG_STATUS_SUCCESS;
}

dbg_status_t os_if_read_configuration_dword(
    IN    u8_t    domain,
    IN    u16_t   bus,
    IN    u8_t    device,
    IN    u8_t    function,
    IN    u32_t   register_number,        /* register offset */
    OUT   u32_t * dword_read              /* output buffer */
    )
{
        char line[200];
        char cmd[60];
        char device_bdf[20];
        char reg_offset_line_start[20];
        char *token;
        FILE *pipe;
        u32_t i, result = 0;
        u32_t tmp;

        /* build lspci command for configuration space data */
        sprintf(device_bdf, "%04x:%02x:%02x.%01x", domain, bus, device, function);
        sprintf(cmd, "lspci -s %s -xxx", device_bdf);
        sprintf(reg_offset_line_start, "%02x", register_number & 0xF0);

        pipe= popen(cmd, "r");
        if (!pipe) {
                printf("BW error!\n");
                return DBG_STATUS_FAILURE;
        }

        /* parse each line in the output to find index in pci_info_ar */
        while (fgets(line, sizeof(line), pipe)) {                        
                        token=strtok(line, ":");                        
                        if (0 == strcmp(reg_offset_line_start, token)) {
                               if (register_number & 0xF > 0xC) {
                                       printf("BW - error - not supported\n");
                                       pclose(pipe);
                                       return DBG_STATUS_FAILURE;
                               }
                               for (i = 0; i < (register_number & 0xF); i++) {
                                       token=strtok(NULL, " ");
                               }
                               token=strtok(NULL, " ");
                               sscanf(token, "%x", &tmp);
                               result = tmp ;
                               token=strtok(NULL, " ");
                               sscanf(token, "%x", &tmp);
                               result |= tmp << 8;
                               token=strtok(NULL, " ");
                               sscanf(token, "%x", &tmp);
                               result |= tmp << 16;
                               token=strtok(NULL, " ");
                               sscanf(token, "%x", &tmp);
                               result |= tmp << 24;
                          }
        }

        *dword_read = result;

        pclose(pipe);

        return DBG_STATUS_SUCCESS;
}

