
/****************************************************************************
 * Copyright(c) 2011-2012 Broadcom Corporation, all rights reserved
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved
 * Proprietary and Confidential Information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without
 * the prior written consent of QLogic Corporation.
 *
 * Name:        uart.c
 *
 * Description: Interrupt driven UART driver for PC
 *
 ****************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <errno.h>
#include "bcmtype.h"
#include "sync.h"
#include "uart.h"
#include "uart_var.h"

/* supported baud rate */
static const u32_t baud_list[] =
{
    600,
    1200,
    4800,
    9600,
    19200,
    38400,
    57600,
    115200,
    0
};

/* supported baud rate */
static const u32_t baud_names[] =
{
    B600,
    B1200,
    B4800,
    B9600,
    B19200,
    B38400,
    B57600,
    B115200,
    0
};

/* per port structure */
static port_t pss[COM_PORT_MAX+1];

/* port hardware config */
static u8_t port_min = 0;
static u8_t port_max = COM_PORT_MAX;

extern u8_t g_usb;

/****************************************************************************
 * implementation
 ****************************************************************************/

/*
 *
 */
ua_status_t os_if_uart_init(void)
{
    return UA_STATUS_OK;
}

/*
 * os_if_uart_open: open a com port and return a handle if successful
 */
ua_status_t os_if_uart_open(
    ua_handle_t * h,
    u8_t          port,    /* 1 (COM1), 2 (COM2), 3, 4 */
    u32_t         baud,    /* 600, 1200, 4800, 9600, 19200, 38400, 57600, 115200 */
    char *        setting, /* "8N1", "7E1", "8O1" */
    u8_t          mode     /* UAM_TEXT, UAM_BINARY */
    )
{
    char port_name[64];
    port_t* ps = &pss[port];
    int fd, i;
    u8_t nbits;
    u32_t baud_name;
    struct termios my_termios;

    /* validate the port */
    if ((port < port_min) || (port > port_max))
    {
        printf("validate port failed\n");
        return UA_STATUS_NODEV;
    }

    if (ps->initialized)
    {
	//os_if_uart_close((ua_handle_t)port);
	return UA_STATUS_PORT_INUSE;
    }

    ps->proto.mode = mode;

    /* validate the baud rate */
    i = 0;

    while (baud > baud_list[i])
    {
        ++i;
    }

    if (baud != baud_list[i])
    {
        return UA_STATUS_BAD_BAUD;
    }

    ps->proto.baud = baud;
    baud_name = baud_names[i];

    /* validate data bit */
    if ((setting == 0) || (*setting == 0))
    {
        return UA_STATUS_BAD_DATABIT;
    }

    nbits = (u8_t)(*setting - '0');

    if ((nbits < 6) || (nbits > 8))
    {
        return UA_STATUS_BAD_DATABIT;
    }

    ps->proto.data_nbits = nbits;

    /* validate parity */
    if (setting[1] == 0)
    {
        return UA_STATUS_BAD_PARITY;
    }

    switch (setting[1] | 0x20)
    {
    case 'n':
        ps->proto.parity = 0;
        break;

    case 'e':
        // ps->proto.parity = UA_LCR_EPAR | UA_LCR_PARITY;
        break;

    case 'o':
        // ps->proto.parity = UA_LCR_PARITY;
        break;

    default:
        return UA_STATUS_BAD_PARITY;
        break;
    }

    /* validate stop bit */
    if (setting[2] == 0)
    {
        return UA_STATUS_BAD_STOPBIT;
    }

    nbits = (u8_t)(setting[2] - '0');

    if ((nbits > 2) || (nbits == 0))
    {
        return UA_STATUS_BAD_STOPBIT;
    }

    //ps->proto.stop_nbits = nbits;

	if (g_usb) {
		sprintf(port_name, "/dev/ttyUSB%d", port-1);               
	} else {
		sprintf(port_name, "/dev/tty0%d", port-1);
	}

    fd = open(port_name, O_RDWR | O_NOCTTY | O_NDELAY);
    if (fd < 0)
    {
        return UA_STATUS_FAILURE;
    }

    tcflush(fd, TCIFLUSH);

    memset(&my_termios, 0, sizeof(my_termios));

    my_termios.c_cflag = CREAD | CLOCAL;
    cfsetospeed (&my_termios, baud_name);
    cfsetispeed (&my_termios, baud_name);

    my_termios.c_cflag &= ~PARENB;
    my_termios.c_cflag &= ~CSTOPB;
    my_termios.c_cflag &= ~CSIZE;
    my_termios.c_cflag |= CS8;

    my_termios.c_lflag  &= ~(ECHO | ICANON | IEXTEN | ISIG);

    my_termios.c_cc[VMIN] = 1; 
    my_termios.c_cc[VTIME] = 0;

    tcsetattr(fd, TCSANOW, &my_termios);

    ps->uart_fd = fd;
    ps->initialized = 1;
    *h = (ua_handle_t)port;

    return UA_STATUS_OK;
}

ua_status_t os_if_uart_close(ua_handle_t h)
{
    port_t* ps = &pss[h];
    close(ps->uart_fd);

    ps->initialized = 0;

    return UA_STATUS_OK;
}

/*
 * os_if_uart_write: non-blocking transmit interface. Buffer the string of characters
 *           in the transmit buffer and drive them out using interrupt.
 */
ua_status_t os_if_uart_write(ua_handle_t h,
                             u8_t *      buf,
                             u16_t       len)
{
    port_t* ps = &pss[h];
    int iOut, i;

    // validate port
    if ((h < port_min) || (h > COM_PORT_MAX) || (pss[h].initialized == 0))
    return UA_STATUS_NODEV;

    /* write data to com port */
    iOut = write(ps->uart_fd, buf, len);

    if (iOut != len)
    {
        printf("error writing to serialport");
        return UA_STATUS_FAILURE;
    }

    return UA_STATUS_OK;
}

void os_if_uart_flush_rb(ua_handle_t h,
                         u32_t       max_buf_size)
{
    port_t* ps = &pss[h];

    tcflush(ps->uart_fd, TCIFLUSH);
    return;
}

/*
 * os_if_uart_read: non-blocking read interface. Copy the buffered input characters
 *          to the caller's buffer.
 */
ua_status_t os_if_uart_read(ua_handle_t h,
                            u8_t *      buf,
                            u16_t       len,
                            u16_t *     outlen)
{
    port_t* ps = &pss[h];
    int iIn;

    iIn = read(ps->uart_fd, buf, len);

    if (iIn < 0)
    {
        if (errno == EAGAIN)
        {
            /* assume no response generated */
            printf("no response generated\n");
            *outlen = 0;
            return UA_STATUS_OK;
        }
        else
        {
            printf("read error\n");
            return UA_STATUS_FAILURE;
        }
    }
    else
    {
        *outlen = iIn;
    }

    return UA_STATUS_OK;
}

/*
 * os_if_uart_flush: flush all the data enqueued in the transmit buffer. This call
 *           is blocking call and it will return only if all data are sent
 *           or flush timeout is reached.
 */
ua_status_t os_if_uart_flush(ua_handle_t h,
                             u16_t       to_sec)
{
    return UA_STATUS_OK;
}

u32_t os_if_uart_get_baud (ua_handle_t h)
{
    return 19200;
}

