#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <syscall.h>
#include <errno.h>
#include <fcntl.h>
#include "bcmtype.h"
#include "edebug_types.h"
#include "dbg_os_utils.h"
#include "os_utils.h"
#include "mem_track.h"
#include "dbg_hmem_def.h"
#include "man_types.h"
#include "dev_info.h"
#include "license.h"
#include "shmem.h"
#include "dbg_shmem.h"
#include "tcl.h"
#include "tcl_al_common.h"

/* externals */
extern int dev_mem_fd;
extern Tcl_Interp *globalInterp;

/* Macros */
#define DEREF_MEM_ENTRY(e) \
          (mem_entry_t *) (((u8_t *) (e)) - sizeof(mem_entry_t))

dbg_status_t os_if_map_iomem (
        u32_t         nbytes,
        dbg_phys_addr_t   phy_addr,
        dbg_virt_addr_t * virt_addr
        )
{
    dbg_phys_addr_t page_size = getpagesize();
    dbg_phys_addr_t page_offset;
    /* If page_size is not a power of 2 we are in trouble! */
    dbg_phys_addr_t  page_mask = page_size - 1;
    int dev_mem_fd2;

    page_offset = phy_addr & page_mask;
    nbytes += page_offset;


    dev_mem_fd2 = open("/dev/mem", O_RDWR);
    if (dev_mem_fd2 == -1) {
	    printf("ERROR - /dev/mem could not be opened\n");
	    return DBG_STATUS_FAILURE;
    }

    *virt_addr = (dbg_virt_addr_t)mmap(NULL, nbytes, PROT_READ | PROT_WRITE,
		      MAP_SHARED | MAP_LOCKED, dev_mem_fd2, phy_addr - page_offset);
    if (*virt_addr == (dbg_virt_addr_t)MAP_FAILED) {
	    printf("mmap failed: %d\n", errno);
	    return DBG_STATUS_FAILURE;
    }

    if((*virt_addr) & page_mask)
    {
	    printf("ERROR: (*virt_addr) & page_mask != 0\n");
    }

    *virt_addr += page_offset;

    return DBG_STATUS_SUCCESS;
}


dbg_status_t os_if_unmap_iomem (
    dbg_virt_addr_t virt_addr,
    u32_t            nbytes
    )
{
    dbg_virt_addr_t page_size = getpagesize();
    dbg_virt_addr_t page_offset;
    /* If page_size is not a power of 2 we are in trouble! */
    dbg_virt_addr_t  page_mask = page_size - 1;

    page_offset = virt_addr & page_mask;
    nbytes += page_offset;

    if (munmap((void*)(virt_addr & (~page_mask)), (int)nbytes)) {
	    printf("munmap failed: %d\n", errno);
	    return DBG_STATUS_FAILURE;
    }

    return DBG_STATUS_SUCCESS;
}

dbg_status_t os_if_alloc_pcontig(
     IN    u32_t           nbytes,
     OUT   dbg_virt_addr_t   * virt_addr_p,  /* User space virtual address */
     OUT   dbg_phys_addr_t   * phy_addr_p   /* PCI bus address */
        )
{
    mem_entry_t * entry = NULL, *aligned_entry = NULL;
    dbg_phys_addr_t aligned_phys;
    dbg_virt_addr_t virt_addr;
    u32_t val_to_write, val_read = 1;
    dbg_dev_info_t *cur_dev_p = XDBG_CUR_DEV;

    /* Need 2 extra cache lines as allocated buffer may be exactly on the
       boundaries of the cache line and entry + sizeof(mem_entry_t) may be on the next cache line.
       Thereby for correct alignment we need one more extra page.  */
    u32_t mem_sz = 2*(CACHE_LINE_SIZE_MASK+1) + nbytes + sizeof(mem_entry_t) + SIG_SZ;
    dbg_phys_addr_t kern_phys_addr;

    /* send to bnx2x command to get physical addr */
    val_to_write = EDEBUG_DRIVER_IF_OP_CODE_GET_PHYS_ADDR;
    val_to_write |= DBG_FUNC_NUM(*cur_dev_p) << 3;
    WRITE_TO_DRIVER_LOW(val_to_write);
    WRITE_TO_DRIVER_HIGH(nbytes);

    /* wait for ack+response from bnx2x */
    while (val_read & 0x7) {
	    READ_FROM_DRIVER_LOW(&val_read);
	    dbg_os_msleep(100);
    }

    /* get full response from bnx2x and insert it to kern_phys_addr */
    kern_phys_addr = val_read;
    READ_FROM_DRIVER_HIGH(&val_read);
    // kern_phys_addr |= ((dbg_phys_addr_t)val_read << 32);

    /* send to bnx2x command to get bus addr */
    val_to_write = EDEBUG_DRIVER_IF_OP_CODE_GET_BUS_ADDR;
    val_to_write |= DBG_FUNC_NUM(*cur_dev_p) << 3;
    WRITE_TO_DRIVER_LOW(val_to_write);
    WRITE_TO_DRIVER_HIGH(0);

    /* wait for ack+response from bnx2x */
    val_read = 1;
    while (val_read & 0x7) {
	    READ_FROM_DRIVER_LOW(&val_read);
	    dbg_os_msleep(100);
    }

    /* get full response from bnx2x and insert it to kern_phys_addr */
    *phy_addr_p = val_read;
    READ_FROM_DRIVER_HIGH(&val_read);
    // *phy_addr_p |= ((dbg_phys_addr_t)val_read << 32);

    /* Map it into the virt space */
    if (os_if_map_iomem(mem_sz, kern_phys_addr, &virt_addr) != DBG_STATUS_SUCCESS) {

	    printf("Failed to map %d bytes (phys. cont.)\n",
			mem_sz);
	    return DBG_STATUS_FAILURE;
    }

    entry = (mem_entry_t *)virt_addr;

    /* Perform alignment */
    aligned_phys = ((*phy_addr_p + sizeof(mem_entry_t)) + CACHE_LINE_SIZE_MASK) &
			(~((dbg_phys_addr_t)CACHE_LINE_SIZE_MASK));
    if(aligned_phys < *phy_addr_p) {
	    printf("ERROR - aligned_phys < *phy_addr_p\n");
    }

    *virt_addr_p = (dbg_virt_addr_t) ((int_ptr_t)(((u8_t*)entry) + aligned_phys - *phy_addr_p));

    /* Write down some houskeeping info - will be needed for correct releasing the memory */
    aligned_entry = ((mem_entry_t*)(*virt_addr_p)) - 1;
    if (aligned_entry < entry) {
	    printf("ERROR: aligned_entry < entry \n");
    }

    aligned_entry->start_addr = (void*)entry;
    aligned_entry->size = mem_sz;
    aligned_entry->phys_addr = kern_phys_addr;
    aligned_entry->bus_addr = *phy_addr_p;

    *phy_addr_p = aligned_phys;

    while (1) {
	    dbg_os_msleep(1000);

    }

    return DBG_STATUS_SUCCESS;
}


dbg_status_t os_if_free_pcontig(
    IN    dbg_virt_addr_t     virt_addr
    )
{
    mem_entry_t * entry = DEREF_MEM_ENTRY(virt_addr);;
    u32_t val_to_write, val_read = 1;
    dbg_dev_info_t *cur_dev_p = XDBG_CUR_DEV;

    munlock(entry->start_addr, entry->size);

    /* Unmap the virtual mem. mapping */
    os_if_unmap_iomem((dbg_virt_addr_t)(entry->start_addr), entry->size);

    /* send to bnx2x command to free physical memory */
    val_to_write = EDEBUG_DRIVER_IF_OP_CODE_FREE_PHYS_MEM;
    val_to_write |= DBG_FUNC_NUM(*cur_dev_p) << 3;
    WRITE_TO_DRIVER_LOW(val_to_write);

    /* wait for ack from bnx2x */
    while (val_read & 0x7) {
	    READ_FROM_DRIVER_LOW(&val_read);
	    dbg_os_msleep(100);
    }

    return DBG_STATUS_SUCCESS;
}

