/****************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved
 * Proprietary and Confidential Information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of QLogic Corporation. 
 *
 * Name:        uart.c
 *
 * Description: Interrupt driven UART driver for PC
 * 
 ****************************************************************************/
#include <stdio.h>
#include <windows.h>
#include "bcmtype.h"
#include "sync.h"
#include "uart.h"
#include "uart_var.h"

/* global serial comm variables */
u32_t windebug_serial_baud;

/* supported baud rate */
static const u32_t baud_list[] =
{
    0
};

/* supported baud rate */
static const u32_t baud_names[] =
{
    0
};

/* per port structure */
static port_t pss[COM_PORT_MAX+1];

/* port hardware config */
static u8_t port_min = 0;
static u8_t port_max = COM_PORT_MAX;

/****************************************************************************
 * implementation
 ****************************************************************************/

/*
 * 
 */
ua_status_t
os_if_uart_init(
    void
    )
{
        return UA_STATUS_OK;
}

/*
 * os_if_uart_open: open a com port and return a handle if successful
 */
ua_status_t
os_if_uart_open(
    ua_handle_t* h,
    u8_t port,          /* 1 (COM1), 2 (COM2), 3, 4 */
    u32_t baud,         /* 600, 1200, 4800, 9600, 19200, 38400, 57600, 115200 */
    char* setting,      /* "8N1", "7E1", "8O1" */
    u8_t mode           /* UAM_TEXT, UAM_BINARY */
    )
{
	HANDLE hSerial;				// windows serial file handle
	DCB dcbSerialParams = {0};
	char str[] = "\\\\.\\COM  "; // For COM > 10 requires prefix \\.\

    char portStr[3] = {0};
	COMMTIMEOUTS no_timeout = {
		500,               /* ReadIntervalTimeout */
		500,               /* ReadTotalTimeoutMultiplier */
		500,               /* ReadTotalTimeoutConstant */
		500,               /* WriteTotalTimeoutMultiplier */
		500,               /* WriteTotalTimeoutConstant */
	};

	windebug_serial_baud = baud;

    if (port > MAX_UART_HANDLES) {
        fprintf(stderr, "windebug:  serial port numbers larger than %d are not supported\n", MAX_UART_HANDLES);
		return UA_STATUS_NODEV;
    }

    /* Add to str "COM" the port number e.g. "COM4" "COM6" */
    sprintf(portStr, "%u", port);
	strncpy(&str[7], portStr, strlen(portStr));

	/* equivalent of tcl script source:  set fd [::open $str r+] */
	hSerial = CreateFile(
		str,
		GENERIC_READ | GENERIC_WRITE,
		0,
		0,
		OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,
		0		
		);
	*h = (ua_handle_t)hSerial;

	if (hSerial == INVALID_HANDLE_VALUE) {
		/* an error has occured. */
		switch (GetLastError()) {
			case ERROR_FILE_NOT_FOUND:
				fprintf(stderr, "windebug:  serial port %d does not exist on this machine\n",port);
				return UA_STATUS_NODEV;
			case ERROR_ACCESS_DENIED:
				fprintf(stderr, "windebug:  serial port %d access denied\n",port);
				return UA_STATUS_PORT_INUSE;
			default:
				fprintf(stderr, "windebug:  serial port %d failed to open\n",port);
				return UA_STATUS_FAILURE;
		}
	}

	/* this was taken from TCL serial comm source code */
	SetupComm(hSerial, 4096, 4096);
	PurgeComm(hSerial, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);

	/* timeout is not used:  there will be blocking */
	SetCommTimeouts(hSerial, &no_timeout);

	/* further configurations:
	 equivalent of tcl script source:  fconfigure $fd -mode $baud,n,8,1 */
	dcbSerialParams.DCBlength=sizeof(dcbSerialParams);
	if (!GetCommState(hSerial, &dcbSerialParams)) {
		fprintf(stderr, "windebug:  error getting comm state for com%d\n",port);
		return UA_STATUS_FAILURE;
	}

	/*	we will set the first 4 fields with function "buildCommDCBProc",  with following order:	[baud=b][parity=p][data=d][stop=s], for example: "110,n,8,1"
	parameter 'setting' should be of length 3.  for example: "8N1", "7E1", "8O1" 
	setting[0] == data
	setting[1] == parity
	setting[2] == stop

	first we set the baud to default of 110.   than we will overwrite it.
	*/
	str[0] = '1';
	str[1] = '1';
	str[2] = '0';
	str[3] = ',';
	str[4] = setting[1];
	str[5] = ',';
	str[6] = setting[0];
	str[7] = ',';
	str[8] = setting[2];
	str[9] = '\0';

	if (BuildCommDCB((PCTSTR)str,&dcbSerialParams) == FALSE) {
		fprintf(stderr, "windebug:  settings for comm port are incorrect\n");
		return UA_STATUS_FAILURE;
	}

	/* overwrite baud 110 */
	dcbSerialParams.BaudRate = baud;

	/*	the rest of the fields will be set manually. */
	/* Default settings for serial communications */ 
	/* this was taken from TCL serial comm source code ,  function: SerialSetOptionProc*/
	dcbSerialParams.fBinary = TRUE;
	dcbSerialParams.fErrorChar = FALSE;
	dcbSerialParams.fNull = FALSE;
	dcbSerialParams.fAbortOnError = FALSE;

	if(!SetCommState(hSerial, &dcbSerialParams)){
		fprintf(stderr, "windebug:  error setting comm state for com%d\n",port);
		return UA_STATUS_FAILURE;
	}


	/* neglected operations that tcl performs:
		1. further configurations made in the tcl script level:
			fconfigure $fd -translation auto -buffering line
			fconfigure $fd -blocking 0
		2.	SerialSetOptionProc() - more options to configure the channel exist there. 
		however with a breakpoint check i confirmed that none of them are called 
		when opening, reading and writing on port.  dani.
	*/

	/* equivalent of tcl script source: 	fconfigure $fd -translation auto -buffering line */
	/* was not implemented */

	/* equivalent of tcl script source: 	fconfigure $fd -blocking 0 */
	/* was not implemented */

	return UA_STATUS_OK;
}

/*
 * os_if_uart_flush: flush all the data enqueued in the transmit buffer. This call
 *           is blocking call and it will return only if all data are sent
 *           or flush timeout is reached.
 */
ua_status_t
os_if_uart_flush(
    ua_handle_t h,
    u16_t to_sec
    )
{
        return UA_STATUS_OK;
}

u32_t os_if_uart_get_baud (
    IN   ua_handle_t h
    )
{    
    return windebug_serial_baud;
}

ua_status_t os_if_uart_read (
	ua_handle_t h,
	u8_t* buf,
	u16_t len,
	u16_t *outlen
	)
{
	DWORD bytes_read;
	HANDLE hSerial = (HANDLE)h;
	
	if (ReadFile(hSerial, buf, len, &bytes_read, NULL)) {
		*outlen = (u16_t)bytes_read;		
		return UA_STATUS_OK;
	} else {
		/* some error occured */
		fprintf(stderr, "windebug: some error occured when reading from serial port.\n");        		
		return UA_STATUS_FAILURE;
	}
}

ua_status_t os_if_uart_write (
	ua_handle_t h,
	u8_t* buf,
	u16_t len	
	)
{
	DWORD bytes_written;
	HANDLE hSerial = (HANDLE)h;

	if (WriteFile(hSerial, buf, len, &bytes_written, NULL)) {
		return UA_STATUS_OK;
	} else {
		/* some error occured */
		//fprintf(stderr, "windebug: some error occured when writing to serial port.\n");		
		return UA_STATUS_FAILURE;
	}
}

ua_status_t os_if_uart_close (
	ua_handle_t h
	)
{
    HANDLE hSerial = (HANDLE)h;
	
	if (CloseHandle(hSerial) == FALSE) {
		return UA_STATUS_FAILURE;
	} else {
		return UA_STATUS_OK;
	}	
}

void
os_if_uart_flush_rb(
    IN ua_handle_t h,
    IN u32_t max_buf_size
    )
{
        PurgeComm((HANDLE)h, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
        return;UA_STATUS_OK;
}

