 /****************************************************************************
 * Copyright(c) 2011-2014 Broadcom Corporation, all rights reserved
 * Copyright(c) 2014-2015 QLogic Corporation, all rights reserved
 * Proprietary and Confidential Information.
 *
 * This source file is the property of QLogic Corporation, and
 * may not be copied or distributed in any isomorphic form without 
 * the prior written consent of QLogic Corporation. 
 *
 * Name:        tcl_mcp_grc.c
 *
 * Description: This module contains routines to monitor MCP GRCs in the everest chip.
 *
 *
 ****************************************************************************/
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "tcl_al.h"
#include "tcl_util.h"            
#include "tcl_al_cmd_handlers.h"
#include "tcl_al_common.h"
#include "parser.h"
#include "bcmtype.h"
#include "license.h"
#include "mcp_shmem.h"
#include "disputil.h"
#include "tcl_mcp.h"
#include "tcl_al.h"
#include "mcp_fio.h"
#include "57712_reg.h"
#include "util.h"


/*
 *------------------------------------------------------------------------
 * GetMcpGrcHelp --
 *
 *     Display Help info for a MCP GRC register
 *
 * Results:
 *     A const string
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static char const * GetMcpGrcHelp(u32_t addr)
{
    char const * str;
                                                                   
    // The MCP block sits in the common area in Everest. As with all common blocks, it may be addressed using any of 8 functions.
    // There are several types of resolution functions in MCP and its sub-blocks:
    // SPLIT2 = bit has two registers. One is accessed when even functions are accessed. A different value is access when odd
    // functions are accessed. These bits are written and read independently
    // SPLIT4 = bit has four registers. This allows per-port/per-path access for 2 ports and 2 paths. These registers are accessed
    // based on function. Func[0] selects the path and func[1] selects the port. These bits are written and read independently
    // SPLIT8 = bit has eight registers. There is one register per function. These bits are written and read independently
    // SHARE = a single register is provided. All function views see the exact same register. Software has to be aware that what
    // it reads may not be what it has written if the other port overwrites the previously written value.

    switch (addr)
    {
    case MCP_REG_MCPR_MCP_CONTROL:         
        str =  "\nMCP_REG_MCPR_MCP_CONTROL - (Offset: 0x80080; Width: 32)\n"
               "  SHARE: These are basic configurations for the MCP block.\n"
               "bit 31 MCP_ISOLATE This bit is set by the driver before it sets the MCP_RESET bit. When set\n"
               "  this bit causes UMP DMA Engines to abort the current transfers and\n"
               "  disables MCP's GRC Master interface. This bit should cleared by the driver\n"
               "  when the MCP reset completes. \tRW\t0\n"
               "bit 30 SMBUS_SEL When set this bit select new SMBUS block to interface to the SMBUS pins.\n"
               "  When clear it selects the legacy SMBUS (ASF). This bit is set by the driver\n"
               "  before it sets the MCP_RESET bit. NOTE: The legacy SMBUS block was\n"
               "  removed for E2, so this bit no longer has any effect.\tRW\t0\n"
               "bit 29-0 unused0 RO 0\n";
        break;
    case MCP_REG_MCPR_MCP_ATTENTION_STATUS:         
        str =  "\nMCP_REG_MCPR_MCP_ATTENTION_STATUS - (Offset: 0x80084; Width: 32)\n"
               "  SPLIT4: This register shows the status of the bits that generate attention from the MCP. Only the drv_doorbell\n"
               "  field is split. Others are SHARE.\n"
               "bit 31 CPU_EVENT This bit is set any time an internal CPU event that requires the\n"
               "  driver's attention happens.\tRO\t0\n"
               "bit 30 WATCHDOG_TIMEOUT This bit is set when the watchdog timer expires. This bit reflects\n"
               "  state of the WATCHDOG_ATTN bit. When this bit is written as '1', the\n"
               "  value will return to '0'.\tWC\t0\n"
               "bit 29 DRV_DOORBELL This bit is set when the MCP posted message for the driver in the\n"
               "  driver's mailbox (that is it reflects state of the DRV_DOORBELL bit).\n"
               "  The mapping of the bits in the DRIVER_DOORBELL register to the\n"
               "  port/path based bits here depends on whether the chip is in 2-port\n"
               "  mode or 4-port mode. In 2-port mode, even doorbells map to\n"
               "  path0/port0 and odd doorbells map to path1/port0. In 4-port mode,\n"
               "  0,4 -> path0/port0; 1,5 -> path1/port0; 2,6 -> path0/port1, and 3,7\n"
               "  -> path1/port1.\tRO\t0\n"
               "bit 28-0 unused0 \tRO\t0\n";
        break;
    case MCP_REG_MCPR_MCP_HEARTBEAT_CONTROL:         
        str =  "\nMCP_REG_MCPR_MCP_HEARTBEAT_CONTROL - (Offset: 0x80088; Width: 32)\n"
               "  SHARE: Control for MCP heartbeat feature.\n"
               "bit 31 MCP_HEARTBEAT_ENABLE When this bit is set by the driver it indicates to MCP that it should\n"
               "  start incrementing the MCP_HEARTBEAT register. The MCP reports\n"
               "  the increment period to the driver using MCP_HEARTBEAT_STATUS\n"
               "  register.\tRW\t0\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_MCP_HEARTBEAT_STATUS:    
        str =  "\nMCP_REG_MCPR_MCP_HEARTBEAT_STATUS - (Offset: 0x8008C; Width: 32)\n"
               "  SHARE: Status of MCP heartbeat feature.\n"
               "bit 31 VALID When set this bit validates bits 10-0 of this register.\tRW\t0\n"
               "bit 30-11 unused0\tRO\t0\n"
               "bit 10-0 MCP_HEARTBEAT_PERIOD This value is written by the MCP and indicates (in ms) to the\n"
               "  driver MCP_HEARTBEAT increment period. This is just rough\n"
               "  increment period as estimated by the firmware and not the exact\n"
               "  increment period.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_MCP_HEARTBEAT:       
        str =  "\nMCP_REG_MCPR_MCP_HEARTBEAT - (Offset: 0x80090; Width: 32)\n"
               "  SHARE: Doorbell registoer for MCP heartbeat feature.\n"
               "bit 31 MCP_HEARTBEAT_RESET When set this bit resets the heartbeat counter. Typically used by\n"
               "  the MCP or the driver.\tSC\t0\n"
               "bit 30 MCP_HEARTBEAT_INC When set this bit causes MCP heartbeat counter to increment.\n"
               "  Typically used by the MCP.\tSC\t0\n"
               "bit 29-0 MCP_HEARTBEAT_COUNT This is free running counter incremented roughly with the period\n"
               "  that is specified in MCP_HEARTBEAT_STATUS register.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_WATCHDOG_RESET:         
        str =  "\nMCP_REG_MCPR_WATCHDOG_RESET - (Offset: 0x80094; Width: 32)\n"
               "  SHARE: Doorbell to reset the watchog timer.\n"
               "bit 31 WATCHDOG_RESET When set this bit resets the watchdog timer. Typically used by the MCP\n"
               "  or the driver.\tSC\t0\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_WATCHDOG_CONTROL:         
        str =  "\nMCP_REG_MCPR_WATCHDOG_CONTROL - (Offset: 0x80098; Width: 32)\n"
               "  SHARE: Control for the watchog timer.\n"
               "bit 31 WATCHDOG_ENABLE When set this bit enables the watchdog timer. Typically used by\n"
               "  the driver.\tRW\t0\n"
               "bit 30 MCP_RST_ENABLE When set this bit enables the watchdog timer to reset the MCP\n"
               "  instead of halting it. The watchdog hardware must set\n"
               "MCP_ISOLATE bit before it resets the MCP.\tSC\t0\n"
               "bit 29 WATCHDOG_ATTN When set indicates that watchdog timer has reached 0 and that it\n"
               "  requires driver's attention. Low to high transition on this bit should\n"
               "  generate MCP attention toward the HC which will send it back to\n"
               "  the driver using driver status block. Cleared when the watchdog\n"
               "  timer is reset.\tRO\t0\n"
               "bit 28 unused0\tRO\t0\n"
               "bit 27-0 WATCHDOG_TIMEOUT These bits specify the watchdog timeout period with respect to the\n"
               "  core_clk. Using core clock should provide timeout period scaling in\n"
               "  case where core/cpu clocks are slowed down. Reset value is 100ms\tRW\t0x989680\n";
        break;
    case MCP_REG_MCPR_ACCESS_LOCK:         
        str =  "\nMCP_REG_MCPR_ACCESS_LOCK - (Offset: 0x8009C; Width: 32)\n"
               "  SPLIT8: Access Lock Register.\n"
               "bit 31 LOCK Driver writes '1' to this bit in order to obtain the lock over the shared\n"
               "  resources within the chip. The actual \"lock\" is implemented in hardware\n"
               "  using the state machine that keeps track of who is the owner of the lock.\n"
               "  Only the owner of the lock can release the lock. When read by the driver as\n"
               "  '1' (after it was written with '1' it tells the driver that it obtained the lock.\n"
               "  If read as '0' it means that the other driver holds the lock. Driver writes '0'\n"
               "  to this bit to release the lock in case that it owns the lock.\tRW\t0\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_TOE_ID:         
        str =  "\nMCP_REG_MCPR_TOE_ID - (Offset: 0x800A0; Width: 32)\n"
               "  SPLIT2: Function ID register\n"
               "bit 31 FUNCTION_ID This bit tells driver the PCIE function that is associated with. '0'\n"
               "  corresponds to even PCIE functions '1' to odd PCIE functions. Since the\n"
               "  value is different for both functions, the reset value is shown as unknown.\tRO\tX\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_MAILBOX_CFG:         
        str =  "\nMCP_REG_MCPR_MAILBOX_CFG - (Offset: 0x800A4; Width: 32)\n"
               "  SPLIT2: Configuration for mailbox trigger.\n"
               "bit 31-20 MAILBOX_SIZE Mailbox size in 32-bit words. Default mailbox size is 1KB.\tRW\t0x100\n"
               "bit 19-14 unused0\tRO\t0\n"
               "bit 13-0 MAILBOX_OFFSET Offset (in 32-bit words) of the mailbox within the MCP scratchpad. There\n"
               "  are two reset values. Register that corresponds to even functions should\n"
               "  have reset value of 0x3EC0 and register that corresponds to odd functions\n"
               "  should have reset value of 0x3F00\tRW\t0x3ec0\n";
        break;
    case MCP_REG_MCPR_MAILBOX_CFG_OTHER_FUNC:         
        str =  "\nMCP_REG_MCPR_MAILBOX_CFG_OTHER_FUNC - (Offset: 0x800A8; Width: 32)\n"
               "  SPLIT2: Configuration for mailbox trigger.\n"
               "bit 31-20 MAILBOX_SIZE Mailbox size in 32-bit words. Default mailbox size is 1KB.\tRW\t0x100\n"
               "bit 19-14 unused0\tRO\t0\n"
               "bit 13-0 MAILBOX_OFFSET Offset (in 32-bit words) of the mailbox within the MCP scratchpad. There\n"
               "  are two reset values. Register that corresponds to even functions should\n"
               "  have reset value of 0x3EC0 and register that corresponds to odd functions\n"
               "  should have reset value of 0x3F00\tRW\t0x3ec0\n";
        break;
    case MCP_REG_MCPR_MCP_DOORBELL:         
        str =  "\nMCP_REG_MCPR_MCP_DOORBELL - (Offset: 0x800AC; Width: 32)\n"
               "  SPLIT8: Doorbell that is written by firmware.\n"
               "bit 31 MCP_DOORBELL Set by the driver to generate the MCP's event. Low to high transition on\n"
               "  this bit should generate event toward the HC which will send it back to the\n"
               "  MCP using MCP's events_bits bus. Cleared by the MCP when it services the\n"
               "  message.\tRW\t0\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_DRIVER_DOORBELL:
        str =  "\nMCP_REG_MCPR_DRIVER_DOORBELL - (Offset: 0x800B0; Width: 32)\n"
               "  SPLIT8: Doorbell that is written by driver.\n"
               "bit 31 DRIVER_DOORBELL Set by the MCP (or the other driver using\n"
               "  DRIVER_DOORBELL_OTHER_FUNC) to generate driver's attention. Low to\n"
               "  high transition on this bit should generate MCP attention toward the HC\n"
               "  which will send it back to the driver using driver status block. Cleared by\n"
               "  the driver when it services the message.\tRW\t0\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_DRIVER_DOORBELL_OTHER_FUNC:
        str =  "\nMCP_REG_MCPR_DRIVER_DOORBELL_OTHER_FUNC - (Offset: 0x800B4; Width: 32)\n"
               "  SPLIT2: Doorbell that is written by other function driver.\n"
               "bit 31 DRIVER_DOORBELL Set by the MCP (or the other driver using\n"
               "  DRIVER_DOORBELL_OTHER_FUNC) to generate driver's attention. Low to\n"
               "  high transition on this bit should generate MCP attention toward the HC\n"
               "  which will send it back to the driver using driver status block. Cleared by\n"
               "  the driver when it services the message. Even functions write to\n"
               "  driver_doorbell[1]. Odd functions write to driver_doorbell[0].\tRW\t0\n"
               "bit 30-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_PORT4MODE_EN:
        str =  "\nMCP_REG_MCPR_PORT4MODE_EN - (Offset: 0x800B8; Width: 32)\n"
               "  SHARE: Enable 4 port mode.\n"
               "bit 31-1 unused0\tRO\t0\n"
               "bit 0 FOUR_PORT_MODE When set this bit enables 4 port mode. When cleared, MCP is in 2 port\n"
               "  mode.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_MCP_VFID:
        str =  "\nMCP_REG_MCPR_MCP_VFID - (Offset: 0x800BC; Width: 32)\n"
               "  SHARE: VFID settings.\n"
               "bit 31 PATH_FORCE\tRW\t0\n"
               "bit 30-21 unused2\tRO\t0\n"
               "bit 20 PATHID\tRW\t0\n"
               "bit 19-17 unused1\tRO\t0\n"
               "bit 16 VFID_VALID\tRW\t0\n"
               "bit 15-6 unused0\tRO\t0\n"
               "bit 5-0 VFID\tRW\t0\n";
        break;
    case MCP_REG_MCPR_CPU_MODE:
        str =  "\nMCP_REG_MCPR_CPU_MODE - (Offset: 0x85000; Width: 32)\n"
               "  SHARE: CPU Internal registers.\n"
               "bit 31-16 unused3\tRO\t0\n"
               "bit 15 SPAD_UNDERFLOW_HALT_ENA When this bit is set, the CPU will halt when state bit 11 is\n"
               "  set.\tRW\t0x1\n"
               "bit 14 unused2\tRO\t0\n"
               "bit 13 FIO_ABORT_HALT_ENA When this bit is set, the CPU will halt when a abort is\n"
               "  indicated from any \"Fast IO\" space peripheral.\tRW\t0x1\n"
               "bit 12 BAD_INST_HALT_ENA When this bit is set, the CPU will halt when any condition\n"
               "  that causes bit 6 in the CPU state register to be set occurs.\n"
               "  This bit is cleared by an interrupt.\tRW\t0x1\n"
               "bit 11 BAD_DATA_HALT_ENA When this bit is set, the CPU will halt when any condition\n"
               "  that causes bit 5 in the CPU state register to be set occurs.\n"
               "  This bit is cleared by an interrupt.\tRW\t0x1\n"
               "bit 10 SOFT_HALT When this bit is set, the CPU will halt. This bit is cleared by\n"
               "  an exception or reset. If the processor does not have a ROM,\n"
               "  then this bit will reset to set so that no code is executed\n"
               "  from the scratchpad. If the processor does have a ROM, this\n"
               "  bit resets a cleared so that the processor executes from ROM\n"
               "  after reset.\tRW\t0\n"
               "bit 9-8 unused1\tRO\t0\n"
               "bit 7 INTERRUPT_ENA When this bit is set to 1, the interrupt is enabled. When this\n"
               "  bit is zero, any interrupt will be ignored. This bit can also be\n"
               "  set by writing the interrupt_enable register\tRW\t0\n"
               "bit 6 MSG_BIT1 This is a simple\tRW\tbit.\tRW\t0x1\n"
               "bit 5-4 unused0\tRO\t0\n"
               "bit 3 PAGE_0_INST_ENA This bit enables the processor to halt and to latch the value\n"
               "  of bit 4 of the state register when an instruction references\n"
               "  the first 256 bytes of memory space (page 0). This bit is\n"
               "  cleared by an interrupt or reset.\tRW\t0\n"
               "bit 2 PAGE_0_DATA_ENA This bit enables the processor to halt and to latch the value\n"
               "  of bit 3 of the state register when data references the first\n"
               "  256 bytes of memory space (page 0). This bit is cleared by an\n"
               "  interrupt or reset.\tRW\t0\n"
               "bit 1 STEP_ENA When this bit is set, the processor is allowed to execute one\n"
               "  cycle regardless of any halt conditions. If the halting\n"
               "  condition still exists, the CPU will halt again after the one\n"
               "  cycle, otherwise, it will resume normal operation.\tRW\t0\n"
               "bit 0 LOCAL_RST When this bit is written to a 1, the processor will reset as if\n"
               "  from power-up state. All \"Reset\" value of registers will be\n"
               "  assigned.\tSC\t0\n";
        break;
    case MCP_REG_MCPR_CPU_STATE:
        str =  "\nMCP_REG_MCPR_CPU_STATE - (Offset: 0x85004; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 BLOCKED_READ This bit indicates that a blocking data cache miss occurred,\n"
               "  causing the CPU to stall while data is fetched from memory. This\n"
               "  is intended as a debugging tool. No state is saved other than the\n"
               "  fact that the miss occurred. This bit is cleared by writing a 1 to\n"
               "  this bit position.\tWC\t0\n"
               "bit 30-16 unused3\tRO\t0\n"
               "bit 15 INST_FETCH_STALL This bit is set while the processor is stalled due to instruction\n"
               "  fetch.\tRO\t0\n"
               "bit 14 DATA_ACCESS_STALL This bit is set while the processor is stalled due to data access.\tRO\t0\n"
               "bit 13 unused2\tRO\t0\n"
               "bit 12 INTERRRUPT This bit is each time an interrupt input is asserted, regardless of\n"
               "  the interrupt enable bit (bit 7, mode).\tWC\t0\n"
               "bit 11 SPAD_UNDERFLOW This bit is each time an attempt is made to access the underflow\n"
               "  area of the Scratchpad.\tWC\t0\n"
               "bit 10 SOFT_HALTED This bit is set while the processor is halted due to the setting of\n"
               "  bit 10 in the mode register.\tRO\t0\n"
               "bit 9 unused1\tRO\t0\n"
               "bit 8 FIO_ABORT_HALTED This bit is set while the processor is halted due to the generation\n"
               "  of a abort condition by one, or more, \"Fast IO\" devices within the\n"
               "  CPU block. This will only happen if halt is enabled by bit 13 in the\n"
               "  mode register.\tRO\t0\n"
               "bit 7 ALIGN_HALTED This bit is set while the processor is halted due to bad memory\n"
               "  alignment problem on a load or store instruction. This bit is\n"
               "  cleared by writing a 1 to this bit position.\tWC\t0\n"
               "bit 6 BAD_PC_HALTED This bit is set while the processor is halted due to bad value in\n"
               "  the Program Counter (PC). This bit is cleared by writing a 1 to\n"
               "  this bit position.\tWC\t0\n"
               "bit 5 BAD_DATA_ADDR_HALTED This bit is set while the processor is halted due to bad data\n"
               "  reference address. This bit is cleared by writing a 1 to this bit\n"
               "  position.\tWC\t0\n"
               "bit 4 PAGE_0_INST_HALTED This bit is set while the processor is halted due to executing an\n"
               "  instruction within page 0 (the first 256 bytes) of memory. This bit\n"
               "  is cleared by writing a 1 to this bit position.\tWC\t0\n"
               "bit 3 PAGE_0_DATA_HALTED This bit is set while the processor is halted due to accessing data\n"
               "  within page 0 (the first 256 bytes) of memory. This bit is cleared\n"
               "  by writing a 1 to this bit position.\tWC\t0\n"
               "bit 2 BAD_INST_HALTED This bit is set while the processor is halted due fetching an\n"
               "  invalid instruction. This bit is cleared by writing a 1 to this bit\n"
               "  position.\tWC\t0\n"
               "bit 1 unused0\tRO\t0\n"
               "bit 0 BREAKPOINT This bit is set while the processor is halted due reaching a\n"
               "  hardware breakpoint as enabled in the mode register. This bit is\n"
               "  cleared by writing a 1 to this bit position.\tWC\t0\n";
        break;
    case MCP_REG_MCPR_CPU_EVENT_MASK:
        str =  "\nMCP_REG_MCPR_CPU_EVENT_MASK - (Offset: 0x85008; Width: 32)\n"
               "  SHARE: This register provides one bit for each state register bit to enable it into the equation for generation the\n"
               "  TX Processor Attention output. The reset value of 1 masks all halt conditions from generating an attention.\n"
               "bit 31-13 unused2\tRO\t0\n"
               "bit 12 INTERRUPT_MASK This bit attention when bit 12 of the state register is set.\tRW\t0\n"
               "bit 11 SPAD_UNDERFLOW_MASK This bit attention when bit 11 of the state register is set.\tRW\t0\n"
               "bit 10 SOFT_HALTED_MASK This bit enables soft halts to generate Attention output.\tRW\t0x1\n"
               "bit 9 unused1\tRO\t0\n"
               "bit 8 FIO_ABORT_MASK This bit enables the attention output when bit 8 of the\n"
               "  state register is set.\tRW\t0x1\n"
               "bit 7 ALIGN_HALTED_MASK This bit enables alignment errors to generate Attention\n"
               "  output.\tRW\t0\n"
               "bit 6 BAD_PC_HALTED_MASK This bit enables invalid PC values to generate Attention\n"
               "  output.\tRW\t0\n"
               "bit 5 BAD_DATA_ADDR_HALTED_MASK This bit enables invalid data addresses to generate\n"
               "  Attention output.\tRW\t0\n"
               "bit 4 PAGE_0_INST_HALTED_MASK This bit enables page 0 instructions to generate Attention\n"
               "  output.\tRW\t0\n"
               "bit 3 PAGE_0_DATA_HALTED_MASK This bit enables page 0 data access to generate Attention\n"
               "  output.\tRW\t0\n"
               "bit 2 BAD_INST_HALTED_MASK This bit enables invalid instruction decodes to generate\n"
               "  Attention output.\tRW\t0\n"
               "bit 1 unused0\tRO\t0\n"
               "bit 0 BREAKPOINT_MASK This bit enables breakpoints to generate Attention output.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_CPU_PROGRAM_COUNTER:
        str =  "\nMCP_REG_MCPR_CPU_PROGRAM_COUNTER - (Offset: 0x8501C; Width: 32) -\n"
               "  (RW/Reset: 0)\n"
               "  SHARE: This register allows the program counter to read at any time. The value can be modified when the\n"
               "  processor is halted. Writes will also clear any pending instruction in the decode stage of the pipeline. Bits 31-2\n"
               "  are implemented. '1's written to bits 1-0 are ignored. If the processor has a ROM, then the reset value of this\n"
               "  register points to the internal ROM. If the processor does not have a ROM, then this reset value points to the scratchpad area.\n";
        break;
    case MCP_REG_MCPR_CPU_INSTRUCTION:
        str =  "\nMCP_REG_MCPR_CPU_INSTRUCTION - (Offset: 0x85020; Width: 32) - (RW/Reset: 0)\n"
               "  SHARE: This register allows access instruction in the decode sate of the pipeline while the processor is halted.\n"
               "  This register is only intended for debugging use. This register may be used to replace a halt instruction with some\n"
               "  other instruction after the halt has been executed.\n";
        break;
    case MCP_REG_MCPR_CPU_DATA_ACCESS:
        str =  "\nMCP_REG_MCPR_CPU_DATA_ACCESS - (Offset: 0x85024; Width: 32) - (RO/Reset: 0)\n"
               "  SHARE: This register allows access to the address of the current data access of the processor. It is only valid when\n"
               "  the processor is doing a load or a store. Normally this will be used for debug of stalls caused by firmware accesses\n"
               "  to invalid memory areas.\n";
        break;
    case MCP_REG_MCPR_CPU_INTERRUPT_ENABLE:
        str =  "\nMCP_REG_MCPR_CPU_INTERRUPT_ENABLE - (Offset: 0x85028; Width: 32) -\n"
               "  (WO/Reset: 0)\n"
               "  SHARE: Any write to this register will enable CPU Interrupts (set bit 7 in mode register). This register is intended\n"
               "  to allow a way to return from an interrupt service routine (ISR) using only 2 general purpose registers. MIPS\n"
               "  conventions reserve registers 26 and 27 (k0 and k1) for use by an interrupt handler. At the end of an ISR, k0\n"
               "  should be loaded with the return address from the CPU Interrupt Saved PC register. Then k1 should be loaded with\n"
               "  the address of the CPU Interrupt Enable register. The last 2 instructions in the ISR should be a jump register (jr)\n"
               "  to k0 followed immediately by a store word (sw) to k1. This ensures that we can't respond to another interrupt\n"
               "  until we are safely out of the ISR. Interrupts can also be enabled through the CPU Mode Register. They can be\n"
               "  disabled only through the CPU Mode Register. Each time this register is written, bit 7 of the mode register is set.\n"
               "  The data value of the write is not used. The read value of this register is always zero.\n";
        break;
    case MCP_REG_MCPR_CPU_INTERRUPT_VECTOR:
        str =  "\nMCP_REG_MCPR_CPU_INTERRUPT_VECTOR - (Offset: 0x8502C; Width: 32) -\n"
               "  (RW/Reset: 0x8000020)\n"
               "  SHARE: This register sets the program counter value that will be loaded when an interrupt is performed due to the\n"
               "  interrupt input.\n";
        break;
    case MCP_REG_MCPR_CPU_INTERRUPT_SAVED_PC:
        str =  "\nMCP_REG_MCPR_CPU_INTERRUPT_SAVED_PC - (Offset: 0x85030; Width: 32) -\n"
               "  (RO/Reset: 0)\n"
               "  SHARE: This register reports the PC that was saved during the execution of an interrupt.\n";
        break;
    case MCP_REG_MCPR_CPU_HW_BREAKPOINT:
        str =  "\nMCP_REG_MCPR_CPU_HW_BREAKPOINT - (Offset: 0x85034; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-2 ADDRESS This field sets the 32-bit word on which the * hardware breakpoint will\n"
               "  execute.\tRW\t0\n"
               "bit 1 unused0\tRO\t0\n"
               "bit 0 DISABLE Reset: 1 When this bit is set, the hardware breakpoint feature is disabled.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_CPU_DEBUG_VECT_PEEK:
        str =  "\nMCP_REG_MCPR_CPU_DEBUG_VECT_PEEK - (Offset: 0x85038; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-28 2_SEL 4 bit select for the peek value of the set-2 debug visibility vector.\tRW\t0\n"
               "bit 27 2_PEEK_EN When this bit is '0', then the debug visiblity mux is controlled by the\n"
               "  setting in the misc. block and is available on the visibility output pins.\n"
               "  When this bit is '1', then the mux is controlled by 1_SEL.\tRW\t0\n"
               "bit 26-16 2_VALUE 11 bit set-2 debug visibility vector value. This value is selected by the\n"
               "  2_SEL value and enabled by 2_PEEK_EN. This value is undefined if\n"
               "  2_PEEK_EN is '0'.\tRO\t0\n"
               "bit 15-12 1_SEL 4 bit select for the peek value of the set-1 debug visibility vector.\tRW\t0\n"
               "bit 11 1_PEEK_EN When this bit is '0', then the debug visiblity mux is controlled by the\n"
               "  setting in the misc. block and is available on the visibility output pins.\n"
               "  When this bit is '1', then the mux is controlled by 1_SEL.\tRW\t0\n"
               "bit 10-0 1_VALUE 11 bit set-1 debug visibility vector value. This value is selected by the\n"
               "  1_SEL value and enabled by 1_PEEK_EN. This value is undefined if\n"
               "  1_PEEK_EN is '0'.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_CPU_LAST_BRANCH_ADDR:
        str =  "\nMCP_REG_MCPR_CPU_LAST_BRANCH_ADDR - (Offset: 0x85048; Width: 32)\n"
               "  SHARE: This register indicates that address and branch type of the last branch that was taken. This register is for\n"
               "  debug use only.\n"
               "bit 31-2 LBA This value indicates the address of the last branch that was taken. An\n"
               "  offset as indicated by the type field must be subtracted from this value.\tRO\t0\n"
               "bit 1 TYPE This bit indicates the type of branch that * was last taken.\tRO\t0\n"
               "  Value Name Description\n"
               "   0 JUMP 4 bytes must be subtracted from the LBA value to\n"
               "  determine the actual address of the branch instruction\n"
               "  that caused this register to load.\n"
               "   1 BRANCH 8 bytes must be subtracted from the LBA value to\n"
               "  determine the actual address of the branch instruction\n"
               "  that caused this register to load.\n"
               "bit 0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_HMCPQ_FTQ_CMD:
        str =  "\nMCP_REG_MCPR_HMCPQ_FTQ_CMD - (Offset: 0x853B8; Width: 32)\n"
               "  SPLIT: FTQ control registers. The depth of each Host Management Processor FTQ is 32 records.\n"
               "bit 31 BUSY This bit reads as '1' when the FTQ controller is busy and can not respond\n"
               "  to any of the command bits below. This bit must be '0' when POP,\n"
               "  INTERVENE_CLR, and ADD command bits below are written as '1'.\tRO\t0\n"
               "bit 30 POP When this bit is written as a '1', the top value of the FTQ is popped\n"
               "  (removed). This should only be done if the top value has the Intervene bit\n"
               "  set as indicated in the ctl register. Writing this bit as a '0' has no effect.\n"
               "  The BUSY bit must be '0' when this bit is written to a '1'. This bit always\n"
               "  reads as '0'.\tSC\t0\n"
               "bit 29 INTERVENE_CLR When this bit is written as a '1', the intervene bit in the top value of the\n"
               "  FTQ is cleared. This should only be done if the top value has the Intervene\n"
               "  bit set as indicated in the ctl register. Writing this bit as a '0' has no\n"
               "  effect. Clearing the intervene bit will allow the value to be processed by\n"
               "  the TX DMA block normally. The BUSY bit must be a '0' when this bit is\n"
               "  written to a '1'. This bit always reads as '0'.\tSC\t0\n"
               "bit 28 ADD_DATA When this bit is written as a '1', the current data holding register is\n"
               "  written to the bottom of the FTQ. The WR_TOP bit must be cleared for this\n"
               "  to work. Writing this bit as a '0' has no effect. The BUSY bit must be a '0'\n"
               "  when this bit is written to a '1'. This bit always reads as a '0'.\tSC\t0\n"
               "bit 27 ADD_INTERVEN When this bit is written as a '1', the current data holding register is\n"
               "  written to the bottom of the FTQ just as when the ADD_DATA bit is set,\n"
               "  except that the intervene bit is set on the entry. The WR_TOP bit must be\n"
               "  cleared for this to work. Writing this bit as a '0' has no effect. The BUSY\n"
               "  bit must be a '0' when this bit is written to a '1'. This bit always reads as a\n"
               "  '0'.\tSC\t0\n"
               "bit 26 RD_DATA When this bit is written as a '1', the data holding register is updated with\n"
               "  the data from the FTQ entry pointed to by the OFFSET register. The BUSY\n"
               "  bit will remain set until the read is complete and the data area can be\n"
               "  read. Writing this bit as a '0' has no effect. The BUSY bit must be a '0'\n"
               "  when this bit is written to a '1'. This bit always reads as a '0'.\tSC\t0\n"
               "bit 25 SFT_RESET When this bit is written as a '1', the FTQ block will be reset. All data in the\n"
               "  FTQ will be lost. This bit is intended for diagnostic use only. Setting this\n"
               "  bit may cause down stream state machines to be confused because the\n"
               "  \"Valid\" output may be de-asserted unexpectedly.\tSC\t0\n"
               "bit 24-11 unused0\tRO\t0\n"
               "bit 10 WR_TOP This bit controls the operation of the exchanges between the FTQ and the\n"
               "  holding register (data).\tRW\t0\n"
               "  Value Name Description\n"
               "   0 0 When '0', RD_DATA commands, will read the top\n"
               "  entry in the FTQ to the holding register. ADD_DATA\n"
               "  and ADD_INTERVENE commands will write the\n"
               "  holding register to the bottom of the FTQ.\n"
               "   1 1 When '1', both reads and writes of the holding register\n"
               "  will access the entry as specified in the OFFSET value.\n"
               "  This should only be done while the chip is idle or an\n"
               "  intervene record is at the top of the queue. This mode\n"
               "  is used for either modifying the top of the queue or as\n"
               "  a debug tool for reading the complete queue.\n"
               "bit 9-0 OFFSET This value is used when the WR_TOP bit is set to '1' and controls the\n"
               "  offset into the FTQ that is being accessed for transfers between the FTQ\n"
               "  and the holding register. When this value is zero, the top of the queue\n"
               "  entry will be accessed. When this value is 1, the top-1 value is being\n"
               "  accessed. This value must never be programmed larger than the default\n"
               "  value of the MAX_DEPTH field.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_HMCPQ_FTQ_CTL:
        str =  "\nMCP_REG_MCPR_HMCPQ_FTQ_CTL - (Offset: 0x853BC; Width: 32)\n"
               "  SPLIT:\n"
               "bit 31-22 CUR_DEPTH This value indicates the current FTQ depth being used. A value of zero\n"
               "  indicates that the FTQ is empty.\tRO\t0\n"
               "bit 21-12 MAX_DEPTH This value controls the amount of the FTQ that will be available and\n"
               "  essentially shortens the queue depth. The default value is for maximum\n"
               "  queue depth. There are two reasons to shorten the queue. First, it can\n"
               "  decrease total chip latency. Second, it reserves some space in the queue\n"
               "  so that old entries can be viewed after a hang. This is because they will\n"
               "  not be written over as the system invariably backs-up. The default value\n"
               "  of this register the maximum number of entries that will fit into the FTQ.\tRW\tX\n"
               "bit 11-3 unused0\tRO\t0\n"
               "bit 2 FORCE_INTERVENE When this bit is '1', it will force the intervene bit to be set on each entry\n"
               "  written to the FTQ by any of the hardware interfaces. The register\n"
               "  control of the intervene bit on writes still works normally (See\n"
               "  ADD_INTERVENE). When this bit is '0', the intervene input no the\n"
               "  hardware interfaces works normally.\tRW\t0\n"
               "bit 1 OVERFLOW When this bit is set, it indicates that a request was made to write the\n"
               "  FTQ when is was at its programmed maximum depth. This is to indicate\n"
               "  that the chip was backed-up in this area at some time since the bit was\n"
               "  last cleared. Writing this bit to a '1' will clear the over-flow status.\n"
               "  Writing this bit as a '0' has no effect.\tWC\t0\n"
               "bit 0 INTERVENE This bit indicates the intervene status of the top entry in the FTQ. If set,\n"
               "  it indicates that the top entry has the intervene bit set and requires\n"
               "  firmware intervention.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_MCPQ_FTQ_CMD:
        str =  "\nMCP_REG_MCPR_MCPQ_FTQ_CMD - (Offset: 0x853F8; Width: 32)\n"
               "  SPLIT: FTQ control registers. The depth of each Management Processor FTQ is 32 records.\n"
               "bit 31 BUSY This bit reads as '1' when the FTQ controller is busy and can not respond\n"
               "  to any of the command bits below. This bit must be '0' when POP,\n"
               "  INTERVENE_CLR, and ADD command bits below are written as '1'.\tRO\t0\n"
               "bit 30 POP When this bit is written as a '1', the top value of the FTQ is popped\n"
               "  (removed). This should only be done if the top value has the Intervene bit\n"
               "  set as indicated in the ctl register. Writing this bit as a '0' has no effect.\n"
               "  The BUSY bit must be '0' when this bit is written to a '1'. This bit always\n"
               "  reads as '0'.\tSC\t0\n"
               "bit 29 INTERVENE_CLR When this bit is written as a '1', the intervene bit in the top value of the\n"
               "  FTQ is cleared. This should only be done if the top value has the Intervene\n"
               "  bit set as indicated in the ctl register. Writing this bit as a '0' has no\n"
               "  effect. Clearing the intervene bit will allow the value to be processed by\n"
               "  the TX DMA block normally. The BUSY bit must be a '0' when this bit is\n"
               "  written to a '1'. This bit always reads as '0'.\tSC\t0\n"
               "bit 28 ADD_DATA When this bit is written as a '1', the current data holding register is\n"
               "  written to the bottom of the FTQ. The WR_TOP bit must be cleared for this\n"
               "  to work. Writing this bit as a '0' has no effect. The BUSY bit must be a '0'\n"
               "  when this bit is written to a '1'. This bit always reads as a '0'.\tSC\t0\n"
               "bit 27 ADD_INTERVEN When this bit is written as a '1', the current data holding register is\n"
               "  written to the bottom of the FTQ just as when the ADD_DATA bit is set,\n"
               "  except that the intervene bit is set on the entry. The WR_TOP bit must be\n"
               "  cleared for this to work. Writing this bit as a '0' has no effect. The BUSY\n"
               "  bit must be a '0' when this bit is written to a '1'. This bit always reads as a\n"
               "  '0'.\tSC\t0\n"
               "bit 26 RD_DATA When this bit is written as a '1', the data holding register is updated with\n"
               "  the data from the FTQ entry pointed to by the OFFSET register. The BUSY\n"
               "  bit will remain set until the read is complete and the data area can be\n"
               "  read. Writing this bit as a '0' has no effect. The BUSY bit must be a '0'\n"
               "  when this bit is written to a '1'. This bit always reads as a '0'.\tSC\t0\n"
               "bit 25 SFT_RESET When this bit is written as a '1', the FTQ block will be reset. All data in the\n"
               "  FTQ will be lost. This bit is intended for diagnostic use only. Setting this\n"
               "  bit may cause down stream state machines to be confused because the\n"
               "  \"Valid\" output may be de-asserted unexpectedly.\tSC\t0\n"
               "bit 24-11 unused0\tRO\t0\n"
               "bit 10 WR_TOP This bit controls the operation of the exchanges between the FTQ and the\n"
               "  holding register (data).\tRW\t0\n"
               "  Value Name Description\n"
               "  0\t0 When '0', RD_DATA commands, will read the top\n"
               "  \tentry in the FTQ to the holding register. ADD_DATA\n"
               "  \tand ADD_INTERVENE commands will write the\n"
               "  \tholding register to the bottom of the FTQ.\n"
               "  1\t1 When '1', both reads and writes of the holding register\n"
               "  \twill access the entry as specified in the OFFSET value.\n"
               "  \tThis should only be done while the chip is idle or an\n"
               "  \tintervene record is at the top of the queue. This mode\n"
               "  \tis used for either modifying the top of the queue or as\n"
               "  \ta debug tool for reading the complete queue.\n"
               "bit 9-0 OFFSET This value is used when the WR_TOP bit is set to '1' and controls the\n"
               "  offset into the FTQ that is being accessed for transfers between the FTQ\n"
               "  and the holding register. When this value is zero, the top of the queue\n"
               "  entry will be accessed. When this value is 1, the top-1 value is being\n"
               "  accessed. This value must never be programmed larger than the default\n"
               "  value of the MAX_DEPTH field.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_MCPQ_FTQ_CTL:
        str =  "\nMCP_REG_MCPR_MCPQ_FTQ_CTL - (Offset: 0x853FC; Width: 32)\n"
               "  SPLIT:\n"
               "bit 31-22 CUR_DEPTH This value indicates the current FTQ depth being used. A value of zero\n"
               "  indicates that the FTQ is empty.\tRO\t0\n"
               "bit 21-12 MAX_DEPTH This value controls the amount of the FTQ that will be available and\n"
               "  essentially shortens the queue depth. The default value is for maximum\n"
               "  queue depth. There are two reasons to shorten the queue. First, it can\n"
               "  decrease total chip latency. Second, it reserves some space in the queue\n"
               "  so that old entries can be viewed after a hang. This is because they will\n"
               "  not be written over as the system invariably backs-up. The default value\n"
               "  of this register the maximum number of entries that will fit into the FTQ.\tRW\tX\n"
               "bit 11-3 unused0\tRO\t0\n"
               "bit 2 FORCE_INTERVENE When this bit is '1', it will force the intervene bit to be set on each entry\n"
               "  written to the FTQ by any of the hardware interfaces. The register\n"
               "  control of the intervene bit on writes still works normally (See\n"
               "  ADD_INTERVENE). When this bit is '0', the intervene input no the\n"
               "  hardware interfaces works normally.\tRW\t0\n"
               "bit 1 OVERFLOW When this bit is set, it indicates that a request was made to write the\n"
               "  FTQ when is was at its programmed maximum depth. This is to indicate\n"
               "  that the chip was backed-up in this area at some time since the bit was\n"
               "  last cleared. Writing this bit to a '1' will clear the over-flow status.\n"
               "  Writing this bit as a '0' has no effect.\tWC\t0\n"
               "bit 0 INTERVENE This bit indicates the intervene status of the top entry in the FTQ. If set,\n"
               "  it indicates that the top entry has the intervene bit set and requires\n"
               "  firmware intervention.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_NVM_COMMAND:
        str =  "\nMCP_REG_MCPR_NVM_COMMAND - (Offset: 0x86400; Width: 32)\n"
               "  The NVM block sits in the common area in Xinan. As with all common blocks, they may be addressed as either\n"
               "  function 0 or function 1. There are two types of resolution functions in NVM block.\n"
               "  SPLIT = bit has two registers. On is access when function 0 is accessed. A different value is access when\n"
               "  function 1 is accessed. These bits are written and read independently\n"
               "  SHARE = a single register is provided. Both function 0 and function 1 views see the exact same register.\n"
               "  Software has to be aware that what it reads may not be what it has written if the other port overwrites the\n"
               "  previously written value.\n"
               "  SHARE: NVM Command register.\n"
               "bit 31-23 unused3\tRO\t0\n"
               "bit 22 MODE_256 The 256B page size mode disable bit. A 256 byte page mode has been\n"
               "  added to the block. This mode is normally on. The mode helps convert a\n"
               "bit 264B page Atmel part to act more like a 256B page part. For reads, the\n"
               "  controller transparently closes the page after byte location 0xFF and opens\n"
               "  the next page. For writes, it is the FW or SW responsiblity to close the\n"
               "  page at 0xFF and start a new operation on the next page. When this bit is\n"
               "  written as '1' when the FIRST bit is set, the 256B page mode is disabled\n"
               "  for the next operation. It is self-clearing when both the LAST bit is set and\n"
               "  the DONE bit is asserted. Effects Atmel only. No effect with ST devices.\tSC\t0\n"
               "bit 21 RD_STATUS The read status command bit. When set, the flash controller will read the\n"
               "  status register from the external flash device\tRW\t0\n"
               "bit 20 RD_ID The read ID command bit. When set, the flash controller will read the ID\n"
               "  register from the external flash device. This is specifically for ST devices.\n"
               "  Setting this bit for Atmel devices will give the same results as RD_STATUS.\tRW\t0\n"
               "bit 19-18 unused2\tRO\t0\n"
               "bit 17 WRDI The write disable command bit. Set '1' will make flash interface state\n"
               "  machine Generate wrdi_cmd to flash device through SPI interface to set\n"
               "  Flash device to be write-disabled. Used for the device with protection\n"
               "  function.\tRW\t0\n"
               "bit 16 WREN The write enable command bit. Set '1' will make flash interface state\n"
               "  machine Generate wren_cmd to flash device through SPI interface to set\n"
               "  Flash device to be write-enabled. Used for the device with protection\n"
               "  function\tRW\t0\n"
               "bit 15-9 unused1\tRO\t0\n"
               "bit 8 LAST When this bit is set, the next command sequence will be interpreted as\n"
               "  the last one of a burst and any cleanup work will be done. This means that\n"
               "  the buffer will be written to flash memory if needed on a write.\tRW\t0\n"
               "bit 7 FIRST This bit is passed to the SEE_FSM or SPI_FSM if the pass_mode bit is set.\tRW\t0\n"
               "bit 6 ERASE The erase command bit. Set high to execute an erase. This bit is ignored if\n"
               "  the wr is clear.\tRW\t0\n"
               "bit 5 WR The Write/Not_Read command bit. Set high to execute write or erase.\tRW\t0\n"
               "bit 4 DOIT Command from software to start the defined command. The done bit must\n"
               "  be clear before setting this bit. This bit is self clearing and will remain set\n"
               "  while the command is active.\tSC\t0\n"
               "bit 3 DONE Sequence completion bit that is asserted when the command requested by\n"
               "  assertion of the doit bit has completed. done Will be cleared while the\n"
               "  command is in progress. done Will stay asserted until doit is reasserted or\n"
               "  the done bit is cleared by writing a 1 to the done bit. The done bit is the\n"
               "  FLSH_ATTN signal.\tWC\t0\n"
               "bit 2-1 unused0\tRO\t0\n"
               "bit 0 RST When set, the entire NVM state machine is reset. This bit is self clearing. SC 0\n";
        break;
    case MCP_REG_MCPR_NVM_STATUS:
        str =  "\nMCP_REG_MCPR_NVM_STATUS - (Offset: 0x86404; Width: 32)\n"
               "  SHARE: NVM Status register.\n"
               "bit 31-5 unused0\tRO\t0\n"
               "bit 4-0 SPI_FSM_STATE Enumeration:\tRO\t0\n"
               "  Value Name Description\n"
               "   0 SPI_IDLE\n"
               "   1 SPI_CMD0\n"
               "   2 SPI_CMD1\n"
               "   3 SPI_CMD_FINISH0\n"
               "   4 SPI_CMD_FINISH1\n"
               "   5 SPI_ADDR0\n"
               "   6 SPI_WRITE_DATA0\n"
               "   7 SPI_WRITE_DATA1\n"
               "   8 SPI_WRITE_DATA2\n"
               "   9 SPI_READ_DATA0\n"
               "   10 SPI_READ_DATA1\n"
               "   11 SPI_READ_DATA2\n"
               "   12 SPI_READ_STATUS_RDID0\n"
               "   13 SPI_READ_STATUS_RDID1\n"
               "   14 SPI_READ_STATUS_RDID2\n"
               "   15 SPI_READ_STATUS_RDID3\n"
               "   16 SPI_READ_STATUS_RDID4\n"
               "   17 SPI_CHECK_BUSY0\n"
               "   18 SPI_ST_WREN\n"
               "   19 SPI_WAIT\n";
        break;
    case MCP_REG_MCPR_NVM_WRITE:
        str =  "\nMCP_REG_MCPR_NVM_WRITE - (Offset: 0x86408; Width: 32)\n"
               "  SHARE: NVM data write register.\n"
               "bit 31-0 NVM_WRITE_VALUE 32 bits of write data use when write commands are executed. When\n"
               "  bitbang_mode is set, only the bottom 6 bits have any effect as\n"
               "  described below.\tRW\t0\n"
               "  Value Name Description\n"
               "   0 BIT_BANG When in bit-bang mode, the bottom 6 bits of this\n"
               "  register control the output value of the external\n"
               "  pins.\n"
               "   1 SI When in bit-bang mode, this bit controls the SI\n"
               "  output value.\n"
               "   2 SO When in bit-bang mode, this bit controls the SO\n"
               "  output value.\n"
               "   4 CS_B When in bit-bang mode, this bit controls the CS_B\n"
               "  output value.\n"
               "   8 SCLK When in bit-bang mode, this bit controls the SCLK\n"
               "  output value.\n";
        break;
    case MCP_REG_MCPR_NVM_ADDR:
        str =  "\nMCP_REG_MCPR_NVM_ADDR - (Offset: 0x8640C; Width: 32)\n"
               "  SHARE: NVM address register.\n"
               "bit 31-24 unused0\tRO\t0\n"
               "bit 23-0 NVM_ADDR_VALUE 24 bit address value used in read, write and erase operations. When in\n"
               "  bit-bang mode, the bottom 6 bits control the output enable for each pin.\tRW\t0\n"
               "  Value Name Description\n"
               "   0 BIT_BANG When in bit-bang mode, the bottom 6 bits of this\n"
               "  register control the output enable value of the\n"
               "  external pins.\n"
               "   1 SI When in bit-bang mode, this bit controls the enables\n"
               "  the SI pin as an output when '0'.\n"
               "   2 SO When in bit-bang mode, this bit controls the enables\n"
               "  the SO pin as an output when '0'.\n"
               "   4 CS_B When in bit-bang mode, this bit controls the enables\n"
               "  the CS_B pin as an output when '0'.\n"
               "   8 SCLK When in bit-bang mode, this bit controls the enables\n"
               "  the SCLK pin as an output when '0'.\n";
        break;
    case MCP_REG_MCPR_NVM_READ:
        str =  "\nMCP_REG_MCPR_NVM_READ - (Offset: 0x86410; Width: 32)\n"
               "  SHARE: NVM data read register.\n"
               "bit 31-0 NVM_READ_VALUE 32 bits of read data use when read commands are executed. When\n"
               "  bitbang_mode is set, bits 0 to 5 reflect the current input value of the\n"
               "  external pins.\tRO\t0\n"
               "  Value Name Description\n"
               "   0 BIT_BANG When in bit-bang mode, the bottom 6 bits of this\n"
               "  return the input value of the external pins.\n"
               "   1 SI When in bit-bang mode, this bit returns the current\n"
               "  value of the SI pin.\n"
               "   2 SO When in bit-bang mode, this bit returns the current\n"
               "  value of the SO pin.\n"
               "   4 CS_B When in bit-bang mode, this bit returns the current\n"
               "  value of the CS_B pin.\n"
               "   8 SCLK When in bit-bang mode, this bit returns the current\n"
               "  value of the SCLK pin.\n";
        break;
    case MCP_REG_MCPR_NVM_CFG1:
        str =  "\nMCP_REG_MCPR_NVM_CFG1 - (Offset: 0x86414; Width: 32)\n"
               "  SHARE: NVM configuration one register.\n"
               "bit 31 COMPAT_BYPASSS Legacy bit. Acts as dummy R/W bit.\tRW\t0\n"
               "bit 30 FW_FLASH_TYPE_EN Legacy strap_control[1] bit. Read only set to 1, indicating FLASH has\n"
               "  already been configured.This register has no hardware function, but can\n"
               "  be modified by firmware.\tRO\t0x1\n"
               "bit 29 FW_USTRAP_3 Legacy strap_value[3]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 28 FW_USTRAP_2 Legacy strap_value[2]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 27 FW_USTRAP_0 Legacy strap_value[0]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 26 FW_USTRAP_1 Legacy strap_value[1]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 25 FLASH_SIZE Legacy strap_value[3]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 24 PROTECT_MODE Legacy strap_value[2]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 23 STRAP_CONTROL_0 Legacy strap_control[1] bit. Read only set to 1, indicating FLASH has\n"
               "  already been configured.\tRO\t0x1\n"
               "bit 22 unused0\tRO\t0\n"
               "bit 21-11 SEE_CLK_DIV Legacy value. Read only.\tRO\t0x60\n"
               "bit 10-7 SPI_CLK_DIV Divisor used to create all \"1x\" time for all Flash Interface I/O pin timing\n"
               "  definition A divisor of 0 means that an SCLK will transition at minimum\n"
               "  of each CORE_CLK rising edge. The equation to calculate the clock\n"
               "  frequency for SCLK is :CORE_CLK frequency/((spi_clk_div +1) *2).\tRW\t0x4\n"
               "bit 6-4 STATUS_BIT Bit offset in status command response to interpret as the \"ready\" flag.\n"
               "  For Atmel, this defaults to 3'h7. For ST, this defaults to 3'h0. NOTE: For\n"
               "  ST, the status value of 1'b0 means \"ready\". For Atmel, the status value\n"
               "  of 1 means \"ready\". This is automatically interpreted by hardware. This\n"
               "  value is self-configured on reset based on the strap values. It can be\n"
               "  overriden.\tRW\tX\n"
               "bit 3 BITBANG_MODE Enable bit-bang mode to control pins.\tRW\t0\n"
               "bit 2 PASS_MODE Enable pass-thru mode to the byte level SPI state machine. When this\n"
               "  mode is enabled, the controller can send/recieve single bytes at the\n"
               "  SPI level. All upper level functions of the controller state machine are\n"
               "  disabled.\tRW\t0\n"
               "bit 1 BUFFER_MODE Legacy strap_value[0]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel or ST.\tRO\tX\n"
               "bit 0 FLASH_MODE Legacy strap_value[1]. Read only. Set based on new strap values to\n"
               "  indicate either Atmel o\tRO\tX\n";
        break;
    case MCP_REG_MCPR_NVM_CFG2:
        str =  "\nMCP_REG_MCPR_NVM_CFG2 - (Offset: 0x86418; Width: 32)\n"
               "  SHARE: NVM configuration two register.\n"
               "bit 31-24 READ_ID Flash Read ID register command. This command is used to read the ID\n"
               "  register from ST devices. Reset value depends on strap values. (0x57 for\n"
               "  Atmel; 0x9f for ST)\tRW\t0\n"
               "bit 23-16 STATUS_CMD Flash status command. This command is used to poll the \"ready\" status of\n"
               "  the flash part after many of the commands. Reset value is 0x9Fh if\n"
               "  flash_mode=1, 0x57h if buffer_mode=1, and 0x5h if protect_mode =1.\tRW\t0\n"
               "bit 15-8 unused0\tRO\t0\n"
               "bit 7-0 ERASE_CMD Flash block erase command. \"ready\" status will be polled for after this\n"
               "  command. Reset value is 0x20h if flash_mode=1, 0x81h if buffer_mode=1,\n"
               "  and 0xd8h if protect_mode=1.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_NVM_CFG3:
        str =  "\nMCP_REG_MCPR_NVM_CFG3 - (Offset: 0x8641C; Width: 32)\n"
               "  SHARE: NVM configuration three register.\n"
               "bit 31-24 READ_CMD This is the flash/seeprom read command. Following this command, any\n"
               "  number of bytes may be read up to the end of the flash memory. For\n"
               "  SEEPROM (flash_mode=0), this is SEEPROM read command. Bit[26:25] is\n"
               "  address bit A1 and A0 of SEEPROM. User should modify those two bits\n"
               "  base on the value of A1 and A0 assigned to this SEEPROM device. Reset\n"
               "  value is 0xFF if flash_mode=1, 0x68 if buffer_mode=1, 0x3 if\n"
               "  protect_mode=1, and 0xA1 otherwise.\tRW\t0\n"
               "bit 23-16 unused0\tRO\t0\n"
               "bit 15-8 WRITE_CMD Command to write one byte to the flash array or SSRAM buffer, depending\n"
               "  on the value of buffer_mode. If BUFFER_MODE is not active, then this\n"
               "  command will poll for \"ready\" status when complete. For SEEPROM\n"
               "  (flash_mode=0), this is SEEPROM write command. Bit[10:9] is address bit\n"
               "  A1 and A0 of SEEPROM. User should modify those two bits base on the\n"
               "  value of A1 and A0 assigned to this SEEPROM device. Reset value is 0x10\n"
               "  if flash_mode=1, 0x83 if buffer_mode=1, and 0x2 if protect_mode=1,\n"
               "  0xA0 otherwise.\tRW\t0\n"
               "bit 7-0 BUFFER_RD_CMD Transfer flash device page to its internal buffer command. For Atmel\n"
               "  devices, this command is issued automatically upon the \"FIRST\" write. It\n"
               "  is not issued for ST devices since they automatically do this operation\n"
               "  internally. (0x53 for Atmel; 0x53 for ST)\tRW\t0\n";
        break;
    case MCP_REG_MCPR_NVM_SW_ARB:
        str =  "\nMCP_REG_MCPR_NVM_SW_ARB - (Offset: 0x86420; Width: 32)\n"
               "  SPLIT: NVM arbitration register. This register provides aribtration resoures for both functions. The register is split,\n"
               "  so arbitration depends on what function you are accessing as.\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15 REQ3 This is the current status of requester 3. When this bit is one, it means\n"
               "  that REQ_SET3 has been set since REQ_CLR3.\tRO\t0\n"
               "bit 14 REQ2 This is the current status of requester 2. When this bit is one, it means\n"
               "  that REQ_SET2 has been set since REQ_CLR2.\tRO\t0\n"
               "bit 13 REQ1 This is the current status of requester 1. When this bit is one, it means\n"
               "  that REQ_SET1 has been set since REQ_CLR1.\tRO\t0\n"
               "bit 12 REQ0 This is the current status of requester 0. When this bit is one, it means\n"
               "  that REQ_SET0 has been set since REQ_CLR0.\tRO\t0\n"
               "bit 11 ARB_ARB3 when REQ3 arbitration is won, this bit will be read as 1, when an operation\n"
               "  is complete, then the CLR_ARB3 must be written to clear this bit.\tRO\t0\n"
               "bit 10 ARB_ARB2 when REQ2 arbitration is won, this bit will be read as 1, when an operation\n"
               "  is complete, then the CLR_ARB2 must be written to clear this bit.\tRO\t0\n"
               "bit 9 ARB_ARB1 when REQ1 arbitration is won, this bit will be read as 1, when an operation\n"
               "  is complete, then the CLR_ARB1 must be written to clear this bit.\tRO\t0\n"
               "bit 8 ARB_ARB0 when REQ0 arbitration is won, this bit will be read as 1, when an operation\n"
               "  is complete, then the CLR_ARB0 must be written to clear this bit. At that\n"
               "  point, the next Arb bit will read as 1. At any time, only one of the\n"
               "  ARB[7:0] bits will be read as a 1. Arb0 has highest priority, and Arb7 has\n"
               "  lowest priority.\tRO\t0\n"
               "bit 7 ARB_REQ_CLR3 Write this bit as a '1' to clear req3 bit. SC 0\n"
               "bit 6 ARB_REQ_CLR2 Write this bit as a '1' to clear req2 bit. SC 0\n"
               "bit 5 ARB_REQ_CLR1 Write this bit as a '1' to clear req1 bit. SC 0\n"
               "bit 4 ARB_REQ_CLR0 Write this bit as a '1' to clear req0 bit. SC 0\n"
               "bit 3 ARB_REQ_SET3 Set Software Arbitration request Bit 3. This bit is set by writing a '1' to this\n"
               "  bit position.\tSC\t0\n"
               "bit 2 ARB_REQ_SET2 Set Software Arbitration request Bit 2. This bit is set by writing a '1' to this\n"
               "  bit position.\tSC\t0\n"
               "bit 1 ARB_REQ_SET1 Set Software Arbitration request Bit 1. This bit is set by writing a '1' to this\n"
               "  bit position.\tSC\t0\n"
               "bit 0 ARB_REQ_SET0 Set Software Arbitration request Bit 0. This bit is set by writing a '1' to this\n"
               "  bit position.\tSC\t0\n";
        break;
    case MCP_REG_MCPR_NVM_ACCESS_ENABLE:
        str =  "\nMCP_REG_MCPR_NVM_ACCESS_ENABLE - (Offset: 0x86424; Width: 32)\n"
               "  SHARE: NVM access enable register.\n"
               "bit 31-2 unused0\tRO\t0\n"
               "bit 1 WR_EN This bit is no longer required in Xinan. Write enable to flash is always 1.\tRO\t0x1\n"
               "bit 0 EN This bit is no longer required in Xinan. Access to flash is always enabled.\tRO\t0x1\n";
        break;
    case MCP_REG_MCPR_NVM_WRITE1:
        str =  "\nMCP_REG_MCPR_NVM_WRITE1 - (Offset: 0x86428; Width: 32)\n"
               "  SHARE: NVM write1 configuration register.\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-8 WRDI_CMD Flash write disable command when device with protection function is used.\n"
               "  This command will be issued by the flash interface state machine through\n"
               "  SPI interface To flash device, and make the flash device write-disabled.\tRW\t0\n"
               "bit 7-0 WREN_CMD Flash write enable command when device with protection function is used.\n"
               "  This command will be issued by the flash interface state machine through\n"
               "  SPI interface To flash device, and make the flash device write-enabled.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_NVM_CFG4:
        str =  "\nMCP_REG_MCPR_NVM_CFG4 - (Offset: 0x8642C; Width: 32)\n"
               "  SHARE: NVM configuration four register.\n"
               "bit 31-7 RESERVED Reserved for future use\tRW\t0\n"
               "bit 6 STATUS_BIT_POLARITY This bit determines how the status bit of the device status\n"
               "  register is interpreted by hardware. If 0, then 0 means \"ready\".\n"
               "  If 1, then 1 means \"ready\". For Atmel, this defaults to 1. For ST,\n"
               "  this defaults to 0. This value is self-configured on reset based\n"
               "  on the strap values. It can be overriden.\tRW\tX\n"
               "bit 5-4 MODE_256_EMPTY_BIT_LOC Bit location for hardware to insert an empty address bit when\n"
               "  MODE_256 is not set with Atmel devices. This value is\n"
               "  self-configured on reset based on the external device. NOTE:\n"
               "  Max Atmel device size is 64 Mbit due to different bus protocols.\tRW\tX\n"
               "  Value Name Description\n"
               "   0 BIT8 For 264 Byte pages\n"
               "   1 BIT9 For 528 Byte pages\n"
               "   2 BIT10 For 1056 Byte pages\n"
               "   3 BIT11 For 2112 Byte pages\n"
               "   3 FLASH_VENDOR This bit is self-configured on reset based on the strap values. It\n"
               "  can be overriden.\tRW\tX\n"
               "  Value Name Description\n"
               "   0 ST\n"
               "   1 ATMEL\n"
               "   2-0 FLASH_SIZE Size of the external flash device. This information is not used by\n"
               "  FLSH hardware. It is only used by software. This value is\n"
               "  self-configured on reset based on the external device.\tRW\tX\n"
               "  Value Name Description\n"
               "   0 1MBIT\n"
               "   1 2MBIT\n"
               "   2 4MBIT\n"
               "   3 8MBIT\n"
               "   4 16MBIT\n"
               "   5 32MBIT\n"
               "   6 64MBIT\n"
               "   7 128MBIT\n";
        break;
    case MCP_REG_MCPR_NVM_RECONFIG:
        str =  "\nMCP_REG_MCPR_NVM_RECONFIG - (Offset: 0x86430; Width: 32)\n"
               "  SHARE: NVM re-configuration register.\n"
               "bit 31 RECONFIG_DONE This bit is 0 on reset. After software finishes reconfiguring FLSH,\n"
               "  they will set this bit to 1 to indicate that FLSH has been\n"
               "  reconfigured. This bit has no hardware functionality.\tRW\t0\n"
               "bit 30-8 RESERVED Reserved for future use\tRW\t0\n"
               "bit 7-4 RECONFIG_STRAP_VALUE Used by software to numerically encode how the FLSH has been\n"
               "  reconfigured. On reset, this register is set to the same value as\n"
               "  ORIG_STRAP_VALUE. These bits have no hardware functionality.\tRW\tX\n"
               "bit 3-0 ORIG_STRAP_VALUE Strap value from iologic pins. Only bit[0] is used. Bits[3:1] are for\n"
               "  future support.\tRO\tX\n"
               "  Value Name Description\n"
               "  0\t0  ST\n"
               "  1\t1  ATMEL\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_CTRL:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_CTRL - (Offset: 0x86800; Width: 32)\n"
               "bit 31 ACTIVE This bit is set to '1' when the expansion ROM engine is actively operating\n"
               "  on a expansion ROM request.\tRO\t0\n"
               "bit 30 READ_TIMEOUT This bit is set to '1' when a read timeout happens.\tWC\t0\n"
               "bit 29 ARB_TIMEOUT This bit is set to '1' when an arbitration timeout happens.\tWC\t0\n"
               "bit 28 CACHE_VALID This bit is set to '1' when the cache is valid.\tWC\t0\n"
               "bit 27-22 unused2\tRO\t0\n"
               "bit 21-16 STATE The current state of expansion ROM engine, for debugging purposes.\tRO\t0\n"
               "bit 15-6 unused1\tRO\t0\n"
               "bit 5-4 ARB_NUM Request number to use to arbitrate for expansion ROM access to the flash\n"
               "  controller\tRW\t0x3\n"
               "bit 3-2 unused0\tRO\t0\n"
               "bit 1 BFRD When this bit is set to '1', the expansion ROM engine will utilize the\n"
               "  buffered mode address translation mode in the flash controller to adjust\n"
               "  the address for flash devices with 264 byte blocks, spaced every 512\n"
               "  bytes.\tRW\t0\n"
               "bit 0 ENA Enable bit for the expansion ROM engine. When '1', the expansion ROM\n"
               "  engine will automatically service expansion ROM requests. When this bit is\n"
               "  cleared, the engine will not service a new request and the on chip cpu will\n"
               "  have to take over the chores.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_BADDR:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_BADDR - (Offset: 0x86804; Width: 32)\n"
               "bit 31-24 unused1\tRO\t0\n"
               "bit 23-2 VALUE Image base address. The expansion ROM engine fetches the values for the\n"
               "  expansion ROM interface starting at this base address.\tRW\t0\n"
               "bit 1-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_CFG:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_CFG - (Offset: 0x86808; Width: 32)\n"
               "bit 31-8 unused0\tRO\t0\n"
               "bit 7-4 READ_TIMEOUT_SHFT Read timeout value. The value of 0x20 is shifted left number of bits\n"
               "  determined by this value to determine the number of core clocks to\n"
               "  wait for the expansion ROM interface to complete a single word read\n"
               "  before giving up.\tRW\t0xf\n"
               "bit 3-0 ARB_TIMEOUT_SHFT Arbitration timeout value. The value of 0x20 is shifted left number of\n"
               "  bits determined by this value to determine the number of times the\n"
               "  nvm_sw_arb register inside the expansion ROM interface will be polled\n"
               "  before giving up on arbitration.\tRW\t0xf\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_ADR:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_ADR - (Offset: 0x8680C; Width: 32)\n"
               "bit 31 REQ This bit will be set if there is a pending request for action. This bit is\n"
               "  equivalent to the EXP_ROM_ATTN attention word bit.\tRO\t0\n"
               "bit 30-26 ACT_FUNC These bits indicate which function is currently accessing the expansion\n"
               "  Rom.\tRO\t0\n"
               "bit 25-24 ADDR_SIZE The size of the PCI BAR rom read request. This value ranges from 1 to 3\n"
               "  dwords\tRO\t0\n"
               "bit 23-2 ADDRESS This value is the address requested by the current (or last) PCI ROM\n"
               "  Expansion cycle request. When the PCI block detect a decode of the ROM\n"
               "  BAR area, it will place the offset from the BAR value in this register and\n"
               "  re-try the PCI bus to make the requester wait.\tRO\t0\n"
               "bit 1-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_DATA0:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_DATA0 - (Offset: 0x86810; Width: 32) -\n"
               "  (RO/Reset: 0)\n"
               "  This register shows the first dword for the expansion ROM access. This is for debug purposes only.\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_DATA1:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_DATA1 - (Offset: 0x86814; Width: 32) -\n"
               "  (RO/Reset: 0)\n"
               "  This register shows the second dword for the expansion ROM access. This is for debug purposes only.\n";
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_DATA2:
        str =  "\nMCP_REG_MCPR_ERNGN_EXP_ROM_DATA2 - (Offset: 0x86818; Width: 32) -\n"
               "  (RO/Reset: 0)\n"
               "  This register shows the third dword for the expansion ROM access. This is for debug purposes only.\n";
        break;
    case MCP_REG_MCPR_SMBUS_CONFIG:
        str =  "\nMCP_REG_MCPR_SMBUS_CONFIG - (Offset: 0x88000; Width: 32)\n"
               "  SHARE: All registers in the SMB block are shared.\n"
               "bit 31 RESET When this bit is set it will reset SMBUS block to its default state. SC 0\n"
               "bit 30 SMB_EN When this bit is '1', the SMBUS block is enabled for operation. When\n"
               "  set the SMBUS block will abort current transaction in compliance\n"
               "  with the SMBUS master and slave behavior at the end of the current\n"
               "  transaction and stop responding to the SMBUS master/slave\n"
               "  transactions.\tRW\t0\n"
               "bit 29 BIT_BANG_EN When this bit is '1', the SMBUS block is placed into bit-bang mode.\n"
               "  SMBUS interface pins are controlled using Bit-Bang Control Register.\tRW\t0\n"
               "bit 28 EN_NIC_SMB_ADDR_0 When this bit is '1' the SMBUS block responds to slave address\n"
               "  7'b0000000.\tRW\t0\n"
               "bit 27 PROMISCOUS_MODE When this bit is '1' the SMBUS block responds to all SMBUS\n"
               "  transactions regardless of the slave address.\tRW\t0\n"
               "bit 26 TIMESTAMP_CNT_EN When this bit is '1' the TIMESTAMP counter is enabled. When '0' the\n"
               "  counter holds its value.\tRW\t0\n"
               "bit 25-20 unused2\tRO\t0\n"
               "bit 19-16 MASTER_RTRY_CNT This bit indicates a number of retries in case where SMBUS block\n"
               "  acted as a master and lost SMBUS arbitration. HW will retry\n"
               "  transaction as many times as specified in this register before it\n"
               "  reports lost of arbitration status to the firmware. When this field is\n"
               "  0 firmware will not do any retries but it will report loss of arbitration\n"
               "  on the initial attempt.\tRW\t0xf\n"
               "bit 15-10 unused1\tRO\t0\n"
               "bit 9 ARP_EN1 When this bit is set the SMBUS block will respond to ARP that is it\n"
               "  to SMBUS Device Default Address (7'b1100001) and reslove\n"
               "  NIC_SMB_ADDR1 using ARP.\tRW\t0\n"
               "bit 8 ARP_EN0 When this bit is set the SMBUS block will respond to ARP that is it\n"
               "  to SMBUS Device Default Address (7'b1100001) and reslove\n"
               "  NIC_SMB_ADDR0 using ARP.\tRW\t0\n"
               "bit 7 HW_ARP_ASSIGN_ADDR When this bit is set HW will service the ARP Assign Address\n"
               "  command, set the AR_FLAG[1:0] and AV_FLAG[1:0] flags, and\n"
               "  program the NIC_SMB_ADDR[1:0] values.\tRW\t0\n"
               "bit 6-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_TIMING_CONFIG:
        str =  "\nMCP_REG_MCPR_SMBUS_TIMING_CONFIG - (Offset: 0x88004; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 MODE_400 When this bit is set the SMBUS block operates in 400KHz mode.\n"
               "  When cleared SMBUS operates in 100KHz mode.\tRW\t0\n"
               "bit 30-24 RANDOM_SLAVE_STRETCH These bits specify time for which clock low time will be stretched\n"
               "  after each byte (that is ACK bit) when the SMBUS block acts as a\n"
               "  slave. This is useful in \"legacy mode\" to allow firmware time to\n"
               "  handle the data. Note that this time contributes to the slave\n"
               "  TLOW:SEXT time, that is combined random and periodic slave\n"
               "  stretch should not exceed 25ms. Register has 1ms resolution.\n"
               "  Default is 25ms.\tRW\t0x19\n"
               "bit 23-16 PERIODIC_SLAVE_STRETCH These bits specify time for which each clock period low time will\n"
               "  be stretched when the SMBUS block acts as a slave. Note that a\n"
               "  cumulative clock low extend time (TLOW:SEXT) for which slave\n"
               "  device is allowed to stretch the clock from the beginning to end\n"
               "  of the message (that is from START to STOP) is 25ms. For\n"
               "  example, if the message is Block Write transaction 36B long\n"
               "  allowed periodic stretch would be 25ms/(9*36) ~= 77us. This is\n"
               "  assuming that random slave stretching is not used. Register has\n"
               "  1us resolution.\tRW\t0\n"
               "bit 15-8 SMBUS_IDLE_TIME These bits specify the time for which both SMBCLK and SMBDAT\n"
               "  must be high before a master can assume that bus is free.\n"
               "  Register has 1us resolution. Default is 50us.\tRW\t0x32\n"
               "bit 7-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_ADDRESS:
        str =  "\nMCP_REG_MCPR_SMBUS_ADDRESS - (Offset: 0x88008; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 EN_NIC_SMB_ADDR3 When this bit is '1' NIC_SMB_ADDR3 will be used as a slave address to\n"
               "  match for incoming messages.\tRW\t0\n"
               "bit 30-24 NIC_SMB_ADDR3 This is the fourth SM Bus addresses which will be used to match for\n"
               "  incoming messages. When the PROMISCOUS_MODE bit is '1', this value\n"
               "  is ignored.\tRW\t0\n"
               "bit 23 EN_NIC_SMB_ADDR2 When this bit is '1' NIC_SMB_ADDR2 will be used as a slave address to\n"
               "  match for incoming messages.\tRW\t0\n"
               "bit 22-16 NIC_SMB_ADDR2 This is the third SM Bus addresses which will be used to match for\n"
               "  incoming messages. When the PROMISCOUS_MODE bit is '1', this value\n"
               "  is ignored. 0x0 This address is also used for ARP. The address will be\n"
               "  resolved using ARP when the ARP_EN1 bit is set.\tRW\t0\n"
               "bit 15 EN_NIC_SMB_ADDR1 When this bit is '1' NIC_SMB_ADDR1 will be used as a slave address to\n"
               "  match for incoming messages.\tRW\t0\n"
               "bit 14-8 NIC_SMB_ADDR1 This is the second SM Bus addresses which will be used to match for\n"
               "  incoming messages. When the PROMISCOUS_MODE bit is '1', this value\n"
               "  is ignored.\tRW\t0\n"
               "bit 7 EN_NIC_SMB_ADDR0 When this bit is '1' NIC_SMB_ADDR0 will be used as a slave address to\n"
               "  match for incoming messages.\tRW\t0\n"
               "bit 6-0 NIC_SMB_ADDR0 This is the first SM Bus addresses which will be used to match for\n"
               "  incoming messages. This address is also used for ARP that is this is\n"
               "  the address that will be resolved using ARP when the ARP function is\n"
               "  enabled. When the PROMISCOUS_MODE bit is '1', this value is ignored.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_FIFO_CONTROL:
        str =  "\nMCP_REG_MCPR_SMBUS_MASTER_FIFO_CONTROL - (Offset: 0x8800C; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 MASTER_RX_FIFO_FLUSH When this bit is set HW will flush Master RX FIFO when the\n"
               "  current RX transaction completes.\tSC\t0\n"
               "bit 30 MASTER_TX_FIFO_FLUSH When this bit is set HW will flush Master TX FIFO when\n"
               "  current TX transaction completes.\tSC\t0\n"
               "bit 29-23 unused2\tRO\t0\n"
               "bit 22-16 MASTER_RX_PKT_COUNT Number of packets in the Master RX FIFO.\tRO\t0\n"
               "bit 15 unused1\tRO\t0\n"
               "bit 14-8 MASTER_RX_FIFO_THRESHOLD When the Master Receive FIFO hits this threshold the SMBUS\n"
               "  block will generate an event for the control processor. When\n"
               "  set to 0x0 event generation is disabled. Threshold is\n"
               "  specified with the byte resolution.\tRW\t0\n"
               "bit 7-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_FIFO_CONTROL:
        str =  "\nMCP_REG_MCPR_SMBUS_SLAVE_FIFO_CONTROL - (Offset: 0x88010; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 SLAVE_RX_FIFO_FLUSH When this bit is set HW will flush Slave RX FIFO when the\n"
               "  current RX transaction completes.\tSC\t0\n"
               "bit 30 SLAVE_TX_FIFO_FLUSH When this bit is set HW will flush Slave TX FIFO when current\n"
               "  TX transaction completes.\tSC\t0\n"
               "bit 29-23 unused2\tRO\t0\n"
               "bit 22-16 SLAVE_RX_PKT_COUNT Number of packets in the Slave RX FIFO.\tRO\t0\n"
               "bit 15 unused1\tRO\t0\n"
               "bit 14-8 SLAVE_RX_FIFO_THRESHOLD When the Slave Receive FIFO hits this threshold the SMBUS\n"
               "  block will generate an event for the control processor. When\n"
               "  set to 0x0 event generation is disabled. Threshold is specified\n"
               "  with the byte resolution.\tRW\t0\n"
               "bit 7-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_BIT_BANG_CONTROL:
        str =  "\nMCP_REG_MCPR_SMBUS_BIT_BANG_CONTROL - (Offset: 0x88014; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 SMBCLK_IN This bit reflects the current input value of the SMBCLK pin. When the\n"
               "  SMBCLK pin is high, this bit will read as '1'. When the SMBCLK pin is low,\n"
               "  this pin will read as '0'.\tRO\t0\n"
               "bit 30 SMBCLK_OUT_EN When the SM Bus interface is configured for bit-bang mode, this bit\n"
               "  controlls the output enable for the CLK pin. When this bit is '0', the CLK\n"
               "  pin will drive low. When this bit is '1', the CLK pin will float.\tRW\t0x1\n"
               "bit 29 SMBDAT_IN This bit reflects the current input value of the SMBDAT pin. When the\n"
               "  SMBDAT pin is high, this bit will read as '1'. When the SMBDAT pin is low,\n"
               "  this pin will read as '0'.\tRO\t0\n"
               "bit 28 SMBDAT_OUT_EN When the SMBUS interface is configured for bit-bang mode, this bit\n"
               "  controlls the output enable for the SMBDAT pin. When this bit is '0', the\n"
               "  SMBDAT pin will drive low. When this bit is '1', the SMBDAT pin will float.\tRW\t0x1\n"
               "bit 27-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_WATCHDOG:
        str =  "\nMCP_REG_MCPR_SMBUS_WATCHDOG - (Offset: 0x88018; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 WATCHDOG This value counts down to zero once each second and sets the WG_TO bit\n"
               "  when it reaches zero. The counter stops when it reaches zero.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_HEARTBEAT:
        str =  "\nMCP_REG_MCPR_SMBUS_HEARTBEAT - (Offset: 0x8801C; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 HEARTBEAT This value counts down to zero once each second and sets the HB_TO bit\n"
               "  when it reaches zero. The counter stops when it reaches zero.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_POLL_ASF:
        str =  "\nMCP_REG_MCPR_SMBUS_POLL_ASF - (Offset: 0x88020; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 POLL_ASF This value counts down to zero once each 5 msec. and sets the PA_TO bit\n"
               "  when it reaches zero. The counter stops when it reaches zero.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_POLL_LEGACY:
        str =  "\nMCP_REG_MCPR_SMBUS_POLL_LEGACY - (Offset: 0x88024; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 POLL_LEGACY This value counts down to zero once each 250 msec. and sets the PL_TO\n"
               "  bit when it reaches zero. The counter stops when it reaches zero.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_RETRAN:
        str =  "\nMCP_REG_MCPR_SMBUS_RETRAN - (Offset: 0x88028; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-8 unused0\tRO\t0\n"
               "bit 7-0 RETRAN This value counts down to zero once each second and sets the RT_TO bit\n"
               "  when it reaches zero. The counter stops when it reaches zero.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_TIMESTAMP:
        str =  "\nMCP_REG_MCPR_SMBUS_TIMESTAMP - (Offset: 0x8802C; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-0 TIMESTAMP This value counts up once each second and rolls to zero each time it\n"
               "  passes 0xffffffff. This counter only counts when the TIMESTAMP_CNT_EN\n"
               "  bit is '1'.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_COMMAND:
        str =  "\nMCP_REG_MCPR_SMBUS_MASTER_COMMAND - (Offset: 0x88030; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 START_BUSY This bit is self clearing. When written to a '1', the currently programmed\n"
               "  SMBUS transaction will activate. Writing this bit as a '0' has no effect.\n"
               "  This bit must be read as a '0' before setting it to prevent un-predictable\n"
               "  results.\tSC\t0\n"
               "bit 30 ABORT Transaction Abort. This bit can be set at any time by the firmware or the\n"
               "  driver in order to abort the transaction. The HW will abort transaction in\n"
               "  compliance with the SMBUS rules and clear the bit when done.\tSC\t0\n"
               "bit 29-28 unused1\tRO\t0\n"
               "bit 27-25 STATUS These bits encode status of the last master transaction. Valid when\n"
               "  START_BUSY is cleared after it was set\tRO\t0\n"
               "  Value Name Description\n"
               "   0 000 Transaction completed successfully\n"
               "   1 001 Lost Arbitration. Firmware should restart transaction,\n"
               "     if required.\n"
               "   2 010 NACK detected after (slave address) first byte.\n"
               "  Indicates that slave device is off-line.\n"
               "   3 011 NACK detected after byte other than first. Indicates\n"
               "  that slave device maybe busy.\n"
               "   4 100 TTIMEOUT_MIN exceeded. Indicates that slave\n"
               "  device held bus for more then 25ms (assuming that\n"
               "  our clock driver is functioning properly).\n"
               "   5 101 TLOW:MEXT exceeded. Used for non-standard\n"
               "  transactions where payload is longer than the TX\n"
               "  FIFO. Indicates FIFO under-run (firmware didn't\n"
               "  provide additional data within 10ms).\n"
               "   6 110 TLOW:MEXT exceeded. Indicates that read\n"
               "  transaction failed due the Master Rx FIFO being full\n"
               "  for more than 10ms.\n"
               "   7 111 Reserved\n"
               "bit 24-13 unused0\tRO\t0\n"
               "bit 12-9 SMBUS_PROTOCOL SMBUS Protocol encoded as\tRW\t0\n"
               "  Value Name Description\n"
               "   0 0000 Quick Command\n"
               "   1 0001 Send Byte\n"
               "   2 0010 Receive Byte\n"
               "   3 0011 Write Byte\n"
               "   4 0100 Read Byte\n"
               "   5 0101 Write Word\n"
               "   6 0110 Read Word\n"
               "   7 0111 Block Write\n"
               "   8 1000 Block Read\n"
               "   9 1001 Process Call\n"
               "   10 1010 Block Write-Block Read Process Call\n"
               "   11 1011 Host Notify\n"
               "bit 8 PEC PEC should be checked or calculated for this transaction.\tRW\t0\n"
               "bit 7-0 RD_BYTE_COUNT This is number of bytes that SMBUS block should read from the slave in\n"
               "  Block Write - Block Read Process Call or Block Read. If this field is 0 the\n"
               "  SMBUS block will assume that first byte received from the slave is a Byte\n"
               "  Count. If different than 0 HW will use this value as an indication as\n"
               "  where to insert NACK and STOP that is end the transaction.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_COMMAND:
        str =  "\nMCP_REG_MCPR_SMBUS_SLAVE_COMMAND - (Offset: 0x88034; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 START _BUSY This bit is self clearing. When written to a '1', the currently\n"
               "  programmed SMBUS transaction will activate. Writing this bit as a '0' has\n"
               "  no effect. This bit must be read as a '0' before setting it to prevent\n"
               "  un-predictable results. Used only for read response.\tSC\t0\n"
               "bit 30 ABORT Transaction Abort. This bit can be set at any time by the firmware or the\n"
               "  driver in order to abort the transaction. The HW will abort transaction in\n"
               "  compliance with the SMBUS rules and clear the bit when done. When set\n"
               "  and the slave transaction is received the HW will ACK the address and\n"
               "  float the bus thereafter.\tSC\t0\n"
               "bit 29-26 unused2\tRO\t0\n"
               "bit 25-23 STATUS These bits encode status of the last slave transaction. Valid when\n"
               "  START_BUSY is cleared after it was set\tRO\t0\n"
               "  Value Name Description\n"
               "   0 000 Transaction completed successfully\n"
               "   5 101 TTIMEOUT_MIN exceeded. Indicates that clock was\n"
               "  held low for more then RANDOM_SLAVE_STRETCH by\n"
               "  the SMBUS or 25ms by the master.\n"
               "   7 111 Master aborted read transaction. HW will flush the\n"
               "  Slave Transmit FIFO.\n"
               "bit 22-9 unused1\tRO\t0\n"
               "bit 8 PEC PEC should be calculated for this transaction.\tRW\t0\n"
               "bit 7-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_EVENT_ENABLE:
        str =  "\nMCP_REG_MCPR_SMBUS_EVENT_ENABLE - (Offset: 0x88038; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 MASTER_RX_FIFO_FULL_EN When set enables MASTER_RX_FIFO_FULL to generate\n"
               "  smbus event.\tRW\t0\n"
               "bit 30 MASTER_RX_THRESHOLD_HIT_EN When set enables MASTER_RX_THRESHOLD_HIT to\n"
               "  generate smbus event.\tRW\t0\n"
               "bit 29 MASTER_RX_EVENT_EN When set enables MASTER_RX_EVENT to generate smbus\n"
               "  event.\tRW\t0\n"
               "bit 28 MASTER_START_BUSY_EN When set enables generation of a smbus event on master\n"
               "  START_BUSY 1 to 0 transition.\tRW\t0\n"
               "bit 27 MASTER_TX_UNDERRUN_EN When set enables generation of a smbus event when\n"
               "  Master Tx FIFO becomes empty and less then PKT_LENGTH\n"
               "  bytes were output on the SMBUS.\tRW\t0\n"
               "bit 26 SLAVE_RX_FIFO_FULL_EN When set enables SLAVE_RX_FIFO_FULL to generate smbus\n"
               "  event.\tRW\t0\n"
               "bit 25 SLAVE_RX_THRESHOLD_HIT_EN When set enables SLAVE_RX_THRESHOLD_HIT to generate\n"
               "  smbus event.\tRW\t0\n"
               "bit 24 SLAVE_RX_EVENT_EN Enables SLAVE_RX_EVENT to generate smbus event.\tRW\t0\n"
               "bit 23 SLAVE_START_BUSY_EN When set enables generation of a smbus event on slave\n"
               "  START_BUSY 1 to 0 transition.\tRW\t0\n"
               "bit 22 SLAVE_TX_UNDERRUN_EN When set enables generation of a smbus event when Slave\n"
               "  Tx FIFO becomes empty and less then PKT_LENGTH bytes\n"
               "  were output on the SMBUS.\tRW\t0\n"
               "bit 21 SLAVE_RD_EVENT_EN Enables SLAVE_RD_EVENT to generate smbus event.\tRW\t0\n"
               "bit 20 SLAVE_ARP_EVENT_EN When set enables hardware to generate smbus event any\n"
               "  time and ARP command is received and ARP_EN0 or\n"
               "  ARP_EN1 bit is set.\tRW\t0\n"
               "bit 19-5 unused0\tRO\t0\n"
               "bit 4 RETRANSMIT_TO_EN When set enables Retransmit Timer to generate smbus\n"
               "  event.\tRW\t0\n"
               "bit 3 POLL_LEGACY_TO_EN When set enables Legacy Sensor Poll Timer to generate\n"
               "  smbus event.\tRW\t0\n"
               "bit 2 POLL_ASF_TO_EN When set enables ASF Sensor Poll Timer to generate smbus\n"
               "  event.\tRW\t0\n"
               "bit 1 HEARTBEAT_TO_EN When set enables Heartbeat Timer to generate smbus\n"
               "  event.\tRW\t0\n"
               "bit 0 WATCHDOG_TO_EN When set enables Watchdog Timer to generate smbus\n"
               "  event.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_EVENT_STATUS:
        str =  "\nMCP_REG_MCPR_SMBUS_EVENT_STATUS - (Offset: 0x8803C; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 MASTER_RX_FIFO_FULL This bit is set when the master receive FIFO become full.\n"
               "  Writing a '1' to this position will clear this bit. When this bit is\n"
               "  '1', the SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n"
               "bit 30 MASTER_RX_THRESHOLD_HIT This bit is set when the master receive FIFO is equal to or\n"
               "  larger than the Master RX_FIFO_THRESHOLD. Writing a '1' to\n"
               "  this position will clear this bit. When this bit is '1', the\n"
               "  SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n"
               "bit 29 MASTER_RX_EVENT This bit is set when the master receive FIFO holds at least one\n"
               "  valid transaction. Writing a '1' to this position will clear this\n"
               "  bit. When this bit is '1', the SMB0_EVENT bit will be '1' in each\n"
               "  processor.\tWC\t0\n"
               "bit 28 MASTER_START_BUSY This bit is set when master START_BUSY transitions from 1 to\n"
               "  0. Writing a '1' to this position will clear this bit. When this bit\n"
               "  is '1', the SMB0_EVENT bit will be '1' in each processor. 0x\tWC\t0\n"
               "bit 27 MASTER_TX_UNDERRUN This bit is set when Master Tx FIFO becomes empty and less\n"
               "  then PKT_LENGTH bytes were output on the SMBUS.\tWC\t0\n"
               "bit 26 SLAVE_RX_FIFO_FULL This bit is set when the slave receive FIFO become full.\n"
               "  Writing a '1' to this position will clear this bit. When this bit is\n"
               "  '1', the SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n"
               "bit 25 SLAVE_RX_THRESHOLD_HIT This bit is set when the slave receive FIFO is equal to or larger\n"
               "  than the Slave RX_FIFO_THRESHOLD. Writing a '1' to this\n"
               "  position will clear this bit. When this bit is '1', the\n"
               "  SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n"
               "bit 24 SLAVE_RX_EVENT This bit is set when the slave receive FIFO holds at least one\n"
               "  valid transaction. Writing a '1' to this position will clear this\n"
               "  bit. When this bit is '1', the SMB0_EVENT bit will be '1' in each\n"
               "  processor.\tWC\t0\n"
               "bit 23 SLAVE_START_BUSY This bit is set when slave START_BUSY transitions from 1 to 0.\n"
               "  Writing a '1' to this position will clear this bit. When this bit is\n"
               "  '1', the SMB0_EVENT bit will be '1' in each processor. 0x\tWC\t0\n"
               "bit 22 SLAVE_TX_UNDERRUN This bit is set when Slave Tx FIFO becomes empty and less\n"
               "  then PKT_LENGTH bytes were output on the SMBUS.\tWC\t0\n"
               "bit 21 SLAVE_RD_EVENT This bit is set when slave hardware detected read transaction\n"
               "  directed toward the SMBUS block. Writing a '1' to this position\n"
               "  will clear this bit. When this bit is '1', the SMB0_EVENT bit will\n"
               "  be '1' in each processor.\tWC\t0\n"
               "bit 20 SLAVE_ARP_EVENT This bit set when slave hardware received an ARP command\n"
               "  and ARP_EN0 or ARP_EN1 bit is set.\tWC\t0\n"
               "bit 19-5 unused0\tRO\t0\n"
               "bit 4 RETRANSMIT_TO This bit changes to '1' each time the RETRANSMIT timer\n"
               "  reaches zero. Writing a '1' to this position will clear this bit.\n"
               "  When this bit is '1', the SMB0_EVENT bit will be '1' in each\n"
               "  processor.\tWC\t0\n"
               "bit 3 POLL_LEGACY_TO This bit changes to '1' each time the POLL_LEGACY timer\n"
               "  reaches zero. Writing a '1' to this position will clear this bit.\n"
               "  When this bit is '1', the SMB0_EVENT bit will be '1' in each\n"
               "  processor.\tWC\t0\n"
               "bit 2 POLL_ASF_TO This bit changes to '1' each time the POLL_ASF timer reaches\n"
               "  zero. Writing a '1' to this position will clear this bit. When this\n"
               "  bit is '1', the SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n"
               "bit 1 HEARTBEAT_TO This bit changes to '1' each time the HEARTBEAT timer reaches\n"
               "  zero. Writing a '1' to this position will clear this bit. When this\n"
               "  bit is '1', the SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n"
               "bit 0 WATCHDOG_TO This bit changes to '1' each time the WATCHDOG timer reaches\n"
               "  zero. Writing a '1' to this position will clear this bit. When this\n"
               "  bit is '1', the SMB0_EVENT bit will be '1' in each processor.\tWC\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_DATA_WRITE:
        str =  "\nMCP_REG_MCPR_SMBUS_MASTER_DATA_WRITE - (Offset: 0x88040; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 WR_STATUS 0 - Byte other then last in an WMBUS transaction 1 - End of\n"
               "  SMBUS transaction\tRW\t0\n"
               "bit 30-8 unused0\tRO\t0\n"
               "bit 7-0 MASTER_SMBUS_WR_DATA This is a software interface to the SMBUS Master Transmit FIFO.\n"
               "  Software should use this register to provide data for an SMBUS\n"
               "  write transaction. Data should be written before the SMBUS\n"
               "  Command is programmed.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_DATA_READ:
        str =  "\nMCP_REG_MCPR_SMBUS_MASTER_DATA_READ - (Offset: 0x88044; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-30 RD_STATUS Enumeration:\tRO\t0\n"
               "  Value Name Description\n"
               "   0 00 FIFO Empty\n"
               "   1 01 Start of SMBUS transaction\n"
               "   2 10 Byte other then first or last in an SMBUS\n"
               "  transaction\n"
               "   3 11 End of SMBUS transaction\n"
               "bit 29 PEC_ERR PEC error. This bit indicates status of the PEC checking. HW will\n"
               "  check the PEC only in case where PEC bit in SMBUS Master\n"
               "  Command Register was set for rhe transaction This field is valid\n"
               "  only when RD_STATUS = 2'b11.\tRO\t0\n"
               "bit 28-8 unused0\tRO\t0\n"
               "bit 7-0 MASTER_SMBUS_RD_DATA SMBUS Read Data in Network Byte Order (MSB first). This is a\n"
               "  software interface to the SMBUS Master Receive FIFO. Software\n"
               "  should use this register to read data received from the SMBUS.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_DATA_WRITE:
        str =  "\nMCP_REG_MCPR_SMBUS_SLAVE_DATA_WRITE - (Offset: 0x88048; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 WR_STATUS Enumeration:\tRW\t0\n"
               "  Value Name Description\n"
               "   0 0 Byte other then last in an SMBUS transaction\n"
               "   1 1 End of SMBUS transaction\n"
               "bit 30-8 unused0\tRO\t0\n"
               "bit 7-0 SLAVE_SMBUS_WR_DATA This is a software interface to the SMBUS Slave Transmit FIFO.\n"
               "  Software should use this register to provide data for an SMBUS\n"
               "  read transaction.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_DATA_READ:
        str =  "\nMCP_REG_MCPR_SMBUS_SLAVE_DATA_READ - (Offset: 0x8804C; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-30 RD_STATUS Enumeration:\tRO\t0\n"
               "  Value Name Description\n"
               "   0 00 FIFO Empty\n"
               "   1 01 Start of SMBUS transaction\n"
               "   2 10 Byte other then first or last in an SMBUS\n"
               "  transaction\n"
               "   3 11 End of SMBUS transaction\n"
               "bit 29-28 ERR_STATUS This field is valid only when RD_STATUS = 2'b11.\tRO\t0\n"
               "  Value Name Description\n"
               "  0 00 Transaction completed successfully\n"
               "  1 01 PEC error. This bit indicates status of the PEC\n"
               "  checking. HW assumes that last byte of the\n"
               "  transaction is always PEC byte and sets this bit\n"
               "  accordingly. It is up to the software to\n"
               "  recognize (based on the Command Code) if the\n"
               "  particular transaction has a PEC or not. If it\n"
               "  does not software should ignore this bit.\n"
               "   2 10 FIFO overflow. SMBUS transaction timed out due\n"
               "  to the receive FIFO being full for more than\n"
               "  RANDOM_SLAVE_STRETCH time.\n"
               "   3 11 General SMBUS transaction error. This is used\n"
               "  for all errors not covered by b01 and b10 errors.\n"
               "bit 27-8 unused0\tRO\t0\n"
               "bit 7-0 SLAVE_SMBUS_RD_DATA This is a software interface to the SMBUS Slave Receive FIFO.\n"
               "  Software should use this register to read data received from the\n"
               "  SMBUS.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_ARP_STATE:
        str =  "\nMCP_REG_MCPR_SMBUS_ARP_STATE - (Offset: 0x88080; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-6 unused1\tRO\t0\n"
               "bit 5 AR_FLAG1 This bit should be set by firmware before ARP_EN1 bit is set. This bit is\n"
               "  typically initialized to '0'.\tRW\t0\n"
               "bit 4 AV_FLAG1 This bit should be set by firmware before ARP_EN1 bit is set. This bit is\n"
               "  typically set to '1' based on the NVRAM content that is if device supports\n"
               "  Persistent Slave Address and cleared otherwise.\tRW\t0\n"
               "bit 3-2 unused0\tRO\t0\n"
               "bit 1 AR_FLAG0 This bit should be set by firmware before ARP_EN0 bit is set. This bit is\n"
               "  typically initialized to '0'.\tRW\t0\n"
               "bit 0 AV_FLAG0 This bit should be set by firmware before ARP_EN0 bit is set. This bit is\n"
               "  typically set to '1' based on the NVRAM content that is if device supports\n"
               "  Persistent Slave Address and cleared otherwise.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_3:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID0_3 - (Offset: 0x88090; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
               "bit 31-24 BYTE_15 UDID_0 byte 15.\tRW\t0\n"
               "bit 23-16 BYTE_14 UDID_0 byte 14.\tRW\t0\n"
               "bit 15-8 BYTE_13 UDID_0 byte 13.\tRW\t0\n"
               "bit 7-0 BYTE_12 UDID_0 byte 12.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_2:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID0_2 - (Offset: 0x88094; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
               "bit 31-24 BYTE_11 UDID_0 byte 11.\tRW\t0\n"
               "bit 23-16 BYTE_10 UDID_0 byte 10.\tRW\t0\n"
               "bit 15-8 BYTE_9 UDID_0 byte 9.\tRW\t0\n"
               "bit 7-0 BYTE_8 UDID_0 byte 8.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_1:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID0_1 - (Offset: 0x88098; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
               "bit 31-24 BYTE_7 UDID_0 byte 7.\tRW\t0\n"
               "bit 23-16 BYTE_6 UDID_0 byte 6.\tRW\t0\n"
               "bit 15-8 BYTE_5 UDID_0 byte 5.\tRW\t0\n"
               "bit 7-0 BYTE_4 UDID_0 byte 4.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_0:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID0_0 - (Offset: 0x8809C; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
               "bit 31-24 BYTE_3 UDID_0 byte 3.\tRW\t0\n"
               "bit 23-16 BYTE_2 UDID_0 byte 2.\tRW\t0\n"
               "bit 15-8 BYTE_1 UDID_0 byte 1.\tRW\t0\n"
               "bit 7-0 BYTE_0 UDID_0 byte 0.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_3:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID1_3 - (Offset: 0x880A0; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN1 bit is set.\n"
               "bit 31-24 BYTE_15 UDID_1 byte 15.\tRW\t0\n"
               "bit 23-16 BYTE_14 UDID_1 byte 14.\tRW\t0\n"
               "bit 15-8 BYTE_13 UDID_1 byte 13.\tRW\t0\n"
               "bit 7-0 BYTE_12 UDID_1 byte 12.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_2:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID1_2 - (Offset: 0x880A4; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN1 bit is set.\n"
               "bit 31-24 BYTE_11 UDID_1 byte 11.\tRW\t0\n"
               "bit 23-16 BYTE_10 UDID_1 byte 10.\tRW\t0\n"
               "bit 15-8 BYTE_9 UDID_1 byte 9.\tRW\t0\n"
               "bit 7-0 BYTE_8 UDID_1 byte 8.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_1:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID1_1 - (Offset: 0x880A8; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN1 bit is set.\n"
               "bit 31-24 BYTE_7 UDID_1 byte 7.\tRW\t0\n"
               "bit 23-16 BYTE_6 UDID_1 byte 6.\tRW\t0\n"
               "bit 15-8 BYTE_5 UDID_1 byte 5.\tRW\t0\n"
               "bit 7-0 BYTE_4 UDID_1 byte 4.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_0:
        str =  "\nMCP_REG_MCPR_SMBUS_UDID1_0 - (Offset: 0x880AC; Width: 32)\n"
               "  SHARE: This register should be set by firmware before ARP_EN1 bit is set.\n"
               "bit 31-24 BYTE_3 UDID_1 byte 3.\tRW\t0\n"
               "bit 23-16 BYTE_2 UDID_1 byte 2.\tRW\t0\n"
               "bit 15-8 BYTE_1 UDID_1 byte 1.\tRW\t0\n"
               "bit 7-0 BYTE_0 UDID_1 byte 0.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_CMD:
        str =  "\nMCP_REG_MCPR_UMP_UMP_CMD - (Offset: 0x98000; Width: 32)\n"
               "  SHARE: All registers in the UMP block are shared.\n"
               "bit 31 INGRESS_MAC_RESET When this bit is '1', a soft reset will be performed to the Ingress\n"
               "  MAC in the UMP block.\tRW\t0\n"
               "bit 30 EGRESS_MAC_RESET When this bit is '1', a soft reset will be performed to the Egress\n"
               "  MAC in the UMP block.\tRW\t0\n"
               "bit 29 INGRESS_FIO_RESET When this bit is '1', a soft reset will be performed to the FIO\n"
               "  interface of the UMP Ingress interface.\tSC\t0\n"
               "bit 28 EGRESS_FIO_RESET When this bit is '1', a soft reset will be performed to the FIO\n"
               "  interface of the UMP Egress interface.\tSC\t0\n"
               "bit 27-22 unused3\tRO\t0\n"
               "bit 21-17 CMD_IPG This bit programms the interpacket gap for the Ingress packets.\n"
               "  The IPG is calculated in nibbles, so a value of 24 indicates 12bytes\n"
               "  of IPG.\tRW\t0x18\n"
               "bit 16 EGRESS_PKT_FLUSH When this bit is '1', the packet at the head of the UMP Egress FIFO\n"
               "  will be flushed and if there are more packets in the FIFO,\n"
               "  appropriate actions will be taken. This allows the firmware to look\n"
               "  at packets of interest only and flush a packet by a single command\n"
               "  rather than read out the packet word by word.\tSC\t0\n"
               "bit 15 NO_PLUS_TWO When this bit is '1', the \"plus-two\" alignment for both Ingress and\n"
               "  Egress traffic is disabled. This means that the packet will be\n"
               "  aligned such that byte 0 of the packet appears in bits [31:24] of\n"
               "  the rd_data and wr_data registers. When this bit is '0', the\n"
               "  \"plus-two\" alignment for both Ingress and Egress traffic is enabled.\n"
               "  This means that the packet will be aligned such that byte 0 of the\n"
               "  packet appears in bits [15:8] of the rd_data and wr_data\n"
               "  registers.\tRW\t0\n"
               "bit 14 INGRESS_RESET When this bit is written as '1', the Ingress state machine is reset\n"
               "  to the idle state. This bit should be written as '1' when a\n"
               "  transmitter wishes to recall a transmit operation that has been\n"
               "  requested by a write to the frm_wr_ctl register.\tSC\t0\n"
               "bit 13 AUTO_DRIVE When this bit is '1', a mode is enabled where the Ingress pins of\n"
               "  the UMP interface will automatically drive from one clock before a\n"
               "  Ingress packet until the clock after the Ingress packet has been\n"
               "  transmitted. When this bit is '0', the tri-state control is done by\n"
               "  the INGRESS_DRIVE bit of this register. For this mode to operate\n"
               "  correctly, the INGRESS_DRIVE bit must be '0'.\tRW\t0\n"
               "bit 12-10 unused2\tRO\t0\n"
               "bit 9 SW_PAUSE When this bit is '1', it forces the \"pause needed\" input to the MAC\n"
               "  to '1'. This may be used by FW to force pause packets to be\n"
               "  generated during times when it knows that it can not process\n"
               "  received packets properly.\tRW\t0\n"
               "bit 8 INGRESS_DRIVE If this bit is set, then the Ingress pins of the UMP interface will be\n"
               "  forced to drive reguardless of the setting of the AUTO_DRIVE bit.\tRW\t0\n"
               "bit 7 unused1\tRO\t0\n"
               "bit 6 INGRESS_MAC_DISABLE When this bit is set, it disables the transmit function of the UMP\n"
               "  MAC. Setting this bit to '1' will reset the UMP port transmitter to a\n"
               "  known state. This bit must be set to '0' to allow transmission.\tRW\t0x1\n"
               "bit 5 EGRESS_MAC_DISABLE When this bit is set, it disables the receive function of the UMP\n"
               "  MAC. Setting this bit to '1' will reset the UMP port receiver to a\n"
               "  known state. This bit must be set to '0' to allow reception of new\n"
               "  frames.\tRW\t0x1\n"
               "bit 4 unused0\tRO\t0\n"
               "bit 3 MAC_LOOPBACK When this bit is '1', Ingress data will be looped back to the Egress\n"
               "  at the RMII interface. This feature only works in RMII mode\tRW\t0\n"
               "bit 2 FC_EN When this bit is '1', it indicates that flow control is enabled. If this\n"
               "  bit is '0', no attempt will be made to prevent packet drop.\tRW\t0\n"
               "bit 1 INGRESS_FIFO_ENABLED This bit controlls the current enable status of the Ingress side of\n"
               "  the UMP FIFO block. If this bit is '1', it indicates that the Ingress\n"
               "  FIFO is enabled When this bit is '0', then FIFO will be reset\n"
               "  immediately. If the transmit MAC is currently transmitting a packet\n"
               "  and does not get all the data needed for the packet, it will underun\tRW\t0\n"
               "  and transmit JAM for the packet and then stop.\n"
               "bit 0 EGRESS_FIFO_ENABLED This bit controlls the current enable status of recevie UMP FIFO\n"
               "  block. If this bit is '1', it indicates that the RX FIFO is enabled\n"
               "  When this bit is '0', the reset FIFO will be reset immediately. Any\n"
               "  partial or complete packets in the FIFO will be lost. This bit has no\n"
               "  effect on the reception of frames by the EGRESS MAC. Use\n"
               "  EGRESS_DISABLE for that purpose.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_CONFIG:
        str =  "\nMCP_REG_MCPR_UMP_UMP_CONFIG - (Offset: 0x98004; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-25 unused3\tRO\t0\n"
               "bit 24 OLD_BCNT_RDY When set '1', UMP logic will revert to the Xinan 90nm method of\n"
               "  calculating BCNT_RDY which does not account for the fact that the\n"
               "  writing of the byte count consumes 1 word in the FIFO. This bit is\n"
               "  available only from 65nm A0 onwards.\tRW\t0\n"
               "bit 23-16 unused2\tRO\t0\n"
               "bit 15-8 INGRESS_WORD_ACCM This field allows the UMP Ingress engine to wait for up-to 256 words\n"
               "  of data to accumulate before starting to send the data out. A word is\n"
               "  equal to 4bytes. The default for this register is 0, but this is not a\n"
               "  valid setting for cut through mode. FW needs to program the\n"
               "  appropriate value > 0. A good value is 8 which allows for 32bytes to\n"
               "  be accumulated before sending the packet out. Safest value is 16\n"
               "  which ensures that a min sized packet is in the FIFO before the\n"
               "  packet starts streaming out.\tRW\t0\n"
               "bit 7 INGRESS_MODE When set '1', the UMP Transmit engine will operate in a cut through\n"
               "  mode. It will wait for INGRESS_WORD_ACCM bytes to accumulate\n"
               "  before starting to send the packet out. When not set, the engine will\n"
               "  function in a store and forward mode.\tRW\t0\n"
               "bit 6 RVMII_MODE When this bit is '1' and RMII_MODE is '0', the UMP port will be put\n"
               "  into \"PHY\" clocking/timing mode where it will emulate the timing of a\n"
               "  PHY MII connection. The clock inputs become clock outputs and the\n"
               "  data pins of the MII are syncronous to those outputs. In this mode,\n"
               "  only single UMP can be connected to a BMC.\tRW\t0\n"
               "bit 5 unused1\tRO\t0\n"
               "bit 4 RMII_MODE When this bit is '1', the UMP interface will run in RMII mode rather\n"
               "  than normal MII mode.\tRW\t0\n"
               "bit 3-0 unused0\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_FC_TRIP:
        str =  "\nMCP_REG_MCPR_UMP_UMP_FC_TRIP - (Offset: 0x98008; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-25 unused1\tRO\t0\n"
               "bit 24-16 XOFF_TRIP When the Egress FIFO has less than XOFF_TRIP*16 bytes remaining, the\n"
               "  \"pause needed\" input to the UMP MAC will be asserted. This value is the\n"
               "  low-water mark for flow control generated by the UMP MAC. If this value is\n"
               "  0, then the \"pause needed\" input to the UMP MAC will be controlled soley\n"
               "  by the SW_PAUSE bit.\tRW\t0xc8\n"
               "bit 15-9 unused0\tRO\t0\n"
               "bit 8-0 XON_TRIP When the Egress FIFO has at least XON_TRIP*16 bytes available, the\n"
               "  \"pause needed\" input to the UMP MAC will be de-asserted. This value is\n"
               "  the high-water mark for flow control generated by the UMP MAC. A value is\n"
               "  the largest valid value that may be programmed and will de-assert the\n"
               "  \"pause needed\" signal only when the Egress FIFO is empty.\tRW\t0x124\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS - (Offset: 0x98090; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-27 EGRESS_FIFO_STATE Enumeration:\tRO\t0\n"
               "  Value Name Description\n"
               "   0 IDLE IDLE\n"
               "   1 READY READY\n"
               "   2 BUSY BUSY\n"
               "   3 EXTRA_RD Extra Read\n"
               "   4 LATCH_IP_HDR Latch IP Header\n"
               "bit 26-14 unused0\tRO\t0\n"
               "bit 13-3 BCNT Current Egress packet length in Byte count.\tRO\t0\n"
               "bit 2 FIFO_EMPTY When this bit is set, it indicates that Egress FIFO is empty and\n"
               "  firmware or software should not read rd_dat any further until this bit\n"
               "  clears.\tRO\t0\n"
               "bit 1 FRM_IN_PRO When this bit is set, it indicates that current packet reading is in\n"
               "  progress. Cleared when all packet data has been read for a particular\n"
               "  packet.\tRO\t0\n"
               "bit 0 NEW_FRM When this bit is set, it indicates that a new packet is ready in Egress\n"
               "  FIFO. Always 0 when FRM_IN_PRO bit is set.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_DATA:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_DATA - (Offset: 0x98094; Width: 32) -\n"
               "  (RO/Reset: 0)\n"
               "  SHARE: Next 32-bit data word of a packet. For a new packet and the first read, the right two bytes contain the\n"
               "  first two data bytes of the packet. This means that the 3rd byte of the packet is in bits 31:24 of the 2nd data\n"
               "  word read. The two extra bytes in the 1st read makes the L3 and above frame headers 32-bit aligned with the\n"
               "  frm_rd_data register. In other words the packet is shifted by 2bytes so the L3 headers line up on 32-bit\n"
               "  boundaries for ease of processing.\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_FRM_WR_CTL:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_FRM_WR_CTL - (Offset: 0x98098; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-30 INGRESS_FIFO_STATE Enumeration:\tRO\t0\n"
               "  Value Name Description\n"
               "   0 IDLE IDLE\n"
               "   1 WAIT WAIT\n"
               "   2 BUSY BUSY\n"
               "   3 EXTRA_WR Extra Write\n"
               "bit 29-14 unused0\tRO\t0\n"
               "bit 13-3 BCNT This register must be written with the packet length before the\n"
               "  ump_frm_wr_dat register is written with any data. For this value to\n"
               "  be accepted, the SOF bit must be set on the same registger write.\tRW\t0\n"
               "bit 2 BCNT_RDY When this bit is set, it indicates that the BCNT field can accept\n"
               "  another write as a request for space for another transmit frame. This\n"
               "  bit is logically or'ed with the FIFO_RDY bit to create the value of the\n"
               "  FTQ1_VALID event bit. This bit must be '1' when the firmware writes\n"
               "  a '1' to the NEW_FRM bit. This bit will clear back to '0' each time a\n"
               "  new BCNT value is written and will only set back to '1' once the\n"
               "  NEW_FRM bit has cleared at the end of the complete packet write.\tRO\t0\n"
               "bit 1 FIFO_RDY This bit is set when the INGRESS_FIFO can accept all the data\n"
               "  requested by a write tot he BCNT value. This value will read as '1'\n"
               "  when there is sufficient space in the Ingress FIFO to accept all the\n"
               "  bytes described by the latest BCNT value. This bit will clear to '0'\n"
               "  once all the data requested in BCNT has been written to the\n"
               "  ump_frm_wr_data register. This bit is logically or'ed with the\n"
               "  BCNT_RDY bit to create the value of the FTQ1_VALID event bit.\tRO\t0\n"
               "bit 0 NEW_FRM This bit must be set for the BCNT field to be accepted as the length\n"
               "  of a new packet. This bit will clear to zero when the complete packet\n"
               "  has been written to the ump_frm_wr_data register.\tSC\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_FRM_WR_DATA:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_FRM_WR_DATA - (Offset: 0x9809C; Width: 32)\n"
               "  - (WO/Reset: 0)\n"
               "  SHARE: Next 32-bit data word of a transmit packet. Only the right two bytes of the first data word of each packet\n"
               "  are valid. This means that the 3rd byte of the packet is in bits 31:24 of the 2nd data word write. This offset of 2\n"
               "  bytes makes the L3 and above frame headers 32-bit aligned with the frm_wr_data register. This register must\n"
               "  only be written when FIFO_RDY bit is '1' in the ump_frm_wr_ctrl register. A read operation on this register will\n"
               "  always return 0s.\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FRAME_TYPE:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_FRAME_TYPE - (Offset: 0x980A0; Width: 32) -\n"
               "  (RO/Reset: 0)\n"
               "  SHARE: This register provides the ethernet type field of the packet when a new packet is ready to be read from\n"
               "  the rd_data regsiter. Reading this register has no effect on the FIFO status.\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_FIFO_REMAINING_WORDS:
        str =  "\nMCP_REG_MCPR_UMP_UMP_FIFO_REMAINING_WORDS - (Offset: 0x980A4; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 INGRESS_FIFO_OVERFLOW This bit indicates that there was an overflow in the UMP Ingress\n"
               "  FIFO. This works only from Xinan 65nm A0 onwards.\tWC\t0\n"
               "bit 30 INGRESS_FIFO_UNDERFLOW This bit indicates that there was an underflow in the UMP\n"
               "  Ingress FIFO. This works only from Xinan 65nm A0 onwards.\tWC\t0\n"
               "bit 29-26 unused1\tRO\t0\n"
               "bit 25-16 INGRESS_FIFO_DEPTH Remainig 32-bit words in Ingress UMP FIFO.\tRO\t0\n"
               "bit 15 EGRESS_FIFO_OVERFLOW This bit indicates that there was an overflow in the UMP Egress\n"
               "  FIFO. This works only from Xinan 65nm A0 onwards.\tWC\t0\n"
               "bit 14 EGRESS_FIFO_UNDERFLOW This bit indicates that there was an underflow in the UMP\n"
               "  Egress FIFO. This works only from Xinan 65nm A0 onwards.\tWC\t0\n"
               "bit 13-11 unused0\tRO\t0\n"
               "bit 10-0 EGRESS_FIFO_DEPTH Remainig 32-bit words in Egress UMP FIFO.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_PTRS:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_PTRS - (Offset: 0x980A8; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 UPDATE_WRPTR Update the write pointer of Ingress FIFO with\n"
               "  egress_write_pointer.\tSC\t0\n"
               "bit 30-28 unused1\tRO\t0\n"
               "bit 27-16 EGRESS_FIFO_WR_PTR Write pointer of the Egress FIFO in 32-bit words. This field is\n"
               "  writable for debug purpose and only when there is atleast one packet\n"
               "  in the FIFO.\tRW\t0\n"
               "bit 15 UPDATE_RDPTR Update the read pointer of Egress FIFO with egress_read_pointer. SC 0\n"
               "bit 14-12 unused0\tRO\t0\n"
               "bit 11-0 EGRESS_FIFO_RD_PTR Read pointer of the Egress FIFO in 32-bit words. This field is writable\n"
               "  for debug purpose and only when there is atleast one packet in the\n"
               "  FIFO.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_PTRS:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_PTRS - (Offset: 0x980AC; Width: 32)\n"
               "  SHARE:\n"
               "bit 31 UPDATE_WRPTR Update the write pointer of Ingress FIFO with\n"
               "  ingress_write_pointer.\tSC\t0\n"
               "bit 30-27 unused1\tRO\t0\n"
               "bit 26-16 INGRESS_FIFO_WR_PTR Write pointer of the Ingress FIFO in 32-bit words. This field is\n"
               "  writable for debug purpose and only when there is atleast one\n"
               "  packet in the FIFO.\tRW\t0\n"
               "bit 15 UPDATE_RDPTR Update the read pointer of Ingress FIFO with\n"
               "  ingress_read_pointer.\tSC\t0\n"
               "bit 14-11 unused0\tRO\t0\n"
               "bit 10-0 INGRESS_FIFO_RD_PTR Read pointer of the Ingress FIFO in 32-bit words. This field is\n"
               "  writable for debug purpose and only when there is atleast one\n"
               "  packet in the FIFO.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_PACKET_SA_0:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_PACKET_SA_0 - (Offset: 0x980B4; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 EGRESS_SA Upper 16 bits of the Source Address of a packet received on the UMP port.\n"
               "  The SA will be used by the firmware to decide which of the two gigabit\n"
               "  ports will be used to transmit the packet.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_PACKET_SA_1:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_PACKET_SA_1 - (Offset: 0x980B8; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-0 EGRESS_SA Lower 32 bits of the Source Address of a packet received on the UMP port.\n"
               "  The SA will be used by the firmware to decide which of the two gigabit\n"
               "  ports will be used to transmit the packet.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_BURST_COMMAND:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_BURST_COMMAND - (Offset: 0x980BC; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-30 unused1\tRO\t0\n"
               "bit 29-16 RBUF_OFFSET This field programs offset into the RBUF cluster from where the DMA\n"
               "  engine needs to copy data. The offset needs to be 32-bit aligned. The\n"
               "  FW is required to program the rbuf_cluster field before doing this\n"
               "  operation.\tRW\t0\n"
               "bit 15 unused0\tRO\t0\n"
               "bit 14-13 INGRESS_PORT_EXT This bit selects the 2 MSb's of which main MAC port the packet was\n"
               "  received.\tRW\t0\n"
               "bit 12-2 DMA_LENGTH This field programs the length of the data that needs to be\n"
               "  transferred from the RBUF. The length is represented in byte units.\tRW\t0\n"
               "bit 1 INGRESS_PORT This bit selects the LSb of which main MAC port the packet was\n"
               "  received.\tRW\t0\n"
               "bit 0 INGRESS_DMA_START Setting this bit to '1' starts the DMA engine that reads data from the\n"
               "  RBUF and writes to UMP Ingress MAC. This is a self clearing bit that\n"
               "  clears after the DMA operation is completed.\tSC\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_RBUF_CLUSTER:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_RBUF_CLUSTER - (Offset: 0x980C0; Width: 32)\n"
               "  SHARE:\n"
               "  Bit Name Description Mode Reset\n"
               "bit 31-25 unused0\tRO\t0\n"
               "bit 24-0 RBUF_CLUSTER This field programs the RBUF cluster.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_VLAN:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_VLAN - (Offset: 0x980C4; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-18 unused0\tRO\t0\n"
               "bit 17 VLAN_DEL Setting this field to '1' causes VLAN Tag to be removed from the\n"
               "  ingress packet.\tRW\t0\n"
               "bit 16 VLAN_INS Setting this field to '1' causes VLAN Tag to be inserted in the ingress\n"
               "  packet.\tRW\t0\n"
               "bit 15-0 INGRESS_VLAN_TAG This field is used to insert VLAN Tag into a ingress packet if vlan_ins is\n"
               "  set.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_BURST_STATUS:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_BURST_STATUS - (Offset: 0x980C8; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-2 unused0\tRO\t0\n"
               "bit 1-0 RESULT This bit indicates the status of the current burst command. Bit gets cleared\n"
               "  when this register is read or INGRESS_BURST_START is set to '1'\n"
               "  AC 0\n"
               "  Value Name Description\n"
               "   0 BUSY DMA is in progress\n"
               "   1 DONE DMA transfer completed successfully\n"
               "   2 ERR DMA transfer interrupted due to MCP_ISOLATE\n"
               "   3 ERR1 DMA transfer interrupted due to the FIFO under-run\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_BURST_COMMAND:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_BURST_COMMAND - (Offset: 0x980CC; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-29 unused1\tRO\t0\n"
               "bit 28-16 TPBUF_OFFSET This field programs offset into the TPBUF memory, this is the memory\n"
               "  to which the DMA engine will move the Egress Data.TPBUF in Xinan is\n"
               "  organized as a 16-byte wide memory and FW should ensure that the\n"
               "  offset is 16-byte aligned. The offset field in this register as it is points\n"
               "  to [TPBUF_ADDR_WIDTH-1:2] of the tpbuf address, in other words the\n"
               "  offset is defined in 4-byte increments. To make sure that the offset is\n"
               "bit 16-byte aligned, the lower two bits (17,16) need to be set to 0.\tRW\t0\n"
               "bit 15-14 unused0\tRO\t0\n"
               "bit 13 EGRESS_PORT_EXT This bit selects the MSb of which main MAC port the packet is sent out.\tRW\t0\n"
               "bit 12-2 DMA_LENGTH This field programs the length of the data that needs to be transferred\n"
               "  to the TPBUF. The length is represented in byte units.\tRW\t0\n"
               "bit 1 EGRESS_PORT This bit selects the LSb of which main MAC port the packet is sent out.\tRW\t0\n"
               "bit 0 EGRESS_DMA_START Setting this bit to '1' starts the DMA engine that reads data from UMP\n"
               "  Egress MAC and writes to TPBUF. This is a self clearing bit that clears\n"
               "  after the DMA operation is completed.\tSC\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_VLAN:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_VLAN - (Offset: 0x980D0; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-18 unused0\tRO\t0\n"
               "bit 17 VLAN_DEL Setting this field to '1' causes VLAN Tag to be removed from the egress\n"
               "  packet.\tRW\t0\n"
               "bit 16 VLAN_INS Setting this field to '1' causes VLAN Tag to be inserted in the egress\n"
               "  packet.\tRW\t0\n"
               "bit 15-0 EGRESS_VLAN_TAG This field is used to insert VLAN Tag into a egress packet if vlan_ins is\n"
               "  set.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_BURST_STATUS:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_BURST_STATUS - (Offset: 0x980D4; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-2 unused0\tRO\t0\n"
               "bit 1-0 RESULT This bit indicates the status of the current burst command. Bit gets cleared\n"
               "  when this register is read or EGRESS_BURST_START is set to '1'\n"
               "  AC 0\n"
               "  Value Name Description\n"
               "   0 BUSY DMA is in progress\n"
               "   1 DONE DMA transfer completed successfully\n"
               "   2 ERR0 DMA transfer interrupted due to MCP_ISOLATE\n"
               "   3 RSVD Reserved\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC - (Offset: 0x980D8; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-24 EGRESS_DROP_CNT This field keeps the count of how many packets were dropped by the\n"
               "  UMP MAC for lack of FIFO space.\tRO\t0\n"
               "bit 23-16 EGRESS_ERROR_CNT This field keeps the count of how many error packets were received by\n"
               "  the UMP MAC.\tRO\t0\n"
               "bit 15-0 EGRESS_GOOD_CNT This field keeps the count of how many good packets were received by\n"
               "  the UMP MAC.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC - (Offset: 0x980DC; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 INGRESS_PKT_CNT This field keeps the count of how many packets were transmitted by the\n"
               "  UMP MAC.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_ARB_CMD:
        str =  "\nMCP_REG_MCPR_UMP_UMP_ARB_CMD - (Offset: 0x980E0; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 UMP_ARB_TIMEOUT This field indicates the value in number of Ingress clock cycles that\n"
               "  the arbitration master will wait before re-starting the arbitration\n"
               "  process.\tRW\t0x400\n"
               "bit 15 UMP_ARB_FC_DISABLE Setting this bit disables the feature to send XOFF/XON ordered\n"
               "  sets on the arbiter interface. This feature is enabled by default and\n"
               "  allows a UMP port that wants to send a XOFF/XON frame to use the\n"
               "  sideband interface to accelerate sending the command on the MII\n"
               "  bus. This helps reducing the size of the Rx FIFO needed when\n"
               "  multiple UMP ports are connected to the BMC.\tRW\t0\n"
               "bit 14 unused1\tRO\t0\n"
               "bit 13 UMP_ARB_TOKEN_VALID This bit indicates if the arbiter has a valid token.\tRO\t0\n"
               "bit 12-8 UMP_ARB_TOKEN_IPG This field is a programmable inter-packet gap for when the token is\n"
               "  sent out.\tRW\t0x14\n"
               "bit 7 UMP_ARB_AUTOBYPASS Setting this field to '1' causes the UMP port to cut latency when\n"
               "  forwarding a token.\tRW\t0\n"
               "bit 6 UMP_ARB_BYPASS Setting this field to '1' the HW arbitration logic to function in\n"
               "  bypass mode. This allows UMP ports that don't have the firmware\n"
               "  running to be automatically bypassed. Firmware should also set\n"
               "  this bit when there is nothing to send. This will reduce the latency\n"
               "  of the token around the loop.\tRW\t0x1\n"
               "bit 5 UMP_ARB_START Setting this field to '1' causes the hardware arbitration scheme to\n"
               "  begin. Any UMP port can re-start the arbitration.\tSC\t0\n"
               "bit 4 UMP_ARB_DISABLE Setting this field to '1' causes the hardware arbitration scheme to\n"
               "  be disabled. This bit should be set when there is only one UMP port\n"
               "  on the board and tokens need not be passed out.\tRW\t0\n"
               "bit 3 unused0\tRO\t0\n"
               "bit 2-0 UMP_ID This field is used to set the ID of the current UMP port. UMP port\n"
               "  with the lowest ID value is the master of the HW based arbitration.\tRW\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_FRAME_COUNT:
        str =  "\nMCP_REG_MCPR_UMP_UMP_FRAME_COUNT - (Offset: 0x980EC; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-21 unused1\tRO\t0\n"
               "bit 20-16 INRESS_FRAME_COUNT This field keeps the count of how many packets are present in the\n"
               "  INGRESS FIFO at any given time. This is provided for debug\n"
               "  purposes and only available in Xinan 65nm onwards.\tRO\t0\n"
               "bit 15-7 unused0\tRO\t0\n"
               "bit 6-0 EGRESS_FRAME_COUNT This field keeps the count of how many packets are present in the\n"
               "  EGRESS FIFO at any given time. This is provided for debug purposes\n"
               "  and only available in Xinan 65nm onwards.\tRO\t0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC_AC:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC_AC - (Offset: 0x980F0; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-24 EGRESS_DROP_CNT This field keeps the count of how many packets were dropped by the\n"
               "  UMP MAC for lack of FIFO space.\n"
               "  AC 0\n"
               "bit 23-16 EGRESS_ERROR_CNT This field keeps the count of how many error packets were received by\n"
               "  the UMP MAC.\n"
               "  AC 0\n"
               "bit 15-0 EGRESS_GOOD_CNT This field keeps the count of how many good packets were received by\n"
               "  the UMP MAC.\n"
               "  AC 0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC_AC:
        str =  "\nMCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC_AC - (Offset: 0x980F4; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-16 unused0\tRO\t0\n"
               "bit 15-0 INGRESS_PKT_CNT This field keeps the count of how many packets were transmitted by the\n"
               "  UMP MAC.\n"
               "  AC 0\n";
        break;
    case MCP_REG_MCPR_UMP_UMP_EVENT:
        str =  "\nMCP_REG_MCPR_UMP_UMP_EVENT - (Offset: 0x980F8; Width: 32)\n"
               "  SHARE:\n"
               "bit 31-21 unused1\tRO\t0\n"
               "bit 20 EGRESS_FRAME_DROP_EVENT_EN Event Enable bit.\tRW\t0\n"
               "bit 19 EGRESSBURST_DONE_EVENT_EN Event Enable bit.\tRW\t0\n"
               "bit 18 INGRESSBURST_DONE_EVENT_EN Event Enable bit.\tRW\t0\n"
               "bit 17 EGRESS_RDY_EVENT_EN Event Enable bit.\tRW\t0\n"
               "bit 16 INGRESS_RDY_EVENT_EN Event Enable bit.\tRW\t0\n"
               "bit 15-5 unused0\tRO\t0\n"
               "bit 4 EGRESS_FRAME_DROP_EVENT\tWC\t0\n"
               "bit 3 EGRESSBURST_DONE_EVENT\tWC\t0\n"
               "bit 2 INGRESSBURST_DONE_EVENT\tWC\t0\n"
               "bit 1 EGRESS_RDY_EVENT\tRO\t0\n"
               "bit 0 INGRESS_RDY_EVENT\tRO\t0\n";
        break;
    default: 
        str = NULL;
        break;
    }
    return str;
}// GetMcpGrcHelp

/*
 *------------------------------------------------------------------------
 * DecodeMcpGrcRegDetail --
 *
 *     Decode MCP GRC register with details (with option help info)
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
void DecodeMcpGrcRegDetail(Tcl_Obj* rc_ptr, u32_t addr, u32_t data, u32_t help)
{
    static char  disp_str[1000];
    char const * help_str = GetMcpGrcHelp(addr);

    if (help_str)
    {
        if (help)
        {
            Tcl_AppendStringsToObj( rc_ptr, help_str, NULL);
        }
        else
        {
            // just display the register name
            sscanf(help_str,"%s- (Offset:", disp_str);
            Tcl_AppendStringsToObj( rc_ptr, "\n",disp_str, NULL);
        }
    }
    else
    {
        // nothing to do
        return;
    }
                                                                   
    switch (addr)
    {
    case MCP_REG_MCPR_MCP_CONTROL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x) %s %s %s %s",
            addr, data, 
            "\n\tSMBUS_SEL    :", (u32_t)(data & MCPR_MCP_CONTROL_SMBUS_SEL        ) ? "ON":"OFF",
            "\n\tMCP_ISOLATE  :", (u32_t)(data & MCPR_MCP_CONTROL_MCP_ISOLATE      ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MCP_ATTENTION_STATUS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x) %s %s %s %s %s %s",
            addr, data, 
            "\n\tDRV_DOORBELL       :", (u32_t)(data & MCPR_MCP_ATTENTION_STATUS_DRV_DOORBELL  ) ? "ON":"OFF",
            "\n\tWATCHDOG_TIMEOUT   :", (u32_t)(data & MCPR_MCP_ATTENTION_STATUS_WATCHDOG_TIMEOUT  ) ? "ON":"OFF",
            "\n\tCPU_EVENT          :", (u32_t)(data & MCPR_MCP_ATTENTION_STATUS_CPU_EVENT  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MCP_HEARTBEAT_CONTROL: 
        // SHARE: Control for MCP heartbeat feature.
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data,
            "\n\tHEARTBEAT_ENABLE  :", (u32_t)(data & MCPR_MCP_HEARTBEAT_CONTROL_MCP_HEARTBEAT_ENABLE  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MCP_HEARTBEAT_STATUS:
        // SHARE: Status of MCP heartbeat feature     
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %s",
            addr, data,
            "\n\tMCP_HEARTBEAT_PERIOD (ms):", (u32_t)(data & MCPR_MCP_HEARTBEAT_STATUS_MCP_HEARTBEAT_PERIOD) >> MCPR_MCP_HEARTBEAT_STATUS_MCP_HEARTBEAT_PERIOD_BITSHIFT, 
            "\n\tVALID   :", (u32_t)(data & MCPR_MCP_HEARTBEAT_STATUS_VALID   ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MCP_HEARTBEAT: 
        // SHARE: Doorbell registoer for MCP heartbeat feature  
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %s %s %s",
            addr, data,
            "\n\tCOUNT  :", (u32_t)(data & MCPR_MCP_HEARTBEAT_MCP_HEARTBEAT_COUNT) >> MCPR_MCP_HEARTBEAT_MCP_HEARTBEAT_COUNT_BITSHIFT, 
            "\n\tINC    :", (u32_t)(data & MCPR_MCP_HEARTBEAT_MCP_HEARTBEAT_INC ) ? "ON":"OFF",
            "\n\tRESET  :", (u32_t)(data & MCPR_MCP_HEARTBEAT_MCP_HEARTBEAT_RESET  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_WATCHDOG_RESET: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tRESET:", (u32_t)(data & MCPR_WATCHDOG_RESET_WATCHDOG_RESET) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_WATCHDOG_CONTROL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %s %s %s %s %s",
            addr, data, 
            "\n\tWATCHDOG_TIMEOUT   :", (u32_t)(data & MCPR_WATCHDOG_CONTROL_WATCHDOG_TIMEOUT) >> MCPR_WATCHDOG_CONTROL_WATCHDOG_TIMEOUT_BITSHIFT,
            "\n\tWATCHDOG_ATTN      :", (u32_t)(data & MCPR_WATCHDOG_CONTROL_WATCHDOG_ATTN  ) ? "ON":"OFF",
            "\n\tMCP_RST_ENABLE     :", (u32_t)(data & MCPR_WATCHDOG_CONTROL_MCP_RST_ENABLE ) ? "ON":"OFF",
            "\n\tWATCHDOG_ENABLE    :", (u32_t)(data & MCPR_WATCHDOG_CONTROL_WATCHDOG_ENABLE) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_ACCESS_LOCK: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tLOCK             :", (u32_t)(data & MCPR_ACCESS_LOCK_LOCK             ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_TOE_ID: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tFUNCTION_ID   :", (u32_t)(data & MCPR_TOE_ID_FUNCTION_ID           ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MAILBOX_CFG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %u",
            addr, data, 
            "\n\tOFFSET:", (u32_t)(data & MCPR_MAILBOX_CFG_MAILBOX_OFFSET) >> MCPR_MAILBOX_CFG_MAILBOX_OFFSET_BITSHIFT,
            "\n\tSIZE  :", (u32_t)(data & MCPR_MAILBOX_CFG_MAILBOX_SIZE  ) >> MCPR_MAILBOX_CFG_MAILBOX_SIZE_BITSHIFT);
        break;
    case MCP_REG_MCPR_MAILBOX_CFG_OTHER_FUNC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %u",
            addr, data, 
            "\n\tOFFSET:", (u32_t)(data & MCPR_MAILBOX_CFG_OTHER_FUNC_MAILBOX_OFFSET) >> MCPR_MAILBOX_CFG_OTHER_FUNC_MAILBOX_OFFSET_BITSHIFT, 
            "\n\tSIZE  :", (u32_t)(data & MCPR_MAILBOX_CFG_OTHER_FUNC_MAILBOX_SIZE) >> MCPR_MAILBOX_CFG_OTHER_FUNC_MAILBOX_SIZE_BITSHIFT);
        break;
    case MCP_REG_MCPR_MCP_DOORBELL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tDOORBELL:", (u32_t)(data & MCPR_MCP_DOORBELL_MCP_DOORBELL    ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_DRIVER_DOORBELL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tDOORBELL:", (u32_t)(data & MCPR_DRIVER_DOORBELL_DRIVER_DOORBELL ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_DRIVER_DOORBELL_OTHER_FUNC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tDOORBELL:", (u32_t)(data & MCPR_DRIVER_DOORBELL_OTHER_FUNC_DRIVER_DOORBELL  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_PORT4MODE_EN: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\t4-port mode:", (u32_t)(data & MCPR_PORT4MODE_EN_FOUR_PORT_MODE  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MCP_VFID: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %s %s %s %s %s",
            addr, data, 
            "\n\tVFID               :", (u32_t)(data & MCPR_MCP_VFID_VFID        ) >> MCPR_MCP_VFID_VFID_BITSHIFT, 
            "\n\tVFID_VALID         :", (u32_t)(data & MCPR_MCP_VFID_VFID_VALID     ) ? "ON":"OFF",
            "\n\tPATHID             :", (u32_t)(data & MCPR_MCP_VFID_PATHID         ) ? "ON":"OFF",
            "\n\tPATH_FORCE         :", (u32_t)(data & MCPR_MCP_VFID_PATH_FORCE     ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_CPU_MODE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tLOCAL_RST           :", (u32_t)(data & MCPR_CPU_MODE_LOCAL_RST           ) ? "ON":"OFF",
            "\n\tSTEP_ENA            :", (u32_t)(data & MCPR_CPU_MODE_STEP_ENA            ) ? "ON":"OFF",
            "\n\tPAGE_0_DATA_ENA     :", (u32_t)(data & MCPR_CPU_MODE_PAGE_0_DATA_ENA     ) ? "ON":"OFF",
            "\n\tPAGE_0_INST_ENA     :", (u32_t)(data & MCPR_CPU_MODE_PAGE_0_INST_ENA     ) ? "ON":"OFF",
            "\n\tMSG_BIT1            :", (u32_t)(data & MCPR_CPU_MODE_MSG_BIT1            ) ? "ON":"OFF",
            "\n\tINTERRUPT_ENA       :", (u32_t)(data & MCPR_CPU_MODE_INTERRUPT_ENA       ) ? "ON":"OFF",
            "\n\tSOFT_HALT           :", (u32_t)(data & MCPR_CPU_MODE_SOFT_HALT           ) ? "ON":"OFF",
            "\n\tBAD_DATA_HALT_ENA   :", (u32_t)(data & MCPR_CPU_MODE_BAD_DATA_HALT_ENA   ) ? "ON":"OFF",
            "\n\tBAD_INST_HALT_ENA   :", (u32_t)(data & MCPR_CPU_MODE_BAD_INST_HALT_ENA   ) ? "ON":"OFF",
            "\n\tFIO_ABORT_HALT_ENA  :", (u32_t)(data & MCPR_CPU_MODE_FIO_ABORT_HALT_ENA  ) ? "ON":"OFF",
            "\n\tSPAD_UNDERFLOW_HALT_ENA:", (u32_t)(data & MCPR_CPU_MODE_SPAD_UNDERFLOW_HALT_ENA) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_CPU_STATE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tBREAKPOINT           :", (u32_t)(data & MCPR_CPU_STATE_BREAKPOINT         ) ? "ON":"OFF",
            "\n\tBAD_INST_HALTED      :", (u32_t)(data & MCPR_CPU_STATE_BAD_INST_HALTED    ) ? "ON":"OFF",
            "\n\tPAGE_0_DATA_HALTED   :", (u32_t)(data & MCPR_CPU_STATE_PAGE_0_DATA_HALTED ) ? "ON":"OFF",
            "\n\tPAGE_0_INST_HALTED   :", (u32_t)(data & MCPR_CPU_STATE_PAGE_0_INST_HALTED ) ? "ON":"OFF",
            "\n\tBAD_DATA_ADDR_HALTED :", (u32_t)(data & MCPR_CPU_STATE_BAD_DATA_ADDR_HALTED  ) ? "ON":"OFF",
            "\n\tBAD_PC_HALTED        :", (u32_t)(data & MCPR_CPU_STATE_BAD_PC_HALTED      ) ? "ON":"OFF",
            "\n\tALIGN_HALTED         :", (u32_t)(data & MCPR_CPU_STATE_ALIGN_HALTED       ) ? "ON":"OFF",
            "\n\tFIO_ABORT_HALTED     :", (u32_t)(data & MCPR_CPU_STATE_FIO_ABORT_HALTED   ) ? "ON":"OFF",
            "\n\tSOFT_HALTED          :", (u32_t)(data & MCPR_CPU_STATE_SOFT_HALTED        ) ? "ON":"OFF",
            "\n\tSPAD_UNDERFLOW       :", (u32_t)(data & MCPR_CPU_STATE_SPAD_UNDERFLOW     ) ? "ON":"OFF",
            "\n\tINTERRRUPT           :", (u32_t)(data & MCPR_CPU_STATE_INTERRRUPT         ) ? "ON":"OFF",
            "\n\tDATA_ACCESS_STALL    :", (u32_t)(data & MCPR_CPU_STATE_DATA_ACCESS_STALL  ) ? "ON":"OFF",
            "\n\tINST_FETCH_STALL     :", (u32_t)(data & MCPR_CPU_STATE_INST_FETCH_STALL   ) ? "ON":"OFF",
            "\n\tBLOCKED_READ         :", (u32_t)(data & MCPR_CPU_STATE_BLOCKED_READ       ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_CPU_EVENT_MASK: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tBREAKPOINT_MASK            :", (u32_t)(data & MCPR_CPU_EVENT_MASK_BREAKPOINT_MASK  ) ? "ON":"OFF",
            "\n\tBAD_INST_HALTED_MASK       :", (u32_t)(data & MCPR_CPU_EVENT_MASK_BAD_INST_HALTED_MASK) ? "ON":"OFF",
            "\n\tPAGE_0_DATA_HALTED_MASK    :",(u32_t)(data & MCPR_CPU_EVENT_MASK_PAGE_0_DATA_HALTED_MASK  ) ? "ON":"OFF",
            "\n\tPAGE_0_INST_HALTED_MASK    :",(u32_t)(data & MCPR_CPU_EVENT_MASK_PAGE_0_INST_HALTED_MASK  ) ? "ON":"OFF",
            "\n\tBAD_DATA_ADDR_HALTED_MASK  :", (u32_t)(data & MCPR_CPU_EVENT_MASK_BAD_DATA_ADDR_HALTED_MASK  ) ? "ON":"OFF",
            "\n\tBAD_PC_HALTED_MASK         :", (u32_t)(data & MCPR_CPU_EVENT_MASK_BAD_PC_HALTED_MASK  ) ? "ON":"OFF",
            "\n\tALIGN_HALTED_MASK          :", (u32_t)(data & MCPR_CPU_EVENT_MASK_ALIGN_HALTED_MASK) ? "ON":"OFF",
            "\n\tFIO_ABORT_MASK             :", (u32_t)(data & MCPR_CPU_EVENT_MASK_FIO_ABORT_MASK) ? "ON":"OFF",
            "\n\tSOFT_HALTED_MASK           :", (u32_t)(data & MCPR_CPU_EVENT_MASK_SOFT_HALTED_MASK ) ? "ON":"OFF",
            "\n\tSPAD_UNDERFLOW_MASK        :", (u32_t)(data & MCPR_CPU_EVENT_MASK_SPAD_UNDERFLOW_MASK ) ? "ON":"OFF",
            "\n\tINTERRUPT_MASK             :", (u32_t)(data & MCPR_CPU_EVENT_MASK_INTERRUPT_MASK) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_CPU_PROGRAM_COUNTER: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);       
        break;
    case MCP_REG_MCPR_CPU_INSTRUCTION: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_CPU_DATA_ACCESS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);               
        break;
    case MCP_REG_MCPR_CPU_INTERRUPT_ENABLE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);          
        break;
    case MCP_REG_MCPR_CPU_INTERRUPT_VECTOR: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);          
        break;
    case MCP_REG_MCPR_CPU_INTERRUPT_SAVED_PC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);         
        break;
    case MCP_REG_MCPR_CPU_HW_BREAKPOINT: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %08x",
            addr, data, 
            "\n\tDISABLE :", (u32_t)(data & MCPR_CPU_HW_BREAKPOINT_DISABLE ) ? "ON":"OFF",
            "\n\tADDRESS :", (u32_t)(data & MCPR_CPU_HW_BREAKPOINT_ADDRESS ) >> MCPR_CPU_HW_BREAKPOINT_ADDRESS_BITSHIFT);
        break;
    case MCP_REG_MCPR_CPU_DEBUG_VECT_PEEK: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %x %s %x %s %s %s %x",
            addr, data, 
            "\n\t1_VALUE   :", (u32_t)(data & MCPR_CPU_DEBUG_VECT_PEEK_1_VALUE) >> MCPR_CPU_DEBUG_VECT_PEEK_1_VALUE_BITSHIFT,
            "\n\t1_PEEK_EN :", (u32_t)(data & MCPR_CPU_DEBUG_VECT_PEEK_1_PEEK_EN) ? "ON":"OFF",
            "\n\t1_SEL     :", (u32_t)(data & MCPR_CPU_DEBUG_VECT_PEEK_1_SEL ) >> MCPR_CPU_DEBUG_VECT_PEEK_1_SEL_BITSHIFT, 
            "\n\t2_VALUE   :", (u32_t)(data & MCPR_CPU_DEBUG_VECT_PEEK_2_VALUE) >> MCPR_CPU_DEBUG_VECT_PEEK_2_VALUE_BITSHIFT, 
            "\n\t2_PEEK_EN :", (u32_t)(data & MCPR_CPU_DEBUG_VECT_PEEK_2_PEEK_EN) ? "ON":"OFF",
            "\n\t2_SEL     :", (u32_t)(data & MCPR_CPU_DEBUG_VECT_PEEK_2_SEL ) >> MCPR_CPU_DEBUG_VECT_PEEK_2_SEL_BITSHIFT);
        break;
    case MCP_REG_MCPR_CPU_LAST_BRANCH_ADDR: 
        sprintf(disp_str, 
        "\n(MCP%05x=%08x)%s %s %s %08x",
        addr, data, 
            "\n\tADDR_TYPE :", (u32_t)(data & MCPR_CPU_LAST_BRANCH_ADDR_TYPE    ) ? "BRANCH":"JUMP", 
            "\n\tADDR_LBA  :", (u32_t)(data & MCPR_CPU_LAST_BRANCH_ADDR_LBA) );
        break;
    case MCP_REG_MCPR_HMCPQ_FTQ_CMD: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
            addr, data, 
            "\n\tOFFSET        :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_OFFSET      ) >> MCPR_HMCPQ_FTQ_CMD_OFFSET_BITSHIFT,
            "\n\tWR_TOP        :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_WR_TOP         ) ? "RD":"R/W", 
            "\n\tSFT_RESET     :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_SFT_RESET      ) ? "ON":"OFF",
            "\n\tRD_DATA       :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_RD_DATA        ) ? "ON":"OFF",
            "\n\tADD_INTERVEN  :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_ADD_INTERVEN   ) ? "ON":"OFF",
            "\n\tADD_DATA      :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_ADD_DATA       ) ? "ON":"OFF",
            "\n\tINTERVENE_CLR :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_INTERVENE_CLR  ) ? "ON":"OFF",
            "\n\tPOP           :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_POP            ) ? "ON":"OFF",
            "\n\tBUSY          :", (u32_t)(data & MCPR_HMCPQ_FTQ_CMD_BUSY           ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_HMCPQ_FTQ_CTL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %u %s %u",
            addr, data, 
            "\n\tINTERVENE      :", (u32_t)(data & MCPR_HMCPQ_FTQ_CTL_INTERVENE      ) ? "ON":"OFF",
            "\n\tOVERFLOW       :", (u32_t)(data & MCPR_HMCPQ_FTQ_CTL_OVERFLOW       ) ? "ON":"OFF",
            "\n\tFORCE_INTERVENE:", (u32_t)(data & MCPR_HMCPQ_FTQ_CTL_FORCE_INTERVENE) ? "ON":"OFF",
            "\n\tMAX_DEPTH      :", (u32_t)(data & MCPR_HMCPQ_FTQ_CTL_MAX_DEPTH   ) >> MCPR_HMCPQ_FTQ_CTL_MAX_DEPTH_BITSHIFT, 
            "\n\tCUR_DEPTH      :", (u32_t)(data & MCPR_HMCPQ_FTQ_CTL_CUR_DEPTH   ) >> MCPR_HMCPQ_FTQ_CTL_CUR_DEPTH_BITSHIFT);
        break;
    case MCP_REG_MCPR_MCPQ_FTQ_CMD: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tOFFSET        :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_OFFSET       ) >> MCPR_MCPQ_FTQ_CMD_OFFSET_BITSHIFT, 
            "\n\tWR_TOP        :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_WR_TOP          ) ? "RD":"R/W", 
            "\n\tSFT_RESET     :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_SFT_RESET       ) ? "ON":"OFF",
            "\n\tRD_DATA       :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_RD_DATA         ) ? "ON":"OFF",
            "\n\tADD_INTERVEN  :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_ADD_INTERVEN    ) ? "ON":"OFF",
            "\n\tADD_DATA      :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_ADD_DATA        ) ? "ON":"OFF",
            "\n\tINTERVENE_CLR :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_INTERVENE_CLR   ) ? "ON":"OFF",
            "\n\tPOP           :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_POP             ) ? "ON":"OFF",
            "\n\tBUSY          :", (u32_t)(data & MCPR_MCPQ_FTQ_CMD_BUSY            ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_MCPQ_FTQ_CTL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %u %s %u",
            addr, data, 
            "\n\tINTERVENE      :", (u32_t)(data & MCPR_MCPQ_FTQ_CTL_INTERVENE       ) ? "ON":"OFF",
            "\n\tOVERFLOW       :", (u32_t)(data & MCPR_MCPQ_FTQ_CTL_OVERFLOW        ) ? "ON":"OFF",
            "\n\tFORCE_INTERVENE:", (u32_t)(data & MCPR_MCPQ_FTQ_CTL_FORCE_INTERVENE ) ? "ON":"OFF",
            "\n\tMAX_DEPTH      :", (u32_t)(data & MCPR_MCPQ_FTQ_CTL_MAX_DEPTH    ) >> MCPR_MCPQ_FTQ_CTL_MAX_DEPTH_BITSHIFT,
            "\n\tCUR_DEPTH      :", (u32_t)(data & MCPR_MCPQ_FTQ_CTL_CUR_DEPTH    ) >> MCPR_MCPQ_FTQ_CTL_CUR_DEPTH_BITSHIFT);
        break;
    case MCP_REG_MCPR_NVM_COMMAND: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tRST        :", (u32_t)(data & MCPR_NVM_COMMAND_RST              ) ? "ON":"OFF",
            "\n\tDONE       :", (u32_t)(data & MCPR_NVM_COMMAND_DONE             ) ? "ON":"OFF",
            "\n\tDOIT       :", (u32_t)(data & MCPR_NVM_COMMAND_DOIT             ) ? "ON":"OFF",
            "\n\tWR         :", (u32_t)(data & MCPR_NVM_COMMAND_WR               ) ? "ON":"OFF",
            "\n\tERASE      :", (u32_t)(data & MCPR_NVM_COMMAND_ERASE            ) ? "ON":"OFF",
            "\n\tFIRST      :", (u32_t)(data & MCPR_NVM_COMMAND_FIRST            ) ? "ON":"OFF",
            "\n\tLAST       :", (u32_t)(data & MCPR_NVM_COMMAND_LAST             ) ? "ON":"OFF",
            "\n\tWREN       :", (u32_t)(data & MCPR_NVM_COMMAND_WREN             ) ? "ON":"OFF",
            "\n\tWRDI       :", (u32_t)(data & MCPR_NVM_COMMAND_WRDI             ) ? "ON":"OFF",
            "\n\tRD_ID      :", (u32_t)(data & MCPR_NVM_COMMAND_RD_ID            ) ? "ON":"OFF",
            "\n\tRD_STATUS  :", (u32_t)(data & MCPR_NVM_COMMAND_RD_STATUS        ) ? "ON":"OFF",
            "\n\tMODE_256   :", (u32_t)(data & MCPR_NVM_COMMAND_MODE_256         ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_NVM_STATUS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, "\n\tFSM_STATE:", 
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_IDLE               ? "IDLE" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_CMD0               ? "CMD0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_CMD1               ? "CMD1" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_CMD_FINISH0        ? "CMD_FINISH0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_CMD_FINISH1        ? "CMD_FINISH1" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_ADDR0              ? "ADDR0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_WRITE_DATA0        ? "WRITE_DATA0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_WRITE_DATA1        ? "WRITE_DATA1" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_WRITE_DATA2        ? "WRITE_DATA2" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_DATA0         ? "READ_DATA0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_DATA1         ? "READ_DATA1" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_DATA2         ? "READ_DATA2" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID0  ?"READ_STATUS_RDID0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID1  ?"READ_STATUS_RDID1" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID2  ?"READ_STATUS_RDID2" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID3  ?"READ_STATUS_RDID3" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID4  ?"READ_STATUS_RDID4" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_CHECK_BUSY0        ? "CHECK_BUSY0" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_ST_WREN            ? "ST_WREN" :
                ((u32_t)(data & MCPR_NVM_STATUS_SPI_FSM_STATE) == MCPR_NVM_STATUS_SPI_FSM_STATE_SPI_WAIT               ? "WAIT" : "INVALID")))))))))))))))))))));
        break;
    case MCP_REG_MCPR_NVM_WRITE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
                "\n\tBIT_BANG :",(u32_t)(data & MCPR_NVM_WRITE_NVM_WRITE_VALUE_BIT_BANG) ? "ON":"OFF",
                "\n\tSI       :",(u32_t)(data & MCPR_NVM_WRITE_NVM_WRITE_VALUE_SI      ) ? "ON":"OFF",
                "\n\tSO       :",(u32_t)(data & MCPR_NVM_WRITE_NVM_WRITE_VALUE_SO      ) ? "ON":"OFF",
                "\n\tCS_B     :",(u32_t)(data & MCPR_NVM_WRITE_NVM_WRITE_VALUE_CS_B    ) ? "ON":"OFF",
                "\n\tSCLK     :",(u32_t)(data & MCPR_NVM_WRITE_NVM_WRITE_VALUE_SCLK    ) ? "ON":"OFF" );
        break;
    case MCP_REG_MCPR_NVM_ADDR: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
                "\n\tBIT_BANG :",(u32_t)(data & MCPR_NVM_ADDR_NVM_ADDR_VALUE_BIT_BANG) ? "ON":"OFF",
                "\n\tSI       :",(u32_t)(data & MCPR_NVM_ADDR_NVM_ADDR_VALUE_SI      ) ? "ON":"OFF",
                "\n\tSO       :",(u32_t)(data & MCPR_NVM_ADDR_NVM_ADDR_VALUE_SO      ) ? "ON":"OFF",
                "\n\tCS_B     :",(u32_t)(data & MCPR_NVM_ADDR_NVM_ADDR_VALUE_CS_B    ) ? "ON":"OFF",
                "\n\tSCLK     :",(u32_t)(data & MCPR_NVM_ADDR_NVM_ADDR_VALUE_SCLK    ) ? "ON":"OFF" );
        break;
    case MCP_REG_MCPR_NVM_READ: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
                "\n\tBIT_BANG :",(u32_t)(data & MCPR_NVM_READ_NVM_READ_VALUE_BIT_BANG)  ? "ON":"OFF",
                "\n\tSI       :",(u32_t)(data & MCPR_NVM_READ_NVM_READ_VALUE_SI      )  ? "ON":"OFF",
                "\n\tSO       :",(u32_t)(data & MCPR_NVM_READ_NVM_READ_VALUE_SO      )  ? "ON":"OFF",
                "\n\tCS_B     :",(u32_t)(data & MCPR_NVM_READ_NVM_READ_VALUE_CS_B    )  ? "ON":"OFF",
                "\n\tSCLK     :",(u32_t)(data & MCPR_NVM_READ_NVM_READ_VALUE_SCLK    )  ? "ON":"OFF" );
        break;
    case MCP_REG_MCPR_NVM_CFG1: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %x %s %x %s %x %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tFLASH_MODE       :", (u32_t)(data & MCPR_NVM_CFG1_FLASH_MODE          ) ? "ON":"OFF",
            "\n\tBUFFER_MODE      :", (u32_t)(data & MCPR_NVM_CFG1_BUFFER_MODE         ) ? "ON":"OFF",
            "\n\tPASS_MODE        :", (u32_t)(data & MCPR_NVM_CFG1_PASS_MODE           ) ? "ON":"OFF",
            "\n\tBITBANG_MODE     :", (u32_t)(data & MCPR_NVM_CFG1_BITBANG_MODE        ) ? "ON":"OFF",
            "\n\tSTATUS_BIT       :", (u32_t)(data & MCPR_NVM_CFG1_STATUS_BIT       ) >> MCPR_NVM_CFG1_STATUS_BIT_BITSHIFT,
            "\n\tSPI_CLK_DIV      :", (u32_t)(data & MCPR_NVM_CFG1_SPI_CLK_DIV      ) >> MCPR_NVM_CFG1_SPI_CLK_DIV_BITSHIFT,
            "\n\tSEE_CLK_DIV      :", (u32_t)(data & MCPR_NVM_CFG1_SEE_CLK_DIV      ) >> MCPR_NVM_CFG1_SEE_CLK_DIV_BITSHIFT, 
            "\n\tSTRAP_CONTROL_0  :", (u32_t)(data & MCPR_NVM_CFG1_STRAP_CONTROL_0     ) ? "ON":"OFF",
            "\n\tPROTECT_MODE     :", (u32_t)(data & MCPR_NVM_CFG1_PROTECT_MODE        ) ? "ON":"OFF",
            "\n\tFLASH_SIZE       :", (u32_t)(data & MCPR_NVM_CFG1_FLASH_SIZE          ) ? "ON":"OFF",
            "\n\tFW_USTRAP_1      :", (u32_t)(data & MCPR_NVM_CFG1_FW_USTRAP_1         ) ? "ON":"OFF",
            "\n\tFW_USTRAP_0      :", (u32_t)(data & MCPR_NVM_CFG1_FW_USTRAP_0         ) ? "ON":"OFF",
            "\n\tFW_USTRAP_2      :", (u32_t)(data & MCPR_NVM_CFG1_FW_USTRAP_2         ) ? "ON":"OFF",
            "\n\tFW_USTRAP_3      :", (u32_t)(data & MCPR_NVM_CFG1_FW_USTRAP_3         ) ? "ON":"OFF",
            "\n\tFW_FLASH_TYPE_EN :", (u32_t)(data & MCPR_NVM_CFG1_FW_FLASH_TYPE_EN    ) ? "ON":"OFF",
            "\n\tCOMPAT_BYPASSS   :", (u32_t)(data & MCPR_NVM_CFG1_COMPAT_BYPASSS      ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_NVM_CFG2: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x %s %x",
            addr, data, 
            "\n\tERASE_CMD     :", (u32_t)(data & MCPR_NVM_CFG2_ERASE_CMD        ) >> MCPR_NVM_CFG2_ERASE_CMD_BITSHIFT,
            "\n\tSTATUS_CMD    :", (u32_t)(data & MCPR_NVM_CFG2_STATUS_CMD       ) >> MCPR_NVM_CFG2_STATUS_CMD_BITSHIFT, 
            "\n\tREAD_ID       :", (u32_t)(data & MCPR_NVM_CFG2_READ_ID          ) >> MCPR_NVM_CFG2_READ_ID_BITSHIFT);
        break;
    case MCP_REG_MCPR_NVM_CFG3: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x %s %x",
            addr, data, 
            "\n\tBUFFER_RD_CMD  :", (u32_t)(data & MCPR_NVM_CFG3_BUFFER_RD_CMD    ) >> MCPR_NVM_CFG3_BUFFER_RD_CMD_BITSHIFT, 
            "\n\tWRITE_CMD      :", (u32_t)(data & MCPR_NVM_CFG3_WRITE_CMD        ) >> MCPR_NVM_CFG3_WRITE_CMD_BITSHIFT,
            "\n\tREAD_CMD       :", (u32_t)(data & MCPR_NVM_CFG3_READ_CMD         ) >> MCPR_NVM_CFG3_READ_CMD_BITSHIFT);
        break;
    case MCP_REG_MCPR_NVM_SW_ARB: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tARB_REQ_SET0  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_SET0      ) ? "ON":"OFF",
            "\n\tARB_REQ_SET1  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_SET1      ) ? "ON":"OFF",
            "\n\tARB_REQ_SET2  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_SET2      ) ? "ON":"OFF",
            "\n\tARB_REQ_SET3  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_SET3      ) ? "ON":"OFF",
            "\n\tARB_REQ_CLR0  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_CLR0      ) ? "ON":"OFF",
            "\n\tARB_REQ_CLR1  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_CLR1      ) ? "ON":"OFF",
            "\n\tARB_REQ_CLR2  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_CLR2      ) ? "ON":"OFF",
            "\n\tARB_REQ_CLR3  :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_REQ_CLR3      ) ? "ON":"OFF",
            "\n\tARB_ARB0      :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_ARB0          ) ? "ON":"OFF",
            "\n\tARB_ARB1      :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_ARB1          ) ? "ON":"OFF",
            "\n\tARB_ARB2      :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_ARB2          ) ? "ON":"OFF",
            "\n\tARB_ARB3      :", (u32_t)(data & MCPR_NVM_SW_ARB_ARB_ARB3          ) ? "ON":"OFF",
            "\n\tREQ0          :", (u32_t)(data & MCPR_NVM_SW_ARB_REQ0              ) ? "ON":"OFF",
            "\n\tREQ1          :", (u32_t)(data & MCPR_NVM_SW_ARB_REQ1              ) ? "ON":"OFF",
            "\n\tREQ2          :", (u32_t)(data & MCPR_NVM_SW_ARB_REQ2              ) ? "ON":"OFF",
            "\n\tREQ3          :", (u32_t)(data & MCPR_NVM_SW_ARB_REQ3              ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_NVM_ACCESS_ENABLE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s ",
            addr, data, 
            "\n\tENABLE_EN     :", (u32_t)(data & MCPR_NVM_ACCESS_ENABLE_EN     ) ? "ON":"OFF",
            "\n\tENABLE_WR_EN  :", (u32_t)(data & MCPR_NVM_ACCESS_ENABLE_WR_EN  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_NVM_WRITE1: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x",
            addr, data, 
            "\n\tWREN_CMD   :", (u32_t)(data & MCPR_NVM_WRITE1_WREN_CMD) >> MCPR_NVM_WRITE1_WREN_CMD_BITSHIFT,
            "\n\tWRDI_CMD   :", (u32_t)(data & MCPR_NVM_WRITE1_WRDI_CMD) >> MCPR_NVM_WRITE1_WRDI_CMD_BITSHIFT);
        break;
    case MCP_REG_MCPR_NVM_CFG4: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %x",
            addr, data, 
            "\n\tFLASH_SIZE          :", 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_1MBIT  ? "1MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_2MBIT  ? "2MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_4MBIT  ? "4MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_8MBIT  ? "8MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_16MBIT ? "16MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_32MBIT ? "32MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_64MBIT ? "64MBIT" : 
                ((u32_t)(data & MCPR_NVM_CFG4_FLASH_SIZE) == MCPR_NVM_CFG4_FLASH_SIZE_128MBIT? "128MBIT" : "INVALID")))))))),            
            "\n\tFLASH_VENDOR        :", (u32_t)(data & MCPR_NVM_CFG4_FLASH_VENDOR  ) ? "ATMEL":"ST", 
            "\n\tMODE_256_EMPTY_BIT_LOC:", 
                ((u32_t)(data & MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) >> MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BITSHIFT == MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT8  ? "BIT8 For 264 Byte pages"   : 
                ((u32_t)(data & MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) >> MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BITSHIFT == MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT9  ? "BIT9 For 528 Byte pages"   : 
                ((u32_t)(data & MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) >> MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BITSHIFT == MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT10 ? "BIT10 For 1056 Byte pages" : 
                ((u32_t)(data & MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) >> MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BITSHIFT == MCPR_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT11 ? "BIT11 For 2112 Byte pages" : "INVALID")))), 
            "\n\tSTATUS_BIT_POLARITY :", (u32_t)(data & MCPR_NVM_CFG4_STATUS_BIT_POLARITY ) ? "1 means ready":"0 means ready", 
            "\n\tRESERVED            :", (u32_t)(data & MCPR_NVM_CFG4_RESERVED         ) >> MCPR_NVM_CFG4_RESERVED_BITSHIFT);
        break;
    case MCP_REG_MCPR_NVM_RECONFIG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %x %s %x %s %s ",
            addr, data, 
            "\n\tORIG_STRAP_VALUE       :", data & MCPR_NVM_RECONFIG_ORIG_STRAP_VALUE  ? "ATMEL":"ST",
            "\n\tRECONFIG_STRAP_VALUE   :", (u32_t)(data & MCPR_NVM_RECONFIG_RECONFIG_STRAP_VALUE) >> MCPR_NVM_RECONFIG_RECONFIG_STRAP_VALUE_BITSHIFT, 
            "\n\tRESERVED               :", (u32_t)(data & MCPR_NVM_RECONFIG_RESERVED     ) >> MCPR_NVM_RECONFIG_RESERVED_BITSHIFT, 
            "\n\tRECONFIG_DONE          :", (u32_t)(data & MCPR_NVM_RECONFIG_RECONFIG_DONE   ) ? "CFG":"RESET");
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_CTRL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %x %s %x %s %s %s %s %s %s %s %s",
            addr, data, 
            "\n\tENA          :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_ENA       ) ? "ON":"OFF",
            "\n\tBFRD         :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_BFRD      ) ? "ON":"OFF",
            "\n\tARB_NUM      :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_ARB_NUM) >> MCPR_ERNGN_EXP_ROM_CTRL_ARB_NUM_BITSHIFT, 
            "\n\tSTATE        :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_STATE  ) >> MCPR_ERNGN_EXP_ROM_CTRL_STATE_BITSHIFT, 
            "\n\tCACHE_VALID  :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_CACHE_VALID  ) ? "ON":"OFF",
            "\n\tARB_TIMEOUT  :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_ARB_TIMEOUT  ) ? "ON":"OFF",
            "\n\tREAD_TIMEOUT :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_READ_TIMEOUT ) ? "ON":"OFF",
            "\n\tACTIVE       :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CTRL_ACTIVE    ) ? "ON":"OFF");                   
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_BADDR: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x",
            addr, data, 
            "\n\tBADDR_VALUE :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_BADDR_VALUE ) >> MCPR_ERNGN_EXP_ROM_BADDR_VALUE_BITSHIFT); 
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_CFG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x",
            addr, data, 
            "\n\tCFG_ARB_TIMEOUT_SHFT :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CFG_ARB_TIMEOUT_SHFT) >> MCPR_ERNGN_EXP_ROM_CFG_ARB_TIMEOUT_SHFT_BITSHIFT,
            "\n\tCFG_READ_TIMEOUT_SHFT:", (u32_t)(data & MCPR_ERNGN_EXP_ROM_CFG_READ_TIMEOUT_SHFT) >> MCPR_ERNGN_EXP_ROM_CFG_READ_TIMEOUT_SHFT_BITSHIFT);
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_ADR: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x %s %x %s %s",
            addr, data, 
            "\n\tADR_ADDRESS    :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_ADR_ADDRESS ) >> MCPR_ERNGN_EXP_ROM_ADR_ADDRESS_BITSHIFT, 
            "\n\tADR_ADDR_SIZE  :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_ADR_ADDR_SIZE) >> MCPR_ERNGN_EXP_ROM_ADR_ADDR_SIZE_BITSHIFT, 
            "\n\tADR_ACT_FUNC   :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_ADR_ACT_FUNC) >> MCPR_ERNGN_EXP_ROM_ADR_ACT_FUNC_BITSHIFT, 
            "\n\tADR_REQ        :", (u32_t)(data & MCPR_ERNGN_EXP_ROM_ADR_REQ        ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_DATA0: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);       
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_DATA1: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)", 
            addr, data);          
        break;
    case MCP_REG_MCPR_ERNGN_EXP_ROM_DATA2: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_SMBUS_CONFIG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %u %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tHW_ARP_ASSIGN_ADDR :", (u32_t)(data & MCPR_SMBUS_CONFIG_HW_ARP_ASSIGN_ADDR ) ? "ON":"OFF",
            "\n\tARP_EN0            :", (u32_t)(data & MCPR_SMBUS_CONFIG_ARP_EN0         ) ? "ON":"OFF",
            "\n\tARP_EN1            :", (u32_t)(data & MCPR_SMBUS_CONFIG_ARP_EN1         ) ? "ON":"OFF",
            "\n\tMASTER_RTRY_CNT    :", (u32_t)(data & MCPR_SMBUS_CONFIG_MASTER_RTRY_CNT) >> MCPR_SMBUS_CONFIG_MASTER_RTRY_CNT_BITSHIFT,
            "\n\tTIMESTAMP_CNT_EN   :", (u32_t)(data & MCPR_SMBUS_CONFIG_TIMESTAMP_CNT_EN) ? "ON":"OFF",
            "\n\tPROMISCOUS_MODE    :", (u32_t)(data & MCPR_SMBUS_CONFIG_PROMISCOUS_MODE ) ? "ON":"OFF",
            "\n\tEN_NIC_SMB_ADDR_0  :", (u32_t)(data & MCPR_SMBUS_CONFIG_EN_NIC_SMB_ADDR_0  ) ? "ON":"OFF",
            "\n\tBIT_BANG_EN        :", (u32_t)(data & MCPR_SMBUS_CONFIG_BIT_BANG_EN     ) ? "ON":"OFF",
            "\n\tSMB_EN             :", (u32_t)(data & MCPR_SMBUS_CONFIG_SMB_EN          ) ? "ON":"OFF",
            "\n\tRESET              :", (u32_t)(data & MCPR_SMBUS_CONFIG_RESET           ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_TIMING_CONFIG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x %s %x %s %s",
            addr, data, 
            "\n\tCONFIG_SMBUS_IDLE_TIME         :", (u32_t)(data & MCPR_SMBUS_TIMING_CONFIG_SMBUS_IDLE_TIME) >> MCPR_SMBUS_TIMING_CONFIG_SMBUS_IDLE_TIME_BITSHIFT,
            "\n\tCONFIG_PERIODIC_SLAVE_STRETCH  :", (u32_t)(data & MCPR_SMBUS_TIMING_CONFIG_PERIODIC_SLAVE_STRETCH) >> MCPR_SMBUS_TIMING_CONFIG_PERIODIC_SLAVE_STRETCH_BITSHIFT,
            "\n\tCONFIG_RANDOM_SLAVE_STRETCH    :", (u32_t)(data & MCPR_SMBUS_TIMING_CONFIG_RANDOM_SLAVE_STRETCH) >> MCPR_SMBUS_TIMING_CONFIG_RANDOM_SLAVE_STRETCH_BITSHIFT, 
            "\n\tCONFIG_MODE_400                :", (u32_t)(data & MCPR_SMBUS_TIMING_CONFIG_MODE_400 ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_ADDRESS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %x %s %s %s %x %s %s %s %x %s %s",
            addr, data, 
            "\n\tNIC_SMB_ADDR0     :", (u32_t)(data & MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR0) >> MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR0_BITSHIFT, 
            "\n\tEN_NIC_SMB_ADDR0  :", (u32_t)(data & MCPR_SMBUS_ADDRESS_EN_NIC_SMB_ADDR0  ) ? "ON":"OFF",
            "\n\tNIC_SMB_ADDR1     :", (u32_t)(data & MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR1) >> MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR1_BITSHIFT, 
            "\n\tEN_NIC_SMB_ADDR1  :", (u32_t)(data & MCPR_SMBUS_ADDRESS_EN_NIC_SMB_ADDR1  ) ? "ON":"OFF",
            "\n\tNIC_SMB_ADDR2     :", (u32_t)(data & MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR2) >> MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR2_BITSHIFT,
            "\n\tEN_NIC_SMB_ADDR2  :", (u32_t)(data & MCPR_SMBUS_ADDRESS_EN_NIC_SMB_ADDR2  ) ? "ON":"OFF",
            "\n\tNIC_SMB_ADDR3     :", (u32_t)(data & MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR3) >> MCPR_SMBUS_ADDRESS_NIC_SMB_ADDR3_BITSHIFT,
            "\n\tEN_NIC_SMB_ADDR3  :", (u32_t)(data & MCPR_SMBUS_ADDRESS_EN_NIC_SMB_ADDR3  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_FIFO_CONTROL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x %s %s %s %s ",
            addr, data, 
            "\n\tRX_FIFO_THRESHOLD:", (u32_t)(data & MCPR_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_FIFO_THRESHOLD) >> MCPR_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_FIFO_THRESHOLD_BITSHIFT, 
            "\n\tRX_PKT_COUNT     :", (u32_t)(data & MCPR_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_PKT_COUNT) >> MCPR_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_PKT_COUNT_BITSHIFT,
            "\n\tTX_FIFO_FLUSH    :", (u32_t)(data & MCPR_SMBUS_MASTER_FIFO_CONTROL_MASTER_TX_FIFO_FLUSH  ) ? "ON":"OFF",
            "\n\tRX_FIFO_FLUSH    :", (u32_t)(data & MCPR_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_FIFO_FLUSH  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_FIFO_CONTROL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %x %s %s %s %s ",
            addr, data, 
            "\n\tRX_FIFO_THRESHOLD:", (u32_t)(data & MCPR_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_FIFO_THRESHOLD) >> MCPR_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_FIFO_THRESHOLD_BITSHIFT,
            "\n\tRX_PKT_COUNT     :", (u32_t)(data & MCPR_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_PKT_COUNT) >> MCPR_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_PKT_COUNT_BITSHIFT,
            "\n\tTX_FIFO_FLUSH    :", (u32_t)(data & MCPR_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_TX_FIFO_FLUSH  ) ? "ON":"OFF",
            "\n\tRX_FIFO_FLUSH    :", (u32_t)(data & MCPR_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_FIFO_FLUSH  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_BIT_BANG_CONTROL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tSMBDAT_OUT_EN:", (u32_t)(data & MCPR_SMBUS_BIT_BANG_CONTROL_SMBDAT_OUT_EN) ? "ON":"OFF",
            "\n\tSMBDAT_IN    :", (u32_t)(data & MCPR_SMBUS_BIT_BANG_CONTROL_SMBDAT_IN) ? "ON":"OFF",
            "\n\tSMBCLK_OUT_EN:", (u32_t)(data & MCPR_SMBUS_BIT_BANG_CONTROL_SMBCLK_OUT_EN) ? "ON":"OFF",
            "\n\tSMBCLK_IN    :", (u32_t)(data & MCPR_SMBUS_BIT_BANG_CONTROL_SMBCLK_IN) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_WATCHDOG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data); 
        break;
    case MCP_REG_MCPR_SMBUS_HEARTBEAT: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data); 
        break;
    case MCP_REG_MCPR_SMBUS_POLL_ASF: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data); 
        break;
    case MCP_REG_MCPR_SMBUS_POLL_LEGACY: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_SMBUS_RETRAN: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_SMBUS_TIMESTAMP: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_COMMAND: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tRD_BYTE_COUNT :", (u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_RD_BYTE_COUNT) >> MCPR_SMBUS_MASTER_COMMAND_RD_BYTE_COUNT_BITSHIFT, 
            "\n\tPEC           :", (u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_PEC     ) ? "ON":"OFF",
            "\n\tSMBUS_PROTOCOL:", 
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0000  ? "Quick Command" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0001  ? "Send Byte" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0010  ? "Receive Byte" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0011  ? "Write Byte" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0100  ? "Read Byte" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0101  ? "Write Word" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0110  ? "Read Word" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0111  ? "Block Write" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1000  ? "Block Read" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1001  ? "Process Call" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1010  ?"Block Write-Block Read Process Call" :
                 (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) >> MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1011  ? "Host Notify" : "invalid")))))))))))),
            "\n\tSTATUS        :", 
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_000  ? "Transaction completed successfully":
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_001  ? "Lost Arbitration.":
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_010  ? "NACK detected after (slave address) first byte.":
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_011  ? "NACK detected after byte other than first":
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_100  ? "TTIMEOUT_MIN exceeded":
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_101  ? "TLOW:MEXT exceeded. Used for non-standard transactions where payload is longer than the TX":
                (((u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_STATUS) >> MCPR_SMBUS_MASTER_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_COMMAND_STATUS_110  ? "TLOW:MEXT exceeded failed due the Master Rx FIFO being full for more than 10ms":"Reserved\n"))))))),
            "\n\tABORT         :", (u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_ABORT   ) ? "ON":"OFF",
            "\n\tSTART_BUSY    :", (u32_t)(data & MCPR_SMBUS_MASTER_COMMAND_START_BUSY ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_COMMAND: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s",
            addr, data, 
            "\n\tPEC           :", (u32_t)(data & MCPR_SMBUS_SLAVE_COMMAND_PEC      ) ? "ON":"OFF",
            "\n\tSTATUS        :", 
                  (((u32_t)(data & MCPR_SMBUS_SLAVE_COMMAND_STATUS) >> MCPR_SMBUS_SLAVE_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_COMMAND_STATUS_000  ? "Transaction completed successfully":
                  (((u32_t)(data & MCPR_SMBUS_SLAVE_COMMAND_STATUS) >> MCPR_SMBUS_SLAVE_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_COMMAND_STATUS_101  ? "TTIMEOUT_MIN exceeded":
                  (((u32_t)(data & MCPR_SMBUS_SLAVE_COMMAND_STATUS) >> MCPR_SMBUS_SLAVE_COMMAND_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_COMMAND_STATUS_111  ? "Master aborted read transaction":"Invalid"))),
            "\n\tABORT         :", (u32_t)(data & MCPR_SMBUS_SLAVE_COMMAND_ABORT    ) ? "ON":"OFF",
            "\n\tSTART         :", (u32_t)(data & MCPR_SMBUS_SLAVE_COMMAND_START    ) ? "ON":"OFF");      
        break;
    case MCP_REG_MCPR_SMBUS_EVENT_ENABLE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tWATCHDOG_TO_EN             :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_WATCHDOG_TO_EN  ) ? "ON":"OFF",
            "\n\tHEARTBEAT_TO_EN            :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_HEARTBEAT_TO_EN ) ? "ON":"OFF",
            "\n\tPOLL_ASF_TO_EN             :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_POLL_ASF_TO_EN  ) ? "ON":"OFF",
            "\n\tPOLL_LEGACY_TO_EN          :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_POLL_LEGACY_TO_EN) ? "ON":"OFF",
            "\n\tRETRANSMIT_TO_EN           :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_RETRANSMIT_TO_EN) ? "ON":"OFF",
            "\n\tSLAVE_ARP_EVENT_EN         :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_ARP_EVENT_EN  ) ? "ON":"OFF",
            "\n\tSLAVE_RD_EVENT_EN          :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_RD_EVENT_EN) ? "ON":"OFF",
            "\n\tSLAVE_TX_UNDERRUN_EN       :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_TX_UNDERRUN_EN  ) ? "ON":"OFF",
            "\n\tSLAVE_START_BUSY_EN        :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_START_BUSY_EN  ) ? "ON":"OFF",
            "\n\tSLAVE_RX_EVENT_EN          :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_RX_EVENT_EN) ? "ON":"OFF",
            "\n\tSLAVE_RX_THRESHOLD_HIT_EN  : ", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_RX_THRESHOLD_HIT_EN  ) ? "ON":"OFF",
            "\n\tSLAVE_RX_FIFO_FULL_EN      :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_SLAVE_RX_FIFO_FULL_EN  ) ? "ON":"OFF",
            "\n\tMASTER_TX_UNDERRUN_EN      :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_MASTER_TX_UNDERRUN_EN  ) ? "ON":"OFF",
            "\n\tMASTER_START_BUSY_EN       :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_MASTER_START_BUSY_EN  ) ? "ON":"OFF",
            "\n\tMASTER_RX_EVENT_EN         :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_MASTER_RX_EVENT_EN  ) ? "ON":"OFF",
            "\n\tMASTER_RX_THRESHOLD_HIT_EN :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_MASTER_RX_THRESHOLD_HIT_EN  ) ? "ON":"OFF",
            "\n\tMASTER_RX_FIFO_FULL_EN     :", (u32_t)(data & MCPR_SMBUS_EVENT_ENABLE_MASTER_RX_FIFO_FULL_EN  ) ? "ON":"OFF");       
        break;
    case MCP_REG_MCPR_SMBUS_EVENT_STATUS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tWATCHDOG_TO            :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_WATCHDOG_TO  ) ? "ON":"OFF",
            "\n\tHEARTBEAT_TO           :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_HEARTBEAT_TO ) ? "ON":"OFF",
            "\n\tPOLL_ASF_TO            :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_POLL_ASF_TO  ) ? "ON":"OFF",
            "\n\tPOLL_LEGACY_TO         :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_POLL_LEGACY_TO  ) ? "ON":"OFF",
            "\n\tRETRANSMIT_TO          :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_RETRANSMIT_TO) ? "ON":"OFF",
            "\n\tSLAVE_ARP_EVENT        :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_ARP_EVENT ) ? "ON":"OFF",
            "\n\tSLAVE_RD_EVENT         :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_RD_EVENT  ) ? "ON":"OFF",
            "\n\tSLAVE_TX_UNDERRUN      :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_TX_UNDERRUN) ? "ON":"OFF",
            "\n\tSLAVE_START_BUSY       :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_START_BUSY) ? "ON":"OFF",
            "\n\tSLAVE_RX_EVENT         :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_RX_EVENT  ) ? "ON":"OFF",
            "\n\tSLAVE_RX_THRESHOLD_HIT :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_RX_THRESHOLD_HIT  ) ? "ON":"OFF",
            "\n\tSLAVE_RX_FIFO_FULL     :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_SLAVE_RX_FIFO_FULL  ) ? "ON":"OFF",
            "\n\tMASTER_TX_UNDERRUN     :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_MASTER_TX_UNDERRUN  ) ? "ON":"OFF",
            "\n\tMASTER_START_BUSY      :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_MASTER_START_BUSY) ? "ON":"OFF",
            "\n\tMASTER_RX_EVENT        :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_MASTER_RX_EVENT ) ? "ON":"OFF",
            "\n\tMASTER_RX_THRESHOLD_HIT:",(u32_t)(data & MCPR_SMBUS_EVENT_STATUS_MASTER_RX_THRESHOLD_HIT  ) ? "ON":"OFF",
            "\n\tMASTER_RX_FIFO_FULL    :", (u32_t)(data & MCPR_SMBUS_EVENT_STATUS_MASTER_RX_FIFO_FULL  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_DATA_WRITE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s ",
            addr, data, 
            "\n\tMASTER_SMBUS_WR_DATA   :", (u32_t)(data & MCPR_SMBUS_MASTER_DATA_WRITE_MASTER_SMBUS_WR_DATA) >> MCPR_SMBUS_MASTER_DATA_WRITE_MASTER_SMBUS_WR_DATA_BITSHIFT, 
            "\n\tWR_STATUS              :", (u32_t)(data & MCPR_SMBUS_MASTER_DATA_WRITE_WR_STATUS  ) ? "End of SMBUS transaction":"Byte other then last in an WMBUS transaction 1");
        break;
    case MCP_REG_MCPR_SMBUS_MASTER_DATA_READ: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %s",
            addr, data, 
            "\n\tMASTER_SMBUS_RD_DATA   :", (u32_t)(data & MCPR_SMBUS_MASTER_DATA_READ_MASTER_SMBUS_RD_DATA) >> MCPR_SMBUS_MASTER_DATA_READ_MASTER_SMBUS_RD_DATA_BITSHIFT,
            "\n\tPEC_ERR                :", (u32_t)(data & MCPR_SMBUS_MASTER_DATA_READ_PEC_ERR  ) ? "ON":"OFF",
            "\n\tRD_STATUS              :", 
              (((u32_t)(data & MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS) >> MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_00 ? "FIFO Empty": 
              (((u32_t)(data & MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS) >> MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_01 ? "Start of SMBUS transaction": 
              (((u32_t)(data & MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS) >> MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_10 ? "Byte other then first or last in an SMBUS transaction": 
              (((u32_t)(data & MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS) >> MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_BITSHIFT) == MCPR_SMBUS_MASTER_DATA_READ_RD_STATUS_11 ? "End of SMBUS transaction": "Invalid")))));
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_DATA_WRITE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s ",
            addr, data, 
            "\n\tSLAVE_SMBUS_WR_DATA: ", (u32_t)(data & MCPR_SMBUS_SLAVE_DATA_WRITE_SLAVE_SMBUS_WR_DATA) >> MCPR_SMBUS_SLAVE_DATA_WRITE_SLAVE_SMBUS_WR_DATA_BITSHIFT,
            "\n\tWR_STATUS          :", (u32_t)(data & MCPR_SMBUS_SLAVE_DATA_WRITE_WR_STATUS) ? "End of SMBUS transaction":"Byte other then last in an SMBUS transaction");
        break;
    case MCP_REG_MCPR_SMBUS_SLAVE_DATA_READ: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s",
            addr, data, 
            "\n\tSLAVE_SMBUS_RD_DATA:", (u32_t)(data & MCPR_SMBUS_SLAVE_DATA_READ_SLAVE_SMBUS_RD_DATA) >> MCPR_SMBUS_SLAVE_DATA_READ_SLAVE_SMBUS_RD_DATA_BITSHIFT,
            "\n\tERR_STATUS         :",  
              (((data & MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS) >> MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_00 ? "FIFO Empty": 
              (((data & MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS) >> MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_01 ? "Start of SMBUS transaction": 
              (((data & MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS) >> MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_10 ? "Byte other then first or last in an SMBUS transaction": 
              (((data & MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS) >> MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_BITSHIFT) == MCPR_SMBUS_SLAVE_DATA_READ_ERR_STATUS_11 ? "End of SMBUS transaction": "Invalid")))));
        break;
    case MCP_REG_MCPR_SMBUS_ARP_STATE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s%s %s%s %s%s %s%s",
            addr, data, 
            "\n\tAV_FLAG0     :", (data & MCPR_SMBUS_ARP_STATE_AV_FLAG0  ) ? "ON":"OFF",
            "\n\tAR_FLAG0     :", (data & MCPR_SMBUS_ARP_STATE_AR_FLAG0  ) ? "ON":"OFF",
            "\n\tAV_FLAG1     :", (data & MCPR_SMBUS_ARP_STATE_AV_FLAG1  ) ? "ON":"OFF",
            "\n\tAR_FLAG1     :", (data & MCPR_SMBUS_ARP_STATE_AR_FLAG1  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_3: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_12     :", (u32_t)(data & MCPR_SMBUS_UDID0_3_BYTE_12  ) >> MCPR_SMBUS_UDID0_3_BYTE_12_BITSHIFT,
            "\n\tBYTE_13     :", (u32_t)(data & MCPR_SMBUS_UDID0_3_BYTE_13  ) >> MCPR_SMBUS_UDID0_3_BYTE_13_BITSHIFT,
            "\n\tBYTE_14     :", (u32_t)(data & MCPR_SMBUS_UDID0_3_BYTE_14  ) >> MCPR_SMBUS_UDID0_3_BYTE_14_BITSHIFT, 
            "\n\tBYTE_15     :", (u32_t)(data & MCPR_SMBUS_UDID0_3_BYTE_15  ) >> MCPR_SMBUS_UDID0_3_BYTE_15_BITSHIFT);
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_2: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_8      :", (u32_t)(data & MCPR_SMBUS_UDID0_2_BYTE_8      ) >> MCPR_SMBUS_UDID0_2_BYTE_8_BITSHIFT, 
            "\n\tBYTE_9      :", (u32_t)(data & MCPR_SMBUS_UDID0_2_BYTE_9      ) >> MCPR_SMBUS_UDID0_2_BYTE_9_BITSHIFT, 
            "\n\tBYTE_10     :", (u32_t)(data & MCPR_SMBUS_UDID0_2_BYTE_10     ) >> MCPR_SMBUS_UDID0_2_BYTE_10_BITSHIFT, 
            "\n\tBYTE_11     :", (u32_t)(data & MCPR_SMBUS_UDID0_2_BYTE_11     ) >> MCPR_SMBUS_UDID0_2_BYTE_11_BITSHIFT); 
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_1: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_4      :", (u32_t)(data & MCPR_SMBUS_UDID0_1_BYTE_4      ) >> MCPR_SMBUS_UDID0_1_BYTE_4_BITSHIFT, 
            "\n\tBYTE_5      :", (u32_t)(data & MCPR_SMBUS_UDID0_1_BYTE_5      ) >> MCPR_SMBUS_UDID0_1_BYTE_5_BITSHIFT, 
            "\n\tBYTE_6      :", (u32_t)(data & MCPR_SMBUS_UDID0_1_BYTE_6      ) >> MCPR_SMBUS_UDID0_1_BYTE_6_BITSHIFT,
            "\n\tBYTE_7      :", (u32_t)(data & MCPR_SMBUS_UDID0_1_BYTE_7      ) >> MCPR_SMBUS_UDID0_1_BYTE_7_BITSHIFT);
        break;
    case MCP_REG_MCPR_SMBUS_UDID0_0: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_0      :", (u32_t)(data & MCPR_SMBUS_UDID0_0_BYTE_0      ) >> MCPR_SMBUS_UDID0_0_BYTE_0_BITSHIFT, 
            "\n\tBYTE_1      :", (u32_t)(data & MCPR_SMBUS_UDID0_0_BYTE_1      ) >> MCPR_SMBUS_UDID0_0_BYTE_1_BITSHIFT, 
            "\n\tBYTE_2      :", (u32_t)(data & MCPR_SMBUS_UDID0_0_BYTE_2      ) >> MCPR_SMBUS_UDID0_0_BYTE_2_BITSHIFT,
            "\n\tBYTE_3      :", (u32_t)(data & MCPR_SMBUS_UDID0_0_BYTE_3      ) >> MCPR_SMBUS_UDID0_0_BYTE_3_BITSHIFT);
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_3: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_12     :", (u32_t)(data & MCPR_SMBUS_UDID1_3_BYTE_12     ) >> MCPR_SMBUS_UDID1_3_BYTE_12_BITSHIFT,
            "\n\tBYTE_13     :", (u32_t)(data & MCPR_SMBUS_UDID1_3_BYTE_13     ) >> MCPR_SMBUS_UDID1_3_BYTE_13_BITSHIFT,
            "\n\tBYTE_14     :", (u32_t)(data & MCPR_SMBUS_UDID1_3_BYTE_14     ) >> MCPR_SMBUS_UDID1_3_BYTE_14_BITSHIFT, 
            "\n\tBYTE_15     :", (u32_t)(data & MCPR_SMBUS_UDID1_3_BYTE_15     ) >> MCPR_SMBUS_UDID1_3_BYTE_15_BITSHIFT); 
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_2: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_8      :", (u32_t)(data & MCPR_SMBUS_UDID1_2_BYTE_8      ) >> MCPR_SMBUS_UDID1_2_BYTE_8_BITSHIFT, 
            "\n\tBYTE_9      :", (u32_t)(data & MCPR_SMBUS_UDID1_2_BYTE_9      ) >> MCPR_SMBUS_UDID1_2_BYTE_9_BITSHIFT, 
            "\n\tBYTE_10     :", (u32_t)(data & MCPR_SMBUS_UDID1_2_BYTE_10     ) >> MCPR_SMBUS_UDID1_2_BYTE_10_BITSHIFT, 
            "\n\tBYTE_11     :", (u32_t)(data & MCPR_SMBUS_UDID1_2_BYTE_11     ) >> MCPR_SMBUS_UDID1_2_BYTE_11_BITSHIFT);
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_1: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_4      :", (u32_t)(data & MCPR_SMBUS_UDID1_1_BYTE_4      ) >> MCPR_SMBUS_UDID1_1_BYTE_4_BITSHIFT, 
            "\n\tBYTE_5      :", (u32_t)(data & MCPR_SMBUS_UDID1_1_BYTE_5      ) >> MCPR_SMBUS_UDID1_1_BYTE_5_BITSHIFT, 
            "\n\tBYTE_6      :", (u32_t)(data & MCPR_SMBUS_UDID1_1_BYTE_6      ) >> MCPR_SMBUS_UDID1_1_BYTE_6_BITSHIFT,
            "\n\tBYTE_7      :", (u32_t)(data & MCPR_SMBUS_UDID1_1_BYTE_7      ) >> MCPR_SMBUS_UDID1_1_BYTE_7_BITSHIFT);
        break;
    case MCP_REG_MCPR_SMBUS_UDID1_0: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %02x %s %02x %s %02x %s %02x",
            addr, data, 
            "\n\tBYTE_0      :", (u32_t)(data & MCPR_SMBUS_UDID1_0_BYTE_0      ) >> MCPR_SMBUS_UDID1_0_BYTE_0_BITSHIFT, 
            "\n\tBYTE_1      :", (u32_t)(data & MCPR_SMBUS_UDID1_0_BYTE_1      ) >> MCPR_SMBUS_UDID1_0_BYTE_1_BITSHIFT, 
            "\n\tBYTE_2      :", (u32_t)(data & MCPR_SMBUS_UDID1_0_BYTE_2      ) >> MCPR_SMBUS_UDID1_0_BYTE_2_BITSHIFT,
            "\n\tBYTE_3      :", (u32_t)(data & MCPR_SMBUS_UDID1_0_BYTE_3      ) >> MCPR_SMBUS_UDID1_0_BYTE_3_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_CMD: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %u %s %s %s %s %s %s %s %s",
            addr, data, 
            "\n\tEGRESS_FIFO_ENABLED    :", (u32_t)(data & MCPR_UMP_UMP_CMD_EGRESS_FIFO_ENABLED ) ? "ON":"OFF",
            "\n\tINGRESS_FIFO_ENABLED   :", (u32_t)(data & MCPR_UMP_UMP_CMD_INGRESS_FIFO_ENABLED) ? "ON":"OFF",
            "\n\tFC_EN                  :", (u32_t)(data & MCPR_UMP_UMP_CMD_FC_EN            ) ? "ON":"OFF",
            "\n\tMAC_LOOPBACK           :", (u32_t)(data & MCPR_UMP_UMP_CMD_MAC_LOOPBACK     ) ? "ON":"OFF",
            "\n\tEGRESS_MAC_DISABLE     :", (u32_t)(data & MCPR_UMP_UMP_CMD_EGRESS_MAC_DISABLE  ) ? "ON":"OFF",
            "\n\tINGRESS_MAC_DISABLE    :", (u32_t)(data & MCPR_UMP_UMP_CMD_INGRESS_MAC_DISABLE ) ? "ON":"OFF",
            "\n\tINGRESS_DRIVE          :", (u32_t)(data & MCPR_UMP_UMP_CMD_INGRESS_DRIVE    ) ? "ON":"OFF",
            "\n\tSW_PAUSE               :", (u32_t)(data & MCPR_UMP_UMP_CMD_SW_PAUSE         ) ? "ON":"OFF",
            "\n\tAUTO_DRIVE             :", (u32_t)(data & MCPR_UMP_UMP_CMD_AUTO_DRIVE       ) ? "ON":"OFF",
            "\n\tINGRESS_RESET          :", (u32_t)(data & MCPR_UMP_UMP_CMD_INGRESS_RESET    ) ? "ON":"OFF",
            "\n\tNO_PLUS_TWO            :", (u32_t)(data & MCPR_UMP_UMP_CMD_NO_PLUS_TWO      ) ? "ON":"OFF",
            "\n\tEGRESS_PKT_FLUSH       :", (u32_t)(data & MCPR_UMP_UMP_CMD_EGRESS_PKT_FLUSH ) ? "ON":"OFF",
            "\n\tCMD_IPG (nibbles)      :", (u32_t)(data & MCPR_UMP_UMP_CMD_CMD_IPG       ) >> MCPR_UMP_UMP_CMD_CMD_IPG_BITSHIFT, 
            "\n\tEGRESS_FIO_RESET       :", (u32_t)(data & MCPR_UMP_UMP_CMD_EGRESS_FIO_RESET ) ? "ON":"OFF",
            "\n\tINGRESS_FIO_RESET      :", (u32_t)(data & MCPR_UMP_UMP_CMD_INGRESS_FIO_RESET) ? "ON":"OFF",
            "\n\tEGRESS_MAC_RESET       :", (u32_t)(data & MCPR_UMP_UMP_CMD_EGRESS_MAC_RESET ) ? "ON":"OFF",
            "\n\tINGRESS_MAC_RESET      :", (u32_t)(data & MCPR_UMP_UMP_CMD_INGRESS_MAC_RESET) ? "ON":"OFF");                  
        break;
    case MCP_REG_MCPR_UMP_UMP_CONFIG: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %u %s %s ",
            addr, data, 
            "\n\tRMII_MODE          :", (u32_t)(data & MCPR_UMP_UMP_CONFIG_RMII_MODE     ) ? "ON":"OFF",
            "\n\tRVMII_MODE         :", (u32_t)(data & MCPR_UMP_UMP_CONFIG_RVMII_MODE    ) ? "ON":"OFF",
            "\n\tINGRESS_MODE       :", (u32_t)(data & MCPR_UMP_UMP_CONFIG_INGRESS_MODE  ) ? "ON":"OFF",
            "\n\tINGRESS_WORD_ACCM  :", (u32_t)(data & MCPR_UMP_UMP_CONFIG_INGRESS_WORD_ACCM) >> MCPR_UMP_UMP_CONFIG_INGRESS_WORD_ACCM_BITSHIFT, 
            "\n\tOLD_BCNT_RDY       :", (u32_t)(data & MCPR_UMP_UMP_CONFIG_OLD_BCNT_RDY  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_UMP_UMP_FC_TRIP: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %u",
            addr, data, 
            "\n\tXON_TRIP  :", (u32_t)(data & MCPR_UMP_UMP_FC_TRIP_XON_TRIP  ) >> MCPR_UMP_UMP_FC_TRIP_XON_TRIP_BITSHIFT, 
            "\n\tXOFF_TRIP :", (u32_t)(data & MCPR_UMP_UMP_FC_TRIP_XOFF_TRIP ) >> MCPR_UMP_UMP_FC_TRIP_XOFF_TRIP_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %u %s %s ",
            addr, data, 
            "\n\tNEW_FRM            :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_NEW_FRM) ? "ON":"OFF",
            "\n\tFRM_IN_PRO         :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_FRM_IN_PRO  ) ? "ON":"OFF",
            "\n\tFIFO_EMPTY         :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_FIFO_EMPTY  ) ? "ON":"OFF",
            "\n\tBCNT               :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_BCNT) >> MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_BCNT_BITSHIFT, 
            "\n\tEGRESS_FIFO_STATE  :", 
                  ((u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) >> MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BITSHIFT == MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_IDLE         ? "IDLE" :
                  ((u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) >> MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BITSHIFT == MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_READY        ? "READY" : 
                  ((u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) >> MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BITSHIFT == MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BUSY         ? "BUSY" :
                  ((u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) >> MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BITSHIFT == MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_EXTRA_RD     ? "EXTRA_RD":
                  ((u32_t)(data & MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) >> MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BITSHIFT == MCPR_UMP_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_LATCH_IP_HDR ? "LATCH_IP_HDR" : "Invalid"))))));
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FRM_RD_DATA: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);                         
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_FRM_WR_CTL: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %u %s %s ",
            addr, data, 
            "\n\tNEW_FRM            :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_NEW_FRM ) ? "ON":"OFF",
            "\n\tFIFO_RDY           :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_FIFO_RDY) ? "ON":"OFF",
            "\n\tBCNT_RDY           :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_BCNT_RDY) ? "ON":"OFF",
            "\n\tBCNT               :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_BCNT) >> MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_BCNT_BITSHIFT,
            "\n\tINGRESS_FIFO_STATE :", 
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) >> MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_BITSHIFT) == MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_IDLE      ?"IDLE":
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) >> MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_BITSHIFT) == MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_WAIT      ?"WAIT":
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) >> MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_BITSHIFT) == MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_BUSY      ?"BUSY":
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) >> MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_BITSHIFT) == MCPR_UMP_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_EXTRA_WR  ?"EXTRA_WR":"Invalid")))));
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_FRM_WR_DATA: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);                        
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FRAME_TYPE: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data); 
        break;
    case MCP_REG_MCPR_UMP_UMP_FIFO_REMAINING_WORDS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %s %s %s %s %u %s %s %s %s",
            addr, data, 
            "\n\tEGRESS_FIFO_DEPTH      :", (u32_t)(data & MCPR_UMP_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_DEPTH) >> MCPR_UMP_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_DEPTH_BITSHIFT,
            "\n\tEGRESS_FIFO_UNDERFLOW  :", (u32_t)(data & MCPR_UMP_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_UNDERFLOW  ) ? "ON":"OFF",
            "\n\tEGRESS_FIFO_OVERFLOW   :", (u32_t)(data & MCPR_UMP_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_OVERFLOW  ) ? "ON":"OFF",
            "\n\tINGRESS_FIFO_DEPTH     :", (u32_t)(data & MCPR_UMP_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_DEPTH) >> MCPR_UMP_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_DEPTH_BITSHIFT,
            "\n\tINGRESS_FIFO_UNDERFLOW :", (u32_t)(data & MCPR_UMP_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_UNDERFLOW  ) ? "ON":"OFF",
            "\n\tINGRESS_FIFO_OVERFLOW  :", (u32_t)(data & MCPR_UMP_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_OVERFLOW  ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_PTRS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %x %s %s %s %x %s %x",
            addr, data, 
            "\n\tEGRESS_FIFO_RD_PTR :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_RD_PTR) >> MCPR_UMP_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_RD_PTR_BITSHIFT, 
            "\n\tUPDATE_RDPTR       :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FIFO_PTRS_UPDATE_RDPTR  ) ? "ON":"OFF",
            "\n\tEGRESS_FIFO_WR_PTR :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_WR_PTR) >> MCPR_UMP_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_WR_PTR_BITSHIFT,
            "\n\tUPDATE_WRPTR       :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_FIFO_PTRS_UPDATE_WRPTR  ) ? "ON":"OFF",
            "\n\tRead ptr->flat space location:", 
                (u32_t)(data & 0x7FF)*4 + MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_FLAT_SPACE,
            "\n\tWrite ptr->flat space location:", 
                (u32_t)((data >> 16) & 0x7FF)*4 + MCP_REG_MCPR_UMP_UMP_EGRESS_FIFO_FLAT_SPACE) ;
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_PTRS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %x %s %s %s %x %s %x",
            addr, data, 
            "\n\tINGRESS_FIFO_RD_PTR:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_RD_PTR) >> MCPR_UMP_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_RD_PTR_BITSHIFT,
            "\n\tUPDATE_RDPTR       :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FIFO_PTRS_UPDATE_RDPTR  ) ? "ON":"OFF",
            "\n\tINGRESS_FIFO_WR_PTR:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_WR_PTR) >> MCPR_UMP_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_WR_PTR_BITSHIFT, 
            "\n\tUPDATE_WRPTR       :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_FIFO_PTRS_UPDATE_WRPTR  ) ? "ON":"OFF",
            "\n\tRead ptr->flat space location:", 
                (u32_t)(data & 0x3FF)*4 + MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_FLAT_SPACE,
            "\n\tWrite ptr->flat space location:", 
                (u32_t)((data >> 16) & 0x3FF)*4 + MCP_REG_MCPR_UMP_UMP_INGRESS_FIFO_FLAT_SPACE);
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_PACKET_SA_0: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_PACKET_SA_1: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_BURST_COMMAND: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %x %s %x %s %x",
            addr, data, 
            "\n\tINGRESS_DMA_START  :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_COMMAND_INGRESS_DMA_START  ) ? "ON":"OFF",
            "\n\tINGRESS_PORT       :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_COMMAND_INGRESS_PORT  ) ? "ON":"OFF",
            "\n\tDMA_LENGTH         :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_COMMAND_DMA_LENGTH) >> MCPR_UMP_UMP_INGRESS_BURST_COMMAND_DMA_LENGTH_BITSHIFT,
            "\n\tINGRESS_PORT_EXT   :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_COMMAND_INGRESS_PORT_EXT) >> MCPR_UMP_UMP_INGRESS_BURST_COMMAND_INGRESS_PORT_EXT_BITSHIFT, 
            "\n\tRBUF_OFFSET        :", (u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_COMMAND_RBUF_OFFSET) >> MCPR_UMP_UMP_INGRESS_BURST_COMMAND_RBUF_OFFSET_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_RBUF_CLUSTER: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)",
            addr, data);
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_VLAN: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %s",
            addr, data, 
            "\n\tINGRESS_VLAN_TAG:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_VLAN_INGRESS_VLAN_TAG) >> MCPR_UMP_UMP_INGRESS_VLAN_INGRESS_VLAN_TAG_BITSHIFT,
            "\n\tVLAN_INS:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_VLAN_VLAN_INS) ? "ON":"OFF",
            "\n\tVLAN_DEL:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_VLAN_VLAN_DEL) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_BURST_STATUS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
            "\n\tRESULT:", 
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_BITSHIFT) == MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_BUSY  ? "RESULT_BUSY" : 
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_BITSHIFT) == MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_DONE  ? "RESULT_DONE" : 
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_BITSHIFT) == MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_ERR   ? "RESULT_ERR"  : 
                (((u32_t)(data & MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_BITSHIFT) == MCPR_UMP_UMP_INGRESS_BURST_STATUS_RESULT_ERR1  ? "RESULT_ERR1" : "Invalid")))));
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_BURST_COMMAND: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %x %s %s %s %x",
            addr, data, 
            "\n\tEGRESS_DMA_START   :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_COMMAND_EGRESS_DMA_START  ) ? "ON":"OFF",
            "\n\tEGRESS_PORT        :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_COMMAND_EGRESS_PORT  ) ? "ON":"OFF",
            "\n\tDMA_LENGTH         :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_COMMAND_DMA_LENGTH) >> MCPR_UMP_UMP_EGRESS_BURST_COMMAND_DMA_LENGTH_BITSHIFT,
            "\n\tEGRESS_PORT_EXT    :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_COMMAND_EGRESS_PORT_EXT  ) ? "ON":"OFF",
            "\n\tTPBUF_OFFSET       :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_COMMAND_TPBUF_OFFSET) >> MCPR_UMP_UMP_EGRESS_BURST_COMMAND_TPBUF_OFFSET_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_VLAN: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %s",
            addr, data, 
            "\n\tEGRESS_VLAN_TAG:", (u32_t)(data & MCPR_UMP_UMP_EGRESS_VLAN_EGRESS_VLAN_TAG) >> MCPR_UMP_UMP_EGRESS_VLAN_EGRESS_VLAN_TAG_BITSHIFT, 
            "\n\tVLAN_INS :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_VLAN_VLAN_INS ) ? "ON":"OFF",
            "\n\tVLAN_DEL :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_VLAN_VLAN_DEL ) ? "ON":"OFF");
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_BURST_STATUS: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s",
            addr, data, 
                "\n\tEGRESS_BURST_STATUS    :", 
                (((u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_BUSY_BITSHIFT) == MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_BUSY ? "BUSY DMA is in progress":
                (((u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_DONE_BITSHIFT) == MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_DONE ? "DONE DMA transfer completed successfully":
                (((u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_ERR0_BITSHIFT) == MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_ERR0 ? "ERR0 DMA transfer interrupted due to MCP_ISOLATE":
                (((u32_t)(data & MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT) >> MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_RSVD_BITSHIFT) == MCPR_UMP_UMP_EGRESS_BURST_STATUS_RESULT_RSVD ? "RSVD Reserved":"Invalid")))));
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %u %s %u ",
            addr, data, 
            "\n\tEGRESS_GOOD_CNT    :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_STATISTIC_EGRESS_GOOD_CNT) >> MCPR_UMP_UMP_EGRESS_STATISTIC_EGRESS_GOOD_CNT_BITSHIFT,
            "\n\tEGRESS_ERROR_CNT   :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_STATISTIC_EGRESS_ERROR_CNT) >> MCPR_UMP_UMP_EGRESS_STATISTIC_EGRESS_ERROR_CNT_BITSHIFT, 
            "\n\tEGRESS_DROP_CNT    :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_STATISTIC_EGRESS_DROP_CNT) >> MCPR_UMP_UMP_EGRESS_STATISTIC_EGRESS_DROP_CNT_BITSHIFT); 
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u ",
            addr, data, 
            "\n\tINGRESS_PKT_CNT:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_STATISTIC_INGRESS_PKT_CNT) >> MCPR_UMP_UMP_INGRESS_STATISTIC_INGRESS_PKT_CNT_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_ARB_CMD: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %x %s %s %s %s %s %s %s %s %s %x %s %s %s %s %s %u",
            addr, data, 
            "\n\tID             :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ID    ) >> MCPR_UMP_UMP_ARB_CMD_UMP_ID_BITSHIFT,
            "\n\tARB_DISABLE    :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_DISABLE ) ? "ON":"OFF",
            "\n\tARB_START      :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_START) ? "ON":"OFF",
            "\n\tARB_BYPASS     :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_BYPASS  ) ? "ON":"OFF",
            "\n\tARB_AUTOBYPASS :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_AUTOBYPASS ) ? "ON":"OFF",
            "\n\tARB_TOKEN_IPG  :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_TOKEN_IPG) >> MCPR_UMP_UMP_ARB_CMD_UMP_ARB_TOKEN_IPG_BITSHIFT,
            "\n\tARB_TOKEN_VALID:", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_TOKEN_VALID) ? "ON":"OFF",
            "\n\tARB_FC_DISABLE :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_FC_DISABLE ) ? "ON":"OFF",
            "\n\tARB_TIMEOUT    :", (u32_t)(data & MCPR_UMP_UMP_ARB_CMD_UMP_ARB_TIMEOUT) >> MCPR_UMP_UMP_ARB_CMD_UMP_ARB_TIMEOUT_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_FRAME_COUNT: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %u",
            addr, data, 
            "\n\tEGRESS_FRAME_COUNT:", (u32_t)(data & MCPR_UMP_UMP_FRAME_COUNT_EGRESS_FRAME_COUNT) >> MCPR_UMP_UMP_FRAME_COUNT_EGRESS_FRAME_COUNT_BITSHIFT, 
            "\n\tINRESS_FRAME_COUNT:", (u32_t)(data & MCPR_UMP_UMP_FRAME_COUNT_INRESS_FRAME_COUNT) >> MCPR_UMP_UMP_FRAME_COUNT_INRESS_FRAME_COUNT_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_EGRESS_STATISTIC_AC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u %s %u %s %u",
            addr, data, 
            "\n\tGOOD_CNT   :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_STATISTIC_AC_EGRESS_GOOD_CNT) >> MCPR_UMP_UMP_EGRESS_STATISTIC_AC_EGRESS_GOOD_CNT_BITSHIFT, 
            "\n\tERROR_CNT  :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_STATISTIC_AC_EGRESS_ERROR_CNT) >> MCPR_UMP_UMP_EGRESS_STATISTIC_AC_EGRESS_ERROR_CNT_BITSHIFT,
            "\n\tDROP_CNT   :", (u32_t)(data & MCPR_UMP_UMP_EGRESS_STATISTIC_AC_EGRESS_DROP_CNT) >> MCPR_UMP_UMP_EGRESS_STATISTIC_AC_EGRESS_DROP_CNT_BITSHIFT);
        break;
    case MCP_REG_MCPR_UMP_UMP_INGRESS_STATISTIC_AC: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %u",
            addr, data, 
            "\n\tINGRESS_PKT_CNT:", (u32_t)(data & MCPR_UMP_UMP_INGRESS_STATISTIC_AC_INGRESS_PKT_CNT) >> MCPR_UMP_UMP_INGRESS_STATISTIC_AC_INGRESS_PKT_CNT_BITSHIFT); 
        break;
    case MCP_REG_MCPR_UMP_UMP_EVENT: 
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data, 
            "\n\tINGRESS_RDY_EVENT          :", (u32_t)(data & MCPR_UMP_UMP_EVENT_INGRESS_RDY_EVENT ) ? "ON":"OFF",
            "\n\tEGRESS_RDY_EVENT           :", (u32_t)(data & MCPR_UMP_UMP_EVENT_EGRESS_RDY_EVENT  ) ? "ON":"OFF",
            "\n\tINGRESSBURST_DONE_EVENT    :", (u32_t)(data & MCPR_UMP_UMP_EVENT_INGRESSBURST_DONE_EVENT  ) ? "ON":"OFF",
            "\n\tEGRESSBURST_DONE_EVENT     :", (u32_t)(data & MCPR_UMP_UMP_EVENT_EGRESSBURST_DONE_EVENT) ? "ON":"OFF",
            "\n\tEGRESS_FRAME_DROP_EVENT    :", (u32_t)(data & MCPR_UMP_UMP_EVENT_EGRESS_FRAME_DROP_EVENT  ) ? "ON":"OFF",
            "\n\tINGRESS_RDY_EVENT_EN       :", (u32_t)(data & MCPR_UMP_UMP_EVENT_INGRESS_RDY_EVENT_EN ) ? "ON":"OFF",
            "\n\tEGRESS_RDY_EVENT_EN        :", (u32_t)(data & MCPR_UMP_UMP_EVENT_EGRESS_RDY_EVENT_EN  ) ? "ON":"OFF",
            "\n\tINGRESSBURST_DONE_EVENT_EN :", (u32_t)(data & MCPR_UMP_UMP_EVENT_INGRESSBURST_DONE_EVENT_EN  ) ? "ON":"OFF",
            "\n\tEGRESSBURST_DONE_EVENT_EN  :", (u32_t)(data & MCPR_UMP_UMP_EVENT_EGRESSBURST_DONE_EVENT_EN  ) ? "ON":"OFF",
            "\n\tEGRESS_FRAME_DROP_EVENT_EN :", (u32_t)(data & MCPR_UMP_UMP_EVENT_EGRESS_FRAME_DROP_EVENT_EN  ) ? "ON":"OFF");                   
        break;
    default:
        sprintf(disp_str, 
            "\n(MCP%05x=%08x)", addr, data);
        break;
    }
    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
}// DecodeMcpGrcRegDetail        

/*
 *------------------------------------------------------------------------
 * GetMcpFioHelp --
 *
 *     Display Help info for a MCP GRC register
 *
 * Results:
 *     A const string
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
static char const * GetMcpFioHelp(u32_t addr)
{
    char const * str;
                                                                   
    // This block is repeated at 0x80080000 within the MCP address space for with values for function 1 if the
    // address is marked with the SPLIT: attribute below. If the item is marked as SHARE:, then the same value is
    // access in both the 0x8000xxxx area and the 0x8008xxxx area.    
    switch (addr & 0xFFFFF)
    {
    case OFFSETOF(mcp_fio_t, mcpf_events_bits):
        str = "\nevents_bits - 0x80000000\n"
            "SPLIT: This register indicates the current status of all local and\n"
            "  global events that need work by this processor.\n"
            "bit 31 ATTENTIONS_VALID\n"
            "  This bit indicates that one or more bits from the attention\n"
            "  value is true and may need servicing.\tRO 0\n"
            "bit 30 NOT_ENABLED\n"
            "  This bit is the inverse of the enable input from the Misc.\n"
            "  block. This bit will be '1' when all the other event bits are\n"
            "  masked to zero to simulate that there is not actions to be\n"
            "  taken.\tRO 0\n"
            "bit 29-28 unused0\tRO 0\n"
            "bit 27 MAIN_PWR_INT\n"
            "  This bit comes from the HC block and indicates that the\n"
            "  PERST_B has gone low since the event has last been\n"
            "  cleared in the HC block. This bit is only valid in TetonII, B0\n"
            "  or greater.\tRO 0\n"
            "bit 26 - 24 RESERVED This event is reserved and will never be asserted.\tRO 0\n"
            "bit 23 MSI\tRO 0\n"
            "bit 22 SW_TMR_4\n"
            "  When this bit is high, it indicates that the corresponding\n"
            "  Software Timer has reached zero and may need servicing.\tRO 0\n"
            "bit 21 SW_TMR_3\n"
            "  When this bit is high, it indicates that the corresponding\n"
            "  Software Timer has reached zero and may need servicing.\tRO 0\n"
            "bit 20 SW_TMR_2\n"
            "  When this bit is high, it indicates that the corresponding\n"
            "  Software Timer has reached zero and may need servicing.\tRO 0\n"
            "bit 19 SW_TMR_1\n"
            "  When this bit is high, it indicates that the corresponding\n"
            "  Software Timer has reached zero and may need servicing.\tRO 0\n"
            "bit 18 GPIO3\n"
            "  When this bit is high, it indicates that one of the SPIO or\n"
            "  PPIO pins that has been configured to generate an event,\n"
            "  has generated an event. and may need servicing.\tRO 0\n"
            "bit 17 GPIO2\n"
            "  When this bit is high, it indicates that one of the SPIO or\n"
            "  PPIO pins that has been configured to generate an event,\n"
            "  has generated an event. and may need servicing.\tRO 0\n"
            "bit 16 GPIO1\n"
            "  When this bit is high, it indicates that one of the SPIO or\n"
            "  PPIO pins that has been configured to generate an event,\n"
            "  has generated an event. and may need servicing.\tRO 0\n"
            "bit 15 GPIO0\n"
            "  When this bit is high, it indicates that one of the SPIO or\n"
            "  PPIO pins that has been configured to generate an event,\n"
            "  has generated an event. and may need servicing.\tRO 0\n"
            "bit 14 RESERVED1 This event is reserved and will never be asserted.\tRO 0\n"
            "bit 13 RESERVED0 This event is reserved and will never be asserted.\tRO 0\n"
            "bit 12 NIG\tRO 0\n"
            "bit 11 SMB0\n"
            "  When this bit is high, it indicates that the SM Bus controller\n"
            "  (in the MISC block) is generating an interrupt.\tRO 0\n"
            "bit 10 FLASH\n"
            "  When this bit is high, it indicates that the flash block has\n"
            "  completed it current command and is ready for data to be\n"
            "  read from the interface or be instructed for a new command.\tRO 0\n"
            "bit 9 VPD\n"
            "  When this bit is high, it indicates that a the PCI block as\n"
            "  detected an access to the Vital Product Data area of the PCI\n"
            "  Configuration space and firmware is required to pass data\n"
            "  between the NVM interface an the PCI block.\tRO 0\n"
            "bit 8 EXP_ROM\n"
            "  When this bit is high, it indicates that a the PCI bus has\n"
            "  detected a memory read of the ROM BAR area and that\n"
            "  firmware needs to fetch the data requested and pass it to the\n"
            "  PCI block.\tRO 0\n"
            "bit 7 HFTQ1_VALID\n"
            "  This bit will be set when there is a valid entry at the to of\n"
            "  the port1 Management Host FTQ. This bit will automatically\n"
            "  be masked if the central enable input from the GRC is not\n"
            "  set.\tRO 0\n"
            "bit 6 HFTQ0_VALID\n"
            "  This bit will be set when there is a valid entry at the to of\n"
            "  the port0 Management Host FTQ. This bit will automatically\n"
            "  be masked if the central enable input from the GRC is not\n"
            "  set.\tRO 0\n"
            "bit 5 MCP_DOORBELL\n"
            "  SPLIT: This bit indicates that the MCP doorbell was rung.\n"
            "  The bit in the port0 portion of the FIO area will reflect the\n"
            "  for the port0 MCP doorbell register. The case for the\n"
            "  port0 MCP doorbell register. The bit in the\n"
            "  port1 portion of the FIO area will reflect the case for the\n"
            "  port1 MCP doorbell register.\tRO 0\n"
            "bit 4 FLASH_EVENT This bit indicates that there is an flash event.\tRO 0\n"
            "bit 3 SMBUS_EVENT This bit indicates that there is an smbus event.\tRO 0\n"
            "bit 2 UMP_EVENT\n"
            "  This bit indicates that there is an UMP event. FW needs to\n"
            "  look at the UMP event register.\tRO 0\n"
            "bit 1 FTQ1_VALID\n"
            "  This bit will be set when there is a valid entry at the to of\n"
            "  the port1 Management FTQ. This bit will automatically be\n"
            "  masked if the central enable input from the GRC is not set.\tRO 0\n"
            "bit 0 FTQ0_VALID\n"
            "  This bit will be set when there is a valid entry at the to of\n"
            "  the port0 Management FTQ. This bit will automatically be\n"
            "  masked if the central enable input from the GRC is not set.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_attentions_bits):
        str = "\nattentions_bits - 0x80000004\n"
            "SPLIT: This register indicates the current status of all chip-wide error attentions.\n"
            "bit 31-5 unused0\tRO 0\n"
            "bit 4 RSV_ACCESS\tRO 0\n"
            "bit 3 SPIO5\tRO 0\n"
            "bit 2 SPAD_PARITY_ERR\n"
            "  When this bit is high, it indicates that a parity error occurred\n"
            "  in the MCP scratchpad.\tRO 0\n"
            "bit 1 PERST_ASSERTION\tRO 0\n"
            "bit 0 GRC_TIMEOUT\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_event_enable):
        str = "\n   event_enable - 0x80000008 - (RW/Reset: 0)\n"
            "  By setting bits as '1' in this register, events can be un-masked (enabled) to generate interrupts.\n"
            "  SHARE: Bit 0 controls the event_enable for the MCPQ from function 0.\n"
            "  SHARE: Bit 1 controls the event_enable for the MCPQ from function 1.\n"
            "  SPLIT: Bit 2 controls the event from the UMP block.\n"
            "  SPLIT: Bit 3 controls the event from the SMBus block.\n"
            "  SPLIT: Bit 4 controls the event from the Flash block.\n"
            "  SPLIT: Bit 5 controls MCP doorbell event on a per-prt basis.\n"
            "  SPLIT: Bits 31:8 control the remaing events for both functions independently.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_attention_enable):
        str = "\nattention_enable - 0x8000000c - (RW/Reset: 0)\n"
            "SPLIT: By setting bits as '1' in this register, attentions can be un-masked (enabled) to generate interrupts. For\n"
            "  this to work, the ATTENTION_VALID bit in the event_mask register must be un-masked as well.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_fio_status):
        str = "\n fio_status - 0x80000010\n"
            "  SHARE:\n"
            "bit 31-2 unused0\tRO 0\n"
            "bit 1 FORCE_ENA\n"
            "  This bit forces enable to the block so that all the events will be\n"
            "  visible regardless of it the block is enabled or not. Essentially,\n"
            "  setting this bit to '1' disables the event register masking feature of\n"
            "  the central enable input.\tRW 0\n"
            "bit 0 ENABLED\n"
            "  This bit indicates the current \"enable\" status from the central enable\n"
            "  block. If this bit is '0', then the firmware on this processor should\n"
            "  stop what it is doing at a convenient place and poll this bit until it\n"
            "  becomes '1' again. The event register assists in this by automatically\n"
            "  masking all bits (except the NOT_ENABLED bit to zero when the\n"
            "  block is disabled. This is used for stress testing of the design.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_interrupt_status):
        str = "\ninterrupt_status - 0x80000014\n"
            "  SHARE:\n"
            "bit 31-8 unused0\tRO 0\n"
            "bit 7 ATTN1_E1\n"
            "  This bit is high when attention_bits for port 1, path 1 (corresponding\n"
            "  register is 0x80180004) is source of the interrupt.\tRO 0\n"
            "bit 6 EVENT1_E1\n"
            "  This bit is high when events_bits for port 1, path 1 (corresponding\n"
            "  register is 0x80180000) is source of the interrupt.\tRO 0\n"
            "bit 5 ATTN0_E1\n"
            "  This bit is high when attention_bits for port 0, path 1 (corresponding\n"
            "  register is 0x80100004) is source of the interrupt.\tRO 0\n"
            "bit 4 EVENT0_E1\n"
            "  This bit is high when events_bits for port 0, path 1 (corresponding\n"
            "  register is 0x80100000) is source of the interrupt.\tRO 0\n"
            "bit 3 ATTN1_E0\n"
            "  This bit is high when attention_bits for port 1, path 0 (corresponding\n"
            "  register is 0x80080004) is source of the interrupt.\tRO 0\n"
            "bit 2 EVENT1_E0\n"
            "  This bit is high when events_bits for port 1, path 0 (corresponding\n"
            "  register is 0x80080000) is source of the interrupt.\tRO 0\n"
            "bit 1 ATTN0_E0\n"
            "  This bit is high when attention_bits for port 0, path 0 (corresponding\n"
            "  register is 0x80000004) is source of the interrupt.\tRO 0\n"
            "bit 0 EVENT0_E0\n"
            "  This bit is high when events_bits for port 0, path 0 (corresponding\n"
            "  register is 0x80000000) is source of the interrupt.\tRO 0\n";
        break;
#ifndef UEFI
    case (OFFSETOF(mcp_fio_t, mcpf_mcp_hc_inc_stat)+8*sizeof(u32_t)):
        str = "\nmcp_hc_inc_stat[8] - 0x80000040 - (WO/Reset: 0)\n"
            "  SHARE: This is a command register to increment the generic statistic counters. Each time one of these 8\n"
            "  addresses is written, the corresponding generic statistic for the Management Processor will be incremented by\n"
            "  one (if selected).\n";
        break;
#endif
    case OFFSETOF(mcp_fio_t, mcpf_free_counter_value):
        str = "\nfree_counter_value - 0x80000070 - (RO/Reset: 0)\n"
            "  SHARE: This value is a free-running counter, running at on the CPU clock. The counter is reset only by\n"
            "  CORE reset. This counter only exists in TetonII, B0 or greater. This value is a\tRO zero in all other revisions.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_bits_status1):
        str = "\nmcpq_bits_status1 - 0x80000100\n"
            "SPLIT:\n"                                                              
                        "bit 31-28 unused0\tRO 0\n"
            "bit 27 L2_CRC L2 CRC Error\tRW 0\n"                                    
                        "bit 26 NO_VLAN Untagged packet\tRW 0\n"
            "bit 25 VLAN Tagged packet\tRW 0\n"
            "bit 24 VLAN_ID2 VLAN #3\tRW 0\n"
            "bit 23 VLAN_ID1 VLAN #2\tRW 0\n"
            "bit 22 VLAN_ID0 VLAN #1\tRW 0\n"
            "bit 21 TCP2 Destination TCP port #3\tRW 0\n"
            "bit 20 TCP1 Destination TCP port #2\tRW 0\n"
            "bit 19 TCP0 Destination TCP port #1\tRW 0\n"
            "bit 18 NTBS_T_DST NetBIOS TCP destination port 137, 138, 139\tRW 0\n"
            "bit 17 UDP2 Destination UDP port #3\tRW 0\n"
            "bit 16 UDP1 Destination UDP port #2\tRW 0\n"
            "bit 15 UDP0 Destination UDP port #1\tRW 0\n"
            "bit 14 NTBS_U_DST NetBIOS UDP destination port 137, 138, 139\tRW 0\n"
            "bit 13 DHCP DHCP UDP/TCP ports 67, 68, 546, 547, 647, 847\tRW 0\n"
            "bit 12 RMCP RMCP UDP ports 0x26f and 0x298\tRW 0\n"                
            "bit 11 NTBS_T_SRC NetBIOS TCP source port 137, 138, 139\tRW 0\n"
            "bit 10 NTBS_U_SRC NetBIOS UDP source port 137, 138, 139\tRW 0\n"
            "bit 9 IP2 Destination IP address #3\tRW 0\n"
            "bit 8 IP1 Destination IP address #2\tRW 0\n"
            "bit 7 IP0 Destination IP address #1\tRW 0\n"
            "bit 6 ARP ARP\tRW 0\n"
            "bit 5 MAC2 Destination MAC address #3\tRW 0\n"
            "bit 4 MAC1 Destination MAC address #2\tRW 0\n"
            "bit 3 MAC0 Destination MAC address #1\tRW 0\n"
            "bit 2 UNCST Unicast MAC address\tRW 0\n"
            "bit 1 MLCST Multicast MAC address\tRW 0\n"
            "bit 0 BRCST Broadcast MAC address\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_pkt_len):                                     
        str = "\n   mcpq_pkt_len - 0x80000104\n"
            "SPLIT: This is the byte length of frame, including CRC\n"
            "bit 15-14 unused0\tRO 0\n"
            "bit 13-0 MASK\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_vlan_tag):
        str = "\n  mcpq_vlan_tag - 0x80000106 - (RO/Reset: 0)\n"
            "SPLIT: This field has the VLAN tag value in it.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_bits_status2):
        str = "\nmcpq_bits_status2 - 0x80000108\n"
            "SPLIT:\n"                                                          
            "bit 31-25 unused0\tRO 0\n"
            "bit 24 MAC5 Destination MAC address #5\tRO 0\n"
            "bit 23 MAC4 Destination MAC address #4\tRO 0\n"
            "bit 22-7 OUTER_VLAN_MASK\tRO 0\n"
            "bit 6 MF_IPV6_MLCST IPv6 Multicast address\tRO 0\n"
            "bit 5 MF_MAC3 Destination MAC address #3\tRO 0\n"                  
            "bit 4 MF_OUTER_VLAN_ID Outer VLAN\tRO 0\n"
            "bit 3 MF_NO_OUTER_VLAN Regular Tagged packet\tRO 0\n"
            "bit 2 MF_OUTER_VLAN Double Tagged packet\tRO 0\n"
            "bit 1-0 CLUSTER_MASK\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_bits_status3):
        str = "\nmcpq_bits_status3 - 0x8000010c\n"
            "SPLIT:\n"
            "bit 18 PF1_VLAN PF outer VLAN\tRO 0\n"
            "bit 16 PF0_VLAN PF outer VLAN\tRO 0\n"
            "bit 13 VNTAG1\tRO 0\n"                                             
            "bit 12 VNTAG0\tRO 0\n"                                             
            "bit 11 LLDP\tRO 0\n"
            "bit 9 ICMPV6 ICMPv6\tRO 0\n"
            "bit 8 ICMPV4 ICMPv4\tRO 0\n"
            "bit 7 ARP ARP\tRO 0\n"                                             
            "bit 3 Multicast\tRO 0\n"                                             
            "bit 2 ETYPE1 Ethertype 2\tRO 0\n"
            "bit 1 ETYPE0 Ethertype 1\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_cmd):
        str = "\nmcpq_cmd - 0x80000138\n"
            "SPLIT: This is the control register for the read port on the FTQ.\n"
            "bit 31 unused1\tRO 0\n"
            "bit 30 MCPQ_CMD_POP\n"
            "  This bit always reads as '0'. When written as a '1', the top entry\n"
            "  on the FTQ will be popped and the next entry will be brought to\n"
            "  the registers above. The value of the FTQ0_VALID bit in the\n"
            "  events_bits register must be cleared when this bit is set and then\n"
            "  re-enabled if there is another FTQ entry. The processor may\n"
            "  read the events_bits register on the next instruction.\tSC 0\n"
            "bit 29-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_bits_status1):
        str = "\nhmcpq_bits_status1 - 0x80000140\n"
            "SPLIT:\n"
            "bit 31-28 unused0\tRO 0\n"
            "bit 27 L2_CRC L2 CRC Error\tRW 0\n"
            "bit 26 NO_VLAN Untagged packet\tRW 0\n"
            "bit 25 VLAN Tagged packet\tRW 0\n"
            "bit 24 VLAN_ID2 VLAN #3\tRW 0\n"
            "bit 23 VLAN_ID1 VLAN #2\tRW 0\n"
            "bit 22 VLAN_ID0 VLAN #1\tRW 0\n"
            "bit 21 TCP2 Destination TCP port #3\tRW 0\n"
            "bit 20 TCP1 Destination TCP port #2\tRW 0\n"                       
            "bit 19 TCP0 Destination TCP port #1\tRW 0\n"
            "bit 18 NTBS_T_DST NetBIOS TCP destination port 137, 138, 139\tRW 0\n"
            "bit 17 UDP2 Destination UDP port #3\tRW 0\n"
            "bit 16 UDP1 Destination UDP port #2\tRW 0\n"
            "bit 15 UDP0 Destination UDP port #1\tRW 0\n"                       
            "bit 14 NTBS_U_DST NetBIOS UDP destination port 137, 138, 139\tRW 0\n"
            "bit 13 DHCP DHCP UDP/TCP ports 67, 68, 546, 547, 647, 847\tRW 0\n"
            "bit 12 RMCP RMCP UDP ports 0x26f and 0x298\tRW 0\n"
            "bit 11 NTBS_T_SRC NetBIOS TCP source port 137, 138, 139\tRW 0\n"   
            "bit 10 NTBS_U_SRC NetBIOS UDP source port 137, 138, 139\tRW 0\n"
            "bit 9 IP2 Destination IP address #3\tRW 0\n"
            "bit 8 IP1 Destination IP address #2\tRW 0\n"
            "bit 7 IP0 Destination IP address #1\tRW 0\n"
            "bit 6 ARP ARP\tRW 0\n"
            "bit 5 MAC2 Destination MAC address #3\tRW 0\n"
            "bit 4 MAC1 Destination MAC address #2\tRW 0\n"
            "bit 3 MAC0 Destination MAC address #1\tRW 0\n"
            "bit 2 UNCST Unicast MAC address\tRW 0\n"
            "bit 1 MLCST Multicast MAC address\tRW 0\n"
            "bit 0 BRCST Broadcast MAC address\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_pkt_len):
        str = "\n  hmcpq_pkt_len - 0x80000144\n"
            "SPLIT: This is the byte length of frame, including CRC\n"
            "bit 15-14 unused0\tRO 0\n"
            "bit 13-0 MASK\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_vlan_tag):
        str = "\n hmcpq_vlan_tag - 0x80000146 - (RO/Reset: 0)\n"
            "SPLIT: This field has the VLAN tag value in it.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_bits_status2):
        str = "\nhmcpq_bits_status2 - 0x80000148\n"
            "SPLIT:\n"
            "bit 31-25 unused0\tRO 0\n"
            "bit 24 MAC5 Destination MAC address #5\tRO 0\n"
            "bit 23 MAC4 Destination MAC address #4\tRO 0\n"
            "bit 22-7 OUTER_VLAN_MASK\tRO 0\n"
            "bit 6 MF_IPV6_MLCST IPv6 Multicast address\tRO 0\n"
            "bit 5 MF_MAC3 Destination MAC address #3\tRO 0\n"                  
            "bit 4 MF_OUTER_VLAN_ID Outer VLAN\tRO 0\n"
            "bit 3 MF_NO_OUTER_VLAN Regular Tagged packet\tRO 0\n"
            "bit 2 MF_OUTER_VLAN Double Tagged packet\tRO 0\n"
            "bit 1-0 CLUSTER_MASK\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_bits_status3):
        str = "\nhmcpq_bits_status3 - 0x8000014c\n"                             
            "SPLIT:\n"
            "bit 18 PF1_VLAN PF outer VLAN\tRO 0\n"
            "bit 16 PF0_VLAN PF outer VLAN\tRO 0\n"
            "bit 13 VNTAG1\tRO 0\n"                                             
            "bit 12 VNTAG0\tRO 0\n"                                             
            "bit 11 LLDP\tRO 0\n"
            "bit 9 ICMPV6 ICMPv6\tRO 0\n"
            "bit 8 ICMPV4 ICMPv4\tRO 0\n"
            "bit 7 ARP ARP\tRO 0\n"                                             
            "bit 3 Multicast\tRO 0\n"                                             
            "bit 2 ETYPE1 Ethertype 2\tRO 0\n"
            "bit 1 ETYPE0 Ethertype 1\tRO 0\n";
        break;    
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_cmd):
        str = "\n  hmcpq_cmd - 0x80000178\n"
            "SPLIT: This is the control register for the read port on the Host FTQ.\n"
            "bit 31 unused1\tRO 0\n"
            "bit 30 HMCPQ_CMD_POP\n"                                            
            "  This bit always reads as '0'. When written as a '1', the top entry\n"
            "  on the FTQ will be popped and the next entry will be brought\n"
            "  to the registers above. The value of the FTQ0_VALID bit in\n"
            "  the events_bits register must be cleared when this bit is set\n"
            "  and then re-enabled if there is another FTQ entry. The\n"
            "  processor may read the events_bits register on the next\n"
            "  instruction.\tSC 0\n"
            "bit 29-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_command):
        str = "\nnvm_command - 0x80026400\n"
            "  The NVM block sits in the common area in Xinan. As with all common blocks, they may be addressed as\n"
            "  either function 0 or function 1. There are two types of resolution functions in NVM block.\n"
            "  SPLIT = bit has two registers. On is access when function 0 is accessed. A different value is access\n"
            "  when function 1 is accessed. These bits are written and read independently\n"
            "  SHARE = a single register is provided. Both function 0 and function 1 views see the exact same\n"
            "  register. Software has to be aware that what it reads may not be wha;t it has written if the other port\n"
            "  overwrites the previously written value.\n"
            "  SHARE: NVM Command register.\n"
            "bit 31-23 unused3\tRO 0\n"
            "bit 22 MODE_256\n"
            "  The 256B page size mode disable bit. A 256 byte page mode has\n"
            "  been added to the block. This mode is normally on. The mode helps\n"
            "  convert a 264B page Atmel part to act more like a 256B page part.\n"
            "  For reads, the controller transparently closes the page after byte\n"
            "  location 0xFF and opens the next page. For writes, it is the FW or\n"
            "  SW responsiblity to close the page at 0xFF and start a new\n"
            "  operation on the next page. When this bit is written as '1' when the\n"
            "  FIRST bit is set, the 256B page mode is disabled for the next\n" 
            "  operation. It is self-clearing when both the LAST bit is set and the\n"
            "  DONE bit is asserted. Effects Atmel only. No effect with ST\n"
            "  devices.\tSC 0\n"
            "bit 21 RD_STATUS\n"
            "  The read status command bit. When set, the flash controller will\n"
            "  read the status register from the external flash device\tRW 0\n"
            "bit 20 RD_ID\n"
            "  The read ID command bit. When set, the flash controller will read\n"
            "  the ID register from the external flash device. This is specifically for\n"
            "  ST devices. Setting this bit for Atmel devices will give the same\n"
            "  results as RD_STATUS.\tRW 0\n"
            "bit 19-18 unused2\tRO 0\n"
            "bit 17 WRDI\n"
            "  The write disable command bit. Set '1' will make flash interface\n"
            "  state machine Generate wrdi_cmd to flash device through SPI\n"
            "  interface to set Flash device to be write-disabled. Used for the\n"
            "  device with protection function.\tRW 0\n"
            "bit 16 WREN\n"
            "  The write enable command bit. Set '1' will make flash interface state\n"
            "  machine Generate wren_cmd to flash device through SPI interface\n"
            "  to set Flash device to be write-enabled. Used for the device with\n"
            "  protection function\tRW 0\n"
            "bit 15-9 unused1\tRO 0\n"
            "bit 8 LAST\n"
            "  When this bit is set, the next command sequence will be interpreted\n"
            "  as the last one of a burst and any cleanup work will be done. This\n"
            "  means that the buffer will be written to flash memory if needed on a\n"
            "  write.\tRW 0\n"
            "bit 7 FIRST\n"
            "  This bit is passed to the SEE_FSM or SPI_FSM if the pass_mode bit\n"
            "  is set.\tRW 0\n"
            "bit 6 ERASE\n"                                                     
            "  The erase command bit. Set high to execute an erase. This bit is\n"
            "  ignored if the wr is clear.\tRW 0\n"
            "bit 5 WR\n"
            "  The Write/Not_Read command bit. Set high to execute write or\n"
            "  erase.\tRW 0\n"                                                  
            "bit 4 DOIT\n"
            "  Command from software to start the defined command. The done\n"
            "bit must be clear before setting this bit. This bit is self clearing a;nd\n"
            "  will remain set while the command is active.\tSC 0\n"
            "bit 3 DONE\n"
            "  Sequence completion bit that is asserted when the command\n"
            "  requested by assertion of the doit bit has completed. done Will be\n"
            "  cleared while the command is in progress. done Will stay asserted\n"
            "  until doit is reasserted or the done bit is cleared by writing a 1 to the\n"
            "  done bit. The done bit is the FLSH_ATTN signal.\tWC 0\n"
            "bit 2-1 unused0\tRO 0\n"
            "bit 0 RST\n"
            "  When set, the entire NVM state machine is reset. This bit is self\n"
            "  clearing. \tSC 0\n";
        break;                         
    case OFFSETOF(mcp_fio_t, mcpf_nvm_status):
        str = "\n nvm_status - 0x80026404\n"
            "  SHARE: NVM Status register.\n"
            "bit 31-5 unused0\tRO 0\n"
            "bit 4-0 SPI_FSM_STATE\n"
            "  Enumeration:\tRO 0\n"
            "  Value Name Description\n"
            "   0 SPI_IDLE\n"
            "   1 SPI_CMD0\n"
            "   2 SPI_CMD1\n"
            "   3 SPI_CMD_FINISH0\n"
            "   4 SPI_CMD_FINISH1\n"
            "   5 SPI_ADDR0\n"
            "   6 SPI_WRITE_DATA0\n"
            "   7 SPI_WRITE_DATA1\n"
            "   8 SPI_WRITE_DATA2\n"                                            
            "   9 SPI_READ_DATA0\n"
            "   10 SPI_READ_DATA1\n"
            "   11 SPI_READ_DATA2\n"
            "   12 SPI_READ_STATUS_RDID0\n"
            "   13 SPI_READ_STATUS_RDID1\n"
            "   14 SPI_READ_STATUS_RDID2\n"
            "   15 SPI_READ_STATUS_RDID3\n"
            "   16 SPI_READ_STATUS_RDID4\n"
            "   17 SPI_CHECK_BUSY0\n"
            "   18 SPI_ST_WREN\n"
            "   19 SPI_WAIT\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_write):
        str = "\n  nvm_write - 0x80026408\n"                                    
            "  SHARE: NVM data write register.\n"
            "bit 31-0 NVM_WRITE_VALUE\n"
            "  32 bits of write data use when write commands are\n"
            "  executed. When bitbang_mode is set, only the bottom 6 bits\n"
            "  have any effect as described below.\tRW 0\n"
            "  Value Name     Description\n"
            "   0    BIT_BANG When in bit-bang mode, the bottom 6\n"
            "                 bits of this register control the output\n"
            "                 value of the external pins.\n"
            "   1    SI       When in bit-bang mode, this bit controls\n"
            "                 the SI output value.\n"
            "   2    SO       When in bit-bang mode, this bit controls\n"
            "                 the SO output value.\n"
            "   4    CS_B     When in bit-bang mode, this bit controls\n"
            "                 the CS_B output value.\n"
            "   8    SCLK     When in bit-bang mode, this bit controls\n"
            "                 the SCLK output value.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_addr):
        str = "\nnvm_addr - 0x8002640c\n"
            "  SHARE: NVM address register.\n"
            "bit 31-24 unused0\tRO 0\n"
            "bit 23-0 NVM_ADDR_VALUE\n"
            "bit 24 bit address value used in read, write and erase operations.\n"
            "  When in bit-bang mode, the bottom 6 bits control the output\n"
            "  enable for each pin.\tRW 0\n"
            "  Value Name      Description\n"
            "   0    BIT_BANG  When in bit-bang mode, the bottom 6\n"
            "                  bits of this register control the output\n"
            "                  enable value of the external pins.\n"
            "   1    SI        When in bit-bang mode, this bit controls\n"
            "                  the enables the SI pin as an output when '0'.\n"
            "   2    SO        When in bit-bang mode, this bit controls\n"
            "                  the enables the SO pin as an output when '0'.\n"
            "   4    CS_B      When in bit-bang mode, this bit controls\n"
            "                  the enables the CS_B pin as an output when '0'.\n"
            "   8    SCLK      When in bit-bang mode, this bit controls\n"
            "                  the enables the SCLK pin as an output when '0'.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_read):
        str = "\nnvm_read - 0x80026410\n"
            "  SHARE: NVM data read register.\n"
            "bit 31-0 NVM_READ_VALUE\n"
            "  32 bits of read data use when read commands are executed.\n"
            "  When bitbang_mode is set, bits 0 to 5 reflect the current input\n"
            "  value of the external pins.\tRO 0\n"
            "  Value Name      Description\n"
            "   0    BIT_BANG  When in bit-bang mode, the bottom 6 bits\n"
            "                  of this return the input value of the\n"
            "                  external pins.\n"
            "   1    SI        When in bit-bang mode, this bit returns\n"
            "                  the current value of the SI pin.\n"
            "   2    SO        When in bit-bang mode, this bit returns\n"
            "                  the current value of the SO pin.\n"
            "   4    CS_B      When in bit-bang mode, this bit returns\n"
            "                  the current value of the CS_B pin.\n"
            "   8    SCLK      When in bit-bang mode, this bit returns\n"
            "                  the current value of the SCLK pin.\n";
        break; 
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg1):
        str = "\nnvm_cfg1 - 0x80026414\n"
            "  SHARE: NVM configuration one register.\n"
            "bit 31 COMPAT_BYPASSS Legacy bit. Acts as dummy R/W bit.\tRW 0\n"
            "bit 30 FW_FLASH_TYPE_EN\n"
            "  Legacy strap_control[1] bit. Read only set to 1, indicating\n"
            "  FLASH has already been configured.This register has no\n"
            "  hardware function, but can be modified by firmware.\tRO 0x1\n"
            "bit 29 FW_USTRAP_3\n"
            "  Legacy strap_value[3]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"
            "bit 28 FW_USTRAP_2\n"
            "  Legacy strap_value[2]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"                  
            "bit 27 FW_USTRAP_0\n"
            "  Legacy strap_value[0]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"
            "bit 26 FW_USTRAP_1\n"
            "  Legacy strap_value[1]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"
            "bit 25 FLASH_SIZE\n"
            "  Legacy strap_value[3]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"
            "bit 24 PROTECT_MODE\n"
            "  Legacy strap_value[2]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"
            "bit 23 STRAP_CONTROL_0\n"
            "  Legacy strap_control[1] bit. Read only set to 1, indicating\n"
            "  FLASH has already been configured.\tRO 0x1\n"
            "bit 22 unused0\tRO 0\n"
            "bit 21-11 SEE_CLK_DIV Legacy value. Read only.\tRO 0x60\n"
            "bit 10-7 SPI_CLK_DIV\n"
            "  Divisor used to create all \"1x\" time for all Flash Interface\n"
            "  I/O pin timing definition A divisor of 0 means that an SCLK\n"
            "  will transition at minimum of each CORE_CLK rising edge.\n"
            "  The equation to calculate the clock frequency for SCLK is\n"
            "  :CORE_CLK frequency/((spi_clk_div +1) *2).\tRW 0x4\n"
            "bit 6-4 STATUS_BIT\n"
            "  Bit offset in status command response to interpret as the\n"
            "  \"ready\" flag. For Atmel, this defaults to 3'h7. For ST, this\n"
            "  defaults to 3'h0. NOTE: For ST, the status value of 1'b0\n"      
            "  means \"ready\". For Atmel, the status value of 1 means\n"
            "  \"ready\". This is automatically interpreted by hardware. This\n"
            "  value is self-configured on reset based on the strap values.\n"
            "  It can be overriden.\tRW X\n"
            "bit 3 BITBANG_MODE Enable bit-bang mode to control pins.\tRW 0\n"
            "bit 2 PASS_MODE\n"
            "  Enable pass-thru mode to the byte level SPI state machine.\n"
            "  When this mode is enabled, the controller can send/recieve\n"
            "  single bytes at the SPI level. All upper level functions of the\n"
            "  controller state machine are disabled.\tRW 0\n"
            "bit 1 BUFFER_MODE\n"
            "  Legacy strap_value[0]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel or ST.\tRO X\n"
            "bit 0 FLASH_MODE\n"
            "  Legacy strap_value[1]. Read only. Set based on new strap\n"
            "  values to indicate either Atmel o\tRO X\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg2):
        str = "\nnvm_cfg2 - 0x80026418\n"
            "  SHARE: NVM configuration two register.\n"                        
            "bit 31-24 READ_ID\n"
            "  Flash Read ID register command. This command is used to read the\n"
            "  ID register from ST devices. Reset value depends on strap values.\n"
            "  (0x57 for Atmel; 0x9f for ST)\tRW 0\n"
            "bit 23-16 STATUS_CMD\n"
            "  Flash status command. This command is used to poll the \"ready\"\n"
            "  status of the flash part after many of the commands. Reset value is\n"
            "bit 0x9Fh if flash_mode=1, 0x57h if buffer_mode=1, and 0x5h if\n"
            "  protect_mode =1.\tRW 0\n"
            "bit 15-8 unused0\tRO 0\n"
            "bit 7-0 ERASE_CMD\n"
            "  Flash block erase command. \"ready\" status will be polled for after\n"
            "  this command. Reset value is 0x20h if flash_mode=1, 0x81h if\n"
            "  buffer_mode=1, and 0xd8h if protect_mode=1.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg3):
        str = "\nnvm_cfg3 - 0x8002641c\n"
            "  SHARE: NVM configuration three register.\n"
            "bit 31-24 READ_CMD\n"
            "  This is the flash/seeprom read command. Following this\n"
            "  command, any number of bytes may be read up to the end of\n"
            "  the flash memory. For SEEPROM (flash_mode=0), this is\n"         
            "  SEEPROM read command. Bit[26:25] is address bit A1 and\n"
            "  A0 of SEEPROM. User should modify those two bits base on\n"
            "  the value of A1 and A0 assigned to this SEEPROM device.\n"
            "  Reset value is 0xFF if flash_mode=1, 0x68 if buffer_mode=1,\n"
            "bit 0x3 if protect_mode=1, and 0xA1 otherwise.\tRW 0\n"
            "bit 23-16 unused0\tRO 0\n"
            "bit 15-8 WRITE_CMD\n"
            "  Command to write one byte to the flash array or SSRAM\n"
            "  buffer, depending on the value of buffer_mode. If\n"
            "  BUFFER_MODE is not active, then this command will poll\n"
            "  for \"ready\" status when complete. For SEEPROM\n"
            "  (flash_mode=0), this is SEEPROM write command. Bit[10:9]\n"
            "  is address bit A1 and A0 of SEEPROM. User should modify\n"
            "  those two bits base on the value of A1 and A0 assigned to this\n"
            "  SEEPROM device. Reset value is 0x10 if flash_mode=1, 0x83\tRW 0\n"
            "  if buffer_mode=1, and 0x2 if protect_mode=1, 0xA0\n"
            "  otherwise.\n"
            "bit 7-0 BUFFER_RD_CMD\n"
            "  Transfer flash device page to its internal buffer command. For\n"
            "  Atmel devices, this command is issued automatically upon the\n"
            "  \"FIRST\" write. It is not issued for ST devices since they\n"
            "  automatically do this operation internally. (0x53 for Atmel;\n"
            "  0x53 for ST)\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_sw_arb):
        str = "\n nvm_sw_arb - 0x80026420\n"
            "SPLIT: NVM arbitration register. This register provides aribtration resoures for both functions. The register is\n"
            "  split, so arbitration depends on what function you are accessing as.\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15 REQ3\n"
            "  This is the current status of requester 3. When this bit is one, it\n"
            "  means that REQ_SET3 has been set since REQ_CLR3.\tRO 0\n"
            "bit 14 REQ2\n"
            "  This is the current status of requester 2. When this bit is one, it\n"
            "  means that REQ_SET2 has been set since REQ_CLR2.\tRO 0\n"
            "bit 13 REQ1\n"
            "  This is the current status of requester 1. When this bit is one, it\n"
            "  means that REQ_SET1 has been set since REQ_CLR1.\tRO 0\n"
            "bit 12 REQ0\n"
            "  This is the current status of requester 0. When this bit is one, it\n"
            "  means that REQ_SET0 has been set since REQ_CLR0.\tRO 0\n"
            "bit 11 ARB_ARB3\n"
            "  when REQ3 arbitration is won, this bit will be read as 1, when an\n"
            "  operation is complete, then the CLR_ARB3 must be written to\n"
            "  clear this bit.\tRO 0\n"
            "bit 10 ARB_ARB2\n"
            "  when REQ2 arbitration is won, this bit will be read as 1, when an\n"
            "  operation is complete, then the CLR_ARB2 must be written to\n"
            "  clear this bit.\tRO 0\n"
            "bit 9 ARB_ARB1\n"
            "  when REQ1 arbitration is won, this bit will be read as 1, when an\n"
            "  operation is complete, then the CLR_ARB1 must be written to\n"
            "  clear this bit.\tRO 0\n"
            "bit 8 ARB_ARB0\n"
            "  when REQ0 arbitration is won, this bit will be read as 1, when an\n"
            "  operation is complete, then the CLR_ARB0 must be written to\n"
            "  clear this bit. At that point, the next Arb bit will read as 1. At any\n"
            "  time, only one of the ARB[7:0] bits will be read as a 1. Arb0 has\n"
            "  highest priority, and Arb7 has lowest priority.\tRO 0\n"
            "bit 7 ARB_REQ_CLR3 Write this bit as a '1' to clear req3 bit. SC 0\n"
            "bit 6 ARB_REQ_CLR2 Write this bit as a '1' to clear req2 bit. SC 0\n"
            "bit 5 ARB_REQ_CLR1 Write this bit as a '1' to clear req1 bit. SC 0\n"
            "bit 4 ARB_REQ_CLR0 Write this bit as a '1' to clear req0 bit. SC 0\n"
            "bit 3 ARB_REQ_SET3\n"
            "  Set Software Arbitration request Bit 3. This bit is set by writing a\n"
            "  '1' to this bit position. \tSC 0\n"
            "bit 2 ARB_REQ_SET2\n"
            "  Set Software Arbitration request Bit 2. This bit is set by writing a\n"
            "  '1' to this bit position.\tSC 0\n"
            "bit 1 ARB_REQ_SET1\n"
            "  Set Software Arbitration request Bit 1. This bit is set by writing a\n"
            "  '1' to this bit position.\tSC 0\n"
            "bit 0 ARB_REQ_SET0\n"
            "  Set Software Arbitration request Bit 0. This bit is set by writing a\n"
            "  '1' to this bit position.\tSC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_access_enable):
        str = "\nnvm_access_enable - 0x80026424\n"                              
            "  SHARE: NVM access enable register.\n"
            "bit 31-2 unused0\tRO 0\n"
            "bit 1 WR_EN\n"
            "  This bit is no longer required in Xinan. Write enable to flash is\n"
            "  always 1.\tRO 0x1\n"
            "bit 0 EN\n"
            "  This bit is no longer required in Xinan. Access to flash is always\n"
            "  enabled.\tRO 0x1\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_write1):
        str = "\n nvm_write1 - 0x80026428\n"
            "  SHARE: NVM write1 configuration register.\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-8 WRDI_CMD\n"
            "  Flash write disable command when device with protection function\n"
            "  is used. This command will be issued by the flash interface state\n"
            "  machine through SPI interface To flash device, and make the flash\n"
            "  device write-disabled.\tRW 0\n"                                  
            "bit 7-0 WREN_CMD\n"
            "  Flash write enable command when device with protection function\n"
            "  is used. This command will be issued by the flash interface state\n"
            "  machine through SPI interface To flash device, and make the flash\n"
            "  device write-enabled.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg4):
        str = "\nnvm_cfg4 - 0x8002642c\n"
            "  SHARE: NVM configuration four register.\n"
            "bit 31-11 RESERVED Reserved for future use\tRW 0\n"

            "bit 10 SR_TURNAROUND 	When this bit is set, \n"
            "  a turnaround cycle is inserted in between the \n"
            "  address and data phases of a status read for \n"
            "  Atmel devices. This bit should only be set when \n"
            "  the legacy status read command (0x57) is used. \n"
            "  This bit defaults to 0.\tRW 	0\n"
            "bit 9 	PASS_MODE_RELAXED_TIMING 	When this bit \n"
            "  is set, the pass mode data is captured one cycle \n"
            "  later than normal. If using pass mode, this bit \n"
            "  should be set whenever the si_input_relaxed_timing \n"
            "  bit is set. This bit defaults to 0. 	RW 	0\n"
            "bit 8 	SI_INPUT_RELAXED_TIMING 	When this bit is \n"
            "  set, the SI input from the external flash device is \n"
            "  latched one cycle later than normal. This bit \n"
            "  defaults to 0. 	RW 	0\n"
            "bit 7 	FAST 	Fast Mode. When this bit is set in ST \n"
            "  mode, fast read command is used. In Atmel mode, this \n"
            "  bit should be set when using the 0xE8 read command. \n"
            "  It should be cleared when using the 0x68 read command. \n"
            "  This value is self-configured on reset based on the \n"
            "  external device. 	RW 	X\n"
            "bit 6 STATUS_BIT_POLARITY\n"
            "  This bit determines how the status bit of the device\n"
            "  status register is interpreted by hardware. If 0, then\n"
            "bit 0 means \"ready\". If 1, then 1 means \"ready\". For\n"
            "  Atmel, this defaults to 1. For ST, this defaults to 0.\n"
            "  This value is self-configured on reset based on the\n"
            "  strap values. It can be overriden.\tRW X\n"
            "bit 5-4 MODE_256_EMPTY_BIT_LOC\n"
            "  Bit location for hardware to insert an empty address\n"
            "bit when MODE_256 is not set with Atmel devices.\n"
            "  This value is self-configured on reset based on the\n"
            "  external device. NOTE: Max Atmel device size is 64\n"
            "  Mbit due to different bus protocols.\tRW X\n"
            "  Value Name  Description\n"
            "   0    BIT8    For 264 Byte pages\n"
            "   1    BIT9    For 528 Byte pages\n"
            "   2    BIT10   For 1056 Byte pages\n"
            "   3    BIT11   For 2112 Byte pages\n"
            "bit 3 FLASH_VENDOR\n"
            "  This bit is self-configured on reset based on the strap\n"
            "  values. It can be overriden.\n"                                  
            "  Value Name Description\tRW X\n"
            "   0    ST\n"
            "   1    ATMEL\n"
            "bit 2-0 FLASH_SIZE\n"
            "  Size of the external flash device. This information is\n"
            "  not used by FLSH hardware. It is only used by\n"
            "  software. This value is self-configured on reset\n"
            "  based on the external device.\tRW X\n"
            "  Value Name Description\n"
            "   0    1MBIT\n"
            "   1    2MBIT\n"
            "   2    4MBIT\n"
            "   3    8MBIT\n"
            "   4    16MBIT\n"
            "   5    32MBIT\n"
            "   6    64MBIT\n"
            "   7    128MBIT\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_reconfig):
        str = "\n   nvm_reconfig - 0x80026430\n"
            "  SHARE: NVM re-configuration register.\n"
            "bit 31 RECONFIG_DONE\n"
            "  This bit is 0 on reset. After software finishes\n"
            "  reconfiguring FLSH, they will set this bit to 1 to\n"
            "  indicate that FLSH has been reconfigured. This bit has\n"
            "  no hardware functionality.\tRW 0\n"
            "bit 30-8 RESERVED Reserved for future use\tRW 0\n"
            "bit 7-4 RECONFIG_STRAP_VALUE\n"
            "  Used by software to numerically encode how the\n"
            "  FLSH has been reconfigured. On reset, this register is\n"
            "  set to the same value as ORIG_STRAP_VALUE.\n"
            "  These bits have no hardware functionality.\tRW X\n"
            "bit 3-0 ORIG_STRAP_VALUE\n"
            "  Strap value from iologic pins. Only bit[0] is used.\n"
            "  Bits[3:1] are for future support.\n"
            "  Value Name Description\tRO X\n"
            "   0    ST\n"
            "   1    ATMEL\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_config):
        str = "\n   smbus_config - 0x80028000\n"
            "  SHARE:\n"
            "bit 31 RESET\n"
            "  When this bit is set it will reset SMBUS block to its\n"
            "  default state. \tSC 0\n"
            "bit 30 SMB_EN\n"
            "  When this bit is '1', the SMBUS block is enabled for\n"
            "  operation. When set the SMBUS block will abort\n"
            "  current transaction in compliance with the SMBUS\n"
            "  master and slave behavior at the end of the current\n"
            "  transaction and stop responding to the SMBUS\n"
            "  master/slave transactions.\tRW 0\n"
            "bit 29 BIT_BANG_EN\n"                                              
            "  When this bit is '1', the SMBUS block is placed into\n"
            "bit-bang mode. SMBUS interface pins are controlled\n"
            "  using Bit-Bang Control Register.\tRW 0\n"
            "bit 28 EN_NIC_SMB_ADDR_0\n"
            "  When this bit is '1' the SMBUS block responds to slave\n"
            "  address 7'b0000000.\tRW 0\n"
            "bit 27 PROMISCOUS_MODE\n"
            "  When this bit is '1' the SMBUS block responds to all\n"
            "  SMBUS transactions regardless of the slave address.\tRW 0\n"
            "bit 26 TIMESTAMP_CNT_EN\n"                                         
            "  When this bit is '1' the TIMESTAMP counter is enabled.\n"
            "  When '0' the counter holds its value.\tRW 0\n"
            "bit 25-20 unused2\tRO 0\n"
            "bit 19-16 MASTER_RTRY_CNT\n"                           
            "  This bit indicates a number of retries in case where\n"
            "  SMBUS block acted as a master and lost SMBUS\n"
            "  arbitration. HW will retry transaction as many times as\n"
            "  specified in this register before it reports lost of\n"
            "  arbitration status to the firmware. When this field is 0\n"
            "  firmware will not do any retries but it will report loss of\n"
            "  arbitration on the initial attempt.\tRW 0xf\n"
            "bit 15-10 unused1\tRO 0\n"
            "bit 9 ARP_EN1\n"
            "  When this bit is set the SMBUS block will respond to\n"          
            "  ARP that is it to SMBUS Device Default Address\n"
            "  (7'b1100001) and reslove NIC_SMB_ADDR1 using\n"
            "  ARP.\tRW 0\n"
            "bit 8 ARP_EN0\n"
            "  When this bit is set the SMBUS block will respond to\n"
            "  ARP that is it to SMBUS Device Default Address\n"
            "  (7'b1100001) and reslove NIC_SMB_ADDR0 using\n"
            "  ARP.\tRW 0\n"
            "bit 7 HW_ARP_ASSIGN_ADDR\n"
            "  When this bit is set HW will service the ARP Assign\n"
            "  Address command, set the AR_FLAG[1:0] and\n"
            "  AV_FLAG[1:0] flags, and program the\n"
            "  NIC_SMB_ADDR[1:0] values.\tRW 0\n"
            "bit 6-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_timing_config):
        str = "\nsmbus_timing_config - 0x80028004\n"
            "  SHARE:\n"
            "bit 31 MODE_400\n"
            "  When this bit is set the SMBUS block operates in\n"
            "bit 400KHz mode. When cleared SMBUS operates in\n"
            "bit 100KHz mode.\tRW 0\n"
            "bit 30-24 RANDOM_SLAVE_STRETCH\n"
            "  These bits specify time for which clock low time\n"
            "  will be stretched after each byte (that is ACK bit)\n"
            "  when the SMBUS block acts as a slave. This is\n"
            "  useful in \"legacy mode\" to allow firmware time to\n"
            "  handle the data. Note that this time contributes to\n"
            "  the slave TLOW:SEXT time, that is combined\n"
            "  random and periodic slave stretch should not\n"
            "  exceed 25ms. Register has 1ms resolution. Default\n"
            "  is 25ms.\tRW 0x19\n"
            "bit 23-16 PERIODIC_SLAVE_STRETCH\n"
            "  These bits specify time for which each clock period\n"
            "  low time will be stretched when the SMBUS block\n"
            "  acts as a slave. Note that a cumulative clock low\n"
            "  extend time (TLOW:SEXT) for which slave device\n"
            "  is allowed to stretch the clock from the beginning to\n"
            "  end of the message (that is from START to STOP) is\n"
            "bit 25ms. For example, if the message is Block Write\n"
            "  transaction 36B long allowed periodic stretch would\n"
            "  be 25ms/(9*36) ~= 77us. This is assuming that\n"
            "  random slave stretching is not used. Register has\n"             
            "bit 1us resolution.\tRW 0\n"
            "bit 15-8 SMBUS_IDLE_TIME\n"
            "  These bits specify the time for which both\n"
            "  SMBCLK and SMBDAT must be high before a\n"
            "  master can assume that bus is free. Register has 1us\n"
            "  resolution. Default is 50us.\tRW 0x32\n"
            "bit 7-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_address):
        str = "\n  smbus_address - 0x80028008\n"
            "  SHARE:\n"
            "bit 31 EN_NIC_SMB_ADDR3\n"
            "  When this bit is '1' NIC_SMB_ADDR3 will be used as a\n"
            "  slave address to match for incoming messages.\tRW 0\n"
            "bit 30-24 NIC_SMB_ADDR3\n"
            "  This is the fourth SM Bus addresses which will be used to\n"
            "  match for incoming messages. When the\n"
            "  PROMISCOUS_MODE bit is '1', this value is ignored.\tRW 0\n"
            "bit 23 EN_NIC_SMB_ADDR2\n"
            "  When this bit is '1' NIC_SMB_ADDR2 will be used as a\n"
            "  slave address to match for incoming messages.\tRW 0\n"           
            "bit 22-16 NIC_SMB_ADDR2\n"
            "  This is the third SM Bus addresses which will be used to\n"
            "  match for incoming messages. When the\n"
            "  PROMISCOUS_MODE bit is '1', this value is ignored. 0x0\n"
            "  This address is also used for ARP. The address will be\n"
            "  resolved using ARP when the ARP_EN1 bit is set.\tRW 0\n"
            "bit 15 EN_NIC_SMB_ADDR1\n"
            "  When this bit is '1' NIC_SMB_ADDR1 will be used as a\n"
            "  slave address to match for incoming messages.\tRW 0\n"
            "bit 14-8 NIC_SMB_ADDR1\n"
            "  This is the second SM Bus addresses which will be used to\n"
            "  match for incoming messages. When the\n"
            "  PROMISCOUS_MODE bit is '1', this value is ignored.\tRW 0\n"
            "bit 7 EN_NIC_SMB_ADDR0\n"
            "  When this bit is '1' NIC_SMB_ADDR0 will be used as a\n"
            "  slave address to match for incoming messages.\tRW 0\n"
            "bit 6-0 NIC_SMB_ADDR0\n"
            "  This is the first SM Bus addresses which will be used to\n"
            "  match for incoming messages. This address is also used for\n"
            "  ARP that is this is the address that will be resolved using\n"
            "  ARP when the ARP function is enabled. When the\n"
            "  PROMISCOUS_MODE bit is '1', this value is ignored.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_fifo_control):
        str = "\n    smbus_master_fifo_control - 0x8002800c\n"
            "  SHARE:\n"
            "bit 31 MASTER_RX_FIFO_FLUSH\n"
            "  When this bit is set HW will flush Master RX\n"
            "  FIFO when the current RX transaction\n"                              
                        "  completes.\tSC 0\n"
            "bit 30 MASTER_TX_FIFO_FLUSH\n"
            "  When this bit is set HW will flush Master TX\n"
            "  FIFO when current TX transaction completes.\tSC 0\n"
            "bit 29-23 unused2\tRO 0\n"
            "bit 22-16 MASTER_RX_PKT_COUNT Number of packets in the Master RX FIFO.\tRO 0\n"
            "bit 15 unused1\tRO 0\n"
            "bit 14-8 MASTER_RX_FIFO_THRESHOLD\n"
            "  When the Master Receive FIFO hits this\n"
            "  threshold the SMBUS block will generate an\n"
            "  event for the control processor. When set to 0x0\n"
            "  event generation is disabled. Threshold is\n"
            "  specified with the byte resolution.\tRW 0\n"
            "bit 7-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_fifo_control):
        str = "\n     smbus_slave_fifo_control - 0x80028010\n"
            "  SHARE:\n"
            "bit 31 SLAVE_RX_FIFO_FLUSH\n"
            "  When this bit is set HW will flush Slave RX FIFO\n"
            "  when the current RX transaction completes.\tSC 0\n"
            "bit 30 SLAVE_TX_FIFO_FLUSH\n"
            "  When this bit is set HW will flush Slave TX FIFO\n"
            "  when current TX transaction completes.\tSC 0\n"
            "bit 29-23 unused2\tRO 0\n"                                         
            "bit 22-16 SLAVE_RX_PKT_COUNT Number of packets in the Slave RX FIFO.\tRO 0\n"
            "bit 15 unused1\tRO 0\n"
            "bit 14-8 SLAVE_RX_FIFO_THRESHOLD\n"
            "  When the Slave Receive FIFO hits this threshold\n"
            "  the SMBUS block will generate an event for the\n"
            "  control processor. When set to 0x0 event\n"
            "  generation is disabled. Threshold is specified with\tRW 0\n"
            "  the byte resolution.\n"
            "bit 7-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_bit_bang_control):
        str = "\n smbus_bit_bang_control - 0x80028014\n"
            "  SHARE:\n"
            "bit 31 SMBCLK_IN\n"
            "  This bit reflects the current input value of the SMBCLK pin.\n"
            "  When the SMBCLK pin is high, this bit will read as '1'. When\n"
            "  the SMBCLK pin is low, this pin will read as '0'.\tRO 0\n"
            "bit 30 SMBCLK_OUT_EN\n"
            "  When the SM Bus interface is configured for bit-bang mode,\n"
            "  this bit controlls the output enable for the CLK pin. When this\n"
            "bit is '0', the CLK pin will drive low. When this bit is '1', the\n"
            "  CLK pin will float.\tRW 0x1\n"
            "bit 29 SMBDAT_IN\n"
            "  This bit reflects the current input value of the SMBDAT pin.\n"
            "  When the SMBDAT pin is high, this bit will read as '1'. When\n"
            "  the SMBDAT pin is low, this pin will read as '0'.\tRO 0\n"
            "bit 28 SMBDAT_OUT_EN\n"
            "  When the SMBUS interface is configured for bit-bang mode,\n"
            "  this bit controlls the output enable for the SMBDAT pin. When\n"
            "  this bit is '0', the SMBDAT pin will drive low. When this bit is\n"
            "  '1', the SMBDAT pin will float.\tRW 0x1\n"
            "bit 27-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_watchdog):
        str = "\n smbus_watchdog - 0x80028018\n"
            "  SHARE:\n"
            "bit 31-16 unused0\tRO 0\n"                                         
            "bit 15-0 WATCHDOG\n"
            "  This value counts down to zero once each second and sets the\n"
            "  WG_TO bit when it reaches zero. The counter stops when it\n"
            "  reaches zero.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_heartbeat):
        str = "\nsmbus_heartbeat - 0x8002801c\n"
            "  SHARE:\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-0 HEARTBEAT\n"
            "  This value counts down to zero once each second and sets the\n"
            "  HB_TO bit when it reaches zero. The counter stops when it reaches\n"
            "  zero.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_poll_asf):
        str = "\n smbus_poll_asf - 0x80028020\n"
            "  SHARE:\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-0 POLL_ASF\n"
            "  This value counts down to zero once each 5 msec. and sets the\n"
            "  PA_TO bit when it reaches zero. The counter stops when it reaches\n"
            "  zero.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_poll_legacy):
        str = "\nsmbus_poll_legacy - 0x80028024\n"
            "  SHARE:\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-0 POLL_LEGACY\n"
            "  This value counts down to zero once each 250 msec. and sets the\n"
            "  PL_TO bit when it reaches zero. The counter stops when it\n"
            "  reaches zero.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_retran):
        str = "\n   smbus_retran - 0x80028028\n"
            "  SHARE:\n"
            "bit 31-8 unused0\tRO 0\n"
            "bit 7-0 RETRAN\n"
            "  This value counts down to zero once each second and sets the\n"
            "  RT_TO bit when it reaches zero. The counter stops when it reaches\n"
            "  zero.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_timestamp):
        str = "\nsmbus_timestamp - 0x8002802c\n"
            "  SHARE:\n"
            "bit 31-0 TIMESTAMP\n"
            "  This value counts up once each second and rolls to zero each time it\n"
            "  passes 0xffffffff. This counter only counts when the\n"
            "  TIMESTAMP_CNT_EN bit is '1'.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_command):
        str = "\nsmbus_master_command - 0x80028030\n"
            "  SHARE:\n"
            "bit 31 START_BUSY\n"
            "  This bit is self clearing. When written to a '1', the currently\n"
            "  programmed SMBUS transaction will activate. Writing this bit\n"
            "  as a '0' has no effect. This bit must be read as a '0' before\n"
            "  setting it to prevent un-predictable results.\tSC 0\n"
            "bit 30 ABORT\n"
            "  Transaction Abort. This bit can be set at any time by the\n"
            "  firmware or the driver in order to abort the transaction. The\n"
            "  HW will abort transaction in compliance with the SMBUS\tSC 0\n"  
            "  rules and clear the bit when done.\n"
            "bit 29-28 unused1\tRO 0\n"
            "bit 27-25 STATUS\n"
            "  These bits encode status of the last master transaction. Valid\n"
            "  when START_BUSY is cleared after it was set\tRO 0\n"
            "  Value Name Description\n"
            "   0     000 Transaction completed successfully\n"
            "   1     001 Lost Arbitration. Firmware should restart\n"
            "             transaction, if required.\n"
            "   2     010 NACK detected after (slave address) first\n"
            "             byte. Indicates that slave device is off-line.\n"
            "   3     011 NACK detected after byte other than first.\n"
            "             Indicates that slave device maybe busy.\n"
            "   4     100 TTIMEOUT_MIN exceeded. Indicates that\n"
            "             slave device held bus for more then 25ms\n"
            "             (assuming that our clock driver is\n"
            "             functioning properly).\n"
            "   5     101  TLOW:MEXT exceeded. Used for\n"
            "             non-standard transactions where payload is\n"         
            "             longer than the TX FIFO. Indicates FIFO\n"
            "             under-run (firmware didn't provide\n"
            "             additional data within 10ms).\n"
            "   6     110 TLOW:MEXT exceeded. Indicates that\n"
            "             read transaction failed due the Master Rx\n"
            "             FIFO being full for more than 10ms.\n"
            "   7     111 Reserved\n"
            "bit 24-13 unused0\tRO 0\n"
            "bit 12-9 SMBUS_PROTOCOL\n"
            "  SMBUS Protocol encoded as\tRW 0\n"
            "  Value Name Description\n"
            "   0    0000 Quick Command\n"
            "   1    0001 Send Byte\n"
            "   2    0010 Receive Byte\n"
            "   3    0011 Write Byte\n"
            "   4    0100 Read Byte\n"
            "   5    0101 Write Word\n"
            "   6    0110 Read Word\n"
            "   7    0111 Block Write\n"
            "   8    1000 Block Read\n"
            "   9    1001 Process Call\n"
            "   10   1010 Block Write-Block Read Process Call\n"                
            "   11   1011 Host Notify\n"
            "bit 8 PEC PEC should be checked or calculated for this transaction.\tRW 0\n"
            "bit 7-0 RD_BYTE_COUNT\n"
            "  This is number of bytes that SMBUS block should read from\n"
            "   Blockthe slave in Block Write - Block Read Process Call or Block\n"
            "  Read. If this field is 0 the SMBUS block will assume that first\n"
            "  byte received from the slave is a Byte Count. If different than\tRW 0\n"
            "bit 0 HW will use this value as an indication as where to insert\n"
            "  NACK and STOP that is end the transaction.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_command):
        str = "\nsmbus_slave_command - 0x80028034\n"
            "  SHARE:\n"
            "bit 31 START\n"
            "  _BUSY This bit is self clearing. When written to a '1', the currently\n"
            "  programmed SMBUS transaction will activate. Writing this bit as a\n"
            "  '0' has no effect. This bit must be read as a '0' before setting it ;to\n"
            "  prevent un-predictable results. Used only for read response.\tSC 0\n"
            "bit 30 ABORT\n"
            "  Transaction Abort. This bit can be set at any time by the firmware\n"
            "  or the driver in order to abort the transaction. The HW will abort\n"
            "  transaction in compliance with the SMBUS rules and clear the bit\n"
            "  when done. When set and the slave transaction is received the HW\n"
            "  will ACK the address and float the bus thereafter.\tSC 0\n"
            "bit 29-26 unused2\tRO 0\n"                                         
            "bit 25-23 STATUS\n"
            "  These bits encode status of the last slave transaction. Valid when\n"
            "  START_BUSY is cleared after it was set\tRO 0\n"
            "  Value Name Description\n"
            "   0    000  Transaction completed successfully\n"
            "   5    101  TTIMEOUT_MIN exceeded. Indicates that clock\n"
            "             was held low for more then\n"
            "             RANDOM_SLAVE_STRETCH by the SMBUS\n"                  
            "             or 25ms by the master.\n"
            "   7    111  Master aborted read transaction. HW will flush\n"
            "             the Slave Transmit FIFO.\n"
            "bit 22-9 unused1\tRO 0\n"
            "bit 8 PEC PEC should be calculated for this transaction.\tRW 0\n"
            "bit 7-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_event_enable):
        str = "\nsmbus_event_enable - 0x80028038\n"                             
            "  SHARE:\n"
            "bit 31 MASTER_RX_FIFO_FULL_EN\n"
            "  When set enables MASTER_RX_FIFO_FULL\n"
            "  to generate smbus event.\tRW 0\n"
            "bit 30 MASTER_RX_THRESHOLD_HIT_EN\n"
            "  When set enables\n"
            "  MASTER_RX_THRESHOLD_HIT to\n"                                    
            "  generate smbus event.\tRW 0\n"
            "bit 29 MASTER_RX_EVENT_EN\n"
            "  When set enables MASTER_RX_EVENT to\n"
            "  generate smbus event.\tRW 0\n"
            "bit 28 MASTER_START_BUSY_EN\n"
            "  When set enables generation of a smbus event\n"
            "  on master START_BUSY 1 to 0 transition.\tRW 0\n"
            "bit 27 MASTER_TX_UNDERRUN_EN\n"
            "  When set enables generation of a smbus event\n"
            "  when Master Tx FIFO becomes empty and less\n"
            "  then PKT_LENGTH bytes were output on the\n"
            "  SMBUS.\tRW 0\n"
            "bit 26 SLAVE_RX_FIFO_FULL_EN\n"
            "  When set enables SLAVE_RX_FIFO_FULL\n"
            "  to generate smbus event.\tRW 0\n"
            "bit 25 SLAVE_RX_THRESHOLD_HIT_EN\n"
            "  When set enables\n"
            "  SLAVE_RX_THRESHOLD_HIT to generate\n"
            "  smbus event.\tRW 0\n"
            "bit 24 SLAVE_RX_EVENT_EN\n"
            "  Enables SLAVE_RX_EVENT to generate\n"
            "  smbus event.\tRW 0\n"
            "bit 23 SLAVE_START_BUSY_EN\n"
            "  When set enables generation of a smbus event\n"
            "  on slave START_BUSY 1 to 0 transition.\tRW 0\n"
            "bit 22 SLAVE_TX_UNDERRUN_EN\n"
            "  When set enables generation of a smbus event\n"
            "  when Slave Tx FIFO becomes empty and less\n"
            "  then PKT_LENGTH bytes were output on the\n"
            "  SMBUS.\tRW 0\n"
            "bit 21 SLAVE_RD_EVENT_EN\n"
            "  Enables SLAVE_RD_EVENT to generate\n"
            "  smbus event.\tRW 0\n"
            "bit 20 SLAVE_ARP_EVENT_EN\n"
            "  When set enables hardware to generate smbus\n"
            "  event any time and ARP command is received\n"
            "  and ARP_EN0 or ARP_EN1 bit is set.\tRW 0\n"
            "bit 19-5 unused0\tRO 0\n"
            "bit 4 RETRANSMIT_TO_EN\n"
            "  When set enables Retransmit Timer to\n"
            "  generate smbus event.\tRW 0\n"
            "bit 3 POLL_LEGACY_TO_EN\n"
            "  When set enables Legacy Sensor Poll Timer to\n"
            "  generate smbus event.\tRW 0\n"
            "bit 2 POLL_ASF_TO_EN\n"
            "  When set enables ASF Sensor Poll Timer to\n"
            "  generate smbus event.\tRW 0\n"
            "bit 1 HEARTBEAT_TO_EN\n"
            "  When set enables Heartbeat Timer to generate\n"
            "  smbus event.\tRW 0\n"
            "bit 0 WATCHDOG_TO_EN\n"
            "  When set enables Watchdog Timer to generate\n"
            "  smbus event.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_event_status):
        str = "\nsmbus_event_status - 0x8002803c\n"
            "  SHARE:\n"
            "bit 31 MASTER_RX_FIFO_FULL\n"
            "  This bit is set when the master receive FIFO\n"
            "  become full. Writing a '1' to this position will clear\n"
            "  this bit. When this bit is '1', the SMB0_EVENT bit\n"
            "  will be '1' in each processor.\tWC 0\n"
            "bit 30 MASTER_RX_THRESHOLD_HIT\n"                                  
            "  This bit is set when the master receive FIFO is\n"
            "  equal to or larger than the Master\n"
            "  RX_FIFO_THRESHOLD. Writing a '1' to this\n"
            "  position will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n"
            "bit 29 MASTER_RX_EVENT\n"
            "  This bit is set when the master receive FIFO holds\n"
            "  at least one valid transaction. Writing a '1' to this\n"
            "  position will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n"
            "bit 28 MASTER_START_BUSY\n"
            "  This bit is set when master START_BUSY\n"
            "  transitions from 1 to 0. Writing a '1' to this\n"
            "  position will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n"
            "bit 27 MASTER_TX_UNDERRUN\n"
            "  This bit is set when Master Tx FIFO becomes\n"
            "  empty and less then PKT_LENGTH bytes were\n"
            "  output on the SMBUS.\tWC 0\n"
            "bit 26 SLAVE_RX_FIFO_FULL\n"
            "  This bit is set when the slave receive FIFO\n"
            "  become full. Writing a '1' to this position will clear\n"
            "  this bit. When this bit is '1', the SMB0_EVENT bit\n"
            "  will be '1' in each processor.\tWC 0\n"
            "bit 25 SLAVE_RX_THRESHOLD_HIT\n"
            "  This bit is set when the slave receive FIFO is\n"
            "  equal to or larger than the Slave\n"
            "  RX_FIFO_THRESHOLD. Writing a '1' to this\n"
            "  position will clear this bit. When this bit is '1', the\n"       
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n"
            "bit 24 SLAVE_RX_EVENT\n"
            "  This bit is set when the slave receive FIFO holds\n"
            "  at least one valid transaction. Writing a '1' to this\n"
            "  position will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n"
            "bit 23 SLAVE_START_BUSY\n"
            "  This bit is set when slave START_BUSY\n"
            "  transitions from 1 to 0. Writing a '1' to this\n"
            "  position will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\n"
            "bit 0x\tWC 0\n"
            "bit 22 SLAVE_TX_UNDERRUN\n"
            "  This bit is set when Slave Tx FIFO becomes\n"
            "  empty and less then PKT_LENGTH bytes were\n"
            "  output on the SMBUS.\tWC 0\n"
            "bit 21 SLAVE_RD_EVENT\n"
            "  This bit is set when slave hardware detected read\n"
            "  transaction directed toward the SMBUS block.\n"
            "  Writing a '1' to this position will clear this bit.\n"
            "  When this bit is '1', the SMB0_EVENT bit will be\n"
            "  '1' in each processor.\tWC 0\n"
            "bit 20 SLAVE_ARP_EVENT\n"
            "  This bit set when slave hardware received an ARP\n"
            "  command and ARP_EN0 or ARP_EN1 bit is set.\tWC 0\n"
            "bit 19-5 unused0\tRO 0\n"
            "bit 4 RETRANSMIT_TO\n"
            "  This bit changes to '1' each time the\n"
            "  RETRANSMIT timer reaches zero. Writing a '1' to\n"
            "  this position will clear this bit. When this bit is '1',\n"
            "  the SMB0_EVENT bit will be '1' in each\n"
            "  processor.\tWC 0\n"
            "bit 3 POLL_LEGACY_TO\n"
            "  This bit changes to '1' each time the\n"
            "  POLL_LEGACY timer reaches zero. Writing a '1'\n"
            "  to this position will clear this bit. When this bit is\n"
            "  '1', the SMB0_EVENT bit will be '1' in each\n"
            "  processor.\tWC 0\n"
            "bit 2 POLL_ASF_TO\n"
            "  This bit changes to '1' each time the POLL_ASF\n"
            "  timer reaches zero. Writing a '1' to this position\n"
            "  will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n"
            "bit 1 HEARTBEAT_TO\n"
            "  This bit changes to '1' each time the\n"
            "  HEARTBEAT timer reaches zero. Writing a '1' to\n"
            "  this position will clear this bit. When this bit is '1',\n"
            "  the SMB0_EVENT bit will be '1' in each\n"
            "  processor.\tWC 0\n"
            "bit 0 WATCHDOG_TO\n"
            "  This bit changes to '1' each time the WATCHDOG\n"
            "  timer reaches zero. Writing a '1' to this position\n"
            "  will clear this bit. When this bit is '1', the\n"
            "  SMB0_EVENT bit will be '1' in each processor.\tWC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_data_write):
        str = "\nsmbus_master_data_write - 0x80028040\n"
            "  SHARE:\n"
            "bit 31 WR_STATUS\n"
            "  1bit 0 - Byte other then last in an WMBUS transaction 1\n"
            "  - End of SMBUS transaction\tRW 0\n"
            "bit 30-8 unused0\tRO 0\n"
            "bit 7-0 MASTER_SMBUS_WR_DATA\n"                                    
            "  This is a software interface to the SMBUS Master\n"
            "  Transmit FIFO. Software should use this register to\n"
            "  provide data for an SMBUS write transaction. Data\n"
            "  should be written before the SMBUS Command is\n"
            "  programmed.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_data_read):
        str = "\n smbus_master_data_read - 0x80028044\n"
            "  SHARE:\n"
            "bit 31-30 RD_STATUS\n"
            "  Enumeration:\tRO 0\n"
            "  Value Name Description\n"
            "   0     00  FIFO Empty\n"
            "   1     01  Start of SMBUS transaction\n"
            "   2     10  Byte other then first or last in an\n"
            "             SMBUS transaction\n"
            "   3     11  End of SMBUS transaction\n"
            "bit 29 PEC_ERR\n"
            "  PEC error. This bit indicates status of              the PEC\n"
            "  checking. HW will check the PEC only in case where\n"
            "  PEC bit in SMBUS Master Command Register was\n"
            "  set for rhe transaction This field is valid only when\n"
            "  RD_STATUS = 2'b11.\tRO 0\n"
            "bit 28-8 unused0\tRO 0\n"
            "bit 7-0 MASTER_SMBUS_RD_DATA\n"
            "  SMBUS Read Data in Network Byte Order (MSB\n"
            "  first). This is a software interface to the SMBUS\n"
            "  Master Receive FIFO. Software should use this\n"
            "  register to read data received from the SMBUS.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_data_write):
        str = "\n smbus_slave_data_write - 0x80028048\n"
            "  SHARE:\n"
            "bit 31 WR_STATUS\n"
            "  Enumeration:\tRW 0\n"
            "  Value Name Description\n"
            "   0   0     Byte other then last in an SMBUS\n"
            "             transaction\n"
            "   1   1     End of SMBUS transaction\n"
            "bit 30-8 unused0\tRO 0\n"
            "bit 7-0 SLAVE_SMBUS_WR_DATA\n"
            "  This is a software interface to the SMBUS Slave\n"
            "  Transmit FIFO. Software should use this register to\n"
            "  provide data for an SMBUS read transaction.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_data_read):
        str = "\nsmbus_slave_data_read - 0x8002804c\n"
            "  SHARE:\n"
            "bit 31-30 RD_STATUS\n"
            "  Enumeration:\tRO 0\n"
            "  Value Name Description\n"
            "   0      00 FIFO Empty\n"
            "   1      01 Start of SMBUS transaction\n"
            "   2      10 Byte other then first or last in an\n"
            "             SMBUS transaction\n"
            "   3      11 End of SMBUS transaction\n"
            "bit 29-28 ERR_STATUS\n"
            "  This field is valid only when RD_STATUS = 2'b11.\tRO 0\n"
            "  Value Name Description\n"
            "   0    00   Transaction completed successfully\n"
            "   1    01   PEC error. This bit indicates status of\n"
            "             the PEC checking. HW assumes that\n"
            "             last byte of the transaction is always\n"
            "             PEC byte and sets this bit accordingly.\n"
            "             It is up to the software to recognize\n"
            "             (based on the Command Code) if the\n"
            "             particular transaction has a PEC or not.\n"
            "             If it does not software should ignore this\n"
            "             bit.\n"
            "   2    10   FIFO overflow. SMBUS transaction\n"
            "             timed out due to the receive FIFO being\n"
            "             full for more than\n"
            "             RANDOM_SLAVE_STRETCH time.\n"
            "   3    11   General SMBUS transaction error. This\n"
            "             is used for all errors not covered by b01\n"
            "             and b10 errors.\n"
            "bit 27-8 unused0\tRO 0\n"
            "bit 7-0 SLAVE_SMBUS_RD_DATA\n"
            "  This is a software interface to the SMBUS Slave\n"
            "  Receive FIFO. Software should use this register to read\n"
            "  data received from the SMBUS.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_arp_state):
        str = "\nsmbus_arp_state - 0x80028080\n"
            "  SHARE:\n"
            "bit 31-6 unused1\tRO 0\n"
            "bit 5 AR_FLAG1\n"
            "  This bit should be set by firmware before ARP_EN1 bit is set. This\n"
            "bit is typically initialized to '0'.\tRW 0\n"
            "bit 4 AV_FLAG1\n"
            "  This bit should be set by firmware before ARP_EN1 bit is set. This\n"
            "bit is typically set to '1' based on the NVRAM content that is if\n"
            "  device supports Persistent Slave Address and cleared otherwise.\tRW 0\n"
            "bit 3-2 unused0\tRO 0\n"
            "bit 1 AR_FLAG0\n"
            "  This bit should be set by firmware before ARP_EN0 bit is set. This\n;"
            "bit is typically initialized to '0'.\tRW 0\n"
            "bit 0 AV_FLAG0\n"
            "  This bit should be set by firmware before ARP_EN0 bit is set. This\n"
            "bit is typically set to '1' based on the NVRAM content that is if\n"
            "  device supports Persistent Slave Address and cleared otherwise.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_3):
        str = "\n  smbus_udid0_3 - 0x80028090\n"
            "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
            "bit 31-24 BYTE_15 UDID_0 byte 15.\tRW 0\n"
            "bit 23-16 BYTE_14 UDID_0 byte 14.\tRW 0\n"
            "bit 15-8 BYTE_13 UDID_0 byte 13.\tRW 0\n"
            "bit 7-0 BYTE_12 UDID_0 byte 12.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_2):                                    
        str = "\n  smbus_udid0_2 - 0x80028094\n"
            "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
            "bit 31-24 BYTE_11 UDID_0 byte 11.\tRW 0\n"
            "bit 23-16 BYTE_10 UDID_0 byte 10.\tRW 0\n"
            "bit 15-8 BYTE_9 UDID_0 byte 9.\tRW 0\n"
            "bit 7-0 BYTE_8 UDID_0 byte 8.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_1):
        str = "\n  smbus_udid0_1 - 0x80028098\n"
            "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
            "bit 31-24 BYTE_7 UDID_0 byte 7.\tRW 0\n"
            "bit 23-16 BYTE_6 UDID_0 byte 6.\tRW 0\n"
            "bit 15-8 BYTE_5 UDID_0 byte 5.\tRW 0\n"
            "bit 7-0 BYTE_4 UDID_0 byte 4.\tRW 0\n"                ;                             
                case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_0):
        str = "\n  smbus_udid0_0 - 0x8002809c\n"
            "  SHARE: This register should be set by firmware before ARP_EN0 bit is set.\n"
            "bit 31-24 BYTE_3 UDID_0 byte 3.\tRW 0\n"
            "bit 23-16 BYTE_2 UDID_0 byte 2.\tRW 0\n"
            "bit 15-8 BYTE_1 UDID_0 byte 1.\tRW 0\n"
            "bit 7-0 BYTE_0 UDID_0 byte 0.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_3):
        str = "\n  smbus_udid1_3 - 0x800280a0\n"
            "  SHARE: This register should be set by firmware before ARP_EN1 bit is; set.\n"
            "bit 31-24 BYTE_15 UDID_1 byte 15.\tRW 0\n"
            "bit 23-16 BYTE_14 UDID_1 byte 14.\tRW 0\n"
            "bit 15-8 BYTE_13 UDID_1 byte 13.\tRW 0\n"
            "bit 7-0 BYTE_12 UDID_1 byte 12.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_2):
        str = "\n  smbus_udid1_2 - 0x800280a4\n"
            "  SHARE: This register should be set by firmware before ARP_EN1 bit is set.\n"
            "bit 31-24 BYTE_11 UDID_1 byte 11.\tRW 0\n"
            "bit 23-16 BYTE_10 UDID_1 byte 10.\tRW 0\n"
            "bit 15-8 BYTE_9 UDID_1 byte 9.\tRW 0\n"
            "bit 7-0 BYTE_8 UDID_1 byte 8.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_1):
        str = "\n  smbus_udid1_1 - 0x800280a8\n"
            "  SHARE: This register should be set by firmware before ARP_EN1 bit is; set.\n"
            "bit 31-24 BYTE_7 UDID_1 byte 7.\tRW 0\n"
            "bit 23-16 BYTE_6 UDID_1 byte 6.\tRW 0\n"
            "bit 15-8 BYTE_5 UDID_1 byte 5.\tRW 0\n"
            "bit 7-0 BYTE_4 UDID_1 byte 4.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_0):
        str = "\n  smbus_udid1_0 - 0x800280ac\n"
            "  SHARE: This register should be set by firmware before ARP_EN1 bit is set.\n"
            "bit 31-24 BYTE_3 UDID_1 byte 3.\tRW 0\n"
            "bit 23-16 BYTE_2 UDID_1 byte 2.\tRW 0\n"
            "bit 15-8 BYTE_1 UDID_1 byte 1.\tRW 0\n"
            "bit 7-0 BYTE_0 UDID_1 byte 0.\tRW 0\n";
        break;
#ifndef UEFI
    case (OFFSETOF(mcp_fio_t, mcpf_legacy_unused_legacy_smb)+9*sizeof(u32_t)):
        str = "\n  legacy_unused_legacy_smb[9] - 0x80028800 - (RO/Reset: 0)\n"
            "  SHARE: This location provides was used for legacy SMBUS in prior\n"
            "  devices. This is no longer supported\n";
        break;    case OFFSETOF(mcp_fio_t, mcpf_ump_cmd):
        str = "\n ump_cmd - 0x80038000\n"
            "  SHARE:\n"
            "bit 31 INGRESS_MAC_RESET\n"
            "  When this bit is '1', a soft reset will be performed to the\n"
            "  Ingress MAC in the UMP block.\tRW 0\n"
            "bit 30 EGRESS_MAC_RESET\n"
            "  When this bit is '1', a soft reset will be performed to the\n"
            "  Egress MAC in the UMP block.\tRW 0\n"
            "bit 29 INGRESS_FIO_RESET\n"
            "  When this bit is '1', a soft reset will be performed to the\n"
            "  FIO interface of the UMP Ingress interface. \tSC 0\n"
            "bit 28 EGRESS_FIO_RESET\n"
            "  When this bit is '1', a soft reset will be performed to the\n"
            "  FIO interface of the UMP Egress interface.\tSC 0\n"
            "bit 27-22 unused3\tRO 0\n"
            "bit 21-17 CMD_IPG\n"
            "  This bit programms the interpacket gap for the Ingress\n"
            "  packets. The IPG is calculated in nibbles, so a value of\n"
            "bit 24 indicates 12bytes of IPG.\tRW 0x18\n"
            "bit 16 EGRESS_PKT_FLUSH\n"
            "  When this bit is '1', the packet at the head of the UMP\n"       
            "  Egress FIFO will be flushed and if there are more\n"
            "  packets in the FIFO, appropriate actions will be taken.\n"
            "  This allows the firmware to look at packets of interest\n"
            "  only and flush a packet by a single command rather\tSC 0\n"
            "  than read out the packet word by word.\n"
            "bit 15 NO_PLUS_TWO\n"
            "  When this bit is '1', the \"plus-two\" alignment for both\n"
            "  Ingress and Egress traffic is disabled. This means that\n"
            "  the packet will be aligned such that byte 0 of the\n"
            "  packet appears in bits [31:24] of the rd_data and\n"
            "  wr_data registers. When this bit is '0', the \"plus-two\"\n"
            "  alignment for both Ingress and Egress traffic is\n"
            "  enabled. This means that the packet will be aligned\n"
            "  such that byte 0 of the packet appears in bits [15:8] of\n"
            "  the rd_data and wr_data registers.\tRW 0\n"
            "bit 14 INGRESS_RESET\n"
            "  When this bit is written as '1', the Ingress state machine\n"
            "  is reset to the idle state. This bit should be written as '1'\n"
            "  when a transmitter wishes to recall a transmit operation\n"      
            "  that has been requested by a write to the frm_wr_ctl\n"
            "  register.\tSC 0\n"
            "bit 13 AUTO_DRIVE\n"
            "  When this bit is '1', a mode is enabled where the\n"
            "  Ingress pins of the UMP interface will automatically\n"
            "  drive from one clock before a Ingress packet until the\n"
            "  clock after the Ingress packet has been transmitted.\n"          
            "  When this bit is '0', the tri-state control is done by the\n"
            "  INGRESS_DRIVE bit of this register. For this mode\n"
            "  to operate correctly, the INGRESS_DRIVE bit must\n"
            "  be '0'.\tRW 0\n"
            "bit 12-10 unused2\tRO 0\n"
            "bit 9 SW_PAUSE\n"
            "  When this bit is '1', it forces the \"pause needed\" input\n"  
            "  to the MAC to '1'. This may be used by FW to force\n"
            "  pause packets to be generated during times when it\n"
            "  knows that it can not process received packets\n"
            "  properly.\tRW 0\n"
            "bit 8 INGRESS_DRIVE\n"
            "  If this bit is set, then the Ingress pins of the UMP\n"
            "  interface will be forced to drive reguardless of the\n"          
            "  setting of the AUTO_DRIVE bit.\tRW 0\n"
            "bit 7 unused1\tRO 0\n"
            "bit 6 INGRESS_MAC_DISABLE\n"
            "  When this bit is set, it disables the transmit function of\n"
            "  the UMP MAC. Setting this bit to '1' will reset the\n"
            "  UMP port transmitter to a known state. This bit must\n"
            "  be set to '0' to allow transmission.\tRW 0x1\n"                  
            "bit 5 EGRESS_MAC_DISABLE\n"
            "  When this bit is set, it disables the receive function of\n"
            "  the UMP MAC. Setting this bit to '1' will reset the\n"
            "  UMP port receiver to a known state. This bit must be\n"
            "  set to '0' to allow reception of new frames.\tRW 0x1\n"
            "bit 4 unused0\tRO 0\n"
            "bit 3 MAC_LOOPBACK\n"                                              
            "  When this bit is '1', Ingress data will be looped back to\n"
            "  the Egress at the RMII interface. This feature only\n"
            "  works in RMII mode\tRW 0\n"
            "bit 2 FC_EN\n"
            "  When this bit is '1', it indicates that flow control is\n"
            "  enabled. If this bit is '0', no attempt will be made to\n"
            "  prevent packet drop.\tRW 0\n"                                    
            "bit 1 INGRESS_FIFO_ENABLED\n"
            "  This bit controlls the current enable status of the\n"
            "  Ingress side of the UMP FIFO block. If this bit is '1', it\n"
            "  indicates that the Ingress FIFO is enabled When this bit\n"
            "  is '0', then FIFO will be reset immediately. If the\n"
            "  transmit MAC is currently transmitting a packet and\n"
            "  does not get all the data needed for the packet, it will\n"      
            "  underun and transmit JAM for the packet and then\n"
            "  stop.\tRW 0\n"
            "bit 0 EGRESS_FIFO_ENABLED\n"
            "  This bit controlls the current enable status of recevie\n"
            "  UMP FIFO block. If this bit is '1', it indicates that the\n"
            "  RX FIFO is enabled When this bit is '0', the reset FIFO\n"
            "  will be reset immediately. Any partial or complete\n"            
            "  packets in the FIFO will be lost. This bit has no effect\n"
            "  on the reception of frames by the EGRESS MAC. Use\n"
            "  EGRESS_DISABLE for that purpose.\tRW 0\n";
        break;
#endif
    case OFFSETOF(mcp_fio_t, mcpf_ump_config):
        str = "\n ump_config - 0x80038004\n"
            "  SHARE:\n"
            "bit 31-25 unused3\tRO 0\n"
            "bit 24 OLD_BCNT_RDY\n"
            "  When set '1', UMP logic will revert to the Xinan 90nm\n"
            "  method of calculating BCNT_RDY which does not\n"
            "  account for the fact that the writing of the byte count\n"
            "  consumes 1 word in the FIFO. This bit is available only\n"
            "  from 65nm A0 onwards.\tRW 0\n"
            "bit 23-16 unused2\tRO 0\n"
            "bit 15-8 INGRESS_WORD_ACCM\n"
            "  This field allows the UMP Ingress engine to wait for\n"
            "  up-to 256 words of data to accumulate before starting to\n"
            "  send the data out. A word is equal to 4bytes. The default\n"
            "  for this register is 0, but this is not a valid setting for cut\n"
            "  through mode. FW needs to program the appropriate\n"
            "  value > 0. A good value is 8 which allows for 32bytes to\n"
            "  be accumulated before sending the packet out. Safest\n"
            "  value is 16 which ensures that a min sized packet is in the\n"
            "  FIFO before the packet starts streaming out.\tRW 0\n"
            "bit 7 INGRESS_MODE\n"
            "  When set '1', the UMP Transmit engine will operate in a\n"
            "  cut through mode. It will wait for\n"
            "  INGRESS_WORD_ACCM bytes to accumulate before\n"
            "  starting to send the packet out. When not set, the engine\n"
            "  will function in a store and forward mode.\tRW 0\n"
            "bit 6 RVMII_MODE\n"
            "  When this bit is '1' and RMII_MODE is '0', the UMP\n"
            "  port will be put into \"PHY\" clocking/timing mode where\n"
            "  it will emulate the timing of a PHY MII connection. The\n"
            "  clock inputs become clock outputs and the data pins of\n"
            "  the MII are syncronous to those outputs. In this mode,\n"
            "  only single UMP can be connected to a BMC.\tRW 0\n"
            "bit 5 unused1\tRO 0\n"
            "bit 4 RMII_MODE\n"
            "  When this bit is '1', the UMP interface will run in RMII\n"
            "  mode rather than normal MII mode.\tRW 0\n"
            "bit 3-0 unused0\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_fc_trip):
        str = "\nump_fc_trip - 0x80038008\n"
            "  SHARE:\n"
            "bit 31-25 unused1\tRO 0\n"
            "bit 24-16 XOFF_TRIP\n"
            "  When the Egress FIFO has less than XOFF_TRIP*16 bytes\n"
            "  remaining, the \"pause needed\" input to the UMP MAC will be\n"
            "  asserted. This value is the low-water mark for flow control\n"
            "  generated by the UMP MAC. If this value is 0, then the \"pause\n"
            "  needed\" input to the UMP MAC will be controlled soley by the\n"
            "  SW_PAUSE bit.\tRW 0xc8\n"
            "bit 15-9 unused0\tRO 0\n"
            "bit 8-0 XON_TRIP\n"
            "  When the Egress FIFO has at least XON_TRIP*16 bytes available,\n"
            "  the \"pause needed\" input to the UMP MAC will be de-asserted.\n"
            "  This value is the high-water mark for flow control generated by the\n"
            "  UMP MAC. A value is the largest valid value that may be\n"
            "  programmed and will de-assert the \"pause needed\" signal only when\n"
            "  the Egress FIFO is empty.\tRW 0x124\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_frm_rd_status):
        str = "\n     ump_egress_frm_rd_status - 0x80038090\n"
            "  SHARE:\n"
            "bit 31-27 EGRESS_FIFO_STATE\n"
            "  Enumeration:\tRO 0\n"
            "  Value Name Description\n"
            "   0    IDLE IDLE\n"
            "   1    READY READY\n"
            "   2    BUSY BUSY\n"
            "   3    EXTRA_RD Extra Read\n"
            "   4    LATCH_IP_HDR Latch IP Header\n"
            "bit 26-14 unused0\tRO 0\n"
            "bit 13-3 BCNT Current Egress packet length in Byte count.\tRO 0\n"
            "bit 2 FIFO_EMPTY\n"
            "  When this bit is set, it indicates that Egress FIFO is empty\n"
            "  and firmware or software should not read rd_dat any further\n"
            "  until this bit clears.\tRO 0\n"
            "bit 1 FRM_IN_PRO\n"
            "  When this bit is set, it indicates that current packet reading is\n"
            "  in progress. Cleared when all packet data has been read for a\n"
            "  particular packet.\tRO 0\n"
            "bit 0 NEW_FRM\n"
            "  When this bit is set, it indicates that a new packet is ready in\n"
            "  Egress FIFO. Always 0 when FRM_IN_PRO bit is set.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_frm_rd_data):
        str = "\n ump_egress_frm_rd_data - 0x80038094 - (RO/Reset: 0)\n"
            "  SHARE: Next 32-bit data word of a packet. For a new packet and the first read, the right two bytes contain\n"
            "  the first two data bytes of the packet. This means that the 3rd byte of the packet is in bits 31:24 of the 2nd\n"
            "  data word read. The two extra bytes in the 1st read makes the L3 and above frame headers 32-bit aligned\n"
            "  with the frm_rd_data register. In other words the packet is shifted by 2bytes so the L3 headers line up on\n"
            "  32-bit boundaries for ease of processing.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_frm_wr_ctl):
        str = "\n ump_ingress_frm_wr_ctl - 0x80038098\n"
            "  SHARE:\n"
            "bit 31-30 INGRESS_FIFO_STATE\n"
            "  Enumeration:\tRO 0\n"
            "  Value Name Description\n"
            "   0  IDLE IDLE\n"
            "   1  WAIT WAIT\n"
            "   2  BUSY BUSY\n"
            "   3  EXTRA_WR Extra Write\n"
            "bit 29-14 unused0\tRO 0\n"                                         
            "bit 13-3 BCNT\n"
            "  This register must be written with the packet length before\n"
            "  the ump_frm_wr_dat register is written with any data. For\n"
            "  this value to be accepted, the SOF bit must be set on the\n"
            "  same registger write.\tRW 0\n"
            "bit 2 BCNT_RDY\n"
            "  When this bit is set, it indicates that the BCNT field can\n"
            "  accept another write as a request for space for another\n"
            "  transmit frame. This bit is logically or'ed with the\n"
            "  FIFO_RDY bit to create the value of the FTQ1_VALID\n"
            "  event bit. This bit must be '1' when the firmware writes a '1'\n"
            "  to the NEW_FRM bit. This bit will clear back to '0' each\n"
            "  time a new BCNT value is written and will only set back to\n"
            "  '1' once the NEW_FRM bit has cleared at the end of the\n"
            "  complete packet write.\tRO 0\n"
            "bit 1 FIFO_RDY\n"
            "  This bit is set when the INGRESS_FIFO can accept all the\n"
            "  data requested by a write tot he BCNT value. This value\n"
            "  will read as '1' when there is sufficient space in the Ingress\n"
            "  FIFO to accept all the bytes described by the latest BCNT\n"
            "  value. This bit will clear to '0' once all the data requested in\n"
            "  BCNT has been written to the ump_frm_wr_data register.\n"
            "  This bit is logically or'ed with the BCNT_RDY bit to create\n"
            "  the value of the FTQ1_VALID event bit.\tRO 0\n"
            "bit 0 NEW_FRM\n"
            "  This bit must be set for the BCNT field to be accepted as\n"
            "  the length of a new packet. This bit will clear to zero when\n"
            "  the complete packet has been written to the\n"
            "  ump_frm_wr_data register.\tSC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_frm_wr_data):
        str = "\nump_ingress_frm_wr_data - 0x8003809c - (WO/Reset: 0)\n"
            "  SHARE: Next 32-bit data word of a transmit packet. Only the right two bytes of the first data word of each\n"
            "  packet are valid. This means that the 3rd byte of the packet is in bits 31:24 of the 2nd data word write. This\n"
            "  offset of 2 bytes makes the L3 and above frame headers 32-bit aligned with the frm_wr_data register. This\n"
            "  register must only be written when FIFO_RDY bit is '1' in the ump_frm_wr_ctrl register. A read operation\n"
            "  on this register will always return 0s.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_frame_type):
        str = "\nump_egress_frame_type - 0x800380a0 - (RO/Reset: 0)\n"
            "  SHARE: This register provides the ethernet type field of the packet ;when a new packet is ready to be read\n"
            "  from the rd_data regsiter. Reading this register has no effect on the FIFO status.\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_fifo_remaining_words):
        str = "\n     ump_fifo_remaining_words - 0x800380a4\n"
            "  SHARE:\n"
            "bit 31 INGRESS_FIFO_OVERFLOW\n"
            "  This bit indicates that there was an overflow in the\n"
            "  UMP Ingress FIFO. This works only from Xinan\n"
            "bit 65nm A0 onwards.\tWC 0\n"
            "bit 30 INGRESS_FIFO_UNDERFLOW\n"
            "  This bit indicates that there was an underflow in\n"
            "  the UMP Ingress FIFO. This works only from\n"
            "  Xinan 65nm A0 onwards.\tWC 0\n"
            "bit 29-26 unused1\tRO 0\n"
            "bit 25-16 INGRESS_FIFO_DEPTH Remainig 32-bit words in Ingress UMP FIFO.\tRO 0\n"
            "bit 15 EGRESS_FIFO_OVERFLOW\n"
            "  This bit indicates that there was an overflow in the\n"
            "  UMP Egress FIFO. This works only from Xinan\n"
            "bit 65nm A0 onwards.\tWC 0\n"
            "bit 14 EGRESS_FIFO_UNDERFLOW\n"                                    
            "  This bit indicates that there was an underflow in\n"
            "  the UMP Egress FIFO. This works only from Xinan\n"
            "bit 65nm A0 onwards.\tWC 0\n"
            "bit 13-11 unused0\tRO 0\n"
            "bit 10-0 EGRESS_FIFO_DEPTH Remainig 32-bit words in Egress UMP FIFO.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_fifo_ptrs):
        str = "\nump_egress_fifo_ptrs - 0x800380a8\n"
            "  SHARE:\n"
            "bit 31 UPDATE_WRPTR\n"
            "  Update the write pointer of Ingress FIFO with\n"
            "  egress_write_pointer.\tSC 0\n"
            "bit 30-28 unused1\tRO 0\n"
            "bit 27-16 EGRESS_FIFO_WR_PTR\n"
            "  Write pointer of the Egress FIFO in 32-bit words. This\n"
            "  field is writable for debug purpose and only when there is\n"
            "  atleast one packet in the FIFO.\tRW 0\n"
            "bit 15 UPDATE_RDPTR\n"
            "  Update the read pointer of Egress FIFO with\n"
            "  egress_read_pointer.\tSC 0\n"
            "bit 14-12 unused0\tRO 0\n"
            "bit 11-0 EGRESS_FIFO_RD_PTR\n"
            "  Read pointer of the Egress FIFO in 32-bit words. This\n"
            "  field is writable for debug purpose and only when there is\n"
            "  atleast one packet in the FIFO.\tRW 0\n";
        break;                         
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_fifo_ptrs):
        str = "\nump_ingress_fifo_ptrs - 0x800380ac\n"
            "  SHARE:\n"
            "bit 31 UPDATE_WRPTR\n"
            "  Update the write pointer of Ingress FIFO with\n"
            "  ingress_write_pointer.\tSC 0\n"
            "bit 30-27 unused1\tRO 0\n"                                         
            "bit 26-16 INGRESS_FIFO_WR_PTR\n"
            "  Write pointer of the Ingress FIFO in 32-bit words. This\n"
            "  field is writable for debug purpose and only when there\n"
            "  is atleast one packet in the FIFO.\tRW 0\n"
            "bit 15 UPDATE_RDPTR\n"
            "  Update the read pointer of Ingress FIFO with\n"
            "  ingress_read_pointer.\tSC 0\n"
            "bit 14-11 unused0\tRO 0\n"
            "bit 10-0 INGRESS_FIFO_RD_PTR\n"
            "  Read pointer of the Ingress FIFO in 32-bit words. This\n"
            "  field is writable for debug purpose and only when there\n"
            "  is atleast one packet in the FIFO.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_packet_sa_0):
        str = "\n ump_egress_packet_sa_0 - 0x800380b4\n"
            "  SHARE:\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-0 EGRESS_SA\n"
            "  Upper 16 bits of the Source Address of a packet received on the\n"
            "  UMP port. The SA will be used by the firmware to decide which of\n"
            "  the two gigabit ports will be used to transmit the packet.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_packet_sa_1):
        str = "\n ump_egress_packet_sa_1 - 0x800380b8\n"
            "  SHARE:\n"
            "bit 31-0 EGRESS_SA\n"
            "  Lower 32 bits of the Source Address of a packet received on the\n"
            "  UMP port. The SA will be used by the firmware to decide which of\n"
            "  the two gigabit ports will be used to transmit the packet.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_burst_command):
        str = "\n    ump_ingress_burst_command - 0x800380bc\n"
            "  SHARE:\n"
            "bit 31-30 unused1\tRO 0\n"
            "bit 29-16 RBUF_OFFSET\n"
            "  This field programs offset into the RBUF cluster from\n"
            "  where the DMA engine needs to copy data. The offset\n"
            "  needs to be 32-bit aligned. The FW is required to program\n"
            "  the rbuf_cluster field before doing this operation.\tRW 0\n"
            "bit 15 unused0\tRO 0\n"
            "bit 14-13 INGRESS_PORT_EXT\n"
            "  This bit selects the 2 MSb's of which main MAC port the\n"
            "  packet was received.\tRW 0\n"                                    
            "bit 12-2 DMA_LENGTH\n"
            "  This field programs the length of the data that needs to be\n"
            "  transferred from the RBUF. The length is represented in\n"
            "  byte units.\tRW 0\n"
            "bit 1 INGRESS_PORT\n"
            "  This bit selects the LSb of which main MAC port the\n"
            "  packet was received.\tRW 0\n"                                    
            "bit 0 INGRESS_DMA_START\n"
            "  Setting this bit to '1' starts the DMA engine that reads data\n"
            "  from the RBUF and writes to UMP Ingress MAC. This is a\n"
            "  self clearing bit that clears after the DMA operation is\n"      
            "  completed.\tSC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_rbuf_cluster):
        str = "\n     ump_ingress_rbuf_cluster - 0x800380c0\n"
            "  SHARE:\n"
            "bit 31-25 unused0\tRO 0\n"
            "bit 24-0 RBUF_CLUSTER This field programsthe RBUF cluster.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_vlan):
        str = "\nump_ingress_vlan - 0x800380c4\n"
            "  SHARE:\n"
            "bit 31-18 unused0\tRO 0\n"
            "bit 17 VLAN_DEL\n"
            "  Setting this field to '1' causes VLAN Tag to be removed\n"
            "  from the ingress packet.\tRW 0\n"
            "bit 16 VLAN_INS\n"
            "  Setting this field to '1' causes VLAN Tag to be inserted in\n"
            "  the ingress packet.\tRW 0\n"
            "bit 15-0 INGRESS_VLAN_TAG\n"
            "  This field is used to insert VLAN Tag into a ingress packet\n"
            "  if vlan_ins is set.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_burst_status):
        str = "\n     ump_ingress_burst_status - 0x800380c8\n"
            "  SHARE:\n"
            "bit 31-2 unused0\tRO 0\n"                                          
            "bit 1-0 RESULT\n"
            "  This bit indicates the status of the current burst command. Bit gets\n"
            "  cleared when this register is read or INGRESS_BURST_START is\n"
            "  set to '1'\tAC 0\n"
            "  Value Name Description\n"
            "   0    BUSY DMA is in progress\n"
            "   1    DONE DMA transfer completed successfully\n"
            "   2    ERR  DMA transfer interrupted due to\n"
            "             MCP_ISOLATE\n"
            "   3    ERR1 DMA transfer interrupted due to the FIFO\n"
            "             under-run\n"
            "\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_burst_command):
        str = "\n     ump_egress_burst_command - 0x800380cc\n"
            "  SHARE:\n"
            "bit 31-29 unused1\tRO 0\n"
            "bit 28-16 TPBUF_OFFSET\n"
            "  This field programs offset into the TPBUF memory, this is\n"
            "  the memory to which the DMA engine will move the Egress\n"       
            "  Data.TPBUF in Xinan is organized as a 16-byte wide\n"
            "  memory and FW should ensure that the offset is 16-byte\n"
            "  aligned. The offset field in this register as it is points to\n"
            "  [TPBUF_ADDR_WIDTH-1:2] of the tpbuf address, in\n"
            "  other words the offset is defined in 4-byte increments. To\n"
            "  make sure that the offset is 16-byte aligned, the lower two\n"
            "bits (17,16) need to be set to 0.\tRW 0\n"
            "bit 15-14 unused0\tRO 0\n"
            "bit 13 EGRESS_PORT_EXT\n"
            "  This bit selects the MSb of which main MAC port the\n"
            "  packet is sent out.\tRW 0\n"
            "bit 12-2 DMA_LENGTH\n"
            "  This field programs the length of the data that needs to be\n"
            "  transferred to the TPBUF. The length is represented in byte\n"
            "  units.\tRW 0\n"
            "bit 1 EGRESS_PORT\n"
            "  This bit selects the LSb of which main MAC port the\n"
            "  packet is sent out.\tRW 0\n"
            "bit 0 EGRESS_DMA_START\n"                                          
            "  Setting this bit to '1' starts the DMA engine that reads data\n"
            "  from UMP Egress MAC and writes to TPBUF. This is a self\n"
            "  clearing bit that clears after the DMA operation is\n"
            "  completed.\tSC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_vlan):
        str = "\nump_egress_vlan - 0x800380d0\n"
            "  SHARE:\n"
            "bit 31-18 unused0\tRO 0\n"                                         
            "bit 17 VLAN_DEL\n"
            "  Setting this field to '1' causes VLAN Tag to be removed from\n"
            "  the egress packet.\tRW 0\n"
            "bit 16 VLAN_INS\n"
            "  Setting this field to '1' causes VLAN Tag to be inserted in the\n"
            "  egress packet.\tRW 0\n"
            "bit 15-0 EGRESS_VLAN_TAG\n"                                        
            "  This field is used to insert VLAN Tag into a egress packet if\n"
            "  vlan_ins is set.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_burst_status):
        str = "\nump_egress_burst_status - 0x800380d4\n"
            "  SHARE:\n"
            "bit 31-2 unused0\tRO 0\n"
            "bit 1-0 RESULT\n"
            "  This bit indicates the status of the current burst command. Bit gets\n"
            "  cleared when this register is read or EGRESS_BURST_START is\n"
            "  set to '1'\tAC 0\n"
            "  Value Name Description\n"
            "   0    BUSY DMA is in progress\n"
            "   1    DONE DMA transfer completed successfully\n"
            "   2    ERR0 DMA transfer interrupted due to MCP_ISOLATE\n"
            "   3    RSVD Reserved\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_statistic):
        str = "\nump_egress_statistic - 0x800380d8\n"
            "  SHARE:\n"
            "bit 31-24 EGRESS_DROP_CNT\n"
            "  This field keeps the count of how many packets were\n"
            "  dropped by the UMP MAC for lack of FIFO space.\tRO 0\n"
            "bit 23-16 EGRESS_ERROR_CNT\n"
            "  This field keeps the count of how many error packets were\n"
            "  received by the UMP MAC.\tRO 0\n"
            "bit 15-0 EGRESS_GOOD_CNT\n"                                        
            "  This field keeps the count of how many good packets were\n"
            "  received by the UMP MAC.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_statistic):
        str = "\nump_ingress_statistic - 0x800380dc\n"
            "  SHARE:\n"                                                        
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-0 INGRESS_PKT_CNT\n"
            "  This field keeps the count of how many packets were\n"
            "  transmitted by the UMP MAC.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_arb_cmd):
        str = "\nump_arb_cmd - 0x800380e0\n"
            "  SHARE:\n"
            "bit 31-16 UMP_ARB_TIMEOUT\n"
            "  This field indicates the value in number of Ingress\n"
            "  clock cycles that the arbitration master will wait\n"
            "  before re-starting the arbitration process.\tRW 0x400\n"
            "bit 15 UMP_ARB_FC_DISABLE\n"
            "  Setting this bit disables the feature to send\n"      
            "  XOFF/XON ordered sets on the arbiter interface. This\n"
            "  feature is enabled by default and allows a UMP port\n"
            "  that wants to send a XOFF/XON frame to use the\n"
            "  sideband interface to accelerate sending the command\n"
            "  on the MII bus. This helps reducing the size of the Rx\n"
            "  FIFO needed when multiple UMP ports are connected\n"
            "  to the BMC.\tRW 0\n"
            "bit 14 unused1\tRO 0\n"
            "bit 13 UMP_ARB_TOKEN_VALID This bit indicates if the arbiter has a valid token.\tRO 0\n"
            "bit 12-8 UMP_ARB_TOKEN_IPG\n"
            "  This field is a programmable inter-packet gap for\n"
            "  when the token is sent out.\tRW 0x14\n"
            "bit 7 UMP_ARB_AUTOBYPASS\n"
            "  Setting this field to '1' causes the UMP port to cut\n"
            "  latency when forwarding a token.\tRW 0\n"
            "bit 6 UMP_ARB_BYPASS\n"
            "  Setting this field to '1' the HW arbitration logic to\n"
            "  function in bypass mode. This allows UMP ports that\n"
            "  don't have the firmware running to be automatically\n"
            "  bypassed. Firmware should also set this bit when there\n"
            "  is nothing to send. This will reduce the latency of the\n"
            "  token around the loop.\tRW 0x1\n"
            "bit 5 UMP_ARB_START\n"
            "  Setting this field to '1' causes the hardware arbitration\n"
            "  scheme to begin. Any UMP port can re-start the\n"
            "  arbitration.\tSC 0\n"
            "bit 4 UMP_ARB_DISABLE\n"
            "  Setting this field to '1' causes the hardware arbitration\n"
            "  scheme to be disabled. This bit should be set when\n"
            "  there is only one UMP port on the board and tokens\n"
            "  need not be passed out.\tRW 0\n"
            "bit 3 unused0\tRO 0\n"
            "bit 2-0 UMP_ID\n"
            "  This field is used to set the ID of the current UMP\n"
            "  port. UMP port with the lowest ID value is the master\n"
            "  of the HW based arbitration.\tRW 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_frame_count):
        str = "\nump_frame_count - 0x800380ec\n"
            "  SHARE:\n"
            "bit 31-21 unused1\tRO 0\n"
            "bit 20-16 INRESS_FRAME_COUNT\n"
            "  This field keeps the count of how many packets are\n"
            "  present in the INGRESS FIFO at any given time. This is\n"
            "  provided for debug purposes and only available in\n"
            "  Xinan 65nm onwards.\tRO 0\n"
            "bit 15-7 unused0\tRO 0\n"
            "bit 6-0 EGRESS_FRAME_COUNT\n"
            "  This field keeps the count of how many packets are\n"
            "  present in the EGRESS FIFO at any given time. This is\n"
            "  provided for debug purposes and only available in\n"
            "  Xinan 65nm onwards.\tRO 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_statistic_ac):
        str = "\nump_egress_statistic_ac - 0x800380f0\n"
            "  SHARE:\n"
            "bit 31-24 EGRESS_DROP_CNT\n"
            "  This field keeps the count of how many packets were\n"
            "  dropped by the UMP MAC for lack of FIFO space.\tAC 0\n"
            "bit 23-16 EGRESS_ERROR_CNT\n"
            "  This field keeps the count of how many error packets were\n"     
            "  received by the UMP MAC.\tAC 0\n"
            "bit 15-0 EGRESS_GOOD_CNT\n"
            "  This field keeps the count of how many good packets were\n"
            "  received by the UMP MAC.\tAC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_statistic_ac):
        str = "\n     ump_ingress_statistic_ac - 0x800380f4\n"
            "  SHARE:\n"
            "bit 31-16 unused0\tRO 0\n"
            "bit 15-0 INGRESS_PKT_CNT\n"
            "  This field keeps the count of how many packets were\n"
            "  transmitted by the UMP MAC.\tAC 0\n";
        break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_event):
        str = "\n  ump_event - 0x800380f8\n"  
            "  SHARE:\n"
            "bit 31-21 unused1\tRO 0\n"
            "bit 20 EGRESS_FRAME_DROP_EVENT_EN Event Enable bit.\tRW 0\n"
            "bit 19 EGRESSBURST_DONE_EVENT_EN Event Enable bit.\tRW 0\n"
            "bit 18 INGRESSBURST_DONE_EVENT_EN Event Enable bit.\tRW 0\n"
            "bit 17 EGRESS_RDY_EVENT_EN Event Enable bit.\tRW 0\n"
            "bit 16 INGRESS_RDY_EVENT_EN Event Enable bit.\tRW 0\n"
            "bit 15-5 unused0\tRO 0\n"
            "bit 4 EGRESS_FRAME_DROP_EVENT\tWC 0\n"
            "bit 3 EGRESSBURST_DONE_EVENT\tWC 0\n"
            "bit 2 INGRESSBURST_DONE_EVENT\tWC 0\n"
            "bit 1 EGRESS_RDY_EVENT\tRO 0\n"  
            "bit 0 INGRESS_RDY_EVENT\tRO 0\n";
        break;
    default: 
        str = NULL;
        break;
    }
    return str;
}// GetMcpFioHelp

/*
 *------------------------------------------------------------------------
 * DecodeMcpFioRegDetail --
 *
 *     Decode MCP GRC register with details (with option help info)
 *
 * Results:
 *     A standard Tcl result
 *
 * Side effects:
 *     None.
 *------------------------------------------------------------------------
 */
void DecodeMcpFioRegDetail(Tcl_Obj* rc_ptr, u32_t addr, u32_t data, u32_t help)
{
    static char  disp_str[1000];
    char const * help_str = GetMcpFioHelp(addr);

    if (help_str)
    {
        if (help)
        {
            Tcl_AppendStringsToObj( rc_ptr, help_str, NULL);
        }
        else
        {
            // just display the register name
            sscanf(help_str,"%s- (Offset:", disp_str);
            Tcl_AppendStringsToObj( rc_ptr, "\n",disp_str, NULL);
        }
    }
    else
    {
        // nothing to do
        return;
    }
                                                                   
    switch (addr & 0xFFFFF)
    {
    case OFFSETOF(mcp_fio_t, mcpf_events_bits):
        sprintf(disp_str, 
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tFTQ0_VALID      :", (data & MCPF_EVENTS_BITS_FTQ0_VALID       ) ? " ON":"OFF", 
            "\n\tFTQ1_VALID      :", (data & MCPF_EVENTS_BITS_FTQ1_VALID       ) ? " ON":"OFF", 
            "\n\tUMP_EVENT       :", (data & MCPF_EVENTS_BITS_UMP_EVENT        ) ? " ON":"OFF", 
            "\n\tSMBUS_EVENT     :", (data & MCPF_EVENTS_BITS_SMBUS_EVENT      ) ? " ON":"OFF", 
            "\n\tFLASH_EVENT     :", (data & MCPF_EVENTS_BITS_FLASH_EVENT      ) ? " ON":"OFF", 
            "\n\tMCP_DOORBELL    :", (data & MCPF_EVENTS_BITS_MCP_DOORBELL     ) ? " ON":"OFF", 
            "\n\tHFTQ0_VALID     :", (data & MCPF_EVENTS_BITS_HFTQ0_VALID      ) ? " ON":"OFF", 
            "\n\tHFTQ1_VALID     :", (data & MCPF_EVENTS_BITS_HFTQ1_VALID      ) ? " ON":"OFF", 
            "\n\tEXP_ROM         :", (data & MCPF_EVENTS_BITS_EXP_ROM          ) ? " ON":"OFF", 
            "\n\tVPD             :", (data & MCPF_EVENTS_BITS_VPD              ) ? " ON":"OFF", 
            "\n\tFLASH           :", (data & MCPF_EVENTS_BITS_FLASH            ) ? " ON":"OFF", 
            "\n\tSMB0            :", (data & MCPF_EVENTS_BITS_SMB0             ) ? " ON":"OFF", 
            "\n\tNIG             :", (data & MCPF_EVENTS_BITS_NIG              ) ? " ON":"OFF", 
            "\n\tRESERVED0       :", (data & MCPF_EVENTS_BITS_RESERVED0        ) ? " ON":"OFF", 
            "\n\tRESERVED1       :", (data & MCPF_EVENTS_BITS_RESERVED1        ) ? " ON":"OFF", 
            "\n\tGPIO0           :", (data & MCPF_EVENTS_BITS_GPIO0            ) ? " ON":"OFF", 
            "\n\tGPIO1           :", (data & MCPF_EVENTS_BITS_GPIO1            ) ? " ON":"OFF", 
            "\n\tGPIO2           :", (data & MCPF_EVENTS_BITS_GPIO2            ) ? " ON":"OFF", 
            "\n\tGPIO3           :", (data & MCPF_EVENTS_BITS_GPIO3            ) ? " ON":"OFF", 
            "\n\tSW_TMR_1        :", (data & MCPF_EVENTS_BITS_SW_TMR_1         ) ? " ON":"OFF", 
            "\n\tSW_TMR_2        :", (data & MCPF_EVENTS_BITS_SW_TMR_2         ) ? " ON":"OFF", 
            "\n\tSW_TMR_3        :", (data & MCPF_EVENTS_BITS_SW_TMR_3         ) ? " ON":"OFF", 
            "\n\tSW_TMR_4        :", (data & MCPF_EVENTS_BITS_SW_TMR_4         ) ? " ON":"OFF", 
            "\n\tMSI             :", (data & MCPF_EVENTS_BITS_MSI              ) ? " ON":"OFF", 
            "\n\tRESERVED2       :", (data & MCPF_EVENTS_BITS_RESERVED2        ) ? " ON":"OFF", 
            "\n\tRESERVED3       :", (data & MCPF_EVENTS_BITS_RESERVED3        ) ? " ON":"OFF", 
            "\n\tRESERVED4       :", (data & MCPF_EVENTS_BITS_RESERVED4        ) ? " ON":"OFF", 
            "\n\tMAIN_PWR_INT    :", (data & MCPF_EVENTS_BITS_MAIN_PWR_INT     ) ? " ON":"OFF", 
            "\n\tNOT_ENABLED     :", (data & MCPF_EVENTS_BITS_NOT_ENABLED      ) ? " ON":"OFF", 
            "\n\tATTENTIONS_VALID:", (data & MCPF_EVENTS_BITS_ATTENTIONS_VALID ) ? " ON":"OFF");   
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_attentions_bits):
    case OFFSETOF(mcp_fio_t, mcpf_attention_enable):
        sprintf(disp_str, 
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tGRC_TIMEOUT    :", (data & MCPF_ATTENTIONS_BITS_GRC_TIMEOUT     ) ? " ON":"OFF", 
            "\n\tPERST_ASSERTION:", (data & MCPF_ATTENTIONS_BITS_PERST_ASSERTION ) ? " ON":"OFF", 
            "\n\tSPAD_PARITY_ERR:", (data & MCPF_ATTENTIONS_BITS_SPAD_PARITY_ERR ) ? " ON":"OFF", 
            "\n\tSPIO5          :", (data & MCPF_ATTENTIONS_BITS_SPIO5           ) ? " ON":"OFF", 
            "\n\tRSV_ACCESS     :", (data & MCPF_ATTENTIONS_BITS_RSV_ACCESS      ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_event_enable):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %06x ",
            addr, data,
            "\n\tSHARE: event_enable for the MCPQ from function 0 :", (data & (1<<0)) ? " ON":"OFF",
            "\n\tSHARE: event_enable for the MCPQ from function 1 :", (data & (1<<1)) ? " ON":"OFF",
            "\n\tSPLIT: event from the UMP block                  :", (data & (1<<2)) ? " ON":"OFF",
            "\n\tSPLIT: event from the SMBus block                :", (data & (1<<3)) ? " ON":"OFF",
            "\n\tSPLIT: event from the Flash block                :", (data & (1<<4)) ? " ON":"OFF",
            "\n\tSPLIT: MCP doorbell event on a per-prt basis     :", (data & (1<<5)) ? " ON":"OFF",
            "\n\tSPLIT: the remaing events for both functions independently  :", (u32_t)(data >> 8));
            break;
    case OFFSETOF(mcp_fio_t, mcpf_fio_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s ",
            addr, data,
            "\n\tENABLED  :", (data & MCPF_FIO_STATUS_ENABLED  ) ? " ON":"OFF", 
            "\n\tFORCE_ENA:", (data & MCPF_FIO_STATUS_FORCE_ENA) ? " ON":"OFF");        
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_interrupt_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tEVENT0_E0:", (data & MCPF_INTERRUPT_STATUS_EVENT0_E0 ) ? " ON":"OFF", 
            "\n\tATTN0_E0 :", (data & MCPF_INTERRUPT_STATUS_ATTN0_E0  ) ? " ON":"OFF", 
            "\n\tEVENT1_E0:", (data & MCPF_INTERRUPT_STATUS_EVENT1_E0 ) ? " ON":"OFF", 
            "\n\tATTN1_E0 :", (data & MCPF_INTERRUPT_STATUS_ATTN1_E0  ) ? " ON":"OFF", 
            "\n\tEVENT0_E1:", (data & MCPF_INTERRUPT_STATUS_EVENT0_E1 ) ? " ON":"OFF", 
            "\n\tATTN0_E1 :", (data & MCPF_INTERRUPT_STATUS_ATTN0_E1  ) ? " ON":"OFF", 
            "\n\tEVENT1_E1:", (data & MCPF_INTERRUPT_STATUS_EVENT1_E1 ) ? " ON":"OFF", 
            "\n\tATTN1_E1 :", (data & MCPF_INTERRUPT_STATUS_ATTN1_E1  ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_free_counter_value):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u",
            addr, data,
            "\n\tfree_counter_value :", data);
            break;
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_bits_status1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tBRCST     :", (data & MCPF_MCPQ_BITS_STATUS1_BRCST     ) ? " ON":"OFF",
            "\n\tMLCST     :", (data & MCPF_MCPQ_BITS_STATUS1_MLCST     ) ? " ON":"OFF",
            "\n\tUNCST     :", (data & MCPF_MCPQ_BITS_STATUS1_UNCST     ) ? " ON":"OFF",
            "\n\tMAC0      :", (data & MCPF_MCPQ_BITS_STATUS1_MAC0      ) ? " ON":"OFF",
            "\n\tMAC1      :", (data & MCPF_MCPQ_BITS_STATUS1_MAC1      ) ? " ON":"OFF",
            "\n\tMAC2      :", (data & MCPF_MCPQ_BITS_STATUS1_MAC2      ) ? " ON":"OFF",
            "\n\tARP       :", (data & MCPF_MCPQ_BITS_STATUS1_ARP       ) ? " ON":"OFF",
            "\n\tIP0       :", (data & MCPF_MCPQ_BITS_STATUS1_IP0       ) ? " ON":"OFF",
            "\n\tIP1       :", (data & MCPF_MCPQ_BITS_STATUS1_IP1       ) ? " ON":"OFF",
            "\n\tIP2       :", (data & MCPF_MCPQ_BITS_STATUS1_IP2       ) ? " ON":"OFF",
            "\n\tNTBS_U_SRC:", (data & MCPF_MCPQ_BITS_STATUS1_NTBS_U_SRC) ? " ON":"OFF",
            "\n\tNTBS_T_SRC:", (data & MCPF_MCPQ_BITS_STATUS1_NTBS_T_SRC) ? " ON":"OFF",
            "\n\tRMCP      :", (data & MCPF_MCPQ_BITS_STATUS1_RMCP      ) ? " ON":"OFF",
            "\n\tDHCP      :", (data & MCPF_MCPQ_BITS_STATUS1_DHCP      ) ? " ON":"OFF",
            "\n\tNTBS_U_DST:", (data & MCPF_MCPQ_BITS_STATUS1_NTBS_U_DST) ? " ON":"OFF",
            "\n\tUDP0      :", (data & MCPF_MCPQ_BITS_STATUS1_UDP0      ) ? " ON":"OFF",
            "\n\tUDP1      :", (data & MCPF_MCPQ_BITS_STATUS1_UDP1      ) ? " ON":"OFF",
            "\n\tUDP2      :", (data & MCPF_MCPQ_BITS_STATUS1_UDP2      ) ? " ON":"OFF",
            "\n\tNTBS_T_DST:", (data & MCPF_MCPQ_BITS_STATUS1_NTBS_T_DST) ? " ON":"OFF",
            "\n\tTCP0      :", (data & MCPF_MCPQ_BITS_STATUS1_TCP0      ) ? " ON":"OFF",
            "\n\tTCP1      :", (data & MCPF_MCPQ_BITS_STATUS1_TCP1      ) ? " ON":"OFF",
            "\n\tTCP2      :", (data & MCPF_MCPQ_BITS_STATUS1_TCP2      ) ? " ON":"OFF",
            "\n\tVLAN_ID0  :", (data & MCPF_MCPQ_BITS_STATUS1_VLAN_ID0  ) ? " ON":"OFF",
            "\n\tVLAN_ID1  :", (data & MCPF_MCPQ_BITS_STATUS1_VLAN_ID1  ) ? " ON":"OFF",
            "\n\tVLAN_ID2  :", (data & MCPF_MCPQ_BITS_STATUS1_VLAN_ID2  ) ? " ON":"OFF",
            "\n\tVLAN      :", (data & MCPF_MCPQ_BITS_STATUS1_VLAN      ) ? " ON":"OFF",
            "\n\tNO_VLAN   :", (data & MCPF_MCPQ_BITS_STATUS1_NO_VLAN   ) ? " ON":"OFF",
            "\n\tL2_CRC    :", (data & MCPF_MCPQ_BITS_STATUS1_L2_CRC    ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t,  mcpf_mcpq_pkt_len):   //mcpf_mcpq_bits_status2):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %x %s %s %s %s %s %s %s %s %s %s %s %x",
            addr, data,
            "\n\tMCPQ_PKT_LEN  :", (u32_t)(data >> 16) & MCPF_MCPQ_PKT_LEN_MASK,    // 16 msb                 
            "\n\tCLUSTER_MASK    :", (u32_t)(data & MCPF_MCPQ_BITS_STATUS2_CLUSTER_MASK) ,
            "\n\tMF_OUTER_VLAN   :", (data & MCPF_MCPQ_BITS_STATUS2_MF_OUTER_VLAN    ) ? " ON":"OFF", 
            "\n\tMF_NO_OUTER_VLAN:", (data & MCPF_MCPQ_BITS_STATUS2_MF_NO_OUTER_VLAN ) ? " ON":"OFF", 
            "\n\tMF_OUTER_VLAN_ID:", (data & MCPF_MCPQ_BITS_STATUS2_MF_OUTER_VLAN_ID ) ? " ON":"OFF", 
            "\n\tMF_MAC3         :", (data & MCPF_MCPQ_BITS_STATUS2_MF_MAC3          ) ? " ON":"OFF", 
            "\n\tMF_IPV6_MLCST   :", (data & MCPF_MCPQ_BITS_STATUS2_MF_IPV6_MLCST    ) ? " ON":"OFF", 
            "\n\tOUTER_VLAN_MASK :", (u32_t)(data & MCPF_MCPQ_BITS_STATUS2_OUTER_VLAN_MASK ) >>7);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_bits_status3):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s",
            addr, data,
            "\n\tETYPE0  :", (data & MCPF_MCPQ_BITS_STATUS3_ETYPE0    ) ? " ON":"OFF", 
            "\n\tETYPE1  :", (data & MCPF_MCPQ_BITS_STATUS3_ETYPE1    ) ? " ON":"OFF", 
            "\n\tALL_MLCS:", (data & MCPF_MCPQ_BITS_STATUS3_ALL_MLCST ) ? " ON":"OFF", 
            "\n\tARP     :", (data & MCPF_MCPQ_BITS_STATUS3_ARP       ) ? " ON":"OFF", 
            "\n\tICMPV4  :", (data & MCPF_MCPQ_BITS_STATUS3_ICMPV4    ) ? " ON":"OFF", 
            "\n\tICMPV6  :", (data & MCPF_MCPQ_BITS_STATUS3_ICMPV6    ) ? " ON":"OFF", 
            "\n\tLLDP    :", (data & MCPF_MCPQ_BITS_STATUS3_LLDP      ) ? " ON":"OFF", 
            "\n\tVNTAG0  :", (data & MCPF_MCPQ_BITS_STATUS3_VNTAG0    ) ? " ON":"OFF", 
            "\n\tVNTAG1  :", (data & MCPF_MCPQ_BITS_STATUS3_VNTAG1    ) ? " ON":"OFF", 
            "\n\tPF0_VLAN:", (data & MCPF_MCPQ_BITS_STATUS3_PF0_VLAN  ) ? " ON":"OFF", 
            "\n\tPF1_VLAN:", (data & MCPF_MCPQ_BITS_STATUS3_PF1_VLAN  ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_mcpq_cmd):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tMCPQ_CMD_POP:", (data & MCPF_MCPQ_CMD_MCPQ_CMD_POP ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_bits_status1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tBRCST     :", (data & MCPF_HMCPQ_BITS_STATUS1_BRCST     ) ? " ON":"OFF",
            "\n\tMLCST     :", (data & MCPF_HMCPQ_BITS_STATUS1_MLCST     ) ? " ON":"OFF",
            "\n\tUNCST     :", (data & MCPF_HMCPQ_BITS_STATUS1_UNCST     ) ? " ON":"OFF",
            "\n\tMAC0      :", (data & MCPF_HMCPQ_BITS_STATUS1_MAC0      ) ? " ON":"OFF",
            "\n\tMAC1      :", (data & MCPF_HMCPQ_BITS_STATUS1_MAC1      ) ? " ON":"OFF",
            "\n\tMAC2      :", (data & MCPF_HMCPQ_BITS_STATUS1_MAC2      ) ? " ON":"OFF",
            "\n\tARP       :", (data & MCPF_HMCPQ_BITS_STATUS1_ARP       ) ? " ON":"OFF",
            "\n\tIP0       :", (data & MCPF_HMCPQ_BITS_STATUS1_IP0       ) ? " ON":"OFF",
            "\n\tIP1       :", (data & MCPF_HMCPQ_BITS_STATUS1_IP1       ) ? " ON":"OFF",
            "\n\tIP2       :", (data & MCPF_HMCPQ_BITS_STATUS1_IP2       ) ? " ON":"OFF",
            "\n\tNTBS_U_SRC:", (data & MCPF_HMCPQ_BITS_STATUS1_NTBS_U_SRC) ? " ON":"OFF",
            "\n\tNTBS_T_SRC:", (data & MCPF_HMCPQ_BITS_STATUS1_NTBS_T_SRC) ? " ON":"OFF",
            "\n\tRMCP      :", (data & MCPF_HMCPQ_BITS_STATUS1_RMCP      ) ? " ON":"OFF",
            "\n\tDHCP      :", (data & MCPF_HMCPQ_BITS_STATUS1_DHCP      ) ? " ON":"OFF",
            "\n\tNTBS_U_DST:", (data & MCPF_HMCPQ_BITS_STATUS1_NTBS_U_DST) ? " ON":"OFF",
            "\n\tUDP0      :", (data & MCPF_HMCPQ_BITS_STATUS1_UDP0      ) ? " ON":"OFF",
            "\n\tUDP1      :", (data & MCPF_HMCPQ_BITS_STATUS1_UDP1      ) ? " ON":"OFF",
            "\n\tUDP2      :", (data & MCPF_HMCPQ_BITS_STATUS1_UDP2      ) ? " ON":"OFF",
            "\n\tNTBS_T_DST:", (data & MCPF_HMCPQ_BITS_STATUS1_NTBS_T_DST) ? " ON":"OFF",
            "\n\tTCP0      :", (data & MCPF_HMCPQ_BITS_STATUS1_TCP0      ) ? " ON":"OFF",
            "\n\tTCP1      :", (data & MCPF_HMCPQ_BITS_STATUS1_TCP1      ) ? " ON":"OFF",
            "\n\tTCP2      :", (data & MCPF_HMCPQ_BITS_STATUS1_TCP2      ) ? " ON":"OFF",
            "\n\tVLAN_ID0  :", (data & MCPF_HMCPQ_BITS_STATUS1_VLAN_ID0  ) ? " ON":"OFF",
            "\n\tVLAN_ID1  :", (data & MCPF_HMCPQ_BITS_STATUS1_VLAN_ID1  ) ? " ON":"OFF",
            "\n\tVLAN_ID2  :", (data & MCPF_HMCPQ_BITS_STATUS1_VLAN_ID2  ) ? " ON":"OFF",
            "\n\tVLAN      :", (data & MCPF_HMCPQ_BITS_STATUS1_VLAN      ) ? " ON":"OFF",
            "\n\tNO_VLAN   :", (data & MCPF_HMCPQ_BITS_STATUS1_NO_VLAN   ) ? " ON":"OFF",
            "\n\tL2_CRC    :", (data & MCPF_HMCPQ_BITS_STATUS1_L2_CRC    ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_pkt_len):  //    u16_t mcpf_hmcpq_vlan_tag
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %x ",
            addr, data,
            "\n\tHMCPQ_PKT_LEN  :", (data >> 16) & MCPF_HMCPQ_PKT_LEN_MASK,    // 16 msb                 
            "\n\tHMCPQ_VLAN_TAG :", data & 0xFFFF);         // 16 lsb
            break;
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_bits_status2):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %07x %s %s %s %s",
            addr, data,
            "\n\tCLUSTER_MASK    :", (data & MCPF_HMCPQ_BITS_STATUS2_CLUSTER_MASK      ) ? " ON":"OFF",
            "\n\tMF_OUTER_VLAN   :", (data & MCPF_HMCPQ_BITS_STATUS2_MF_OUTER_VLAN     ) ? " ON":"OFF", 
            "\n\tMF_NO_OUTER_VLAN:", (data & MCPF_HMCPQ_BITS_STATUS2_MF_NO_OUTER_VLAN  ) ? " ON":"OFF", 
            "\n\tMF_OUTER_VLAN_ID:", (data & MCPF_HMCPQ_BITS_STATUS2_MF_OUTER_VLAN_ID  ) ? " ON":"OFF", 
            "\n\tMF_MAC3         :", (data & MCPF_HMCPQ_BITS_STATUS2_MF_MAC3           ) ? " ON":"OFF", 
            "\n\tMF_IPV6_MLCST   :", (data & MCPF_HMCPQ_BITS_STATUS2_MF_IPV6_MLCST     ) ? " ON":"OFF", 
            "\n\tOUTER_VLAN_MASK :", (u32_t)(data & MCPF_HMCPQ_BITS_STATUS2_OUTER_VLAN_MASK) >>7,
            "\n\tMF_MAC4         :", (data & MCPF_HMCPQ_BITS_STATUS2_MAC4           ) ? " ON":"OFF",
            "\n\tMF_MAC5         :", (data & MCPF_HMCPQ_BITS_STATUS2_MAC5           ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_hmcpq_bits_status3):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tETYPE0   :", (data & MCPF_HMCPQ_BITS_STATUS3_ETYPE0    ) ? " ON":"OFF",
            "\n\tETYPE1   :", (data & MCPF_HMCPQ_BITS_STATUS3_ETYPE1    ) ? " ON":"OFF",
            "\n\tALL_MLCST:", (data & MCPF_HMCPQ_BITS_STATUS3_ALL_MLCST ) ? " ON":"OFF",
            "\n\tARP      :", (data & MCPF_HMCPQ_BITS_STATUS3_ARP       ) ? " ON":"OFF",
            "\n\tICMPV4   :", (data & MCPF_HMCPQ_BITS_STATUS3_ICMPV4    ) ? " ON":"OFF",
            "\n\tICMPV6   :", (data & MCPF_HMCPQ_BITS_STATUS3_ICMPV6    ) ? " ON":"OFF",
            "\n\tLLDP     :", (data & MCPF_HMCPQ_BITS_STATUS3_LLDP      ) ? " ON":"OFF",
            "\n\tVNTAG0   :", (data & MCPF_HMCPQ_BITS_STATUS3_VNTAG0    ) ? " ON":"OFF",
            "\n\tVNTAG1   :", (data & MCPF_HMCPQ_BITS_STATUS3_VNTAG1    ) ? " ON":"OFF",
            "\n\tPF0_VLAN :", (data & MCPF_HMCPQ_BITS_STATUS3_PF0_VLAN  ) ? " ON":"OFF",
            "\n\tPF1_VLAN :", (data & MCPF_HMCPQ_BITS_STATUS3_PF1_VLAN  ) ? " ON":"OFF");
            break;               

    case OFFSETOF(mcp_fio_t, mcpf_nvm_command):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tRST      :", (data & MCPF_NVM_COMMAND_RST      ) ? " ON":"OFF",
            "\n\tDONE     :", (data & MCPF_NVM_COMMAND_DONE     ) ? " ON":"OFF",
            "\n\tDOIT     :", (data & MCPF_NVM_COMMAND_DOIT     ) ? " ON":"OFF",
            "\n\tWR       :", (data & MCPF_NVM_COMMAND_WR       ) ? " ON":"OFF",
            "\n\tERASE    :", (data & MCPF_NVM_COMMAND_ERASE    ) ? " ON":"OFF",
            "\n\tFIRST    :", (data & MCPF_NVM_COMMAND_FIRST    ) ? " ON":"OFF",
            "\n\tLAST     :", (data & MCPF_NVM_COMMAND_LAST     ) ? " ON":"OFF",
            "\n\tWREN     :", (data & MCPF_NVM_COMMAND_WREN     ) ? " ON":"OFF",
            "\n\tWRDI     :", (data & MCPF_NVM_COMMAND_WRDI     ) ? " ON":"OFF",
            "\n\tRD_ID    :", (data & MCPF_NVM_COMMAND_RD_ID    ) ? " ON":"OFF",
            "\n\tRD_STATUS:", (data & MCPF_NVM_COMMAND_RD_STATUS) ? " ON":"OFF",
            "\n\tMODE_256 :", (data & MCPF_NVM_COMMAND_MODE_256 ) ? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tnvm_status:", 
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_IDLE               ? "IDLE "             :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_CMD0               ? "CMD0 "             :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_CMD1               ? "CMD1 "             :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_CMD_FINISH0        ? "CMD_FINISH0 "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_CMD_FINISH1        ? "CMD_FINISH1 "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_ADDR0              ? "ADDR0  "           :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_WRITE_DATA0        ? "WRITE_DATA0 "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_WRITE_DATA1        ? "WRITE_DATA1 "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_WRITE_DATA2        ? "WRITE_DATA2 "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_DATA0         ? "READ_DATA0  "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_DATA1         ? "READ_DATA1  "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_DATA2         ? "READ_DATA2  "      :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID0  ? "READ_STATUS_RDID0" :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID1  ? "READ_STATUS_RDID1" :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID2  ? "READ_STATUS_RDID2" :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID3  ? "READ_STATUS_RDID3" :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_READ_STATUS_RDID4  ? "READ_STATUS_RDID4" :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_CHECK_BUSY0        ? "CHECK_BUSY0  "     :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_ST_WREN            ? "ST_WREN   "        :
            ((data & MCPF_NVM_STATUS_SPI_FSM_STATE) == MCPF_NVM_STATUS_SPI_FSM_STATE_SPI_WAIT               ? "WAIT      "        :"Invalid")))))))))))))))))))));
            break;
    case OFFSETOF(mcp_fio_t, mcpf_nvm_write):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tNVM_WRITE_VALUE :", 
            ((data & MCPF_NVM_WRITE_NVM_WRITE_VALUE) == MCPF_NVM_WRITE_NVM_WRITE_VALUE_BIT_BANG ?  "BIT_BANG" :
            ((data & MCPF_NVM_WRITE_NVM_WRITE_VALUE) == MCPF_NVM_WRITE_NVM_WRITE_VALUE_SI       ?  "SI      " :
            ((data & MCPF_NVM_WRITE_NVM_WRITE_VALUE) == MCPF_NVM_WRITE_NVM_WRITE_VALUE_SO       ?  "SO      " :
            ((data & MCPF_NVM_WRITE_NVM_WRITE_VALUE) == MCPF_NVM_WRITE_NVM_WRITE_VALUE_CS_B     ?  "CS_B    " :
            ((data & MCPF_NVM_WRITE_NVM_WRITE_VALUE) == MCPF_NVM_WRITE_NVM_WRITE_VALUE_SCLK     ?  "SCLK    ":"Invalid"))))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_addr):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tNVM_ADDR_VALUE :", 
            ((data & MCPF_NVM_ADDR_NVM_ADDR_VALUE) == MCPF_NVM_ADDR_NVM_ADDR_VALUE_BIT_BANG ? "BIT_BANG" :
            ((data & MCPF_NVM_ADDR_NVM_ADDR_VALUE) == MCPF_NVM_ADDR_NVM_ADDR_VALUE_SI       ? "SI      " :
            ((data & MCPF_NVM_ADDR_NVM_ADDR_VALUE) == MCPF_NVM_ADDR_NVM_ADDR_VALUE_SO       ? "SO      " :
            ((data & MCPF_NVM_ADDR_NVM_ADDR_VALUE) == MCPF_NVM_ADDR_NVM_ADDR_VALUE_CS_B     ? "CS_B    " :
            ((data & MCPF_NVM_ADDR_NVM_ADDR_VALUE) == MCPF_NVM_ADDR_NVM_ADDR_VALUE_SCLK     ? "SCLK    " : "Invalid"))))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_read):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tNVM_READ_VALUE :", 
            ((data & MCPF_NVM_READ_NVM_READ_VALUE) == MCPF_NVM_READ_NVM_READ_VALUE_BIT_BANG  ?"BIT_BANG" :
            ((data & MCPF_NVM_READ_NVM_READ_VALUE) == MCPF_NVM_READ_NVM_READ_VALUE_SI        ?"SI      " :
            ((data & MCPF_NVM_READ_NVM_READ_VALUE) == MCPF_NVM_READ_NVM_READ_VALUE_SO        ?"SO      " :
            ((data & MCPF_NVM_READ_NVM_READ_VALUE) == MCPF_NVM_READ_NVM_READ_VALUE_CS_B      ?"CS_B    " :
            ((data & MCPF_NVM_READ_NVM_READ_VALUE) == MCPF_NVM_READ_NVM_READ_VALUE_SCLK      ?"SCLK    " : "Invalid"))))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %x %s %x %s %x %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tFLASH_MODE      :", (data & MCPF_NVM_CFG1_FLASH_MODE         )? " ON":"OFF",
            "\n\tBUFFER_MODE     :", (data & MCPF_NVM_CFG1_BUFFER_MODE        )? " ON":"OFF",
            "\n\tPASS_MODE       :", (data & MCPF_NVM_CFG1_PASS_MODE          )? " ON":"OFF",
            "\n\tBITBANG_MODE    :", (data & MCPF_NVM_CFG1_BITBANG_MODE       )? " ON":"OFF",
            "\n\tSTATUS_BIT      :", (u32_t)(data & MCPF_NVM_CFG1_STATUS_BIT) >>4,
            "\n\tSPI_CLK_DIV     :", (u32_t)(data & MCPF_NVM_CFG1_SPI_CLK_DIV) >>7,
            "\n\tSEE_CLK_DIV     :", (u32_t)(data & MCPF_NVM_CFG1_SEE_CLK_DIV ) >>11,
            "\n\tSTRAP_CONTROL_0 :", (data & MCPF_NVM_CFG1_STRAP_CONTROL_0    )? " ON":"OFF",
            "\n\tPROTECT_MODE    :", (data & MCPF_NVM_CFG1_PROTECT_MODE       )? " ON":"OFF",
            "\n\tFLASH_SIZE      :", (data & MCPF_NVM_CFG1_FLASH_SIZE         )? " ON":"OFF",
            "\n\tFW_USTRAP_1     :", (data & MCPF_NVM_CFG1_FW_USTRAP_1        )? " ON":"OFF",
            "\n\tFW_USTRAP_0     :", (data & MCPF_NVM_CFG1_FW_USTRAP_0        )? " ON":"OFF",
            "\n\tFW_USTRAP_2     :", (data & MCPF_NVM_CFG1_FW_USTRAP_2        )? " ON":"OFF",
            "\n\tFW_USTRAP_3     :", (data & MCPF_NVM_CFG1_FW_USTRAP_3        )? " ON":"OFF",
            "\n\tFW_FLASH_TYPE_EN:", (data & MCPF_NVM_CFG1_FW_FLASH_TYPE_EN   )? " ON":"OFF",
            "\n\tCOMPAT_BYPASSS  :", (data & MCPF_NVM_CFG1_COMPAT_BYPASSS     )? " ON":"OFF");       
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg2):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %x %s %x",
            addr, data,
            "\n\tERASE_CMD  :", (u32_t)(data & MCPF_NVM_CFG2_ERASE_CMD),
            "\n\tSTATUS_CMD :", (u32_t)(data & MCPF_NVM_CFG2_STATUS_CMD) >>16,
            "\n\tREAD_ID    :", (u32_t)(data & MCPF_NVM_CFG2_READ_ID ) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg3):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %x %s %x",
            addr, data,
            "\n\tBUFFER_RD_CMD:", (u32_t)(data & MCPF_NVM_CFG3_BUFFER_RD_CMD),
            "\n\tWRITE_CMD    :", (u32_t)(data & MCPF_NVM_CFG3_WRITE_CMD ) >>8,
            "\n\tREAD_CMD     :", (u32_t)(data & MCPF_NVM_CFG3_READ_CMD ) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_sw_arb):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tARB_REQ_SET0:", (data & MCPF_NVM_SW_ARB_ARB_REQ_SET0                )? " ON":"OFF",
            "\n\tARB_REQ_SET1:", (data & MCPF_NVM_SW_ARB_ARB_REQ_SET1                )? " ON":"OFF",
            "\n\tARB_REQ_SET2:", (data & MCPF_NVM_SW_ARB_ARB_REQ_SET2                )? " ON":"OFF",
            "\n\tARB_REQ_SET3:", (data & MCPF_NVM_SW_ARB_ARB_REQ_SET3                )? " ON":"OFF",
            "\n\tARB_REQ_CLR0:", (data & MCPF_NVM_SW_ARB_ARB_REQ_CLR0                )? " ON":"OFF",
            "\n\tARB_REQ_CLR1:", (data & MCPF_NVM_SW_ARB_ARB_REQ_CLR1                )? " ON":"OFF",
            "\n\tARB_REQ_CLR2:", (data & MCPF_NVM_SW_ARB_ARB_REQ_CLR2                )? " ON":"OFF",
            "\n\tARB_REQ_CLR3:", (data & MCPF_NVM_SW_ARB_ARB_REQ_CLR3                )? " ON":"OFF",
            "\n\tARB_ARB0    :", (data & MCPF_NVM_SW_ARB_ARB_ARB0                    )? " ON":"OFF",
            "\n\tARB_ARB1    :", (data & MCPF_NVM_SW_ARB_ARB_ARB1                    )? " ON":"OFF",
            "\n\tARB_ARB2    :", (data & MCPF_NVM_SW_ARB_ARB_ARB2                    )? " ON":"OFF",
            "\n\tARB_ARB3    :", (data & MCPF_NVM_SW_ARB_ARB_ARB3                    )? " ON":"OFF",
            "\n\tREQ0        :", (data & MCPF_NVM_SW_ARB_REQ0                        )? " ON":"OFF",
            "\n\tREQ1        :", (data & MCPF_NVM_SW_ARB_REQ1                        )? " ON":"OFF",
            "\n\tREQ2        :", (data & MCPF_NVM_SW_ARB_REQ2                        )? " ON":"OFF",
            "\n\tREQ3        :", (data & MCPF_NVM_SW_ARB_REQ3                        )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_access_enable):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s ",
            addr, data,
            "\n\tENABLE_EN   :", (data & MCPF_NVM_ACCESS_ENABLE_EN                   )? " ON":"OFF",
            "\n\tENABLE_WR_EN:", (data & MCPF_NVM_ACCESS_ENABLE_WR_EN                )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_write1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %x",
            addr, data,
            "\n\tWREN_CMD:", (u32_t)(data & MCPF_NVM_WRITE1_WREN_CMD),
            "\n\tWRDI_CMD:", (u32_t)(data & MCPF_NVM_WRITE1_WRDI_CMD ) >>8);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_cfg4):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %x",
            addr, data,
            "\n\tFLASH_SIZE :", 
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_1MBIT          ?     "1MBIT  " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_2MBIT          ?     "2MBIT  " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_4MBIT          ?     "4MBIT  " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_8MBIT          ?     "8MBIT  " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_16MBIT         ?     "16MBIT " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_32MBIT         ?     "32MBIT " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_64MBIT         ?     "64MBIT " :
            ((data & MCPF_NVM_CFG4_FLASH_SIZE) == MCPF_NVM_CFG4_FLASH_SIZE_128MBIT        ?     "128MBIT" :"Invalid")))))))),
            "\n\tFLASH_VENDOR :", 
            ((data & MCPF_NVM_CFG4_FLASH_VENDOR) == MCPF_NVM_CFG4_FLASH_VENDOR_ST      ?       "ST   "    :
            ((data & MCPF_NVM_CFG4_FLASH_VENDOR) == MCPF_NVM_CFG4_FLASH_VENDOR_ATMEL   ?       "ATMEL"    :"Invalid")),
            "\n\tMODE_256_EMPTY_BIT_LOC :", 
            ((data & MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) == MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT8 ? "BIT8 " :
            ((data & MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) == MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT9 ? "BIT9 " :
            ((data & MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) == MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT10? "BIT10" :
            ((data & MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC) == MCPF_NVM_CFG4_MODE_256_EMPTY_BIT_LOC_BIT11? "BIT11" :"Invalid")))),
            "\n\t:", (data & MCPF_NVM_CFG4_STATUS_BIT_POLARITY           )? " ON":"OFF",
            "\n\t:", (u32_t)(data & MCPF_NVM_CFG4_RESERVED) >>7);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_nvm_reconfig):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %x %s %x %s %s ",
            addr, data,
            "\n\tORIG_STRAP_VALUE:", 
            ((data & MCPF_NVM_RECONFIG_ORIG_STRAP_VALUE) == MCPF_NVM_RECONFIG_ORIG_STRAP_VALUE_ST    ?"ST   " :
            ((data & MCPF_NVM_RECONFIG_ORIG_STRAP_VALUE) == MCPF_NVM_RECONFIG_ORIG_STRAP_VALUE_ATMEL ?"ATMEL":"Invalid")),
            "\n\tRECONFIG_STRAP_VALUE:", (u32_t)(data & MCPF_NVM_RECONFIG_RECONFIG_STRAP_VALUE ) >>4,
            "\n\tRESERVED            :", (u32_t)(data & MCPF_NVM_RECONFIG_RESERVED) >>8,
            "\n\tRECONFIG_DONE       :", (data & MCPF_NVM_RECONFIG_RECONFIG_DONE             )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_config):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %u %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tHW_ARP_ASSIGN_ADDR :", (data & MCPF_SMBUS_CONFIG_HW_ARP_ASSIGN_ADDR        )? " ON":"OFF", 
            "\n\tARP_EN0            :", (data & MCPF_SMBUS_CONFIG_ARP_EN0                   )? " ON":"OFF", 
            "\n\tARP_EN1            :", (data & MCPF_SMBUS_CONFIG_ARP_EN1                   )? " ON":"OFF", 
            "\n\tMASTER_RTRY_CNT    :", (u32_t)(data & MCPF_SMBUS_CONFIG_MASTER_RTRY_CNT ) >>16,
            "\n\tTIMESTAMP_CNT_EN   :", (data & MCPF_SMBUS_CONFIG_TIMESTAMP_CNT_EN          )? " ON":"OFF", 
            "\n\tPROMISCOUS_MODE    :", (data & MCPF_SMBUS_CONFIG_PROMISCOUS_MODE           )? " ON":"OFF", 
            "\n\tEN_NIC_SMB_ADDR_0  :", (data & MCPF_SMBUS_CONFIG_EN_NIC_SMB_ADDR_0         )? " ON":"OFF", 
            "\n\tBIT_BANG_EN        :", (data & MCPF_SMBUS_CONFIG_BIT_BANG_EN               )? " ON":"OFF", 
            "\n\tSMB_EN             :", (data & MCPF_SMBUS_CONFIG_SMB_EN                    )? " ON":"OFF", 
            "\n\tRESET              :", (data & MCPF_SMBUS_CONFIG_RESET                     )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_timing_config):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %u %s %u %s %s ",
            addr, data,
            "\n\tSMBUS_IDLE_TIME       :", (u32_t)(data & MCPF_SMBUS_TIMING_CONFIG_SMBUS_IDLE_TIME) >>8,
            "\n\tPERIODIC_SLAVE_STRETCH:", (u32_t)(data & MCPF_SMBUS_TIMING_CONFIG_PERIODIC_SLAVE_STRETCH) >>16,
            "\n\tRANDOM_SLAVE_STRETCH  :", (u32_t)(data & MCPF_SMBUS_TIMING_CONFIG_RANDOM_SLAVE_STRETCH) >>24,
            "\n\tMODE_400              :", (data & MCPF_SMBUS_TIMING_CONFIG_MODE_400 )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_address):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %s %s %02x %s %s %s %02x %s %s %s %02x %s %s ",
            addr, data,
            "\n\tNIC_SMB_ADDR0   :", (u32_t)(data & MCPF_SMBUS_ADDRESS_NIC_SMB_ADDR0),
            "\n\tEN_NIC_SMB_ADDR0:", (data & MCPF_SMBUS_ADDRESS_EN_NIC_SMB_ADDR0     )? " ON":"OFF",
            "\n\tNIC_SMB_ADDR1   :", (u32_t)(data & MCPF_SMBUS_ADDRESS_NIC_SMB_ADDR1 ) >>8,
            "\n\tEN_NIC_SMB_ADDR1:", (data & MCPF_SMBUS_ADDRESS_EN_NIC_SMB_ADDR1     )? " ON":"OFF",
            "\n\tNIC_SMB_ADDR2   :", (u32_t)(data & MCPF_SMBUS_ADDRESS_NIC_SMB_ADDR2 ) >>16,
            "\n\tEN_NIC_SMB_ADDR2:", (data & MCPF_SMBUS_ADDRESS_EN_NIC_SMB_ADDR2     )? " ON":"OFF",
            "\n\tNIC_SMB_ADDR3   :", (u32_t)(data & MCPF_SMBUS_ADDRESS_NIC_SMB_ADDR3) >>24,
            "\n\tEN_NIC_SMB_ADDR3:", (data & MCPF_SMBUS_ADDRESS_EN_NIC_SMB_ADDR3     )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_fifo_control):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %u %s %s %s %s ",
            addr, data,
            "\n\tRX_FIFO_THRESHOLD:", (u32_t)(data & MCPF_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_FIFO_THRESHOLD ) >>8,
            "\n\tRX_PKT_COUNT     :", (u32_t)(data & MCPF_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_PKT_COUNT ) >>16,
            "\n\tTX_FIFO_FLUSH    :", (data & MCPF_SMBUS_MASTER_FIFO_CONTROL_MASTER_TX_FIFO_FLUSH )? " ON":"OFF", 
            "\n\tRX_FIFO_FLUSH    :", (data & MCPF_SMBUS_MASTER_FIFO_CONTROL_MASTER_RX_FIFO_FLUSH )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_fifo_control):
        sprintf(disp_str,
            "\n(FIO%x=%08x%s %u %s %u %s %s %s %s ",
            addr, data,
            "\n\tRX_FIFO_THRESHOLD:", (u32_t)(data & MCPF_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_FIFO_THRESHOLD) >>8,
            "\n\tRX_PKT_COUNT     :", (u32_t)(data & MCPF_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_PKT_COUNT) >>16,
            "\n\tTX_FIFO_FLUSH    :", (data & MCPF_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_TX_FIFO_FLUSH)? " ON":"OFF", 
            "\n\tRX_FIFO_FLUSH    :", (data & MCPF_SMBUS_SLAVE_FIFO_CONTROL_SLAVE_RX_FIFO_FLUSH)? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_bit_bang_control):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tSMBDAT_OUT_EN:", (data & MCPF_SMBUS_BIT_BANG_CONTROL_SMBDAT_OUT_EN   )? " ON":"OFF",
            "\n\tSMBDAT_IN    :", (data & MCPF_SMBUS_BIT_BANG_CONTROL_SMBDAT_IN       )? " ON":"OFF",
            "\n\tSMBCLK_OUT_EN:", (data & MCPF_SMBUS_BIT_BANG_CONTROL_SMBCLK_OUT_EN   )? " ON":"OFF",
            "\n\tSMBCLK_IN    :", (data & MCPF_SMBUS_BIT_BANG_CONTROL_SMBCLK_IN       )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_watchdog):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x",
            addr, data,
            "\n\tWATCHDOG  :", (u32_t)(data & MCPF_SMBUS_WATCHDOG_WATCHDOG));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_heartbeat):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x ",
            addr, data,
            "\n\tHEARTBEAT:", (u32_t)(data & MCPF_SMBUS_HEARTBEAT_HEARTBEAT));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_poll_asf):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u ",
            addr, data,
            "\n\tPOLL_ASF (x5ms):", (u32_t)(data & MCPF_SMBUS_POLL_ASF_POLL_ASF  ));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_poll_legacy):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u ",
            addr, data,
            "\n\tPOLL_LEGACY (x250ms):", (u32_t)(data & MCPF_SMBUS_POLL_LEGACY_POLL_LEGACY));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_retran):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u ",
            addr, data,
            "\n\tRETRAN  (x1s):", (u32_t)(data & MCPF_SMBUS_RETRAN_RETRAN ));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_timestamp):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u ",
            addr, data,
            "\n\tTIMESTAMP (x1s):", (u32_t)(data & MCPF_SMBUS_TIMESTAMP_TIMESTAMP));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_command):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tRD_BYTE_COUNT :", (u32_t)(data & MCPF_SMBUS_MASTER_COMMAND_RD_BYTE_COUNT),
            "\n\tPEC           :", (data & MCPF_SMBUS_MASTER_COMMAND_PEC   )? " ON":"OFF", 
            "\n\tSMBUS_PROTOCOL:",     
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0000?  "0000" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0001?  "0001" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0010?  "0010" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0011?  "0011" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0100?  "0100" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0101?  "0101" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0110?  "0110" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_0111?  "0111" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1000?  "1000" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1001?  "1001" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1010?  "1010" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL) == MCPF_SMBUS_MASTER_COMMAND_SMBUS_PROTOCOL_1011?  "1011":"Invalid")))))))))))),
            "\n\tSTATUS:", 
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_000? "000" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_001? "001" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_010? "010" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_011? "011" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_100? "100" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_101? "101" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_110? "110" :
            ((data & MCPF_SMBUS_MASTER_COMMAND_STATUS) == MCPF_SMBUS_MASTER_COMMAND_STATUS_111? "111" : "Invalid")))))))),
            "\n\tABORT      :", (data & MCPF_SMBUS_MASTER_COMMAND_ABORT             )? " ON":"OFF",
            "\n\tSTART_BUSY :", (data & MCPF_SMBUS_MASTER_COMMAND_START_BUSY        )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_command):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tPEC   :", (data & MCPF_SMBUS_SLAVE_COMMAND_PEC )? " ON":"OFF",
            "\n\tSTATUS:",
            ((data & MCPF_SMBUS_SLAVE_COMMAND_STATUS) == MCPF_SMBUS_SLAVE_COMMAND_STATUS_000 ? "000" :
            ((data & MCPF_SMBUS_SLAVE_COMMAND_STATUS) == MCPF_SMBUS_SLAVE_COMMAND_STATUS_101 ? "101" :
            ((data & MCPF_SMBUS_SLAVE_COMMAND_STATUS) == MCPF_SMBUS_SLAVE_COMMAND_STATUS_111 ? "111" : "Invalid"))),
            "\n\tABORT:", (data & MCPF_SMBUS_SLAVE_COMMAND_ABORT)? " ON":"OFF",
            "\n\tSTART:", (data & MCPF_SMBUS_SLAVE_COMMAND_START)? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_event_enable):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,                                                                                                 
            "\n\tWATCHDOG_TO_EN            :", (data & MCPF_SMBUS_EVENT_ENABLE_WATCHDOG_TO_EN            )? " ON":"OFF",
            "\n\tHEARTBEAT_TO_EN           :", (data & MCPF_SMBUS_EVENT_ENABLE_HEARTBEAT_TO_EN           )? " ON":"OFF",
            "\n\tPOLL_ASF_TO_EN            :", (data & MCPF_SMBUS_EVENT_ENABLE_POLL_ASF_TO_EN            )? " ON":"OFF",
            "\n\tPOLL_LEGACY_TO_EN         :", (data & MCPF_SMBUS_EVENT_ENABLE_POLL_LEGACY_TO_EN         )? " ON":"OFF",
            "\n\tRETRANSMIT_TO_EN          :", (data & MCPF_SMBUS_EVENT_ENABLE_RETRANSMIT_TO_EN          )? " ON":"OFF",
            "\n\tSLAVE_ARP_EVENT_EN        :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_ARP_EVENT_EN        )? " ON":"OFF",
            "\n\tSLAVE_RD_EVENT_EN         :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_RD_EVENT_EN         )? " ON":"OFF",
            "\n\tSLAVE_TX_UNDERRUN_EN      :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_TX_UNDERRUN_EN      )? " ON":"OFF",
            "\n\tSLAVE_START_BUSY_EN       :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_START_BUSY_EN       )? " ON":"OFF",
            "\n\tSLAVE_RX_EVENT_EN         :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_RX_EVENT_EN         )? " ON":"OFF",
            "\n\tSLAVE_RX_THRESHOLD_HIT_EN :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_RX_THRESHOLD_HIT_EN )? " ON":"OFF",
            "\n\tSLAVE_RX_FIFO_FULL_EN     :", (data & MCPF_SMBUS_EVENT_ENABLE_SLAVE_RX_FIFO_FULL_EN     )? " ON":"OFF",
            "\n\tMASTER_TX_UNDERRUN_EN     :", (data & MCPF_SMBUS_EVENT_ENABLE_MASTER_TX_UNDERRUN_EN     )? " ON":"OFF",
            "\n\tMASTER_START_BUSY_EN      :", (data & MCPF_SMBUS_EVENT_ENABLE_MASTER_START_BUSY_EN      )? " ON":"OFF",
            "\n\tMASTER_RX_EVENT_EN        :", (data & MCPF_SMBUS_EVENT_ENABLE_MASTER_RX_EVENT_EN        )? " ON":"OFF",
            "\n\tMASTER_RX_THRESHOLD_HIT_EN:", (data & MCPF_SMBUS_EVENT_ENABLE_MASTER_RX_THRESHOLD_HIT_EN)? " ON":"OFF",
            "\n\tMASTER_RX_FIFO_FULL_EN    :", (data & MCPF_SMBUS_EVENT_ENABLE_MASTER_RX_FIFO_FULL_EN    )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_event_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tWATCHDOG_TO             :", (data & MCPF_SMBUS_EVENT_STATUS_WATCHDOG_TO            )? " ON":"OFF",                                                                                                                                                                             
            "\n\tHEARTBEAT_TO            :", (data & MCPF_SMBUS_EVENT_STATUS_HEARTBEAT_TO           )? " ON":"OFF",                                                                                                                                                                             
            "\n\tPOLL_ASF_TO             :", (data & MCPF_SMBUS_EVENT_STATUS_POLL_ASF_TO            )? " ON":"OFF",                                                                                                                                                                             
            "\n\tPOLL_LEGACY_TO          :", (data & MCPF_SMBUS_EVENT_STATUS_POLL_LEGACY_TO         )? " ON":"OFF",                                                                                                                                                                             
            "\n\tRETRANSMIT_TO           :", (data & MCPF_SMBUS_EVENT_STATUS_RETRANSMIT_TO          )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_ARP_EVENT         :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_ARP_EVENT        )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_RD_EVENT          :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_RD_EVENT         )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_TX_UNDERRUN       :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_TX_UNDERRUN      )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_START_BUSY        :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_START_BUSY       )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_RX_EVENT          :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_RX_EVENT         )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_RX_THRESHOLD_HIT  :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_RX_THRESHOLD_HIT )? " ON":"OFF",                                                                                                                                                                             
            "\n\tSLAVE_RX_FIFO_FULL      :", (data & MCPF_SMBUS_EVENT_STATUS_SLAVE_RX_FIFO_FULL     )? " ON":"OFF",                                                                                                                                                                             
            "\n\tMASTER_TX_UNDERRUN      :", (data & MCPF_SMBUS_EVENT_STATUS_MASTER_TX_UNDERRUN     )? " ON":"OFF",                                                                                                                                                                             
            "\n\tMASTER_START_BUSY       :", (data & MCPF_SMBUS_EVENT_STATUS_MASTER_START_BUSY      )? " ON":"OFF",                                                                                                                                                                             
            "\n\tMASTER_RX_EVENT         :", (data & MCPF_SMBUS_EVENT_STATUS_MASTER_RX_EVENT        )? " ON":"OFF",                                                                                                                                                                             
            "\n\tMASTER_RX_THRESHOLD_HIT :", (data & MCPF_SMBUS_EVENT_STATUS_MASTER_RX_THRESHOLD_HIT)? " ON":"OFF",                                                                                                                                                                             
            "\n\tMASTER_RX_FIFO_FULL     :", (data & MCPF_SMBUS_EVENT_STATUS_MASTER_RX_FIFO_FULL    )? " ON":"OFF");                                                                                                                                                                            
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_data_write):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s ",
            addr, data,
            "\n\tMASTER_SMBUS_WR_DATA   :",   (u32_t)(data & MCPF_SMBUS_MASTER_DATA_WRITE_MASTER_SMBUS_WR_DATA),
            "\n\tWR_STATUS              :",   (data & MCPF_SMBUS_MASTER_DATA_WRITE_WR_STATUS      )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_master_data_read):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s %s %s ",
            addr, data,
            "\n\tMASTER_SMBUS_RD_DATA  :", (u32_t)(data & MCPF_SMBUS_MASTER_DATA_READ_MASTER_SMBUS_RD_DATA),
            "\n\tPEC_ERR         :", (data & MCPF_SMBUS_MASTER_DATA_READ_PEC_ERR         )? " ON":"OFF",
            "\n\tRD_STATUS       :", 
            ((data & MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS) == MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS_00? "RD_STATUS_00  " :
            ((data & MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS) == MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS_01? "RD_STATUS_01  " :
            ((data & MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS) == MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS_10? "RD_STATUS_10  " :
            ((data & MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS) == MCPF_SMBUS_MASTER_DATA_READ_RD_STATUS_11? "RD_STATUS_11  ":"Invalid")))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_data_write):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s ",
            addr, data,
            "\n\tSLAVE_SMBUS_WR_DATA  :", (u32_t)(data & MCPF_SMBUS_SLAVE_DATA_WRITE_SLAVE_SMBUS_WR_DATA),
            "\n\tWR_STATUS       :", 
            ((data & MCPF_SMBUS_SLAVE_DATA_WRITE_WR_STATUS) == MCPF_SMBUS_SLAVE_DATA_WRITE_WR_STATUS_0? "WR_STATUS_0" :
            ((data & MCPF_SMBUS_SLAVE_DATA_WRITE_WR_STATUS) == MCPF_SMBUS_SLAVE_DATA_WRITE_WR_STATUS_1? "WR_STATUS_1":"Invalid")));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_slave_data_read):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s %s %s ",
            addr, data,
            "\n\tSLAVE_SMBUS_RD_DATA  :", (u32_t)(data & MCPF_SMBUS_SLAVE_DATA_READ_SLAVE_SMBUS_RD_DATA),
            "\n\tERR_STATUS       :", 
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS_00? "00" :
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS_01? "01" :
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS_10? "10" :
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_ERR_STATUS_11? "11" : "Invalid")))),
            "\n\tRD_STATUS        :", 
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS_00? "00" :
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS_01? "01" :
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS_10? "10" :
            ((data & MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS) == MCPF_SMBUS_SLAVE_DATA_READ_RD_STATUS_11? "11" : "Invalid")))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_arp_state):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tAV_FLAG0 :", (data & MCPF_SMBUS_ARP_STATE_AV_FLAG0 )? " ON":"OFF",
            "\n\tAR_FLAG0 :", (data & MCPF_SMBUS_ARP_STATE_AR_FLAG0 )? " ON":"OFF",
            "\n\tAV_FLAG1 :", (data & MCPF_SMBUS_ARP_STATE_AV_FLAG1 )? " ON":"OFF",
            "\n\tAR_FLAG1 :", (data & MCPF_SMBUS_ARP_STATE_AR_FLAG1 )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_3):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_12 :", (u32_t)(data & MCPF_SMBUS_UDID0_3_BYTE_12 ),              
            "\n\tBYTE_13 :", (u32_t)(data & MCPF_SMBUS_UDID0_3_BYTE_13 ) >>8,
            "\n\tBYTE_14 :", (u32_t)(data & MCPF_SMBUS_UDID0_3_BYTE_14 ) >>16,
            "\n\tBYTE_15 :", (u32_t)(data & MCPF_SMBUS_UDID0_3_BYTE_15 ) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_2):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x  ",
            addr, data,
            "\n\tBYTE_8  :", (u32_t)(data & MCPF_SMBUS_UDID0_2_BYTE_8 ),              
            "\n\tBYTE_9  :", (u32_t)(data & MCPF_SMBUS_UDID0_2_BYTE_9 ) >>8,
            "\n\tBYTE_10 :", (u32_t)(data & MCPF_SMBUS_UDID0_2_BYTE_10) >>16,
            "\n\tBYTE_11 :", (u32_t)(data & MCPF_SMBUS_UDID0_2_BYTE_11) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_4  :", (u32_t)(data & MCPF_SMBUS_UDID0_1_BYTE_4),                   
            "\n\tBYTE_5  :", (u32_t)(data & MCPF_SMBUS_UDID0_1_BYTE_5) >>8,
            "\n\tBYTE_6  :", (u32_t)(data & MCPF_SMBUS_UDID0_1_BYTE_6) >>16,
            "\n\tBYTE_7  :", (u32_t)(data & MCPF_SMBUS_UDID0_1_BYTE_7) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid0_0):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_0  :", (u32_t)(data & MCPF_SMBUS_UDID0_0_BYTE_0),                  
            "\n\tBYTE_1  :", (u32_t)(data & MCPF_SMBUS_UDID0_0_BYTE_1) >>8,
            "\n\tBYTE_2  :", (u32_t)(data & MCPF_SMBUS_UDID0_0_BYTE_2) >>16,
            "\n\tBYTE_3  :", (u32_t)(data & MCPF_SMBUS_UDID0_0_BYTE_3) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_3):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_12 :", (u32_t)(data & MCPF_SMBUS_UDID1_3_BYTE_12),             
            "\n\tBYTE_13 :", (u32_t)(data & MCPF_SMBUS_UDID1_3_BYTE_13) >>8,
            "\n\tBYTE_14 :", (u32_t)(data & MCPF_SMBUS_UDID1_3_BYTE_14) >>16,
            "\n\tBYTE_15 :", (u32_t)(data & MCPF_SMBUS_UDID1_3_BYTE_15) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_2):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_8  :", (u32_t)(data & MCPF_SMBUS_UDID1_2_BYTE_8 ),         
            "\n\tBYTE_9  :", (u32_t)(data & MCPF_SMBUS_UDID1_2_BYTE_9 ) >>8,
            "\n\tBYTE_10 :", (u32_t)(data & MCPF_SMBUS_UDID1_2_BYTE_10) >>16,
            "\n\tBYTE_11 :", (u32_t)(data & MCPF_SMBUS_UDID1_2_BYTE_11) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_4  :", (u32_t)(data & MCPF_SMBUS_UDID1_1_BYTE_4),         
            "\n\tBYTE_5  :", (u32_t)(data & MCPF_SMBUS_UDID1_1_BYTE_5) >>8,
            "\n\tBYTE_6  :", (u32_t)(data & MCPF_SMBUS_UDID1_1_BYTE_6) >>16,
            "\n\tBYTE_7  :", (u32_t)(data & MCPF_SMBUS_UDID1_1_BYTE_7) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_smbus_udid1_0):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %02x %s %02x %s %02x %s %02x ",
            addr, data,
            "\n\tBYTE_0  :", (u32_t)(data & MCPF_SMBUS_UDID1_0_BYTE_0),         
            "\n\tBYTE_1  :", (u32_t)(data & MCPF_SMBUS_UDID1_0_BYTE_1) >>8,
            "\n\tBYTE_2  :", (u32_t)(data & MCPF_SMBUS_UDID1_0_BYTE_2) >>16,
            "\n\tBYTE_3  :", (u32_t)(data & MCPF_SMBUS_UDID1_0_BYTE_3) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_cmd):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %u %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tEGRESS_FIFO_ENABLED   :", (data & MCPF_UMP_CMD_EGRESS_FIFO_ENABLED  )? " ON":"OFF", 
            "\n\tINGRESS_FIFO_ENABLED  :", (data & MCPF_UMP_CMD_INGRESS_FIFO_ENABLED )? " ON":"OFF", 
            "\n\tFC_EN                 :", (data & MCPF_UMP_CMD_FC_EN                )? " ON":"OFF", 
            "\n\tMAC_LOOPBACK          :", (data & MCPF_UMP_CMD_MAC_LOOPBACK         )? " ON":"OFF", 
            "\n\tEGRESS_MAC_DISABLE    :", (data & MCPF_UMP_CMD_EGRESS_MAC_DISABLE   )? " ON":"OFF", 
            "\n\tINGRESS_MAC_DISABLE   :", (data & MCPF_UMP_CMD_INGRESS_MAC_DISABLE  )? " ON":"OFF", 
            "\n\tINGRESS_DRIVE         :", (data & MCPF_UMP_CMD_INGRESS_DRIVE        )? " ON":"OFF", 
            "\n\tSW_PAUSE              :", (data & MCPF_UMP_CMD_SW_PAUSE             )? " ON":"OFF", 
            "\n\tAUTO_DRIVE            :", (data & MCPF_UMP_CMD_AUTO_DRIVE           )? " ON":"OFF", 
            "\n\tINGRESS_RESET         :", (data & MCPF_UMP_CMD_INGRESS_RESET        )? " ON":"OFF", 
            "\n\tNO_PLUS_TWO           :", (data & MCPF_UMP_CMD_NO_PLUS_TWO          )? " ON":"OFF", 
            "\n\tEGRESS_PKT_FLUSH      :", (data & MCPF_UMP_CMD_EGRESS_PKT_FLUSH     )? " ON":"OFF", 
            "\n\tCMD_IPG  (in nibbles) :", (u32_t)(data & MCPF_UMP_CMD_CMD_IPG) >>17,
            "\n\tEGRESS_FIO_RESET      :", (data & MCPF_UMP_CMD_EGRESS_FIO_RESET     )? " ON":"OFF", 
            "\n\tINGRESS_FIO_RESET     :", (data & MCPF_UMP_CMD_INGRESS_FIO_RESET    )? " ON":"OFF", 
            "\n\tEGRESS_MAC_RESET      :", (data & MCPF_UMP_CMD_EGRESS_MAC_RESET     )? " ON":"OFF", 
            "\n\tINGRESS_MAC_RESET     :", (data & MCPF_UMP_CMD_INGRESS_MAC_RESET    )? " ON":"OFF"); 
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_config):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %u %s %s ",
            addr, data,
            "\n\tRMII_MODE         :", (data & MCPF_UMP_CONFIG_RMII_MODE                   )? " ON":"OFF",
            "\n\tRVMII_MODE        :", (data & MCPF_UMP_CONFIG_RVMII_MODE                  )? " ON":"OFF",
            "\n\tINGRESS_MODE      :", (data & MCPF_UMP_CONFIG_INGRESS_MODE                )? " ON":"OFF",
            "\n\tINGRESS_WORD_ACCM :", (u32_t)(data & MCPF_UMP_CONFIG_INGRESS_WORD_ACCM ) >>8,
            "\n\tOLD_BCNT_RDY      :", (data & MCPF_UMP_CONFIG_OLD_BCNT_RDY                )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_fc_trip):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %04x %s %04x ",
            addr, data,
            "\n\tXON_TRIP  :", (u32_t)(data & MCPF_UMP_FC_TRIP_XON_TRIP ),
            "\n\tXOFF_TRIP :", (u32_t)(data & MCPF_UMP_FC_TRIP_XOFF_TRIP) >>16);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_frm_rd_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %u %s %s ",
            addr, data,
            "\n\tNEW_FRM       :", (data & MCPF_UMP_EGRESS_FRM_RD_STATUS_NEW_FRM       )? " ON":"OFF",
            "\n\tFRM_IN_PRO    :", (data & MCPF_UMP_EGRESS_FRM_RD_STATUS_FRM_IN_PRO    )? " ON":"OFF",
            "\n\tFIFO_EMPTY    :", (data & MCPF_UMP_EGRESS_FRM_RD_STATUS_FIFO_EMPTY    )? " ON":"OFF",
            "\n\tBCNT          :", (u32_t)(data & MCPF_UMP_EGRESS_FRM_RD_STATUS_BCNT) >>3,
            "\n\tEGRESS_FIFO_STATE  :", 
            ((data & MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) == MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_IDLE        ?"IDLE  "         :
            ((data & MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) == MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_READY       ?"READY  "        :
            ((data & MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) == MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_BUSY        ?"BUSY  "         :
            ((data & MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) == MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_EXTRA_RD    ?"EXTRA_RD  "     :
            ((data & MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE) == MCPF_UMP_EGRESS_FRM_RD_STATUS_EGRESS_FIFO_STATE_LATCH_IP_HDR?"LATCH_IP_HDR  " :"Invalid"))))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_frm_wr_ctl):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %u %s %s ",
            addr, data,
            "\n\tNEW_FRM  :", (data & MCPF_UMP_INGRESS_FRM_WR_CTL_NEW_FRM  )? " ON":"OFF",
            "\n\tFIFO_RDY :", (data & MCPF_UMP_INGRESS_FRM_WR_CTL_FIFO_RDY )? " ON":"OFF",
            "\n\tBCNT_RDY :", (data & MCPF_UMP_INGRESS_FRM_WR_CTL_BCNT_RDY )? " ON":"OFF",
            "\n\tBCNT     :", (u32_t)(data & MCPF_UMP_INGRESS_FRM_WR_CTL_BCNT ) >>3,
            "\n\tINGRESS_FIFO_STATE  :", 
            ((data & MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) == MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_IDLE    ?"IDLE  "   :
            ((data & MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) == MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_WAIT    ?"WAIT  "   :
            ((data & MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) == MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_BUSY    ?"BUSY  "   :
            ((data & MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE) == MCPF_UMP_INGRESS_FRM_WR_CTL_INGRESS_FIFO_STATE_EXTRA_WR?"EXTRA_WR  ":"Invalid")))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_fifo_remaining_words):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %s %s %s %s %x %s %s %s %s ",
            addr, data,
            "\n\tEGRESS_FIFO_DEPTH  :", (u32_t)(data & MCPF_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_DEPTH),
            "\n\tEGRESS_FIFO_UNDERFLOW  :", (data & MCPF_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_UNDERFLOW  )? " ON":"OFF", 
            "\n\tEGRESS_FIFO_OVERFLOW  :", (data & MCPF_UMP_FIFO_REMAINING_WORDS_EGRESS_FIFO_OVERFLOW  )? " ON":"OFF", 
            "\n\tINGRESS_FIFO_DEPTH  :", (u32_t)(data & MCPF_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_DEPTH ) >>16,
            "\n\tINGRESS_FIFO_UNDERFLOW  :", (data & MCPF_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_UNDERFLOW  )? " ON":"OFF", 
            "\n\tINGRESS_FIFO_OVERFLOW  :", (data & MCPF_UMP_FIFO_REMAINING_WORDS_INGRESS_FIFO_OVERFLOW  )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_fifo_ptrs):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %04x %s %s %s %04x %s %s ",
            addr, data,
            "\n\tEGRESS_FIFO_RD_PTR:", (u32_t)(data & MCPF_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_RD_PTR ),
            "\n\tUPDATE_RDPTR      :", (data & MCPF_UMP_EGRESS_FIFO_PTRS_UPDATE_RDPTR      )? " ON":"OFF", 
            "\n\tEGRESS_FIFO_WR_PTR:", (u32_t)(data & MCPF_UMP_EGRESS_FIFO_PTRS_EGRESS_FIFO_WR_PTR ) >>16,
            "\n\tUPDATE_WRPTR      :", (data & MCPF_UMP_EGRESS_FIFO_PTRS_UPDATE_WRPTR      )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_fifo_ptrs):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %04x %s %s %s %04x %s %s ",
            addr, data,
            "\n\tINGRESS_FIFO_RD_PTR  :", (u32_t)(data & MCPF_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_RD_PTR),
            "\n\tUPDATE_RDPTR     :", (data & MCPF_UMP_INGRESS_FIFO_PTRS_UPDATE_RDPTR     )? " ON":"OFF",
            "\n\tINGRESS_FIFO_WR_PTR  :", (u32_t)(data & MCPF_UMP_INGRESS_FIFO_PTRS_INGRESS_FIFO_WR_PTR) >>16,
            "\n\tUPDATE_WRPTR     :", (data & MCPF_UMP_INGRESS_FIFO_PTRS_UPDATE_WRPTR     )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_packet_sa_0):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %08x ",
            addr, data,
            "\n\tUMP_EGRESS_PACKET_SA_0_EGRESS_SA  :", (u32_t)(data & MCPF_UMP_EGRESS_PACKET_SA_0_EGRESS_SA));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_packet_sa_1):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %08x ",
            addr, data,
            "\n\tUMP_EGRESS_PACKET_SA_1_EGRESS_SA  :", (u32_t)(data & MCPF_UMP_EGRESS_PACKET_SA_1_EGRESS_SA));      
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_burst_command):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %x %s %x %s %x",
            addr, data,
            "\n\tINGRESS_DMA_START  :", (data & MCPF_UMP_INGRESS_BURST_COMMAND_INGRESS_DMA_START  )? " ON":"OFF",
            "\n\tINGRESS_PORT       :", (data & MCPF_UMP_INGRESS_BURST_COMMAND_INGRESS_PORT  )? " ON":"OFF", 
            "\n\tDMA_LENGTH         :", (u32_t)(data & MCPF_UMP_INGRESS_BURST_COMMAND_DMA_LENGTH ) >>2,
            "\n\tINGRESS_PORT_EXT   :", (u32_t)(data & MCPF_UMP_INGRESS_BURST_COMMAND_INGRESS_PORT_EXT) >>13,
            "\n\tRBUF_OFFSET        :", (u32_t)(data & MCPF_UMP_INGRESS_BURST_COMMAND_RBUF_OFFSET ) >>16);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_rbuf_cluster):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x",
            addr, data,
            "\n\tUMP_INGRESS_RBUF_CLUSTER_RBUF_CLUSTER  :", (u32_t)(data & MCPF_UMP_INGRESS_RBUF_CLUSTER_RBUF_CLUSTER));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_vlan):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s %s %s ",
            addr, data,
            "\n\tINGRESS_VLAN_TAG      :", (u32_t)(data & MCPF_UMP_INGRESS_VLAN_INGRESS_VLAN_TAG),
            "\n\tVLAN_INS              :", (data & MCPF_UMP_INGRESS_VLAN_VLAN_INS )? " ON":"OFF", 
            "\n\tVLAN_DEL              :", (data & MCPF_UMP_INGRESS_VLAN_VLAN_DEL )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_burst_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tRESULT        :", 
            ((data & MCPF_UMP_INGRESS_BURST_STATUS_RESULT) == MCPF_UMP_INGRESS_BURST_STATUS_RESULT_BUSY?"BUSY " :
            ((data & MCPF_UMP_INGRESS_BURST_STATUS_RESULT) == MCPF_UMP_INGRESS_BURST_STATUS_RESULT_DONE?"DONE " :
            ((data & MCPF_UMP_INGRESS_BURST_STATUS_RESULT) == MCPF_UMP_INGRESS_BURST_STATUS_RESULT_ERR ?"ERR  " :
            ((data & MCPF_UMP_INGRESS_BURST_STATUS_RESULT) == MCPF_UMP_INGRESS_BURST_STATUS_RESULT_ERR1?"ERR1 ":"Invalid")))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_burst_command):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %x %s %s %s %x",
            addr, data,
            "\n\tEGRESS_DMA_START  :", (data & MCPF_UMP_EGRESS_BURST_COMMAND_EGRESS_DMA_START  )? " ON":"OFF", 
            "\n\tEGRESS_PORT       :", (data & MCPF_UMP_EGRESS_BURST_COMMAND_EGRESS_PORT   )? " ON":"OFF",
            "\n\tDMA_LENGTH        :", (u32_t)(data & MCPF_UMP_EGRESS_BURST_COMMAND_DMA_LENGTH) >>2,
            "\n\tEGRESS_PORT_EXT   :", (data & MCPF_UMP_EGRESS_BURST_COMMAND_EGRESS_PORT_EXT  )? " ON":"OFF", 
            "\n\tTPBUF_OFFSET      :", (u32_t)(data & MCPF_UMP_EGRESS_BURST_COMMAND_TPBUF_OFFSET ) >>16);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_vlan):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s %s %s ",
            addr, data,
            "\n\tEGRESS_VLAN_TAG :", (u32_t)(data & MCPF_UMP_EGRESS_VLAN_EGRESS_VLAN_TAG),
            "\n\tVLAN_INS        :", (data & MCPF_UMP_EGRESS_VLAN_VLAN_INS               )? " ON":"OFF",
            "\n\tVLAN_DEL        :", (data & MCPF_UMP_EGRESS_VLAN_VLAN_DEL               )? " ON":"OFF");
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_burst_status):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s ",
            addr, data,
            "\n\tRESULT         :",
            ((data & MCPF_UMP_EGRESS_BURST_STATUS_RESULT) == MCPF_UMP_EGRESS_BURST_STATUS_RESULT_BUSY?"BUSY" :
            ((data & MCPF_UMP_EGRESS_BURST_STATUS_RESULT) == MCPF_UMP_EGRESS_BURST_STATUS_RESULT_DONE?"DONE" :
            ((data & MCPF_UMP_EGRESS_BURST_STATUS_RESULT) == MCPF_UMP_EGRESS_BURST_STATUS_RESULT_ERR0?"ERR0" :
            ((data & MCPF_UMP_EGRESS_BURST_STATUS_RESULT) == MCPF_UMP_EGRESS_BURST_STATUS_RESULT_RSVD?"RSVD":"Invalid")))));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_statistic):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %u %s %u ",
            addr, data,
            "\n\tGOOD_CNT   :", (u32_t)(data & MCPF_UMP_EGRESS_STATISTIC_EGRESS_GOOD_CNT),
            "\n\tERROR_CNT  :", (u32_t)(data & MCPF_UMP_EGRESS_STATISTIC_EGRESS_ERROR_CNT ) >>16,
            "\n\tDROP_CNT   :", (u32_t)(data & MCPF_UMP_EGRESS_STATISTIC_EGRESS_DROP_CNT) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_statistic):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u",
            addr, data,
            "\n\tUMP_INGRESS_STATISTIC_INGRESS_PKT_CNT  :",  (u32_t)(data & MCPF_UMP_INGRESS_STATISTIC_INGRESS_PKT_CNT));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_arb_cmd):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %x %s %s %s %s %s %s %s %s %s %x %s %s %s %s %s %u ",
            addr, data,
            "\n\tID              :", (u32_t)(data & MCPF_UMP_ARB_CMD_UMP_ID),
            "\n\tARB_DISABLE     :", (data & MCPF_UMP_ARB_CMD_UMP_ARB_DISABLE    )? " ON":"OFF", 
            "\n\tARB_START       :", (data & MCPF_UMP_ARB_CMD_UMP_ARB_START      )? " ON":"OFF", 
            "\n\tARB_BYPASS      :", (data & MCPF_UMP_ARB_CMD_UMP_ARB_BYPASS     )? " ON":"OFF", 
            "\n\tARB_AUTOBYPASS  :", (data & MCPF_UMP_ARB_CMD_UMP_ARB_AUTOBYPASS )? " ON":"OFF", 
            "\n\tARB_TOKEN_IPG   :", (u32_t)(data & MCPF_UMP_ARB_CMD_UMP_ARB_TOKEN_IPG ) >>8,
            "\n\tARB_TOKEN_VALID :", (data & MCPF_UMP_ARB_CMD_UMP_ARB_TOKEN_VALID )? " ON":"OFF", 
            "\n\tARB_FC_DISABLE  :", (data & MCPF_UMP_ARB_CMD_UMP_ARB_FC_DISABLE  )? " ON":"OFF", 
            "\n\tARB_TIMEOUT(clk):", (u32_t)(data & MCPF_UMP_ARB_CMD_UMP_ARB_TIMEOUT) >>16);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_frame_count):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %u ",
            addr, data,
            "\n\tEGRESS_FRAME_COUNT :",  (u32_t)(data & MCPF_UMP_FRAME_COUNT_EGRESS_FRAME_COUNT),
            "\n\tINRESS_FRAME_COUNT :",  (u32_t)(data & MCPF_UMP_FRAME_COUNT_INRESS_FRAME_COUNT) >>16);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_statistic_ac):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u %s %u %s %u",
            addr, data,
            "\n\tGOOD_CNT  :",    (u32_t)(data & MCPF_UMP_EGRESS_STATISTIC_AC_EGRESS_GOOD_CNT ),
            "\n\tERROR_CNT :",    (u32_t)(data & MCPF_UMP_EGRESS_STATISTIC_AC_EGRESS_ERROR_CNT ) >>16,
            "\n\tDROP_CNT  :",    (u32_t)(data & MCPF_UMP_EGRESS_STATISTIC_AC_EGRESS_DROP_CNT) >>24);
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_statistic_ac):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %u ",
            addr, data,
            "\n\tINGRESS_PKT_CNT  :", (u32_t)(data & MCPF_UMP_INGRESS_STATISTIC_AC_INGRESS_PKT_CNT));
            break;               
    case OFFSETOF(mcp_fio_t, mcpf_ump_event):
        sprintf(disp_str,
            "\n(FIO%x=%08x)%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s ",
            addr, data,
            "\n\tINGRESS_RDY_EVENT            :",  (data & MCPF_UMP_EVENT_INGRESS_RDY_EVENT            )? " ON":"OFF", 
            "\n\tEGRESS_RDY_EVENT             :",  (data & MCPF_UMP_EVENT_EGRESS_RDY_EVENT             )? " ON":"OFF", 
            "\n\tINGRESSBURST_DONE_EVENT      :",  (data & MCPF_UMP_EVENT_INGRESSBURST_DONE_EVENT      )? " ON":"OFF", 
            "\n\tEGRESSBURST_DONE_EVENT       :",  (data & MCPF_UMP_EVENT_EGRESSBURST_DONE_EVENT       )? " ON":"OFF", 
            "\n\tEGRESS_FRAME_DROP_EVENT      :",  (data & MCPF_UMP_EVENT_EGRESS_FRAME_DROP_EVENT      )? " ON":"OFF", 
            "\n\tINGRESS_RDY_EVENT_EN         :",  (data & MCPF_UMP_EVENT_INGRESS_RDY_EVENT_EN         )? " ON":"OFF", 
            "\n\tEGRESS_RDY_EVENT_EN          :",  (data & MCPF_UMP_EVENT_EGRESS_RDY_EVENT_EN          )? " ON":"OFF", 
            "\n\tINGRESSBURST_DONE_EVENT_EN   :",  (data & MCPF_UMP_EVENT_INGRESSBURST_DONE_EVENT_EN   )? " ON":"OFF", 
            "\n\tEGRESSBURST_DONE_EVENT_EN    :",  (data & MCPF_UMP_EVENT_EGRESSBURST_DONE_EVENT_EN    )? " ON":"OFF", 
            "\n\tEGRESS_FRAME_DROP_EVENT_EN   :",  (data & MCPF_UMP_EVENT_EGRESS_FRAME_DROP_EVENT_EN   )? " ON":"OFF");
            break;
    case OFFSETOF(mcp_fio_t, mcpf_ump_ingress_frm_wr_data):
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_frame_type):
    case OFFSETOF(mcp_fio_t, mcpf_ump_egress_frm_rd_data):
    default:
        sprintf(disp_str, 
            "\n(FIO%0x=%08x)", addr, data);
        break;
    }
    Tcl_AppendStringsToObj( rc_ptr, disp_str, NULL);
}// DecodeMcpFioRegDetail        
